/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/**
 * The Annotation class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Annotation.hpp 560 2008-11-09 11:12:14Z davidfange $
 */
#ifndef MESORD_ANNOTATION_HPP
#define MESORD_ANNOTATION_HPP 1
  /* For compilation using the Intel C++ compiler, turn off the "value
   * copied to temporary, reference to temporary used" and "operands
   * are evaluated in unspecified order" warnings.
   */
#ifdef __INTEL_COMPILER
#  if  __INTEL_COMPILER >= 800
#    pragma warning(push)
#    pragma warning(disable:383)
#    pragma warning(disable:981)
#  endif
#endif

#include <iostream>
#include <map>
#include <stdexcept>
#include <string>
#include <vector>

#include <sbml/SBMLTypes.h>

#include "mesord.hpp"
#include "definition.hpp"



namespace MesoRD
{

    /**
     * @brief Class for handling annotations that may be present in
     * SBML model definitions.
     *
     * Annotations are used in SBML to allow users to attach
     * arbitrary, application-specific information to models.
     * Annotations are formally unconstrained, but should not be
     * misused; in particular, any information that is essential to a
     * model should not be stored in an annotation.  This class
     * provides functions to deal with SBML annotations.  The primary
     * concern is of course to extract the information stored in
     * annotations in a form that is useful for other components of
     * MesoRD.
     *
     * Annotations are identified by a name.  The root node of an
     * annotation is defined in the SBML namespace, so that
     * annotations do not clash with any other SBML objects.  In the
     * context of MesoRD, the root annotation name is treated as a
     * constant.  The namespace inside an annotation is separate and
     * application-specific.  Any quantities defined in namespaces
     * which MesoRD knows nothing about are silently ignored.
     *
     * The structure of annotations in MesoRD is tree-like: any
     * annotation may have children, which are themselves instances of
     * the Annotation class.  In addition to children, annotations may
     * have attributes.  An attribute is identified by a namespaced
     * name, and associates a value to that name.  Since XML, which is
     * used to define an annotation, can be represented by simple
     * text, both name and value can be represented by C++ string
     * objects.  Hence, the attribute-management in the Annotation
     * class is implemented as an STL map, mapping a string object to
     * another string object.
     *
     * The code in the Annotation class relies on either expat or
     * Xerces-C++ for the actual XML parsing.  If both are present,
     * expat will be used.  For the Xerces-C++, a SAX reader/parser is
     * used.  Note that one may experience strange problems if one
     * uses an SBML library compiled with one XML engine together with
     * MesoRD compiled with another XML engine.
     */
  class Annotation
  {
  private:
    std::map<std::string, std::string> attributes;
    std::vector<const Annotation*>     children;
    std::string                        name;

  public:

      /**
       * @brief Class for reporting non-existent attributes.
       *
       * If the value of a non-existing attribute is requested, this
       * exception will be thrown.  The seriousness of this kind of
       * error is context-dependant: some attributes may be absolutely
       * necessary, while others are optional.
       */
    class AttributeNotFound : public std::runtime_error
    {
    private:
      const Annotation* annotation;
      const char*       name;

    public:

        /**
         * @brief The AttributeNotFound value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param annotation The annotation in which the attribute was
         *                   not found.
         * @param name       The name of the attribute which was not
         *                   found.
         */
      inline AttributeNotFound(const Annotation* annotation, const char* name)
        : std::runtime_error("AttributeNotFound"),
          annotation(annotation), name(name)
        {}


        /**
         * @brief Get the annotation in which the attribute was not
         * found.
         *
         * This function returns a pointer to the annotation in which
         * the requested attribute was not found.  This may be the
         * <code>null</code> pointer
         *
         * @return A pointer to the annotation with the missing
         *         attribute.
         */
      inline const Annotation*
      getAnnotation() const
        {
          return annotation;
        }


        /**
         * @brief Get the name of the attribute which was not found.
         *
         * This function returns a pointer to the name of the
         * attribute which was not found.  This may be the
         * <code>null</code> pointer.
         *
         * @return A pointer to the name of the missing attribute.
         */
      inline const char*
      getName() const
        {
          return name;
        }
    };


      /**
       * @brief Class for reporting errors during XML engine
       * initialisation.
       *
       * The initialisation of the MesoRD Annotation class cannot
       * really fail.  However, initialisation of the backing XML
       * parsers can.  Should there be any problem with the underlying
       * XML engine, this kind of exception is thrown.  In general it
       * means that something which MesoRD does not control has
       * failed. 
       * !!FIXME Pending deletion !!
       */
    class InitialisationError : public std::runtime_error
    {
    private:
      const char* message;

    public:

        /**
         * @brief The InitialisationError value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param message A clue that the underlying XML engine may
         *                have left behind, before it failed.
         */
      inline InitialisationError(const char* message)
        : std::runtime_error("InitialisationError"),
          message(message)
        {}


        /**
         * @brief Get the message possibly left behind by the XML
         * engine.
         *
         * This function returns a pointer to the message that may
         * have been left as a clue by the XML engine before it
         * failed.  This may be the <code>null</code> pointer.
         *
         * @return A pointer to message left behind by the XML engine.
         */
      inline const char*
      getMessage() const
        {
          return message;
        }
    };


      /**
       * @brief Class for reporting malformed attribute values.
       *
       * This exception is thrown if there are parsing difficulties in
       * extracting attribute values.  For example trying to parse
       * <code>true</code> as a double precision floating number is
       * really difficult.  It will result in this kind of exception
       * being thrown.  This class is a template class, even though
       * templates are not really used; the template merely gives
       * information about the destination type of the failed
       * conversion.
       */
    template<class T> class InvalidFormat : public std::runtime_error
    {
    private:
      const char* value;

    public:

        /**
         * @brief The InvalidFormat value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that the
         * argument may be the <code>null</code> pointer.  Invalid
         * pointers on the other hand should be avoided, as they have
         * the potential to cause much confusion.
         *
         * @param value The value which could not be expressed in the
         *              given format.
         */
      inline InvalidFormat(const char* value)
        : std::runtime_error("InvalidFormat"),
          value(value)
        {}


        /**
         * @brief Get the textual representation of the value which
         * could not be converted.
         *
         * This function returns a pointer to a textual representation
         * of the value which could not be converted into to the given
         * format.  This may be the <code>null</code> pointer.
         *
         * @return A pointer to a textual representation of the value
         *         which could not be converted.
         */
      inline const char*
      getValue() const
        {
          return value;
        }
    };


      /**
       * @brief Class for reporting XML parsing errors.
       *
       * This exception is thrown if there are parsing difficulties in
       * parsing the XML.  If the XML of the annotation is malformed,
       * we expect this exception to be thrown.
       */
    class ParseError : public std::runtime_error
    {
    private:
      unsigned int line;
      const char*  message;

    public:

        /**
         * @brief The ParseError value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param line    The line number where the parser ran into
         *                problems.
         * @param message A clue that the underlying XML engine may
         *                have left behind, before it failed.
         */
      inline ParseError(const unsigned int line, const char* message)
        : std::runtime_error("ParseError"),
          line(line), message(message)
        {}


        /**
         * @brief Get the line number where the parser ran into
         * problems.
         *
         * This function returns the line number where the parser ran
         * into problems.  This may be an invalid or nonsensical
         * number, because some underlying code may have reformatted
         * the input internally.
         *
         * @return The line number where the parser ran into problems.
         */
      inline unsigned int
      getLine() const
        {
          return line;
        }


        /**
         * @brief Get the message possibly left behind by the XML
         * engine.
         *
         * This function returns a pointer to the message that may
         * have been left as a clue by the XML engine before it
         * failed.  This may be the <code>null</code> pointer.
         *
         * @return A pointer to message left behind by the XML engine.
         */
      inline const char*
      getMessage() const
        {
          return message;
        }
    };


      /**
       * @brief Annotation class constructor.
       *
       * This will construct a new instance of the Annotation class,
       * with the given name, no attributes and no children.
       *
       * @param name The name of the new annotation.
       */
    inline Annotation(const std::string& name)
      : name(name)
      {
        attributes.clear();
        children.clear();
      }


      /**
       * @brief Annotation class value constructor.
       *
       * The annotation field is found in the SBML abstract type
       * SBase.  Since all other types in SBML are derived from SBase,
       * this means that any SBML type can have an annotation.  This
       * constructor builds an instance of the Annotation class from
       * the data found in the SBase object.  Theoretically, there may
       * be multiple annotations in an SBase object.  This constructor
       * will disregard any such annotations except the first.
       *
       * @param s A pointer to the SBase instance holding the relevant
       *          annotations.
       */
      //Annotation(const ::XMLNode* n) throw(InitialisationError, ParseError);
    Annotation(const ::XMLNode* n);

      /**
       * @brief Annotation class destructor.
       *
       * This function will free the attributes and, recursively, all
       * the children of this instance of the Annotation class.
       */
    inline ~Annotation()
      {
        attributes.clear();
        for (size_t i = 0; i < children.size(); i++)
          delete children[i];
        children.clear();
      }


      /**
       * @brief Add an attribute to the Annotation class instance.
       *
       * An attribute is really just a name-value tuple.  Both of
       * these quantities are standard C++ strings.  This function
       * just inserts the value at the given name in the attribute map
       * of the Annotation class instance.  Should there already be an
       * attribute with the given name, its value will simply be
       * replaced.
       *
       * @param name  The name of the attribute to add.
       * @param value The value of the attribute to add.
       */
    inline void
    addAttribute(const std::string& name, const std::string& value)
      {
        attributes[name] = value;
      }


      /**
       * @brief Add a child to the Annotation class instance.
       *
       * This function just adds the child to the list of children.
       * There is nothing to prevent the repeated addition of the same
       * pointer to the list of children.
       *
       * @param child A pointer to the child to add to the list of
       *              children.
       */
    inline void
    addChild(const Annotation* child)
      {
        children.push_back(child);
      }


      /**
       * @brief Get the @p i:th child of the annotation.
       *
       * Returns a pointer to the @p i:th child of the annotation.  If
       * the annotation does not have @p i children, the
       * <code>null</code> pointer is returned.
       *
       * @param  i The index of the child to get.
       * @return   The @p i:th child of the annotation provided it
       *           exists.
       */
    const Annotation*
    getChild(const size_t i) const
      {
        if (i < children.size())
          return children[i];
        return 0;
      }


      /**
       * @brief Get name of the annotation.
       *
       * Returns a reference to the name of the annotation.  This name
       * does not include any namespace information, and is, what an
       * XML engine might have called a <i>local name</i>.
       *
       * @return The name of the annotation.
       */
    const std::string&
    getName() const
      {
        return name;
      }


      /**
       * @brief Get the number of children in the annotation.
       *
       * This function could be equivalent to
       * <code>getFamilySize()</code> which would return the the
       * number of children plus two.  However, that would introduce a
       * heap of difficulties: in some families one or more parents
       * may be permanently absent.  Furthermore, modern families can
       * have more than two parents.  If a parent has children that
       * are shared with other families, we may end up with with a
       * fractional number of children.  Perhaps
       * <code>getNumChildren()</code> is good enough.
       *
       * @return The number of children in the annotation.
       */
    size_t
    getNumChildren() const
      {
        return children.size();
      }


      /**
       * @brief Get the value of a boolean-valued attribute.
       *
       * This function returns the boolean value of an attribute
       * called @p name in the annotation.  If there is no attribute
       * called @p name, the function throws an AttributeNotFound
       * exception.  If the found attribute does not evaluate to a
       * boolean, that is any of the case insensitive strings
       * <code>true</code> or <code>false</code>, the function throws
       * an InvalidFormat exception.
       *
       * @param name The name of the sought attribute.
       * @return     The boolean value.
       */
    inline bool
    getBoolAttribute(const std::string& name) const
      {
        std::string value = getStringAttribute(name);
        if (strcasecmp(value.c_str(), "true") == 0)
          return true;
        else if (strcasecmp(value.c_str(), "false") == 0)
          return false;
        throw InvalidFormat<bool>(value.c_str());
      }


      /**
       * @brief Get the double precision value of a floating
       * point-valued attribute.
       *
       * This function returns the floating point value in double
       * precision of an attribute called @p name in the annotation.
       * If there is no attribute called @p name, the function throws
       * an AttributeNotFound exception.  If the found attribute does
       * not evaluate to a floating point value, the function throws
       * an InvalidFormat exception.
       *
       * @param name The name of the sought attribute.
       * @return     The floating point value.
       */
    inline double
    getDoubleAttribute(const std::string& name) const
      {
        const char* str;
        double      d;
        int         n;

        str = getStringAttribute(name).c_str();
        if (str && sscanf(str, "%lf%n", &d, &n) >= 1
          && static_cast<size_t>(n) == strlen(str)) {
          return d;
        }	
        throw InvalidFormat<double>(str);
      }


      /**
       * @brief Get the single precision value of a floating
       * point-valued attribute.
       *
       * This function returns the floating point value in single
       * precision of an attribute called @p name in the annotation.
       * If there is no attribute called @p name, the function throws
       * an AttributeNotFound exception.  If the found attribute does
       * not evaluate to a floating point value, the function throws
       * an InvalidFormat exception.
       *
       * @param name The name of the sought attribute.
       * @return     The floating point value.
       */
    inline float
    getFloatAttribute(const std::string& name) const
      {
        const char* str;
        float       f;
        int         n;

        str = getStringAttribute(name).c_str();
	
        if (str && sscanf(str, "%f%n", &f, &n) >= 1
          && static_cast<size_t>(n) == strlen(str)) {
          return f;
        }
        throw InvalidFormat<float>(str);
      }


      /**
       * @brief Get the character string value of an attribute.
       *
       * This function returns the value of an attribute called @p
       * name in the annotation as a C++ standard string object.  If
       * there is no such attribute, the function throws an
       * AttributeNotFound exception.  Note that because XML
       * attributes can always be encoded as character strings, this
       * function will never throw a InvalidFormat exception.
       *
       * @param name The name of the sought attribute.
       * @return     The value as a C++ standard string.
       */
    inline const std::string&
    getStringAttribute(const std::string& name) const
      {
        std::map<std::string, std::string>::const_iterator i;
        if ((i = attributes.find(name)) == attributes.end())
          throw AttributeNotFound(this, name.c_str());
        return i -> second;
      }


      /**
       * @brief Put an annotation to an output stream.
       *
       * This function will put an annotation onto an output stream.
       * The result should be a valid SBML annotation if the root node
       * of the Annotation tree is output, and valid namespaced XML
       * otherwise.
       *
       * @param os The output stream on which to write.
       * @param a  The annotation in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream&
    operator<<(std::ostream& os, const Annotation& a);
  };
}


  /* Restore changed preprocessor directives.
   */
#ifdef __INTEL_COMPILER
#  if  __INTEL_COMPILER >= 800
#    pragma warning(pop)
#  endif
#endif

#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    pragma warning(pop)
#    undef _CRT_SECURE_NO_WARNINGS
#    undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
#  endif
#endif

#endif /* ifndef MESORD_ANNOTATION_HPP */
