/* The Compartment class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Compartment.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include <iostream>
#include <sstream>

#include "mesord.hpp"
#include "definition.hpp"

#include "Compartment.hpp"
#include "CompartmentTable.hpp"
#include "SubVolume.hpp"

#include "reaction/Reaction.hpp"


  /* Initialise the static data members.  Apparently this is not
   * required on some compilers.  However, the ANSI standard says that
   * static data members should always be initialised.
   */
MesoRD::Random::Random* MesoRD::Compartment::prng = 0;


  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Default constructor.
   */
MesoRD::Compartment::Compartment(Random::Random* r)
  : id(""),
    compartmentTable(0),
    outstream_log(0),
    volume(0)
{
  prng = r;
}

  /* Read the compartment from an SBML definition.
   */
MesoRD::Compartment::Compartment(
  CompartmentTable*     c,
  Random::Random*       r,
  const SpeciesTable*   s,
  const ParameterTable* p,
  const UnitTable*      u,
  const Model_t*        m,
  unsigned int          i) throw(MesoRD::Compartment::ModelError)
  : id(""),
    compartmentTable(c),
    outstream_log(0),
    volume(0)
{
  prng = r;

  try {
    fromSBML(s, p, u, m, i);
  } catch (ModelError& e) {
    throw ModelError(e.what());
  }
}


  /* Destructor.  Do-nothing.
   */
MesoRD::Compartment::~Compartment()
{}



  /****************
   * SBML SUPPORT *
   ****************/

  /* This is a lot of code that really does not do much.  In
   * principle, all we do is to build the species table and the
   * reaction table of the compartment.  The error handling
   * complicates things a little.
   */
void
MesoRD::Compartment::fromSBML(const SpeciesTable* s, const ParameterTable* p,
  const UnitTable* u, const Model* m, unsigned int i)
  throw(MesoRD::Compartment::ModelError)
{
  const ::Compartment* c = m -> getCompartment(i);
  unsigned int   j;
  
  if (c && c -> isSetId()) {
    id = c -> getId();
  } else {
    std::stringstream message;
    message << "Compartment " << i << " has no id.";
    throw ModelError((message.str()).c_str());
  }


    /* Parse the species of the model.  Copy the species table from
     * the parent and Update the relevant species in the local copy as
     * we go along.
     */
  speciesTable = *s;
  for (j = 0; j < m -> getNumSpecies(); j++) {
    const ::Species* s = m -> getSpecies(j);
    
    if ( s-> isSetCompartment() ) {
      if (id == s -> getCompartment() ) {
        try{
          speciesTable.updateSpecies(Species(const_cast< ::Species*>(s), 
              this, u, outstream_log));
        } catch(Species::ModelError& e) {
          throw ModelError(e.what());
        }

      } else {
        if ( s -> isSetId() ) {
          if (Species::getUseDoubleInAmount()) {
            speciesTable.setAmountById( s -> getId(), (double)0);
          } else {
            speciesTable.setAmountById( s -> getId(), (unsigned int)0);
          }
        }
      }
    } else {
      std::stringstream message;
      message << "No compartment associated with species " << j << ".";
      throw ModelError((message.str()).c_str());
    }
  }


  if (outstream_log) {
    (*outstream_log) << "The species table for "
                     << m -> getNumSpecies()
                     << " species in compartment "
                     << id
                     << " is:"
                     << std::endl << speciesTable << std::endl;
  }
  
   
    /* Parse the reactions.  We only want to add those reactions that
     * belong to the current compartment.  With all the exceptions
     * thrown around, this part seems perhaps hairier than it really
     * is.
     */
  for (j = 0; j < m -> getNumReactions(); j++) {
    const ::Reaction* r = m -> getReaction(j);
    std::string compartment = Reaction::getCompartment(r, &speciesTable);
    if (compartment == id) {
      try {
        reactionTable.addReaction(
          *(new Reaction(r, p, &speciesTable, outstream_log)));
      } catch (Reaction::ModelError& e) {
        throw ModelError(e.what());
      }
    }
  }
  if (outstream_log) {
    (*outstream_log) << "The reaction table for "
                     <<  m -> getNumReactions()
                     << " reactions in compartment "
                     << id
                     << " is:"
                     << std::endl << reactionTable << std::endl;
  }
}



  /************************
   * SPECIES DISTRIBUTION *
   ************************/

  /* Randomly place the molecules of different species into the sub
   * volumes.  Loop over the species and place each molecule into a
   * randomly determined sub volume.  First we need to make sure that
   * all the sub volumes have capacity to store the required number of
   * different species.
   */

double
MesoRD::Compartment::uniformSpeciesDistributionDeterministic(
  std::vector<SubVolume*>& geometry) const
{
  double total = 0;

  if (geometry.size() > 0 && speciesTable.getSpeciesNumber() > 0) {
    unsigned int i;

    for (i = 0; i < geometry.size(); i++)
      geometry[i] -> setSpeciesNumber(speciesTable.getSpeciesNumber());

    long geometrySize = geometry.size();

    for (i = 0;  i < speciesTable.getSpeciesNumber(); i++) {
      for (long j = 0; j < geometrySize; j++) {
        geometry[j] -> distributeSpecies(i, (double) 
          (speciesTable.getAmountByIndexDouble(i) / geometrySize));

        total += speciesTable.getAmountByIndexDouble(i) / geometrySize;
      }
    }
  }

  return total;
}


  /* FIXME!! Maybe it would be a good idea to do some output here as
   * to how many molecules will be distributed.  The reason is that if
   * the number is horribly large (which may happen if the model is
   * incorrectly defined), this step will take a long, long time.  If
   * there is no output, this gives the impression of a hang.
   */
unsigned long
MesoRD::Compartment::uniformSpeciesDistribution(
  std::vector<SubVolume*>& geometry) const
{
  unsigned long total = 0;

  if (geometry.size() > 0 && speciesTable.getSpeciesNumber() > 0) {
    unsigned int i;

    for (i = 0; i < geometry.size(); i++)
      geometry[i] -> setSpeciesNumber(speciesTable.getSpeciesNumber());

    for (i = 0;  i < speciesTable.getSpeciesNumber(); i++) {
      for (unsigned int j = 0; j < speciesTable.getAmountByIndex(i); j++) {
        geometry[static_cast<size_t>(
                   rint(prng -> randomClosed() * (geometry.size() - 1)))]
          -> distributeSpecies(i, (int) 1);
        total++;
      }
    }
  }

  return total;
}
