/* The Compartment class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Compartment.hpp 556 2008-09-01 18:38:28Z davidfange $
 */
#ifndef MESORD_COMPARTMENT_HPP
#define MESORD_COMPARTMENT_HPP 1

#include <string>
#include <vector>

#include "CompartmentTable.hpp"
#include "ReactionTable.hpp"
#include "SpeciesTable.hpp"
#include "random/Random.hpp"

#ifdef WIN32
#pragma warning(disable: 4290)
#endif

namespace MesoRD {

  class ParameterTable;


    /**
     * Class for keeping track of compartments.
     *
     * A compartment keeps track of its root compartment table.  This
     * is done so that changes to a compartment can be reflected
     * globally if necessary.  The methods that are called for
     * modifying the global species table live in time-critical
     * section -- the sub volume event handling -- so we will not
     * bother with range checking.  Unlike SBML, reactions in MesoRD
     * are local to compartments.  Therefore, a compartment holds its
     * own table of reactions.
     *
     */
  class Compartment {
  private:
    ReactionTable     reactionTable;
    SpeciesTable      speciesTable;
    std::string       id;
    CompartmentTable* compartmentTable;
    std::ostream*     outstream_log;
    double            volume;
    std::string       origUnit;
    

      /**
       * @brief A pointer to the pseudo-random number generator.
       *
       * All instances of the Compartment class share the same
       * instance of the Random class.
       */
    static Random::Random* prng;


  public:


      /**
       * Class for reporting errors during model building.
       *
       * If something went wrong during the model construction, this
       * exception will be thrown.
       */
    class ModelError : public std::runtime_error {

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
    public:
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Default constructor of the Compartment class.
       *
       * Sets everything except for the pseudo-random number generator
       * to null and void.  The compartment is unusable if initialised
       * like this.
       *
       * @param prng Pointer to the pseudo-random number generator.
       */
    Compartment(Random::Random* r);

      /**
       * Construct the compartment from an SBML model.
       *
       * All this method does is to call the fromSBML() method in
       * order to read in the model from the SBML model definition.
       *
       * @param c The compartment table of which this compartment is part.
       * @param r Pointer to the pseudo-random number generator.
       * @param s The species table.  Should be a copy of the global table.
       * @param p The root parameter table.
       * @param u The unit table.
       * @param m The SBML model.
       * @param i The index of this compartment in the SBML model.
       */
    Compartment(
      CompartmentTable*     c,
      Random::Random*       r,
      const SpeciesTable*   s,
      const ParameterTable* p,
      const UnitTable*      u,
      const Model*          m,
      unsigned int          i) throw(ModelError);

    ~Compartment();

      /**
       * Uniformly distribute the species in the sub volumes of the
       * compartment.
       *
       * Once the number of molecules of the different species is
       * known, they can be distributed among the compartment's sub
       * volumes.  Calling this method will do just that.  Note that
       * this method utilises a uniform distribution of molecules.
       *
       * @param geometry The sub volumes of the compartment.
       * @return         The total number of molecules distributed.
       */
    unsigned long uniformSpeciesDistribution(
      std::vector<SubVolume*>& geometry) const;

    double uniformSpeciesDistributionDeterministic(
      std::vector<SubVolume*>& geometry) const;
      /**
       * Get the id of the compartment.
       *
       * The id is what identifies an compartment.
       *
       * @return The id of the compartment.
       */
    inline std::string getId() const
      { return id; }

      /**
       * Get the unit as it was read from the SBML file.  
       *
       * When the SBML is read and the MesoRD creates its own
       * compartment representation the unit is comverted into litres.
       * When the Species are read from the SBML file the
       * initialConcentration is defined in the original unit as it is
       * read from SBML file. This will return this unit and is used
       * when creating new Species.
       */
    inline std::string getOriginalUnit() const
      { return origUnit; }
    
        
      /**
       * Get the volume of the compartment.
       *
       * Note that the volume is returned in litres.  Typically, the
       * volume of an compartment is the number of sub volumes used to
       * construct it, multiplied by the volume of a sub volume.
       *
       * @return The volume of the compartment.
       */
    inline double getVolume() const
      { return volume; }

      /**
       * Get the diffusion constant of a certain species into another
       * compartment.
       *
       * In every compartment, there is a diffusion constant
       * associated with every species and every compartment.  Using
       * this method, one will find the diffusion constant for
       * diffusing a species identified by "index" from the current
       * compartment into compartment identified by name
       * "compartment".  Note that the compartment name can be the
       * name of the current compartment, in which case the return
       * value is to be interpreted as the diffusion constant for
       * diffusing the species between different sub volumes within
       * the current compartment.
       *
       * @return The volume of the compartment.
       */
    inline double getDiffusion(
      const size_t index, const std::string& compartment) const
      { return speciesTable.getDiffusion(index, compartment); }

      /*inline SpeciesTable& getSpeciesTable()
      { return speciesTable; }
      */

/**
       * Build the compartment from an SBML model.
       *
       * A compartment keeps track of the species it contains and the
       * reactions that affect these species.  Calling this method
       * will set up these two tables according to the specifications
       * in the SBML model.
       *
       * @param s The species table.  Should be a pointer to the global table.
       * @param p The root parameter table.
       * @param u The unit table.
       * @param m The SBML model.
       * @param i The index of this compartment in the SBML model.
       */
    void fromSBML(
      const SpeciesTable*   s,
      const ParameterTable* p,
      const UnitTable*      u,
      const Model*          m,
      unsigned int          i) throw(ModelError);

      /**
       * Set the compartment table of the compartment.
       *
       * The every compartment has a pointer back to its compartment
       * table.
       *
       * @param c A pointer to the parent compartment table.
       */
    inline void setCompartmentTable(CompartmentTable* c)
      { this -> compartmentTable = c; }


      /**
       * Set the id of the compartment.
       *
       * The id is what identifies an compartment.
       *
       * @param i The new id of the compartment.
       */
    inline void setId(const std::string& i)
      { id = i; }

      /**
       * Sets the unit defining the SBML file unit.    
       * 
       * This is a "fluke" function, but it allows for easy testing.  
       *
       */
    inline void setOriginalUnit(const std::string u)
      { origUnit = u; }
    
      /**
       * Set the volume of the compartment.
       *
       * Note that the volume has to be given in litres.  Typically,
       * the volume of an compartment is the number of sub volumes
       * used to construct it, multiplied by the volume of a sub
       * volume.
       *
       * @param v The new volume of the compartment.
       */
    inline void setVolume(double v)
      { volume = v; }


      /**
       * Change the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline unsigned int
    changeSpecies(const off_t index, const int change)
      { 
        return speciesTable.changeAmountByIndex(index, change);
          //if (compartmentTable) 
          //return compartmentTable -> decreaseSpecies(index, decrement);
          //return 0; 
      }


      /**
       * Change the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline double
    changeSpecies(const off_t index, const double change)
      { 
        return speciesTable.changeAmountByIndex(index, change);
          //if (compartmentTable) return compartmentTable -> decreaseSpecies(
          //index, decrement); return 0; }
      }


      /**
       * Get the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the sub volume.
       */
    inline unsigned int getSpecies(off_t index) const
      {
        return speciesTable.getAmountByIndex(index);
          //if (compartmentTable) return compartmentTable -> getSpecies(
          //index); return 0; }
      }
    
    inline size_t getSpeciesNumber() const
      { return speciesTable.getSpeciesNumber(); }
    
    inline std::string getSpeciesId(off_t index)
      { return speciesTable.getId(index); }
      /**
       * Set the global number of a certain species to an arbitrary
       * value.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline unsigned int setSpecies(off_t index, unsigned int number)
      {
        return speciesTable.setAmountByIndex(index, number);
          //if (compartmentTable) return compartmentTable -> setSpecies(
          //index, number); return 0; 
      }


      /**
       * Get the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the sub volume.
       */
    inline double getSpeciesDouble(off_t index) const
      {
        return speciesTable.getAmountByIndexDouble(index);
          //if (compartmentTable) return compartmentTable -> getSpeciesDouble(
          //index); return 0; }
      }


      /**
       * Set the global number of a certain species to an arbitrary
       * value.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline double setSpecies(off_t index, double number)
      {
        return speciesTable.setAmountByIndex(index, number);
          //if (compartmentTable) return compartmentTable -> setSpecies(
          //index, number); return 0; }
      }
    


      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing and
       * model building begins, any error reporting and logging
       * information will be silently lost.  It is thus highly
       * recommended to call this method prior to calling fromSBML().
       * Note also that this method will set up the logging stream for
       * any member classes that can make use of one.
       *
       * @param log The logging stream onto which to report errors, etc.
       */
    inline void setOptionLogfile(std::ostream* log)
      { outstream_log = log; }

      /**
       * The SubVolume class needs to access the reaction table.  For
       * faster access, we let the SubVolume class be a friend of the
       * Compartment class.
       */
    friend class SubVolume;
    friend class CubicSubVolume;
    friend class OctahedricSubvolume;
    
    
  };
}
#endif /* ifndef MESORD_COMPARTMENT_HPP */
