/* The CompartmentTable class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CompartmentTable.cpp 561 2008-11-25 20:10:53Z davidfange $
 */
#include <iomanip>
#include <fstream>
#include <iostream>
#include <sstream>

#include "definition.hpp"

#include "Compartment.hpp"
#include "CompartmentTable.hpp"
#include "CubicSubVolume.hpp"
#include "OctahedricSubvolume.hpp"
#include "ParameterTable.hpp"
#include "SubVolume.hpp"
#include "Unit.hpp"
#include "UnitTable.hpp"

#include "csg/CSG.hpp"
#include "csg/CSGNode.hpp"

#include "linalg/Matrix.hpp" // FIXME!!
#include "linalg/Vector.hpp" // FIXME!!

#include "random/Random.hpp"

#include "sparselib/comprow_double.hpp"
#include "sparselib/compcol_double.hpp"
#include "mv/mvvd.hpp"


  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Default constructor.  Make sure we are aware that we do not have
   * a logging stream yet.
   */
MesoRD::CompartmentTable::CompartmentTable()
{
  outstream_log = 0;
}


  /* Destructor.  Do-nothing.
   */
MesoRD::CompartmentTable::~CompartmentTable()
{}



  /****************
   * SBML SUPPORT *
   ****************/

  /* This function was deemed misplaced in the CSG module.  Maybe it
   * is misplaced here as well?  It certainly should not be defined in
   * the empty namespace.
   *
   */
namespace
{
  void
  csgFromSBML(std::ostream*                                  outstream_log,
    const ListOfCompartments*                                 compartments,
    const MesoRD::UnitTable*                                 ut,
    const double                                             size_metres,
    std::map<std::string, std::vector<MesoRD::SubVolume*> >& geometries,
    MesoRD::Random::Random*                                  prng)
    throw(MesoRD::Species::ModelError)
  {

    const MesoRD::Subvolume_t sv_type = MesoRD::cubic; 
                                              // cubic || octahedric

    std::map<std::string, MesoRD::CSG::Node*> forest;

    if (compartments) {
      for (size_t j = 0; j < compartments -> size(); j++) {
        const ::Compartment* c = 
          dynamic_cast<const ::Compartment*> (compartments -> get(j));

        if ( c -> isSetSize() ) {
          if (outstream_log) {
            (*outstream_log) << "WARNING: MesoRD will recompute size"
                             << std::endl;
          }
        }

        if ( c -> isSetVolume() ) {
          if (outstream_log) {
            (*outstream_log) << "WARNING: MesoRD will recompute volume"
                             << std::endl;
          }
        }

        if ( c -> getSpatialDimensions() != 3) {
          if (outstream_log) {
            (*outstream_log) << "WARNING: Reality is three-dimensional"
                             << std::endl;
          }
        }

          /* !! FIXME !! This should throw an exception if there is no id
           */
        if (c -> isSetId()) {
          const std::string        id = c -> getId();
          const MesoRD::Annotation 
            annotation((const_cast< ::Compartment*>(c)) -> getAnnotation());

          if (annotation.getNumChildren() >= 1) {
            const MesoRD::Annotation* child = annotation.getChild(0);
            MesoRD::CSG::Node*        tree  = MesoRD::CSG::fromXML(
              child, ut, size_metres, MesoRD::Matrix::eye(4));

            forest[id] = tree;

          } else {
            if (outstream_log) {
              (*outstream_log) << "ERROR: Compartment \""
                               << id << "\" has no geometry"
                               << std::endl;
            }
          }
        } else {
          if (outstream_log) {
            (*outstream_log) << "ERROR: Compartment without ID"
                             << std::endl;
          }
        }
      }

        /* Resolve compartment references.
         */
      if(outstream_log)
        (*outstream_log) << "Resolving the forest... "<< std::flush;
      
      MesoRD::CSG::resolve(forest);
      
      if(outstream_log)
        (*outstream_log) << "DONE!" << std::endl;
      
    }

      /* This is the important call that will build the geometries.
       * It is <em>very</em> important to remember to set the size
       * once the geometries are built.
       */
    if(outstream_log)
        (*outstream_log) << "Realising subvolumes in geometry... "<< std::flush;
    
    MesoRD::CSG::realise(forest, sv_type, geometries, 0, prng);
    
    if(outstream_log)
      (*outstream_log) << "DONE!" << std::endl;

    MesoRD::SubVolume::setSize(size_metres * 100);
    
    if(outstream_log) {
      (*outstream_log) << "The size of a subvolume is "
                       << size_metres << " metres" << std::endl;
    }
    

    if (outstream_log) {
      std::map<std::string, MesoRD::CSG::Node*>::const_iterator i;

      for (i = forest.begin(); i != forest.end(); i++) {
        const std::string id = i -> first;

        (*outstream_log) << "Compartment \"" << id
                         << "\" has "        << geometries[id].size()
                         << " sub volumes."  << std::endl;
      }
    }

      /* ADDITION TO OUTPUT POVRAY CSG HERE */
    std::string   filename("geometry.pov");
    std::ofstream povray(filename.c_str(), std::ios::trunc);
    if (povray)
      MesoRD::CSG::putPOVRay(forest, povray);
      /* ADDITION TO OUTPUT POVRAY CSG HERE */
  }
}



  /* Create the global species table.  We must then read the
   * geometries before we proceed, because there may be strange
   * interdependencies between them.  For every compartment in the
   * model, set the size.  Then randomly distribute the molecules into
   * the compartments.  Finally, return the vector containing all the
   * sub volumes.
   */
std::vector<MesoRD::SubVolume*>
MesoRD::CompartmentTable::fromSBML(const Model* m, ParameterTable* p,
  const UnitTable* u, double size, const std::string& unit,
  Random::Random* prng)
  throw(
    MesoRD::CompartmentTable::EmptyGeometry,
    MesoRD::CompartmentTable::ModelError,
    MesoRD::CompartmentTable::NoId)
{
  unsigned int i;

    /* Create the global species table.  We do this by adding all the
     * species from the SBML model.  Then we set the set all the
     * amounts to zero.  The amounts in the global species table will
     * be updated as we go along.
     */
  speciesTable.clear();
  try {
    for (i = 0; i < m -> getNumSpecies(); i++) {
      speciesTable.addSpecies(
        *(new Species(const_cast< ::Species*>(m -> getSpecies(i)), 
            0, u, outstream_log)));
    }
  } catch( MesoRD::Species::ModelError e) {
    throw ModelError(e.what());
  }
  speciesTable.clearAmounts();
	
  
    /* Build the geometries of all the compartments in the model.  For
     * every compartment set the logging stream and the volume, which
     * is just the volume of a single sub volume multiplied by the
     * number of sub volumes used to build the geometry of the
     * compartment.  This volume will be recorded as a constant in the
     * global parameter table as well.  We want to store the volume in
     * litres, but the size is stored in cm, so we need to do some
     * conversion here.
     */
  std::map<std::string, std::vector<SubVolume*> > geometries;

    /* Grabbing the correct unit for the subvolumes, before building
     * them.
     */
  const MesoRD::Unit* unit_sv     = u -> getUnit(unit);
  if(!unit_sv) {
    std::stringstream message;
    message << "The unit "
            << unit
            << " used for subvolume size is not defined";
    throw MesoRD::CompartmentTable::ModelError((message.str()).c_str());
  }
  
  double conversion = 0;
  try{
    conversion  = unit_sv -> toUnit(MesoRD::Unit::metre());
  } catch (Unit::InvalidConversion&) {
    delete unit_sv;
    std::stringstream message;
    message << "Cannot convert unit "
            << unit
            << " to a length unit, which is required for subvolume size.";
    throw ModelError((message.str()).c_str());
  }
  
  const double size_metres          = size * conversion;
  csgFromSBML(outstream_log,
    m -> getListOfCompartments(), u, size_metres, geometries, prng);
  

  if (outstream_log) {
    (*outstream_log) << "Every sub volume will require "
                     << sizeof(CubicSubVolume)
                     << " bytes of memory."
                     << std::endl;
  }
  
  compartments.clear();
  for (i = 0; i < m -> getNumCompartments(); i++) {
    const ::Compartment* c = m -> getCompartment(i);
    
    if (c -> isSetId()) {
      std::string             id       = c -> getId();
      std::vector<SubVolume*> geometry = geometries[id];

      if (geometry.size() > 0) {
        Compartment* compartment = new Compartment(prng);

        compartment -> setCompartmentTable(this);

        std::string origUnit = "";        
        if(c -> isSetUnits())
          origUnit = c -> getUnits();

        if (typeid(*geometry[0]) == typeid(CubicSubVolume)) {
            //|| typeid(*geometry[0]) == typeid(OctahedricSubvolume)) { // FIXME!!
          double      size_cubic         = geometry[0] -> getSize();
          double      volume_cubic       =
            size_cubic * size_cubic * size_cubic / 1000;
          double      volume_compartment = 1.0 * geometry.size()
            * volume_cubic;
          std::string volume_name        = id;
          std::string volume_cubic_name  = std::string("volume_subVolume");

          std::string one_molecule_molar_name =
            std::string("one_molecule_molar");

          MesoRD::Unit one_molecule_molar_unit = Unit::mole() / Unit::litre();
          double one_molecule_molar_value = 
            1.0 / (volume_cubic * Constant::Avogadro);

          std::string size_name = "size_subVolume";

          std::string one_molecule_molar_special_name =
            std::string("one_molecule_molar_spc");

          compartment -> setOptionLogfile(outstream_log);
          compartment -> setVolume(volume_compartment);
          compartment -> setOriginalUnit(origUnit);
          
          p -> addReplaceParameter(
            volume_name, volume_name, volume_compartment, 
            Unit::litre(), true);
          
          p -> addReplaceParameter(
            volume_cubic_name, volume_cubic_name, volume_cubic, 
            Unit::litre(), true);

          p -> addReplaceParameter(
            one_molecule_molar_name, one_molecule_molar_name,
            one_molecule_molar_value, one_molecule_molar_unit, true);

          p -> addReplaceParameter(
            size_name, size_name, size_cubic, Unit::centimetre(), true);

            /* Adding the special parameter one_molecule_molar_special
             * which will have different values in stochastic and
             * determinsitc simulations to allow for better
             * compatibilty between determinsitic and stochastic
             * simulations using bimolecular readctions
             */
          double one_molecule_molar_special_value = 0;
          if( SubVolume::getUseDoubleInSpecies()) {
            one_molecule_molar_special_value = 0;
          } else {
            one_molecule_molar_special_value = 
              one_molecule_molar_value;
          }

          p -> addReplaceParameter(
            one_molecule_molar_special_name,
            one_molecule_molar_special_name, 
            one_molecule_molar_special_value,
            one_molecule_molar_unit, true);

          if (outstream_log) {
            (*outstream_log) << "Recorded parameter \""
                             << volume_cubic_name
                             << "\" with value "
                             << volume_cubic
                             << " litres in global parameter table."
                             << std::endl

                             << "Recorded parameter \""
                             << volume_name
                             << "\" with value "
                             << volume_compartment
                             << " litres in global parameter table."
                             << std::endl

                             << "Recorded parameter \""
                             << one_molecule_molar_name
                             << "\" with value "
                             << one_molecule_molar_value
                             << " litres in global parameter table."
                             << std::endl

                             << "Recorded parameter \""
                             << one_molecule_molar_special_name
                             << "\" with value "
                             << one_molecule_molar_special_value
                             << " litres in global parameter table."
                             << std::endl

                             << "Recorded parameter \""
                             << size_name
                             << "\" with value "
                             << size_cubic
                             << " centimetre in global parameter table."
                             << std::endl;
          }
        }

        try {
          compartment -> fromSBML(&speciesTable, p, u, m, i);
        } catch (Compartment::ModelError& e) {
          throw ModelError(e.what());
        }

        compartments.push_back(compartment);


          /* Make sure that every sub volume in the compartment has a
           * valid pointer back to its compartment.
           */
        std::vector<SubVolume*>::iterator j;
        for (j = geometry.begin(); j != geometry.end(); j++)
          (*j) -> setCompartment(compartment);
      } else {
        std::stringstream message;
        message << "No subvolumes in compartment: " << id << std::endl;
        throw EmptyGeometry((message.str()).c_str());
      }
    } else {
      std::stringstream number;
      number << "Compartment number " << i;
      throw NoId((number.str()).c_str());
    }
  }

  if (outstream_log) {
    (*outstream_log) << "This is the final parameter table" << std::endl
                     << *p << std::endl;
  }

    /* Distribute the molecules into each compartment.  We use a
     * uniform random distribution.  The rationale is that if other
     * distribution are needed, one may accomplish that to a
     * reasonable approximation by defining the compartments
     * differently.
     */
  std::list<Compartment*>::iterator j;
  for (j = compartments.begin(); j != compartments.end(); j++) {
    Compartment*            compartment = *j;
    std::string             id          = compartment -> getId();
    std::vector<SubVolume*> geometry    = geometries[id];

    if(SubVolume::getUseDoubleInSpecies()) {
      double total = compartment
        -> uniformSpeciesDistributionDeterministic(geometry);

      if (outstream_log) {
        (*outstream_log) << "Distributed in total "
                         << total
                         << " molecules in compartment \""
                         << id
                         << "\"." << std::endl;
      }

    } else {
      unsigned long total = compartment
        -> uniformSpeciesDistribution(geometry);
      if (outstream_log) {
        (*outstream_log) << "Distributed in total "
                         << total
                         << " molecules in compartment \""
                         << id
                         << "\"." << std::endl;
      }
    }
  }


    /* This is an ugly piece of work to work around the new CSG
     * module.  It just puts all the subvolume from all the
     * compartments into a single vector.
     *
     * FIXME!!
     */
  std::vector<SubVolume*>                                   all_subvolumes;
  std::map<std::string, std::vector<SubVolume*> >::iterator iter;

  for (iter = geometries.begin(); iter != geometries.end(); iter++) {
    std::vector<SubVolume*>::iterator jter;
    for (jter = iter -> second.begin(); jter != iter -> second.end(); jter++)
      all_subvolumes.push_back(*jter);
  }

  return all_subvolumes;
}


  /*******************
   * LOGFILE CONTROL *
   *******************/

  /* Set the logging stream of all compartments in the table.  Then
   * set the logging stream of the table itself.
   */
void
MesoRD::CompartmentTable::setOptionLogfile(std::ostream* log)
{
  std::list<Compartment*>::iterator i;

  for (i = compartments.begin(); i != compartments.end(); i++)
    (*i) -> setOptionLogfile(log);
    
  outstream_log = log;
}



  /*****************
   * CHECKPOINTING *
   *****************/

  /* Writes a formatted summary of the local species table to the
   * supplied stream.
   */

void
MesoRD::CompartmentTable::startCheckpointing(std::ostream& os)
{
  
  const int width = 20;

  os.setf(std::ios_base::left);
  os.fill(' ');

  os << std::setw(width)
     << "Time:" << "\t";  
  
  std::list<Compartment*>::const_iterator compartment;
  for(compartment = compartments.begin(); 
      compartment != compartments.end(); compartment++) {
    for (unsigned int j = 0; j < (*compartment) -> getSpeciesNumber(); j++) {
      os << std::setw(width)
         << (*compartment) -> getSpeciesId(j)
         << "\t";
    }
  }
  
  os << std::endl;
}

  

void
MesoRD::CompartmentTable::checkpointGlobal(
  std::ostream& os, std::valarray<size_t>&, double time) const
{
    //bool      started = false;
  const int width   = 20;
/*  
  os << "Global checkpoint at iteration ";
  for (int i = (iteration.size() - 1); i >= 0; i--) {
    if (started || (iteration[i] > 0)) {
      started = true;
      os << iteration[i];
    }
  }
  os << "." << std::endl;

  os.setf(std::ios_base::right);
  os.fill(' ');
  os << std::setw(width)
     << "Time"
     << " : ";
  os.setf(std::ios_base::left);
  os.fill(' ');
  os << time
     << std::endl;

  for (unsigned int j = 0; j < speciesTable.getSpeciesNumber(); j++) {
    os.setf(std::ios_base::right);
    os.fill(' ');
    os << std::setw(width)
       << speciesTable.getKey(j)
       << " : ";

    os.setf(std::ios_base::left);
    os.fill(' ');

    if(SubVolume::getUseDoubleInSpecies())
    { 
      os << speciesTable.getAmountByIndexDouble(j)
         << std::endl;
    }
    else
    {
      os << speciesTable.getAmountByIndex(j)
         << std::endl;
    }
    
  }
*/

  
  os.setf(std::ios_base::left);
  os.fill(' ');
  
  os << std::setw(width)
     << time << "\t";

  std::list<Compartment*>::const_iterator compartment;
  for(compartment = compartments.begin(); 
      compartment != compartments.end(); compartment++) {
    for (unsigned int j = 0; j < (*compartment) -> getSpeciesNumber(); j++) {
      os << std::setw(width)
         << (*compartment) -> getSpecies(j)
         << "\t";
    }
  }
  os << std::endl;
}


void
MesoRD::CompartmentTable::checkPointGlobalDeterministic(
  std::ostream& os, std::valarray<size_t>&, double time,
  std::vector<SubVolume*>& subVolumes, 
  MV_Vector_double& currentState) const
{
  const int width   = 20;
  
  os.setf(std::ios_base::left);
  os.fill(' ');
  
  os << std::setw(width)
     << time << "\t";

  std::map<std::string, double> species;
  std::vector<std::string> speciesNames;
  std::list<Compartment*>::const_iterator compartment;
  for(compartment = compartments.begin(); 
      compartment != compartments.end(); compartment++) {
    for (unsigned int j = 0; j < (*compartment) -> getSpeciesNumber(); j++) {
      species[(*compartment) -> getSpeciesId(j)] = 0;
      speciesNames.push_back((*compartment) -> getSpeciesId(j));
    }
  }
    
  size_t speciesNumber = (*(compartments.begin())) -> getSpeciesNumber();
  for(unsigned int i = 0; i < subVolumes.size(); i++) {
    for(unsigned int j = 0; j < speciesNumber; j++) {
      species[(subVolumes[i] -> getCompartment()) -> getSpeciesId(j)] 
        += currentState(static_cast<int>(i * speciesNumber + j));
    }
  }


  for(unsigned int k = 0; k < speciesNames.size(); k++) {
    os << std::setw(width)
       << species[speciesNames[k]]
       << "\t";
  }
  os << std::endl;
}


  


  /* Short and sweet...  Well, short, anyway.
   */
void
MesoRD::CompartmentTable::checkpointLocal(
  std::vector<SubVolume*>& subVolumes)
{
  const int width = 10;
 
  
  for (unsigned int i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    std::ofstream* outfile = outfiles_local[speciesTable.getKey(i)];
 
    if (outfile) {
      if (outputSparse ) {
        bool emptyTimePoint = true;
        for (unsigned int j = 0; j < subVolumes.size(); j++) {
          if( subVolumes[j] -> getSpecies(i) > 0) {
            *outfile << std::setw(width)
                     << j << " " << subVolumes[j] -> getSpecies(i) << " ";
            emptyTimePoint = false;
          }
        }
          // if there where no molcules of the current species,
          // indicate this with a sole -1.  
        if (emptyTimePoint) 
          *outfile << "-1";
        

      } else {
        for (unsigned int j = 0; j < subVolumes.size(); j++) {
          *outfile << std::setw(width)
                   << subVolumes[j] -> getSpecies(i) << " ";
        }
      }
      *outfile << std::endl;
    }
  }
}



void
MesoRD::CompartmentTable::checkpointLocalDeterministic(
  std::vector<SubVolume*>& subVolumes, MV_Vector_double& currentState)
{
  const int width = 10;

  size_t speciesNumber = (*(compartments.begin())) -> getSpeciesNumber();
  
  for (unsigned int i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    std::ofstream* outfile = outfiles_local[speciesTable.getKey(i)];
 
    if (outfile) {
      
      for (unsigned int j = 0; j < subVolumes.size(); j++) {
        *outfile << std::setw(width)
                 << currentState(static_cast<int>(i + speciesNumber * j))
		 << " ";
      }
    }
    *outfile << std::endl;
  }
}



  /*****************
   * FILE HANDLING *
   *****************/

  /* The method name says it all.
   */
void
MesoRD::CompartmentTable::closeLocalFiles()
{
  for (unsigned int i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    std::string key = speciesTable.getKey(i);
    outfiles_local[key] -> close();
  }
}


  /* Open files for storing the amount of each species in each sub
   * volume.  Make sure to use the "trunc" argument, otherwise we will
   * just append to files.
   */
void
MesoRD::CompartmentTable::openLocalFiles(std::string& prefix)
{
  outfiles_local.clear();
  for (unsigned int i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    std::string key = speciesTable.getKey(i);
    outfiles_local[key] = new std::ofstream(
      (prefix
        + MESORD_SPECIES_FILE_PREFIX
        + key
        + MESORD_SPECIES_FILE_SUFFIX).c_str(),
      std::ios::trunc);
  }
}

std::map<int, std::ifstream*>
MesoRD::CompartmentTable::openSpeciesFiles(std::string& prefix)
  throw( MesoRD::CompartmentTable::FileNotFoundError )
{
  std::map<int, std::ifstream*> infiles_local;
  for (unsigned int i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    std::string key = speciesTable.getKey(i);
    infiles_local[i] = new std::ifstream(
      (prefix
        + MESORD_SPECIES_FILE_PREFIX
        + key
        + MESORD_SPECIES_FILE_SUFFIX).c_str());
    if(!infiles_local[i]) {
      std::stringstream message;
      message << "Unable to open file: "
              << prefix
              << MESORD_SPECIES_FILE_PREFIX
              << key
              << MESORD_SPECIES_FILE_SUFFIX 
              << " for reading." << std::endl;
      throw(FileNotFoundError((message.str()).c_str()));
    }
  }
  return infiles_local;
}


void
MesoRD::CompartmentTable::closeSpeciesFiles(std::map<int, std::ifstream*>& infiles_species)
{
  std::map<int, std::ifstream*>::const_iterator it;
  for(it = infiles_species.begin(); it != infiles_species.end(); ++it){
    (it -> second) -> close();
  }
}
