/* The CompartmentTable class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CompartmentTable.hpp 556 2008-09-01 18:38:28Z davidfange $
 */
#ifndef MESORD_COMPARTMENTTABLE_HPP
#define MESORD_COMPARTMENTTABLE_HPP 1

#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include <list>
#include <stdexcept>
#include <string>
#include <valarray>

#include "SpeciesTable.hpp"

#include "sparselib/comprow_double.hpp"
#include "sparselib/compcol_double.hpp"
#include "mv/mvvd.hpp"
#include "random/Random.hpp"

namespace MesoRD {

  class Compartment;
  class ParameterTable;
  class SubVolume;
  class UnitTable;


    /**
     * Class for keeping track of all the compartments in the model.
     *
     * Similarly to how things are organised in SBML, there is an
     * entity to keep track of all the compartments.  In MesoRD this
     * entity is called a CompartmentTable and it is defined here.
     * The CompartmentTable keeps an instance of a SpeciesTable.  The
     * reason is that we record the global amount of all the species
     * in this table.  The methods that are called for modifying the
     * global table live in time-critical section -- the sub volume
     * event handling -- so we will not bother with range checking.
     */
  class CompartmentTable {
  private:
    SpeciesTable                          speciesTable;
    std::list<Compartment*>               compartments;
    std::map<std::string, std::ofstream*> outfiles_local;
    std::ostream*                         outstream_log;

    bool                                  outputSparse;
  public:


      /**
       * Class for reporting that the geometry is empty.
       *
       * If the geometry construction results in an empty geometry,
       * this exception will be thrown.  An empty geometry is most
       * likely the result of an incorrectly defined CSG tree.
       */
    class EmptyGeometry : public std::runtime_error {

        /**
         * The EmptyGeometry constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
    public:
      EmptyGeometry(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Class for reporting errors during model building.
       *
       * If something went wrong during the model construction, this
       * exception will be thrown.
       */
    class ModelError : public std::runtime_error {

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
    public:
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Class for reporting that a compartment lacks an id.
       *
       * If the compartment definition in the SBML file lacks an id,
       * this exception will be thrown.  We must have an id on every
       * compartment according to the SBML definition.
       */
    class NoId : public std::runtime_error {
      
        /**
         * The NoId constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      public:
      NoId(const char* what = "")
        : runtime_error(what) {}
    };
    
      /**
       * Class for reporting that the program was unable to open some
       * of the user specified files.
       */
    class FileNotFoundError : public std::runtime_error {
      
        /**
         * The FileNotFoundError constructor.
         *
         * This enables the program to abort or recover from faulty
         * filename (possibly supplied by the user).
         *
         * @param what More information of what went wrong.
         */
      public:
      FileNotFoundError( const char* what = "")
        : runtime_error(what) {}
    };

    
      /**
       * The default constructor for the CompartmentTable class.
       *
       * Makes sure we are aware of the fact that we do not have a
       * logging stream until we are assigned one using
       * setOptionLogFile().
       */
    CompartmentTable();

      /**
       * The CompartmentTable class destructor.
       *
       * Does nothing really.
       */
    ~CompartmentTable();

      /**
       * Attempt to read in all the compartments from an SBML model.
       *
       * This method will create the global species table and attempt
       * to construct all the compartments from the information in the
       * SBML model.  We will need the size of the (cubic) sub volumes
       * used in building the geometry.  The global parameter table is
       * needed because it is global and may be referenced anywhere.
       * The table of units is needed for both geometry construction
       * and further processing.
       *
       * @param m    The SBML information from which to extract the model.
       * @param p    The global parameter table.
       * @param u    The global unit table.  All units are always global.
       * @param size The side length of a cubic sub volume.
       * @param unit The unit of the sub volume size.
       * @param prng Pointer to the pseudo-random number generator.
       */
    std::vector<MesoRD::SubVolume*> fromSBML(
      const Model*       m,
      ParameterTable*    p,
      const UnitTable*   u,
      double             size,
      const std::string&  unit,
      Random::Random*    prng)
      throw(EmptyGeometry, ModelError, NoId);


      /**
       * Change the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline double
    changeSpecies(off_t index, double change)
      {
        return speciesTable.changeAmountByIndex(index, change);
      }


      /**
       * Change the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline unsigned int
    changeSpecies(const off_t index, const int change)
      {
        return speciesTable.changeAmountByIndex(index, change);
      }


      /**
       * Get the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the sub volume.
       */
    inline unsigned int getSpecies(off_t index) const
      { return speciesTable.getAmountByIndex(index); }


      /**
       * Set the global number of a certain species to an arbitrary
       * value.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline unsigned int setSpecies(off_t index, unsigned int number)
      { return speciesTable.setAmountByIndex(index, number); }


      /**
       * Get the global number of a certain species.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the sub volume.
       */
    inline double getSpeciesDouble(off_t index) const
      { return speciesTable.getAmountByIndexDouble(index); }


      /**
       * Set the global number of a certain species to an arbitrary
       * value.
       *
       * This call will effect the global species table.  Note that no
       * checking is performed on the index and the range.  If the
       * index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline double setSpecies(off_t index, double number)
      { return speciesTable.setAmountByIndex(index, number); }

    
    inline std::list<Compartment*>* getCompartments()
      { return &compartments; }
    
      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing and
       * model building begins, any error reporting and logging
       * information will be silently lost.  It is thus highly
       * recommended to call this method prior to calling fromSBML().
       * Note also that this method will set up the logging stream for
       * any member classes that can make use of one.
       *
       * @param l The logging stream onto which to report errors, etc.
       */
    void setOptionLogfile(std::ostream* l);

    void startCheckpointing(std::ostream& os);
    
    
      /**
       * Print a summary of the global species table.
       *
       * This method will print a summary of the global species table
       * to the supplied stream.  The summary contains how many
       * molecules there are of each species in all compartments.
       * This is the same as the number of molecules of each species
       * in the entire system.  The format of the iteration number
       * deserves a little comment, perhaps.  It is an array of long
       * integers.  The value at position i in the array gives a
       * number of multiples of 10^(i).  The total number of
       * iterations is the sum of all these numbers.  Furthermore, if
       * let the index of the largest non-zero entry be i_max.  Then
       * all entries with i < i_max should lie in the range
       * [0,9]. This makes outputting the array as a human-readable
       * number really easy.
       *
       * @param os        The stream onto which to write.
       * @param iteration The iteration number of the checkpoint.
       * @param time      The simulation time of the checkpoint.
       */
    void checkpointGlobal(
      std::ostream& os, std::valarray<size_t>& iteration, double time) const;

      /**
       * Same as checkpointGlobal, but for deterministic simulation.  
       * 
       * This will do a global checkpoint for the deterministic 
       * simulation using the currentState vector instead of the state
       * stored in the subvolumes.  
       *
       * @param os        The stream onto which to write.
       * @param iteration The iteration number of the checkpoint.
       * @param time      The simulation time of the checkpoint.
       * @param subVolumes The list of subvolumes for the system.  
       * @param currentState The current state-vector of the system.  
       */
    void checkPointGlobalDeterministic(
      std::ostream& os, std::valarray<size_t>& iteration, double time,
      std::vector<SubVolume*>& subVolumes, 
      MV_Vector_double& currentState) const;

      /**
       * Output a local checkpoint.
       *
       * Writes a formatted line to each of the species files.  For
       * every species file, the line contains numbers separated by
       * whitespace.  Each number gives the number of the particular
       * species in the particular sub volume.  These sub volumes are
       * ordered the same way the are ordered in the sub volume
       * vector.
       *
       * @param subVolumes The currently valid vector of sub volumes.
       */
    void checkpointLocal(std::vector<SubVolume*>& subVolumes);

      /**
       * Output a local checkpoint for the deterministic simulations.  
       *
       * Does the same thing as localCheckPoint except that this
       * function uses the currentState vector instead of the state saved
       * in the subvolumes.  
       */
    void checkpointLocalDeterministic(std::vector<SubVolume*>& subVolumes,
       MV_Vector_double& currentState);

      /**
       * Closes all locally opened checkpoint files.
       *
       * Close all files previously opened by a call to
       * openLocalFiles().
       */
    void closeLocalFiles();

      /**
       * Open local checkpoint files.
       *
       * Open files for storing the amount of each species in each sub
       * volume.  There will be one file for every species in the
       * model.  The prefix is prepended to all the file names.  The
       * idea is to put path information into the prefix argument.
       *
       * @param prefix A string that is prepended to all file names.
       */
    void openLocalFiles(std::string& prefix);

      /**
       * Open old simulation files to use as an inital state.
       *
       * This will open the species_* files from an old simulation. These
       * files can be used to start an simulation from any state in an old
       * simulation. The files to bn opened are specified by the model.  
       *
       * @param prefix A string that prepended to all file names.
       * @return A map of all the files found.  
       */
    std::map<int, std::ifstream*> openSpeciesFiles(std::string& prefix)
      throw( FileNotFoundError );

      /**
       * Closes the files opened by openSpeciesFileRead.
       *
       * This will close all the files that are given in the argument.
       *
       * @param files The files to close.
       */
    void closeSpeciesFiles(std::map<int, std::ifstream*>& infiles_species);

      /**
       * This will set the flag to enable the compartments
       * to output the subvolumes in a sparse format.
       *
       * @param outputSparse The sparse flag.
       */
    inline void setOptionOutputSparse(bool outputSparse)
      { this -> outputSparse = outputSparse; }
    
  };
}
#endif /* ifndef MESORD_COMPARTMENTTABLE_HPP */
