/* The CubicSubVolume class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CubicSubVolume.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef MESORD_CUBICSUBVOLUME_H
#define MESORD_CUBICSUBVOLUME_H 1

#include <valarray>
#include <map>
#include <vector>

#include "SubVolume.hpp"
#include "csg/CSGPoint.hpp"



namespace MesoRD
{

    /**
     * @brief Class for modelling cubic subvolumes.
     *
     * A cubic subvolume is a special case of the more general
     * subvolume.  The size of a cubic subvolume is the side length of
     * the cube.  In MesoRD, most of the simulation algorithm is
     * independent of the particular geometry of the individual
     * subvolumes.  The single exception is diffusion.  Therefore, the
     * main purpose of this class is to implement diffusion for cubic
     * subvolumes.  In order to handle diffusion, we must keep track
     * of where molecules in a subvolume go when they diffuse out of
     * it.  To this end, the class has an array pointing to the six
     * neighbours of the cube.
     *
     * Subvolumes on the boundary of the geometry are special, as they
     * may have fewer than six <em>proper</em> neighbours.  In these
     * cases, the pointers that would point to nowhere must be set up
     * to loop back onto the subvolume itself.  Boundary subvolumes
     * are in fact their own neighbours: the left neighbour of a left
     * boundary subvolume is the subvolume itself.  Thus, if a
     * molecule diffuses to the left from a left boundary subvolume,
     * it effectively "bounces" on the boundary.
     */
  class CubicSubVolume : public SubVolume
  {

  public:

      /**
       * @brief An enumeration identifying the directions of the
       * neighbours of a cubic subvolume.
       *
       * A cubic subvolume always has six neighbours.  These are
       * called <code>down</code>, <code>east</code>,
       * <code>north</code>, <code>south</code>, <code>west</code> and
       * <code>up</code>.  Using this enumeration, we can refer them
       * to in a more intuitive way using these names, thereby
       * avoiding the hassle of numeric indices.  Because the number
       * of neighbours is fixed, we should never need to resize the
       * array containing the neighbours of a subvolume.
       */
    typedef enum {

        /**
         * This direction points to the neighbour in the negative
         * <code>z</code>-direction
         */
      down,

        /**
         * This direction points to the neighbour in the positive
         * <code>x</code>-direction.
         */
      east,

        /**
         * This direction points to the neighbour in the positive
         * <code>y</code>-direction.
         */
      north,

        /**
         * This direction points to the neighbour in the negative
         * <code>y</code>-direction.
         */
      south,

        /**
         * This direction points to the neighbour in the negative
         * <code>x</code>-direction.
         */
      west,

        /**
         * This direction points to the neighbour in the positive
         * <code>z</code>-direction.
         */
      up

    } NeighbourDirection;


  public:

      /**
       * @brief CubicSubVolume class default constructor.
       *
       * The default constructor will initialise the pointers to the
       * neighbours.  Initially, the neighbour pointers are pointing
       * to <code>this</code>.  Doing this will not only ensure that
       * each new subvolume has no neighbours but also avoids the
       * problems arising from setting the pointers to
       * <code>null</code>.  The diffusion vector is not updated by
       * the constructor.  Thus an instance of the CubicSubVolume
       * class is not ready for use until setSpeciesNumber() has been
       * called.  Note that the number of neighbours is hardcoded into
       * the constructor: a cubic subvolume will <em>always</em> have
       * exactly six neighbours.
       *
       * @param prng    Pointer to the pseudo-random generator.
       * @param size    The length of one side of the cube, given in
       *                centimetres.
       * @param species The number of different species in the
       *                subvolume.
       */
    inline CubicSubVolume(
      Random::Random*      prng,
      const double         size,
      const unsigned short species = 0)
      : SubVolume(prng, size, 6, species)
      {}


      /**
       * @brief CubicSubVolume destructor.
       *
       * Does just about as much as the extent to which you can run
       * Windows on a MIPS processor.  In particular, destructing a
       * subvolume will not automagically destruct its neighbours.
       */
    inline ~CubicSubVolume()
      {}


      /**
       * @brief Make an addition to the diffusion vector for a given
       * species.
       *
       * This function allows other subvolumes to modify the diffusion
       * vector of the current subvolume.  In particular, neighbours
       * can make addition to the diffusion vector for a particular
       * species.  The diffusion vector in deterministic simulation is
       * treated as the total change in each direction.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * @param index The direction in which the diffusion should be
       *              added.
       * @param rate  The diffusion rate to add.
       */
//     inline void
//     addDiffusionRate(const off_t index, const double rate)
//       {
//         diffusion_vector[index] += rate;
//       }


      /**
       * @brief Clear all the values in the diffusion vector.
       *
       * This function will set the entire diffusion vector to zero.
       */
//     inline void
//     clearDiffusionVector()
//       {
//         diffusion_vector = 0.0;
//       }


      /**
       * @brief Compute a diffusion event in the subvolume.
       *
       * A diffusion event depends on the precise geometry of the
       * subvolume.  This function will, given a random number in the
       * interval [0, 1], figure out what species diffuses where,
       * perform the diffusion event and return a pointer to the
       * diffusion target.  Performing the diffusion means decreasing
       * the number of a certain species in the source subvolume and
       * increasing the number of the same species in the target
       * subvolume.
       *
       * @param  rand A random number between zero and one inclusive.
       * @return      A pointer to the target of the diffusion.
       */
    SubVolume*
    eventDiffusion(const double rand);


      /**
       * @brief Get the vector containing the rate of diffusion in all
       * directions.
       *
       * FIXME!! Hattne doesn't understand this!
       * FIXME!! I don't understand "in all directions".
       *
       * @return The total diffusion rate for each species.
       */
//     inline std::valarray<double>&
//     getDiffusionVector()
//       {
//         return diffusion_vector;
//       }


      /**
       * @brief Set the elements of the diffusion matrix.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * @param index the index of the species.
       *
       * @param subVolumeMap Map between CSG points of a subvolume and
       * its index.
       *
       * @param diffusionMatrixData Where the data points for the
       * diffusion matrix will be stored.
       *
       * @param diffusionMatrixRowIndex row index for data in
       * diffusion matrix.
       *
       * @param diffusionMatrixColIndex column index in diffusion
       * matrix.
       */
    void
    getDiffusionVector(
      off_t                             index,
      std::map<CSGPoint, unsigned int>& subVolumeMap,
      std::vector<double>&              diffusionMatrixData,
      std::vector<size_t>&              diffusionMatrixRowIndex,
      std::vector<size_t>&              diffusionMatrixColIndex,
      int                               numberOfSpecies);


      /**
       * @brief Get a pointer to the @p i:th neighbour of the
       * subvolume.
       *
       * Returns a pointer to the @p i:th neighbour of the subvolume.
       * If this is <code>null</code>, then we are in trouble.
       *
       * @param  i The index of the neighbour to get.
       * @return   A pointer to the requested neighbour.
       */
    inline CubicSubVolume*
    getNeighbour(const size_t i) const
      {
        return getNeighbour(static_cast<NeighbourDirection>(i));
      }

      /**
       * @brief Returns a reference to the array of neighbour subvolumes.  
       * 
       */
    inline std::valarray<SubVolume*>& getNeighbourArray()
      { return neighbours_array; }
    

      /**
       * @brief Get a pointer to the neighbouring subvolume in the
       * requested direction.
       *
       * Returns a pointer to the neighbour in the requested
       * direction.  If this is <code>null</code>, then we are in
       * trouble.
       *
       * @param  d The requested direction.
       * @return   A pointer to the requested neighbour.
       */
    inline CubicSubVolume*
    getNeighbour(const NeighbourDirection d) const
      {
        return static_cast<CubicSubVolume*>(neighbours_array[d]);
      }


      /**
       * @brief Set the pointer to the @p i:th neighbour of the
       * subvolume.
       *
       * This function will set or update the neighbour in the @p i:th
       * direction.  Note that a cubic subvolume can only have
       * neighbours that are cubic subvolumes themselves.  Note also
       * that for boundary subvolumes, empty neighbour pointers should
       * point back to the subvolume itself.
       *
       * @param  i The index of the neighbour to set.
       * @param  n A pointer to the new neighbour.
       */
    inline void
    setNeighbour(const size_t i, SubVolume* n)
      {
        CubicSubVolume* csv = dynamic_cast<CubicSubVolume*>(n);
        if (csv)
          neighbours_array[i] = csv;
      }


      /**
       * @brief Set the neighbour of the subvolume in the requested
       * direction.
       *
       * This function will set or update the neighbour in the
       * requested direction.  Note that a cubic subvolume can only
       * have neighbours that are cubic subvolumes themselves.  Note
       * also that for boundary subvolumes, empty neighbour pointers
       * should point back to the subvolume itself.
       *
       * @param d The requested direction.
       * @param n A pointer to the new neighbour.
       */
    inline void
    setNeighbour(const NeighbourDirection d, CubicSubVolume* n)
      {
        neighbours_array[d] = n;
      }


      /**
       * @brief Set the number of species the subvolume can hold.
       *
       * It is crucial that this function is called prior to
       * simulation use of the instance.  The reason is that calling
       * this function will initialise the diffusion vector which
       * holds precomputed diffusion rates for each species in each
       * direction.  If this is not done, strange things will happen,
       * and the complete and utter collapse of the universe as we
       * know it may, not be as distant as we once thought.  It is
       * important that the subvolume's compartment has been set up
       * prior to calling setSpeciesNumber(), as the compartment holds
       * the numerical diffusion rates we need to obtain.
       *
       * @param n The number of species the subvolume will be able to
       *          hold.
       */
    inline void
    setSpeciesNumber(const size_t n)
      {
        SubVolume::setSpeciesNumber(n);
        diffusion_vector.resize(n);
        for (size_t i = 0; i < species.integer -> size(); i++) {
          diffusion_vector[i] = 0.0;
          for (size_t j = 0; j < neighbours; j++) {
            diffusion_vector[i] += compartment
              -> getDiffusion(i,
                neighbours_array[j] -> getCompartment() -> getId());
          }
        }
      }


      /**
       * @brief Update the diffusion constant for the cubic subvolume.
       *
       * This will update the R-value for diffusion events.  It should
       * be called whenever the chemical composition of the subvolume
       * is changed.  Note that no stochasticity is involved in this
       * process.  We try to make this call as cheap as possible by
       * using the diffusion vector computed previously, because this
       * function will be called quite often.
       *
       * It may be tempting to make use of the multiplication operator
       * of valarrays or the inner_product() function from the C++
       * numerics header.  Unfortunately, that will not work, since
       * the two valarrays here are not of the same type.
       */
    inline void
    updateDiffusionConstants()
      {
        double D_tot = 0.0;
        for (size_t i = 0; i < species.integer -> size(); i++)
          D_tot += diffusion_vector[i] * (*species.integer)[i];
        R_diffusion = D_tot / (size * size);
      }

      /**
       * Updates the reaction propensities for the reactions that are
       * defined to span multiple subvolumes. 
       *
       * Here this means that when the reaction radius may be about
       * the same size, or larger, the subvolume size, the reactions
       * may occure over the subvolume boundaries.  
       *
       * First all the neighbours are updated, and finally the
       * reaction propensities for the current subvolume is updated.
       */
    inline void
    updateNeighbourReactionConstants() 
      {
        R_neighbour_reaction = 0; 
        for( size_t i = 0;  i < neighbours_array.size(); i++) {
          if( neighbours_array[i] -> getSpeciesArray() -> sum() != 0)
            R_neighbour_reaction += 
              compartment -> reactionTable.getRates(*getSpeciesArray(), *(neighbours_array[i] -> getSpeciesArray()));
        }
        
        R_neighbour_reaction += (compartment -> reactionTable.getRates(*getSpeciesArray(), *getSpeciesArray()));
        R_neighbour_reaction /= ((double) (neighbours_array.size() + 1));
      }
    
    
      /**
       * @brief Update the diffusion vector.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * This will calculate the flow out from the current subvolume
       * for each species.  Remove this from the current diffusion
       * vector and add it to the correct neighbour.
       */
//     void
//     updateDiffusionVectorDeterministic();
  };
}
#endif /* ifndef MESORD_CUBICSUBVOLUME_H */
