/* The Derivative class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Derivative.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include <stdexcept>
#include <valarray>
#include <vector>

#include "sys/types.h"

#include "SubVolume.hpp"
#include "Derivative.hpp"
#include "linalg/Vector.hpp"

#include "csg/CSGPoint.hpp"

#include "sparselib/comprow_double.hpp"
#include "sparselib/compcol_double.hpp"
#include "mv/mvvd.hpp"

MesoRD::Derivative::Derivative(std::vector<SubVolume*>& subVolumes, 
  CompartmentTable* compartmentTable)
{
  
  this -> subVolumes = subVolumes;
  this -> compartmentTable = compartmentTable;

  tempState = std::valarray<double>((double)0,0);
  
}

MesoRD::Derivative::~Derivative()
{}

MV_Vector_double
MesoRD::Derivative::getCurrentState() const
{
  unsigned int j;
    //the system should have atleast one subvolume so we are fine doing this.
  if(subVolumes.size() > 0) {
    
    unsigned long speciesNumber = subVolumes[0] -> getSpeciesNumber();
    
    MV_Vector_double 
      theState(static_cast<int>(subVolumes.size() * speciesNumber));
    
    for(unsigned int i = 0; i < subVolumes.size(); i++)
    {
      for(j = 0; j < speciesNumber; j++)
      {
        theState(static_cast<int>(i * speciesNumber + j)) =
          subVolumes[i] -> getSpeciesDouble(j);
      }
    }
    return theState;
  } else {
    return MV_Vector_double(0);
  }
  
}

void
MesoRD::Derivative::setCurrentState(MV_Vector_double& theState)
{
  unsigned int j;
  unsigned int i;

  unsigned long speciesNumber = subVolumes[0] -> getSpeciesNumber();
  
  std::valarray<double> totalNrOfSpecies((double) 0, speciesNumber);
  
  SubVolume* tempSubVolume;
  
  

  for(i = 0; i < subVolumes.size(); i++)
  {
    tempSubVolume = subVolumes[i];
    
    for(j = 0; j < speciesNumber; j++)
    {
      tempSubVolume
	-> setSpecies(j, theState(static_cast<int>(i * speciesNumber + j)));
        //tempSubVolume -> clearDiffusionVector();

      totalNrOfSpecies[j]
	+= theState(static_cast<int>(i * speciesNumber + j)); 
    }
  }

  for(i = 0; i< speciesNumber; i++)
  {
    compartmentTable -> setSpecies(i, totalNrOfSpecies[i]);
  } 
}

void
MesoRD::Derivative::getReactionDerivative(MV_Vector_double& theState,
  MV_Vector_double& theDerivative)
{
  
  unsigned int i;
  unsigned long speciesNumber = subVolumes[0] -> getSpeciesNumber();

  for(i = 0; i < subVolumes.size(); i++)
  {
    subVolumes[i] -> getDerivativeVector(theState, 
      theDerivative, i*speciesNumber, tempState); 
  }
}

CompRow_Mat_double
MesoRD::Derivative::getDiffusionMatrix()
{   
  
  std::map<CSGPoint, unsigned int> subVolumeMap;
  unsigned int i;
  unsigned int j;
  size_t speciesNumber = subVolumes[0] -> getSpeciesNumber();
  
  for(i = 0; i < subVolumes.size(); i++)
  {
    subVolumeMap[CSGPoint(
                   subVolumes[i] -> getX(),
                   subVolumes[i] -> getY(),
                   subVolumes[i] -> getZ(), 0.5, 0.5, 0.5)] = i;
  }
  
  
  std::vector<double> diffusionMatrixData;
  std::vector<size_t> diffusionMatrixColIndex;
  std::vector<size_t> diffusionMatrixRowIndex;
  
  
  for(i = 0; i < subVolumes.size(); i++)
  {
    for(j = 0; j < speciesNumber; j++)
    { 
      subVolumes[i] -> getDiffusionVector(j, subVolumeMap,diffusionMatrixData, 
        diffusionMatrixRowIndex, diffusionMatrixColIndex, static_cast<int>(speciesNumber));      
    }
  }
  
  diffusionMatrixColIndex.push_back(diffusionMatrixRowIndex.size());
  
  double* data   = new double[diffusionMatrixData.size()];
  int* row_index   = new int[diffusionMatrixRowIndex.size()];
  int* col_ptr = new int[diffusionMatrixColIndex.size()];

  int nonzeros = 0;
  
  for(i = 0; i < diffusionMatrixData.size(); i++){
    data[i]      = diffusionMatrixData[i];
    nonzeros++;
  }
  
  for(i = 0; i < diffusionMatrixRowIndex.size(); i++)
    row_index[i]   = static_cast<int>(diffusionMatrixRowIndex[i]);
  for(i = 0; i < diffusionMatrixColIndex.size(); i++)
    col_ptr[i] = static_cast<int>(diffusionMatrixColIndex[i]);


  return CompRow_Mat_double(CompCol_Mat_double(
    static_cast<int>(speciesNumber * subVolumes.size()),
    static_cast<int>(speciesNumber * subVolumes.size()),
    nonzeros,
    data,
    row_index,
    col_ptr));
}
