/* The Derivative class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Derivative.hpp 553 2008-05-09 13:28:22Z davidfange $
 */
#ifndef MESORD_DERIVATIVE_HPP
#define MESORD_DERIVATIVE_HPP 1

#include <stdexcept>
#include <valarray>

#include "sys/types.h"

#include "SubVolume.hpp"
#include "CompartmentTable.hpp"

#include "sparselib/compcol_double.hpp"
#include "mv/mvvd.hpp"

namespace MesoRD
{
  
  class SubVolume;
  class CompartmentTable;
  
    /**
     * Class for calculating derivative from the system description.  
     */
  class Derivative 
  {
    
  private:
    std::vector<SubVolume*> subVolumes;
    CompartmentTable* compartmentTable;

    std::valarray<double> tempState;
            
  public:
    
      /** 
       * Default constructor for the Derivative class.  
       * 
       * Does nothing and shall not be used.  
       */
    inline Derivative(){compartmentTable=0;};
    
      /**
       * Value constructor for the Derivative class.  
       *
       * The destructor will set a refernce to the a the subVolumes
       * vector and to the compartmentTable to the system.
       *
       * @param subVolumes A reference to all the subvolumes in the
       * system.  
       * @param compartmentTable A reference to the
       * compartmentTable of the system.
       */
    Derivative(std::vector<SubVolume*>& subVolumes, 
      CompartmentTable* compartmentTable);
    
      /**
       * Destructor for the Derivative class.  
       * 
       * Does nothing.  
       */
    ~Derivative();
    
      /**
       * Get the current state of all the species in all subvolumes.  
       *
       * This will return the number of species in all subvolumes.
       * Can for example be used to get
       * the initial statevector for the whole system before the
       * deterministic simulation starts.  
       * 
       * @return The complete state for all species in all subvolumes.  
       */
    MV_Vector_double getCurrentState() const;
    
      /**
       * Sets a new state for the system.  
       * 
       * This will set the number of all species in all subvolumes
       * accordning the variable theState.  
       * 
       * @param theState The definition of the new state.  
       */
    void setCurrentState(MV_Vector_double& theState);
    
      /** Gets the derivative for the system defined by theState.  
       *
       * This will calculate the derivative for the state defined by
       * theState and store it in derivative.  
       *
       * @param theState The state for which the derivarive will be
       * calculated.
       *
       * @param derivative Used to store the derivative.
       */
    void getReactionDerivative(MV_Vector_double& theState,
      MV_Vector_double& theDerivative);
      
      /**
       * Will ask the subvolumes for the diffusionconstants  to their 
       * neighbours, and then put all of these in a matrix to allow
       * easy update of the statevector due to diffusion.  
       *
       * @return A spare matrix constaining the diffusion for all
       *         speices in all subvolumes.
       */
    CompRow_Mat_double getDiffusionMatrix();

  };
}

#endif
