/* The OctahedricSubvolume class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: OctahedricSubvolume.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <iostream>
#include <map>
#include <vector>

#include "Compartment.hpp"
#include "OctahedricSubvolume.hpp"
#include "csg/CSGPoint.hpp"



  /*************
   * DIFFUSION *
   *************/

  /* We now know that a neighbour can never be zero.  If a subvolume
   * lacks a neighbour in a certain direction, the neighbour must be
   * set up to point to itself.  Should a neighbour nevertheless be
   * <code>null</code>, it is an error which will lead to a
   * segmentation fault.  We do not try to catch that here, because
   * the error is severe enough to motivate complete failure of the
   * program.  Note that we want to ensure that diffusions that have
   * zero probability of happening do not actually happen.
   */
MesoRD::SubVolume*
MesoRD::OctahedricSubvolume::eventDiffusion(const double rand) 
{
  if (R_diffusion > 0) {
    double       D_cum    = 0;
    const double rand_tot = rand * R_diffusion * (size * size);

    for (size_t i = 0; i < species.integer -> size(); i++) {
      const double s = 1.0 * (*species.integer)[i];

      if (s > 0) {
        for (size_t j = 0; j < neighbours; j++) {
          const double D  = s * compartment -> getDiffusion(i,
            neighbours_array[j] -> getCompartment() -> getId());
          D_cum          += D;

          if (D > 0 && rand_tot <= D_cum) {
            OctahedricSubvolume* target = static_cast<OctahedricSubvolume*>(
              neighbours_array[j]);

            changeSpecies(i, int(-1));
            target -> changeSpecies(i, int(+1));
            return target;
          }
        }
      }
    }
  }
  return 0;
}


  /* Looping over all subvolumes calling this function for each
   * species, you will be able to extract a matrix from
   * diffusionMatrixData, diffusionMatrix(Row/Col)Index.  The vectors
   * can be seen as a coordinate compressed sparse matrix.
   */
void
MesoRD::OctahedricSubvolume::getDiffusionVector(off_t index,
  std::map<CSGPoint, unsigned int>& subVolumeMap,
  std::vector<double>& diffusionMatrixData,
  std::vector<size_t>& diffusionMatrixRowIndex,
  std::vector<size_t>& diffusionMatrixColIndex,
  int numberOfSpecies)
{
  double temp_diffusion_rate = 0.0;
  double acumulatedDiffusionRate = 0.0;

  double sizeSquared = size * size;
  int subVolumeIndex = subVolumeMap[CSGPoint(
                                      this -> getX(),
                                      this -> getY(),
                                      this -> getZ(), 0.5, 0.5, 0.5)];
  int neighbourIndex = -1;

  diffusionMatrixColIndex.push_back(diffusionMatrixRowIndex.size());

    /* Looping over all the neighbours and picking out their indices.
     * Theses indices are used to construct the diffusion matrix.
     */
  for (size_t j = 0; j < neighbours; j++) {
    neighbourIndex = subVolumeMap[
      CSGPoint(neighbours_array[j] -> getX(),
        neighbours_array[j] -> getY(),
        neighbours_array[j] -> getZ(), 0.5, 0.5, 0.5)];

      /* Only non-self neighbours make a contribution to the diffusion
       * matrix.
       */
    if(subVolumeIndex != neighbourIndex) {
      temp_diffusion_rate = (compartment
        -> getDiffusion(index,
          neighbours_array[j] -> getCompartment() -> getId())) / sizeSquared;

      acumulatedDiffusionRate -= temp_diffusion_rate;

      diffusionMatrixData.push_back(temp_diffusion_rate);
      diffusionMatrixRowIndex.push_back(
        neighbourIndex * numberOfSpecies + index);
    }
    neighbourIndex = -1;
  }

  diffusionMatrixRowIndex.push_back(subVolumeIndex * numberOfSpecies + index);
  diffusionMatrixData.push_back(acumulatedDiffusionRate);
}
