/* The Parameter class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Parameter.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <stdlib.h>

#include <iostream>

#include "Parameter.hpp"
#include "UnitTable.hpp"
#include "ui/OPrefixStream.hpp"



  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* The constructor.  Do a deep copy of the data.  Remember to copy
   * the trailing null character as well
   */
MesoRD::Parameter::Parameter(const std::string& i, const std::string& n,
  double v, const Unit& u, bool k) 
{
  this -> constant = k;
  this -> id       = i;
  this -> name     = n;
  this -> unit     = u;
  this -> value    = v;
}


  /* Copy constructor.  We must do deep copy.  However, we can use the
   * default constructor to do this.
   */
MesoRD::Parameter::Parameter(const Parameter& p) 
{
  this -> constant = p.constant;
  this -> id       = p.id;
  this -> name     = p.name;
  this -> unit     = p.unit;
  this -> value    = p.value;
}


  /* Just read in the parameter from the SBML definition.
   */
MesoRD::Parameter::Parameter(const ::Parameter* p, const UnitTable* u) 
{
  fromSBML(p, u);
}


  /* Destructor. Do-nothing.
   */
MesoRD::Parameter::~Parameter() 
{}



  /****************
   * SBML SUPPORT *
   ****************/

  /* Check whether the respective fields are set in the SBML.  If they
   * are, get their values, otherwise set some defaults.
   */
void
MesoRD::Parameter::fromSBML(const ::Parameter* p, const UnitTable* ut)
{
  Unit *up;

  constant = p -> getConstant();
  id       = p -> getId();
  name     = p -> getName();

  if (ut && p -> isSetUnits() && (up = ut -> getUnit( p -> getUnits()))) {
    unit = *up;
  } else {
    unit = Unit::dimensionless();
  }

  if ( p -> isSetValue() )
    value = Parameter_getValue(p);
  else
    value = 0;
}



  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML.  Unlike SBML, a parameter
   * must have a value.
   */
namespace MesoRD {
  std::ostream& operator<< (std::ostream& os, const Parameter& p) {
    OPrefixStream opfs_level0(os.rdbuf(), "");

    opfs_level0 << "<parameter id=\"" << p.id << "\"" << std::endl;
    
    if (p.name.length() > 0) {
      opfs_level0
        << "           name=\"" << p.name << "\"" << std::endl;
    }
    
    opfs_level0
      << "           value=\""
      << p.value << "\"" << std::endl;

    if ((p.unit.getId()).length()) {
      opfs_level0
        << "           units=\""
        << p.unit.getId() << "\"" << std::endl;
    }

    opfs_level0
      << "           constant=\""
      << (p.constant ? "true" : "false") << "\"/>" << std::endl;

    return os;
  }
}

