/* The Parameter class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Parameter.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_PARAMETER_H
#define MESORD_PARAMETER_H 1

#include <iostream>
#include <string>

#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include "Unit.hpp"

namespace MesoRD {

  class UnitTable;


    /**
     * Class for parameters.
     *
     * In MesoRD a parameter must have a value.  If no value is given
     * in the definition of the model, the value is assumed to be
     * zero.  This is different from SBML, where parameters do not
     * need to have a defined value.  Just as in SBML parameters need
     * not be constant.  Things are a lot easier if parameters are
     * constant.  Therefore, if a parameter is not a constant, this
     * class needs to be explicitely notified -- parameters are
     * assumed to be constant by default.  Every parameter also has a
     * unit associated with it.
     */
  class Parameter {
  private:
    bool        constant;
    std::string id;
    std::string name;
    Unit        unit;
    double      value;

  public:


      /**
       * Default constructor for the Parameter class.
       *
       * Creates a new parameter given the value supplied.  This
       * method really does not do any magic at all.
       *
       * @param i The id of the new parameter.
       * @param n The name of the new parameter.
       * @param v The value of the new parameter.
       * @param u The unit of the new species.
       * @param k Whether the new parameter is constant or not.
       */
    Parameter(
      const std::string& i = "",
      const std::string& n = "",
      double             v = 0,
      const Unit&        u = Unit::dimensionless(),
      bool               k = true);

    /**
     * Copy constructor for the Parameter class.
     *
     * Creates a copy of the given instance of the Parameter class.
     * This method really does not do any magic at all.
     *
     * @param p The parameter to be copied.
     */
    Parameter(const Parameter& p);

      /**
       * Value constructor for the Parameter class.
       *
       * This constructor will read the definition of the parameter
       * from the SBML model.  The method uses fromSBML() internally.
       *
       * @param p The SBML parameter type to construct from.
       */
    Parameter(const ::Parameter* p, const UnitTable* u);

      /**
       * Destructor for the Parameter class.
       *
       * Does nothing at all.
       */
    ~Parameter();

      /**
       * Read the parameter from an SBML definition.
       *
       * This method will take all the values set in an SBML
       * definition and copy them to the current instance of the
       * Parameter class.  Some fields are optional in SBML.  For such
       * fields the method makes sure to use default values instead.
       *
       * @param p The SBML parameter type to extract values from.
       */
    void fromSBML(const ::Parameter* p, const UnitTable* u);

      /**
       * Checks whether a parameter is constant.
       *
       * Checks whether a parameter is constant or not.  Parameters
       * are constant by default.
       *
       * @return True if the parameter is constant, false otherwise.
       */
    inline bool isConstant() const
      { return constant; }

      /**
       * Checks whether a parameter has a certain name.
       *
       * Tests for equality on the name.  If the name was not set,
       * this method will return false.
       *
       * @param n The name to check for equality.
       * @return  True if the names match, false otherwise.
       */
    inline bool isName(const std::string& n) const
      { return ((name.length() > 0 ) && (name == n)); }

      /**
       * Checks whether a parameter has a certain id.
       *
       * Tests for equality on the id.  If the id was not set, this
       * method will return false.
       *
       * @param i The id to check for equality.
       * @return  True if the id:s match, false otherwise.
       */
    inline bool isId(const std::string& i) const
      { return ((id.length() > 0) && (id == i)); }
    
      /**
       * Get the name of a parameter.
       *
       * Returns the name of a parameter.
       *
       * @return The name of the parameter.
       */
    inline const std::string& getName() const
      { return name; }

      /**
       * Get the id of a parameter.
       *
       * Returns the id of a parameter.
       *
       * @return The id of the parameter.
       */
    inline const std::string& getId() const
      { return id; }

      /**
       * Get the unit of a parameter.
       *
       * Returns the unit of a parameter.
       *
       * @return The unit of the parameter.
       */
    inline const Unit& getUnit() const
      { return unit; }

      /**
       * Get the value of a parameter.
       *
       * Returns the value of a parameter.
       *
       * @return The value of the parameter.
       */
    inline double getValue() const
      { return value; }

      /**
       * Set the whether a parameter should be regarded constant or
       * not.
       *
       * Changes the constant state of the parameter.  Parameters in
       * SBML are constant by default.
       *
       * @param k True if the parameter is constant, false otherwise.
       */
    inline void setConstant(bool k = true)
      { constant = k; }

      /**
       * Set the name of a parameter.
       *
       * Changes the name of a parameter.  A parameter in SBML need
       * not have a name.
       *
       * @param n The new name of the parameter.
       */
    inline void setName(const std::string& n)
      { name = n; }

      /**
       * Set the id of a parameter.
       *
       * Changes the id of a parameter.  A parameter in SBML must have
       * an id.  It seems that this is currently not enforced in
       * libsbml.
       *
       * @param i The new id of the parameter.
       */
    inline void setId(const std::string& i)
      { id = i; }

      /**
       * Set the unit of a parameter.
       *
       * Changes the unit of a parameter.  A parameter in SBML need
       * not have a unit.
       *
       * @param u The new unit of the parameter.
       */
    inline void setUnit(const Unit& u)
      { unit = u; }

      /**
       * Set the value of a parameter.
       *
       * Changes the value of a parameter.  In SBML the value of a
       * parameter is optional.  In MesoRD it is not.  Note that it is
       * always possible to set the value of a parameter using this
       * method, even if the parameter should be a constant.
       *
       * @param v The new value of the parameter.
       */
    inline void setValue(double v)
      { value = v; }
    
      /**
       * Put a parameter to an output stream.
       *
       * This method will put a parameter onto an output stream.  The
       * result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param p  The parameter in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (std::ostream& os, const Parameter& p);
    
  };
}

#endif
