/* The ParameterTable class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ParameterTable.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <stdio.h>
//#include <sbml/util.h>

#include "ParameterTable.hpp"
#include "ui/OPrefixStream.hpp"



  /**********************
   * PARAMETER QUERYING *
   **********************/

  /* Find the parameter and check if it is constant.  Return false by
   * default.
   */
bool
MesoRD::ParameterTable::isConstant(const std::string& id) const
{
  std::list<Parameter>::const_iterator i;

  for (i = parameters.begin(); i != parameters.end(); i++) {
    if (i -> isId(id)) {
      return (i -> isConstant());
    }
  }
  return false;
}


  /* Either a parameter is defined locally or it is defined in the
   * parent parameter table.
   */
bool
MesoRD::ParameterTable::isDefined(const std::string& id) const
{
  if (isLocal(id))
    return true;
  else if (parent)
    return (parent -> isDefined(id));
  
  return false;
}


  /* Only check the local parameter table here.
   */
bool
MesoRD::ParameterTable::isLocal(const std::string& id) const
{
  std::list<Parameter>::const_iterator i;

  for (i = parameters.begin(); i != parameters.end(); i++) {
    if (i -> isId(id)) {
      return true;
    }
  }
  return false;
}



  /*****************************
   * GETTER AND SETTER METHODS *
   *****************************/

  /* Just do a normal addition of the parameter if it is not locally
   * present.  Otherwise, erase the previously existing copy and add a
   * new one.
   */   
void
MesoRD::ParameterTable::addReplaceParameter(
  const std::string& id,
  const std::string& name,
  double             value,
  const Unit&        unit,
  bool               constant)
{ 
  if (!isLocal(id)) {
    parameters.push_back(Parameter(id, name, value, unit, constant));
  } else {
    std::list<Parameter>::iterator i;

    for (i = parameters.begin(); i != parameters.end(); i++) {
      if (i -> isId(id)) {
        i -> setName(name);
        i -> setValue(value);
        i -> setUnit(unit);
        i -> setConstant(constant);
        return;
      }
    }
  }
}


  /* Just do a normal addition of the parameter if it is not locally
   * present.  Otherwise, erase the previously existing copy and add a
   * new one.
   */   
void
MesoRD::ParameterTable::addReplaceParameter(const Parameter& p)
{
  if (!isLocal(p.getId())) {
    parameters.push_back(p);
  } else {
    std::list<Parameter>::iterator i;

    for (i = parameters.begin(); i != parameters.end(); i++) {
      if (i -> isId(p.getId())) {
        i -> setName(p.getName());
        i -> setValue(p.getValue());
        i -> setUnit(p.getUnit());
        i -> setConstant(p.isConstant());
      }
    }
  }
}


  /* Return dimensionless by default.  Remember to check the parent
   * table.
   */
MesoRD::Unit
MesoRD::ParameterTable::getUnit(const std::string &id) const
{
  std::list<Parameter>::const_iterator i;

  for (i = parameters.begin(); i != parameters.end(); i++) {
    if (i -> isId(id)) {
      return (i -> getUnit());
    }
  }

  if (parent)
    return parent -> getUnit(id);
  
  return 0;
}


  /* Return zero by default.  Remember to check the parent table.
   */
double
MesoRD::ParameterTable::getValue(const std::string &id) const
{
  std::list<Parameter>::const_iterator i;

  for (i = parameters.begin(); i != parameters.end(); i++) {
    if (i -> isId(id)) {
      return (i -> getValue());
    }
  }

  if (parent)
    return parent -> getValue(id);
  
  return 0;
}


  /* Does nothing if the parameter could not be found.
   */
void
MesoRD::ParameterTable::setValue(const std::string &id, double value)
{
  std::list<Parameter>::iterator i;

  for (i = parameters.begin(); i != parameters.end(); i++) {
    if (i -> isId(id)) {
      i -> setValue(value);
      return;
    }
  }

  if (parent)
    parent -> setValue(id, value);
}



  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML.
   */
namespace MesoRD {
  std::ostream &operator<< (std::ostream &os, const ParameterTable &p) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "<listOfParameters>" << std::endl;
    for (std::list<Parameter>::const_iterator i = p.parameters.begin();
         i != p.parameters.end(); i++) {
      opfs_level1 << *i;
    }
    opfs_level0 << "</listOfParameters>" << std::endl;
    return os;
  }
}
