/* The ParameterTable class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ParameterTable.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef __ParameterTable_H
#define __ParameterTable_H

#include <iostream>
#include <list>

#include "Parameter.hpp"
#include "Unit.hpp"

namespace MesoRD {

    /**
     * Class for handling scoped parameters.
     *
     * Parameters in SBML can be defined in several scopes.  For
     * instance the kinetic laws that govern how the reactions work
     * can carry local parameter tables.  In order to handle this
     * scoping, a parameter can have a parent parameter table.  When
     * looking up parameters in a parameter table, the local table is
     * always consulted first.  Only if the requested operation fails
     * on the local table, is the parent table consulted.
     */
  class ParameterTable {
  private:
    std::list<Parameter> parameters;
    ParameterTable*      parent;

  public:
      /**
       * The default constructor for the ParameterTable class.
       *
       * A parameter can be initialised with a pointer to its parent
       * parameter table.  Calling this constructor will otherwise
       * make sure that we start out with a fresh table containing no
       * parameters at all.
       *
       * @param p A pointer to the parent parameter table.
       */
    inline ParameterTable(ParameterTable* p = 0)
      { parameters.clear(); parent = p; }
  
      /**
       * Destructor for the ParameterTable class.
       *
       * Does nothing.
       */
    inline ~ParameterTable()
      {}
    
      /**
       * Checks whether a parameter is constant.
       *
       * Checks to see whether the parameter identified by the given
       * id is a constant.  If the parameter is undefined, this method
       * returns false.
       *
       * @param id The id of the parameter to query.
       * @return   True if the parameter is constant, false otherwise.
       */
    bool isConstant(const std::string& id) const;
    
      /**
       * Check whether a parameter is defined.
       *
       * This method checks if the parameter identified by the given
       * id is defined.  Unlike islocal(), this method consults the
       * root parent parameter table as well, but only if the
       * parameter is not defined in locally.  If the parameter is
       * defined in neither table, then it is not defined at all.
       *
       * @param id The id of the sought parameter.
       * @return   True if the parameter is defined, false otherwise.
       */
    bool isDefined(const std::string& id) const;

      /**
       * Check whether a parameter is defined in the local table
       *
       * This method checks whether a parameter identified by the
       * given id is defined in the local parameter table.  The root
       * table is not consulted at all.
       *
       * @param id The id of the sought parameter.
       * @return   True if the parameter is defined, false otherwise.
       */
    bool isLocal(const std::string& id) const;

      /**
       * Add a parameter to the parameter table.
       *
       * This method will add a parameter to the parameter table.  If
       * there already exists a parameter with the given id, nothing
       * is done.
       *
       * @param id       The id of the parameter to add.
       * @param name     The name of the parameter to add.
       * @param value    The value of the parameter to add.
       * @param unit     The unit of the parameter to add.
       * @param constant Whether the new parameter is constant or not.
       */
    inline void addParameter(
      const std::string& id,
      const std::string& name,
      double             value,
      const Unit&        unit,
      bool               constant = true)
      { if (!isLocal(id))
        parameters.push_back(Parameter(id, name, value, unit, constant)); }
    
      /**
       * Add a parameter to the parameter table.
       *
       * This method will add a parameter to the parameter table.  If
       * there already exists a parameter with the given id, nothing
       * is done.
       *
       * @param p The parameter to add.
       */
    inline void addParameter(const Parameter& p)
      { if (!isLocal(p.getId())) parameters.push_back(p); }

      /**
       * Add or replace a parameter to the parameter table.
       *
       * This method will add a parameter to the parameter table.  If
       * there already exists a parameter with the given id, the old
       * parameter is replaced with the new one.
       *
       * @param id       The id of the parameter to add.
       * @param name     The name of the parameter to add.
       * @param value    The value of the parameter to add.
       * @param unit     The unit of the parameter to add.
       * @param constant Whether the new parameter is constant or not.
       */
    void addReplaceParameter(
      const std::string& id,
      const std::string& name,
      double             value,
      const Unit&        unit,
      bool               constant = true);

      /**
       * Add or replace a parameter to the parameter table.
       *
       * This method will add a parameter to the parameter table.  If
       * there already exists a parameter with the given id, the old
       * parameter is replaced with the new one.
       *
       * @param p The parameter to add.
       */
    void addReplaceParameter(const Parameter& p);

      /**
       * Get the unit of a parameter.
       *
       * One should really check that the parameter is defined before
       * calling this method.  If the parameter could not be found,
       * this method silently returns the dimensionless unit.  Note
       * that the root parameter table is consulted as well.
       *
       * @param id The id of the sought parameter.
       * @return   The unit of the parameter, or dimensionless if it is
       *           undefined.
       */
    Unit getUnit(const std::string& id) const;

      /**
       * Get the value of a parameter.
       *
       * One should really check that the parameter is defined before
       * calling this method.  If the parameter could not be found,
       * this method silently return zero.  Note that the root
       * parameter table is consulted as well.
       *
       * @param id The id of the sought parameter.
       * @return   The value of the parameter, or zero if it is undefined.
       */
    double getValue(const std::string& id) const;

      /**
       * Set the value of a parameter.
       *
       * The parameter must be defined before calling this method.  If
       * attempting to set the value of an undefined parameter,
       * nothing will happen.  Note that if a parameter is undefined
       * in the current parameter table, the parent parameter table is
       * consulted.  This process continues recursively until either
       * the parameter is found or there are no more parent parameter
       * tables.  Thus if the parameter "p" is defined both locally
       * and in the parent parameter table, this method will only
       * update the local value.
       *
       * @param id    The id of the sought parameter.
       * @param value The new value.
       */
    void setValue(const std::string& id, double value);

      /**
       * Set the parent parameter table.
       *
       * This method will update the parent parameter table.
       *
       * @param p The new parent parameter table.
       */
    inline void setParent(ParameterTable* p)
      { parent = p; }

      /**
       * Put a parameter table to an output stream.
       *
       * This method will put a parameter table onto an output stream.
       * The result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param p  The parameter table in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (
      std::ostream& os, const ParameterTable& p);
  };
}
#endif
