/* The ReactionTable class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionTable.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <math.h>
#include <stdlib.h>

#include "definition.hpp"
#include "ReactionTable.hpp"
#include "ui/OPrefixStream.hpp"

#include <sbml/util/util.h>

  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* The list clearing can perhaps be done a little nicer.  On the
   * other, here we have plenty of time for stupid code.
   */
MesoRD::ReactionTable::ReactionTable()
  :rates((double) 0, 0)
{
  reactions.clear();

  hash_hits     = 0;
  hash_misses   = 0;
  hash_size     = MESORD_REACTION_HASHTABLE_SIZE;
  outstream_log = 0;

  hash_table.resize(hash_size);
 
  for (unsigned int i = 0; i < hash_size; i++)
    hash_table[i] = 0;
}

  
  /* Destructor.  Do-nothing.
   */
MesoRD::ReactionTable::~ReactionTable()
{}



  /********************
   * REACTION HASHING *
   ********************/

  /* Maybe we should all go on a quest to see if we can find a worse
   * hash function.
   */
size_t
MesoRD::ReactionTable::hash(const std::valarray<unsigned int>& species) const
{
  size_t hash_factor = 1;
  size_t hash_value  = 0;
  
  for (unsigned int i = 0; i < species.size(); i++) {
    hash_value  += species[i] * hash_factor;
    hash_factor *= 2;
  }
  
  return hash_value % hash_size;
}


  /* If the key is already filled in the hash table, remove the old
   * value and add the new one.
   */
void
MesoRD::ReactionTable::hashAddReplace(const ReactionHashEntry* entry)
{
  size_t hash_value = hash(entry -> species);

  if (hash_table[hash_value])
    delete hash_table[hash_value];

  hash_table[hash_value] = entry;
}


  /* The valarray is a funny thing when it comes to elementary
   * operations.
   */
bool
MesoRD::ReactionTable::hashValueEqual(
  const std::valarray<unsigned int>& v1,
  const std::valarray<unsigned int>& v2)
{
  if (v1.size() != v2.size())
    return false;
  
  std::valarray<bool> v = (v1 == v2);
  
  for (unsigned int i = 0; i < v.size(); i++) {
    if (!v[i])
      return false;
  }
  
  return true;
}


  /***********************
   * REACTION EVALUATION *
   ***********************/

  /* First attempt to look up the particular species configuration in
   * the hash table.  If it was not found, construct a new entry and
   * add it.  Return the sum of all reaction rates.
   */
double
MesoRD::ReactionTable::getRates(const std::valarray<unsigned int>& species)
{
  if (reactions.size() == 0)
    return 0;
	
  if (hash_table[hash(species)]) {
    if (hashValueEqual(species, hash_table[hash(species)] -> species)) {
      hash_hits++;
      return (hash_table[hash(species)] -> rates).sum();
    }
  }
  
  unsigned int                        i = 0;
  std::list<Reaction>::const_iterator j;
  ReactionHashEntry*                  entry = new ReactionHashEntry;

  entry -> rates.resize(reactions.size());
  entry -> species.resize(species.size());
  entry -> species = species;
  for (j = reactions.begin(); j != reactions.end(); j++)
    entry -> rates[i++] = j -> getRate(species);

  hashAddReplace(entry);
  hash_misses++;

  return (entry -> rates).sum();
}

/* 
 * This function assumes that there are two reactants, A and B, for each
 * reactions. For each reaction the rate for the combination A1B2 and
 * A2B1 will summed together.
 */
double 
MesoRD::ReactionTable::getRates(const std::valarray<unsigned int>& species_sv_1, 
  const std::valarray<unsigned int>& species_sv_2)
{
  double sum_of_rates = 0;
  
  std::list<Reaction>::iterator j;  
  std::valarray<unsigned int> temp = std::valarray<unsigned int>((unsigned int) 0, species_sv_1.size());

  for( j = reactions.begin(); j != reactions.end(); j++ ) {
    if( j -> getMultiSVReactionAss() ) { 
      std::valarray<unsigned int> reactants = j -> getReactants();
      temp[reactants[0]] = species_sv_2[reactants[0]];
      temp[reactants[1]] = species_sv_1[reactants[1]];
      sum_of_rates += j -> getRate(temp);
      temp[reactants[0]] = species_sv_1[reactants[0]];        
      temp[reactants[1]] = species_sv_2[reactants[1]];
      sum_of_rates += j -> getRate(temp);
    }
  }
  return sum_of_rates/2;
}

/* Does the same as getRates. The only difference is that the reaction
 * rates for each combination is put into a vector instead of beeing
 * summed.
 */
const std::valarray<double>&
MesoRD::ReactionTable::getRatesArray(const std::valarray<unsigned int>& species_sv_1, 
  const std::valarray<unsigned int>& species_sv_2)
{
 
  std::list<Reaction>::iterator j;  
  std::valarray<unsigned int> temp = std::valarray<unsigned int>(species_sv_1);

  rates.resize(2*reactions.size());
  rates = 0;

  unsigned int counter = 0;
  
  for( j = reactions.begin(); j != reactions.end(); j++ ) {
    if( j -> getMultiSVReactionAss() ) {
      std::valarray<unsigned int> reactants = j -> getReactants();
      
      temp[reactants[0]] = species_sv_2[reactants[0]];
      rates[counter] = j -> getRate(temp);
      temp[reactants[0]] = species_sv_1[reactants[0]];

      temp[reactants[1]] = species_sv_2[reactants[1]];
      rates[counter+1] += j -> getRate(temp);
      temp[reactants[1]] = species_sv_1[reactants[1]];

    }
    counter += 2;
  }
  return rates;
}


/*double
MesoRD::ReactionTable::getRate(const std::valarray<unsigned int>& species, size_t reaction_index)
{
  if (reactions.size() == 0)
    return 0;

  return reactions[reaction_index].getRate(species);
}
*/
  /* First attempt to look up the particular species configuration in
   * the hash table.  If it was not found, construct a new entry and
   * add it.  Return an array with all the reaction rates.
   */
const std::valarray<double>&
MesoRD::ReactionTable::getRatesArray(
  const std::valarray<unsigned int>& species)
{

  if (hash_table[hash(species)]) {
    if (hashValueEqual(species, hash_table[hash(species)] -> species)) {
      hash_hits++;
      return hash_table[hash(species)] -> rates;
    }
  }

  unsigned int                        i = 0;
  std::list<Reaction>::const_iterator j;
  ReactionHashEntry*                  entry = new ReactionHashEntry;

  entry -> rates.resize(reactions.size());
  entry -> species.resize(species.size());
  entry -> species = species;
  for (j = reactions.begin(); j != reactions.end(); j++)
    entry -> rates[i++] = j -> getRate(species);

  hashAddReplace(entry);
  hash_misses++;

  return (entry -> rates);
}


double
MesoRD::ReactionTable::getRates(const std::valarray<double>& species)
{
  if (reactions.size() == 0)
    return 0;
  
    //ReactionHashEntry*                  entry = new ReactionHashEntry;
  unsigned int                           i = 0;
  std::list<Reaction>::const_iterator j;

  rates.resize(reactions.size());
  
  for (j = reactions.begin(); j != reactions.end(); j++)
    rates[i++] = j -> getRate(species);

  return (rates).sum();
}


/* Removed the hash function since it wasn't applicable to continious
   number of molecules */

const std::valarray<double>&
MesoRD::ReactionTable::getRatesArray(
  const std::valarray<double>& species)
{
  unsigned int                           i = 0;
  std::list<Reaction>::const_iterator j;
  
  rates.resize(reactions.size());
  
  for (j = reactions.begin(); j != reactions.end(); j++)
    rates[i++] = j -> getRate(species);

  return (rates);
}

  
  /* If the particular index is not found, return a dummy change
   * vector instead.
   */
const std::valarray<int>&
MesoRD::ReactionTable::getSpeciesChangeVector(off_t index) const
{
  std::list<Reaction>::const_iterator i;
  off_t                               j;
  
  for (i = reactions.begin(), j = 0; i != reactions.end(); i++, j++) {
    if (j == index)
      return i -> getSpeciesChangeVector();
  }

  return *(new std::valarray<int>((int)0, 0));
}

  
  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML.
   */
namespace MesoRD {
  std::ostream& operator<< (std::ostream& os, const ReactionTable& r) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
    std::list<Reaction>::const_iterator i;
    
    opfs_level0 << "<listOfReactions>" << std::endl;
    for (i = r.reactions.begin(); i != r.reactions.end(); i++)
      opfs_level1 << (*i);

    opfs_level0 << "</listOfReactions>" << std::endl;
    return os;
  }
}
