/* The ReactionTable class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionTable.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef __ReactionTable_H
#define __ReactionTable_H

#include <iostream>
#include <list>
#include <valarray>
#include <vector>

#include "reaction/Reaction.hpp"

namespace MesoRD {

    /**
     * The ReactionHashEntry struct.
     *
     * We hope to increase performance by hashing frequently computed
     * reaction--species combinations in a hash table.  This is what
     * the hash table entries look like.
     */
  typedef struct ReactionHashEntry {

      /**
       * The species composition for which the associated rates vector
       * is valid.
       */
    std::valarray<unsigned int> species;

      /**
       * The rates for all the reaction of the associated species
       * vector.
       */
    std::valarray<double> rates;
  } ReactionHashEntry;
    

    /**
     * Class keeping track of reactions in a table.
     *
     * The ReactionTable class holds a list of reactions as well as a
     * hash table.  Whenever a reaction rate is requested, we first
     * try to find it in the hash table.  If it is not there, we
     * compute the rate and put it into the hash table.  Thus we hope
     * to enable faster throughput on frequently computed
     * reaction--species combinations.
     */
  class ReactionTable {
  private:

    std::list<Reaction>                   reactions;

    int                                   hash_hits;
    int                                   hash_misses;
    size_t                                hash_size;
    std::vector<const ReactionHashEntry*> hash_table;
    std::ostream*                         outstream_log;

    std::valarray<double>rates;
    
  public:

      /**
       * Constructor for the ReactionTable class.
       *
       * The constructor will empty the list of reactions and
       * initialise the hash table.
       */
    ReactionTable();

      /**
       * Destructor for the ReactionTable class.
       *
       * The destructor does absolutely nothing.
       */
    ~ReactionTable();

      /**
       * Add a reaction to the reaction table.
       *
       * This method will add a reaction to the reaction table.  Note
       * that there is no check preventing the addition of several
       * identical reactions.  Callers simply have to keep track of
       * their actions.
       *
       * @param r The reaction to add.
       */
    void addReaction(const Reaction& r)
      { reactions.push_back(r); }

      /**
       * Computes a hash value of a species array.
       *
       * This hash function needs a little...  well, a lot really...
       * of thought.  It is not very good.
       *
       * @param species The species array which to hash.
       * @return        The computed hash value.
       */
    size_t hash(const std::valarray<unsigned int>& species) const;

      /**
       * Add or replace a value in the hash table.
       *
       * This method first computes the hash value of the entry.  This
       * value is the key.  It then consults the hash table to see if
       * the key is occupied.  If it is the key's value is replaced
       * with the new value.
       *
       * @param entry The entry to put into the hash table.
       */
    void hashAddReplace(const ReactionHashEntry* entry);

      /**
       * Checks equality of two species arrays.
       *
       * This method is actually less trivial than one may think.  The
       * valarray is presumably not really meant for these kinds of
       * operations.
       *
       * @param v1 The first species array.
       * @param v2 The second species array.
       * @return   True if the two arrays are equal, false otherwise.
       */
    bool hashValueEqual(
      const std::valarray<unsigned int>& v1,
      const std::valarray<unsigned int>& v2);
    

      /** 
       * Get the reactions that are allowing interactions in nearest
       * neighbouring subvolumes.
       *
       * Get the reactions that are inter subvolumes reactions, ie
       * reaction where the one of the reacting species resides in a
       * neighbouring subvolume.  
       *
       * @return An array of defining which
       * reactions that are nearest neighbour reactions.
       */    
    inline std::valarray<bool> getMultiSVReactionAssArray()
      {
        std::list<Reaction>::const_iterator i;
        
        std::valarray<bool> multi_sv_reactions(false, reactions.size());
        unsigned int counter = 0;
        for (i = reactions.begin(); i != reactions.end(); i++) {
          multi_sv_reactions[counter] = i -> getMultiSVReactionAss();
          counter++;
        }
        return multi_sv_reactions;
      }


      /** 
       * Get the reactions that are dissociation reactions
       * corresponding to the association reactions defined in
       * getMultiSVReactionAssArray().
       *
       * Get the reactions that are inter subvolumes reactions, ie
       * reaction where the one of the reacting species resides in a
       * neighbouring subvolume.
       *
       * @return Array of dissociation reatctions.  
       */    
    inline std::valarray<bool> getMultiSVReactionDissArray()
      {
        std::list<Reaction>::const_iterator i;
        
        std::valarray<bool> multi_sv_reactions_diss(false, reactions.size());
        unsigned int counter = 0;
        for (i = reactions.begin(); i != reactions.end(); i++) {
          multi_sv_reactions_diss[counter] = i -> getMultiSVReactionDiss();
          counter++;
        }
        return multi_sv_reactions_diss;
      }
      
    
      /**
       * Get the nearest neighbour interaction reaction flag for a specific reaction.  
       *
       *  Get the reactions that are inter subvolumes reactions, ie
       * reaction where the one of the reacting species resides in a
       * neighbouring subvolume.
       *
       * @param index defines which reaction to query.
       *
       * @return the state of the nearest neighbour reaction.  
       */
    inline bool getMultiSVReactionAss(unsigned int index)
      { 
        std::list<Reaction>::const_iterator i = reactions.begin();
        std::advance(i, index);
        
        return i -> getMultiSVReactionAss(); 
      }

        
      /**
       * Get the multi-subvolume reaction flag for a specific reaciton.  
       *
       * @param index defining the reaction to query.  
       * @return the state of the nearest neighbour reactions.  
       */
    inline bool getMultiSVReactionDiss(unsigned int index)
      { 
        std::list<Reaction>::const_iterator i = reactions.begin();
        std::advance(i, index);
        
        return i -> getMultiSVReactionDiss(); 
      }


      /**
       * Get the number of reactions stored in this reactiontable.
       *
       * @return Number of reactions in the reaction table.  
       */    
    inline size_t 
    getNumReactions()
      { return reactions.size(); }
    
      /**
       * Get the sum of all reaction rates.
       *
       * Get the sum of the reaction rates.  This is useful for
       * finding the time until the next reaction, whatever this next
       * reaction may be.
       *
       * @param species The current species vector.
       * @return        The sum of all reaction rates.
       */
    double getRates(const std::valarray<unsigned int>& species);


      /**
       * Get the sum of all that participate in multisubvolume
       * reactions.  This requires the species composition of two
       * different subvolumes.  
       */
    double getRates(const std::valarray<unsigned int>& species_sv_1, 
      const std::valarray<unsigned int>& species_sv_2);
    
    
      /**
       * Get the all reaction rates in an array
       *
       * Get all the reaction rates.  This is useful for finding
       * exactly what the next reaction will be.
       *
       * @param species The current species vector.
       * @return        An array with all the reaction rates.
       */
    const std::valarray<double>& getRatesArray(
      const std::valarray<unsigned int>& species);

      /**
       * Get all the reaction rates that are tagged as nearest neighbour reactions.  
       *
       * Get an array of reaction rates for nearest neighbour reactions.  
       *
       * @param species_sv_1 The species array of the middle subvolume.  
       * @param species_sv_2 The species array of the nearest neighbouring subvolume.  
       * @return             The array of reactions rates.  
       */    
    const std::valarray<double>& getRatesArray(
      const std::valarray<unsigned int>& species_sv_1, 
      const std::valarray<unsigned int>& species_sv_2);
    
    
      /**
       * Get the sum of all reaction rates.
       *
       * Get the sum of the reaction rates.  This is useful for
       * finding the time until the next reaction, whatever this next
       * reaction may be. Only used if RUN_DETERMINISTIC is set at
       * compile.  
       *
       * @param species The current species vector.
       * @return        The sum of all reaction rates.
       */
    double getRates(const std::valarray<double>& species);

      /**
       * Get the all reaction rates in an array
       *
       * Get all the reaction rates.  This is useful for finding
       * exactly what the next reaction will be. Only used if 
       * RUN_DETERMINISTIC is set at compile.  
       *
       * @param species The current species vector.
       * @return        An array with all the reaction rates.
       */
    const std::valarray<double>& getRatesArray(
      const std::valarray<double>& species);


      /**
       * Get the reactants for a specific reactions.  
       * 
       * @param index The reaction for which the reactants are returned.  
       * @return A list of indecies for the reactants of a reactions. 
       */    
    inline std::valarray<unsigned int> getReactants(unsigned int index)
      { std::list<Reaction>::iterator i = reactions.begin();
        std::advance(i, index);
        return i -> getReactants();
        }
        

      /**
       * Get the productss for a specific reactions.  
       * 
       * @param index The products for which the reactants are returned.  
       * @return A list of indecies for the products of a reactions. 
       */
    inline std::valarray<unsigned int> getProducts(unsigned int index)
      {
        std::list<Reaction>::iterator i = reactions.begin();
        std::advance(i, index);
        return i -> getProducts();
        }

      /**
       * Get the species change vector for the particular reaction.
       *
       * One should make sure that the index is in range.  We do not
       * want to check index ranges, because it can take time we are
       * not willing to spend.  Should the index be out of range, a
       * dummy array of undefined length is returned, containing all
       * zeroes.
       *
       * @param index The index of the reaction of interest.
       * @return      The species change vector of the reaction.
       */
    const std::valarray<int>& getSpeciesChangeVector(off_t index) const;
  
      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing
       * begins, any error reporting and logging information will be
       * silently lost.  It is thus highly recommended to call this
       * method prior to calling any model parsing methods.  Note also
       * that this method will set up the logging stream for any
       * member classes that can make use of one.
       *
       * @param log The logging stream onto which to report errors, etc.
       */
    inline void setOptionLogfile(std::ostream* log)
      { for (std::list<Reaction>::iterator i = reactions.begin(); i != reactions.end(); i++)
        i -> setOptionLogfile(log);
      outstream_log = log; }

      /**
       * Put a reaction table to an output stream.
       *
       * This method will put a reaction table onto an output stream.
       * The result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param r  The reaction table in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (std::ostream& os, const ReactionTable& r);
  };
}

#endif
