/* The Species class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Species.cpp 561 2008-11-25 20:10:53Z davidfange $
 */
#include <iostream>
#include <sstream>

#include "mesord.hpp"
#include "definition.hpp"

#include "Annotation.hpp"
#include "Compartment.hpp"
#include "Species.hpp"
#include "Unit.hpp"
#include "UnitTable.hpp"

#include "ui/OPrefixStream.hpp"


bool MesoRD::Species::useDoubleInAmount = false;

  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* True default constructor.  Sets everything to default values.
   */
MesoRD::Species::Species() 
{
  if(useDoubleInAmount)
  {
    amountDouble = 0;
  }
  else
  {
    amountInt = 0;
  }
  compartment   = "";
  constant      = false;
  id            = "";
  name          = "";
  outstream_log = 0;

  diffusion.clear();
}


  /* Copy constructor.
   */
MesoRD::Species::Species(const Species& s) 
{
  this -> useDoubleInAmount = s.useDoubleInAmount;
  if(this -> useDoubleInAmount)
  {
    this -> amountDouble = s.amountDouble;
  }
  else
  {
    this -> amountInt = s.amountInt;
  }
  this -> compartment   = s.compartment;
  this -> constant      = s.constant;
  this -> diffusion     = s.diffusion;
  this -> id            = s.id;
  this -> name          = s.name;
  this -> outstream_log = s.outstream_log;
}


  /* Value constructor.
   */
MesoRD::Species::Species(const std::map<std::string, double>& d,
  const std::string& i, const std::string& n,
  const std::string& c, unsigned int a, bool k)
{
  this -> amountInt     = a;
  this -> compartment   = c;
  this -> constant      = k;
  this -> diffusion     = d;
  this -> id            = i;
  this -> name          = n;
  this -> outstream_log = 0;
}

MesoRD::Species::Species(const std::map<std::string, double>& d,
  const std::string& i, const std::string& n,
  const std::string& c, double a, bool k)
{
  this -> amountDouble  = a;
  this -> compartment   = c;
  this -> constant      = k;
  this -> diffusion     = d;
  this -> id            = i;
  this -> name          = n;
  this -> outstream_log = 0;
}


  /* Initialises the species from the SBML data structures.
   */
MesoRD::Species::Species(
  ::Species* s, const Compartment* c, const UnitTable* u, std::ostream* o)
  throw(MesoRD::Species::ModelError)
{
  outstream_log = o;
  try {
    fromSBML(s, c, u);
  } catch (ModelError &e) {
    throw ModelError(e.what());
  }
}


  /* Destructor.  Do-nothing.
   */
MesoRD::Species::~Species() 
{}



  /************************
   * MANIPULATION METHODS *
   ************************/

void
MesoRD::Species::clearAmount()
{
  if (useDoubleInAmount) {
    amountDouble=0;
  } else {
    amountInt=0;
  }
}



  /* Gets the diffusion rate for the compartment identified by the
   * given id.
   */
double
MesoRD::Species::getDiffusion(const std::string& c) const
{
  if (compartment.length() > 0) {
    std::map<std::string, double>::const_iterator i;

    if ((i = diffusion.find(c)) != diffusion.end())
      return i -> second;
  }
  
  return -1.0;
}


  /* Only change the amount of the species if it will not make the
   * amount negative.
   */
unsigned int
MesoRD::Species::changeAmount(const int change)
{
  if (change > 0 || change < 0 && -1 * change <= static_cast<int>(amountInt))
      return amountInt += change;
  return amountInt;
}


double
MesoRD::Species::changeAmount(const double change)
{
  if (change > 0 || change < 0 && -1 * change <= amountDouble)
    return amountDouble += change;
  return amountDouble;
}



  /****************
   * SBML SUPPORT *
   ****************/

  /* The compartment is needed in order to compute the initial amount
   * of the species in case the initial amount is specified in terms
   * of concentration.
   */
void
MesoRD::Species::fromSBML(
  ::Species* s, const Compartment* c, const UnitTable* ut)
  throw(MesoRD::Species::ModelError)
{
  std::stringstream message;
  
    /* Read in the id.
     */
  if ( s -> isSetId() ) {
    id = s -> getId();
  } else {
    message << "One species is lacking "
            << "id, unable to continue";
    throw ModelError((message.str()).c_str());
  }
  
  
    /* Read in the compartment.
     */
  if (s -> isSetCompartment())
    compartment = s -> getCompartment();
  else {
    message << "Species \""
            << id
            << " has no defined Compartment";
    throw ModelError((message.str()).c_str());
  }
  
    /* Read in the name.  In this implementation, the name is key.
     * !! FIXME !! SpeciesType should be used here instead.
     */
  if ( s -> isSetName()) {
    name = s -> getName();
  } else {
    message << "Species \""
            << id
            << " has no defined Compartment";
    throw ModelError((message.str()).c_str());
  }
  
  
    /* Read in the constant status of the species.
     */
  constant = s -> getConstant();
  

    /* Read the amount.  The amount can be specified in either
     * concentration or substance units. This only makes sense if
     * there is a unittable.
     */
  if(ut) {
    if ( s -> isSetInitialAmount()) {
      MesoRD::Unit* unit_ptr;
      if (s -> isSetSubstanceUnits())
        unit_ptr = ut -> getUnit(s -> getSubstanceUnits());
      else
        unit_ptr = new MesoRD::Unit(MesoRD::Unit::mole());
    
      if(unit_ptr) {
        try {
          double conversion = unit_ptr -> toUnit(Unit::dimensionless());
          if (useDoubleInAmount) {
            amountDouble = rint(conversion * s -> getInitialAmount());
          }else {
            amountInt = static_cast<unsigned int>
              (lrint(conversion * s -> getInitialAmount()));
          }
          delete unit_ptr;
        
        } catch (Unit::InvalidConversion&) {
          delete unit_ptr;
          message << "Cannot use unit "
                  << s -> getSubstanceUnits()
                  << " to obtain species amount.";
          throw ModelError((message.str()).c_str());
        }
      } else {
        message << "Undefined unit "
                << s -> getSubstanceUnits()
                << " for species amount.";
        throw ModelError(message.str().c_str());
      }
    } else if (s -> isSetInitialConcentration()) {
    
        /* Attempt to convert the given units to item per cubic metre
         * and figure out the concentration from there.  If neither
         * unit is set, we assume mole per litre.  This is the default
         * in SBML.
         */
      if (c) {
        double volume    = c -> getVolume();
        double size      = 1;
        double substance = 0;
        double concentration;
      
        std::string    unit_id  = c -> getOriginalUnit();
        MesoRD::Unit*  unit_ptr = ut -> getUnit(unit_id);
        if (unit_ptr) {
          try {
            size = unit_ptr -> toUnit(Unit::litre());
          } catch (Unit::InvalidConversion&) {
            delete unit_ptr;
            message << "Cannot use unit "
                    << s -> getSubstanceUnits()
                    << " to obtain amount for species:"
                    << s -> getId()
                    << " defined on line:"
                    << s -> getLine();
            throw ModelError((message.str()).c_str());
          } 
        } else {
          if (unit_id == "") {
              //!! FIXME, SBML definition dependecy should be defined
              // elsewhere !!
            message << "No unit defined for compartment: "
                    << c -> getId()
                    << " needed for species:"
                    << s -> getId() 
                    << " defined on line:" 
                    << s -> getLine()
                    << ". Check documentation for SBML level 2,"
                    << " version 3. " ;
          } else {
            message << "Undefined unit "
                    << unit_id
                    << " as spatialunit for species:"
                    << s -> getId() 
                    << " defined on line:" 
                    << s -> getLine();
          }
          throw ModelError(message.str().c_str());
        }
        
        if ( s -> isSetSubstanceUnits() )
          unit_ptr = ut -> getUnit(s -> getSubstanceUnits());
        else
          unit_ptr = new MesoRD::Unit(MesoRD::Unit::mole());
      
        if(unit_ptr) {
          try {
            substance = unit_ptr -> toUnit(Unit::dimensionless());
          } catch (Unit::InvalidConversion&) {
            delete unit_ptr;
            message << "Cannot use unit "
                    << s -> getSubstanceUnits()
                    << " to obtain amount for species:"
                    << s -> getId()
                    << " defined on line:"
                    << s -> getLine();
            throw ModelError((message.str()).c_str());
          }
          delete unit_ptr;
        
        } else {
          message << "Undefined unit "
                  << s -> getSubstanceUnits()
                  << " for species amount.";
          throw ModelError(message.str().c_str());
        }
      
        concentration = 
          s -> getInitialConcentration() * size * volume * substance;
      
        if(useDoubleInAmount)
        {
          amountDouble = rint(concentration);
        }
        else
        {
          amountInt = static_cast<unsigned int>(lrint(concentration));
        }
      }
    
    } else {
      std::stringstream message;
      message << "Initial amount for species \""
              << id
              << " not specified";
      throw ModelError((message.str()).c_str());
    }
  
      /* Read in the diffusion rates.
       */
    const ::XMLNode* n = s -> getAnnotation();
  
    Annotation   annotation( n );
  
    for (unsigned int i = 0; i < annotation.getNumChildren(); i++) {
      const Annotation *child = annotation.getChild(i);
    
      Unit *up = 0;
    
      if (child) {
        if (child -> getName() == Identifier::XML::Diffusion::name) {
          try {
          
            double diffusion_unit_multiplier = 1;
          
            std::string c = child -> getStringAttribute(
              Identifier::XML::Diffusion::Attributes::compartment);
          
            std::string units       = child -> getStringAttribute(
              Identifier::XML::Diffusion::Attributes::units);

            if (ut && (up = ut -> getUnit(units))) {
              diffusion_unit_multiplier = 
                up -> toUnit(Unit::centimetreSquaredPerSecond());

            } else {
              message << "Diffusion annotation requires a defined unit";
              throw ModelError((message.str()).c_str());
            }

            diffusion[c]  = diffusion_unit_multiplier *
              (child -> getDoubleAttribute(
                Identifier::XML::Diffusion::Attributes::rate));

          } catch (Annotation::AttributeNotFound &e) {
            message << "Diffusion annotation requires "
                    << "MesoRD:" << e.getName()
                    << " to be defined";
            throw ModelError((message.str()).c_str());
            
          } catch(Annotation::InvalidFormat<double>& e) {
            message << e.what() 
                    << ", while parsing: " 
                    << e.getValue() << std::endl;
            throw ModelError((message.str()).c_str());
            
          } catch(Unit::InvalidConversion&) {
            message << "Cannot use unit "
                    << up -> getId()
                    << " to obtain diffusion rate.";
            throw ModelError((message.str()).c_str());
          }
        }
      }
    }
  }
}






  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML.
   *
   * FIXME!!! Use the Annotation class to do the output
   */
namespace MesoRD {
  std::ostream& operator<< (std::ostream& os, const Species& s) {
    std::map<std::string, double>::const_iterator i;
    std::string                                   leader = "";
    
    for (unsigned int j = 0;
         j < (Identifier::XML::Diffusion::name.length() + 2); j++)
      leader += " ";

    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
    OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");
    OPrefixStream opfs_level3(opfs_level2.rdbuf(), leader.c_str());

    opfs_level0
      << "<species id=\""
      << s.id          << "\""  << std::endl
      << "         compartment=\""
      << s.compartment << "\""  << std::endl
      << "         constant=\""
      << "false"       << "\""  << std::endl
      << "         hasOnlySubstanceUnits=\""
      << "true"        << "\""  << std::endl
      << "         initialAmount=\"";
    
      
      if(s.useDoubleInAmount)
      {  
        opfs_level0 << s.amountDouble      << "\""  << std::endl;
      }
      else
      {
        opfs_level0 << s.amountInt      << "\""  << std::endl;
      }
    
      opfs_level0
       << "         name=\""
       << s.name        << "\""  << std::endl
       << "         substanceUnits=\""
       << "item"        << "\">" << std::endl;

      
      
    opfs_level1 << "<annotation xmlns:" << Identifier::XML::xmlns
                << ">" << std::endl;

    for (i = s.diffusion.begin(); i != s.diffusion.end(); i++) {
      opfs_level2 << "<"   << Identifier::XML::Diffusion::name
                  << " "   << Identifier::XML::Diffusion::Attributes::compartment
                  << "=\"" << i -> first
                  << "\""  << std::endl;

      opfs_level3 << Identifier::XML::Diffusion::Attributes::rate
                  << "=\""  << i -> second
                  << "\"" << std::endl;
      
    }
    
    opfs_level1 << "</annotation>" << std::endl;
    opfs_level0 << "</species>" << std::endl;

    return os;
  }
}
