/* The Species class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Species.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_SPECIES_HPP
#define MESORD_SPECIES_HPP

#include <iostream>
#include <map>
#include <stdexcept>
#include <string>

#include <stdio.h>
#include <sbml/SBMLTypes.h>

namespace MesoRD {

  class Compartment;
  class UnitTable;

    /**
     * Class for modelling species.
     *
     * SBML provides no means to specify the diffusion constants of a
     * species.  Therefore, an MesoRD SBML file must contain an
     * annotation element for every species.  This annotation must
     * define the diffusion constants for the species from a sub
     * volume in the compartment in which it is defined to sub volumes
     * of all the other compartments in the model definition.  A note
     * on initial amounts of molecules: internally we always work with
     * absolute amounts of species, never concentrations.
     */
  class Species {
  private:  
    union 
    {
      unsigned int                amountInt;
      double                      amountDouble;
    };
    bool                          constant;
    std::map<std::string, double> diffusion;
    std::string                   compartment;
    std::string                   id;
    std::string                   name;
    std::ostream*                 outstream_log;    
    static bool                   useDoubleInAmount;
    
  public:


      /**
       * Class for reporting errors during model building.
       *
       * If model building failed, this exception will be thrown.
       * Actually this exception lumps together a lot of other
       * exceptions that may be caught during parsing.
       */
    class ModelError : public std::runtime_error {
    public:

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Default constructor of the Species class.
       *
       * Initialises everything to null and void.
       */
    Species();

      /**
       * Copy constructor for the species class.
       *
       * Constructs a new species as a copy from an already defined
       * species.
       *
       * @param s The species to copy.
       */
    Species(const Species& s);

      /**
       * Value constructor for the Species class.
       *
       * Constructs a new species with the values as given in the
       * argument list.  Note that this constructor doubles as a
       * partial default constructor, as it has the potential to use
       * default values for everything except the diffusion map.
       *
       * @param d A map with diffusion constants to all other compartments.
       * @param i The id of the species.
       * @param n The name of the species.
       * @param c The compartment where the species is defined.
       * @param a The initial amount of the species.
       * @param k Whether the species is constant or not.
       */
    Species(
      const std::map<std::string, double>& d,
      const std::string&                   i = "",
      const std::string&                   n = "",
      const std::string&                   c = "",
      unsigned int                         a = 0,
      bool                                 k = false);

      /**
       * Value constructor for the Species class.
       *
       * Constructs a new species with the values as given in the
       * argument list.  Note that this constructor doubles as a
       * partial default constructor, as it has the potential to use
       * default values for everything except the diffusion map.
       *
       * @param d A map with diffusion constants to all other compartments.
       * @param i The id of the species.
       * @param n The name of the species.
       * @param c The compartment where the species is defined.
       * @param a The initial amount of the species.
       * @param k Whether the species is constant or not.
       */
    Species(
      const std::map<std::string, double>& d,
      const std::string&                   i = "",
      const std::string&                   n = "",
      const std::string&                   c = "",
      double                               a = 0,
      bool                                 k = false);


      /**
       * Value constructor for the species class.
       *
       * Construct a new species from the supplied SBML definition.
       * We need the compartment, because we may need to retrieve its
       * size in order to calculate molecular amounts from
       * concentrations.
       *
       * @param s The SBML species definition.
       * @param c The compartment in which the species lives.
       * @param u The unit table.
       */
    Species(
      ::Species* s, const Compartment* c, 
      const UnitTable* u, std::ostream* o)
      throw(ModelError);
    
      /**
       * Destructor for the Species class.
       *
       * Does nothing.
       */
    ~Species();

      /**
       * Sets the amount of the species to zero.
       *
       * Sets the amount to zero.
       */
    void clearAmount();


      /**
       * Change the amount of the species.
       *
       * @param  change The change.
       * @return        The updated number of species in the subvolume.
       */
    unsigned int
    changeAmount(const int change);


      /**
       * Change the amount of the species.
       *
       * @param  change The decrement.
       * @return        The updated number of species in the subvolume.
       */
    double
    changeAmount(const double change);


      /**
       * Get the amount of the species.
       *
       * Returns the amount of the species.
       *
       * @return The amount of the species.
       */
    inline unsigned int getAmount() const
      { return amountInt; }


      /**
       * Get the amount of the species.
       *
       * Returns the amount of the species.
       *
       * @return The amount of the species.
       */
    inline double getAmountDouble() const
      { return amountDouble; }

      /**
       * Get the compartment of which the species is part.
       *
       * Returns the id of the compartment in which the species lives.
       *
       * @return The id of the compartment which is holding the species.
       */
    inline std::string getCompartment() const
      { return compartment; }

      /**
       * Get the diffusion map.
       *
       * Returns the diffusion map.  The diffusion map keys the
       * diffusion constant for every compartment id.  The diffusion
       * constants are with respect to the compartment in which the
       * current species is defined.
       *
       * @return The diffusion map of the species.
       */
    inline std::map<std::string, double> getDiffusion() const
      { return diffusion; }

      /**
       * Get the diffusion constant for a certain compartment.
       *
       * Returns the diffusion constant for the species when diffusing
       * from the compartment in which it is defined to the
       * compartment as identified by the argument id.  If the
       * requested compartment could not be found, this method returns
       * a negative value, because negative values are not valid
       * diffusion constants in any way.
       *
       * @param c The id of the compartment we are diffusing into.
       * @return  The requested diffusion constant.
       */
    double getDiffusion(const std::string& c) const;

      /**
       * Get the id of the species.
       *
       * Returns the id.
       */
    inline std::string getId() const
      { return id; }

      /**
       * Get the key of the species.
       *
       * Returns the key.  In the case of species, we key on the
       * species' name.  This contrasts with SBML, which only requires
       * the id:s to be unique and imposes no limitations on the
       * names.
       *
       * @return The name of the species.
       */
    inline std::string getKey() const
      { return name; }

      /**
       * Get the name of the species.
       *
       * Returns the name of the species.
       *
       * @return The name of the species.
       */
    inline std::string getName() const
      { return name; }

      /**
       * Checks whether the species is defined in a given compartment.
       *
       * Checks if we are in the compartment as defined by the
       * argument.  If the compartment has not been defined, this
       * method always returns false.
       *
       * @param c The id of the compartment to compare with.
       * @return  True if the compartment matches this species' home
       *          compartment, false otherwise
       */
    inline bool isCompartment(const std::string& c) const
      { return ((compartment.length() > 0) && (compartment == c)); }

      /**
       * Get the constant status of the species.
       *
       * This method returns true if the given species is defined to
       * be constant, false otherwise.
       *
       * @return The constant status of the species.
       */
    inline bool isConstant() const
      { return constant; }

      /**
       * Checks whether the species is identified by a given id.
       *
       * Checks if we are identified by the given id.  If the id has
       * not been defined, this method always returns false.
       *
       * @param i The id to compare with.
       * @return  True if the id:s match, false otherwise.
       */
    inline bool isId(const std::string& i) const
      { return ((id.length() > 0) && (id == i)); }

      /**
       * Checks whether the species is identified by a given key.
       *
       * Checks if we are identified by the given key.  If the key has
       * not been defined, this method always returns false.  For
       * species, the name is key.
       *
       * @param key The name to compare with.
       * @return    True if the names match, false otherwise.
       */
    inline bool isKey(const std::string& key) const
      { return (((this -> name).length() > 0) && (this -> name) == key); }

      /**
       * Checks if the species is identified by the given name.
       *
       * Checks if we are identified by the given name.  If the name
       * has not been defined, this method always returns false.
       *
       * @param n The name to compare with.
       * @return  True if the names match, false otherwise.
       */
    inline bool isName(const std::string& n) const
      { return ((name.length() > 0) && (name == n)); }

      /**
       * Sets the amount of the species.
       *
       * Sets the amount of the species.
       *
       * @param a The new amount of the species.
       * @return  The new amount of the species, after the change.
       */
    inline unsigned int setAmount(unsigned int a)
      { return (amountInt = a); }


      /**
       * Sets the amount of the species.
       *
       * Sets the amount of the species.
       *
       * @param a The new amount of the species.
       * @return  The new amount of the species, after the change.
       */
    inline double setAmount(double a)
      { return (amountDouble = a); }

      /**
       * Sets the compartment of the species.
       *
       * Sets the home compartment of the species.
       *
       * @param c The new compartment of the species.
       */
    inline void setCompartment(const std::string& c)
      { compartment = c; }

      /**
       * Sets the diffusion map.
       *
       * Sets the diffusion map.  The diffusion map keys the diffusion
       * constant for every compartment id.  The diffusion constants
       * are with respect to the compartment in which the current
       * species is defined.
       *
       * @param d The diffusion map of the species.
       */
    inline void setDiffusion(const std::map<std::string, double>& d)
      { diffusion = d; }

      /**
       * Sets the diffusion constant for a certain compartment.
       *
       * Sets the diffusion constant for the species when diffusing
       * from the compartment in which it is defined to the
       * compartment as identified by the argument id.  If the
       * requested compartment could not be found, the program may
       * break.  Here we mean break as in producing wrong results, not
       * break is in causing segmentation fault.
       *
       * @param compartment The id of the compartment we are diffusing into.
       * @param diffusion   The new diffusion constant.
       */
    inline void setDiffusion(const std::string& c, double d)
      { diffusion[c] = d; }

      /**
       * Sets the id of the species
       *
       * Sets the id of the species.
       *
       * @param i The new id of the species.
       */
    inline void setId(const std::string& i)
      { id = i; }

      /**
       * Sets the key of the species.
       *
       * Sets the key of the species.  For species, the name is key.
       *
       * @param k The new key of the species.
       */
    inline void setKey(const std::string& k)
      { name = k; }

      /**
       * Sets the name of the species.
       *
       * Sets the name of the species.
       *
       * @param n The new name of the species.
       */
    inline void setName(const std::string& n)
      { name = n; }

      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing and
       * model building begins, any error reporting and logging
       * information will be silently lost.  It is thus highly
       * recommended to call this method prior to calling fromSBML().
       * Note also that this method will set up the logging stream for
       * any member classes that can make use of one.
       *
       * @param l The logging stream onto which to report errors, etc.
       */
    inline void setOptionLogfile(std::ostream* l)
      { outstream_log = l; }

      /**
       * Read the species from an SBML definition.
       *
       * This method will take all the values set in an SBML
       * definition and copy them to the current instance of the
       * Species class.  Some fields are optional in SBML.  For such
       * fields the method makes sure to use default values instead.
       * We need the compartment, because we may need to retrieve its
       * size in order to calculate molecular amounts from
       * concentrations.
       *
       * @param s The SBML species definition.
       * @param c The compartment in which the species lives.
       * @param u The unit table.
       */
    void fromSBML(
      ::Species* s, const Compartment* c, const UnitTable* u)
      throw(ModelError);

      /**
       * Put a species to an output stream.
       *
       * This method will put a species onto an output stream.  The
       * result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param s  The species in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (std::ostream& os, const Species& s);

      /**
       * Sets the flag useDoubleInAmount which decieds if the number
       * of species should be saved in unsigned int or in double.  
       *
       * @param useDoubleInAmount The flag indicating double.  
       */
    inline static void setUseDoubleInAmount(bool useDouble)
      { useDoubleInAmount = useDouble; }

      /**
       * Gets the flag indicating if the amount of the species should
       * be saved in unsigned int or in double.  
       *
       * @ return The flag indicating double.  
       */
    inline static bool getUseDoubleInAmount()
      { return useDoubleInAmount; }

  };
}

#endif
