/* The SpeciesTable class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SpeciesTable.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include "SpeciesTable.hpp"

#include "ui/OPrefixStream.hpp"



  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Clear the table.
   */
MesoRD::SpeciesTable::SpeciesTable()
{
  clear();
}


  /* Do a deep copy of the species data vector.
   */
MesoRD::SpeciesTable::SpeciesTable(const SpeciesTable& s)
{
  species.clear();
  for (unsigned int i = 0; i < s.species.size(); i++)
    species.push_back(s.species[i]);
}


  /* Destructor.  Do-nothing.
   */
MesoRD::SpeciesTable::~SpeciesTable()
{}



  /*****************
   * QUERY METHODS *
   *****************/

  /* Check all species in the table.
   */
bool
MesoRD::SpeciesTable::isDefined(const std::string& key) const
{
  std::vector<Species>::const_iterator i;

  for (i = species.begin(); i != species.end(); i++) {
    if (i -> isKey(key)) {
      return true;
    }
  }

  return false;
}


  /* Check all species in the table.
   */
bool
MesoRD::SpeciesTable::isDefinedById(const std::string& id) const
{
  std::vector<Species>::const_iterator i;

  for (i = species.begin(); i != species.end(); i++) {
    if (i -> isId(id)) {
      return true;
    }
  }

  return false;
}



 /******************
   * GETTER METHODS *
   ******************/

  /* Search through the species table until the species is found.
   */
off_t
MesoRD::SpeciesTable::getIndex(const std::string& key) const 
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isKey(key))
      return i;
  }

  return -1;
}


  /* Search through the species table until the species is found.
   */
off_t
MesoRD::SpeciesTable::getIndexById(const std::string& id) const 
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isId(id))
      return i;
  }

  return -1;
}



  /******************************
   * TABLE MANIPULATION METHODS *
   ******************************/
  
  /* Look through the entire table until the species is found.  It
   * probably makes sense to return zero if the species was not found.
   */
double
MesoRD::SpeciesTable::changeAmountById(
  const std::string& id, const double change)
{
  for (size_t i = 0; i < species.size(); i++) {
    if (species[i].isId(id))
      return species[i].changeAmount(change);
  }

  return 0;
}


unsigned int
MesoRD::SpeciesTable::changeAmountById(
  const std::string& id, const int change)
{
  for (size_t i = 0; i < species.size(); i++) {
    if (species[i].isId(id))
      return species[i].changeAmount(change);
  }

  return 0;
}


double
MesoRD::SpeciesTable::changeAmountByName(
  const std::string& name, const double change)
{
  for (size_t i = 0; i < species.size(); i++) {
    if (species[i].isName(name))
      return species[i].changeAmount(change);
  }

  return 0;
}


unsigned int
MesoRD::SpeciesTable::changeAmountByName(
  const std::string& name, const int change)
{
  for (size_t i = 0; i < species.size(); i++) {
    if (species[i].isName(name))
      return species[i].changeAmount(change);
  }

  return 0;
}


  /* Look through the entire table until the species is found.  It
   * probably makes sense to return zero if the species was not found.
   */

unsigned int
MesoRD::SpeciesTable::setAmountById(const std::string& id, unsigned int amount)
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isId(id))
      return species[i].setAmount(amount);
  }

  return 0;
}


double
MesoRD::SpeciesTable::setAmountById(const std::string& id, double amount)
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isId(id))
      return species[i].setAmount(amount);
  }

  return 0;
}


  /* Look through the entire table until the species is found.  It
   * probably makes sense to return zero if the species was not found.
   */

unsigned int
MesoRD::SpeciesTable::setAmountByName(
  const std::string& name, unsigned int amount)
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isName(name))
      return species[i].setAmount(amount);
  }

  return 0;
}


double
MesoRD::SpeciesTable::setAmountByName(
  const std::string& name, double amount)
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isName(name))
      return species[i].setAmount(amount);
  }

  return 0;
}


  /* Look through the entire table until the species is found.
   */
void
MesoRD::SpeciesTable::updateSpecies(const std::map<std::string,
  double>& d, const std::string& i, const std::string& n,
  const std::string& c, unsigned int a)
{
  for (unsigned int j = 0; j < species.size(); j++) {
    if (species[j].isKey(n)) {
      species[j].setAmount(a);
      species[j].setCompartment(c);
      species[j].setDiffusion(d);
      species[j].setId(i);
      species[j].setName(n);
    }
  }
}

void
MesoRD::SpeciesTable::updateSpecies(const std::map<std::string,
  double>& d, const std::string& i, const std::string& n,
  const std::string& c, double a)
{
  for (unsigned int j = 0; j < species.size(); j++) {
    if (species[j].isKey(n)) {
      species[j].setAmount(a);
      species[j].setCompartment(c);
      species[j].setDiffusion(d);
      species[j].setId(i);
      species[j].setName(n);
    }
  }
}



  /* Look through the entire table until the species is found.
   */
void
MesoRD::SpeciesTable::updateSpecies(const Species& s)
{
  for (unsigned int i = 0; i < species.size(); i++) {
    if (species[i].isKey(s.getKey())) {
      if(species[i].getUseDoubleInAmount()) {  
        species[i].setAmount(s.getAmountDouble());
      } else {
        species[i].setAmount(s.getAmount());
      }
      species[i].setCompartment(s.getCompartment());
      species[i].setDiffusion(s.getDiffusion());
      species[i].setId(s.getId());
      species[i].setName(s.getName());
    }
  }
}



  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML
   */
namespace MesoRD {
  std::ostream& operator<< (std::ostream& os, const SpeciesTable& s) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "<listOfSpecies>" << std::endl;
    
    for (unsigned int i = 0; i < s.species.size(); i++)
      opfs_level1 << s.species[i];
    
    opfs_level0 << "</listOfSpecies>" << std::endl;
    return os;
  }
}
