/* The SpeciesTable class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SpeciesTable.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef __SpeciesTable_H
#define __SpeciesTable_H

#include <iostream>
#include <vector>

#include "Species.hpp"


namespace MesoRD {

    /**
     * The SpeciesTable class.
     *
     * Analogous to the SBML list of species, MesoRD has the
     * SpeciesTable class.  The species table holds all the species in
     * a compartment or the entire system, depending of the were and
     * how it is instantiated.  Note that the indices in the species
     * table should generally not be tampered with.  Several functions
     * in the code rely on these indices to be fixed throughout the
     * entire run.  Many aspects of this class are time critical.
     * Therefore, we do not do any range checking on indices.
     *
     * A notable difference between SBML and MesoRD is that species in
     * MesoRD are keyed on name, not on id.  The reason is that we may
     * have the same species in two compartments.  SBML requires these
     * two instances of the same species to have unique id:s, when in
     * fact they really are the same from our point of view.  There is
     * no such restriction on a species' name.  Thus we always refer
     * to species by their names.
     */
  class SpeciesTable {
  private:
    std::vector<Species> species;

  public:


      /**
       * Default constructor for the SpeciesTable class.
       *
       * A new SpeciesTable is initially empty.
       */
    SpeciesTable();

      /**
       * Copy constructor for the SpeciesTable class.
       *
       * Note that this constructor will do a deep copy of the species
       * table.
       *
       * @param s The species table to copy from.
       */
    SpeciesTable(const SpeciesTable& s);

      /**
       * Destructor for the SpeciesTable class.
       *
       * Does nothing at all.
       */
    ~SpeciesTable();

      /**
       * Add a species to the species table.
       *
       * We only ever add species to the species table if they are not
       * already defined.  Remember that species are keyed on name,
       * not id.
       *
       * @param d The new species' diffusion constant.
       * @param i The new species' id.
       * @param n The new species' name.
       * @param c The containing compartment of the new species.
       * @param a The initial amount of the new species.
       */
    inline void addSpecies(
      const std::map<std::string, double>& d,
      const std::string&                   i,
      const std::string&                   n,
      const std::string&                   c,
      unsigned int                         a)
      { if (!isDefined(n)) species.push_back(Species(d, i, n, c, a)); }

      /**
       * Add a species to the species table.
       *
       * We only ever add species to the species table if they are not
       * already defined.  Remember that species are keyed on name,
       * not id.
       *
       * @param s The species to add.
       */
    inline void addSpecies(const Species& s)
      { if (!isDefined(s.getKey())) species.push_back(s); }

      /**
       * Clear the species table.
       *
       * Calling this method will empty the species table.  There are
       * no species in the table after this call.
       */
    inline void clear()
      { species.clear(); }

      /**
       * Clear the amount for all species in the table.
       *
       * Calling this method will set the amount of all species in the
       * table to zero.  Its use is rather limited...  and specific.
       */
    inline void clearAmounts()
      { for (unsigned int i = 0; i < species.size(); i++)
        species[i].clearAmount(); }

      /**
       * Gets the constant status of the given species.
       *
       * This method returns true if the given species is defined to
       * be constant, false otherwise.  The index must be ensured to
       * be in range, otherwise strange things may happen.
       *
       * @param index The index of the species.
       * @return      The constant status of the species in question.
       */ 
    inline bool isConstantByIndex(off_t index) const
      { return species[index].isConstant(); }

      /**
       * Checks whether a species is defined.
       *
       * This method checks whether the species identified by the
       * given key is defined in the species table.  For species, the
       * name is key.
       *
       * @param key The name of the species.
       * @return    True if the species is defined, false otherwise.
       */
    bool isDefined(const std::string& key) const;

      /**
       * Checks whether a species is defined.
       *
       * This method checks whether the species identified by the
       * given id is defined in the species table.
       *
       * @param id The name of the species.
       * @return   True if the species is defined, false otherwise.
       */
    bool isDefinedById(const std::string& id) const;

      /**
       * Gets the amount of the given species.
       *
       * This method returns the amount of the given species.  The
       * index must be ensured to be in range, otherwise strange
       * things may happen.
       *
       * @param index The index of the species.
       * @return      The amount of the species in question.
       */ 
    inline unsigned int getAmountByIndex(off_t index) const
      { return species[index].getAmount(); }

      /**
       * Gets the amount of the given species.
       *
       * This method returns the amount of the given species.  The key
       * must be ensured to be defined, otherwise strange things may
       * happen.
       *
       * @param name The name of the species.
       * @return     The amount of the species in question.
       */ 
    inline unsigned int getAmountByName(const std::string& name) const
      { return species[getIndex(name)].getAmount(); }

      /**
       * Gets the id of the compartment of the given species.
       *
       * This method returns the id of the compartment of the given
       * species.  The index must be ensured to be in range, otherwise
       * strange things may happen.
       *
       * @param index The index of the species.
       * @return      The id of the compartment of the species in question.
       */ 
    inline std::string getCompartmentByIndex(off_t index) const
      { return species[index].getCompartment(); }

      /**
       * Gets the amount of the given species.
       *
       * This method returns the amount of the given species.  The
       * index must be ensured to be in range, otherwise strange
       * things may happen.
       *
       * @param index The index of the species.
       * @return      The amount of the species in question.
       */ 
    inline double getAmountByIndexDouble(off_t index) const
      { return species[index].getAmountDouble(); }

      /**
       * Gets the amount of the given species.
       *
       * This method returns the amount of the given species.  The key
       * must be ensured to be defined, otherwise strange things may
       * happen.
       *
       * @param name The name of the species.
       * @return     The amount of the species in question.
       */ 
    inline double getAmountByNameDouble(const std::string& name) const
      { return species[getIndex(name)].getAmountDouble(); }


      /**
       * Gets the diffusion constant of a species into a given
       * compartment.
       *
       * This method returns the diffusion constant of the given
       * species into a given compartment.  The index must be in range
       * and the compartment must be defined, otherwise strange things
       * may happen.
       *
       * @param index       The index of the species.
       * @param compartment The id of the compartment to diffuse into.
       * @return            The diffusion constant.
       */ 
    inline double getDiffusion(
      const size_t index, const std::string& compartment) const
      { return species[index].getDiffusion(compartment); }

      /**
       * Gets the id of the given species.
       *
       * This method returns the id of the given species.  The index
       * must be ensured to be in range, otherwise strange things may
       * happen.
       *
       * @param index The index of the species.
       * @return      The id of the species in question.
       */ 
    inline std::string getId(off_t index) const
      { return species[index].getId(); }

      /**
       * Gets the index of the given species.
       *
       * This method returns the index of the given species.  If the
       * key, that is the species name, is not defined anywhere in the
       * species table, the method returns a negative value.
       *
       * @param key The key of the species.
       * @return    The index of the species, or negative if undefined.
       */ 
    off_t getIndex(const std::string& key) const;
 
     /**
       * Gets the index of the given species.
       *
       * This method returns the index of the given species.  If the
       * id is not defined anywhere in the species table, the method
       * returns a negative value.
       *
       * @param id The id of the species.
       * @return   The index of the species, or negative if undefined.
       */ 
    off_t getIndexById(const std::string& id) const;

      /**
       * Gets the key of the given species.
       *
       * This method returns the name of the given species.  The index
       * must be ensured to be in range, otherwise strange things may
       * happen.  Note that for species, the name is key.
       *
       * @param index The index of the species.
       * @return      The name of the species in question.
       */ 
    inline std::string getKey(off_t index) const
      { return species[index].getKey(); }

      /**
       * Gets the number of species in the species table.
       *
       * Returns the number of species in the species table.  This is
       * just the size of the species table.
       *
       * @return The number of species in the species table.
       */
    inline size_t getSpeciesNumber() const
      { return species.size(); }


      /**
       * Change the number of a certain species.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  id     The id of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    double
    changeAmountById(const std::string& id, const double decrement);


      /**
       * Change the number of a certain species.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  id     The id of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    unsigned int
    changeAmountById(const std::string& id, const int change);


      /**
       * Change the number of a certain species.
       *
       * Note that no checking is performed on the index and the
       * range.  If the index is out of range, the program may break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the subvolume.
       */
    inline double
    changeAmountByIndex(const off_t index, const double change)
      {
        return species[index].changeAmount(change);
      }


      /**
       * Change the number of a certain species.
       *
       * Note that no checking is performed on the index and the
       * range.  If the index is out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the subvolume.
       */
    inline unsigned int
    changeAmountByIndex(const off_t index, const int change)
      {
        return species[index].changeAmount(change);
      }


      /**
       * Change the number of a certain species.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  name   The name of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the subvolume.
       */
    double
    changeAmountByName(const std::string& name, const double change);


      /**
       * Change the number of a certain species.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  name   The name of the requested species.
       * @param  change The change
       * @return        The updated number of species in the subvolume.
       */
    unsigned int
    changeAmountByName(const std::string& name, const int change);


      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * Note that no checking is performed on the index and the
       * range.  If the index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline unsigned int setAmountByIndex(off_t index, unsigned int amount)
      { return species[index].setAmount(amount); }

      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  id      The id of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    unsigned int setAmountById(const std::string& id, unsigned int amount);

      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  name    The name of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    unsigned int setAmountByName(const std::string& name, unsigned int amount);

      /**
       * Updates the species as identified by the key with the given
       * values.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.  Otherwise all values except the name are
       * replaced with the values as given in the argument list.
       *
       * @param d The new diffusion vector.
       * @param i The new id.
       * @param n The name of the species to update.
       * @param c The new compartment.
       * @param a The new amount.
       */
    void updateSpecies(
      const std::map<std::string, double>& d,
      const std::string&                   i,
      const std::string&                   n,
      const std::string&                   c,
      unsigned int                         a);

      /**
       * Updates the species as identified by the key with the given
       * values.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param s The new species that will replace the old with the
       *          same name.
       */
    void updateSpecies(const Species& s);


      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * Note that no checking is performed on the index and the
       * range.  If the index is out of range, the program may break.
       *
       * @param  index   The index of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    inline double setAmountByIndex(off_t index, double amount)
      { return species[index].setAmount(amount); }

      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  id      The id of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    double setAmountById(const std::string& id,double amount);

      /**
       * Set the number of a certain species to an arbitrary value.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.
       *
       * @param  name    The name of the requested species.
       * @param  amount  The new absolute amount of the species.
       * @return         The updated number of species in the sub volume.
       */
    double setAmountByName(const std::string& name, double amount);

      /**
       * Updates the species as identified by the key with the given
       * values.
       *
       * If the species is not found, nothing is done and the method
       * will return zero.  Otherwise all values except the name are
       * replaced with the values as given in the argument list.
       *
       * @param d The new diffusion vector.
       * @param i The new id.
       * @param n The name of the species to update.
       * @param c The new compartment.
       * @param a The new amount.
       */
    void updateSpecies(
      const std::map<std::string, double>& d,
      const std::string&                   i,
      const std::string&                   n,
      const std::string&                   c,
      double                               a);

    
      /**
       * Put a species table to an output stream.
       *
       * This method will put a species table onto an output stream.
       * The result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param s  The species table in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (std::ostream& os, const SpeciesTable& s);
  };
}

#endif
