/* The SubVolume class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SubVolume.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <stdio.h>

#include <cmath>
#include <iostream>
#include <limits>

#include "Compartment.hpp"
#include "SubVolume.hpp"


  /* From the mv++ package.
   */
#ifdef DETERMINISTIC
#  include "mv/mvvd.hpp"
#endif


  /* Initialise the static data members.  Apparently this is not
   * required on some compilers.  However, the ANSI standard says that
   * static data members should always be initialised.
   */
size_t                  MesoRD::SubVolume::neighbours         = 0;
MesoRD::Random::Random* MesoRD::SubVolume::prng               = 0;
double                  MesoRD::SubVolume::size               = 0;
bool                    MesoRD::SubVolume::useDoubleInSpecies = false;


  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* For technical reasons one should always provide a non-inline
   * implementation for virtual destructors, even if it is trivial.
   * This will make sure that the compiler generates a vtable.
   * Without a vtable, there will be unresolved symbols when it comes
   * to linking.
   */ 
MesoRD::SubVolume::~SubVolume()
{
  if (useDoubleInSpecies)
    delete species.real;
  else
    delete species.integer;
}



  /********************
   * EVENT MANAGEMENT *
   ********************/

  /* Note that we may safely reuse the random number drawn for
   * determining weather we are dealing with diffusion or reaction.
   * We must, however, remember to rescale it properly.
   */
MesoRD::SubVolume*
MesoRD::SubVolume::event()
{
  if (R_diffusion > 0 || R_reaction > 0 || R_neighbour_reaction > 0) {
    const double rand      = prng -> randomClosed();
    const double prob_rxn  = R_reaction / (R_diffusion + R_reaction + R_neighbour_reaction);
    const double prob_diff = R_diffusion / (R_diffusion + R_reaction + R_neighbour_reaction);

    if (rand < prob_diff) {
      return eventDiffusion((rand - 0.0) / prob_diff);

    } else { 
      if(rand < prob_diff + prob_rxn) {
        eventReaction((rand - prob_diff) / (prob_rxn));
        return 0;
      } else {
        return eventNeighbourReaction((rand - (prob_diff + prob_rxn))/(1 - (prob_diff + prob_rxn)));
      }
    }
  }

  return 0;
}


  /* This code should actually be very similar to the code for
   * executing diffusion events.  Note that we want to ensure that
   * reactions that have zero probability of happening do not actually
   * happen.
   */
MesoRD::SubVolume*
MesoRD::SubVolume::eventReaction(const double rand)
{
  if (R_reaction > 0) {
    double       R_cum    = 0;
    const double rand_tot = rand * R_reaction;
    
    std::valarray<bool> multi_sv_reactions_ass = compartment 
      -> reactionTable.getMultiSVReactionAssArray();      
    
    std::valarray<bool> multi_sv_reactions_diss = 
      compartment -> reactionTable.getMultiSVReactionDissArray();
    
    std::valarray<double> rates = compartment
      -> reactionTable.getRatesArray(*species.integer);
    
    for (size_t i = 0; i < rates.size(); i++) {
      if( multi_sv_reactions_ass[i] == 0) {
        R_cum += rates[i];
        
        if (rates[i] > 0 && rand_tot <= R_cum) {
          
          std::valarray<int> change =
            compartment -> reactionTable.getSpeciesChangeVector(i);
          
          if( multi_sv_reactions_diss[i] == 1) {
            std::valarray<unsigned int> products = compartment -> reactionTable.getProducts(i);
            std::valarray<unsigned int> reactants = compartment -> reactionTable.getReactants(i);
            
            double diss_rand_dir = prng -> randomClosed();
            double diss_rand_prod = prng -> randomClosed();
            
            double direction = ((double)1) / ((double)(neighbours+1));
            for( size_t k = 0; k < neighbours; k++) {
              if( diss_rand_dir < direction ) {
                if( diss_rand_prod >= 0.5) {
                  neighbours_array[k] -> changeSpecies(products[0], change[products[0]]);
                  changeSpecies(products[1], change[products[1]]);
                  changeSpecies(reactants[0], change[reactants[0]]);
                  return neighbours_array[k];
                } else {
                  neighbours_array[k] -> changeSpecies(products[1], change[products[1]]);
                  changeSpecies(products[0], change[products[0]]);
                  changeSpecies(reactants[0], change[reactants[0]]);
                  return neighbours_array[k];
                }
              } else {
                direction += direction;
              }
            }
              /* This is the last direction, ie the same subvolume. */
            changeSpecies(products[1], change[products[1]]);
            changeSpecies(products[0], change[products[0]]);
            changeSpecies(reactants[0], change[reactants[0]]);
            return 0;
            
            
          } else {
            for (size_t j = 0; j < species.integer -> size(); j++)
              changeSpecies(j, change[j]);
            return 0 ;
          }
        }
      }
    }
  }
}



  /****************************
   * DETERMINISTIC SIMULATION *
   ****************************/
#ifdef DETERMINISTIC

  /* This will add the reaction rates to the diffusion vector and and
   * set the total derivative to the derivative vector.
   *
   * FIXME!! Hattne doesn't understand this!
   */
void
MesoRD::SubVolume::getDerivativeVector(MV_Vector_double& theState,
  MV_Vector_double& derivative, size_t writePosition,
  std::valarray<double>& tempState)
{
  tempState.resize(getSpeciesNumber());

  for (size_t k = 0; k < tempState.size(); k++)
    tempState[k] = theState(static_cast<int>(writePosition + k));

  std::valarray<double> rates =
    (compartment -> reactionTable).getRatesArray(tempState);

  for(size_t j = 0; j < rates.size(); j++) {
    if(rates[j] > 0) {
      std::valarray<int> change =
        (compartment -> reactionTable).getSpeciesChangeVector(j);

      for (size_t i = 0; i < getSpeciesNumber(); i++) {
        derivative(static_cast<int>(writePosition + i))
	  += rates[j] * change[i];
      }
    }
  }
}
#endif //DETERMINISTIC
