/* The SubVolume class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SubVolume.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_SUBVOLUME_HPP
#define MESORD_SUBVOLUME_HPP 1

#include <sys/types.h>

#include <stdexcept>
#include <valarray>
#include <vector>
#include <limits>

#ifdef HAVE_CONFIG_H //|| defined(WIN32)
#  include "../config.h"
#endif

#include "Compartment.hpp"
#include "csg/CSGPoint.hpp"
#include "random/Random.hpp"
#include "definition.hpp"

#ifdef DETERMINISTIC
#  include "mv/mvvd.hpp"
#endif

namespace MesoRD
{
    /**
     * @brief Abstract class for modelling subvolumes.
     *
     * There are a few aspects of subvolumes that do not depend on the
     * particular geometry.  For instance, reaction events can be
     * handled the same way irrespective of the geometry of the
     * subvolume, whereas diffusion events cannot.  This class
     * attempts to cover those parts that are independent of geometry.
     * There are some general points to note about this class and its
     * subclasses.
     *
     * <ul>
     *
     * <li>The subvolume size -- the interpretation of which is
     * defined by the particular geometry of the sub volume -- is a
     * <em>static</em> member of this class.  This means that all
     * subvolumes must have the same size.  In practise, it also means
     * that all subvolumes in the system must be of the same
     * kind.</li>
     *
     * <li>We never do any, possibly time-consuming, logging or sanity
     * checks on the inputs.  If the numbers fed to this class are
     * wrong, there is probably a mistake somewhere else in the
     * program, and aborting execution entirely is probably just as
     * good as trying to give misleading error messages.</li>
     *
     * <li>We also do not try to catch any errors here.  If there are
     * errors, than there is a bug in the program, and we may as well
     * let the application cause a segmentation violation and make
     * this fact painfully obvious.</li>
     *
     * <li>The SubVolume class heavily relies on the STL valarray
     * template for array data storage.  The valarray class is a
     * strange, but standardised, beast in need of delicate taming.
     * For instance, it is safe to copy valarrays using the usual
     * assignment operator, but only as long as one knows the
     * valarrays are of the same size.  If they are not, the result is
     * undefined.</li>
     *
     * </ul>
     *
     * It is perhaps not obvious why each subvolume needs to keep
     * track of its position in the queue.  After all, the queue keeps
     * track of the order of all the subvolumes, so we are really
     * storing the same information in two different places.  The
     * reason is that in the case of diffusion events we need to
     * quickly find the target subvolume in the queue, so that it can
     * be resorted.  If the subvolume knows its queue position we need
     * not scan the entire queue to find it, but can look it up in
     * constant time.
     */
  class SubVolume
  {

  protected:

      /**
       * @brief The neighbours array holds pointers to the neighbours
       * of the subvolume.
       *
       * The neighbours array will always hold <code>neighbours</code>
       * elements.
       */
    std::valarray<SubVolume*> neighbours_array;

      /**
       * @brief The diffusion vector holds diffusion rates for
       * directions or species.
       *
       * In deterministic simulation all subvolumes hold a vector of
       * diffusions rates in all direction out off the subvolume.
       *
       * The diffusion vector in stochastic simulation stores the
       * accumulated diffusion rates for each species: the probability
       * that species <code>i</code> will diffuse to a neighbouring
       * subvolume is proportional to
       * <code>diffusion_vector[i]</code>.  The reason for this
       * overhead storage is that we may save on some potentially
       * expensive memory accesses when computing diffusion rates.
       * Whenever the number of species a subvolume can accommodate is
       * changed, the diffusion vector must be updated.
       */
    std::valarray<double> diffusion_vector;

      /**
       * @brief The species vector which holds the chemical
       * composition of the subvolume.
       *
       * In stochastic simulation, there will be an integer number of
       * each species in the subvolume.  In deterministic simulation,
       * the number of species can be any real number.  This union
       * member holds pointers to the two different valarrays used for
       * this purpose: a valarray of int for the stochastic models and
       * valarray of double for the deterministic model.  Since only
       * the pointers are stored here, an instance in stochastic
       * simulation will require as much space as an instance in
       * deterministic simulation
       */
    union {
      std::valarray<unsigned int>* integer;
      std::valarray<double>*       real;
    } species;

      /**
       * @brief A pointer to the compartment to which this subvolume
       * belongs.
       */
    Compartment* compartment;

      /**
       * @brief The current position in the event queue of the
       * subvolume.
       */
    size_t queue_index;

      /**
       * @brief The cumulative diffusion rate for the current
       * composition of the subvolume.
       *
       * This is proportional to the probability that the next event
       * in the subvolume will be a diffusion event.
       */
    double R_diffusion;

    
    
      /**
       * @brief The cumulative reaction rate for the current
       * composition of the subvolume.
       *
       * This is proportional to the probability that the next event
       * in the subvolume will be a reaction event.
       */
    double R_reaction;
    
    double R_neighbour_reaction;

      /**
       * @brief The absolute time when the next event in the subvolume
       * will occur.
       */
    double time;

      /**
       * @brief The <code>x</code>-coordinate of the subvolume.
       *
       * A subvolume has a location in space.  This attribute holds
       * the Cartesian <code>x</code>-coordinate, in arbitrary units,
       * of the centre of the subvolume.
       */
    float x;

      /**
       * @brief The <code>y</code>-coordinate of the subvolume.
       *
       * A subvolume has a location in space.  This attribute holds
       * the Cartesian <code>y</code>-coordinate, in arbitrary units,
       * of the centre of the subvolume.
       */
    float y;

      /**
       * @brief The <code>z</code>-coordinate of the subvolume.
       *
       * A subvolume has a location in space.  This attribute holds
       * the Cartesian <code>z</code>-coordinate, in arbitrary units,
       * of the centre of the subvolume.
       */
    float z;

      /**
       * @brief The number of neighbours of the subvolume.
       *
       * The number of neighbours of a subvolume depends on the
       * geometry of a subvolume.  For instance, a cubic subvolume has
       * six neighbours, while a truncated octahedron has fourteen.
       */
    static size_t neighbours;

      /**
       * @brief A pointer to the pseudo-random number generator.
       *
       * All instances of the Subvolume class share the same instance
       * of the Random class.
       */
    static Random::Random* prng;

      /**
       * @brief The size, given in units of centimetres, of the
       * current subvolume.
       *
       * The exact interpretation of the size of a subvolume depends
       * on the particular geometry of the subvolume.  For instance,
       * for cubic subvolumes, the size gives the side length of the
       * cube.  Since this is a static attribute, all subvolumes will
       * have the same size.
       */
    static double size;

      /**
       * @brief A flag to indicate whether the chemical composition is
       * represented by floating point numbers or not.
       *
       * This flag indicates whether we use floating point or integer
       * arithmetic for the representation of the number of each
       * species in the subvolume.  For deterministic simulation, we
       * represent the chemical composition by floating point numbers.
       * In the case of stochastic simulation, we use integer
       * arithmetic.
       */
    static bool useDoubleInSpecies;


  protected:

      /**
       * @brief Update the diffusion constants of the current
       * subvolume.
       *
       * Updating diffusion constants depends on the geometry of the
       * subvolume.  Alas, this function cannot be implemented here.
       * Any concrete subclass must implement this.
       */
    virtual void
    updateDiffusionConstants() = 0;

      /**
       * @brief Update the reaction rates due to neighbour reactions
       * of the current subvolume.  These reactions will depend on the
       * concentrations in the adjecent subvolumes.
       *
       * Updating the rates for molecules with large reaction radius
       * in comparision to the subvolume size.  In these cases the
       * reaction rates will depend on the concentration in
       * neighbouring subvolumes.
      */
    virtual void
    updateNeighbourReactionConstants() = 0;
    

      /**
       * @brief Update the reaction constants of the current
       * subvolume.
       *
       * Calling this function will recompute the reaction constants.
       * These are needed to determine what will happen next in the
       * subvolume, and when it will happen.  Whenever the composition
       * of the subvolume is changed, one must call this function to
       * update the reaction constants accordingly.  Note that we do
       * not need any random numbers in order to find new reaction
       * constants.  The function includes spatial correction, where a
       * reaction may have limited probablity of reacting inside a
       * subvolume.  The probability to react inside the subvolume is
       * restricted by P_intra_sv_reaction in file_name.hpp
       */
    inline void
    updateReactionConstants()
      {
        R_reaction = 0;

        std::valarray<bool> multi_sv_reactions = 
          compartment -> reactionTable.getMultiSVReactionAssArray();      
        std::valarray<double> reaction_rates = 
          compartment -> reactionTable.getRatesArray(*species.integer);

          /* Here only the reactions that are non-multisubvolume
           * association reactions should be counted.  */
        for( unsigned int i = 0; i < reaction_rates.size(); i++) {
          if( multi_sv_reactions[i] == 0) {
            R_reaction += reaction_rates[i];
          }
        }
      }


  public:

      /**
       * @brief Default constructor for the SubVolume class
       *
       * The constructor will allocate and initialise the data needed
       * for a single subvolume.  A subvolume can hold an integer
       * number of unique chemical species.  Although it is possible
       * to change the number of different species runtime, it is not
       * recommended, since memory management can be expensive.
       *
       * @param r Pointer to the pseudo-random number generator.
       * @param s Size of the subvolume in centimetres.
       * @param n The number of neighbours of the subvolume.
       * @param m The number of species the subvolume can hold.
       */
    SubVolume(
      Random::Random*      r,
      const double         s,
      const size_t         n,
      const unsigned short m)
      : neighbours_array(n),
        diffusion_vector(m),
        compartment(0),
        queue_index(0),
        R_diffusion(0),
        R_reaction(0),
        R_neighbour_reaction(0),
        time(0),
        x(0),
        y(0),
        z(0)
      {
        neighbours_array = this;
        neighbours = n;
        prng       = r;
        size       = s;
        if (useDoubleInSpecies)
          species.real    = new std::valarray<double>(m);
        else
	  species.integer = new std::valarray<unsigned int>(m);
      }


      /**
       * @brief Destructor for the SubVolume class.
       *
       * The destructor will delete the species vector, which holds
       * the chemical composition of the subvolume.  For technical
       * reasons, destructors of abstract classes should not be
       * defined inline.
       */
    virtual ~SubVolume();


      /**
       * @brief Make an addition to the diffusion vector for a given
       * species.
       *
       * In deterministic simulation all subvolumes will hold a vector
       * of diffusions in all direction out of the subvolume.  This
       * function allows other subvolumes to modify the diffusion
       * vector of the current subvolume.  In particular, neighbours
       * can make addition to the diffusion vector for a particular
       * species.  The diffusion vector in deterministic simulation is
       * treated as the total change in each direction.
       *
       * @param index The direction in which the diffusion should be
       *              added.
       * @param rate  The diffusion rate to add.
       */
//     virtual void
//     addDiffusionRate(const off_t index, const double rate) = 0;


      /**
       * @brief Clear all the values in the diffusion vector.
       *
       * In deterministic simulation all subvolumes will hold a vector
       * of diffusions in all direction out off the subvolume. This
       * function sets all these diffusion rates to zero.
       *
       */
//     virtual void
//     clearDiffusionVector() = 0;


      /**
       * @brief Changes the number of a certain species in a
       * subvolume.
       *
       * MesoRD keeps track of all the species globally as well as
       * locally.  Therefore, when modifying the contents of a
       * subvolume, this function makes sure to reflect the change not
       * only in the current subvolume, but globally as well.  Note
       * that no checking is performed on the index: if the index is
       * out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline double
    changeSpecies(const off_t index, const double change)
      {
        if (change > 0.0
          || change < 0.0 && -1.0 * change <= (*species.real)[index]) {
          compartment -> changeSpecies(index, change);
          (*species.real)[index] += change;
        }
        return (*species.real)[index];
      }


      /**
       * @brief Changes the number of a certain species in a
       * subvolume.
       *
       * MesoRD keeps track of all the species globally as well as
       * locally.  Therefore, when modifying the contents of a
       * subvolume, this function makes sure to reflect the change not
       * only in the current subvolume, but globally as well.  Note
       * that no checking is performed on the index: if the index is
       * out of range, the program will break.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline unsigned int
    changeSpecies(const off_t index, const int change)
      {
        if (change > 0 || change < 0
          && -1 * change <= static_cast<int>((*species.integer)[index])) {
          compartment -> changeSpecies(index, change);
          (*species.integer)[index] += change;
        }
        return (*species.integer)[index];
      }


      /**
       * @brief Change the number for a certain species without
       * incrementing the total number of molecules in the
       * compartment.
       *
       * This has the same function as changeSpecies() except that
       * this function does <em>not</em> increase the global number of
       * molecules in the compartment.  This is used during the
       * initial distribution of molecules into subvolumes, when the
       * total number of molecules is already known to the compartment
       * and should instead be distributed to the subvolumes.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline double
    distributeSpecies(const off_t index, const double change)
      {
        if (change > 0.0
          || change < 0.0 && -1.0 * change <= (*species.real)[index])
          (*species.real)[index] += change;
        return (*species.real)[index];
      }


      /**
       * @brief Change the number for a certain species without
       * incrementing the total number of molecules in the
       * compartment.
       *
       * This has the same function as changeSpecies() except that
       * this function does <em>not</em> change the global number of
       * molecules in the compartment.  This is used during the
       * initial distribution of molecules into subvolumes, when the
       * total number of molecules is already known to the compartment
       * and should instead be distributed to the subvolumes.
       *
       * @param  index  The index of the requested species.
       * @param  change The change.
       * @return        The updated number of species in the
       *                subvolume.
       */
    inline unsigned int
    distributeSpecies(const off_t index, const int change)
      {
        if (change > 0 || change < 0
          && -1 * change <= static_cast<int>((*species.integer)[index]))
          (*species.integer)[index] += change;
        return (*species.integer)[index];
      }


      /**
       * @brief Compute a general event in the subvolume.
       *
       * An event may either be a diffusion event or a reaction event.
       * When this function is called, we determine, using a single
       * random number, which kind of event will occur next.  This
       * requires that the reaction and diffusion constants reflect
       * the current composition of the subvolume.  Having found the
       * next event, we execute it: if it is a reaction, the
       * composition of the current subvolume alone will change
       * through a call to eventReaction().  If it is a diffusion
       * event, eventDiffusion() is called, and a molecule will move
       * from this subvolume to another, so that in total two
       * subvolumes change.  The function returns a pointer to a
       * subvolume, which, apart from the current subvolume, must be
       * re-sorted due to the change in chemical composition.  If no
       * other subvolumes than the current will need re-sorting, that
       * is, if the event was a reaction, the function returns
       * <code>null</code>.  The function will only return
       * non-<code>null</code> in the case of diffusion events.
       *
       * @return Pointer to the target subvolume in case of
       *         diffusion, <code>null</code> otherwise.
       */
    SubVolume*
    event();


      /**
       * @brief Execute a diffusion event in the subvolume.
       *
       * A diffusion event depends on the precise geometry of the
       * subvolume.  The implementation of this function will
       * therefore depend on the kind of subvolume used.  Any concrete
       * class inhering from this class must therefore implement the
       * eventDiffusion() function.
       *
       * @param  rand A random number from zero to one inclusive.
       * @return      Pointer to the target subvolume.
       */
    virtual SubVolume* eventDiffusion(const double rand) = 0;


      /**
       * @brief Execute a reaction event in the subvolume.
       *
       * Using the supplied random number, a reaction is chosen.  This
       * will then change the chemical composition in the subvolume
       * according to the details of the particular reaction.
       *
       * @param rand A random number from zero to one inclusive.
       */
    SubVolume* eventReaction(const double rand);


      /**
       * @brief Execute a reaction event in the subvolume, using the
       * species number from neighbouring subvolumes.
       * 
       * Using the supplied random number a reaction is sampled.  The
       * sampled reaction will change the chemical composition in this
       * subvolume and the neighbouring subvolume choosen in the
       * reactions.
      */
    virtual SubVolume*
    eventNeighbourReaction(const double rand);
    
      /**
       * @brief Get a pointer to the compartment to which the
       * subvolume belongs.
       *
       * Every sub volume belongs to a compartment.  Calling this
       * function returns a pointer to the compartment of this
       * particular subvolume.
       *
       * @return A pointer to the compartment to which the subvolume
       *         belongs.
       */
    inline Compartment*
    getCompartment() const
      {
        return compartment;
      }


      /**
       * @brief Get a pointer to the @p i:th neighbour of the
       * subvolume.
       *
       * Returns a pointer to the @p i:th neighbour of the subvolume.
       * Since layout of the neighbours varies with the particular
       * type of subvolume involved, this function cannot be
       * implemented here.
       *
       * @param  i The index of the neighbour to get.
       * @return   A pointer to the @p i:th neighbour of the subvolume.
       */
    virtual SubVolume*
    getNeighbour(const size_t i) const = 0;

    
      /**
       * @brief Get a reference to the array of subvolumes.  
       * 
       * Returns a reference to the valarray of neighbour subvolumes
       * that is stored in each subvolume.  
       *
       * @return A reference to the list of neighbours.  
      */
    virtual std::valarray<SubVolume*>& 
    getNeighbourArray() = 0;
    

      /**
       * @brief Get the number of neighbours of the subvolume.
       *
       * Returns the number of neighbours of the subvolume.  This
       * number varies with the particular type of subvolume involved.
       * Inheriting classes will set it by means of the constructor.
       *
       * @return The number of neighbours of the subvolume.
       */
    inline static size_t
    getNeighboursNumber()
      {
        return neighbours;
      }


      /**
       * @brief Get the queue index of the subvolume.
       *
       * The System class stores pointers to all subvolumes, sorted in
       * order of increasing time until the next event.  The queue
       * index, as returned by this function, indicates the position
       * of the current subvolume in the time queue.  Manipulating
       * queue positions is really what sorting is all about.
       *
       * @return The position in the time queue.
       */
    inline size_t
    getQueueIndex() const
      {
        return queue_index;
      }


      /**
       * @brief Get the size, in centimetres, of the subvolume.
       *
       * How the size of a subvolume is to be interpreted depends on
       * its particular geometry.  For instance, in the case of cubic
       * subvolumes, the size is the side length of the subvolume.  It
       * is important to remember that the size of a sub volume is
       * given in centimetres.
       *
       * @return The size of the subvolume in centimetres.
       */
    inline double
    getSize() const
      {
        return size;
      }


      /**
       * @brief Get the number of a certain species in a subvolume.
       *
       * This function will return the number of molecules of a
       * certain species using floating point arithmetic for
       * deterministic simulation.  Note that no checking is performed
       * on the index and the range.  If the index is out of range,
       * the program will break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the
       *                subvolume.
       */
    inline double
    getSpeciesDouble(const off_t index) const
      {
        return (*species.real)[index];
      }


      /**
       * @brief Get the number of a certain species in a subvolume.
       *
       * This function will return the number of molecules of a
       * certain species in the subvolume using integer arithmetic for
       * stochastic simulation.  Note that no checking is performed on
       * the index and the range.  If the index is out of range, the
       * program will break.
       *
       * @param  index  The index of the requested species.
       * @return        The number of the requested species in the
       *                subvolume.
       */
    inline unsigned int
    getSpecies(const off_t index) const
      {
        return (*species.integer)[index];
      }

    
    inline std::valarray<unsigned int>* 
    getSpeciesArray() const
      {
        return species.integer;
      }
    
      /**
       * @brief Get the number of different species in the subvolume.
       *
       * This function returns the number of different species the
       * subvolume can accommodate.  This number does not reveal
       * anything about the actual chemical composition of the
       * subvolume.
       *
       * @return The number of different species the subvolume can
       *         hold.
       */
    inline size_t
    getSpeciesNumber() const
      {
        return useDoubleInSpecies
          ? species.real -> size() : species.integer -> size();
      }


      /**
       * @brief Get the absolute time of the next event in the
       * subvolume.
       *
       * Note that the time returned is not the time <em>until</em>
       * the next event, but the absolute time <em>of</em> the next
       * event.  This absolute time is measured from time zero, which
       * is the starting time of the simulation.
       *
       * @return The time of the next event in the subvolume.
       */
    inline double
    getTime() const
      {
        return time;
      }


      /**
       * @brief Get the Cartesian <code>x</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will return the Cartesian <code>x</code>-coordinate
       * of the centre of the subvolume.  The coordinate system in
       * which subvolumes live is not rigorously defined.
       *
       * @return The Cartesian <code>x</code>-coordinate of the
       *         subvolume's centre.
       */
    inline float
    getX() const
      {
        return x;
      }


      /**
       * @brief Get the Cartesian <code>y</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will return the Cartesian <code>y</code>-coordinate
       * of the centre of the subvolume.  The coordinate system in
       * which subvolumes live is not rigorously defined.
       *
       * @return The Cartesian <code>y</code>-coordinate of the
       *         subvolume's centre.
       */
    inline float
    getY() const
      {
        return y;
      }


      /**
       * @brief Get the Cartesian <code>z</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will return the Cartesian <code>z</code>-coordinate
       * of the centre of the subvolume.  The coordinate system in
       * which subvolumes live is not rigorously defined.
       *
       * @return The Cartesian <code>z</code>-coordinate of the
       *         subvolume's centre.
       */
    inline float
    getZ() const
      {
        return z;
      }


      /**
       * @brief Poll the flag that indicates whether the chemical
       * composition is represented by floating point numbers or not.
       *
       * In deterministic simulation, we represent the chemical
       * composition by floating point numbers.  In the case of
       * stochastic simulation, we can use integer arithmetic.
       *
       * @return True for floating point arithmetic, false for integer
       *         arithmetic.
       */
    inline static bool
    getUseDoubleInSpecies()
      {
        return useDoubleInSpecies;
      }


      /**
       * @brief Set the pointer to the compartment to which the
       * subvolume belongs.
       *
       * Every subvolume belongs to a compartment.  Calling this
       * function updates the compartment of this particular
       * subvolume.
       *
       * @param c A pointer to the new compartment.
       */
    inline void
    setCompartment(Compartment* c)
      {
        compartment = c;
      }


      /**
       * @brief Set the pointer to the @p i:th neighbour of the
       * subvolume.
       *
       * This function will set or update the neighbour in the @p i:th
       * direction.  Since the layout of the neighbours varies with
       * the particular type of subvolume involved, this function
       * cannot be implemented here.
       *
       * @param  i The index of the neighbour to set.
       * @param  n A pointer to the new neighbour.
       */
    virtual void
    setNeighbour(const size_t i, SubVolume* n) = 0;


      /**
       * @brief Set the queue index of the subvolume.
       *
       * The System class stores pointers to all subvolumes, sorted in
       * order of increasing time until the next event.  The queue
       * index, as set by this function, indicates the position of the
       * current subvolume in the time queue.  Manipulating queue
       * positions is really what sorting is all about.
       *
       * @param index The position in the time queue.
       */
    inline void
    setQueueIndex(const off_t index)
      {
        queue_index = index;
      }


      /**
       * @brief Sets the size, in centimetres, of the subvolume.
       *
       * How the size of a subvolume is to be interpreted depends on
       * its particular geometry.  For instance, in the case of cubic
       * subvolumes, the size is the side length of the subvolume.  It
       * is important to remember that the size of a subvolume is
       * given in centimetres.
       *
       * @param s The new size of the subvolume.
       */
    inline static void
    setSize(const double s)
      {
        size = s;
      }


      /**
       * @brief Set the number of a certain species in a subvolume to
       * an arbitrary value.
       *
       * This function will set the number of species for
       * deterministic simulation.  Note that no checking is performed
       * on the index and the range.  If the index is out of range,
       * the program will break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the
       *                 subvolume.
       */
    inline double
    setSpecies(const off_t index, const double number)
      {
        return (*species.real)[index] = number;
      }


      /**
       * @brief Set the number of a certain species in a subvolume to
       * an arbitrary value.
       *
       * This function will set the number of species for stochastic
       * simulation.  Note that no checking is performed on the index
       * and the range.  If the index is out of range, the program
       * will break.
       *
       * @param  index   The index of the requested species.
       * @param  number  The new absolute value of the species.
       * @return         The updated number of species in the
       *                 subvolume.
       */
    inline unsigned int
    setSpecies(const off_t index, const unsigned int number)
      {
        return (*species.integer)[index] = number;
      }


      /**
       * @brief Set the number of different species a subvolume may
       * contain.
       *
       * Just like the resize() function of many of the container
       * classes in the STL, this function changes the capacity of the
       * subvolume.  Incidentally, all this function does is to call
       * resize() on the container holding the species.
       *
       * @param n The new number of species a subvolume will be able
       *          to hold.
       */
    inline virtual void
    setSpeciesNumber(const size_t n)
      {
        useDoubleInSpecies
          ? species.real -> resize(n) : species.integer -> resize(n);
      }


      /**
       * @brief Set the flag that indicates whether the chemical
       * composition is represented by floating point numbers or not.
       *
       * In deterministic simulation, we represent the chemical
       * composition by floating point numbers.  In the case of
       * stochastic simulation, we can use integer arithmetic.
       *
       * @param udis True for floating point arithmetic, false for
       *             integer arithmetic.
       */
    inline static void
    setUseDoubleInSpecies(const bool udis)
      {
        useDoubleInSpecies = udis;
      }


      /**
       * @brief Set the Cartesian <code>x</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will set the Cartesian <code>x</code>-coordinate of
       * the centre of the subvolume.  The coordinate system in which
       * subvolumes live is not rigorously defined.
       *
       * @param x_new The new Cartesian <code>x</code>-coordinate of
       *              the subvolume's centre.
       */
    inline void
    setX(const float x_new)
      {
        x = x_new;
      }


      /**
       * @brief Set the Cartesian <code>y</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will set the Cartesian <code>y</code>-coordinate of
       * the centre of the subvolume.  The coordinate system in which
       * subvolumes live is not rigorously defined.
       *
       * @param y_new The new Cartesian <code>y</code>-coordinate of
       *              the subvolume's centre.
       */
    inline void
    setY(const float y_new)
      {
        y = y_new;
      }


      /**
       * @brief Set the Cartesian <code>z</code>-coordinate of the
       * subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will set the Cartesian <code>z</code>-coordinate of
       * the centre of the subvolume.  The coordinate system in which
       * subvolumes live is not rigorously defined.
       *
       * @param z_new The new Cartesian <code>z</code>-coordinate of
       *              the subvolume's centre.
       */
    inline void
    setZ(const float z_new)
      {
        z = z_new;
      }


      /**
       * @brief Set the Cartesian <code>x</code>-, <code>y</code>- and
       * <code>z</code>-coordinates of the subvolume.
       *
       * Every subvolume has a location in space.  Calling this
       * function will set the Cartesian coordinates of the centre of
       * the subvolume.  The coordinate system in which subvolumes
       * live is not rigorously defined.
       *
       * @param x_new The new Cartesian <code>x</code>-coordinate of
       *              the subvolume's centre.
       * @param y_new The new Cartesian <code>y</code>-coordinate of
       *              the subvolume's centre.
       * @param z_new The new Cartesian <code>z</code>-coordinate of
       *              the subvolume's centre.
       */
    inline void
    setXYZ(const float x_new, const float y_new, const float z_new)
      {
        x = x_new;
        y = y_new;
        z = z_new;
      }


      /**
       * @brief Update the time until the next event in the subvolume.
       *
       * In order to find the time until the next event, we recompute
       * new diffusion and reaction constants.  We may then compute
       * the time interval to the next event by drawing a random
       * number from an exponential distribution.  The absolute time
       * until the next event is then found by adding this obtained
       * time interval to the current time.
       *
       * When drawing the random number from the exponential
       * distribution, we must weigh it using the diffusion and
       * reaction constants.  Because we are dividing by something
       * that could be zero -- and it will be zero if there are no
       * molecules in the sub volume -- we can gracefully avoid
       * division by zero by returning the system's idea of infinity
       * instead.  We use an open interval random number in order to
       * avoid possible oddities around the boundaries.
       *
       * @param time_now The current time in seconds.
       * @return         The absolute time until the next event.
       */
    inline double
    updateTime(const double time_now)
      {
        updateReactionConstants();
        updateNeighbourReactionConstants();
        updateDiffusionConstants();
        const double d = R_reaction + R_diffusion + R_neighbour_reaction;
        if (fabs(d) >= std::numeric_limits<double>::min())
          time = time_now - log(prng -> randomOpen()) / d;
        else
          time = std::numeric_limits<double>::infinity();
        return time;
      }


#ifdef DETERMINISTIC
      /**
       * @brief Get the reaction-derivative-vector for a subvolume.
       *
       * This method will write the reaction derivate for each species
       * to the derivative vector. The reaction derivative includes the
       * contribution from each reaction to the current state.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * @param theState Holds the current state.
       * @param derivative Vector to write the derivative result to.
       * @param writePosition Keeps track of where in the state vector we are.
       * @param tempState Temporary write space.
       */
    void
    getDerivativeVector(MV_Vector_double& theState,
      MV_Vector_double& derivative, size_t writePosition,
      std::valarray<double>& tempState);


      /**
       * @brief Update the current state by using an time step adaptive
       * Euler-forward.
       *
       * @param theState The current state of the system.  
       * @param timeStep The largest time step to use (system wide time step.  
       * @param writePosition The position in the state vector where this
       *                      subvolume should write its data.   
       */
//     void updatetateByReaction(MV_Vector_double& theState, double timeStep,
//       int writePosition);


      /**
       * @brief Get the vector containing the rate of diffusion in all
       * directions.
       * 
       * FIXME!! Hattne doesn't understand this!
       *
       * @return The total diffusion rate for each species.
       */
//     virtual std::valarray<double>&
//     getDiffusionVector() = 0;


      /**
       * @brief Updates the diffusion_vector.
       *
       * This will calculate to flow out from the current subvolume
       * for each species. Remove this from the current diffusion_vector
       * and add it to the correct neighbour.  
       */ 
//     virtual void updateDiffusionVectorDeterministic() = 0;


      /**
       * Read the a matrix which contains all the diffusion in each
       * direction.  Specified by a matrix.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * @param index The current species index.
       * @param subVolumeMap Map of subvolume coordinates,
       *                     used to find neighbours.
       * @param diffusionMatrixData Stores the diffusion rate.
       * @param diffusionMatrixRowIndex Stores the row index for the diffusion.
       * @param diffusionMatrixColIndex Stores the column index for the
       *                                diffusion.
       * @param numberOfSpecies Number of species in a subvolume.
       */
    virtual void
    getDiffusionVector(
      off_t                             index,
      std::map<CSGPoint, unsigned int>& subVolumeMap,
      std::vector<double>&              diffusionMatrixData,
      std::vector<size_t>&              diffusionMatrixRowIndex,
      std::vector<size_t>&              diffusionMatrixColIndex,
      int                               numberOfSpecies) = 0;


      /**
       * @brief Update the diffusion vector.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * This will calculate the flow out from the current subvolume
       * for each species.  Remove this from the current diffusion
       * vector and add it to the correct neighbour.
       */
//     virtual void
//     updateDiffusionVectorDeterministic() = 0;


      /**
       * Update the current state by using an time step adaptive
       * Euler-forward.
       *
       * FIXME!! Hattne doesn't understand this!
       *
       * @param theState The current state of the system.
       * @param timeStep The largest time step to use (system wide time step.
       * @param writePosition The position in the state vector where this
       *                      subvolume should write its data.
       */
    void
    updateStateByReaction(
      MV_Vector_double& theState, double timeStep, int writePosition);
#endif //DETERMINISTIC
  };
}
#endif /* ifndef MESORD_SUBVOLUME_H */
