/* The Symbol class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Symbol.hpp 540 2008-04-12 14:30:36Z davidfange $
 */
#ifndef MesoRD_Symbol_H
#define MesoRD_Symbol_H

#include <string>


namespace MesoRD
{


    /**
     * Base class for symbols.
     *
     * The objects of SBML have certain features in common.  In SBML
     * this is captured by the SBase object, from which all other
     * objects inherit.  Objects in MesoRD are used slightly
     * different, thus a different base object is in place.  This base
     * object is the "Symbol".  All symbols which MesoRD needs to care
     * about must have a key which uniquely defines the symbol in the
     * appropriate namespace.  In MesoRD this key corresponds to
     * either the object's SBML name or the object's SBML id.  Both of
     * these quantities are character strings.  Symbols are stored in
     * a symbol table.  The symbol table facilitates look-up of
     * symbols, just like the symbol table in a compiler.
     *
     * Symbols small and uncomplicated objects that are used
     * frequently throughout the MesoRD code, so speed is important.
     * We hope this is achieved by defining most of the implementation
     * inline.  Furthermore, this class will not throw any exceptions.
     */
  class Symbol 
  {
  protected:

      /**
       * Either the id or the name uniquely defines a symbol in its
       * appropriate namespace.  If a symbol is uniquely identified by
       * its id, the the id is said to be its key.  Just like in SBML,
       * the id is a string.  MesoRD, however, being a C++ program,
       * uses the standard string class to represent the id.
       */
    std::string id;

      /**
       * Either the id or the name uniquely defines a symbol in its
       * appropriate namespace.  If a symbol is uniquely identified by
       * its name, the the name is said to be its key.  Just like in
       * SBML, the name is a string.  MesoRD, however, being a C++
       * program, uses the standard string class to represent the id.
       */
    std::string name;

  public:

      /**
       * Value constructor for the Symbol class.
       *
       * This constructor will initialise the properties associated
       * with a symbol, namely the id and the name.  There are no
       * default values for the two parameters; one must specify them
       * explicitly when constructing a new symbol.
       *
       * @param id   The id of the new symbol.
       * @param name The name of the new symbol.
       */
    inline Symbol(const std::string& id, const std::string& name)
      : id(id), name(name)
      {}


      /**
       * Destructor for the Symbol class.
       *
       * Does nothing at all.  For technical reasons, destructors of
       * abstract classes should not be defined inline.
       */
    virtual ~Symbol();


      /**
       * Get the id of the symbol.
       *
       * This function will return the id of the symbol.  This is a
       * string that may or may not uniquely define the symbol in the
       * current namespace.
       *
       * @return The id of the symbol.
       */
    inline const std::string&
    getId() const
      {
        return id;
      }


      /**
       * Get the key of the symbol.
       *
       * The key is the property which uniquely defines the symbol in
       * its namespace.  The key can be either the id or the name.  A
       * class which inherits from this base class therefore must
       * define the getKey() function.  Typically, this would be done
       * by having it call and return the value from either the
       * getId() or the getName() function.
       *
       * @return The key of the symbol, which may be either its id or
       *         its name.
       */
    virtual const std::string&
    getKey() const = 0;


      /**
       * Get the name of the symbol.
       *
       * This function will return the name of the symbol.  This is a
       * string that may or may not uniquely define the symbol in the
       * current namespace.
       *
       * @return The name of the symbol.
       */
    inline const std::string&
    getName() const
      {
        return name;
      }


      /**
       * Check whether the symbol has a given id.
       *
       * This function return true if the symbol has the given id.
       * The id is a string that may or may not uniquely define the
       * symbol in the current namespace.
       *
       * @param id The id to compare against the symbol's id.
       * @return   True if the id:s match, false otherwise.
       */
    inline bool
    isId(const std::string& id) const
      {
        return this -> id == id;
      }


      /**
       * Check whether the symbol has a given key.
       *
       * This function return true if the symbol has the given id.
       * The key can be either the id or the name.  A class which
       * inherits from this base class therefore must define the
       * isKey() function.  Typically, this would be done by having it
       * call and return the value from either the isId() or the
       * isName() function.
       *
       * @param  key The key to compare against the symbol's key.
       * @return     True if the keys match, false otherwise.
       */
    virtual bool
    isKey(const std::string& key) const = 0;


      /**
       * Check whether the symbol has a given name.
       *
       * This function return true if the symbol has the given name.
       * The name is a string that may or may not uniquely define the
       * symbol in the current namespace.
       *
       * @param name The name to compare against the symbol's name.
       * @return     True if the names match, false otherwise.
       */
    inline bool
    isName(const std::string& name) const
      {
        return this -> name == name;
      }


      /**
       * Set the id of the symbol
       *
       * This function sets the id of the symbol to the supplied
       * value.  The id is a string that may or may not uniquely
       * define the symbol in the current namespace.
       *
       * @param id The new id of the symbol.
       */
    inline void
    setId(const std::string& id)
      {
        this -> id = id;
      }


      /**
       * Set the key of the symbol.
       *
       * This function sets the key of the symbol to the supplied
       * value.  The key can be either the id or the name.  A class
       * which inherits from this base class therefore must define the
       * setKey() function.  Typically, this would be done by having
       * it call and return the value from either the setId() or the
       * setName() function.
       *
       * @param key The new key of the symbol.
       */
    virtual void
    setKey(const std::string& key) = 0;


      /**
       * Set the name of the symbol.
       *
       * This function sets the name of the symbol to the supplied
       * value.  The name is a string that may or may not uniquely
       * define the symbol in the current namespace.
       *
       * @param name The new name of the symbol.
       */
    inline void
    setName(const std::string& name)
      {
        this -> name = name;
      }


      /**
       * Equality operator for symbols.
       *
       * This function will compare two symbols for equality.  Two
       * symbols are considered equal if and only if their keys are
       * identical.
       *
       * @param s The symbol to compare with.
       * @return  True if the symbols are equal, false otherwise.
       */
    inline bool
    operator==(const Symbol& s) const
      {
        return getKey() == s.getKey();
      }


      /**
       * Equality operator for symbols.
       *
       * This function will compare two symbols for inequality.  Two
       * symbols are considered equal if and only if their keys are
       * not identical.
       *
       * @param s The symbol to compare with.
       * @return  True if the symbols are not equal, false otherwise.
       */
    inline bool
    operator!=(const Symbol& s) const
      {
        return *this != s;
      }


      /**
       * Less-than operator for symbols.
       *
       * This function will compare two symbols to see if the left
       * hand symbol is less than the right hand symbol.  The
       * comparison is based on the keys of the two symbols only.
       *
       * @param s The symbol to compare with.
       * @return  True if the left hand symbol is less than the right
       *          hand symbol, false otherwise.
       */
    inline bool
    operator<(const Symbol& s) const
      {
        return getKey() < s.getKey();
      }


      /**
       * Less-than-or-equal operator for symbols.
       *
       * This function will compare two symbols to see if the left
       * hand symbol is less than or equal to the right hand symbol.
       * The comparison is based on the keys of the two symbols only.
       *
       * @param s The symbol to compare with.
       * @return  True if the left hand symbol is less than or equal to
       *          the right hand symbol, false otherwise.
       */
    inline bool
    operator<=(const Symbol& s) const
      {
        return *this < s || *this == s;
      }


      /**
       * Greater-than operator for symbols.
       *
       * This function will compare two symbols to see if the left
       * hand symbol is greater than the right hand symbol.  The
       * comparison is based on the keys of the two symbols only.
       *
       * @param s The symbol to compare with.
       * @return  True if the left hand symbol is greater than the right
       *          hand symbol, false otherwise.
       */
    inline bool
    operator>(const Symbol& s) const
      {
        return !(*this < s) && (*this != s);
      }


      /**
       * Greater-than-or-equal operator for symbols.
       *
       * This function will compare two symbols to see if the left
       * hand symbol is greater than or equal to the right hand
       * symbol.  The comparison is based on the keys of the two
       * symbols only.
       *
       * @param s The symbol to compare with.
       * @return  True if the left hand symbol is greater than or equal
       *          to the right hand symbol, false otherwise.
       */
    inline bool operator>=(const Symbol& s) const
      {
        return !(*this < s);
      }
  };
}
#endif
