/* The SymbolTable class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SymbolTable.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef MesoRD_SymbolTable_H
#define MesoRD_SymbolTable_H

#include <list>
#include <map>
#include <stdexcept>

#include "Symbol.hpp"


namespace MesoRD
{


    /**
     * Class for managing symbols.
     *
     * Symbols, as defined in the Symbol class, need to be stored and
     * retrieved in order to be useful.  The symbol table serves just
     * that purpose.  It stores and indexes symbols by their key
     * property.  Symbols can be looked up, by querying the symbol
     * table for symbols with a certain key, id or name.  Even though
     * the symbols are indexed by their key, it is possible to
     * retrieve symbols by id or name as well.  However, querying by a
     * property other than the key may be a bit more expensive.  While
     * a symbol table holds at most one symbol with a given key, there
     * may be several symbols with a certain name or id.  If one or
     * more symbols with the sought property exist in the table, they
     * are returned in an appropriate container.
     *
     * A symbol table corresponds to a single namespace for the type
     * of object in question.  Namespaces, however, can be nested.
     * Thus, symbol tables can be nested as well.  Nesting is
     * implemented by allowing each symbol table have a parent symbol
     * table.  This class has different functions in order to query
     * single symbol tables or complete stacks of symbol tables.  If
     * one only wishes to consider the current, or local, symbol table
     * one should call the getLocalX() functions.  The getX()
     * functions on the other hand will recurse through the complete
     * stack of tables, returning all matching symbols in the entire
     * stack.
     */
  class SymbolTable
  {
  protected:

      /**
       * Since symbol tables may be nested, a symbol table may have a
       * parent.  The parent as defined in the SymbolTable class is a
       * pointer to another instance of the SymbolTable class.  If a
       * symbol table does not have a parent, this pointer is null.
       */
    SymbolTable* parent;

      /**
       * In essence, a symbol table maps keys to symbols.  In this
       * implementation a STL map object is used to accomplish this
       * mapping.  Irrespective of whether a symbol table is keyed by
       * id or name, the mapping is from a string to symbol pointer.
       */
    std::map<std::string, Symbol*> symbols;

  public:

      /**
       * Default constructor for the SymbolTable class.
       *
       * By default, a symbol table does not have a parent.
       * Initially, a symbol table contains no symbols.
       */
    inline SymbolTable()
      : parent(0)
      {
        symbols.clear();
      }


      /**
       * Destructor for the SymbolTable class.
       *
       * Does nothing at all.  For technical reasons, destructors of
       * abstract classes should not be defined inline.
       */
    virtual ~SymbolTable();


      /**
       * Adds a symbol to the symbol table.
       *
       * This function will add a symbol to the symbol table.  If the
       * table already contains a symbol with the key of the new
       * symbol, nothing is done.
       *
       * @param s The new symbol to add to the table.
       * @return  True if the symbol was added, false otherwise.
       */
    inline bool
    addSymbol(Symbol* s)
      {
        std::string key = s -> getKey();
        if (!containsLocal(key)) {
          symbols[key] = s;
          return true;
        }
        return false;
      }


      /**
       * Check whether a symbol identified by the given key is defined
       * in any of the symbol tables.
       *
       * This function checks if the symbol identified by the given
       * key is defined in the local symbol table or any of its
       * parents.  Unlike the containsLocal() functions, this function
       * recursively consults the parent symbol tables as well.
       *
       * @param key The key of the sought symbol.
       * @return    True if the symbol is defined, false otherwise.
       */
    inline bool
    contains(const std::string& key) const
      {
        return containsLocal(key) || parent && parent -> contains(key);
      }


      /**
       * Check whether a symbol identified by the given id is defined
       * in any of the symbol tables.
       *
       * This function checks if the symbol identified by the given id
       * is defined in the local symbol table or any of its parents.
       * Unlike the containsLocal() functions, this function
       * recursively consults the parent symbol tables as well.
       *
       * @param id The id of the sought symbol.
       * @return   True if the symbol is defined, false otherwise.
       */
    inline bool
    containsById(const std::string& id) const
      {
        return containsLocalById(id) || parent && parent -> containsById(id);
      }


      /**
       * Check whether a symbol identified by the given name is
       * defined in any of the symbol tables.
       *
       * This function checks if the symbol identified by the given
       * name is defined in the local symbol table or any of its
       * parents.  Unlike the containsLocal() functions, this function
       * recursively consults the parent symbol tables as well.
       *
       * @param name The name of the sought symbol.
       * @return     True if the symbol is defined, false otherwise.
       */
    inline bool
    containsByName(const std::string& name) const
      {
        return containsLocalByName(name)
          || parent && parent -> containsByName(name);
      }


      /**
       * Check whether a symbol identified by the given key is defined
       * in the local symbol table.
       *
       * This function checks if the symbol identified by the given
       * key is defined in the local symbol table.  Note that this
       * function will not recurse into the parent symbol tables if
       * the symbol is undefined in the local table.
       *
       * @param key The key of the sought symbol.
       * @return    True if the symbol is defined locally, false
       *            otherwise.
       */
    inline bool
    containsLocal(const std::string& key) const
      {
        return symbols.find(key) != symbols.end();
      }


      /**
       * Check whether a symbol identified by the given id is defined
       * in the local symbol table.
       *
       * This function checks if the symbol identified by the given id
       * is defined in the local symbol table.  Note that this
       * function will not recurse into the parent symbol tables if
       * the symbol is undefined in the local table.
       *
       * @param id The id of the sought symbol.
       * @return   True if the symbol is defined locally, false
       *           otherwise.
       */
    inline bool
    containsLocalById(const std::string& id) const
      {
        std::map<std::string, Symbol*>::const_iterator i;
        for (i = symbols.begin(); i != symbols.end(); i++) {
          if (i -> second -> isId(id))
            return true;
        }
        return false;
      }


      /**
       * Check whether a symbol identified by the given name is
       * defined in the local symbol table.
       *
       * This function checks if the symbol identified by the given
       * name is defined in the local symbol table.  Note that this
       * function will not recurse into the parent symbol tables if
       * the symbol is undefined in the local table.
       *
       * @param name The name of the sought symbol.
       * @return     True if the symbol is defined locally, false
       *             otherwise.
       */
    inline bool
    containsLocalByName(const std::string& name) const
      {
        std::map<std::string, Symbol*>::const_iterator i;
        for (i = symbols.begin(); i != symbols.end(); i++) {
          if (i -> second -> isName(name))
            return true;
        }
        return false;
      }


      /**
       * Get the parent symbol table.
       *
       * This function will return a pointer to the parent symbol
       * table of the current symbol table.  Note that a symbol need
       * not have a parent symbol table.  In such cases, this function
       * will return the null pointer.  Because the pointer returned
       * is not a const pointer, this function is not defined const.
       *
       * @return A pointer to the parent symbol table.
       */
    inline SymbolTable*
    getParent()
      {
        return parent;
      }


      /**
       * Find the symbol corresponding to the given key.
       *
       * This function will return a pointer to the symbol
       * corresponding to the given key.  This function will not
       * recurse into parent symbol tables in cases where the symbol
       * is not defined locally.  Note that since keys uniquely define
       * symbols in their symbol tables, this function will at most
       * return a single symbol.  If no appropriate symbol could be
       * found, the function will return the null pointer.  Because
       * the pointer returned is not a const pointer, this function is
       * not defined const.
       *
       * @param  key The key of the sought symbol.
       * @return     The unit corresponding to the given key.
       */
    virtual inline Symbol*
    getLocalSymbol(const std::string& key)
      {
        std::map<std::string, Symbol*>::const_iterator i;
        if ((i = symbols.find(key)) != symbols.end())
          return i -> second;
        return 0;
      }


      /**
       * Find the symbols corresponding to the given id.
       *
       * This function will return a list of pointers to symbols
       * corresponding to the given id.  This function will not
       * recurse into parent symbol tables in cases where the symbol
       * is not defined locally.  There may be any number of such
       * symbols.  In case there are no symbols matching the id, an
       * empty list is returned.  Note that it is not as efficient to
       * get symbols by name or id, as to get them by key.  This may
       * be surprising, as exactly one of the former has to be
       * equivalent to the latter.  Because the pointer returned is
       * not a const pointer, this function is not defined const.
       *
       * @param id The id of the sought symbol.
       * @return   A list of symbols corresponding to the given id.
       */
    virtual std::list<Symbol*>
    getLocalSymbolById(const std::string& id);


      /**
       * Find the symbols corresponding to the given name.
       *
       * This function will return a list of pointers to symbols
       * corresponding to the given name.  This function will not
       * recurse into parent symbol tables in cases where the symbol
       * is not defined locally.  There may be any number of such
       * symbols.  In case there are no symbols matching the name, an
       * empty list is returned.  Note that it is not as efficient to
       * get symbols by name or id, as to get them by key.  This may
       * be surprising, as exactly one of the former has to be
       * equivalent to the latter.  Because the pointer returned is
       * not a const pointer, this function is not defined const.
       *
       * @param name The name of the sought symbol.
       * @return     A list of symbols corresponding to the given name.
       */
    virtual std::list<Symbol*>
    getLocalSymbolByName(const std::string& name);


      /**
       * Find the symbol corresponding to the given key.
       *
       * This function will return a pointer to the symbol
       * corresponding to the given key.  The symbol may be defined in
       * the current symbol table or any of its parents.  Note that
       * since keys uniquely define symbols in their symbol tables,
       * this function will at most return a single symbol.  If no
       * appropriate symbol could be found, the function will return
       * the null pointer.  Because the pointer returned is not a
       * const pointer, this function is not defined const.
       *
       * @param  key The key of the sought symbol.
       * @return     The symbol corresponding to the given key.
       */
    virtual inline Symbol*
    getSymbol(const std::string& key)
      {
        Symbol* s = getLocalSymbol(key);
        if (!s && parent)
          return parent -> getSymbol(key);
        return s;
      }


      /**
       * Find the symbols corresponding to the given id.
       *
       * This function will return a list of pointers to symbols
       * corresponding to the given id.  The symbol may be defined in
       * the current symbol table or any of its parents.  There may be
       * any number of such symbols.  In case there are no symbols
       * matching the id, an empty list is returned.  Note that it is
       * not as efficient to get symbols by name or id, as to get them
       * by key.  This may be surprising, as exactly one of the former
       * has to be equivalent to the latter.  Because the pointer
       * returned is not a const pointer, this function is not defined
       * const.
       *
       * @param id The id of the sought symbol.
       * @return   A list of symbols corresponding to the given id.
       */
    virtual inline std::list<Symbol*>
    getSymbolById(const std::string& id)
      {
        std::list<Symbol*> l = getLocalSymbolById(id);
        if (parent) {
          std::list<Symbol*> lp = parent -> getSymbolById(id);
          l.merge(lp);
        }
        return l;
      }


      /**
       * Find the symbols corresponding to the given name.
       *
       * This function will return a list of pointers to symbols
       * corresponding to the given name.  The symbol may be defined
       * in the current symbol table or any of its parents.  There may
       * be any number of such symbols.  In case there are no symbols
       * matching the name, an empty list is returned.  Note that it
       * is not as efficient to get symbols by name or id, as to get
       * them by key.  This may be surprising, as exactly one of the
       * former has to be equivalent to the latter.  Because the
       * pointer returned is not a const pointer, this function is not
       * defined const
       *
       * @param name The name of the sought symbol.
       * @return     A list of symbols corresponding to the given name.
       */
    virtual inline std::list<Symbol*>
    getSymbolByName(const std::string& name)
      {
        std::list<Symbol*> l = getLocalSymbolByName(name);
        if (parent) {
          std::list<Symbol*> lp = parent -> getSymbolByName(name);
          l.merge(lp);
        }
        return l;
      }


      /**
       * Locally remove a symbol identified by a given key.
       *
       * This function will remove the symbol identified by the given
       * key from the local symbol table.  This function does not
       * consider any parent symbol tables.  If the symbol is not
       * defined in the local table, nothing happens.
       *
       * @param key The key of the symbol to remove.
       */
    inline void
    removeLocalSymbol(const std::string& key)
      {
        if (containsLocal(key))
          delete symbols[key];
      }


      /**
       * Locally remove a symbol identified by a given id.
       *
       * This function will remove any symbols identified by the given
       * id from the local symbol table.  This function does not
       * consider any parent symbol tables.  If the symbol is not
       * defined in the local table, nothing happens.
       *
       * @param id The id of the symbol to remove.
       */
    inline void
    removeLocalSymbolById(const std::string& id)
      {
        std::map<std::string, Symbol*>::iterator i;
        for (i = symbols.begin(); i != symbols.end(); i++) {
          if (i -> second -> isId(id)) {
            std::map<std::string, Symbol*>::iterator j = i--;
            symbols.erase(j);
          }
        }
      }


      /**
       * Locally remove a symbol identified by a given name.
       *
       * This function will remove any symbols identified by the given
       * name from the local symbol table.  This function does not
       * consider any parent symbol tables.  If the symbol is not
       * defined in the local table, nothing happens.
       *
       * @param name The name of the symbol to remove.
       */
    inline void
    removeLocalSymbolByName(const std::string& name)
      {
        std::map<std::string, Symbol*>::iterator i;
        for (i = symbols.begin(); i != symbols.end(); i++) {
          if (i -> second -> isName(name)) {
            std::map<std::string, Symbol*>::iterator j = i--;
            symbols.erase(j);
          }          
        }
      }


      /**
       * Remove a symbol identified by a given key from the symbol
       * table stack.
       *
       * This function will remove the symbol identified by the given
       * key from the entire symbol table stack.  If the symbol is not
       * defined in any of the symbol tables in the stack, nothing
       * happens.
       *
       * @param key The key of the symbol to remove.
       */
    inline void
    removeSymbol(const std::string& key)
      {
        removeLocalSymbol(key);
        if (parent)
          parent -> removeSymbol(key);
      }


      /**
       * Remove a symbol identified by a given id from the symbol
       * table stack.
       *
       * This function will remove any symbols identified by the given
       * key from the entire symbol table stack.  If the symbol is not
       * defined in any of the symbol tables in the stack, nothing
       * happens.
       *
       * @param id The id of the symbol to remove.
       */
    inline void
    removeSymbolById(const std::string& id)
      {
        removeLocalSymbolById(id);
        if (parent)
          parent -> removeSymbolById(id);
      }


      /**
       * Remove a symbol identified by a given name from the symbol
       * table stack.
       *
       * This function will remove any symbols identified by the given
       * name from the entire symbol table stack.  If the symbol is
       * not defined in any of the symbol tables in the stack, nothing
       * happens.
       *
       * @param name The name of the symbol to remove.
       */
    inline void
    removeSymbolByName(const std::string& name)
      {
        removeLocalSymbolByName(name);
        if (parent)
          parent -> removeSymbolByName(name);
      }


      /**
       * Replace a symbol identified by a given key with another in
       * the entire symbol table stack.
       *
       * This function will replace the symbol identified by the given
       * key with another.  Any symbols with the given key in the
       * entire symbol table stack are subject to replacement.  If the
       * symbol is not defined in a symbol table table, nothing is
       * done.
       *
       * @param key    The key for the symbol to replace
       * @param symbol The replacement symbol.
       * @return       True if at least one replacement was done, false
       *               otherwise.
       */
    inline bool
    replaceSymbol(const std::string& key, Symbol* symbol)
      {
        bool replaced = false;
        if (containsLocal(key)) {
          symbols[key] = symbol;
          replaced     = true;
        }
        if (parent)
          return replaced | parent -> replaceSymbol(key, symbol);
        return replaced;
      }


      /**
       * Locally replace a symbol identified by a given key with
       * another.
       *
       * This function will replace the symbol identified by the given
       * key with another.  This function does not consider any parent
       * symbol tables.  If the symbol is not defined in a symbol
       * table table, nothing is done.
       *
       * @param key    The key for the symbol to replace
       * @param symbol The replacement symbol.
       * @return       True if a symbol was replaced, false otherwise.
       */
    inline bool
    replaceLocalSymbol(const std::string& key, Symbol* symbol)
      {
        if (containsLocal(key)) {
          symbols[key] = symbol;
          return true;
        }
        return false;
      }


      /**
       * Set the parent symbol table.
       *
       * This function will set the parent of the symbol table.  Note
       * that a symbol need not have a parent symbol table.  In such
       * cases, the argument should be the null pointer.
       *
       * @param p A pointer to the parent symbol table.
       */
    inline void
    setParent(SymbolTable* p)
      {
        parent = p;
      }
  };
}
#endif
