/* The System class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 * 
 * This file is part of MesoRD.  
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: System.cpp 561 2008-11-25 20:10:53Z davidfange $
 */
#include <limits>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <math.h>
#include <sbml/SBMLReader.h>

#include "mesord.hpp"
#include "definition.hpp"
#include "System.hpp"

#include "ui/ProgressBar.hpp"

#ifdef DETERMINISTIC
#include "Derivative.hpp"
#include "integration/Euler.hpp"
#include "integration/EulerAdaptive.hpp"
#include "integration/EulerBackward.hpp"
#include "integration/RungeKutta4.hpp"
#include "integration/Trapez.hpp"
#include "integration/BDF2.hpp"

#include "sparselib/compcol_double.hpp"
#include "sparselib/comprow_double.hpp"
#include "mv/mvvd.hpp"
#endif //DETERMINISTIC

#include "linalg/Vector.hpp"



  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Default constructor.  Make sure everything points to null and so
   * on.  In particular, make sure the queue pointer is set to null,
   * otherwise we may run into believing we have a queue, when, in
   * fact, we do not.
   */
MesoRD::System::System() 
  : compartmentTable(), parameterTable(), unitTable()
{
  checkpoint_iterations = 0;
  checkpoint_time       = 0;
  max_iterations        = 0;
  max_time              = 0;

  subVolumes.clear();
  outfile_prefix  = "";
  outfile_time    = 0;
  outstream_log   = 0;
  progressBar     = 0;
  queue           = 0;
  random          = new Random::MersenneTwister();
}


  /* Value constructor.  This is probably what will be used most of
   * the time.  Make sure to set the queue pointer to null.
   */
MesoRD::System::System(const Model_t* m, double size, const std::string& unit)
{
  checkpoint_iterations = 0;
  checkpoint_time       = 0;
  max_iterations        = 0;
  max_time              = 0;
  
  subVolumes.clear();
  outfile_prefix  = "";
  outfile_time    = 0;
  outstream_log   = 0;
  progressBar     = 0;
  queue           = 0;
  random          = new Random::MersenneTwister();
  fromSBML(m, size, unit);
}


  /* Destructor.  Just free the queue.
   */
MesoRD::System::~System() 
{
  if (random)
    delete random;
  freeQueue();
}



  /******************************
   * QUEUE SORTING -- PROTECTED *
   ******************************/

  /* This C-style function and its accompanying static sub volume
   * vector pointer is only public within the current translation
   * unit.  It should not, be used outside this file.  Its only use is
   * as a comparison function for the operating system's quick sort
   * function.  The function compares the times of two queue entries.
   */
static std::vector<MesoRD::SubVolume*>* subVolumes_static = 0;

extern "C" 
{
  static int
  queueComparison(const void* i, const void* j) {
    double time_i = (*subVolumes_static)[
      *((MesoRD::SystemQueueEntry*)i)] -> getTime();
    double time_j = (*subVolumes_static)[
      *((MesoRD::SystemQueueEntry*)j)] -> getTime();

    if (time_i < time_j)
      return -1;
    else if (time_i > time_j)
      return 1;
    else
      return 0;
  }
}



  /* Note that we cannot reuse loop indices the way the C++ standard
   * says we should, because some compilers (...) do not tolerate
   * that.  Here we need the comparison function defined above for the
   * qsort() call.  Note that once the time queue is sorted, every sub
   * volume needs to know its position in the time queue.
  */
void
MesoRD::System::sortAllSubVolumes(const double) 
{
  unsigned int i;
  allocateQueue();

  for (i = 0; i < subVolumes.size(); i++)
    queue[i] = i;

  subVolumes_static = &subVolumes;
  qsort(
    &queue[0], subVolumes.size(), sizeof(SystemQueueEntry), queueComparison);

  for (i = 0; i < subVolumes.size(); i++) {
    subVolumes[queue[i]] -> setQueueIndex(i);
  }
}


  /* There are two basic cases: if the time for the next event has
   * increased, the sub volume should be moved down in the tree, if it
   * has decreased it should move up.  We cannot move a sub volume up
   * or down if it is at the top or the bottom of the tree
   * respectively.  This method really implements a kind of bubble
   * sort in the tree.
   */
void
MesoRD::System::sortOneSubVolume(const size_t index_queue)
{
  double time_subVolume = getQueueTime(index_queue);
  size_t index_above    = (index_queue - 1) / 2;
  size_t index_below    = 2 * (index_queue + 1) - 1;
  size_t index_current  = index_queue;

  if (
    (((index_below + 0) < subVolumes.size())
      && (time_subVolume > getQueueTime(index_below + 0)))
    || (((index_below + 1) < subVolumes.size())
      && (time_subVolume > getQueueTime(index_below + 1)))) {

      /* We will move the sub volume down the tree.  If we are at the
       * bottom of the tree, there is nothing we can do.  Otherwise
       * find the sub volume below the current one which with the
       * smallest time.  If the time for that sub volume is less than
       * the time for the one we are sorting, swap them and go for
       * another spin in the loop.
       */
    for ( ; ; ) {
      index_below = 2 * (index_current + 1) - 1;
      if (((index_below + 1) < subVolumes.size())
        && (getQueueTime(index_below + 0) > getQueueTime(index_below + 1))) {
        index_below++;
      }

      if ((index_below >= subVolumes.size())
        || (getQueueTime(index_below) >= time_subVolume)) {
        break;
      }

      swapInQueue(index_current, index_below);
      index_current = index_below;
    }
  } else {

      /* We will move the sub volume up the tree.  If we are at the
       * top of the tree, there is nothing we can do.  Otherwise find
       * the sub volume above the current one.  If the time for that
       * sub volume is more than the time for the one we are sorting,
       * swap them and go for another spin in the loop.
       */
    for ( ; ; ) {
      index_above = (index_current - 1) / 2;
      if ((index_current <= 0)
        || (getQueueTime(index_above) <= time_subVolume)) {
        break;
      }

      swapInQueue(index_current, index_above);
      index_current = index_above;
    }
  }
}


  /* If we have to resort two sub volumes, we may save on some bubble
   * sorting if we swap the two queue entries first.  We estimate that
   * if the time difference between the updated sub volumes and to the
   * old sub volumes after a swap would be less than before, a swap is
   * a good idea.  Note that we must save a pointer to the second sub
   * volume, because resorting the first sub volume may change the
   * queue position of the second.
   */
void
MesoRD::System::sortTwoSubVolumes(const size_t i, const size_t j)
{
  if (   i < j && getQueueTime(i) > getQueueTime(j)
      || i > j && getQueueTime(i) < getQueueTime(j)) {
    swapInQueue(i, j);
  }

  SubVolume* second = subVolumes[queue[j]];
  sortOneSubVolume(i);
  sortOneSubVolume(second -> getQueueIndex());
}


  /* Text book example of swapping.  Here "i" and "j" are indices in
   * the queue which will be swapped.
   */
void
MesoRD::System::swapInQueue(const size_t i, const size_t j)
{
  subVolumes[queue[i]] -> setQueueIndex(j);
  subVolumes[queue[j]] -> setQueueIndex(i);

  SystemQueueEntry tmp = queue[i];
  queue[i]             = queue[j];
  queue[j]             = tmp;
}



  /**************
   * STATISTICS *
   **************/

  /* Create a list of indices containing all the indices of all the
   * species in the system.  Then do a histogram on those.
   */
void
MesoRD::System::histogram(std::ostream& os) const
{
  std::list<off_t> indices;
  
  indices.clear();
  for (off_t i = 0; i < (off_t)subVolumes[0] -> getSpeciesNumber(); i++) {
    indices.push_back(i);
  }
  
  histogram(indices, os);
}


  /* Loop through all the sub volumes.  For every species index in the
   * indices list sum up the number of molecules in the sub volume.
   * Then construct the histogram and dump it to the requested output
   * stream.
   */
void
MesoRD::System::histogram(
  const std::list<off_t>& indices, std::ostream& os) const
{
  std::map<unsigned int, unsigned long> histogram;
  
  histogram.clear();
  for (unsigned int i = 0; i < subVolumes.size(); i++) {
    unsigned int                                    total = 0;
    std::list<off_t>::const_iterator                j;
    std::map<unsigned int, unsigned long>::iterator k;

    for (j = indices.begin(); j != indices.end(); j++) {
      total += subVolumes[i] -> getSpecies(*j);
    }
    
    if ((k = histogram.find(total)) == histogram.end()) {
      histogram[total] = 1;
    } else {
      histogram[total]++;
    }
  }

  std::map<unsigned int, unsigned long>::const_iterator l;
  for (l = histogram.begin(); l != histogram.end(); l++) {
    os << "  " << l -> first
       << ": " << l -> second
       << std::endl;
  }
}



  /****************
   * SBML SUPPORT *
   ****************/


  /* First we parse the SBML in order to generate an SBML document.
   * Then we attempt to build the model in our own representation from
   * this.  We try to be nice in reporting the errors to the log file.
   * The reporting should obey the GNU error reporting standards.
   */ 
void
MesoRD::System::fromSBML(
  const std::string& filename, double size, const std::string& unit)
  throw(MesoRD::System::ModelError)
{
  unsigned int errors   = 0;
  unsigned int warnings = 0;
  unsigned int fatals   = 0;

  if (filename.length() > 0) {
    
    SBMLDocument* d = readSBML(filename.c_str());
    
    if (d -> getNumErrors() > 0) {
      unsigned int    sbml_errors;
      
      sbml_errors  = d -> getNumErrors();
      
      for(unsigned int i = 0; i < sbml_errors; i++){
        if((d -> getError(i)) -> isError())
          errors++;
        if((d -> getError(i)) -> isWarning())
          warnings++;
        if((d -> getError(i)) -> isFatal())
          fatals++;
        
      }
      if (outstream_log) {
        (*outstream_log) << "There "
                         << ((warnings > 1) ? "were" : "was")
                         << " "                    << warnings
                         << ((warnings > 1) ? " warnings" : " warning")
                         << " while parsing file " << filename
                         << "!"                    << std::endl;
      
        for (unsigned int i = 0; i < sbml_errors; i++) {
          const SBMLError* sbml_error = d -> getError(i);
          
          if(sbml_error -> isWarning()) {
            
            (*outstream_log) << "  SBML WARNING "
                             << i
                             << " at line "
                             << sbml_error -> getLine()
                             << ":"
                             << sbml_error -> getColumn()
                             << " "
                             << sbml_error -> getMessage()
                             << std::endl;
          }
        }
      
        
        (*outstream_log) << "There "
                         << ((errors > 1) ? "were" : "was")
                         << " "                    << errors
                         << ((errors > 1) ? " errors" : " error")
                         << " while parsing file " << filename
                         << "!"                    << std::endl;
        
        for (unsigned int i = 0; i < sbml_errors; i++) {
          const SBMLError* sbml_error = d -> getError(i);
          
          if( sbml_error -> isError() ){
            (*outstream_log) << "  SBML ERROR "
                             << i
                             << " at line "
                             << sbml_error -> getLine()
                             << ":"
                             << sbml_error -> getColumn()
                             << " "
                             << sbml_error -> getMessage()
                             << std::endl;
          }
        }
        
        (*outstream_log) << "There "
                         << ((fatals > 1) ? "were" : "was")
                         << " "                    << fatals
                         << " fatal "
                         << ((fatals > 1) ? "errors" : "error")
                         << " while parsing file " << filename
                         << "!"                    << std::endl;

        for (unsigned int i = 0; i < sbml_errors; i++) {
          const SBMLError* sbml_error = d -> getError(i);
          
          if( sbml_error -> isFatal() ){
            (*outstream_log) << "  SBML FATAL ERROR "
                             << i
                             << " at line "
                             << sbml_error -> getLine()
                             << ":"
                             << sbml_error -> getColumn()
                             << " "
                             << sbml_error -> getMessage()
                             << std::endl;
          }
        }
      } 
    }
    
    
      /* We check for exceptions here and throw them on to whoever
       * called us.
       */
    if (d && (errors == 0) && (fatals == 0)) {
      try {
	Model* m = d -> getModel();

        fromSBML(m, size, unit);
      } catch (ModelError& e) {
        throw ModelError(e.what());
      }
    } else {
      throw ModelError("Problems during SBML parsing.");
    }
  }
}


  /* Start out by freeing the sub volumes and the queue, because we
   * are rebuilding the entire lot.  Then parse the units.  We may
   * need the units for many entities that can be defined in what
   * follows, so we read them in first.  Second, parse the parameters.
   * We really should worry about the previously parsed units.
   * Finally construct the table of the compartments.  This will in
   * turn build the geometry, parse the reactions and do whatever else
   * is required to make the system ready to run.
   */
void
MesoRD::System::fromSBML(
  const Model* m, double size, const std::string& unit)
  throw(MesoRD::System::ModelError)
{
  unsigned int i;

  subVolumes.clear();
  freeQueue();

  unitTable.fromSBML(m -> getListOfUnitDefinitions());
  if (outstream_log) {
    (*outstream_log) << "This is the root unit table "
                     << std::endl
                     << unitTable
                     << std::endl;
  }
  
  for (i = 0; i < m -> getNumParameters(); i++) {
    parameterTable.addParameter(
      Parameter(m -> getParameter(i), &unitTable));
  }
  if (outstream_log) {
    (*outstream_log) << "This is the root parameter table"
                     << std::endl
                     << parameterTable
                     << std::endl;
  }

    /* Build the geometry pass throw any exceptions on to whoever
     * called us.
     */
  compartmentTable.setOptionOutputSparse(outputSparse);
  try {
    subVolumes = compartmentTable.fromSBML(
      m, &parameterTable, &unitTable, size, unit, random);
  } catch (CompartmentTable::ModelError& e) {
    throw ModelError(e.what());
    
  } catch (CompartmentTable::EmptyGeometry& e) {
    throw ModelError(e.what());
    
  } catch(CompartmentTable::NoId& e) {
    std::stringstream message;
    message << "No id for " << e.what();
    throw ModelError((message.str()).c_str());
  }

    /* A little statistics.
     */
  if (outstream_log) {
    (*outstream_log) << "We have in total "
                     << subVolumes.size()
                     << " sub volumes to work on"
                     << std::endl;
  }  
}


  /* Writes the geometry information to file.  Not very complicated at
   * all.
   */
void
MesoRD::System::writeGeometry(const std::string& filename)
{
  std::ofstream outfile(filename.c_str(), std::ios::trunc);
  for (unsigned int i = 0; i < subVolumes.size(); i++) {
    outfile << std::setw(10) << subVolumes[i] -> getX() << " "
            << std::setw(10) << subVolumes[i] -> getY() << " "
            << std::setw(10) << subVolumes[i] -> getZ() << " " 
            << std::setw(10) 
            << (subVolumes[i] -> getCompartment()) -> getId() 
            << std::endl;
  }
  outfile.close();
}

/* Reads the state from an earlier output from the program. This
 * assumes that the same sbml file and the same parameters has been
 * used in the the simulations. This will most likely quietly do tons
 * of error otherwise.  Start out the function by clearing all the
 * concentrations that the parsing of the model has produced.
 */
void 
MesoRD::System::parseOldState(const size_t lineNumber)
  throw( MesoRD::System::FileNotFoundError,
    MesoRD::System::ModelError )
{
  std::ifstream geometryFile(
     (outfile_prefix + MESORD_GEOMETRYFILE_NAME).c_str());

  
  if(!geometryFile) {
    std::stringstream message;
    message << "Unable to open file: "
            << outfile_prefix
            << MESORD_GEOMETRYFILE_NAME
            << " for reading" << std::endl;
    throw FileNotFoundError((message.str()).c_str());
  }  
  std::map<int, std::ifstream*> infiles_species;
  try {
    infiles_species =
      compartmentTable.openSpeciesFiles(outfile_prefix);
  } catch( CompartmentTable::FileNotFoundError& e) {
    throw FileNotFoundError(e.what());
  }
  
  std::map<int, std::ifstream*>::const_iterator it;
  
  for(it = infiles_species.begin(); it != infiles_species.end(); ++it){
    for(unsigned int i = 0; i < (unsigned int)lineNumber; i++){
      std::string dummy; 
      if(!(std::getline(*(it -> second), dummy))) {
        throw ModelError("No state aviable at this timepoint");
      }
    }
  }
  
  std::map<CSGPoint, unsigned int> subVolumeMap;
  for(unsigned int i = 0; i < subVolumes.size(); i++) {
    subVolumeMap[CSGPoint(
                   subVolumes[i] -> getX(),
                   subVolumes[i] -> getY(),
                   subVolumes[i] -> getZ(), 0.5, 0.5, 0.5)] = i;
  }
  
  
    /* Ressetting the compartment species numbers */
  std::list<Compartment*>* compartments = compartmentTable.getCompartments();
  
  std::list<Compartment*>::const_iterator compartment;
  for(compartment = compartments -> begin(); 
      compartment != compartments -> end(); compartment++) {
    for (unsigned int j = 0; j < (*compartment) -> getSpeciesNumber(); j++) {
      if(SubVolume::getUseDoubleInSpecies()) {
        (*compartment) ->setSpecies(j, (double)0);
      } else {
        (*compartment) ->setSpecies(j, (unsigned int)0);
      }
    }
  }

  
  
  for(unsigned int k = 0; k < subVolumes.size(); k++) {
    size_t speciesNumber = 
      (subVolumes[k] -> getCompartment()) -> getSpeciesNumber();
    for(unsigned int j = 0; j < speciesNumber; j++) {
      if(SubVolume::getUseDoubleInSpecies()) {
        subVolumes[k] -> setSpecies(j, double(0));
      }
      else {
        subVolumes[k] -> setSpecies(j, (unsigned int)(0));
      }
    }     
  }
  
  
  double       X,Y,Z;  
  double       amountDouble;
  unsigned int amountInt;

  std::valarray<double> 
    totalAmountDouble((double)0, infiles_species.size());
  std::valarray<unsigned int> 
    totalAmountInt((unsigned int)0, infiles_species.size());

  std::vector<std::vector<double> > geometry;

  while(geometryFile >> X){
    geometryFile >> Y;
    geometryFile >> Z;
    
    std::vector<double> geometry_line(3,0);
    geometry_line[0] = X;
    geometry_line[1] = Y;
    geometry_line[2] = Z;
    std::string foo;
    geometryFile >> foo;
    geometry.push_back(geometry_line);
  }  


  if(geometry.size() != subVolumeMap.size()) {
    throw ModelError("New and old geometry does not match");
  }
  
  std::map<unsigned int, unsigned int> subVolumeIndecies;
   
  std::vector<std::vector<double> >::const_iterator geometry_it;
  
  unsigned int geometryIndex = 0;
  
  for(geometry_it = geometry.begin(); 
      geometry_it != geometry.end(); 
      geometry_it++) {
    
    X = (*geometry_it)[0];
    Y = (*geometry_it)[1];
    Z = (*geometry_it)[2];

    subVolumeIndecies[geometryIndex++] = 
      subVolumeMap[CSGPoint((float)X,(float)Y,(float)Z,(float)0.5,(float)0.5,(float)0.5)];
  }

  if(outputSparse) {
      /* If the sparse flag is set read the old state as if it was in 
       * sparse
       */
    for(it = infiles_species.begin(); it != infiles_species.end(); ++it) {
      std::string line;
	  std::getline(*(it -> second), line);
      std::istringstream streamline(line);
      while(streamline >> geometryIndex) {
        if(!(streamline >> amountInt)) { 
          throw ModelError("Unable to read amount for all species");
        }
        totalAmountInt[it->first] += amountInt;
        subVolumes[subVolumeIndecies[geometryIndex]] 
          -> changeSpecies(it->first, (int)amountInt);
      }
    }
  } else {
    
      /* code for non sparse format.
       */
    if(SubVolume::getUseDoubleInSpecies()) {
      for(it = infiles_species.begin(); it != infiles_species.end(); ++it) {
        geometryIndex = 0;
        std::string line;
		std::getline(*(it -> second), line);
        std::istringstream streamline(line);
        while( streamline >> amountDouble ) {
          totalAmountDouble[it->first] += amountDouble;
          subVolumes[subVolumeIndecies[geometryIndex++]] 
            -> changeSpecies(it->first, amountDouble);
        }
      }
    } else {
      for(it = ++it; it != infiles_species.end(); ++it) {
        geometryIndex = 0;
        std::string line;
		std::getline(*(it -> second), line);
        std::istringstream streamline(line);
        while( streamline >> amountInt ) {
          totalAmountInt[it->first] += amountInt;
          subVolumes[subVolumeIndecies[geometryIndex++]] 
            -> changeSpecies(it->first, (int)amountInt);
        }
      }
    }
  }
  
  for(it = infiles_species.begin(); it != infiles_species.end(); ++it){
    if(SubVolume::getUseDoubleInSpecies()){
      compartmentTable.setSpecies(it->first, totalAmountDouble[it -> first]);
    }
    else{
      compartmentTable.setSpecies(it->first, totalAmountInt[it -> first]);
    }
  }

  geometryFile.close();
  compartmentTable.closeSpeciesFiles(infiles_species);
}




  /* Set the logging stream of compartment table.  Then set the
   * logging stream of the system itself.
   */
void
MesoRD::System::setOptionLogfile(std::ostream* log)
{
  compartmentTable.setOptionLogfile(log);
  outstream_log = log;
}



  /**************
   * SIMULATION *
   **************/

  /* Do some initial setup for the simulation to start.  Then we give
   * the logging stream a short summary of the checkpointing settings,
   * write the geometry file and open the files we need for
   * checkpointing.  The rest of the comments are along the way.
   *
   */
void
MesoRD::System::simulationInit()
  throw( MesoRD::System::FileNotFoundError,
    MesoRD::System::ModelError )
{
  
  if (outstream_log) {
    (*outstream_log)   << "checkpoint iterations : ";
    if (checkpoint_iterations <= 0)
      (*outstream_log) << "[disabled]"                << std::endl;
    else
      (*outstream_log) << checkpoint_iterations       << std::endl;

    (*outstream_log)   << "      checkpoint time : ";
    if (checkpoint_time <= 0)
      (*outstream_log) << "[disabled]"               << std::endl;
    else
      (*outstream_log) << checkpoint_time            << std::endl;

    (*outstream_log)   << "       max iterations : "
                       << max_iterations             << "e+"
                       << max_iterations_scale       << std::endl
                       << "             max time : "
                       << max_time                   << std::endl;
  }
  
    /* parse the output files from an old simulation and grab the 
     * state at timepoint readOldStateTime
     */
  if( readOldState ) {
    (*outstream_log)  << "Starting simulation from saved checkpoint nr: "
                      << readOldStateTime            << std::endl;
    
    try {
      parseOldState(static_cast<size_t>(readOldStateTime));
    } catch( ModelError& e) {
      throw ModelError(e.what());
    } catch( FileNotFoundError& e) {
      throw FileNotFoundError(e.what());
    }
  }
  
  writeGeometry(outfile_prefix + MESORD_GEOMETRYFILE_NAME);
  openTimeFile(outfile_prefix + MESORD_TIMEFILE_NAME);
  compartmentTable.openLocalFiles(outfile_prefix);

    /* Initial Sorting: Sort the time queue.  We start at simulation
     * time 0.
     */
  if (outstream_log) {
    (*outstream_log) << "Starting initial sorting..."
                     << std::endl;
  }
  
  for (unsigned int i = 0; i < subVolumes.size(); i++) {
    subVolumes[i] -> updateTime(0);
  }
  
  sortAllSubVolumes(0.0);

  if (outstream_log) {
    (*outstream_log) << "Initial sorting done."
                     << std::endl;
  }
}


   /* End the simulation session by closing all the oopened files.
    */
void
MesoRD::System::simulationFinalise()
{
      /* Close the files.  Everybody should treat their operating system
     * the way they would want to be treated themselves.
     */
  closeTimeFile();
  compartmentTable.closeLocalFiles();
}
  
  /*
   * The iterations array works like an integer of arbitrary size.
   * Every element should be an integer in the range [0, 9] and it
   * represents the number of ten-to-the-power-of-index multiples of
   * the current iteration count.
   */
void
MesoRD::System::simulate()
{
 
  unsigned int          i;
  size_t                last_checkpoint_iterations
    = 2 * checkpoint_iterations;
  double                last_checkpoint_time = 0;
  double                last_checkpoint_time_inc = 0;
  size_t                last_progressbar_iterations
    = 2 * Constant::iterations_progress_stochastic;
  std::valarray<size_t> iterations_array(size_t(0), max_iterations_scale + 1);

    /* We should only print the headers if checkpoint is turned on */
  if (checkpoint_iterations > 0 || checkpoint_time > 0 )
    compartmentTable.startCheckpointing(*outstream_log);


    /* The main loop.  If <code>max_iterations</code> and
     * <code>max_time</code> is less than zero, there will be an
     * infinite loop.  Otherwise stop as soon as any one of them
     * exceeds its limit.  If the progress bar dies, the simulator
     * should die as well.  This is because it should be possible to
     * stop the simulation by interacting with the progress bar.
     * Thank you for playing, please call again.
     */
  while (
    (max_iterations <= 0
      || iterations_array[max_iterations_scale] < max_iterations)
    && (max_time <= 0
      || getQueueTime(0) <= max_time)
    && (!progressBar
      || progressBar -> isAlive())) {
    size_t     this_index = queue[0];
    double     this_time  = getQueueTime(0);
    SubVolume* diffusion_target;


      /* If the time has come to write out a checkpoint, do so now.
       * If we have a logging stream, we will write a global
       * checkpoint to the log.
       */
    if (
      ((checkpoint_iterations > 0)
        && (last_checkpoint_iterations >=  checkpoint_iterations))
      || ((checkpoint_time > 0)
        && ((this_time - last_checkpoint_time) > checkpoint_time))) {

      double nr_checkpoints = floor( (this_time - last_checkpoint_time_inc)/checkpoint_time );
      
      for( unsigned int nr = 0; nr < nr_checkpoints; nr++) {
        last_checkpoint_time_inc += checkpoint_time;
        
        checkpointTime( last_checkpoint_time_inc );
        
        if(fullStateCheckPoint)
          compartmentTable.checkpointLocal(subVolumes);
        
        if (outstream_log) {
          compartmentTable.checkpointGlobal(
            *outstream_log, iterations_array, last_checkpoint_time_inc);
        }
      }
      last_checkpoint_iterations = 0;
      last_checkpoint_time       = this_time;
    }
    
      

      last_checkpoint_iterations++;
      
      
        /* Increase the iteration count by one.  This is done by
         * incrementing the digits in <code>iterations_array</code>,
         * from low indices to high indices.  Once an element overflows
         * decimally (i.e. becomes bigger than 9), it is spilled to
         * next, all the way up to the
         * <code>max_iterations_scale</code>:th element.
         */
      iterations_array[0]++;
      for (i = 0; i < max_iterations_scale && iterations_array[i] >= 10; i++) {
        iterations_array[i]     -= 10;
        iterations_array[i + 1] += 1;
      }

      /* Execute the event in top sub volume.  Update the time of the
       * changed sub volumes and resort them.  The swapping done in
       * sortTwoSubvolumes() yields a speedup of between 0% and 10%,
       * depending on how diffusion intensive the system is.
       *
       * For the implementation with spatial corrections that allow
       * for reactions over subvolume boundaries, the
       * reaction-propensities for the multi-subvolume reactions needs
       * to be recalculated at every iteration.  The reason is that
       * the multi-subvolume reactions depend on the concentration of
       * molcules in neighbouring subvolumes.  
       *
       * FIXME!! The sorting could be much more cpu-cycle conservative.  
       */
    if ((diffusion_target = subVolumes[this_index] -> event())) {
      
      subVolumes[this_index] -> updateTime(this_time);
      sortOneSubVolume(subVolumes[this_index] -> getQueueIndex());
      
      diffusion_target       -> updateTime(this_time);
      sortOneSubVolume(diffusion_target -> getQueueIndex());

      for (size_t i = 0; i < subVolumes[this_index] -> getNeighboursNumber(); i++) {
        if( subVolumes[this_index] -> getNeighbour(i) -> getSpeciesArray() -> sum() != 0 &&
            subVolumes[this_index] != diffusion_target) {
          
          subVolumes[this_index] -> getNeighbour(i) -> updateTime(this_time);
          sortOneSubVolume(subVolumes[this_index] -> getNeighbour(i) -> getQueueIndex());
        }
        
        if( diffusion_target -> getNeighbour(i) -> getSpeciesArray() -> sum() != 0 &&
          diffusion_target -> getNeighbour(i) != subVolumes[this_index] ) {
          diffusion_target -> getNeighbour(i) -> updateTime(this_time);
          sortOneSubVolume(diffusion_target -> getNeighbour(i) -> getQueueIndex());
        }
      }
      
    } else {
      subVolumes[this_index] -> updateTime(this_time);
      sortOneSubVolume(0);
      for (size_t i = 0; i < subVolumes[this_index] -> getNeighboursNumber(); i++) {
        subVolumes[this_index] -> getNeighbour(i) -> updateTime(this_time);
        sortOneSubVolume(
          subVolumes[this_index] -> getNeighbour(i) -> getQueueIndex());
      }
    }


      /* If it is time to update the progress bar, do so now.
       * Remember to reset the counter if the bar is updated in this
       * iteration, and increase the counter if it is not.
       */
    if (progressBar) {
      last_progressbar_iterations++;

      if (last_progressbar_iterations
        > Constant::iterations_progress_stochastic) {
        const double f_time       = this_time / max_time;
        double       f_iterations = 0;

        for (i = 0; i < iterations_array.size(); i++) {
          f_iterations
            += iterations_array[i] * pow(double(10), static_cast<double>(i));
        }
        f_iterations /= max_iterations
          * pow(double(10), static_cast<double>(max_iterations_scale));

        last_progressbar_iterations = 0;
        progressBar -> setFraction(std::max(f_iterations, f_time));
        progressBar -> setIteration(iterations_array);
        progressBar -> setTime(this_time);
      }
    }
  }
}


#ifdef DETERMINISTIC
void
MesoRD::System::simulateDeterministic()
{  
  unsigned int        i;
  size_t              last_checkpoint_iterations
    = 2 * checkpoint_iterations;
  double              last_checkpoint_time
    = -1.0 * (std::numeric_limits<double>::max());
  size_t              last_progressbar_iterations
    = 2 * Constant::iterations_progress_deterministic;
  std::valarray<size_t> iterations_array(size_t(0), max_iterations_scale + 1);
  double              current_time = 0;

    /* Make instances of the objects needed for time-stepping */
  Derivative derivative(subVolumes, &compartmentTable);

  CompRow_Mat_double diffusionMatrix = derivative.getDiffusionMatrix();  
  
  MV_Vector_double currentState = derivative.getCurrentState();
  
  Integration *integration = 0;

  
  switch(Integration::getIntegrationMethod()){
    case Integration::Euler:
      integration = new Euler(&derivative, &diffusionMatrix, timeStep, 
        outstream_log);
      break;
    case Integration::EulerBackward:
      integration = new EulerBackward(&derivative, &diffusionMatrix, 
        timeStep, outstream_log, itMetConvergenceTolerance, 
        itMetMaxIterations, currentState);
      break;
    case Integration::EulerAdaptive:
      integration = new EulerAdaptive(&derivative, &diffusionMatrix, timeStep, 
        outstream_log);
      break;
    case Integration::RungeKutta4:
      integration = new RungeKutta4(&derivative, &diffusionMatrix, timeStep, 
        outstream_log);
      break;
    case Integration::Trapezoid:
      integration = new Trapez(&derivative, &diffusionMatrix, 
        timeStep, outstream_log, itMetConvergenceTolerance, 
        itMetMaxIterations, currentState);
      break;
    case Integration::BDF2:
      integration = new BDF2(&derivative, &diffusionMatrix, 
        timeStep, outstream_log, itMetConvergenceTolerance, 
        itMetMaxIterations, currentState);
      break;
  }

    /* We should only print headers if checkpointing is turned on */
  if (checkpoint_iterations > 0 || checkpoint_time > 0 )
    compartmentTable.startCheckpointing(*outstream_log);


    /* The main loop.  If <code>max_iterations</code> and
     * <code>max_time</code> is less than zero, there will be an
     * infinite loop.  Otherwise stop as soon as any one of them
     * exceeds its limit.  If the progress bar dies, the simulator
     * should die as well.  This is because it should be possible to
     * stop the simulation by interacting with the progress bar.
     * Thank you for playing, please call again.
     */
  while (
    (max_iterations <= 0
      || iterations_array[max_iterations_scale] < max_iterations)
    && (max_time <= 0
      || current_time <= max_time)
    && (!progressBar
      || progressBar -> isAlive())) {

    current_time += integration -> updateState(currentState);


    double       this_time  = current_time;


      /* If it is time to update the progress bar, do so now.
       * Remember to reset the counter if the bar is updated in this
       * iteration, and increase the counter if it is not.
       */
    if (progressBar) {
      last_progressbar_iterations++;

      if (last_progressbar_iterations
        > Constant::iterations_progress_deterministic) {
        const double f_time       = this_time / max_time;
        double       f_iterations = 0;

        for (i = 0; i < iterations_array.size(); i++) {
          f_iterations
            += iterations_array[i] * pow(double(10), static_cast<double>(i));
        }
        f_iterations /= max_iterations
          * pow(double(10), static_cast<double>(max_iterations_scale));

        last_progressbar_iterations = 0;
        progressBar -> setFraction(std::max(f_iterations, f_time));
        progressBar -> setIteration(iterations_array);
        progressBar -> setTime(this_time);
      }
    }


      /* If the time has come to write out a checkpoint, do so now.
       * If we have a logging stream, we will write a global
       * checkpoint to the log.
       */
    if (
      ((checkpoint_iterations > 0)
        && (last_checkpoint_iterations >=  checkpoint_iterations))
      || ((checkpoint_time > 0)
        && ((this_time - last_checkpoint_time) >= checkpoint_time))) {
      

      last_checkpoint_iterations = 0;
      last_checkpoint_time       = this_time;
      
      checkpointTime(this_time);
      
      if( fullStateCheckPoint )
        compartmentTable.checkpointLocalDeterministic(
          subVolumes, currentState);

      if (outstream_log) {
        compartmentTable.checkPointGlobalDeterministic(
          *outstream_log, iterations_array, this_time, 
          subVolumes, currentState);
      }

    }
    last_checkpoint_iterations++;


      /* Increase the iteration count by one.  This is done by
       * incrementing the digits in <code>iterations_array</code>,
       * from low indices to high indices.  Once an element overflows
       * decimally (i.e. becomes bigger than 9), it is spilled to
       * next, all the way up to the
       * <code>max_iterations_scale</code>:th element.
       */
    iterations_array[0]++;
    for (i = 0; i < max_iterations_scale && iterations_array[i] >= 10; i++) {
      iterations_array[i]     -= 10;
      iterations_array[i + 1] += 1;
    }
  }
  delete integration;
} 
#endif //DETERMINISTIC
