/* The System class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: System.hpp 551 2008-05-08 06:53:54Z davidfange $
 */
#ifndef MESORD_SYSTEM_HPP
#define MESORD_SYSTEM_HPP 1

#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>

#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include "CompartmentTable.hpp"
#include "ReactionTable.hpp"
#include "ParameterTable.hpp"
#include "SubVolume.hpp"
#include "UnitTable.hpp"

#include "csg/CSG.hpp"
#include "random/MersenneTwister.hpp"


namespace MesoRD {

  class ProgressBar;
  

    /**
     * The SystemQueueEntry typedef.
     *
     * Defining a new class for managing the time queue of the system
     * seem like massive overkill.  This struct will do just fine: it
     * keeps track of the index of a sub volume.  Thus,
     * for any given time queue index we can find the corresponding
     * sub volume by using the index as defined in this struct.
     * Because a sub volume also carries its time queue index, we can
     * easily find a time queue entry given a particular sub volume.
     * Note here that the index used in this struct is not the index
     * in the queue, but the index in the vector of sub volumes.
     */
  typedef off_t SystemQueueEntry;


    /**
     * The System class.
     *
     * This class is a high level class of MesoRD.  It uses lots of
     * other classes to actually drive the simulation.  In principle,
     * the main() function of the program will use this class to read
     * in the model definition from an SBML file, set some additional
     * options such as the checkpointing frequency, and then call
     * simulate() in order to run the simulation.
     *
     * In this class we see that things can go wrong as an effect of
     * user action.  It makes a lot of sense to report these errors
     * back to the user, so that the user can take corrective
     * measures.  To this end we use the logging stream.  The logging
     * stream should be set up by whomever instantiates this class.
     * There should optimally be no outputting to standard out or
     * standard error in this class.  Such output is lost on Windows
     * and also of Unices if the application does not run from a
     * terminal.
     *
     * It was once intended to be able to support more than one kind
     * of sub volume geometry.  This support stops right here, because
     * in this class we begin to assume that the sub volumes are
     * indeed cubic.
     */
  class System {
  private:
    std::vector<SubVolume*> subVolumes;

    CompartmentTable        compartmentTable;
    ParameterTable          parameterTable;
    UnitTable               unitTable;

    std::string             outfile_prefix;
    std::string             integrationMethod;

    std::ofstream*          outfile_time;
    std::ostream*           outstream_log;
    ProgressBar*            progressBar;
    SystemQueueEntry*       queue;
    Random::Random*         random;

    size_t                  checkpoint_iterations;
    size_t                  max_iterations;
    size_t                  max_iterations_scale;
    size_t                  itMetMaxIterations;

    double                  checkpoint_time;
    double                  max_time;
    double                  timeStep;
    double                  itMetConvergenceTolerance;
    long                    readOldStateTime;
    bool                    deterministicSimulation;
    bool                    readOldState;
    bool                    outputSparse;
    bool                    fullStateCheckPoint;

  protected:

      /**
       * Allocate the time queue for the system.
       *
       * The systems keeps pointers to all its sub volumes in a queue.
       * Thus, the size of the queue has to equal the number of sub
       * volumes in the system.  Calling this method will free any
       * previously allocated queue and set up a new queue which
       * fulfils this requirement.  It is thus important that the
       * vector of sub volumes is set up correctly prior to calling
       * this method.
       */
    inline void allocateQueue()
      { if (subVolumes.size() > 0) {
        freeQueue(); queue = new SystemQueueEntry[subVolumes.size()]; } }

      /**
       * Frees the time queue of the system.
       *
       * Calling this method will free the time queue of the system.
       * The pointer to the queue is reset to null, indicating that
       * there is no queue allocated.  This should be called by the
       * destructor.
       */
    inline void freeQueue()
      { if (queue) delete []queue; queue = 0; }
    
      /**
       * Get the time of sub volume at a given index in the time
       * queue.
       *
       * This quick method gets the absolute time of the next event in
       * a sub volume, given the sub volume's index in the time queue.
       * Note that no boundary checking is performed here.
       *
       * @param index The time queue index of the sub volume.
       * @return      The time of the next event in the sub volume.
       */
    inline double getQueueTime(const size_t index) 
      { return (subVolumes[queue[index]] -> getTime()); }
        
      /**
       * Sort all sub volumes in order of increasing time.
       *
       * This method will initialise the time queue given the list of
       * sub volumes.  Once the times until the next reaction have
       * been determined, the queue is sorted in order of increasing
       * time using quicksort.  Note that we thus need not call
       * updateTime() on the sub volumes prior to calling
       * sortAllSubVolumes(), as the latter method will call it for
       * every sub volume in the system.  We need the current
       * simulation time in order to find the absolute time of the
       * next event in each sub volume.
       *
       * @param time_now The current simulation time.
       */
    void sortAllSubVolumes(const double time_now);

      /**
       * Resort one particular entry of the time queue.
       *
       * This method will move the requested queue entry around in the
       * queue until the queue is again sorted.  Note that
       * updateTime() should be called on the sub volume we will
       * resort prior to calling sortOneSubVolumue().
       *
       * @param index_queue The old time queue index of the sub volume.
       */
    void sortOneSubVolume(const size_t index_queue);

      /**
       * Resort two entries in the time queue.
       *
       * This method will move the requested queue entries around in
       * the queue until the queue is again sorted.  Note that
       * updateTime() should be called on the sub volume we will
       * resort prior to calling sortOneSubVolumue().
       *
       * @param i The old time queue index of the first sub volume.
       * @param j The old time queue index of the second sub volume.
       */
    void sortTwoSubVolumes(const size_t i, const size_t j);

      /**
       * Swaps two entries in the time queue.
       *
       * This method will swap two entries in the time queue.  It is
       * used by the sorting functions defined above.
       *
       * @param i Queue index of the first queue entry to swap.
       * @param j Queue index of the second queue entry to swap.
       */
    void swapInQueue(const size_t i, const size_t j);

  public:


      /**
       * Class for reporting errors during model building.
       *
       * If model building failed, this exception will be thrown.
       * Actually this exception lumps together a lot of other
       * exceptions that may be caught during parsing.
       */
    class ModelError : public std::runtime_error {
    public:

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };

       /**
       * Class for reporting that the program was unable to open some
       * of the user specified files.
       */
    class FileNotFoundError : public std::runtime_error {
      
        /**
         * The FileNotFoundError constructor.
         *
         * This enables the program to abort or recover from faulty
         * filename (possibly supplied by the user).
         *
         * @param what More information of what went wrong.
         */
      public:
      FileNotFoundError( const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * The default constructor for the System class.
       *
       * This constructor will initialise everything into null and
       * void.  I am afraid it is not very useful.  Everything has to
       * be set to proper values using the remaining methods in this
       * class.
       */
    System();

      /**
       * Value constructor for the system class.
       *
       * This constructor will read the system from a pointer to an
       * SBML model.  The one-more-thing we need to build the model
       * that is not captured in SBML is the size and unit of the
       * individual sub volumes.  Note that this method assumes that
       * the sub volumes are cubic.  There are a few settings that are
       * not contained in the SBML model description, but that
       * nevertheless control how this instance of the System class
       * should operate.  Examples of such options are the frequency
       * at which checkpointing is performed.  These things are not
       * handled by this constructor, but by other methods in this
       * class.
       *
       * @param m    A pointer to the SBML model which defines the system.
       * @param size The side length of a cube used to build the geometry.
       * @param unit The unit which is used to specify the above size.
       */
    System(const Model_t* m, double size, const std::string& unit);

      /**
       * The System class destructor.
       *
       * This will free all the resources allocated to the system.
       * The only resource is in fact the time queue.
       */
    ~System();

      /**
       * Attempt to read in the model from an SBML file.
       *
       * This method will open the file as specified by the supplied
       * file name.  Following this, libsbml functions are used to
       * parse the SBML.  Finally, we try to extract what information
       * we need in order to rebuild the model in our own
       * representation.  Note again that this method assumes that
       * cubic sub volumes are used.
       *
       * Plenty of things can go wrong here, because we cannot expect
       * anybody to write perfect model definitions all the time.
       * Therefore we try to be a little verbose when things actually
       * do go bad.  The place to be verbose is the log file.  Thus,
       * in order to benefit from error reporting at all, the log file
       * must be set up prior to calling this method.
       *
       * @param filename The file name from which to read the SBML model.
       * @param size     The side length of a cube used to build the geometry.
       * @param unit     The unit which is used to specify the above size.
       */
    void fromSBML(
      const std::string& filename,
      double             size,
      const std::string& unit)
      throw(ModelError);

      /**
       * Attempt to build the model from an SBML model instance.
       *
       * This method is somewhat more low-level than the previous.
       * The objective however is the same.  The only difference is
       * that this method needs a parsed SBML model instance instead
       * of the name to a file containing one.  Actually, the previous
       * method will do the parsing of the SBML file, and then pass on
       * control to this method.
       *
       * @param m    The SBML model instance which we will use.
       * @param size The side length of a cube used to build the geometry.
       * @param unit The unit which is used to specify the above size.
       * @param prng Pointer to the pseudo-random number generator.
       */
    void fromSBML(
      const Model_t*     m,
      double             size,
      const std::string& unit)
      throw(ModelError);

      /**
       * Get a reference to a vector containing all the sub volumes of
       * the system.
       *
       * This method should be called by classes that whish to do
       * real-time analysis or visualisation of the system.  These
       * callers cannot modify the reference because it is const, and
       * in a multithreaded environment, one should never assume that
       * the values read are exactly true in any way, unless one can
       * control the running simulation and the analysis code with
       * mutexes.
       *
       * @return A vector containing all the sub volumes of the system.
       */
    inline const std::vector<SubVolume*>& getSubVolumeList() const
      { return subVolumes; }

      /**
       * Get the timestep for a deterministic simulation.  
       *
       * @return The timestep for a deterministic simulation.  
       */
    inline double getOptionTimeStep() const
      { return timeStep; }
    

      /**
       * Get the flag indicating if this is a deterministic simulation.  
       * 
       * @return The bool flag indicating deterministic simulation.  
       */
    inline bool getOptionDeterministicSimulation()
      { return deterministicSimulation; }
    

      /**
       * Output a histogram using all present species.
       *
       * For every integer, this method will count the number of sub
       * volumes containing that number of molecules.  The histogram
       * is outputted to the requested stream.  All species are
       * counted in the generation of the histogram.
       *
       * @param os The output stream on which to put the histogram.
       */
    void histogram(std::ostream& os) const;

      /**
       * Output a histogram using selected species only.
       *
       * For every integer, this method will count the number of sub
       * volumes containing that number of molecules.  The histogram
       * is outputted to the requested stream.  Only those species,
       * the indices of which are given in the argument list are
       * regarded.
       *
       * @param indices A list with the indices which we should count.
       * @param os      The output stream on which to put the histogram.
       */
    void histogram(const std::list<off_t>& indices, std::ostream& os) const;


      /**
       * @brief Set the number of iterations between checkpoints.
       *
       * This function will set the number of iterations to perform
       * between checkpoints.  If this is set to anything less or
       * equal to zero, no iteration-based checkpointing will be done.
       * If checkpointing is enabled, a checkpoint will always be done
       * for the first iteration step.  Further checkpoints are done
       * after the set amount of iterations or the set amount of time,
       * whichever comes first.
       *
       * @param ci The number of iterations between checkpoints.
       */
    inline void
    setOptionCheckpointIterations(const size_t ci)
      {
        checkpoint_iterations = ci;
      }


      /**
       * @brief Set the simulation time interval between checkpoints.
       *
       * This function will set the time interval, in simulation time
       * seconds, between checkpoints.  If this is set to anything
       * less or equal to zero, no time-based checkpointing will be
       * done.  If checkpointing is enabled, a checkpoint will always
       * be done for the first iteration step.  Further checkpoints
       * are done after the set amount of iterations or the set amount
       * of time, whichever comes first.
       *
       * @param ct Simulation time, in seconds, between checkpoints.
       */
    inline void
    setOptionCheckpointTime(const double ct)
      {
        checkpoint_time = ct;
      }


      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing and
       * model building begins, any error reporting and logging
       * information will be silently lost.  It is thus highly
       * recommended to call this method prior to calling fromSBML().
       * Note also that this method will set up the logging stream for
       * any member classes that can make use of one.
       *
       * @param l The logging stream onto which to report errors, etc.
       */
    void setOptionLogfile(std::ostream* l);
    
      /**
       * Sets the maximal number of iterations to perform.
       *
       * This will set the maximal number of iterations to perform.
       * If it is set to anything negative, there is no upper bound on
       * the number of iterations.  When simulating, the simulation
       * will stop as soon as the maximal number of iterations or the
       * simulation time is exceeded.
       *
       * @param max_iterations The maximal number of iterations.
       */
    inline void setOptionMaxIterations(const size_t iterations, 
      const size_t scale)
      { max_iterations = iterations; max_iterations_scale = scale; }

      /**
       * Sets the simulation time.
       *
       * This will set the simulation time.  If it is set to anything
       * negative, there is no upper bound on the simulation time.
       * When simulating, the simulation will stop as soon as the
       * maximal number of iterations or the simulation time is
       * exceeded.
       *
       * @param mt The simulation time.
       */
    inline void setOptionMaxTime(double mt)
      { max_time = mt; }

      /**
       * Set the out file prefix.
       *
       * This prefix will be prepended to all out files written
       * through this class.  Typically, it should contain path
       * information or something along those lines.
       *
       * @param prefix The prefix prepended to all out files.
       */
    inline void setOptionOutfilePrefix(const std::string& prefix)
      { outfile_prefix = prefix; }

      /**
       * Sets the progress bar pointer.
       *
       * If the System instance has been provided with a valid pointer
       * to an instance of the ProgressBar class, the simulation()
       * method will periodically do callback using the progress bar's
       * setFraction() method.  This enables the use of progress bars,
       * to indicate progress (quelle surprise).  Furthermore, the
       * progress bar can terminate the simulation method.  Should the
       * progress bar's isAlive() method return false at any point,
       * execution of the simulation method will cease gracefully.
       *
       * @param p Pointer to the ProgressBar instance to use.
       */
    inline void setOptionProgressBar(ProgressBar* p)
      { progressBar = p; }


      /**
       * Sets the timestep in a deterministic simulation.  
       *
       * @param ts The timestep for deterministic simulation.  
       */
    inline void setOptionTimeStep(double ts)
      { timeStep = ts; }
    
      /**
       * Sets the option to run a deterministic simulation.  
       * 
       * @param ds Flag indicating to run a deterministic simulation.  
       */
    inline void setOptionDeterministicSimulation(bool ds)
      { deterministicSimulation = ds; }

    /**
     * Sets the method by which the timestep is done.  
     *
     * @param intMeth The name of the integrationmethod.  Ie 'euler', 
     *                'eulerbackwards', possible more.  
     */
    inline void setOptionIntegrationMethod(std::string intMeth)
      { integrationMethod = intMeth; }

    /**
     * Sets the cut off for the tolerance in the iterative method used in 
     * the euler-backwards method.  The tolerance for a linear system of 
     * equations Ax=b is defined as tol=norm(r)/norm(b) where r = b-Ax.  
     *
     * @param tolerance The tolerance in the iterative method.  
     */
    inline void setOptionItMetConvergenceTolerance(double tolerance)
      { itMetConvergenceTolerance = tolerance; }

    /**
     * Sets the maximum number of iterations that the iterative method
     * will do before bailing out.  
     *
     * @param maxIterations The maximum number of iterations in the iterative
     *                      method.  
     */
    inline void setOptionItMetMaxIterations(const size_t maxIterations)
      { itMetMaxIterations = maxIterations; }

      /**
       * Sets the flag indicating the a state from an old simulation
       * should be used as initial state for starting the current
       * simulation.
       *
       * @param readOldState The old state flag.
       */
    inline void setOptionReadOldState(bool readOldState) 
      { this -> readOldState = readOldState; }

      /**
       * This will set the timepoint which defines which of the states
       * from the old simlation that should be used. The state is choosen
       * so that the state dumped at the specified timepoint is read.
       *
       * @param readOldStateTime The timepoint to use.
       */
    inline void setOptionReadOldStateTime(const size_t readOldStateTime) 
      { this -> readOldStateTime = readOldStateTime; }

      /**
       * This will set the sparse flag to enable the simulation to output
       * the subvolume data in a sparse format.
       *
       * @param outputSparse The sparse flag.
       */
    inline void setOptionOutputSparse(bool outputSparse) 
      { this -> outputSparse = outputSparse; }
    

      /**
       * This will set the fullstate checkpoint option.  
       * 
       * @param fullStateCheckPoint The fullstate checkpoint flag.  
       */
    inline void setOptionFullStateCheckPoint(bool fullStateCheckPoint) 
      { this -> fullStateCheckPoint = fullStateCheckPoint; }

      /**
       * Initiate the simulation
       *
       * This function will do error checking, optionally read states
       * from old simulation and setup the output files for the 
       * simulations.  
       */
    void simulationInit()
      throw( FileNotFoundError, ModelError );


      /**
       * Do post simulation cleanup.
       *
       * This will close the files opened by the simulation.  
       */
    void simulationFinalise();
    

      /**
       * Simulate the system.
       *
       * This is probably the most important method of this class.  It
       * will simulate the system, as described in the accompanying
       * documents.  In short, we find the sub volume that is at the
       * top of the time queue, and execute an event in it.  This will
       * change the time of the sub volume, and possibly another sub
       * volume if the event turned out to be a diffusion event.  The
       * affected sub volumes are resorted and we go for another spin
       * in the loop.  Thus we proceed, until we either exceed the
       * maximal number of iterations, simulation time or until the
       * progress bar signals that we should stop.  At certain
       * intervals we do checkpointing, writing the current state of
       * the system to file.
       */
    void simulate();

      /**
       * Simulate a deterministic system.  
       * 
       * This is the corresponding simulate method for simulating 
       * deterministic systems.  It will make use of both the derivate class 
       * and the integrate class. The derivate class is able to produce
       * the change in all the parameters in the state vector and the 
       * Integrate class will calculate the new state.  
       */
    void simulateDeterministic();
    

      /**
       * Write the geometry information to a file.
       *
       * The format of the file written is not very complicated.  The
       * sub volumes are emitted in the order they appear in the
       * system's sub volume vector.  For each sub volume, a line is
       * written which is of the form "x<ws>y<ws>z", where "x", "y"
       * and "z" are the x-, y- and z-coordinates of the centre of the
       * sub volume and "<ws>" denotes an arbitrary amount of white
       * space.
       *
       * @param filename Name of the file to which to write the geometry.
       */
    void writeGeometry(const std::string& filename);

      /**
       * Read a state from an old simluation allowing the user to start
       * from it.
       *
       * This will open the output-files used in a earlier simulation and
       * read the species_* files. At the time specified in the argument
       * the state will be read in and distributed to the subvolumes.
       *
       * @param lineNumber The line on the species_* files at which the
       *                   state can be fould.
       */
    void parseOldState(const size_t lineNumber) 
      throw( FileNotFoundError, ModelError );
    
      /**
       * Write a time checkpoint.
       *
       * The format of the time checkpoint file is not very
       * complicated.  The simulation time for each successive
       * checkpoint is written to the file on a separate line, one
       * line for each checkpoint.  Note that the simulation time
       * checkpoint file must be opened prior to this call.  The file
       * opening can be accomplished using the openTimeFile() method
       * in this class.
       *
       * @param time The simulation time of the checkpoint.
       */
    inline void checkpointTime(double time)
      { if (outfile_time) (*outfile_time) << time << std::endl; }
    
      /**
       * Closes the time checkpoint file.
       *
       * This should be done before terminating the simulate() method.
       * It is strictly speaking not necessary, because most modern
       * operating systems will close any open files on program
       * termination.
       */
    inline void closeTimeFile()
      { if (outfile_time) { outfile_time -> close(); delete outfile_time; } }

      /**
       * Opens the time checkpoint file.
       *
       * This method must be called before calling the
       * checkpointTime() method.  The method will open the named file
       * for writing, removing any of its previous contents in the
       * process.  For looks, one should call closeTimeFile() before
       * leaving the simulate() method.
       *
       * @param filename The name of the time checkpoint file to open.
       */
    inline void openTimeFile(const std::string& filename)
      { closeTimeFile(); outfile_time = new std::ofstream(
        filename.c_str(), std::ios::trunc); }
  };
}
#endif /* ifndef MESORD_SYSTEM_H */
