/* The Thread class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 * * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Thread.cpp 561 2008-11-25 20:10:53Z davidfange $
 */

#include "System.hpp"
#include "Thread.hpp"

#if HAVE_VIEWER
#  include "Visual.hpp"
#endif

#include "ui/GetOpt.hpp"

#ifdef WIN32
#  include "ui/WoeProgressBar.hpp"
#else
//#  include "ui/GraphicProgressBar.hpp"
#  include "ui/TextProgressBar.hpp"
#endif

#include "mesord.hpp"

#ifdef WIN32
#include "windows.h"
#pragma comment(lib,"kernel32.lib")
#endif

  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Initialisation.
   */
MesoRD::Thread::Thread()
{
  running = false;

#ifdef HAVE_PTHREAD
  handle  = 0;

  pthread_attr_init(&attributes);
  pthread_cond_init(&initialised_condition, NULL);
  pthread_mutex_init(&initialised_mutex, NULL);

  pthread_mutex_lock(&initialised_mutex);
  initialised_variable = false;
  pthread_mutex_unlock(&initialised_mutex);
#endif

#ifdef WIN32
  id     = 0;
  handle = NULL;

  initialised_event = CreateEvent(NULL, TRUE, FALSE, NULL);
  ResetEvent(initialised_event);
  //  initialised_mutex = CreateMutex(NULL, false, NULL);
//  WaitForSingleObject(initialised_mutex, INFINITE);
//  initialised_variable = false;
//  ReleaseMutex(initialised_mutex);
#endif
}

MesoRD::Thread::~Thread()
{
  cancel();
}



  /*********************
   * THREAD MANAGEMENT *
   *********************/

  /* Cancel the thread and hope that it was not doing anything
   * critically important.
   */
#ifdef HAVE_PTHREAD
void
MesoRD::Thread::cancel()
#endif
#ifdef WIN32
DWORD
MesoRD::Thread::cancel()
#endif
{
#ifdef WIN32
  DWORD returnCode;
#endif

  if (running) {
#ifdef HAVE_PTHREAD

  int   error;

  char* error_string;
  
  if ((error = pthread_cancel(handle))) {
    error_string = strerror(error);
  }

  running = false;

    /* wait for the thread to cancel, this may take especially long
     * time if there are alot of subvolumes and molecules to
     * visualize.
     */
  pthread_join(handle, NULL);
  
#endif
#ifdef WIN32
    if (!TerminateThread(handle, 0)) {
//    std::string error_string = getLastErrorString();
      if (!GetExitCodeThread(handle, &returnCode)) {
//    std::string error_string = getLastErrorString();    
        if (!CloseHandle(handle)) {
//    std::string error_string = getLastErrorString();
        }
      }
    }
    running = false;
#endif
  }
  
#ifdef HAVE_PTHREAD
  return;
#endif
#ifdef WIN32
  return returnCode;
#endif
}



  /* Create the thread.  Make sure that the "running" flag indicates
   * the current state of the thread.
   */
#ifdef HAVE_PTHREAD
void
MesoRD::Thread::create(MESORD_THREAD_FUNCTION function, void* arguments)
#endif
#ifdef WIN32
void
MesoRD::Thread::create(LPTHREAD_START_ROUTINE function, LPVOID arguments)
#endif
{
#ifdef HAVE_PTHREAD
  if (pthread_create(&handle, &attributes, function, arguments)) {
//      char* error_string = strerror(error);
    running = false; // throw an exception
  } else {
    running = true;
  }
#endif

#ifdef WIN32
  if ((handle = CreateThread(
         NULL,
         0,
         function,
         arguments,
         0,
         (LPDWORD)(&id)))) {
    running = true;
  } else {
//    std::string error_string = getLastErrorString();
    running = false; // throw an exception
  }
#endif

  return;
}


  /* Wait for the thread to terminate.
   */
#ifdef HAVE_PTHREAD
void*
MesoRD::Thread::join()
#endif
#ifdef WIN32
DWORD
MesoRD::Thread::join()
#endif
{
    
#ifdef HAVE_PTHREAD
  void* returnCode = 0;
#endif
#ifdef WIN32
  DWORD returnCode = 0;
#endif

  if (running) {
#ifdef HAVE_PTHREAD
    if (pthread_join(handle, &returnCode)) {
//      char* error_string = strerror(error);
    }
#endif
#ifdef WIN32
    WaitForSingleObject(handle, INFINITE);

    if (!GetExitCodeThread(handle, &returnCode)) {
//    std::string error_string = getLastErrorString();    
      if (!CloseHandle(handle)) {
//    std::string error_string = getLastErrorString();
      }
    }
#endif
    running = false;
  }
  
  return returnCode;
}


  /* Mark the initialisation as done.
   */
void
MesoRD::Thread::setInitialised(bool initialised)
{
#ifdef HAVE_PTHREAD
  pthread_mutex_lock(&initialised_mutex);
  initialised_variable = initialised;
  pthread_cond_signal(&initialised_condition);
  pthread_mutex_unlock(&initialised_mutex);
#endif
#ifdef WIN32
  if (initialised)
	  SetEvent(initialised_event);
  else
	  ResetEvent(initialised_event);
#endif
}


  /* Wait until the initialisation is done.  The polling for the Win32 code is horribly stupid.
   */
void
MesoRD::Thread::waitForInitialisation()
{
#ifdef HAVE_PTHREAD
  pthread_mutex_lock(&initialised_mutex);
  while (!initialised_variable)
    pthread_cond_wait(&initialised_condition, &initialised_mutex);
  pthread_mutex_unlock(&initialised_mutex);  
#endif
#ifdef WIN32
  WaitForSingleObject(initialised_event, INFINITE);
#endif
}



  /*******************
   * PROGRESS THREAD *
   *******************/

  /* The argument struct for the progress bar thread.
   */
typedef struct arguments_progress {
#ifdef WIN32
  int                     nCmdShow;
  HINSTANCE               hInstance;
  HWND                    hwnd;
#endif
  bool                    graphic;
  MesoRD::System*         system;
  MesoRD::ThreadProgress* threadProgress;
} arguments_progress;


  /* Needs a pointer to the running System instance.  The reason for
   * this is that the progress bar must register itself with the
   * system in order to of any use.  Note that the run() method in the
   * ProgressBar class does not return until the ProgressBar instance
   * has gone to completion, or if the simulation was aborted.  So we
   * wait, and once run() is done, we delete the ProgressBar instance
   * and kill the thread.
   */
extern "C" 
{
#ifdef HAVE_PTHREAD
  static void*
  function_progress(void* arguments)
#endif
#ifdef WIN32
    static DWORD WINAPI
  function_progress(LPVOID arguments)
#endif
  {
    if (arguments) {
      MesoRD::ProgressBar*    progressBar = 0;
      MesoRD::System*         system      = ((arguments_progress *)arguments)
        -> system;
      MesoRD::ThreadProgress* thread      = ((arguments_progress *)arguments)
        -> threadProgress;

      if (system && thread) {
#ifdef HAVE_PTHREAD
        if (((arguments_progress *)arguments) -> graphic)
          progressBar = 0; //new GraphicProgressBar(); // FIXME!!
        else
          progressBar = new MesoRD::TextProgressBar();
#endif
#ifdef WIN32
        progressBar = new MesoRD::WoeProgressBar(
          ((arguments_progress *)arguments) -> hInstance,
          ((arguments_progress *)arguments) -> hwnd,
          ((arguments_progress *)arguments) -> nCmdShow);
#endif

        system -> setOptionProgressBar(progressBar);
        thread -> setInitialised(true);

        progressBar -> run();


          /* Do not delete the progress bar instance, as it must hang
           * around for long enough for the simulator to poll it.
           * This could certainly be solved much nicer, when this
           * class is revised.
           */
        for ( ; ; )
#ifdef WIN32
          Sleep(1);
#else
			sleep(1);
#endif
/*
        if (progressBar)
          delete progressBar;
*/
      }
    }

    return 0;
  }
}



  /* Constructor.
   */
MesoRD::ThreadProgress::ThreadProgress()
  : Thread()
{}


  /* Destructor.
   */
MesoRD::ThreadProgress::~ThreadProgress()
{}


  /* Create function.  In Windows, the progress bar is always
   * graphical, so we con safely hardcode that argument to true.
   */
#ifdef HAVE_PTHREAD
void
MesoRD::ThreadProgress::createProgress(bool graphic, System* system)
#endif
#ifdef WIN32
void
MesoRD::ThreadProgress::createProgress(
  HINSTANCE hInstance, HWND hwnd, int nCmdShow, System* system)
#endif
{  
  arguments_progress arguments;

#ifdef HAVE_PTHREAD
  arguments.graphic        = graphic;
  arguments.system         = system;
  arguments.threadProgress = this;
#endif
#ifdef WIN32
  arguments.nCmdShow       = nCmdShow;
  arguments.hInstance      = hInstance;
  arguments.hwnd           = hwnd;
  arguments.graphic        = true;
  arguments.system         = system;
  arguments.threadProgress = this;
#endif

#ifdef HAVE_PTHREAD
  Thread::create(function_progress, (void*)&arguments);
#endif
#ifdef WIN32
  Thread::create(function_progress, (LPVOID)&arguments);
#endif

  waitForInitialisation();

  return;
}



  /********************
   * SIMULATOR THREAD *
   ********************/

/* !! FIXME!! Pending removal, no need for a simulator thread.
typedef struct arguments_simulator {
  MesoRD::System*          system;
  MesoRD::ThreadSimulator* threadSimulator;
} arguments_simulator;
*/

  /* Needs a pointer to the system instance to simulate.  The system
   * must be initialsed before the thread is started.
   */
/*extern "C" 
{
#ifdef HAVE_PTHREAD
  static void*
  function_simulator(void* arguments)
#endif
#ifdef WIN32
    static DWORD WINAPI
  function_simulator(LPVOID arguments)
#endif
  {
    if (arguments) {
      MesoRD::System*          system = ((arguments_simulator*)arguments) -> system;
      MesoRD::ThreadSimulator* thread = ((arguments_simulator*)arguments) -> threadSimulator;

      if (system && thread) {
          //std::cout << "Starting Simulator thread..." << std::endl; //<< arguments << " and " << ((arguments_simulator *)arguments) -> system << " and " << ((arguments_simulator *)arguments) -> threadSimulator << std::endl;
        thread -> setInitialised(true);
#ifdef DETERMINISTIC
        if(system -> getOptionDeterministicSimulation())
        { 
          system -> simulateDeterministic();
        }
        else
        {
          system -> simulate();
        }
#else
        system -> simulate();
        
#endif
        
      }
    }
    return 0;
  }
}

*/

  /* Constructor.
   */
/*MesoRD::ThreadSimulator::ThreadSimulator()
  : Thread()
{}
*/

  /* Destructor.
   */
/*MesoRD::ThreadSimulator::~ThreadSimulator()
{}
*/

  /* Create function.
   */
/* #ifdef HAVE_PTHREAD
void
MesoRD::ThreadSimulator::createSimulator(System* system)
#endif
#ifdef WIN32
void
MesoRD::ThreadSimulator::createSimulator(System* system)
#endif
{  
  arguments_simulator arguments;

  arguments.system          = system;
  arguments.threadSimulator = this;


#ifdef HAVE_PTHREAD
  Thread::create(function_simulator, (void*)&arguments);
#endif
#ifdef WIN32
  Thread::create(function_simulator, (LPVOID)&arguments);
#endif

  waitForInitialisation();

  return;
}
*/

  /*****************
   * VIEWER THREAD *
   *****************/
#if HAVE_VIEWER
typedef struct arguments_visualiser {
#ifdef WIN32
  int                       nCmdShow;
  HINSTANCE                 hInstance;
  HWND                      hwnd;
#endif
  MesoRD::GetOpt*           getOpt;
  MesoRD::System*           system;
  MesoRD::ThreadVisualiser* threadVisualiser;
} arguments_visualiser;


  /* Needs a pointer to the viewer instance.  The viewer instance must
   * be initialised before the thread is started.  We return NULL
   * instead of 0 in honour of the C-style calling conventions used
   * for the pthread functions.
   */
extern "C" 
{
#ifdef HAVE_PTHREAD
  static void*
  function_visualiser(void* arguments)
#endif
#ifdef WIN32
    static DWORD WINAPI
  function_visualiser(LPVOID arguments)
#endif
  {
    if (arguments) {
      MesoRD::GetOpt*           getOpt
        = ((arguments_visualiser*)arguments) -> getOpt;
      MesoRD::System*           system
        = ((arguments_visualiser*)arguments) -> system;
      MesoRD::ThreadVisualiser* thread
        = ((arguments_visualiser*)arguments) -> threadVisualiser;
      if (system) {
        MesoRD::Visual* visual;

        try {
#ifdef HAVE_PTHREAD
          visual = new MesoRD::Visual(system -> getSubVolumeList());
#endif
#ifdef WIN32
          visual = new MesoRD::Visual(
            ((arguments_visualiser *)arguments) -> hInstance,
            ((arguments_visualiser *)arguments) -> nCmdShow,
            ((arguments_visualiser *)arguments) -> system -> getSubVolumeList());
#endif
        } catch (MesoRD::Visual::InitialisationError& e) {
          std::cerr << "Could not start viewer thread." << std::endl
                    << "  EXCEPTION: " << e.what()      << std::endl;
          return 0;
        }
      
        if (getOpt) {
          visual -> setColours(getOpt -> getColours());
          visual -> setSpeciesDimension(getOpt -> getOptionSpeciesDimension());
        }
        
        thread -> setInitialised(true);

        for ( ; ; ) {
          visual -> animateStep();
        }

        delete visual;
      }
    }
  
    return 0;
  }
}



  /* Constructor.
   */
MesoRD::ThreadVisualiser::ThreadVisualiser()
  : Thread()
{}


  /* Destructor.
   */
MesoRD::ThreadVisualiser::~ThreadVisualiser()
{}


  /* Create function.
   */
#ifdef HAVE_PTHREAD
void
MesoRD::ThreadVisualiser::createVisualiser(GetOpt* getOpt, System* system)
#endif
#ifdef WIN32
void
MesoRD::ThreadVisualiser::createVisualiser(
  HINSTANCE hInstance, HWND hwnd, int nCmdShow, GetOpt* getOpt, System* system)
#endif
{  
  arguments_visualiser arguments;

#ifdef HAVE_PTHREAD
  arguments.getOpt           = getOpt;
  arguments.system           = system;
  arguments.threadVisualiser = this;
#endif
#ifdef WIN32
  arguments.nCmdShow         = nCmdShow;
  arguments.hInstance        = hInstance;
  arguments.hwnd             = hwnd;
  arguments.getOpt           = getOpt;
  arguments.system           = system;
  arguments.threadVisualiser = this;
#endif

#ifdef HAVE_PTHREAD
   Thread::create(function_visualiser, (void*)&arguments);
#endif


#ifdef WIN32
  Thread::create(function_visualiser, (LPVOID)&arguments);
#endif

  waitForInitialisation();

  return;
}
#endif
