/* The Unit class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Unit.cpp 554 2008-05-25 19:55:15Z davidfange $
 */
#include <math.h>

#include <algorithm>
#include <iostream>
#include <string>

#include "Unit.hpp"
#include "ui/OPrefixStream.hpp"


  /* Reduces a unit to a product of basic or supplementary units
   * according to rules found in "Physics Handbook".  This is not very
   * difficult, if one has such a table of unit definitions at hand.
   * The function starts by checking whether the respective fields are
   * set in the SBML.  If they are, their values are retrieved,
   * otherwise attempts are made to use some sensible defaults.  Note
   * that we must ensure that we start adding units to an initially
   * empty vector.  The remainder of the function essentially only
   * features one huge switch statement, handling each possible unit
   * factor that can be used in composing the unit.  Note that
   * exponents on the muliplier are applied directly, so that the
   * internal representation of MesoRDs multipliers do not depend on
   * the exponent of the unit.
   */
void
MesoRD::Unit::fromSBML(const UnitDefinition* ud)
{
  if( ud -> isSetId() )
    id =   ud -> getId();

  if( ud -> isSetName() )
    name = ud -> getName();
  
  factors.clear();
  for (unsigned int i = 0; i < ud -> getNumUnits(); i++) {
    
    const ::Unit *u = ud -> getUnit(i);
    
    int    e = u -> getExponent();
    double m = pow(u -> getMultiplier() * pow((double)10, u -> getScale()), e);
    
    switch ( u -> getKind() ) {

        /* Ampere is a basic unit.
         */
      case UNIT_KIND_AMPERE:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::ampere,
            m));
        break;

          /* Becquerel is equivalent to s^(-1).
           */
      case UNIT_KIND_BECQUEREL:
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::second,
            m));
        break;

          /* Candela is a basic unit.
           */
      case UNIT_KIND_CANDELA:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::ampere,
            m));
        break;

          /* Coulomb is equivalent to A^(+1) * s^(+1).
           */
      case UNIT_KIND_COULOMB:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::ampere,
            m));
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Dimensionless is a basic unit in this context.
           */
      case UNIT_KIND_DIMENSIONLESS:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::dimensionless,
            m));
        break;

          /* Coulomb is equivalent to A^(+2) * s^(+4) * kg^(-1) *
           * m^(-2).
           */
      case UNIT_KIND_FARAD:
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::ampere,
            m));
        factors.push_back(
          UnitFactor(
            +4 * e,
            UnitFactor::second,
            1.0));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::kilogram,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::metre,
            1.0));
        break;

          /* Gram is equivalent to 10^(-3) * kg^(+1).
           */
      case UNIT_KIND_GRAM:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m * pow(1.0e-3, e)));
        break;

          /* Gray is equivalent to m^(+2) * s^(-2).
           */
      case UNIT_KIND_GRAY:
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            m));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Henry is equivalent to kg^(+1) * m^(+2) * A^(-2) *
           * s^(-2).
           */
      case UNIT_KIND_HENRY:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::ampere,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Hertz is equivalent to s^(-1).
           */
      case UNIT_KIND_HERTZ:
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::second,
            m));
        break;

          /* Once invalid -- always invalid.
           */
      case UNIT_KIND_INVALID:
        break;

          /* Item is equivalent to dimensionless.
           */
      case UNIT_KIND_ITEM:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::dimensionless,
            m));
        break;

          /* Joule is equivalent to kg^(+1) * m^(+2) * s^(-2).
           */
      case UNIT_KIND_JOULE:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Katal is equivalent to (Avogadro's number) *
           * dimensionless^(+1) * s^(-1)
           */
      case UNIT_KIND_KATAL:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::dimensionless,
            m * pow(Constant::Avogadro, e)));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Kelvin is a basic unit.
           */
      case UNIT_KIND_KELVIN:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kelvin,
            m));
        break;

          /* Kilogram is a basic unit.
           */
      case UNIT_KIND_KILOGRAM:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        break;

          /* Litre is equivalent to 10^(-3) * m^(+3).
           */
      case UNIT_KIND_LITER:
      case UNIT_KIND_LITRE:
        factors.push_back(
          UnitFactor(
            +3 * e,
            UnitFactor::metre,
            m * pow(1.0e-3, e)));
        break;

          /* Lumen is equivalent to cd^(+1) * sr^(+1).
           */
      case UNIT_KIND_LUMEN:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::candela,
            m));
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::steradian,
            1.0));
        break;

          /* Lux is equivalent to cd^(+1) * sr^(+1) * m^(-2).
           */
      case UNIT_KIND_LUX:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::candela,
            m));
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::steradian,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::metre, (double)0));
        break;

          /* Metre is a basic unit.
           */
      case UNIT_KIND_METER:
      case UNIT_KIND_METRE:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::metre,
            m));
        break;

          /* Mole is equivalent to (Avogadro's number) *
           * dimensionless^(+1).
           */
      case UNIT_KIND_MOLE:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::dimensionless,
            m * pow(Constant::Avogadro, e)));
        break;

          /* Newton is equivalent to kg^(+1) * m^(+1) * s^(-2).
           */
      case UNIT_KIND_NEWTON:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Ohm is equivalent to kg^(+1) * m^(+2) * A^(-2) *
           * s^(-3).
           */
      case UNIT_KIND_OHM:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::ampere,
            1.0));
        factors.push_back(
          UnitFactor(
            -3 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Pascal is equivalent to kg^(+1) * m^(-1) * s^(-2).
           */
      case UNIT_KIND_PASCAL:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Radian is a basic unit in this context.
           */
      case UNIT_KIND_RADIAN:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::radian,
            m));
        break;

          /* Second is a basic unit.
           */
      case UNIT_KIND_SECOND:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::second,
            m));
        break;

          /* Siemens is equivalent to A^(+2) * s^(+3) * kg^(-1) *
           * m^(-2).
           */
      case UNIT_KIND_SIEMENS:
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::ampere,
            m));
        factors.push_back(
          UnitFactor(
            +3 * e,
            UnitFactor::second,
            1.0));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::kilogram,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::metre,
            1.0));
        break;

          /* Sievert is equivalent to m^(+2) * s^(-2).
           */
      case UNIT_KIND_SIEVERT:
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            m));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Steradian is a basic unit in this context.
           */
      case UNIT_KIND_STERADIAN:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::steradian,
            m));
        break;

          /* Tesla is equivalent to kg^(+1) * A^(-1) * m^(-1) *
           * s^(-2).
           */
      case UNIT_KIND_TESLA:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::ampere,
            1.0));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Volt is equivalent to kg^(1) * m^(2) * A^(-1) *
           * s^(-3).
           */
      case UNIT_KIND_VOLT:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::ampere,
            1.0));
        factors.push_back(
          UnitFactor(
            -3 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Watt is equivalent to kg^(+1) * m^(+2) * s^(-3).
           */
      case UNIT_KIND_WATT:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -3 * e,
            UnitFactor::second,
            1.0));
        break;

          /* Weber is equivalent to kg^(+1) * m^(+2) * A^(-1) *
           * s^(-2).
           */
      case UNIT_KIND_WEBER:
        factors.push_back(
          UnitFactor(
            +1 * e,
            UnitFactor::kilogram,
            m));
        factors.push_back(
          UnitFactor(
            +2 * e,
            UnitFactor::metre,
            1.0));
        factors.push_back(
          UnitFactor(
            -1 * e,
            UnitFactor::ampere,
            1.0));
        factors.push_back(
          UnitFactor(
            -2 * e,
            UnitFactor::second,
            1.0));
        break;
            
    }
  }

  fold();
}


  /* Attempts to find a set of conversion parameters to convert from
   * the current unit to the supplied unit.  The basic idea is to
   * check that for every unit factor in the "from" unit, there is a
   * corresponding, compatible unit factor in the "to" unit.
   * Trivially, two unit factors are compatible if they are completely
   * identical.  If the factors are not completely identical, at least
   * their kinds have to be identical for them to match.  In addition,
   * the two factors must have the same exponent, or both be
   * dimensionless, in which case the exponent does not matter.  If
   * there are matches for all the factors in the left hand unit, the
   * two units are in fact compatible.  This follows because if the
   * units are folded correctly, there can be only one factor of each
   * kind in the unit.  In the implementation, note that the "i"
   * index, unlike the "j" index, is signed.  The reason is the "i--"
   * done at the end of the loop.  If "i" were unsigned, it may cause
   * strange overflow instabilities.
   */
double
MesoRD::Unit::toUnit(const Unit& u) const
  throw(InvalidConversion)
{
  
  double                  conversion = 1.0;
  std::vector<UnitFactor> from(this -> factors);
  std::vector<UnitFactor> to(u.factors);

  
  for (int i = 0; i < (signed)from.size(); i++) {
    for (unsigned int j = 0; j < to.size(); j++) {
      bool              match = false;
      const UnitFactor& f     = from[i];
      const UnitFactor& t     = to[j];

      if (f == t) {
        match = true;

      } else if (f.getKind() == t.getKind()
        && (f.getExponent() == t.getExponent()
          || f.getKind() == UnitFactor::dimensionless)) {
        
          conversion *= f.getMultiplier() / t.getMultiplier();
          match = true;
      }

      if (match) {
                
        unsigned int k;
        
        for (k = i; k < from.size() - 1; k++)
          from[k] = from[k + 1];
        from.pop_back();

        for (k = j; k < to.size() - 1; k++)
          to[k] = to[k + 1];
        to.pop_back();

        i--;        
        break;
      }
    }
  }

  if (from.size() != 0 || to.size() != 0)
    throw InvalidConversion(this, &u);
  return conversion;
}


  /* The general idea is to compare all unique pairs of factors in the
   * unit definition, and, if possible, fold them into a single
   * factor.  If any factor is dimensionless or, equivalently, has
   * exponent of zero, it may be regarded as being of any kind.  In
   * particular, it may take the kind of the factor currently compared
   * with.  During folding, exponents are added and multipliers
   * multiplied. If folding two factors yields a dimensionless first
   * factor, we break out of the inner loop and repeat the last step
   * in the outer loop for the new dimensionless factor.  This means
   * that newly formed dimensionless factors are handled immediately.
   * Finally, all multipliers are collected into the multiplier of the
   * first factor of the unit.
   */
void
MesoRD::Unit::fold()
{
  for (int i = 0; i < (signed)factors.size(); i++) {
    for (unsigned int j = 0; j < factors.size(); j++) {
      if (i != (signed)j) {
        bool        folded = false;
        UnitFactor& fi     = factors[i];
        UnitFactor& fj     = factors[j];

        if (fi.getExponent() == 0
          || fi.getKind() == UnitFactor::dimensionless) {
          fi.setExponent(0);
          fi.setKind(fj.getKind());
        }

        if (fj.getExponent() == 0
          || fj.getKind() == UnitFactor::dimensionless) {
          fj.setExponent(0);
          fj.setKind(fi.getKind());
        }

        if (fi.getKind() == fj.getKind()) {

            //FIXME!! isn't this taken care of above??
          int true_exponent_i = fi.getKind() == UnitFactor::dimensionless
            ? 0 : fi.getExponent();
          int true_exponent_j = fj.getKind() == UnitFactor::dimensionless
            ? 0 : fj.getExponent();

          fi.setExponent(true_exponent_i + true_exponent_j);
          fi.setMultiplier(fi.getMultiplier() * fj.getMultiplier());
         
          for (unsigned int k = j; k < (factors.size() - 1); k++)
            factors[k] = factors[k + 1];
          factors.pop_back();
          folded = true;
        }
        if(folded && fi.getExponent() == 0) {
          i--;
          break;
        }
        
      }
    }
  }
  

    /* Gather multiplier to the foremost factor */
  std::sort(factors.begin(), factors.end());
  for (unsigned int l = 1; l < factors.size(); l++) {
    if (fabs(factors[l].getMultiplier() - 1.0)
      >= std::numeric_limits<double>::epsilon()) {
      factors[0].setMultiplier(
        factors[0].getMultiplier() * factors[l].getMultiplier());
      factors[l].setMultiplier(1.0);
    }
  }
}



  /* Check that the two units have the same number of factors in their
   * definitions.  Then, for every factor in the left hand unit, check
   * that there is a corresponding, compatible factor in the right
   * hand unit.  As soon as we fail to find a match, we may
   * return false.  If we have found matches for all the factors in
   * the left hand unit, we know that the two units are in fact
   * identical, because if the units are folded correctly, there can
   * be only one factor of each kind in any unit.
   */
bool
MesoRD::Unit::operator== (const Unit& u) const
{
  if (factors.size() != u.factors.size())
    return false;
  
  for (unsigned int i = 0; i < factors.size(); i++) {
    unsigned int j;
    for (j = 0; j < factors.size(); j++) {
      if (factors[i] == u.factors[j])
        break;
      
      if (j >= factors.size())
        return false;
    }
  }
  return true;
}



  /* Output operator.  Should be a valid SBML "unitDefinition".  Note
   * that since we do not use the "scale" field, but instead collapse
   * it into a multiplier, the scale will always be zero.
   */
namespace MesoRD {
  std::ostream&
  operator<< (std::ostream& os, const Unit& u)
  {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
    OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");

    opfs_level0 << "<unitDefinition id=\"" << u.id << "\"";
    if (u.name.length() > 0)
      opfs_level0 << std::endl << "                name=\"" << u.name << "\"";
    opfs_level0 << ">" << std::endl;

    opfs_level1 << "<listOfUnits>" << std::endl;

    std::vector<UnitFactor>::const_iterator i;
    for (i = u.factors.begin(); i != u.factors.end(); i++)
      opfs_level2 << *i << std::endl;

    opfs_level1 << "</listOfUnits>" << std::endl;
    opfs_level0 << "</unitDefinition>" << std::endl;
    return os;
  }
}
