/* The Unit class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Unit.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_UNIT_H
#define MESORD_UNIT_H 1

#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>
#include <algorithm>
#include <limits>
#include <math.h>


#include "Symbol.hpp"
#include "UnitFactor.hpp"

#include "definition.hpp"

#include <sbml/SBMLTypes.h>

namespace MesoRD
{


    /**
     * Class for units.
     *
     * A unit in MesoRD, or equivalently a unit definition in SBML, is
     * a product of one or more basic unit factors.  Unit factors are
     * defined by the UnitFactor class.  The Unit class is essentially
     * a vector of unit factors; the unit is the product of the
     * factors in this vector.
     *
     * The Unit class here provides some features that would be nice
     * to have in SBML.  It seems like libSBML has included some unit
     * handling (as of libsbml-3.1.1), but it does still not provide
     * the same functionality as this class.  Using the Unit class it
     * is for instance possible to convert among different units.
     * Some basic operators, such as addition, multiplication and
     * exponentiation, can be applied to units.
     */
  class Unit : public Symbol
  {
  private:
    std::vector<UnitFactor> factors;

  public:
    
      /**
         * The multiplier needed for the conversion.  When converting
         * from one unit to another, the first unit should be
         * multiplied by this value.
         */
      double multiplier;

      /**
       * Class for reporting invalid unit conversion attempts.
       *
       * This exception is thrown on attempting to convert one unit to
       * anther, incompatible, unit.  For instance, a metre cannot be
       * converted to kilogramme.  Note that exception classes should
       * never throw exceptions themselves
       */
    class InvalidConversion : public std::runtime_error
    {
    private:
      const Unit* from;
      const Unit* to;

    public:


        /**
         * The InvalidConversion value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be null pointers.  Invalid pointers on
         * the other hand should be avoided, as they have the
         * potential to cause much confusion.
         *
         * @param from The unit which one could not convert from.
         * @param to   The unit which one could not convert to.
         */
      inline InvalidConversion(const Unit* from, const Unit* to)
        : std::runtime_error("InvalidConversion"),
          from(from), to(to)
        {}


        /**
         * Get the unit one could not convert from.
         *
         * This function returns a pointer to the unit which could not
         * be converted from.  This may be the null pointer.
         *
         * @return A pointer to the unit one could not convert from.
         */
      inline const Unit*
      getFrom() const
        {
          return from;
        }


        /**
         * Get the unit one could not convert to.
         *
         * This function returns a pointer to the unit which could not
         * be converted to.  This may be the null pointer.
         *
         * @return A pointer to the unit one could not convert to.
         */
      inline const Unit*
      getTo() const
        {
          return to;
        }
    };


      /**
       * Default constructor for the Unit class.
       *
       * Creates a new unit with an empty unit factor vector.  Note
       * that this is not quite the same as setting the unit to
       * dimensionless.  Maybe the unit should be set to dimensionless
       * in order to avoid have these null units hanging around.  See
       * also comment on the dimensionless static.
       */
      inline Unit()
        : Symbol("", "")
      {
        factors.clear();
      }


      /**
       * Value constructor for the Unit class.
       *
       * Copies all of the supplied data into the unit.  This
       * constructor relies on STL vector assignment to work properly.
       *
       * @param factors The product vector of unit factors.
       * @param id      The id of the new unit.
       * @param name    The name of the new unit.
       */
    inline Unit(const std::vector<UnitFactor>& factors,
      const std::string& id = "", const std::string& name = "")
      : Symbol(id, name), factors(factors)
      {}


      /**
       * Value constructor for the Unit class.
       *
       * This constructor will construct the unit from data given in
       * an SBML model.  The function uses fromSBML() internally.
       *
       * @param ud The SBML unit definition to construct from.
       */
    inline Unit(const UnitDefinition* ud)
      : Symbol("", "")
      {
        fromSBML(ud);
      }


      /**
       * Copy constructor for the Unit class. 
       *
       * @param u The unit object to be copied.  
       */
    inline Unit(const Unit &u) 
      : Symbol(u.id, u.name), factors(u.factors)
      {
        std::sort(factors.begin(), factors.end());
      }
    

      /**
       * Return a "centimetre squared per second" unit.
       *
       * This function will create a "centimetre squared per second"
       * unit and return it.
       *
       * @return The "centimetre squared per second" unit.
       */
    inline static Unit
    centimetreSquaredPerSecond()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+2, UnitFactor::metre,  1.0e-4));
        f.push_back(UnitFactor(-1, UnitFactor::second, 1.0e+0));
        Unit u(f);
        return u;
      }


      /**
       * Return a "dimensionless" unit.
       *
       * This function will create a "dimensionless" unit and return
       * it.  It should be used instead of creating a new unit with
       * the default constructor, as the latter would not necessarily
       * be dimensionless.
       *
       * @return The "dimensionless" unit.
       */
    inline static Unit
    dimensionless()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+0, UnitFactor::dimensionless, 1.0e+0));
        Unit u(f);
        return u;
      }


      /**
       * Return a "litre" unit.
       *
       * This function will create a "litre" unit and return it.  A
       * litre is one thousandth of a cubic metre.
       *
       * @return The "litre" unit.
       */
    inline static Unit
    litre()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+3, UnitFactor::metre, 1.0e-3));
        Unit u(f);
        return u;
      }


      /**
       * Return a "metre" unit.
       *
       * This function will create a "metre" unit and return it.
       * Metre is a basic SI unit.
       *
       * @return The "metre" unit.
       */
    inline static Unit
    metre()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+1, UnitFactor::metre, 1.0e+0));
        Unit u(f);
        return u;
      }

      /**
       * Return a "centimetre" unit.  
       *
       * This function will create a "centimetre" unit and return it.  
       *
       * @return The "centimetre" unit.  
       */
    inline static Unit
    centimetre()
      {
       std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+1, UnitFactor::metre, 1.0e-2));
        Unit u(f);
        return u;
      } 
    
      /**
       * Return a mole unit.
       *
       * This function will create a "mole" unit and return it.  A
       * "mole" is nothing but a dimensionless unit with a multiplier.
       *
       * @return The "mole" unit.
       */
    inline static Unit
    mole()
      {  
        std::vector<UnitFactor> f;
        f.push_back(
          UnitFactor(+1, UnitFactor::dimensionless, Constant::Avogadro));
        Unit u(f);
        return u;
      }


      /**
       * Return a "molar per second" unit.
       *
       * This function will create a "molar per second" unit and
       * return it.
       *
       * @return The "molar per second" unit.
       */
    inline static Unit
    molarPerSecond()
      {
        std::vector<UnitFactor> f;
        f.push_back(
          UnitFactor(+1, UnitFactor::dimensionless, Constant::Avogadro));
        f.push_back(UnitFactor(-3, UnitFactor::metre,  1.0e+3));
        f.push_back(UnitFactor(-1, UnitFactor::second, 1.0e+0));
        Unit u(f);
        u.fold();
        return u;
      }


      /**
       * Return a "per molar per second" unit.
       *
       * This function will create a "per molar per second" unit and
       * return it.
       *
       * @return The "per molar per second" unit.
       */
    inline static Unit
    perMolarPerSecond()
      {
        std::vector<UnitFactor> f;
        f.push_back(
          UnitFactor(
            +1, UnitFactor::dimensionless, 1.0 / Constant::Avogadro));
        f.push_back(UnitFactor(+3, UnitFactor::metre,  1.0e-3));
        f.push_back(UnitFactor(-1, UnitFactor::second, 1.0e+0));
        Unit u(f);
        u.fold();
        return u;
      }


      /**
       * Return a "per second" unit.
       *
       * This function will create a "per second" unit and return it.
       *
       * @return The "per second" unit.
       */
    inline static Unit
    perSecond()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(-1, UnitFactor::second, 1.0e+0));
        Unit u(f);
        return u;
      }


      /**
       * Return a "radian" unit.
       *
       * This function will create a "radian" unit and return it.
       * Radian is a supplementary SI unit.
       *
       * @return The "metre" unit.
       */
    inline static Unit
    radian()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+1, UnitFactor::radian, 1.0e+0));
        Unit u(f);
        return u;
      }


      /**
       * Return a "second" unit.
       *
       * This function will create a "second" unit and return it.
       *
       * @return The "second" unit.
       */
    inline static Unit
    second()
      {
        std::vector<UnitFactor> f;
        f.push_back(UnitFactor(+1, UnitFactor::second, 1.0e+0));
        Unit u(f);
        return u;
      }


      /**
       * Read the unit from an SBML definition.
       *
       * This function will take all the values set in an SBML unit
       * definition and copy them to the current instance of the Unit
       * class.  This essentially means that the vector of unit
       * factors is rebuilt.  If there is data in the current instance
       * prior to calling this function, it will be lost.  Some fields
       * are optional in SBML.  For such fields the function attempts
       * to use sensible default values.
       *
       * @param ud The SBML unit definition to extract values from.
       */
    void
    fromSBML(const UnitDefinition* ud);


      /**
       * Attempts to fold a unit in order to reduce redundancies.
       *
       * The purpose is to perform certain obvious simplifications in
       * units.  For instance, "one square metre per metre", is more
       * efficiently expressed as "metre".  Such redundancies can be
       * introduced when a unit is re-expressed using only SI-units --
       * an example of this is given in the "User's Guide".  The
       * reason we need a function for redundancy reduction, is that
       * we may otherwise have several units which all really are the
       * same.
       *
       * Maybe this function should have been called "simplify"?
       */
    void
    fold();


      /**
       * Get the key of a unit.
       *
       * For units, the id is key.
       *
       * @return The key of the unit.  For units, the id is key.
       */
    inline const std::string&
    getKey() const
      {
        return getId();
      }


      /**
       * Check whether the supplied string matches the key of the
       * unit.
       *
       * Check whether the supplied string matches the key of the
       * unit.  For units, the id is key.
       *
       * @param  key The key to check against.
       * @return     True if the unit is identified by the key,
       *             false otherwise.
       */
    inline bool
    isKey(const std::string& key) const
      {
        return isId(key);
      }


      /**
       * Set the key of a unit
       *
       * For units, the id is key.
       *
       * @param key The new key of the unit.
       */
    inline void
    setKey(const std::string& key)
      {
        setId(key);
      }


      /**
       * Check whether a unit is dimensionless or not.
       *
       * This function will test to see if a unit is dimensionless.
       * If it is, the function will return true, otherwise it returns
       * false.  A unit is dimensionless if it has only one factor,
       * and this factor evaluates to "dimensionless".  Maybe units
       * with no unit factors should be regarded as dimensionless as
       * well.
       *
       * @return True if the function is dimensionless, false
       *         otherwise.
       */
    inline bool
    isDimensionless()
      {
        return (factors.size() == 1
          && factors[0].getKind() == UnitFactor::dimensionless);
      }


      /**
       * Attempt to convert the unit to another.
       *
       * Sometimes we need to convert the value of a parameter with
       * some strange unit into a value with a known unit.  In order
       * to do this, one needs the a conversion multiplier.  If the
       * conversion cannot be completed, the function will throw an
       * exception.  Note that units should be reduced and folded
       * before attempting to convert them.
       *
       * @param u The unit to convert to.
       * @return  The conversion multiplier for the conversion.
       */
    double
    toUnit(const Unit& u) const throw(InvalidConversion);


      /**
       * Addition operator for units.
       *
       * Two units can be added only if they are compatible.  If they
       * can be added, the new unit will be identical to any one of
       * the original units.  The two units must be folded in order
       * for addition to work.
       *
       * @param u The unit to add with.
       * @return  The unit after addition.
       */
    inline Unit
    operator+(const Unit& u) const throw(InvalidConversion)
      {
        try {
          toUnit(u);
        } catch (InvalidConversion&) {
          throw InvalidConversion(this, &u);
        }
        return *this;
      }


      /**
       * Subtraction operator for units.
       *
       * Two units can be subtracted only if they are compatible.  If
       * they can be subtracted, the new unit will be identical to any
       * one of the original units.  The two units must be folded in
       * order for subtraction to work.
       *
       * @param u The unit to subtract with.
       * @return  The unit after subtraction.
       */
    inline Unit
    operator-(const Unit& u) const throw(InvalidConversion)
      {
        try {
          toUnit(u);
        } catch (InvalidConversion&) {
          throw InvalidConversion(this, &u);
        }
        return *this;
      }


      /**
       * Division operator for units.
       *
       * It is always possible to divide two units.  The result of a
       * unit division is a new unit in which the factors of the two
       * original units are concatenated, but the definitions in the
       * right hand unit have their exponents negated and their
       * multipliers inverted.  Note that this function will fold the
       * resulting unit.
       *
       * @param u The unit to divide with.
       * @return  The unit after division.
       */
    Unit
    operator/(const Unit& u) const
      {
        Unit v(*this);
        for (unsigned int i = 0; i < u.factors.size(); i++) {
          UnitFactor w(
            -1 * u.factors[i].getExponent(),
            u.factors[i].getKind(),
            1.0 / u.factors[i].getMultiplier());
          v.factors.push_back(w);
        }
        v.fold();
        return v;
      }


      /**
       * Multiplication operator for units.
       *
       * It is always possible to multiply two units.  The result of a
       * unit multiplication is a new unit in which the factors of the
       * two original units are concatenated.  Note that this function
       * will fold the resulting unit.
       *
       * @param u The unit to multiply with.
       * @return  The unit after multiplication.
       */
    inline Unit
    operator*(const Unit& u) const
      {
        Unit v(*this);
        for (unsigned int i = 0; i < u.factors.size(); i++)
          v.factors.push_back(u.factors[i]);
        v.fold();
        return v;
      }


      /**
       * Power operator for units.
       *
       * This function will raise the unit to the specified integer
       * power.  For positive powers it is equivalent to multiplying
       * the unit "p" times with itself.  For negative powers, it is
       * equivalent to dividing dimensionless "p" times by the unit.
       *
       * @param p The power to raise the unit to.
       * @return  The unit after the power operation has been performed.
       */
    inline Unit
    operator^(const double p) throw(InvalidConversion)
      {
        for (size_t i = 0; i < factors.size(); i++) {
          const double exponent_double = factors[i].getExponent() * p;
          const long   exponent_long    = lrint(exponent_double);
          if(fabs(exponent_double / exponent_long - double(1.0))
	     <= double(10.0) * std::numeric_limits<double>::epsilon()) {
            factors[i].setExponent(static_cast<int>(exponent_long));
            factors[i].setMultiplier(pow(factors[i].getMultiplier(), p));
          } else if(factors[i].getExponent() == 0) {
            factors[i].setMultiplier(pow(factors[i].getMultiplier(), p));
          } else {
            throw InvalidConversion(this, NULL);
          }
        }
        Unit u(*this);
        return u;
      }


      /**
       * Equality operator for units.
       *
       * Two units are equal if and only if all their factors are
       * equal.  The two units must be folded and reduced in order for
       * this to work.  Note that two units are not necessarily equal
       * just because there "id" or "name" are equal.
       *
       * @param u The unit with which to test for equality.
       * @return  A boolean, true if the units are equal, false
       *          otherwise.
       */
    bool operator==(const Unit& u) const;


      /**
       * Inequality operator for units.
       *
       * Two units are unequal if they are not equal.  Thus we define
       * inequality in terms of equality.  The two units must be
       * folded and reduced in order for this to work.
       *
       * @param u The unit with which to test for inequality.
       * @return  A boolean, true if the units are not equal,
       *          false otherwise.
       */
    inline bool
    operator!=(const Unit& u) const
      {
        return !(*this == u);
      }


      /**
       * Put a unit to an output stream.
       *
       * This function will put a unit onto an output stream.  The
       * result should be a valid SBML "unitDefinition".  Note that
       * units are output the way MesoRD sees them, that is, as
       * products of basic SI units.
       *
       * @param os The output stream on which to write.
       * @param u  The unit in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream&
    operator<<(std::ostream& os, const Unit& u);
  };
}
#endif
