/* The UnitFactor class.
 *
 * Copyright 2004 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: UnitFactor.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <limits>
#include <math.h>

#include "UnitFactor.hpp"


  /* Just return the string corresponding to the unit.  We cannot
   * return a reference, because we do not want to return references
   * to local data.  There should not be a need to use constant string
   * variables or anything like that here: unless incorrectly spelt,
   * the names should never change.  Not including a default label in
   * the switch allows us to catch inconsistencies should the set of
   * supported SI units ever change.
   */
std::string
MesoRD::UnitFactor::getKindString() const
{
  switch(kind) {
    case ampere:
      return "ampere";

    case candela:
      return "candela";

    case dimensionless:
      return "dimensionless";

    case kelvin:
      return "kelvin";

    case kilogram:
      return "kilogram";

    case metre:
      return "metre";

    case radian:
      return "radian";

    case second:
      return "second";

    case steradian:
      return "steradian";
  }
  return "";
}


  /* The exponent and kind must match exactly.  The
   * multiplier, on the other hand, must merely be sufficiently close.
   * Both of the exponents 0 means comparision of two dimensionless, 
   * independent of kind. 
   */
bool
MesoRD::UnitFactor::operator==(const UnitFactor& uf) const
{
  if (exponent == uf.exponent && kind    == uf.kind) {
    if (fabs(multiplier / uf.multiplier - (double)1.0)
      <= 10 * std::numeric_limits<double>::epsilon()) {
      return true;
    }
  } else if (exponent == 0 && uf.exponent == 0) {
    if (fabs(multiplier / uf.multiplier - (double)1.0)
      <= 10 * std::numeric_limits<double>::epsilon()) {
      return true;
    }
  }
  
  return false;
}


  /* Output operator.  Should be a valid SBML "unit".
   */
namespace MesoRD {
  std::ostream&
  operator<< (std::ostream& os, const UnitFactor& uf)
  {
    os << "<unit kind=\""       << uf.getKindString()
       << "\""   << std::endl
       << "      exponent=\""   << uf.exponent
       << "\""   << std::endl
       << "      multiplier=\"" << uf.multiplier
       << "\""   << std::endl
       << "      scale=\"0\"/>";
    return os;
  }
}
