/* The UnitFactor class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: UnitFactor.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_UNITFACTOR_HPP
#define MESORD_UNITFACTOR_HPP

#include <iostream>
#include <string>


namespace MesoRD
{

    /**
     * The UnitFactor class.
     *
     * Units in MesoRD, like units in SBML, are arbitrary products of
     * a set of predefined units.  The set of predefined units in
     * MesoRD, however, is much smaller than the corresponding set in
     * SBML.  The reason is that SBML supports several "redundant"
     * units.  These units are redundant because they can be expressed
     * as products of SI units.  One example is the unit "Gray" which
     * is equivalent to "metre squared per second squared".
     *
     * While these redundant units make SBML easier to use, they
     * create a problem for applications.  It is, for instance,
     * possible to define a unit "Joule per Newton", which is
     * equivalent to metre.  MesoRD must understand that "Joule per
     * Newton" and "metre" are really the same thing, so that
     * quantities given in one unit can be compared to quantities
     * given in another.
     *
     * MesoRD solves the problem by expressing every unit as a product
     * of SI units.  This class implements the basic factor of a unit;
     * an instance of the UnitFactor class corresponds to a possibly
     * exponentiated, offset and scaled SI unit.  A complex unit, as
     * captured by the Unit class in MesoRD, is composed by combining
     * several instances of the UnitFactor class by multiplication or
     * division.
     *
     * Another difference between units in MesoRD and units in SBML is
     * that MesoRD uses a single floating point value to scale a unit.
     * SBML allows for scaling by arbitrary real multipliers and
     * powers of ten separately.  This is an implementation
     * difference, and should not need to be of concern to users of
     * the unit classes.
     */
  class UnitFactor
  {
  public:

      /**
       * This enumeration lists the units that MesoRD uses internally.
       * All these units are either basic or supplementary SI units.
       * Any unit encountered throughout the model will be converted
       * to a product of possibly scaled and offset factors of these
       * units.
       */
    typedef enum UnitKind {
      ampere,
      candela,
      dimensionless,
      kelvin,
      kilogram,
      metre,
      radian,
      second,
      steradian
    } UnitKind;

  private:
    int      exponent;
    UnitKind kind;
    double   multiplier;

  public:

      /**
       * Value constructor for the UnitFactor class.
       *
       * Creates a new unit factor based on the supplied values.
       *
       * @param exponent   The exponent of the unit factor.
       * @param kind       The kind of the unit factor.
       * @param multiplier The multiplier of the unit factor.
       */
    inline UnitFactor(
      int exponent, UnitKind kind, double multiplier)
      : exponent(exponent), kind(kind), multiplier(multiplier)
      {}


      /**
       * Get the exponent of the unit factor.
       *
       * Just like in SBML, units can have exponents.  These can be
       * positive or negative.  This function will return the exponent
       * for the unit factor.
       *
       * @return The exponent of the unit factor.
       */
    inline int
    getExponent() const
      {
        return exponent;
      }


      /**
       * Get the kind of the unit factor.
       *
       * Unlike an SBML unit, units in MesoRD are built out of basic
       * SI units only.  Thus, the kind returned here is a basic SI
       * unit.
       *
       * @return The kind of the unit factor.
       */
    inline UnitKind
    getKind() const
      {
        return kind;
      }


      /**
       * Get the kind of the unit factor.
       *
       * Unlike an SBML unit, units in MesoRD are built up out of
       * basic SI units only.  Thus, the kind returned here is a basic
       * SI unit.  This function will return a textual representation
       * of the kind of the unit factor.
       *
       * @return A textual representation of the unit factor.
       */
    std::string
    getKindString() const;


      /**
       * Get the multiplier of the unit factor.
       *
       * The multiplier field is used to scale units by an arbitrary
       * real-numbered factor.  It is the single multiplier of a unit
       * factor in MesoRD.  This approach is different from the one
       * chosen in SBML, which in addition to the multiplier also
       * includes a "scale", that controls multiples by powers of ten
       * separately.
       *
       * @return The multiplier of the unit factor.
       */
    inline double
    getMultiplier() const
      {
        return multiplier;
      }


      /**
       * Set the exponent of the unit factor.
       *
       * This function will set the exponent of the unit factor to the
       * supplied value.
       *
       * @param exponent The new exponent of the unit factor.
       */
    inline void
    setExponent(const int exponent)
      {
        this -> exponent = exponent;
      }


      /**
       * Set the kind of the unit factor.
       *
       * This function will set the kind of the unit factor to the
       * supplied value..
       *
       * @param kind The new kind of the unit factor.
       */
    inline void
    setKind(const UnitKind kind)
      {
        this -> kind = kind;
      }


      /**
       * Set the multiplier of the unit factor.
       *
       * This function will set the multiplier of the unit factor to
       * the supplied value.
       *
       * @param multiplier The new multiplier of the unit factor.
       */
    inline void
    setMultiplier(const double multiplier)
      {
        this -> multiplier = multiplier;
      }


      /**
       * Less-than operator for the UnitFactor class
       *
       * The less-than operator implies a certain order among unit
       * factors.  The application of this order is purely technical;
       * in real life we do not need to worry if a kilogramme is
       * greater or less than metre.  The order allows MesoRD to store
       * and sort unit factors in a list, so that for instance
       * identical units will always have their factors in the same
       * order.  Unit factors are ordered primarily after the sign of
       * the exponent: positive exponents precede negative exponents.
       * Unit factors with exponent of the same sign are ordered
       * alphabetically.
       *
       * @param uf The unit factor to compare with.
       * @return   True if the left instance is less than the right.
       */
    inline bool
    operator<(const UnitFactor& uf) const
      {
        if (exponent * uf.exponent < 0 ) {
          if( exponent > 0) {
            return true;
          }
          else {
            return false;
          }
        }
        return getKindString() < uf.getKindString();
      }


      /**
       * Greater-than operator for the UnitFactor class.
       *
       * The greater-than operator implies a certain order among unit
       * factors.  This allows MesoRD to store and sort unit factors
       * in a list, so that for instance identical units will always
       * have their factors in the same order.
       *
       * @param uf the unit factor to compare with.
       * @return   True if the left instance is greater than the right.
       */
    inline bool
    operator>(const UnitFactor& uf) const
      {
        if (exponent * uf.exponent < 0 ){
          if( exponent < 0 ) {
            return true;
          }
          else {
            return false;
          }
        }
        
        return getKindString() > uf.getKindString();
      }


      /**
       * Equality operator for unit factors.
       *
       * Two unit factors are equal if and only if the exponent and the
       * kind are equal.  Really, the multipliers
       * should be identical as well, but that is difficult to
       * implement; due to the floating point nature of the
       * multipliers, there may be truncation errors.
       *
       * @param uf The unit factor to test for equality with.
       * @return   A boolean, true if the unit factors are equal, false
       *           otherwise.
       */
    bool
    operator==(const UnitFactor& uf) const;


      /**
       * Inequality operator for unit factors.
       *
       * Two units are unequal if they are not equal.  Thus we define
       * inequality in terms of equality.
       *
       * @param uf The unit factor to test for inequality with.
       * @return   A boolean, true if the unit factors are not equal,
       *           false otherwise.
       */
    inline bool
    operator!=(const UnitFactor& uf) const
      {
        return !(*this == uf);
      }


      /**
       * Put a unit factor to an output stream.
       *
       * This function will put a unit factor onto an output stream.
       * The result should be a valid SBML "unit".  Note that units
       * are output the way MesoRD sees them, that is, as products of
       * basic SI units.
       *
       * @param os The output stream on which to write.
       * @param uf The unit factor in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream&
    operator<<(std::ostream& os, const UnitFactor& uf);
  };
}
#endif
