/* The UnitTable class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: UnitTable.hpp 558 2008-09-06 15:52:52Z davidfange $
 */
#ifndef MESORD_UNITTABLE_HPP
#define MESORD_UNITTABLE_HPP 1

#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include <list>
#include <map>
#include <string>

#include "SymbolTable.hpp"
#include "Unit.hpp"


namespace MesoRD
{

    /**
     * Class for managing several unit definitions.
     *
     * This class implements the symbol table for units.  Units, as
     * implemented in the Unit class, are keyed on their id:s.  Since
     * any units are global to the entire model, there can be only one
     * unit table.  This single unit table cannot have a parent.  Lo
     * and behold, here is the UnitTable class.
     */
  class UnitTable : public SymbolTable {
  public:

      /**
       * Default constructor for the UnitTable class.
       *
       * Makes sure the table of units contains the SBML built-in
       * units and their default values.  The SBML built-in units are
       * <code>area</code>, <code>length</code>,
       * <code>substance</code>, <code>time</code> and
       * <code>volume</code>.  Uses the default constructor of the
       * SymbolTable class.
       */
    inline UnitTable()
      : SymbolTable()
      {
        symbols["area"]      = new Unit(Unit::metre() ^ 2);
        symbols["length"]    = new Unit(Unit::metre());
        symbols["substance"] = new Unit(Unit::mole());
        symbols["time"]      = new Unit(Unit::second());
        symbols["volume"]    = new Unit(Unit::litre());
      }


      /**
       * Value constructor for the UnitTable class.
       *
       * This constructor will read the definitions of all the units
       * from the SBML model.  These units are added to the table in
       * turn.  The function uses fromSBML() internally.
       *
       * @param uds The list of SBML unit definitions to construct
       *            from.
       */
    inline UnitTable(const ListOfUnitDefinitions* uds)
      : SymbolTable()
      { 
        fromSBML(uds);
      }


      /**
       * Read all the units from an SBML definition.
       *
       * This function will take a list of SBML unit definitions,
       * convert them to MesoRD unit objects and add the latter to the
       * current unit table.  Note that all units are reduced and
       * folded prior to addition to the table.  The implementation of
       * this function contains a C-style cast.
       *
       * @param uds The list of SBML unit definitions to add to the
       *            table.
       */
      /*inline void
        fromSBML(const ListOfUnitDefinitions* uds)
        {
        for (unsigned int i = 0; i < uds -> size(); i++) {
        Unit *u = new Unit(dynamic_cast<const UnitDefinition*>(uds -> get(i)));
        u -> fold();
        symbols[u -> getKey()] = u;
        }
        }
      */
    void fromSBML(const ListOfUnitDefinitions* uds);

      /**
       * Find all possible length units defined in the table.
       *
       * This function will return a list of keys of units which
       * evaluate to units of length.  The list cannot be empty,
       * because "metre" which is a standard SBML unit is always
       * considered to be defined.  This is useful if we want to know
       * what units we can use to specify length.
       *
       * @return A list of keys of units which evaluate to length.
       */
    std::list<std::string>
    getLengthUnits() const;


      /**
       * Find the unit corresponding to the given key.
       *
       * This function will return the unit corresponding to the given
       * key.  In addition to the units defined in the model, this
       * function knows about all the standard units defined in SBML.
       * For units, the id is key.  If the unit could not be found,
       * the function will return null.
       *
       * @param key The key for the sought unit.  For units, the id is
       *            key.
       * @return    The unit corresponding to the given key.
       */
    Unit*
    getUnit(const std::string& key) const;


      /**
       * Put a unit table to an output stream.
       *
       * This function will put a unit table onto an output stream.
       * The result should be valid SBML.
       *
       * @param os The output stream on which to write.
       * @param ut The unit table in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream&
    operator<<(std::ostream& os, const UnitTable& ut);
  };
}
#endif
