/* The Visual class.
 *
 * Copyright 2004-2006 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Visual.cpp 554 2008-05-25 19:55:15Z davidfange $
 */

  /* The unistd.h header is needed on Unix for the sleep() function.
   */
#include <cmath>
#include <fstream>
#include <iomanip>
#include <sstream>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    pragma warning(disable:4244)
#  endif
#endif


#include "CubicSubVolume.hpp"
#include "OctahedricSubvolume.hpp"
#include "Visual.hpp"

#include "definition.hpp"
#include "mesord.hpp"



  /*********************
   * SWAPBUFFERS MACRO *
   *********************/

  /* Swapping the double buffered display is done differently on
   * different systems.
   */
#ifdef HAVE_X
#  define SWAPBUFFERS glXSwapBuffers(display, window)
#endif

#ifdef WIN32
#  define SWAPBUFFERS SwapBuffers(ghDC)
#endif



  /*****************************************************************
   * FUNCTION DECLARATIONS (MESSAGE LOOP CALLBACK -- WINDOWS ONLY) *
   *****************************************************************/

  /* If this function could be a friend of the Visual class, things
   * would be a little cleaner.  Trying that results in a broken, but
   * optimised, binary on Windows.
   */
#ifdef WIN32
LONG WINAPI MainWndProc(HWND, UINT, WPARAM, LPARAM);
#endif



  /***********************************************************
   * STATIC CLASSES FOR MESSAGE LOOP CALLBACK (WINDOWS ONLY) *
   ***********************************************************/

#ifdef WIN32
static MesoRD::Visual* visualStaticClass;
#endif



  /***********************
   * QUATERNION ALGEBRA *
   ***********************/

  /**
   * @brief Compute the norm of a quaternion.
   *
   * Quaternion norms are computed very much like vector norms.  This
   * particular norms is the square root of the sum of the squared
   * <code>x</code>-, <code>y</code>-, <code>z</code>- and
   * <code>w</code>-components.
   *
   * @param q The quaternion for which the norm is sought.
   * @return  The norm of the quaternion.
   */
inline double
norm(const MesoRD::Visual::Quaternion<double>& q)
{
  return sqrt(q.x * q.x + q.y * q.y + q.z * q.z + q.w * q.w);
}


  /**
   * @copydoc norm
   */
inline float
norm(const MesoRD::Visual::Quaternion<float>& q)
{
  return sqrtf(q.x * q.x + q.y * q.y + q.z * q.z + q.w * q.w);
}


  /**
   * @brief Normalise a quaternion to unit length.
   *
   * This function normalises a quaternion to unit length by dividing
   * all components by the norm of the quaternion.  Note that the
   * argument quaternion is normalised; the return value is just a
   * reference to the same, normalised quaternion.
   *
   * @param q The quaternion to normalise.
   * @return  The normalised quaternion.
   */
inline const MesoRD::Visual::Quaternion<double>&
normalise(MesoRD::Visual::Quaternion<double>& q)
{
  const double n = norm(q);
  if (fabs(n) >= std::numeric_limits<double>::min()) {
    q.w /= n;
    q.x /= n;
    q.y /= n;
    q.z /= n;
  }
  return q;
}


  /**
   * @copydoc normalise
   */
inline const MesoRD::Visual::Quaternion<float>&
normalise(MesoRD::Visual::Quaternion<float>& q)
{
  const float n = norm(q);
  if (fabsf(n) >= std::numeric_limits<float>::min()) {
    q.w /= n;
    q.x /= n;
    q.y /= n;
    q.z /= n;
  }
  return q;
}


  /**
   * @brief Convert a quaternion to OpenGL style rotation.
   *
   * Converting a quaternion to OpenGL-style rotation parameters is
   * almost trivial.  The cosine of the <code>w</code>-component
   * corresponds to half the rotation angle, in radians.  The function
   * will convert the rotation angle to degrees, to ease the use with
   * OpenGL.  The <code>x</code>-, <code>y</code>- and
   * <code>z</code>-components of the quaternion correspond to the
   * <code>x</code>-, <code>y</code>- and <code>z</code>-components of
   * the vector about which to rotate.
   *
   * @param q     The quaternion to convert.
   * @param v     The rotation axis.
   * @param angle The rotation angle, in degrees.
   */
inline void
toOpenGL(const MesoRD::Visual::Quaternion<double>& q,
  MesoRD::Visual::Coordinate<double>& v, double& angle)
{
  const double len = sqrt(q.x * q.x + q.y * q.y + q.z * q.z);

  if (fabs(len) >= std::numeric_limits<double>::min()) {
    v.x   = q.x / len;
    v.y   = q.y / len;
    v.z   = q.z / len;
    angle = double(2) * double(180) * acos(q.w) / M_PI;
  } else {
    v.x   = double(0);
    v.y   = double(0);
    v.z   = double(1);
    angle = double(0);
  }
}


  /**
   * @copydoc toOpenGL
   */
inline void
toOpenGL(const MesoRD::Visual::Quaternion<float>& q,
  MesoRD::Visual::Coordinate<float>& v, float& angle)
{
  const float len = sqrtf(q.x * q.x + q.y * q.y + q.z * q.z);

  if (fabsf(len) >= std::numeric_limits<float>::min()) {
    v.x   = q.x / len;
    v.y   = q.y / len;
    v.z   = q.z / len;
    angle = float(2) * float(180) * acosf(q.w) / float(M_PI);
  } else {
    v.x   = float(0);
    v.y   = float(0);
    v.z   = float(1);
    angle = float(0);
  }
}


  /**
   * @brief multiplication operator for quaternions.
   *
   * Performs quaternion multiplication as discussed elsewhere.  Note
   * that we would not really gain anything by implementing
   * multiplication using the <code>*=</code> operator, since the
   * computation of any element depends on <em>all</em> elements of
   * both operands.
   *
   * @param q The left multiplication operand.
   * @param r The right multiplication operand.
   * @return  The product quaternion, q * r.
   */
template<class T>
inline MesoRD::Visual::Quaternion<T>
operator*(const MesoRD::Visual::Quaternion<T>& q,
  const MesoRD::Visual::Quaternion<T>& r)
{
  MesoRD::Visual::Quaternion<T> s;

  s.w = q.w * r.w - q.x * r.x - q.y * r.y - q.z * r.z;
  s.x = q.x * r.w + q.w * r.x + q.y * r.z - q.z * r.y;
  s.y = q.y * r.w + q.w * r.y + q.z * r.x - q.x * r.z;
  s.z = q.z * r.w + q.w * r.z + q.x * r.y - q.y * r.x;

  return s;
}



  /**********************************************************
   * OPENGL CALL LIST GENERATORS & GENERAL HELPER FUNCTIONS *
   **********************************************************/

  /**
   * @brief Creates the call list for an arrowhead.
   *
   * We use OpenGL call lists for rendering all things that are not
   * trivial.  Arrowheads, used for drawing connectors, are modelled
   * by cones.  A cone in turn is nothing but a cylinder with the top
   * radius equal to zero.  The cone as drawn by the call list
   * generated here is placed along the <code>z</code>-axis, with its
   * base at <code>z</code> = 0 and has its cap at <code>z</code> = @p
   * height.
   *
   * @param height The height, or length, of the arrowhead.
   * @param radius The radius of the base of the arrowhead.
   */
namespace
{
  inline void
  createArrowhead(const GLdouble height, const GLdouble radius)
  {
    GLUquadricObj* q = gluNewQuadric();

    glNewList(MesoRD::VIEWER_LIST_ARROWHEAD, GL_COMPILE);
    gluQuadricDrawStyle(q, GLU_LINE);
    gluQuadricNormals(q, GLU_FLAT);
    gluCylinder(q, radius, 0, height,
      MesoRD::VIEWER_ARROWHEAD_SLICES, MesoRD::VIEWER_ARROWHEAD_STACKS);
    glEndList();
  }
}


  /**
   * @brief Generate all possible locations for the species in a cube.
   *
   * Given the maximal number of species that are supposed to fit
   * along a single dimension in the subvolume, this function
   * generates the coordinates for all the species that will fit.
   * Trivially, for cubic subvolumes, all points in the cube are
   * inside the subvolume.
   *
   * @param d         The maximal number of species along a single
   *                  dimension.
   * @param locations A list which holds all possible species
   *                  locations in the subvolume.
   */
namespace
{
  template<class T>
  inline void
  createCubicLocations(const size_t d,
    std::list<MesoRD::Visual::Coordinate<T> >& locations)
  {
    const float dist = float(1) / d;

    locations.clear();
    for (size_t x = 0; x < d; x++) {
      for (size_t y = 0; y < d; y++) {
        for (size_t z = 0; z < d; z++) {
          const MesoRD::Visual::Coordinate<T> p(
            0.5 * dist + dist * x - 0.5,
            0.5 * dist + dist * y - 0.5,
            0.5 * dist + dist * z - 0.5);

          locations.push_back(p);
        }
      }
    }
  }
}


  /**
   * @brief Creates the call list for a sphere.
   *
   * We use OpenGL call lists for rendering all things that are not
   * trivial.  Spheres are used for rendering molecules.  Whenever the
   * number of molecules that will fit into a subvolume is changed,
   * the radius of the call list for the sphere needs to be changed as
   * well so that the required number of molecules have the space they
   * need.
   *
   * @param radius The radius of the sphere.
   */
namespace
{
  inline void
  createSphere(const GLdouble radius)
  {
    GLUquadricObj* q = gluNewQuadric();

    glNewList(MesoRD::VIEWER_LIST_SPHERE, GL_COMPILE);
    gluQuadricDrawStyle(q, GLU_FILL);
    gluSphere(
      q, radius, MesoRD::VIEWER_SPHERE_SLICES, MesoRD::VIEWER_SPHERE_STACKS);
    glEndList();
  }
}


  /**
   * @brief Generate all possible locations for the species in a
   * truncated octahedron.
   *
   * Given the maximal number of species that are supposed to fit
   * along a single dimension in the subvolume, this function
   * generates the coordinates for all the species that will fit.  For
   * octahedric subvolumes, all points in the cube, minus those that
   * fall outside the planes |<code>x</code>| + |<code>y</code>| +
   * |<code>z</code>| = 1, are inside the subvolume.
   *
   * @param d         The maximal number of species along a single
   *                  dimension.
   * @param locations A list which holds all possible species
   *                  locations in the subvolume.
   */
namespace
{
  template<class T>
  inline void
  createOctahedricLocations(const size_t d,
    std::list<MesoRD::Visual::Coordinate<T> >& locations)
  {
    const float dist = float(1) / d;

    locations.clear();
    for (size_t x = 0; x < d; x++) {
      for (size_t y = 0; y < d; y++) {
        for (size_t z = 0; z < d; z++) {
          const MesoRD::Visual::Coordinate<T> p(
            (float)0.5 * dist + dist * x - 0.5,
            (float)0.5 * dist + dist * y - 0.5,
            (float)0.5 * dist + dist * z - 0.5);

          if (fabs(p.x) + fabs(p.y) + fabs(p.z) <= 1)
            locations.push_back(p);
        }
      }
    }
  }
}


  /**
   * @brief Generate the call list for the system bounding box.
   *
   * We use OpenGL call lists for rendering all things that are not
   * trivial.  A wire representation is used for rendering the
   * bounding box of the entire system.  Since the system geometry,
   * and therefore, the system bounding box, does not change during
   * the course of the simulation, the call list can be fixed early.
   * A box can be drawn using two loops for the opposing squares, and
   * four lines, connecting the two squares.
   *
   * @param min The minimum <code>x</code>-, <code>y</code>- and
   *            <code>z</code>-coordinates of the system geometry.
   * @param max The maximum <code>x</code>-, <code>y</code>- and
   *            <code>z</code>-coordinates of the system geometry.
   */
namespace
{
  inline void
  createWireBoundingBox(const MesoRD::Visual::Coordinate<float>& min,
    const MesoRD::Visual::Coordinate<float>& max)
  {
    glNewList(MesoRD::VIEWER_LIST_BOUNDINGBOX, GL_COMPILE);

    glBegin(GL_LINE_LOOP);
    glVertex3f(min.x, min.y, min.z);
    glVertex3f(min.x, max.y, min.z);
    glVertex3f(max.x, max.y, min.z);
    glVertex3f(max.x, min.y, min.z);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f(min.x, min.y, max.z);
    glVertex3f(min.x, max.y, max.z);
    glVertex3f(max.x, max.y, max.z);
    glVertex3f(max.x, min.y, max.z);
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(min.x, min.y, min.z);
    glVertex3f(min.x, min.y, max.z);
    glVertex3f(min.x, max.y, min.z);
    glVertex3f(min.x, max.y, max.z);
    glVertex3f(max.x, max.y, min.z);
    glVertex3f(max.x, max.y, max.z);
    glVertex3f(max.x, min.y, min.z);
    glVertex3f(max.x, min.y, max.z);
    glEnd();

    glEndList();
  }
}


  /**
   * @brief Generate the call list for a cube.
   *
   * We use OpenGL call lists for rendering all things that are not
   * trivial.  Wire representations are used for rendering the
   * boundaries of different kinds of subvolumes.  For cubic
   * subvolumes, the subvolume boundary is a cube.  Since the boundary
   * is completely independent, the call list can be fixed early.  A
   * cube can be drawn using two loops for the opposing squares, and
   * four lines, connecting the two squares.  The squares are located
   * at <code>d_s</code> from the centre of the cube.
   */
namespace
{
  inline void
  createWireCube()
  {
    const float d_s = 0.5;

    glNewList(MesoRD::VIEWER_LIST_SUBVOLUME, GL_COMPILE);

    glBegin(GL_LINE_LOOP);
    glVertex3f(-d_s, -d_s, -d_s);
    glVertex3f(-d_s, +d_s, -d_s);
    glVertex3f(+d_s, +d_s, -d_s);
    glVertex3f(+d_s, -d_s, -d_s);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f(-d_s, -d_s, +d_s);
    glVertex3f(-d_s, +d_s, +d_s);
    glVertex3f(+d_s, +d_s, +d_s);
    glVertex3f(+d_s, -d_s, +d_s);
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(-d_s, -d_s, -d_s);
    glVertex3f(-d_s, -d_s, +d_s);
    glVertex3f(-d_s, +d_s, -d_s);
    glVertex3f(-d_s, +d_s, +d_s);
    glVertex3f(+d_s, +d_s, -d_s);
    glVertex3f(+d_s, +d_s, +d_s);
    glVertex3f(+d_s, -d_s, -d_s);
    glVertex3f(+d_s, -d_s, +d_s);
    glEnd();

    glEndList();
  }
}


  /**
   * @brief Generate the call list for a truncated octahedron.
   *
   * We use OpenGL call lists for rendering all things that are not
   * trivial.  Wire representations are used for rendering the
   * boundaries of different kinds of subvolumes.  For octahedric
   * subvolumes, the subvolume boundary is a truncated octahedron.
   * Since the boundary is completely independent, the call list can
   * be fixed early.  A truncated octahedron can be drawn using six
   * loops for the three sets of opposing squares where the octahedron
   * is truncated, and twelve lines, connecting the six squares.  The
   * six squares are located at <code>d_s</code> from the centre of
   * the truncated octahedron.  The eight hexagons which result from
   * the twelve lines are located at <code>d_h</code> from the centre
   * of the truncated octahedron
   */
namespace
{
  inline void
  createWireOctahedron()
  {
    const float d_s = 0.50;
    const float d_h = 0.25;

    glNewList(MesoRD::VIEWER_LIST_SUBVOLUME, GL_COMPILE);

    glBegin(GL_LINE_LOOP);
    glVertex3f(+d_s,  0,   -d_h);
    glVertex3f(+d_s, -d_h,  0);
    glVertex3f(+d_s,  0,   +d_h);
    glVertex3f(+d_s, +d_h,  0);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f(-d_s,  0,   -d_h);
    glVertex3f(-d_s, -d_h,  0);
    glVertex3f(-d_s,  0,   +d_h);
    glVertex3f(-d_s, +d_h,  0);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f( 0,   +d_s, -d_h);
    glVertex3f(-d_h, +d_s,  0);
    glVertex3f( 0,   +d_s, +d_h);
    glVertex3f(+d_h, +d_s,  0);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f( 0,   -d_s, -d_h);
    glVertex3f(-d_h, -d_s,  0);
    glVertex3f( 0,   -d_s, +d_h);
    glVertex3f(+d_h, -d_s,  0);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f( 0,   -d_h, +d_s);
    glVertex3f(-d_h,  0,   +d_s);
    glVertex3f( 0,   +d_h, +d_s);
    glVertex3f(+d_h,  0,   +d_s);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3f( 0,   -d_h, -d_s);
    glVertex3f(-d_h,  0,   -d_s);
    glVertex3f( 0,   +d_h, -d_s);
    glVertex3f(+d_h,  0,   -d_s);
    glEnd();

    glBegin(GL_LINES);
    glVertex3f(+d_s, +d_h,  0);
    glVertex3f(+d_h, +d_s,  0);
    glVertex3f(+d_s, -d_h,  0);
    glVertex3f(+d_h, -d_s,  0);
    glVertex3f(-d_s, +d_h,  0);
    glVertex3f(-d_h, +d_s,  0);
    glVertex3f(-d_s, -d_h,  0);
    glVertex3f(-d_h, -d_s,  0);

    glVertex3f(+d_s,  0,   +d_h);
    glVertex3f(+d_h,  0,   +d_s);
    glVertex3f(+d_s,  0,   -d_h);
    glVertex3f(+d_h,  0,   -d_s);
    glVertex3f(-d_s,  0,   +d_h);
    glVertex3f(-d_h,  0,   +d_s);
    glVertex3f(-d_s,  0,   -d_h);
    glVertex3f(-d_h,  0,   -d_s);

    glVertex3f( 0,   +d_s, +d_h);
    glVertex3f( 0,   +d_h, +d_s);
    glVertex3f( 0,   +d_s, -d_h);
    glVertex3f( 0,   +d_h, -d_s);
    glVertex3f( 0,   -d_s, +d_h);
    glVertex3f( 0,   -d_h, +d_s);
    glVertex3f( 0,   -d_s, -d_h);
    glVertex3f( 0,   -d_h, -d_s);
    glEnd();

    glEndList();
  }
}


  /**
   * @brief Draws a connector between the current point and a target
   * point.
   *
   * This function will draw an arrow from the current subvolume, the
   * centre of which is in the current origin, to the target
   * subvolume, the centre of which lies at @p d with respect to the
   * current origin.  The arrow is not drawn exactly between the
   * centres, but a quarter of a subvolume side length away.  The
   * arrow has an arrowhead pointing to the target subvolume.
   *
   * Drawing the line from the current position and almost all the way
   * to @p d is not so difficult; it is done by the first two
   * glVertex3f() calls.  Because the untransformed arrowhead is drawn
   * along the <code>z</code>-axis, we must rotate it about a vector
   * that is orthogonal to the <code>z</code>-axis <em>and</em> the
   * directional vector, @p d.  This vector, called <code>o_n</code>
   * in the code below, is computed as the cross product of the
   * <code>z</code>-axis and @p d.  The rotation angle is given by the
   * dot product of the <code>z</code>-axis and @p d.
   *
   * @param d The target point, if the current point is the origin.
   */
namespace
{
  inline void
  drawConnector(const MesoRD::Visual::Coordinate<float>& d)
  {
    const float d_len = sqrtf(d.x * d.x + d.y * d.y + d.z * d.z);

    if (d_len > 0) {
      const MesoRD::Visual::Coordinate<float> d_n(
        d.x / d_len,
        d.y / d_len,
        d.z / d_len);
      const MesoRD::Visual::Coordinate<float> start(
        0.25 * d_n.x,
        0.25 * d_n.y,
        0.25 * d_n.z);
      const MesoRD::Visual::Coordinate<float> stop(
        (d_len - 0.25) * d_n.x,
        (d_len - 0.25) * d_n.y,
        (d_len - 0.25) * d_n.z);

      GLfloat     angle = 0;
      const float o_len = sqrtf(d_n.x * d_n.x + d_n.y * d_n.y);
      MesoRD::Visual::Coordinate<float> o_n(1, 0, 0);

      if (o_len > 0) {
        o_n.x = float(-1) * d_n.y / o_len;
        o_n.y = float(+1) * d_n.x / o_len;
        angle = fmodf(float(180) * acosf(d_n.z) / float(M_PI) + float(360),
		      float(360));
      } else {
        angle = d_n.z >= 0 ? float(0) : float(180);
      }

      glBegin(GL_LINES);
      glVertex3f(start.x, start.y, start.z);
      glVertex3f(stop.x,  stop.y,  stop.z);
      glEnd();

      glPushMatrix();
      glTranslatef(stop.x,  stop.y,  stop.z);
      glRotatef(angle, o_n.x, o_n.y, o_n.z);
      glCallList(MesoRD::VIEWER_LIST_ARROWHEAD);
      glPopMatrix();
    }
  }
}


  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* The constructor does all the operating system dependant stuff to
   * handle the windows.  The constructor has different arguments
   * depending on what system we are compiling on.  Note that we must
   * initialise the subvolume vector here, because C++ will not let us
   * have dangling references.  Hopefully, the default values set by
   * the constructor are suitable enough.
   */
#ifdef HAVE_X
MesoRD::Visual::Visual(
  const std::vector<SubVolume*>& v)
#endif
#ifdef WIN32
MesoRD::Visual::Visual(
  HINSTANCE hInstance, int nCmdShow, const std::vector<SubVolume*>& v)
#endif
    throw(MesoRD::Visual::InitialisationError)
  : connectors(false),
    fast(false),
    grid(true),
    molecules(true),
    moving(false),
    z_far(0),
    z_near(0),
    height(VIEWER_WINDOW_INITAL_HEIGHT),
    width(VIEWER_WINDOW_INITAL_WIDTH),
    screenshots(0),
    species_dimension(4),
    colours(v[0] -> getSpeciesNumber(), 0),
    subvolumes(v)
{
#ifdef WIN32
  //MSG      msg;
  WNDCLASS wndclass;

  visualStaticClass = this;

    /* Register the frame class.  Throw an exception if it could not
     * be done.
     */
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wndclass.hIcon         = LoadIcon(hInstance, szAppName);
  wndclass.hInstance     = hInstance;
  wndclass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
  wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
  wndclass.lpszClassName = szAppName;
  wndclass.lpszMenuName  = szAppName;
  wndclass.style         = 0;

  if (!RegisterClass(&wndclass))
    throw InitialisationError("RegisterClass()", getLastErrorString().c_str());

    /* Create the frame, or equivalently, the window.  Throw an
     * exception if it could not be done.
     */ 
  ghWnd = CreateWindow(
    szAppName,
    VIEWER_WINDOW_TITLE.c_str(),
    WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
    CW_USEDEFAULT,
    CW_USEDEFAULT,
    width,
    height,
    NULL,
    NULL,
    hInstance,
    NULL);

  if (!ghWnd)
    throw InitialisationError("CreateWindow()", getLastErrorString().c_str());
 
    /* Show and update main window.  ShowWindow() cannot really fail,
     * so no need to check for errors.
     */	
  ShowWindow(ghWnd, nCmdShow);
  if (!UpdateWindow(ghWnd))
    throw InitialisationError("UpdateWindow()", getLastErrorString().c_str());


#else
  GLXContext           ctx;
  Window               root;
  XSetWindowAttributes attr;
  XTextProperty        textProperty;
  XVisualInfo*         visinfo;

  unsigned long attrMask;
  const char*   wmName          = VIEWER_WINDOW_TITLE.c_str();
  int           visAttributes[] = { GLX_RGBA,
                                    GLX_RED_SIZE,     1,
                                    GLX_GREEN_SIZE,   1,
                                    GLX_BLUE_SIZE,    1,
                                    GLX_DOUBLEBUFFER, None };

    /* Get the target display.  Because the argument is
     * <code>NULL</code> we attempt to open the display specified in
     * the <code>DISPLAY</code> environment variable.  Throw an
     * exception if it could not be done.  Then we use the nifty macro
     * to get the root window.
     */
  if (!(display = XOpenDisplay(NULL)))
    throw InitialisationError("XOpenDisplay()", "");
  root = RootWindow(display, 0);

    /* Choose GLX visual / pixel format.  Throw an exception if it
     * could not be done.
     */
  visinfo = glXChooseVisual(display, 0, visAttributes);
  if (!visinfo)
    throw InitialisationError("glXChooseVisual", "");

    /* Create the window.  The parent window will be the root window.
     * While the window manager may position the window as it pleases,
     * we request it to be placed at the top, left corner of the
     * screen.
     */
  attr.background_pixel = 0;
  attr.border_pixel     = 0;
  attr.colormap         = XCreateColormap(
    display, root, visinfo -> visual, AllocNone);
  attr.event_mask       = ButtonMotionMask | ButtonPressMask
    | ButtonReleaseMask | ExposureMask | KeyPressMask | StructureNotifyMask;
  attrMask              = CWBackPixel | CWBorderPixel
    | CWColormap | CWEventMask;

  window = XCreateWindow(display, root, 0, 0, width, height, 0,
    visinfo -> depth, InputOutput, visinfo -> visual, attrMask, &attr);

  if (!window)
    throw InitialisationError("XCreateWindow()", "");

    /* Set the window title.  Throw an exception if it could not be
     * done.
     */
  if (XmbTextListToTextProperty(display,
        const_cast<char**>(&wmName), 1, XStringStyle, &textProperty))
    throw InitialisationError("XmbTextListToTextProperty()", "");
  XSetWMName(display, window, &textProperty);

    /* Make sure to catch delete messages from the window manager.
     */
  deleteAtom = XInternAtom(display, "WM_DELETE_WINDOW", false);
  XSetWMProtocols(display, window, &deleteAtom, 1);

    /* Display the window.  Then create the GLX rendering context.
     * Finally, bind the rendering context and the window.  Throw an
     * exception if it could not be done.
     */
  XMapWindow(display, window);
  ctx = glXCreateContext(display, visinfo, NULL, True);
  if (!ctx)
    throw InitialisationError("glXCreateContext()", "");
  glXMakeCurrent(display, window, ctx);
#endif

    /* Compute the diameter of the geometry.  We assume that the
     * geometry is constant throughout, otherwise we would have to
     * repeat this step whenever the geometry changes.  We use the
     * diameter to set the initial camera position and the location of
     * the <code>z</code>-clipping planes.
     */
  float             diameter = 2;
  Coordinate<float> max;
  Coordinate<float> min;
  std::vector<SubVolume*>::const_iterator i;

  max.x = max.y = max.z = float(-1) * std::numeric_limits<float>::max();
  min.x = min.y = min.z = float(+1) * std::numeric_limits<float>::max();

  for (i = subvolumes.begin(); i != subvolumes.end(); i++) {
    max.x = MesoRD::max(max.x, (*i) -> getX());
    min.x = MesoRD::min(min.x, (*i) -> getX());
    max.y = MesoRD::max(max.y, (*i) -> getY());
    min.y = MesoRD::min(min.y, (*i) -> getY());
    max.z = MesoRD::max(max.z, (*i) -> getZ());
    min.z = MesoRD::min(min.z, (*i) -> getZ());
  }

  createWireBoundingBox(min, max);
  diameter = MesoRD::max(diameter,
    MesoRD::max(fabsf(max.x), MesoRD::max(fabsf(min.x),
                  MesoRD::max(fabsf(max.y), MesoRD::max(fabsf(min.y),
                                MesoRD::max(fabsf(max.z), fabsf(min.z)))))));

  z_near   = float(+2);
  z_far    = float(+1) * VIEWER_INITIAL_SLAB_FACTOR * diameter;
  camera.z = float(+1) * VIEWER_INITIAL_ZOOM_FACTOR * diameter;

    /* Set some OpenGL parameters and the scene lighting.  The
     * location of the single light source is fixed at
     * <code>light</code>.  Flat shading should be faster than smooth
     * shading.  We use white as clear colour here.  Then set up the
     * initial view port and projection matrix using reshape().
     */
  glClearColor(1, 1, 1, 0);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0, GL_POSITION, light);
  glShadeModel(GL_FLAT);

  reshape(width, height);

    /* Now create the call lists and the locations of species in a
     * subvolume.  The arrowhead is static in the sense that is does
     * not depend on what kind of subvolume is used, nor the number of
     * species in a subvolume.
     */
  createArrowhead(0.2, 0.1);
  setSpeciesDimension(species_dimension);
}


  /* The destructor is quite possibly the simplest function of this
   * class.  It just deletes the list of colours that may have been
   * allocated during the lifetime of the instance.  In the case of
   * the X Window System, we also close the display.
   */
MesoRD::Visual::~Visual()
{
  while (!colours.empty()) {
    if (colours.back())
      delete colours.back();
    colours.pop_back();
  }

    /* Close the display.
     */
#ifdef HAVE_X
  XCloseDisplay(display);
#endif
}



  /***********************************
   * DISPLAY & INTERACTION (GENERAL) *
   ***********************************/

  /* Process the event queue or message queue in a way that is highly
   * dependant on what operating system we are on.  Note that we do a
   * sleep on Unixes here, to prevent the viewer from consuming too
   * much time, and starve any other threads we may have running on
   * the system.  On Windows we need to call the window procedure to
   * do the actual processing.
   */
int
MesoRD::Visual::animateStep()
{
#ifdef HAVE_X
  unsigned char      buffer[20];
  KeySym             keySym;
  State              state;
  XComposeStatus     composeStatus;
  XEvent             event;
  const unsigned int sleep_time = 20;

  while (XPending(display) > 0) {
    XNextEvent(display, &event);

    switch (event.type) {
      case ButtonPress:
        state.x     = event.xbutton.x;
        state.y     = event.xbutton.y;
        state.flags = State::button | State::rotation
          | (event.xmotion.state & ShiftMask     ? State::slab        : 0)
          | (event.xmotion.state & ControlMask   ? State::translation : 0)
          | (event.xmotion.state & ShiftMask     ? State::zoom        : 0);
        motion(state);
        moving = true;
        break;

      case ButtonRelease:
        state.x     = 0;
        state.y     = 0;
        state.flags = 0;
        motion(state);
        moving = false;
        break;

      case ClientMessage:
        break;

      case ConfigureNotify:
        if (static_cast<signed int>(height) != event.xconfigure.height
          || static_cast<signed int>(width) != event.xconfigure.width)
          reshape(event.xconfigure.width, event.xconfigure.height);
        break;

      case Expose:
        break;

      case KeyPress:
        if (XLookupString(reinterpret_cast<XKeyEvent*>(&event),
              reinterpret_cast<char*>(buffer), 1, &keySym, &composeStatus) > 0)
          keyboard(buffer[0]);
        break;

      case MotionNotify:
        if (moving) {
          state.x     = event.xbutton.x;
          state.y     = event.xbutton.y;
          state.flags = State::button | State::rotation
            | (event.xmotion.state & ShiftMask     ? State::slab        : 0)
            | (event.xmotion.state & ControlMask   ? State::translation : 0)
            | (event.xmotion.state & ShiftMask     ? State::zoom        : 0);
          motion(state);
        }
        break;
    }
  }

  refresh();
  glXWaitGL();
  usleep(sleep_time);

  return 0;
#endif
#ifdef WIN32
  MSG msg;

  while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) == TRUE) {
    if (GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    } else {
      return TRUE;
    }
  }

  refresh();
  return TRUE;
#endif
}


  /* The main window procedure, only used on Windows.  Generally, we
   * should always make sure we always return sensible values.  Due to
   * some weird optimisation issues, this function cannot be made a
   * friend of the Visual class.  The window procedure is actually
   * very similar to the way the Unix function works.  Because this
   * function is not a member of the Visual class, we need to use the
   * statically defined Visual instance whenever we call functions
   * from the viewer class.  This also means that this function will
   * be utterly confused if there are several instances of the Visual
   * class.  Also note that we do not really want to throw exceptions
   * here, because it is not entirely clear where to catch them.
   * Instead we rely on the Windows MessageBox() call to report errors
   * to users.
   */ 
#ifdef WIN32
LONG WINAPI
MainWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  int                   pixelformat;
  LONG                  lRet = 1;
  PAINTSTRUCT           ps;
  PIXELFORMATDESCRIPTOR pfd;
  RECT                  rect;
  MesoRD::Visual::State                 state;

  switch (uMsg) {
    case WM_CHAR:
      visualStaticClass -> keyboard(wParam);
      break;

    case WM_CLOSE:
      break;

    case WM_CREATE:
      visualStaticClass -> ghDC = GetDC(hWnd);
      pfd.nSize                 = sizeof(PIXELFORMATDESCRIPTOR);
      pfd.nVersion              = 1;
      pfd.dwFlags               =
        PFD_DOUBLEBUFFER | PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
      pfd.dwLayerMask           = PFD_MAIN_PLANE;
      pfd.iPixelType            = PFD_TYPE_COLORINDEX;
      pfd.cColorBits            = 8;
      pfd.cDepthBits            = 16;
      pfd.cAccumBits            = 0;
      pfd.cStencilBits          = 0;

      if (!(pixelformat = ChoosePixelFormat(
              visualStaticClass -> ghDC, &pfd))) {
        MessageBox(
          NULL, "ChoosePixelFormat()", getLastErrorString().c_str(), MB_OK);
        PostQuitMessage(0);
        return 1;
      }

      if (SetPixelFormat(
            visualStaticClass -> ghDC, pixelformat, &pfd) == FALSE) {
        MessageBox(
          NULL, "SetPixelFormat()", getLastErrorString().c_str(), MB_OK);
        PostQuitMessage(0);
        return 1;
      }

      visualStaticClass -> ghRC = wglCreateContext(visualStaticClass -> ghDC);
      wglMakeCurrent(visualStaticClass -> ghDC, visualStaticClass -> ghRC);

      if (!GetClientRect(hWnd, &rect)) {
        MessageBox(
          NULL, "GetClientRect()", getLastErrorString().c_str(), MB_OK);
        PostQuitMessage(0);
        return 1;
      }

      visualStaticClass -> reshape(rect.right, rect.bottom);
      break;

    case WM_DESTROY:
      if (visualStaticClass -> ghRC)
        wglDeleteContext(visualStaticClass -> ghRC);
      if (visualStaticClass -> ghDC)
        ReleaseDC(hWnd, visualStaticClass -> ghDC);
      PostQuitMessage(0);
      break; 

        /* We should return zero after processing this message.
         */
    case WM_LBUTTONUP:
      state.x     = 0;
      state.y     = 0;
      state.flags = 0;
      visualStaticClass -> motion(state);
      visualStaticClass  -> setMoving(false);
      lRet = 0;
      break;

        /* We should return zero after processing this message.
         */
    case WM_LBUTTONDOWN:
      state.x     = GET_X_LPARAM(lParam);
      state.y     = GET_Y_LPARAM(lParam);
      state.flags = (true       ? MesoRD::Visual::State::button      : 0)
        | (true                 ? MesoRD::Visual::State::rotation    : 0)
        | (wParam & MK_SHIFT    ? MesoRD::Visual::State::slab        : 0)
        | (wParam & MK_CONTROL  ? MesoRD::Visual::State::translation : 0)
        | (wParam & MK_SHIFT    ? MesoRD::Visual::State::zoom        : 0);
      visualStaticClass -> motion(state);
      visualStaticClass -> setMoving(true);
      lRet = 0;
      break;

        /* We should return zero after processing this message.
         */
    case WM_MOUSEMOVE:
      if (wParam & MK_LBUTTON) {
        state.x     = GET_X_LPARAM(lParam);
        state.y     = GET_Y_LPARAM(lParam);
        state.flags = (true       ? MesoRD::Visual::State::button      : 0)
          | (true                 ? MesoRD::Visual::State::rotation    : 0)
          | (wParam & MK_SHIFT    ? MesoRD::Visual::State::slab        : 0)
          | (wParam & MK_CONTROL  ? MesoRD::Visual::State::translation : 0)
          | (wParam & MK_SHIFT    ? MesoRD::Visual::State::zoom        : 0);
        visualStaticClass -> motion(state);
      }
      lRet = 0;
      break;

    case WM_PAINT:
      BeginPaint(hWnd, &ps);
      EndPaint(hWnd, &ps);
      break;

    case WM_SIZE:
      GetClientRect(hWnd, &rect);
      visualStaticClass -> reshape(rect.right, rect.bottom);
      break;

    default:
      lRet = DefWindowProc(hWnd, uMsg, wParam, lParam);
      break;
  }

  return lRet;
}
#endif



  /***********************************
   * DISPLAY & INTERACTION (GENERAL) *
   ***********************************/

  /* Process a key press event.  The keyboard is used to toggle
   * display of connectors, grid and species.
   */
void
MesoRD::Visual::keyboard(const unsigned char key)
{
  std::ostringstream filename;

  switch (key) {
    case 'c':
    case 'C':
      connectors = !connectors;
      break;

    case 'f':
    case 'F':
      fast = !fast;
      break;

    case 'g':
    case 'G':
      grid = !grid;
      break;

    case 'm':
    case 'M':
      molecules = !molecules;
      break;

    case 's':
    case 'S':
      filename << "mesord_" << ++screenshots << ".pov";
      dumpPOVRay(filename.str());
      break;

    default:
      break;
  }
}


  /* This function tracks mouse movement.  We only ever need to do
   * anything if the mouse button is pressed, and has been pressed
   * during the last mouse movement.  If that is the case, the mouse
   * pointer has been "dragged".  If the mouse pointer has been
   * dragged, we will perform one or two of the rotation, slab,
   * translation or zoom operations.  Exactly which we will do depends
   * on the state of some qualifiers, such as the control and shift
   * keys.  Note that it really only makes sense to slab and zoom at
   * the same time.
   */
void
MesoRD::Visual::motion(const State& state)
{
  static bool previous_button = false;
  static int  previous_x      = -1;
  static int  previous_y      = -1;

  if (state.flags && state.flags & State::button && previous_button) {
    const int delta_x = state.x - previous_x;
    const int delta_y = state.y - previous_y;

    if (state.flags & State::translation) {

        /* If translation is active the camera is simply translated in
         * the <code>x</code><code>y</code>-plane.
         */
      camera.x -= VIEWER_MOUSE_FACTOR_X * delta_x;
      camera.y += VIEWER_MOUSE_FACTOR_Y * delta_y;

    } else if (state.flags & (State::slab | State::zoom)) {

        /* If slabbing is active, we are changing the
         * <code>z</code>-clipping planes.  However, we must never
         * allow the back clipping plane to lie in front of the front
         * clipping plane.  Changing the clipping planes means our
         * viewing frustum changes, so we need to call reshape() to
         * bring about an update.
         */
      if (state.flags & State::slab) {
        z_far = MesoRD::max(z_near, z_far - VIEWER_MOUSE_FACTOR_Z * delta_x);
        reshape(width, height);
      }

        /* If zooming is active, we are moving the camera along the
         * <code>z</code>-axis.  Note that the camera will not move
         * beyond the <code>z</code>-clipping planes.
         */
      if (state.flags & State::zoom) {
        camera.z = MesoRD::max(z_near, MesoRD::min(z_far,
                                 camera.z - VIEWER_MOUSE_FACTOR_Z * delta_y));
      }

    } else if (state.flags & State::rotation) {

        /* If rotation is active, we compute the angular displacement
         * and cumulatively update the orientation of the camera.
         * Camera orientation is stored as a quaternion; updating it
         * simply requires a quaternion postmultiplication.
         */
      const float angle_x = fmodf(
        VIEWER_MOUSE_FACTOR_Y * delta_y + float(360), 360);
      const float angle_y = fmodf(
        VIEWER_MOUSE_FACTOR_X * delta_x + float(360), 360);

      orientation = Quaternion<float>(angle_x, angle_y, float(0))
        * orientation;
    }
  }

  previous_button = state.flags & State::button;
  previous_x      = state.x;
  previous_y      = state.y;
}


  /* Draw the entire system on the display using OpenGL's model view
   * matrix.  First, clear the display, while making sure to honour
   * double buffering.  Then set up the camera and render the whole
   * system.  In terms of visualisation, the system consists of lots
   * of subvolumes which may have a grid, connectors and species.  The
   * grid is a visual representation of the boundary of a subvolume.
   * The connectors are arrows that point to the centres of
   * neighbouring subvolumes.  The species are supposed to give an
   * idea of the chemical composition of the subvolume.  The rendering
   * of all of these three components is optional.
   */
void
MesoRD::Visual::refresh() const
{
  float             angle;
  Coordinate<float> axis;
  std::vector<SubVolume*>::const_iterator i;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

    /* Set the orientation and location of the camera.
     */
  toOpenGL(orientation, axis, angle);
  glTranslatef(-camera.x, -camera.y, -camera.z);
  glRotatef(angle, axis.x, axis.y, axis.z);

  if (fast && moving) {
    GLfloat colour_black[] = {0, 0, 0, 0};
    glMaterialfv(GL_FRONT, GL_DIFFUSE, colour_black);
    glCallList(VIEWER_LIST_BOUNDINGBOX);

  } else {

      /* Big loop for all the subvolumes.
       */
    for (i = subvolumes.begin(); i != subvolumes.end(); i++) {
      const SubVolume* sv = *i;

        /* Make sure the current position is the centre of the
         * subvolume.
         */
      glPushMatrix();
      glTranslatef(sv -> getX(), sv -> getY(), sv -> getZ());

        /* Draw the grid if we are asked to.
         */
      if (grid) {
        GLfloat colour_black[] = {0, 0, 0, 0};
        glMaterialfv(GL_FRONT, GL_DIFFUSE, colour_black);
        glCallList(VIEWER_LIST_SUBVOLUME);
      }

        /* Draw the connectors, if we are asked to.  This is not very
         * efficient, but the connectors are really only intended for
         * geometry debugging.
         */
      if (connectors) {
        const GLfloat colour_red[] = {1, 0, 0, 0};

        glMaterialfv(GL_FRONT, GL_DIFFUSE, colour_red);

        for (size_t j = 0; j < sv -> getNeighboursNumber(); j++) {
          const SubVolume*           n = sv -> getNeighbour(j);
          const Coordinate<float> d(
            n -> getX() - sv -> getX(),
            n -> getY() - sv -> getY(),
            n -> getZ() - sv -> getZ());
          drawConnector(d);
        }
      }

        /* Draw the molecules inside the subvolume.  The
         * <code>colour</code> variable gives the index into the
         * <code>colours</code> table for the colour that will be used
         * for the species.  The <code>colour</code> variable is
         * <em>also</em> the index of the species we are currently
         * drawing.  If <code>colour</code> is out of range, or the
         * colour table at <code>colour</code> is <code>null</code>,
         * the species will not be rendered.  At any given time,
         * <code>count</code> keeps track of how the maximum number of
         * molecules of <em>every</em> species that have been rendered
         * into the subvolume.  We keep track of this, because there
         * could be cases where there are millions of molecules of one
         * species but only a few of a another.  If we were to draw
         * molecules into the subvolume one species at the time, the
         * subvolume would appear to be filled with molecules of only
         * the most abundant species.  If we instead cycle through the
         * species, adding one molecule of every species at the time,
         * we make sure that the subvolume is rendered so that the
         * less prominent species appear as well.
         */
      if (molecules) {
        unsigned int colour = 0;
        unsigned int count  = 0;
        std::list<Coordinate<float> >::const_iterator i;

        for (i = locations.begin(); i != locations.end(); i++) {
          const Coordinate<float>& p     = *i;
          bool                     drawn = false;

          for (size_t j = 0; !drawn && j < colours.size(); j++) {
            if (colour < colours.size() && colours[colour]
              && (SubVolume::getUseDoubleInSpecies()
                && count < sv -> getSpeciesDouble(colour) - 1
                || !SubVolume::getUseDoubleInSpecies()
                && count < sv -> getSpecies(colour))) {

              glMaterialfv(GL_FRONT, GL_DIFFUSE, colours[colour]);
              glPushMatrix();
              glTranslatef(p.x, p.y, p.z);
              glCallList(VIEWER_LIST_SPHERE);
              glPopMatrix();

              drawn = true;
            }

            if ((colour = ++colour
		 % static_cast<unsigned int>(sv -> getSpeciesNumber())) == 0)
              count++;
          }

          if (!drawn)
            break;
        }
      }

      glPopMatrix();
    }
  }

    /* Flush and swap the buffers when we are done.  Life would be
     * really jerky if it was not for double buffering.
     */
  glFlush();
  SWAPBUFFERS;
}


  /* Dump the entire system to a POV-Ray file.  This closely mirrors
   * what is done in the refresh() function.  Basically, first the
   * camera is set up, and a couple of macros are defined to make the
   * actual scene definition easier.  The radius calculation of the
   * species is duplicated from the setSpeciesDimension() function.
   * The scene of the system consists of lots of subvolumes which may
   * have a grid, connectors and species.  The grid is a visual
   * representation of the boundary of a subvolume.  The connectors
   * are arrows that point to the centres of neighbouring subvolumes.
   * The species are supposed to give an idea of the chemical
   * composition of the subvolume.  The rendering of all of these
   * three components is optional.
   */
void
MesoRD::Visual::dumpPOVRay(const std::string filename) const
{
  float             angle           = 0;
  Coordinate<float> axis            = Coordinate<float>(0, 0, 0);
  const float       angle_view      = float(2) * atanf(float(2) / z_near)
    * float(180) / float(M_PI);
  const float       radius          = float(1)
    / (float(2) * species_dimension);
  const int         float_precision = 6;
  const int         float_width     = float_precision + 7;


    /* Open file for output and check for success.  Return on error
     * since there is nothing we can do.  Set the precision for
     * floating point numbers on output.
     */
  std::ofstream povray(filename.c_str(), std::ios::trunc);
  if (!povray) {
    std::cout << "Could not open '"<< filename
              << "' for writing."  << std::endl;
    return;
  }
  
  povray.precision(float_precision);


    /* Preamble.  Personally, I find the sample command line really
     * useful -- I always forget how to run <code>povray</code>.
     */
  povray << "// ***********************" << std::endl
         << "// * Generated by MesoRD *" << std::endl
         << "// ***********************" << std::endl
         << "// Run e.g."                << std::endl
         << "//   povray " << filename
         << " +H"          << height
         << " +W"          << width
         << " +FT +A0.1 +D +P -V"        << std::endl
         << "// to generate image."      << std::endl
         << std::endl                    << std::endl;


    /* Camera settings, background and lighting.  The camera is always
     * looking towards the negative <code>z</code>-axis and the
     * background is always white.  Light sources are not handled very
     * well yet -- their use in OpenGL has to be fixed as well!
     */
  povray << "// ************************************"       << std::endl
         << "// * CAMERA SETTINGS, BACKGROUND, ETC *"       << std::endl
         << "// ************************************"       << std::endl
         << "#declare HEIGHT       = " << height     << ";" << std::endl
         << "#declare WIDTH        = " << width      << ";" << std::endl
         << "#declare ASPECT_RATIO = WIDTH / HEIGHT" << ";" << std::endl
         << std::endl
         << "camera"                                        << std::endl
         << "{"                                             << std::endl
         << "  location  < 0,            0,  0>"            << std::endl
         << "  look_at   < 0,            0, -1>"            << std::endl
         << "  direction < 0,            0, -1>"            << std::endl
         << "  right     < ASPECT_RATIO, 0,  0>"            << std::endl
         << "  angle     "
         << std::setw(float_width) << angle_view            << std::endl
         << "}"                                             << std::endl
         << std::endl
         << "background"                                    << std::endl
         << "{"                                             << std::endl
         << "  color rgb <1.0, 1.0, 1.0>"                   << std::endl
         << "}"                                             << std::endl
         << std::endl
         << "light_source"                                  << std::endl
         << "{"                                             << std::endl
         << "  <0, 0, 0>"                                   << std::endl
         << "  rgb <1.0, 1.0, 1.0>"                         << std::endl
         << "}"                                             << std::endl
         << std::endl                                       << std::endl;


    /* Common macros.  These set the general appearance which is
     * applied to all objects.
     */
  povray << "// *******************************"  << std::endl
         << "// * MACROS COMMON TO ALL SHAPES *"  << std::endl
         << "// *******************************"  << std::endl
         << "#default {"                          << std::endl
         << "  finish {"                          << std::endl
         << "    ambient   0.2"                   << std::endl
         << "    diffuse   0.6"                   << std::endl
         << "    specular  1.0"                   << std::endl
         << "    roughness 0.001"                 << std::endl
         << "    metallic"                        << std::endl
         << "  }"                                 << std::endl
         << "}"                                   << std::endl
         << std::endl                             << std::endl;


    /* Set the orientation and location of the camera.  Just like in
     * the OpenGL part, every object is rotated and translated before
     * it is emitted.
     */
  toOpenGL(orientation, axis, angle);
  povray << "// **********************************"           << std::endl
         << "// * ROTATION AND TRANSLATION MACRO *"           << std::endl
         << "// **********************************"           << std::endl
         << "#macro rotate_translate()"                       << std::endl
         << "  #local ANGLE  = "
         << std::setw(float_width) << float(+1) * angle
         << ";"                                               << std::endl
         << "  #local AXIS   = <"
         << std::setw(float_width) << float(+1) * axis.x
         << ", "
         << std::setw(float_width) << float(+1) * axis.y
         << ", "
         << std::setw(float_width) << float(+1) * axis.z
         << ">;"                                              << std::endl
         << "  #local DISP   = <"
         << std::setw(float_width) << float(-1) * camera.x
         << ", "
         << std::setw(float_width) << float(-1) * camera.y
         << ", "
         << std::setw(float_width) << float(-1) * camera.z
         << ">;"                                              << std::endl
         << "  #local VAXISX = vaxis_rotate(x, AXIS, ANGLE);" << std::endl
         << "  #local VAXISY = vaxis_rotate(y, AXIS, ANGLE);" << std::endl
         << "  #local VAXISZ = vaxis_rotate(z, AXIS, ANGLE);" << std::endl
         << "  transform {"                                   << std::endl
         << "    matrix <VAXISX.x, VAXISX.y, VAXISX.z,"       << std::endl
         << "            VAXISY.x, VAXISY.y, VAXISY.z,"       << std::endl
         << "            VAXISZ.x, VAXISZ.y, VAXISZ.z,"       << std::endl
         << "            DISP.x,   DISP.y,   DISP.z>"         << std::endl
         << "  }"                                             << std::endl
         << "  clipped_by"                                    << std::endl
         << "  {"                                             << std::endl
         << "    plane"                                       << std::endl
         << "    {"                                           << std::endl
         << "      <0, 0, +1>, "
         << std::setw(float_width) << -1 * z_near             << std::endl
         << "    }"                                           << std::endl
         << "    plane"                                       << std::endl
         << "    {"                                           << std::endl
         << "      <0, 0, -1>, "
         << std::setw(float_width) << +1 * z_far              << std::endl
         << "    }"                                           << std::endl
         << "  }"                                             << std::endl
         << "#end"                                            << std::endl
         << std::endl                                         << std::endl;


    /* This part is only needed if connectors are to be drawn.
     * Connectors are always rendered in red.
     */
  if (connectors) {
    povray << "#macro connector(p, q)"        << std::endl
           << "  #local HEIGHT = 0.2;"        << std::endl
           << "  #local RADIUS = 0.1;"        << std::endl
           << "  #local DIR    = q - p;"      << std::endl
           << "  #local LEN    = sqrt("
           << "DIR.x * DIR.x + "
           << "DIR.y * DIR.y + "
           << "DIR.z * DIR.z);"               << std::endl
           << "  #local BASE   = p + "
           << "((LEN - HEIGHT) / LEN) * DIR;" << std::endl
           << "  union"                       << std::endl
           << "  {"                           << std::endl
           << "    cone"                      << std::endl
           << "    {"                         << std::endl
           << "      BASE, RADIUS, q, 0"      << std::endl
           << "    }"                         << std::endl
           << "    wire(p, BASE)"             << std::endl
           << "    pigment"                   << std::endl
           << "    {"                         << std::endl
           << "      rgb <1.0, 0.0, 0.0>"     << std::endl
           << "    }"                         << std::endl
           << "    rotate_translate()"        << std::endl
           << "  }"                           << std::endl
           << "#end"                          << std::endl
           << std::endl;
  }


    /* A macro for the species.  These are just spheres with fixed
     * radii and variable pigmentation.  Macros are only emitted for
     * those species which should also be rendered.
     */
  for (size_t i = 0; i < colours.size(); i++) {
    if (colours[i]) {
      povray << "// ***********"                                << std::endl
             << "// * SPECIES *"                                << std::endl
             << "// ***********"                                << std::endl
             << "#macro species_" << i << "(c)"                 << std::endl
             << "  #local RADIUS = " << radius << ";"           << std::endl
             << "  sphere"                                      << std::endl
             << "  {"                                           << std::endl
             << "    <c.x, c.y, c.z>,"                          << std::endl
             << "    RADIUS"                                    << std::endl
             << "    pigment"                                   << std::endl
             << "    {"                                         << std::endl
             << "      rgb <"
             << std::setw(float_width) << colours[i][0] << ", "
             << std::setw(float_width) << colours[i][1] << ", "
             << std::setw(float_width) << colours[i][2] << ">"  << std::endl
             << "    }"                                         << std::endl
             << "    rotate_translate()"                        << std::endl
             << "  }"                                           << std::endl
             << "#end"                                          << std::endl;
    }
  }


    /* Subvolumes are always rendered in black.  The macro determines
     * what kind of subvolumes will be used.  Currently, this is
     * either cubes or octahedra.
     */
  povray << std::endl                    << std::endl
         << "// ***********************" << std::endl
         << "// * SUBVOLUME RENDERING *" << std::endl
         << "// ***********************" << std::endl
         << "#macro wire(p, q)"          << std::endl
         << "  #local RADIUS = 0.02;"    << std::endl
         << "  union"                    << std::endl
         << "  {"                        << std::endl
         << "    cylinder"               << std::endl
         << "    {"                      << std::endl
         << "      <p.x, p.y, p.z>"      << std::endl
         << "      <q.x, q.y, q.z>"      << std::endl
         << "      RADIUS"               << std::endl
         << "    }"                      << std::endl
         << "    sphere"                 << std::endl
         << "    {"                      << std::endl
         << "      <p.x, p.y, p.z>,"     << std::endl
         << "      RADIUS"               << std::endl
         << "    }"                      << std::endl
         << "    sphere"                 << std::endl
         << "    {"                      << std::endl
         << "      <q.x, q.y, q.z>,"     << std::endl
         << "      RADIUS"               << std::endl
         << "    }"                      << std::endl
         << "  }"                        << std::endl
         << "#end"                       << std::endl
         << std::endl;

  povray << "#macro subvolume(c)" << std::endl;
  if (dynamic_cast<const CubicSubVolume*>(*(subvolumes.begin()))) {
    povray << "  #local SIDE = 0.5;" << std::endl
           << "  union"              << std::endl
           << "  {"                  << std::endl;

    for (size_t i1 = 0; i1 < 2; i1++) {
      for (size_t j1 = 0; j1 < 2; j1++) {
        for (size_t k1 = 0; k1 < 2; k1++) {
          for (size_t i2 = i1; i2 < 2; i2++) {
            for (size_t j2 = j1; j2 < 2; j2++) {
              for (size_t k2 = k1; k2 < 2; k2++) {
                if ((i2 - i1) + (j2 - j1) + (k2 - k1) == 1) {
                  povray << "    wire("
                         << "<c.x " << (i1 % 2 == 0 ? '-' : '+') << " SIDE,"
                         << " c.y " << (j1 % 2 == 0 ? '-' : '+') << " SIDE,"
                         << " c.z " << (k1 % 2 == 0 ? '-' : '+') << " SIDE>, "
                         << std::endl
                         << "         "
                         << "<c.x " << (i2 % 2 == 0 ? '-' : '+') << " SIDE,"
                         << " c.y " << (j2 % 2 == 0 ? '-' : '+') << " SIDE,"
                         << " c.z " << (k2 % 2 == 0 ? '-' : '+') << " SIDE>)"
                         << std::endl;
                }
              }
            }
          }
        }
      }
    }

  } else if (dynamic_cast<const OctahedricSubvolume*>(*(subvolumes.begin()))) {
    povray  << "  #local HEIGHT = 0.25;" << std::endl
            << "  #local SIDE   = 0.5;"  << std::endl
            << "  union" << std::endl
            << "  {"     << std::endl

            << "    wire("
            << "<c.x + SIDE,   c.y + 0.0,    c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x + SIDE,   c.y - HEIGHT, c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x + SIDE,   c.y - HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + SIDE,   c.y + 0.0,    c.z + HEIGHT>)"  << std::endl
            << "    wire("
            << "<c.x + SIDE,   c.y + 0.0,    c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + SIDE,   c.y + HEIGHT, c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x + SIDE,   c.y + HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + SIDE,   c.y + 0.0,    c.z - HEIGHT>)"  << std::endl

            << "    wire("
            << "<c.x - SIDE,   c.y + 0.0,    c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x - SIDE,   c.y - HEIGHT, c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y - HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x - SIDE,   c.y - 0.0,    c.z + HEIGHT>)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y + 0.0,    c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x - SIDE,   c.y + HEIGHT, c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y + HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x - SIDE,   c.y + 0.0,    c.z - HEIGHT>)"  << std::endl

            << "    wire("
            << "<c.x + 0.0,    c.y + SIDE,   c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - HEIGHT, c.y + SIDE,   c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + SIDE,   c.z + HEIGHT>)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y + SIDE,   c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x + HEIGHT, c.y + SIDE,   c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + SIDE,   c.z - HEIGHT>)"  << std::endl

            << "    wire("
            << "<c.x + 0.0,    c.y - SIDE,   c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y - SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - HEIGHT, c.y - SIDE,   c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - SIDE,   c.z + HEIGHT>)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y - SIDE,   c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y - SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x + HEIGHT, c.y - SIDE,   c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - SIDE,   c.z - HEIGHT>)"  << std::endl

            << "    wire("
            << "<c.x + 0.0,    c.y - HEIGHT, c.z + SIDE  >, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + 0.0,    c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x - HEIGHT, c.y + 0.0,    c.z + SIDE  >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + HEIGHT, c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y + HEIGHT, c.z + SIDE  >, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + 0.0,    c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + HEIGHT, c.y + 0.0,    c.z + SIDE  >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - HEIGHT, c.z + SIDE  >)"  << std::endl

            << "    wire("
            << "<c.x + 0.0,    c.y - HEIGHT, c.z - SIDE  >, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + 0.0,    c.z - SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x - HEIGHT, c.y + 0.0,    c.z - SIDE  >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + HEIGHT, c.z - SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y + HEIGHT, c.z - SIDE  >, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + 0.0,    c.z - SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + HEIGHT, c.y + 0.0,    c.z - SIDE  >, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - HEIGHT, c.z - SIDE  >)"  << std::endl

            << "    wire("
            << "<c.x + SIDE,   c.y + HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x + SIDE,   c.y - HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y - SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y + HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + SIDE,   c.z + 0.0   >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y - HEIGHT, c.z + 0.0   >, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y - SIDE,   c.z + 0.0   >)"  << std::endl

            << "    wire("
            << "<c.x + SIDE,   c.y + 0.0,    c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + 0.0,    c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + SIDE,   c.y + 0.0,    c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x + HEIGHT, c.y + 0.0,    c.z - SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y + 0.0,    c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + 0.0,    c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x - SIDE,   c.y + 0.0,    c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x - HEIGHT, c.y + 0.0,    c.z - SIDE  >)"  << std::endl

            << "    wire("
            << "<c.x + 0.0,    c.y + SIDE,   c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + HEIGHT, c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y + SIDE,   c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y + HEIGHT, c.z - SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y - SIDE,   c.z + HEIGHT>, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - HEIGHT, c.z + SIDE  >)"  << std::endl
            << "    wire("
            << "<c.x + 0.0,    c.y - SIDE,   c.z - HEIGHT>, " << std::endl
            << "         "
            << "<c.x + 0.0,    c.y - HEIGHT, c.z - SIDE  >)"  << std::endl;
  }

  povray << "    pigment"                 << std::endl
         << "    {"                       << std::endl
         << "      rgb <0, 0, 0>"         << std::endl
         << "    }"                       << std::endl
         << "    rotate_translate()"      << std::endl
         << "  }"                         << std::endl
         << "#end"                        << std::endl;


      /* Big loop for all the subvolumes.  The centre of the current
       * subvolume is stored in the <code>centre</code> coordinate.
       */
  if (subvolumes.size() > 0) {
    povray << std::endl                  << std::endl
           << "// *********************" << std::endl
           << "// * COMPLETE GEOMETRY *" << std::endl
           << "// *********************" << std::endl;

    std::vector<SubVolume*>::const_iterator i;
    for (i = subvolumes.begin(); i != subvolumes.end(); i++) {
      const SubVolume* sv = *i;
      const Coordinate<float> centre(sv -> getX(), sv -> getY(), sv -> getZ());

        /* Draw the grid if we are asked to.  This just calls the
         * subvolume macro.
         */
      if (grid) {
        povray << "subvolume(<"
               << std::setw(float_width) << centre.x << ", "
               << std::setw(float_width) << centre.y << ", "
               << std::setw(float_width) << centre.z << ">)" << std::endl;
      }

        /* Draw the connectors, if we are asked to.  This is not very
         * efficient, but the connectors are really only intended for
         * geometry debugging.
         */
      if (connectors) {
        for (size_t j = 0; j < sv -> getNeighboursNumber(); j++) {
          const SubVolume* n = sv -> getNeighbour(j);
          if (n) {
            const Coordinate<float> neighbour(
              n -> getX(), n -> getY(), n -> getZ());

            if(  fabsf(centre.x - neighbour.x)
                 >= std::numeric_limits<float>::epsilon()
              || fabsf(centre.y - neighbour.y)
                 >= std::numeric_limits<float>::epsilon()
              || fabsf(centre.z - neighbour.z)
                 >= std::numeric_limits<float>::epsilon()) {
              povray << "connector(<"
                     << std::setw(float_width) << centre.x    << ", "
                     << std::setw(float_width) << centre.y    << ", "
                     << std::setw(float_width) << centre.z    << ">, <"
                     << std::setw(float_width) << neighbour.x << ", "
                     << std::setw(float_width) << neighbour.y << ", "
                     << std::setw(float_width) << neighbour.z << ">)"
                     << std::endl;
            }
          }
        }
      }

        /* Draw the molecules inside the subvolume.  The
         * <code>colour</code> variable gives the index into the
         * <code>colours</code> table for the colour that will be used
         * for the species.  The <code>colour</code> variable is
         * <em>also</em> the index of the species we are currently
         * drawing.  This is done just like in the refresh() function.
         */
      if (molecules) {
        unsigned int colour = 0;
        unsigned int count  = 0;
        std::list<Coordinate<float> >::const_iterator i;

        for (i = locations.begin(); i != locations.end(); i++) {
          const Coordinate<float>& p     = *i;
          bool                     drawn = false;

          for (size_t j = 0; !drawn && j < colours.size(); j++) {
            if (colour < colours.size() && colours[colour]) {
              if(SubVolume::getUseDoubleInSpecies()) {
                if(count < sv -> getSpeciesDouble(colour) - 1) {
                  
                  povray << "species_" << colour << "(<"
                         << std::setw(float_width) << centre.x + p.x << ", "
                         << std::setw(float_width) << centre.y + p.y << ", "
                         << std::setw(float_width) << centre.z + p.z << ">)"
                         << std::endl;
                  
                  drawn = true;
                }
              }
              else {
                if(count < sv -> getSpecies(colour)) {
                  
                  povray << "species_" << colour << "(<"
                         << std::setw(float_width) << centre.x + p.x << ", "
                         << std::setw(float_width) << centre.y + p.y << ", "
                         << std::setw(float_width) << centre.z + p.z << ">)"
                         << std::endl;
                  
                  drawn = true;
                }
              }
            }
            

            if ((colour = ++colour
		 % static_cast<unsigned int>(sv -> getSpeciesNumber())) == 0)
              count++;
          }

          if (!drawn)
            break;
        }
      }
      povray << std::endl;
    }
  }

  povray.close();
}


  /* Update the aspect ratio.  This is required whenever the display
   * dimension changes.  Here, updating the aspect ratio means
   * changing the view port and the viewing frustum.  The change is
   * done to the projection matrix.  In fact, this is the only
   * function that should ever touch the projection matrix.  It is
   * safe to call this function multiple times, since the effects are
   * not cumulative.
   */
void
MesoRD::Visual::reshape(const unsigned int width, const unsigned int height)
{
  const float aspect = float(1) * height / width;

  this -> width  = width;
  this -> height = height;

  glViewport(0, 0,
    static_cast<GLsizei>(width),
    static_cast<GLsizei>(height));

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  glFrustum(
    static_cast<GLdouble>(-2),
    static_cast<GLdouble>(+2),
    static_cast<GLdouble>(-2 * aspect),
    static_cast<GLdouble>(+2 * aspect),
    static_cast<GLdouble>(z_near),
    static_cast<GLdouble>(z_far));
}



  /*********************
   * CONTROL FUNCTIONS *
   *********************/

  /* Note that we use <code>null</code> to indicate that a particular
   * species should not be rendered.  The refresh() function should
   * check this, and not render any species who's colour is
   * <code>null</code>.
   */
void
MesoRD::Visual::setColours(const std::vector<VisualColour*>& c)
{
  std::vector<VisualColour*>::const_iterator i;

  while (!colours.empty()) {
    if (colours.back())
      delete colours.back();
    colours.pop_back();
  }

  for (i = c.begin(); i != c.end(); i++) {
    if ((*i) -> red >= 0 && (*i) -> green >= 0 && (*i) -> blue >= 0) {
      GLfloat* colour = new GLfloat[4];

      colour[0] = (*i) -> red;
      colour[1] = (*i) -> green;
      colour[2] = (*i) -> blue;
      colour[3] = 1;

      colours.push_back(colour);
    } else {
      colours.push_back(0);
    }
  }
}


  /* Whenever the number of species that will be drawn into a
   * subvolume changes, the size and the location of the species need
   * to be updated to reflect the change.
   */
void
MesoRD::Visual::setSpeciesDimension(const size_t d)
{
  const double radius = double(1) / (double(2) * d);

  species_dimension = d;
  createSphere(radius);

  if (dynamic_cast<const CubicSubVolume*>(*(subvolumes.begin()))) {
    createCubicLocations(species_dimension, locations);
    createWireCube();
  } else if (dynamic_cast<const OctahedricSubvolume*>(*(subvolumes.begin()))) {
    createOctahedricLocations(species_dimension, locations);
    createWireOctahedron();
  }
}


void
MesoRD::Visual::setMoving(const bool moving)
{ this -> moving = moving;}
