/* The Visual class.
 *
 * Copyright 2004-2006 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Visual.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef MESORD_VISUAL_H
#define MESORD_VISUAL_H 1

#include "mesord.hpp"


  /* For some strange reason, we seem to need to include the valarray
   * header before windows.h on Windows systems.  Also note that the
   * windows.h in turn must be included before the GL headers.
   */
#ifdef HAVE_GL
#  include <GL/gl.h>
#endif
#ifdef HAVE_GLU
#  include <GL/glu.h>
#endif
#ifdef HAVE_X
#  include <GL/glx.h>
#  include <X11/Xlib.h>
#endif
#ifdef WIN32
#  define NOMINMAX
#  include <valarray>
#  include <windows.h>
#  include <windowsx.h>
#  include <GL/gl.h>
#  include <GL/glu.h>
#endif

#include <stdexcept>
#include <list>
#include <vector>

#include "definition.hpp"
#include "SubVolume.hpp"


namespace MesoRD {

    /**
     * @brief The Visual class renders a system of subvolumes using
     * OpenGL.
     *
     * The Visual class deals with the rendering of subvolumes.  We
     * use OpenGL for all the fancy 3D graphics, so the implementation
     * is a lot nicer than it would have been if we had tried to roll
     * our own graphics engine.  In principle, all one has to do is to
     * feed the Visual class a vector of all the subvolumes, and
     * <code>press play on tape</code>.  A limitation that has to do
     * with OpenGL originally being intended for use with C and not
     * C++ is that there should only ever be one instance of the
     * Visual class.
     *
     * Throughout this class, we assume that the geometry is constant.
     * Probably we assume a lot of other, undocumented, things as
     * well.  Also note that this class is incredibly sensitive to the
     * icc optimiser.  Please make sure the class works on Windows
     * when it is modified, even if the change seems trivial!
     */
  class Visual
  {
  public:

      /**
       * @brief The templated Quaternion struct.
       *
       * The Visual class contains a minimal quaternion
       * implementation.  Quaternions make the implementation of
       * intuitive camera orientation much simpler.
       */
    template<class T> struct Quaternion
    {

        /**
         * @brief The <code>w</code>-component of the quaternion.
         */
      T w;

        /**
         * @brief The <code>x</code>-component of the quaternion.
         */
      T x;

        /**
         * @brief The <code>y</code>-component of the quaternion.
         */
      T y;

        /**
         * @brief The <code>z</code>-component of the quaternion.
         */
      T z;

        /**
         * @brief Default Quaternion constructor.
         *
         * The default constructor initialises the quaternion to the
         * unit quaternion.  The unit quaternion is one in the
         * <code>w</code>-direction and zero in the <code>x</code>-,
         * <code>y</code>- and <code>z</code>-directions.
         */
      inline Quaternion()
        : w(T(1)), x(T(0)), y(T(0)), z(T(0))
        {}


        /**
         * @brief Construct a quaternion from Euler angles.
         *
         * This constructor will construct the quaternion equivalent
         * to first yawing around the <code>z</code>-axis by
         * <code>phi</code> degrees, followed by pitching around the
         * <code>y</code>-axis by <code>theta</code> degrees and
         * finally rolling around the <code>x</code>-axis by
         * <code>psi</code> degrees.  Apparently, this is the way it
         * is done in aeronautics.
         *
         * @param psi   Euler roll angle around the
         *              <code>x</code>-axis.
         * @param theta Euler pitch angle around the
         *              <code>y</code>-axis.
         * @param phi   Euler yaw angle around the
         *              <code>z</code>-axis.
         */
      inline Quaternion(const double psi, const double theta, const double phi)
        {
          const double psi_half   = double(M_PI) * psi   / double(2 * 180);
          const double theta_half = double(M_PI) * theta / double(2 * 180);
          const double phi_half   = double(M_PI) * phi   / double(2 * 180);

          const double c_p = cos(theta_half);
          const double c_r = cos(psi_half);
          const double c_y = cos(phi_half);

          const double s_p = sin(theta_half);
          const double s_r = sin(psi_half);
          const double s_y = sin(phi_half);

          w = static_cast<T>(c_r * c_p * c_y + s_r * s_p * s_y);
          x = static_cast<T>(s_r * c_p * c_y - c_r * s_p * s_y);
          y = static_cast<T>(c_r * s_p * c_y + s_r * c_p * s_y);
          z = static_cast<T>(c_r * c_p * s_y - s_r * s_p * c_y);
        }


        /**
         * @copydoc Quaternion
         */
      inline Quaternion(const float psi, const float theta, const float phi)
        {
          const float psi_half   = float(M_PI) * psi   / float(2 * 180);
          const float theta_half = float(M_PI) * theta / float(2 * 180);
          const float phi_half   = float(M_PI) * phi   / float(2 * 180);

          const float c_p = cosf(theta_half);
          const float c_r = cosf(psi_half);
          const float c_y = cosf(phi_half);

          const float s_p = sinf(theta_half);
          const float s_r = sinf(psi_half);
          const float s_y = sinf(phi_half);

          w = c_r * c_p * c_y + s_r * s_p * s_y;
          x = s_r * c_p * c_y - c_r * s_p * s_y;
          y = c_r * s_p * c_y + s_r * c_p * s_y;
          z = c_r * c_p * s_y - s_r * s_p * c_y;
        }



        /**
         * @brief Value constructor for quaternions.
         *
         * There is also a templated value constructor for the
         * Quaternion struct itself.  This constructor allows for
         * conversion of quaternions.
         *
         * @param w The <code>w</code>-value of the new quaternion.
         * @param x The <code>x</code>-value of the new quaternion.
         * @param y The <code>y</code>-value of the new quaternion.
         * @param z The <code>z</code>-value of the new quaternion.
         */
      inline Quaternion(const T& w, const T& x, const T& y, const T& z)
        : w(w), x(x), y(y), z(z)
        {}
    };


      /**
       * @brief The templated Coordinate struct.
       *
       * This struct defines a point in a Cartesian three-dimensional
       * space.  This is useful because sometimes, using a vector or
       * an array type class, is massive overkill.
       */
    template<class T> struct Coordinate
    {

        /**
         * @brief The Cartesian <code>x</code>-component of the point
         * in three-dimensional space.
         */
      T x;

        /**
         * @brief The Cartesian <code>y</code>-component of the point
         * in three-dimensional space.
         */
      T y;

        /**
         * @brief The Cartesian <code>z</code>-component of the point
         * in three-dimensional space.
         */
      T z;

        /**
         * @brief Default Coordinate constructor.
         *
         * The default constructor initialises the elements to the
         * default value of the template type.  This is accomplished
         * by using the default constructor of the template type.
         */
      Coordinate()
        : x(T()), y(T()), z(T())
        {}


        /**
         * @brief Value constructor for Coordinate struct.
         *
         * There is also a templated copy constructor for the
         * Coordinate struct itself.  This constructor allows for
         * conversion of coordinates.
         *
         * @param x The <code>x</code>-component of the coordinate.
         * @param y The <code>y</code>-component of the coordinate.
         * @param z The <code>z</code>-component of the coordinate.
         */
      Coordinate(const T& x, const T& y, const T& z)
        : x(x), y(y), z(z)
        {}
    };


      /**
       * @brief The State struct encodes the user's current state.
       *
       * The state struct contains low-level information, such as the
       * position of the mouse, as well as high-level, viewer-specific
       * data, such as whether the user expects the viewer to rotate,
       * translate, zoom, etc.  The struct is passed on to those
       * functions that need to respond to user input.
       */
    typedef struct State 
    {

        /**
         * @brief The flags defined below give additional information
         * about what action the user is expecting.
         *
         * The State struct contains a flag field, which is a
         * bit-field composed of the flags defined below.
         */
      typedef enum Flags {

          /**
           * @brief True if the mouse button is pressed.
           */
        button = 0x01,

          /**
           * @brief True if rotation is active.
           */
        rotation = 0x02,

          /**
           * @brief True if <code>z</code>-clipping, or slabbing, is
           * active.
           */
        slab = 0x04,

          /**
           * @brief True if translation is active.
           */
        translation = 0x08,

          /**
           * @brief True if zoom is active.
           */
        zoom = 0x10
      } Flags;

        /**
         * @brief The <code>x</code>-coordinate of the mouse pointer.
         */
      int x;

        /**
         * @brief The <code>y</code>-coordinate of the mouse pointer.
         */
      int y;

        /**
         * @brief The flags are a boolean field giving additional
         * information.
         *
         * The flags field should be a result of boolean
         * <code>OR</code> of the values defined in the Flags typedef.
         */
      int flags;
    } State;


  private:
    bool   connectors;
    bool   fast;
    bool   grid;
    bool   molecules;
    bool   moving;
    float  z_far;
    float  z_near;

    /**
     * @brief The height of the window.
     *
     * This is an <code>unsigned int</code>, and not a
     * <code>size_t</code>, because that is what XCreateWindow()
     * expects.
     */
    unsigned int height;

    /**
     * @brief The width of the window.
     *
     * This is an <code>unsigned int</code>, and not a
     * <code>size_t</code>, because that is what XCreateWindow()
     * expects.
     */
    unsigned int width;
    size_t screenshots;
    size_t species_dimension;

    Coordinate<float>                camera;
    Quaternion<float>                orientation;
    std::list<Coordinate<float> >    locations;
    std::vector<GLfloat*>            colours;
    const std::vector<SubVolume*>&   subvolumes;

#ifdef HAVE_X
    Atom     deleteAtom;
    Display* display;
    Window   window;
#endif
  public:
#ifdef WIN32
    HWND  ghWnd;
    HDC   ghDC;
    HGLRC ghRC;
#endif


      /**
       * @brief Class for reporting errors during initialisation.
       *
       * If something goes wrong during initialisation, and there are
       * quite a few things that can, this exception will be thrown.
       * The exception allows the catcher to see what call failed,
       * and, in some cases, obtain a diagnostic message left behind
       * by the failing call.  Note that on X windowing systems, the
       * diagnostic is generally not set.  The reason is that MesoRD
       * relies on the built-in error handler of the X windows
       * implementation.
       */
    class InitialisationError : public std::runtime_error
    {
    private:
      const char* call;
      const char* diagnostic;

    public:

        /**
         * @brief The InitialisationError constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param call       The call that failed.
         * @param diagnostic A possible diagnostic message which
         *                   attempts to explain the failure.
         */
      inline InitialisationError(const char* call, const char* diagnostic)
        : std::runtime_error("InitialisationError"),
        call(call), diagnostic(diagnostic)
        {}
      

        /**
         * @brief Get the name of the call that failed.
         *
         * This function returns a pointer to the name of the call
         * that failed.  This may be the <code>null</code> pointer
         *
         * @return A pointer to the name of the failing call.
         */
      inline const char*
        getCall() const
        {
          return call;
        }



        /**
         * @brief Get the diagnostic message of the failed call.
         *
         * This function returns a pointer to the diagnostic message,
         * possibly explaining why the call failed.  This may be the
         * <code>null</code> pointer.  On X windowing systems, it
         * generally will be the <code>null</code> pointer.
         *
         * @return A pointer to the diagnostic message of the failed
         *         call.
         */
      inline const char*
        getDiagnostic() const
        {
          return diagnostic;
        }

    };


      /**
       * @brief The one and only constructor for the Visual class.
       *
       * This class must be constructed with a reference to a vector
       * containing all the subvolumes.  If this vector has zero
       * length, then construction will fail.  Therefore, one
       * <em>must</em> ensure that the subvolume vector is non-empty
       * prior to instantiating this class.  The constructor is a
       * nasty piece of work, because it is completely different on
       * Windows compared to the Unixes.  Note that the required
       * arguments to this constructor vary somewhat on different
       * windowing systems.  I have to admit that while this class
       * suffers from extreme optimisation problems on Windows when
       * using icc, it is actually simpler on Windows.
       *
       * @param v A vector of all the subvolumes to render.
       */
#ifdef HAVE_X
    Visual(const std::vector<SubVolume*>& v) throw(InitialisationError);
#endif
#ifdef WIN32
    Visual(
      HINSTANCE hInstance, int nCmdShow, const std::vector<SubVolume*>& v)
      throw(InitialisationError);
#endif


      /**
       * @brief The destructor for the Visual class.
       *
       * The destructor frees any resources allocated during the
       * lifetime of the Visual class.  In the case of the X windowing
       * system, it also closes the display.
       */
    ~Visual();


      /**
       * @brief Process all pending events and update the display as
       * appropriate.
       *
       * This function should be called repeatedly in order to animate
       * the display and act responsively to user interaction.  Just
       * like the constructor, it is quite system-dependent.  The
       * basic idea, however, is the same on all supported systems:
       * process mouse movement, mouse button presses and releases,
       * key presses and so on, and make sure to call the appropriate
       * functions to handle these events.  Then render the system,
       * according to whatever input the user provided
       *
       * @return On Unixes, the function always returns zero.  On
       *         Windows the function always returns <code>TRUE</code>.
       */
    int
    animateStep();


      /**
       * @brief Dump the display to a POV-Ray file.
       *
       * This function works just like the refresh() function, expect
       * it does not output to the display, but to a named file.  It
       * will render the vector of subvolumes, including the grid,
       * connectors, molecules and all.
       *
       * @bug There are still unresolved issues with the placement of
       * the light sources.
       *
       * @param filename The name of the file to write.
       */
    void
    dumpPOVRay(const std::string filename) const;


      /**
       * @brief Process a key press.
       *
       * This function will take whatever action is required when a
       * key is pressed.  This involves updating some of the member
       * attributes of the class.
       *
       * @param key The character corresponding to the pressed key.
       */
    void
    keyboard(const unsigned char key);


      /**
       * @brief Process mouse motion.
       *
       * This function will take whatever action is required when the
       * mouse is moved.  In particular, if the mouse is pressed, that
       * is, when the mouse pointer is <i>dragged</i>, this function
       * will make sure to call any appropriate functions to rotate
       * the coordinate system, adjust the <code>z</code>-clipping
       * planes (slabbing), move the camera back and forth (zooming)
       * or displace the camera up, down, left or right (translation).
       * Which of these will happen, depends on the qualifier
       * arguments passed along.  In fact, mouse movement can be
       * completely ignored unless the mouse is dragged, even though
       * one must call this function when the mouse pointer is
       * released so that it knows dragging has stopped.
       *
       * <ul>
       *
       * <li>When rotation is active, the coordinate system of the
       * entire geometry will be rotated around the <code>x</code>-
       * and <code>y</code>-axes with mouse movement.</li>
       *
       * <li>When slabbing is active, moving the mouse left and right
       * will move the back clipping plane backwards and forwards.
       * The front clipping plane is never moved.</li>
       *
       * <li>When translation is active, the camera is translated
       * along the <code>x</code>- and <code>y</code>-axes with mouse
       * movement.</li>
       *
       * <li>When zooming is active, moving the mouse up and down will
       * translate the camera backwards and forwards.  The camera is
       * never moved in front of the front clipping plane.</li>
       *
       * </ul>
       *
       * This function does keep state: it remembers the old position
       * of the mouse whenever it is called.  Therefore, the motion()
       * function <em>will</em> cause problems if the Visual class is
       * ever split onto multiple, simultaneous threads.  Luckily,
       * MesoRD uses a single, separate thread for the viewer.
       *
       * @param state The user's current state.
       */
    void
    motion(const State& state);


      /**
       * @brief Refresh the display.
       *
       * This is probably the most interesting function in this class.
       * Thanks to the people at SGI who gave us OpenGL, this function
       * has no system-dependant code at all.  Basically it will
       * render the vector of subvolumes, including the grid,
       * connectors, molecules and all.  Even though the entire viewer
       * is "just for show", we care a bit about performance here.  In
       * this case, caring about performance means not throwing
       * exceptions.
       */
    void
    refresh() const;


      /**
       * @brief Reshape the display to new dimensions.
       *
       * When the size of the display changes, we need to make sure we
       * update the aspect ratio as well.  If we were to change the
       * window from being square to being tall without updating the
       * aspect ratio, the stuff in the display would become fat.
       *
       * @param width  The new width, in pixels, of the display.
       * @param height The new height, in pixels, of the display.
       */
    void
    reshape(const unsigned int width, const unsigned int height);


      /**
       * @brief Set the colours of species.
       *
       * Species, when rendered in subvolumes, are differentiated by
       * colours.  The colour of each species is set using this
       * function.  The input vector should hold the colour in a
       * VisualColour struct as described elsewhere.  The colour
       * vector must be organised so that its indices match the
       * indices of the species, i.e. the <code>n</code>:th element in
       * the colour vector gives the colour of the <code>n</code>:th
       * species in the species vector.  If a species should not be
       * rendered, on must set at least one of its red, green or blue
       * intensities to a negative value.
       *
       * @param c The colour vector holding the colours of each
       *          species.
       */
    void
    setColours(const std::vector<VisualColour*>& c);


      /**
       * @brief Set the number of molecules to render along one
       * dimension of a subvolume.
       *
       * This function controls how many molecules will fit into each
       * subvolume.  It may be a little counter-intuitive, because the
       * actual number of molecules that will be rendered is something
       * like the third power of the number set here, the exact number
       * depending on the particular kind of subvolume.  The number is
       * best understood as <i>the maximal number of molecules that
       * will fit into a subvolume along any of the <code>x</code>-,
       * <code>y</code>- or <code>z</code>-axes</i>.
       *
       * @param d The number of molecules along one dimension of the
       *          subvolume.
       */
    void
    setSpeciesDimension(const size_t d);

      /**
       * @brief Sets the moving flag, (WIN32 specific)
       *
       * In the windows implementation of the visualizer it uses a
       * callback function and static realisation. Since visual is *
       * not properly inherited a public setter is needed.
       * 
       * @param moving The moving flag
       *
       */
    void
      setMoving(const bool moving);
  };
}

#endif
