/* The CSG class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSG.cpp 555 2008-06-08 18:07:05Z davidfange $
 */
#include <algorithm>
#include <cmath>
#include <limits>

#include "CSG.hpp"
#include "CSGNode.hpp"

#include "../Annotation.hpp"
#include "../CubicSubVolume.hpp"
#include "../OctahedricSubvolume.hpp"
#include "../UnitTable.hpp"

#include "../linalg/Vector.hpp"
#include "../linalg/Matrix.hpp"



  /************************
   * THE TRIPLET TEMPLATE *
   ************************/

namespace
{

    /**
     * @brief The triplet template for handling general three-tuples.
     *
     * The triplet template is a trivial extension of Stroustrup's
     * <code>pair</code> template given in section 17.4.1.2 of "The
     * C++ Programming Language".  The first, second and third
     * elements of any triplet are referred independently of their
     * actual types.  The interface is inspired by g++'s
     * <code>pair</code> implementation, which in turn is a
     * modification of code written by the STL people at SGI, who seem
     * to have borrowed a few lines from HP.
     *
     * @todo This should probably be moved to its own file.
     */
  template<class T1, class T2, class T3> struct triplet
  {

      /**
       * @brief first_type is the first bound type.
       */
    typedef T1 first_type;

      /**
       * @brief second_type is the second bound type.
       */
    typedef T2 second_type;

      /**
       * @brief third_type is the third bound type.
       */
    typedef T3 third_type;

      /**
       *@brief first is a copy of the first object.
       */
    T1 first;

      /**
       * @brief second is a copy of the second object.
       */
    T2 second;

      /**
       * @brief third is a copy of the third object.
       */
    T3 third;


      /**
       * @brief Default constructor for the triplet struct.
       *
       * The default constructor initialises the elements to the
       * default value of their respective types.  This is
       * accomplished by using the default constructors of the
       * respective types.
       *
       * @note Commented out, as it is not currently used.
       */
/*
    triplet()
      : first(T1()), second(T2()), third(T3())
      {}
*/

      /**
       * @brief Value constructor for the triplet struct.
       *
       * Copies the three objects supplied as arguments into the
       * struct.
       *
       * @param a The <code>first</code> value.
       * @param b The <code>second</code> value.
       * @param c The <code>third</code> value.
       */
    triplet(const T1& a, const T2& b, const T3& c)
      : first(a), second(b), third(c)
      {}


      /**
       * @brief Templated copy constructor for the triplet struct.
       *
       * Because the template arguments of the parameter struct are
       * different, this constructor allows for conversion of
       * triplets.  Note that this requires that conversion from
       * <code>U1</code> to <code>T1</code>, <code>U2</code> to
       * <code>T2</code> and <code>U3</code> to <code>T3</code> is
       * defined.
       *
       * @param t The triplet to copy from.
       */
    template<class U1, class U2, class U3>
    triplet(const triplet<U1, U2, U3>& t)
      : first(t.first), second(t.second), third(t.third)
      {}
  };


    /**
     * @brief Equality operator for triplets.
     *
     * Two triplets of the same type are equal iff their members are
     * equal.  We do not handle triplets that are not of the same type
     * here.
     *
     * @param a The left operand for the equality operation.
     * @param b The right operand for the equality operation.
     * @return  True if @p a == @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator==(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return a.first == b.first && a.second == b.second && a.third == b.third;
  }


    /**
     * @brief Less-than operator for triplets.
     *
     * When evaluating less-than for two triplets of the same type,
     * the <code>first</code> has precedence over <code>second</code>
     * and <code>third</code>.  The <code>second</code> has precedence
     * over the <code>third</code>.
     *
     * @param a The left operand for the less-than operation.
     * @param b The right operand for the less-than operation.
     * @return  True if @p a < @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator<(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return a.first < b.first
      || (!(b.first < a.first) && a.second < b.second)
      || (!(b.first < a.first || b.second < a.second) && a.third < b.third);
  }


    /**
     * @brief Inequality operator for triplets.
     *
     * The inequality operator for two triplets of the same type is
     * defined in terms of the equality operator.
     *
     * @param a The left operand for the equality operation.
     * @param b The right operand for the equality operation.
     * @return  True if @p a != @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator!=(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return !(a == b);
  }


    /**
     * @brief Greater-than operator for triplets.
     *
     * The greater-than operator for two triplets of the same type is
     * defined in terms of the less-than operator.
     *
     * @param a The left operand for the greater-than operation.
     * @param b The right operand for the greater-than operation.
     * @return  True if @p a > @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator>(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return b < a;
  }


    /**
     * @brief Less-or-equal-to operator for triplets.
     *
     * The less-or-equal-to operator for two triplets of the same type
     * is defined in terms of the less-than operator.
     *
     * @param a The left operand for the less-or-equal-to operation.
     * @param b The right operand for the less-or-equal-to operation.
     * @return  True if @p a <= @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator<=(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return !(b < a);
  }


    /**
     * @brief Greater-or-equal-to operator for triplets.
     *
     * The greater-or-equal-to operator for two triplets of the same
     * type is defined in terms of the less-than operator.
     *
     * @param a The left operand for the greater-or-equal-to
     *          operation.
     * @param b The right operand for the greater-or-equal-to
     *          operation.
     * @return  True if @p a >= @p b, false otherwise.
     */
  template<class T1, class T2, class T3>
  inline bool
  operator>=(const triplet<T1, T2, T3>& a, const triplet<T1, T2, T3>& b)
  {
    return !(a < b);
  }


    /**
     * @brief A convenience wrapper for creating a triplet from three
     * objects.
     *
     * This function is included for completeness, since the STL
     * implementation used by g++ has one just like it.  The effect is
     * the same as if one were to use the value constructor of the
     * triplet struct.
     *
     * @param a The <code>first</code> object.
     * @param b The <code>second</code> object.
     * @param c The <code>third</code> object.
     * @return  A newly-constructed triplet object of the appropriate
     *          type.
     */
  template<class T1, class T2, class T3>
  inline triplet<T1, T2, T3>
  make_triplet(const T1& a, const T2& b, const T3& c)
  {
    return triplet<T1, T2, T3>(a, b, c);
  }
}



  /****************************
   * MODEL BUILDING TEMPLATES *
   ****************************/

  /* The find templates return a pointer to the subvolume found in @p
   * locations at the supplied coordinate @p coord.  If @p locations
   * does not have an entry for @p coord, the functions return the
   * fallback value, @p fallback.  As an educational feature, one
   * might try to remove the ampersand from the @p locations argument
   * in the argument list.  The resulting performance decrease
   * illustrates the impact references can have.
   *
   * @param coord     The coordinate of the subvolume, either encoded
   *                  in a triplet, or in a vector.
   * @param locations An object which maps locations in space to
   *                  pointers to subvolumes.
   * @param fallback  The pointer that will be returned if @p
   *                  locations does not contain anything at @p coord.
   */
namespace
{
  template<class T>
  inline T*
  find(const triplet<int, int, int>&                            coord,
    const std::map<triplet<int, int, int>, MesoRD::SubVolume*>& locations,
    T*                                                          fallback)
  {
    std::map<triplet<int, int, int>, MesoRD::SubVolume*>::const_iterator i;
    if ((i = locations.find(coord)) != locations.end())
      return static_cast<T*>(i -> second);
    return fallback;
  }

  template<class T>
  inline T*
  find(
    const MesoRD::Vector&                                       coord,
    const std::map<triplet<int, int, int>, MesoRD::SubVolume*>& locations,
    T*                                                          fallback)
  {
    const triplet<int, int, int> t(
      static_cast<int>(rint(coord(0))),
      static_cast<int>(rint(coord(1))),
      static_cast<int>(rint(coord(2))));

    std::map<triplet<int, int, int>, MesoRD::SubVolume*>::const_iterator i;

    if ((i = locations.find(t)) != locations.end())
      return static_cast<T*>(i -> second);
    return fallback;
  }
}


  /* The grid templates return the grid numbers for each supported
   * kind of subvolume.  The <i>grid</i>, and its <i>grid numbers</i>
   * is perhaps a bit of a misnomer -- rather the number is to be
   * interpreted as follows: if the centres of all the subvolumes are
   * projected onto one of coordinate axes of the global coordinate
   * system, the grid number for that axis is the number of projected
   * points per subvolume size in the chosen axis direction.
   * Examples:
   *
   * <ul>
   *
   *   <li>Consider cubic subvolumes of size one; the centre-to-centre
   *   distance of subvolumes along the <code>x</code>-axis is one.
   *   If * the origin is a valid centre location for a subvolume, the
   *   * <code>x</code>-coordinate of the centre of <em>any</em>
   *   subvolume * will be an integer multiple of one.  Thus the grid
   *   number in the * <code>x</code>-axis for cubic subvolumes is
   *   one.</li>
   *
   *   <li>Consider octahedric subvolumes of size one; the
   *   centre-to-centre distance of subvolumes along the *
   *   <code>x</code>-axis is again one.  Again assuming that the
   *   origin * is a valid centre location for a subvolume, the *
   *   <code>x</code>-coordinate of the centre of an arbitrary
   *   subvolume * will be <em>either</em> of the form <code>n</code>
   *   <em>or</em> * <code>n</code> + 0.5, where <code>n</code> is an
   *   integer.  For * every interval of length one along the
   *   <code>x</code>-axis, there * are <em>two</em> possible
   *   <code>x</code>-coordinates for * subvolume centres.  Thus the
   *   grid number in the * <code>x</code>-axis for octahedric
   *   subvolumes is two.</li>
   *
   * </ul>
   *
   * Come to think of it, this is not entirely unrelated to some
   * concepts encountered in crystallography (but then again, the
   * entire concept of "tessellation in space" is probably a distant
   * cousin to crystallography).
   *
   * @return A triplet giving the grid number in the <code>x</code>,
   *         <code>y</code> and <code>z</code>-direction.
   */
namespace
{
  template<class T>
  inline triplet<int, int, int>
  grid()
  {
    return triplet<int, int, int>(0, 0, 0);
  }

  template<>
  inline triplet<int, int, int>
  grid<MesoRD::CubicSubVolume>()
  {
    return triplet<int, int, int>(1, 1, 1);
  }

  template<>
  inline triplet<int, int, int>
  grid<MesoRD::OctahedricSubvolume>()
  {
    return triplet<int, int, int>(2, 2, 2);
  }
}


  /* The direction templates are a set of functions that, for each
   * supported subvolume, will give the directional displacements of
   * and the direction opposite to the argument direction.  Finding
   * the opposite direction is useful when resolving
   * neighbour-relationships of subvolumes: if foo is the <i>left</i>
   * neighbour of bar, then bar is the <i>right</i> neighbour of foo,
   * where <i>left</i> and <i>right</i> are opposite directions.
   *
   * @param direction The direction we are querying
   * @param x         The directional displacement along the
   *                  <code>x</code>-axis
   * @param y         The directional displacement along the
   *                  <code>y</code>-axis.
   * @param z         The directional displacement along the
   *                  <code>z</code>-axis.
   * @return          The direction opposite to @p direction.
   */
namespace
{
  template<class T>
  inline size_t
  direction(const size_t direction, int& x, int& y, int& z)
  {
    x = y = z = 0;
    return direction;
  }

  template<>
  inline size_t
  direction<MesoRD::CubicSubVolume>(
    const size_t direction, int& x, int& y, int& z)
  {
    const MesoRD::CubicSubVolume::NeighbourDirection d
      = static_cast<MesoRD::CubicSubVolume::NeighbourDirection>(
        direction);

    x = y = z = 0;

    switch (d) {

      case MesoRD::CubicSubVolume::down:
        z = -1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::up);

      case MesoRD::CubicSubVolume::east:
        x = +1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::west);

      case MesoRD::CubicSubVolume::north:
        y = +1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::south);

      case MesoRD::CubicSubVolume::south:
        y = -1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::north);

      case MesoRD::CubicSubVolume::west:
        x = -1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::east);

      case MesoRD::CubicSubVolume::up:
        z = +1;
        return static_cast<size_t>(MesoRD::CubicSubVolume::down);
    }

    return direction;
  }

  template<>
  inline size_t
  direction<MesoRD::OctahedricSubvolume>(
    const size_t direction, int& x, int& y, int& z)
  {
    const MesoRD::OctahedricSubvolume::NeighbourDirection d
      = static_cast<MesoRD::OctahedricSubvolume::NeighbourDirection>(
        direction);

    x = y = z = 0;

    switch (d) {

      case MesoRD::OctahedricSubvolume::down:
        z = -2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::up);

      case MesoRD::OctahedricSubvolume::east:
        x = +2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::west);

      case MesoRD::OctahedricSubvolume::north:
        y = +2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::south);

      case MesoRD::OctahedricSubvolume::south:
        y = -2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::north);

      case MesoRD::OctahedricSubvolume::west:
        x = -2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::east);

      case MesoRD::OctahedricSubvolume::up:
        z = +2;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::down);

      case MesoRD::OctahedricSubvolume::eastNorthDown:
        x = +1;
        y = +1;
        z = -1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::westSouthUp);

      case MesoRD::OctahedricSubvolume::eastNorthUp:
        x = +1;
        y = +1;
        z = +1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::westSouthDown);

      case MesoRD::OctahedricSubvolume::eastSouthDown:
        x = +1;
        y = -1;
        z = -1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::westNorthUp);

      case MesoRD::OctahedricSubvolume::eastSouthUp:
        x = +1;
        y = -1;
        z = +1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::westNorthDown);

      case MesoRD::OctahedricSubvolume::westNorthDown:
        x = -1;
        y = +1;
        z = -1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::eastSouthUp);

      case MesoRD::OctahedricSubvolume::westNorthUp:
        x = -1;
        y = +1;
        z = +1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::eastSouthDown);

      case MesoRD::OctahedricSubvolume::westSouthDown:
        x = -1;
        y = -1;
        z = -1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::eastNorthUp);

      case MesoRD::OctahedricSubvolume::westSouthUp:
        x = -1;
        y = -1;
        z = +1;
        return static_cast<size_t>(MesoRD::OctahedricSubvolume::eastNorthDown);
    }

    return direction;
  }
}


  /* Resolve the neighbour-relationships for the subvolume @p sv,
   * located at @p x, @p y, @p z.  For every direction we use the
   * templated direction() function to find directional displacements
   * and the opposite direction.  Using these directional
   * displacements, we can query the @p locations map to find the
   * neighbour.  Note that in every iteration of the loop we set the
   * neighbour-relationships both ways: if foo is a neighbour of bar,
   * then bar is also a neighbour of foo.
   *
   * @param x         The <code>x</code>-coordinate of the subvolume.
   * @param y         The <code>y</code>-coordinate of the subvolume.
   * @param z         The <code>z</code>-coordinate of the subvolume.
   * @param locations An object which maps locations in space to
   *                  pointers to subvolumes.
   * @param sv        The subvolume we are attempting to resolve the
   *                  neighbour-relationships for.
   */
namespace
{
  template<class T>
  inline void
  neighbour(
    const int                                                   x,
    const int                                                   y,
    const int                                                   z,
    const std::map<triplet<int, int, int>, MesoRD::SubVolume*>& locations,
    T*                                                          sv)
  {
    int o_x;
    int o_y;
    int o_z;

    for (size_t dir = 0; dir < static_cast<size_t>(T::getNeighboursNumber());
         dir++) {
      const size_t opp = direction<T>(dir, o_x, o_y, o_z);

      if (o_z < 0 || o_z == 0 && o_y < 0 || o_z == 0 && o_y == 0 && o_x < 0) {
        T* neighbour = find(
          triplet<int, int, int>(x + o_x, y + o_y, z + o_z),
          locations, static_cast<T*>(0));

        if (neighbour) {
          sv        -> setNeighbour(dir, neighbour);
          neighbour -> setNeighbour(opp, sv);
        }
      }
    }
  }
}


  /* Resolves periodic boundary conditions for the subvolume @p sv,
   * located at @p p and contained in @p node.  For every direction,
   * we call the neighbour() function of the node that reported the
   * periodic boundary condition.  This will results in a vector which
   * points to the location of the neighbour we are looking for.
   * Finally, we attempt to find that neighbour, given its supposed
   * location, and set up the neighbour-relationships accordingly.
   *
   * @param p         The location of the subvolume.
   * @param node      The node in the CSG tree that reported the
   *                  periodic boundary condition.
   * @param locations An object which maps locations in space to
   *                  pointers to subvolumes.
   * @param sv        The subvolume we are attempting to resolve the
   *                  periodic boundary conditions for.
   */
namespace
{
  template<class T>
  inline void
  pbc(const MesoRD::Vector&                                     p,
    const MesoRD::CSG::Node*                                    node,
    const std::map<triplet<int, int, int>, MesoRD::SubVolume*>& locations,
    T*                                                          sv)
  {
    const triplet<int, int, int> g = grid<T>();

    int            o_x;
    int            o_y;
    int            o_z;
    MesoRD::Vector v(4);

    for (size_t dir = 0; dir < static_cast<size_t>(T::getNeighboursNumber());
         dir++) {
      direction<T>(dir, o_x, o_y, o_z);

      v(0) = static_cast<double>(o_x) / static_cast<double>(g.first);
      v(1) = static_cast<double>(o_y) / static_cast<double>(g.second);
      v(2) = static_cast<double>(o_z) / static_cast<double>(g.third);
      v(3) = 1.0;
      v    = node -> neighbour(p, v);

      T* neighbour = find(triplet<int, int, int>(
                            static_cast<int>(lrint((v(0) - 0.5) * g.first)),
                            static_cast<int>(lrint((v(1) - 0.5) * g.second)),
                            static_cast<int>(lrint((v(2) - 0.5) * g.third))),
        locations, static_cast<T*>(0));

      if (neighbour)
        sv -> setNeighbour(dir, neighbour);
    }
  }
}


  /* Brief overview of the build template: we loop through the global
   * bounding box, and check each point to see if it is contained in
   * any of the trees.  If the point is contained, then a new
   * subvolume of the template type is created.  Note that the size of
   * a new subvolume is set to zero, which is obviously an invalid
   * size.  The subvolume's neighbours can be set on the fly, because
   * we move through the bounding box in a predetermined fashion --
   * because we build along increasing <code>x</code>, <code>y</code>
   * and <code>z</code>, we can <em>always</em> find the neighbouring
   * subvolume in the negative <code>x</code>-, <code>y</code>- or
   * <code>z</code>-directions.  Once the global bounding box is
   * traversed, we handle subvolume for which PBC is in effect
   * separately.
   *
   * @param global_max The positive extrema of the global bounding
   *                   box.
   * @param global_min The negative extrema of the global bounding
   *                   box.
   * @param all_max    The positive extrema of all the trees in the
   *                   forest.
   * @param all_min    The negative extrema of all the trees in the
   *                   forest.
   * @param forest     The CSG trees from which to build.
   * @param geometries The map in which to store the built subvolumes.
   */
namespace
{
  template<class T>
  inline void
  build(const MesoRD::Vector&                                global_max,
    const MesoRD::Vector&                                    global_min,
    const std::map<std::string, MesoRD::Vector>&             all_max,
    const std::map<std::string, MesoRD::Vector>&             all_min,
    const std::map<std::string, MesoRD::CSG::Node*>&         forest,
    std::map<std::string, std::vector<MesoRD::SubVolume*> >& geometries,
    MesoRD::ProgressBar*    progressBar,
    MesoRD::Random::Random* prng)
  {
    const triplet<int, int, int> g = grid<T>();

    std::list<std::pair<T*, const MesoRD::CSG::Node*> >  boundaries;
    std::map<triplet<int, int, int>, MesoRD::SubVolume*> locations;
    std::map<std::string, MesoRD::CSG::Node*>::const_iterator i;

      /* Clear the geometries for every compartment.  This makes sure
       * that every vector is at least initialised.  We can then
       * safely add to each vector while looping through the bounding
       * box.
       */
    for (i = forest.begin(); i != forest.end(); i++)
      geometries[i -> first].resize(0);

      /* Construct the boundaries.  Note that the three for-loops here
       * will generate the coordinates of the upper-left-top corner of
       * the subvolume.  Therefore, in order not to miss any
       * subvolumes, we may need to adjust the boundaries a little.
       * The progress bar, if present, is called on every iteration in
       * the outermost <code>z</code>-loop.
       */
    const int z_min = g.third * static_cast<int>(lrint(floor(global_min(2))))
      - (g.third - 1);
    const int z_max = g.third * static_cast<int>(lrint(ceil(global_max(2))))
      + (g.third - 1);

    for (int z = z_min; z <= z_max; z += 1) {
      if (progressBar) {
          //std::cout << "  FRACTION " << 1.0 * (z - z_min) / (z_max - z_min + 1) << std::endl; // FIXME!! Remove!

        progressBar -> setFraction(1.0 * (z - z_min) / (z_max - z_min + 1));
      }
      
      const bool even  = (z - z_min) % g.third == 0;
      const int  x_max =
	g.first	* static_cast<int>(lrint(ceil(global_max(0))))
	- (even ? 0 : 1);
      const int  x_min =
	g.first * static_cast<int>(lrint(floor(global_min(0))))
        + (even ? 0 : 1);
      const int  y_max =
	g.second * static_cast<int>(lrint(ceil(global_max(1))))
        - (even ? 0 : 1);
      const int  y_min =
	g.second * static_cast<int>(lrint(floor(global_min(1))))
        + (even ? 0 : 1);

      for (int y = y_min; y <= y_max; y += g.second) {
        for (int x = x_min; x <= x_max; x += g.first) {

            /* Now we have <code>x</code>, <code>y</code> and
             * <code>z</code>, which define the left, upper, top
             * corner of the prospective subvolume.  We can find the
             * centre point of the prospective subvolume,
             * <code>p</code>, as follows.
             */
          T*             sv = 0;
          MesoRD::Vector p  = MesoRD::Vector(4);

	  p(0) = 0.5 * (2 / g.first  * x + 1);
	  p(1) = 0.5 * (2 / g.second * y + 1);
	  p(2) = 0.5 * (2 / g.third  * z + 1);
	  p(3) = 1.0;

            /* Loop through all the trees and check whether any of
             * them contain the point <code>p</code>.
             */
          for (i = forest.begin(); i != forest.end(); i++) {
            const std::string        id   = i -> first;
            const MesoRD::CSG::Node* tree = i -> second;

            std::map<std::string, MesoRD::Vector>::const_iterator max =
              all_max.find(id);
            std::map<std::string, MesoRD::Vector>::const_iterator min =
              all_min.find(id);

              /* If the point lies inside the bounding box for the
               * current tree, check if the point is contained in the
               * tree.  If it is create a new subvolume, and store its
               * location in the <code>locations</code> map.
               * Furthermore, if the point is not only contained, but
               * also a boundary point for where PBC is active, add
               * the subvolume to list of boundary points.
               */
            if (tree && max != all_max.end() && min != all_min.end()
              && p(0) >= (min -> second)(0)
              && p(0) <= (max -> second)(0)
              && p(1) >= (min -> second)(1)
              && p(1) <= (max -> second)(1)
              && p(2) >= (min -> second)(2)
              && p(2) <= (max -> second)(2)) {

              MesoRD::CSG::Node::Containment containment =
                tree -> contains(p);

              if  (containment.type
                == MesoRD::CSG::Node::Containment::exterior) {
                continue;

              } else {
                triplet<int, int, int> t(x, y, z);

                  /* Create a new subvolume.  Note that the new
                   * subvolume has size zero, which is clearly
                   * invalid.
                   */
                sv = new T(prng, 0);
                sv -> setXYZ(static_cast<float>(p(0)),
			     static_cast<float>(p(1)),
			     static_cast<float>(p(2)));

                geometries[id].push_back(sv);
                locations[t] = sv;

                if (containment.type
                  == MesoRD::CSG::Node::Containment::boundary) {

                    /* Subvolume is boundary point, so add it and the
                     * node that contains it, to the list of
                     * boundaries.
                     */
                  boundaries.push_back(
                    std::pair<T*, const MesoRD::CSG::Node*>(
                      sv, containment.node));
                }
              }

                /* If the current compartment has claimed lordship of
                 * this subvolume, stop looking for new masters.
                 */
              break;
            }
          }

            /* Resolve the neighbour-relationships for the subvolume
             * just built.  This is where we use the
             * <code>locations</code> map.
             */
          if (sv)
            neighbour(x, y, z, locations, sv);
        }
      }
    }

    if (progressBar)
      progressBar -> setFraction(1.0);

      /* Handle the boundaries with periodic boundary conditions, if
       * any such subvolumes exist.  Again, the <code>locations</code>
       * map comes in handy.
       */
    while (!boundaries.empty()) {
      MesoRD::Vector           p    = MesoRD::Vector(4);
      T*                       sv   = boundaries.front().first;
      const MesoRD::CSG::Node* node = boundaries.front().second;

      p(0) = sv -> getX();
      p(1) = sv -> getY();
      p(2) = sv -> getZ();
      p(3) = 1.0;

      pbc(p, node, locations, sv);
      boundaries.pop_front();
    }
  }
}



  /**************************
   * ORDINARY CSG FUNCTIONS *
   **************************/

  /* Build the node using the constructors of the respective derived
   * classes.  This is a recursive implementation.  The function will
   * call itself on anything but terminal nodes.  Terminal nodes in
   * this case are the geometric primitives.  The non-terminals are
   * transformation nodes and set operation nodes.  Whenever checking
   * for absolutely mandatory children or attributes, we do not
   * enclose the code in a try-block.  That way, if the attribute is
   * not found, the generated exception will fall through.
   */
MesoRD::CSG::Node*
MesoRD::CSG::fromXML(
  const Annotation* a, const UnitTable* ut, const double size, Matrix T)
{
  Node* node = 0;

  if (a -> getName() == Identifier::XML::Box::name) {

      /* Constructing a new box node.  A box element can have
       * children.  If a box node has children, they must be periodic
       * boundary condition (PBC) elements for the three different
       * axis.  The default is no periodic boundary conditions
       * anywhere.
       */
    bool   pbc_x      = false;
    bool   pbc_y      = false;
    bool   pbc_z      = false;
    double conversion = 1.0;

    const double x =
      a -> getDoubleAttribute(Identifier::XML::Box::Attributes::x);
    const double y =
      a -> getDoubleAttribute(Identifier::XML::Box::Attributes::y);
    const double z =
      a -> getDoubleAttribute(Identifier::XML::Box::Attributes::z);

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);

      if (child && child -> getName() == Identifier::XML::PBC::name) {
        try {
          pbc_x =
            child -> getBoolAttribute(Identifier::XML::PBC::Attributes::x);
        } catch (...) {
          ;
        }
        try {
          pbc_y =
            child -> getBoolAttribute(Identifier::XML::PBC::Attributes::y);
        } catch (...) {
          ;
        }
        try {
          pbc_z =
            child -> getBoolAttribute(Identifier::XML::PBC::Attributes::z);
        } catch (...) {
          ;
        }
        break;
      }
    }

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Box::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::metre());
    } catch (...) {
      conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
    }

    node = new Box(
      x * conversion / size,
      y * conversion / size,
      z * conversion / size,
      pbc_x, pbc_y, pbc_z, T);

  } else if (a -> getName() == Identifier::XML::Compartment::name) {

      /* Constructing a new compartment node.  In order to get a
       * pointer to the referred compartment, we must have built the
       * CSG tree for the relevant compartment.  Since that tree is
       * not known at this point, all we can do is defer the
       * assignment, only noting the relevant identifier for the
       * compartment in question.  The pointer to the compartment will
       * have to be set up during a subsequent traversal of the tree.
       */
    const std::string id =
      a -> getStringAttribute(Identifier::XML::Compartment::id);
    node = new Compartment(id, T);

  } else if (a -> getName() == Identifier::XML::Cone::name) {

      /* Constructing a new cone node.  A cone node can have no
       * children, nor can it have periodic boundary conditions.  We
       * just need to read the radius and the height.
       */
    double conversion = 1.0;

    const double height =
      a -> getDoubleAttribute(Identifier::XML::Cone::Attributes::height);
    const double radius =
      a -> getDoubleAttribute(Identifier::XML::Cone::Attributes::radius);

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Cone::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::metre());
    } catch (...) {
      conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
    }

    node = new Cone(
      height * conversion / size,
      radius * conversion / size,
      T);

  } else if (a -> getName() == Identifier::XML::Cylinder::name) {

      /* Constructing a new cylinder node.  A cylinder element can
       * have one child.  If a cylinder node has a child, it must be a
       * periodic boundary condition (PBC) element for the major axis
       * of the cylinder.  The default is no periodic boundary
       * condition.
       */
    bool   pbc        = false;
    double conversion = 1.0;

    const double height =
      a -> getDoubleAttribute(Identifier::XML::Cylinder::Attributes::height);
    const double radius =
      a -> getDoubleAttribute(Identifier::XML::Cylinder::Attributes::radius);

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);

      if (child && child -> getName() == Identifier::XML::PBC::name) {
        try {
          pbc =
            child -> getBoolAttribute(Identifier::XML::PBC::Attributes::z);
        } catch (...) {
          ;
        }
        break;
      }
    }

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Cylinder::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::metre());
    } catch (...) {
      conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
    }

    node = new Cylinder(
      height * conversion / size,
      radius * conversion / size,
      pbc, T);

  } else if (a -> getName() == Identifier::XML::Difference::name) {

      /* Constructing a new difference node.  A difference element
       * must have exactly two children, corresponding to the left and
       * right operands of the difference operation.  The children are
       * read in recursively.  If a difference node has more than two
       * children, all but the first two are silently disregarded.
       */
    Node* left  = 0;
    Node* right = 0;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);

      if (child && !left) {
        left = fromXML(child, ut, size, T);
      } else if (child && !right) {
        right = fromXML(child, ut, size, T);
        if (right)
          break;
      }
    }

    node = new Difference(left, right);

  } else if (a -> getName() == Identifier::XML::Intersection::name) {

      /* Constructing a new intersection node.  An intersection
       * element can have any number of children, even though the
       * operation makes no sense if the number of children is less
       * than two.  We read in the children recursively one after the
       * other.  The order is in fact irrelevant.
       */
    std::list<Node*> operands;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);
      if (child)
        operands.push_back(fromXML(child, ut, size, T));
    }

    node = new Intersection(operands);

  } else if (a -> getName() == Identifier::XML::Mesh::name) {

      /* Constructing a new mesh node.  A mesh element can have any
       * number of children, each defining a triangle.  Each triangle
       * node in turn must have three children, each defining a vertex
       * of the triangle.  If a triangle node has more than three
       * children, all but the first three are silently ignored.
       */
    std::list<Mesh::Triangle> triangles;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);

      if (child -> getName() == Identifier::XML::Mesh::Triangle::name
        && child -> getNumChildren() >= 3) {
        Mesh::Triangle triangle = { Vector(4), Vector(4), Vector(4) };

        for (size_t j = 0; j < 3; j++) {
          const Annotation* vertex = child -> getChild(j);
	  Vector            p      = Vector(4);

          if (vertex && vertex -> getName()
            == Identifier::XML::Mesh::Triangle::Vertex::name) {

            double conversion = 1.0;

            try {
              const std::string unit_str = vertex -> getStringAttribute(
                Identifier::XML::Mesh::Triangle::Vertex::Attributes::units);
              const Unit*       unit_ptr = ut -> getUnit(unit_str);

              conversion = unit_ptr -> toUnit(Unit::metre());
            } catch (...) {
              conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
            }

	    p(0) = vertex -> getDoubleAttribute(
                Identifier::XML::Mesh::Triangle::Vertex::Attributes::x)
              * conversion / size;
	    p(1) = vertex -> getDoubleAttribute(
                Identifier::XML::Mesh::Triangle::Vertex::Attributes::y)
              * conversion / size;
	    p(2) = vertex -> getDoubleAttribute(
                Identifier::XML::Mesh::Triangle::Vertex::Attributes::z)
              * conversion / size;
	    p(3) = 1.0;

            switch (j) {

              case 0:
                triangle.p0 = p;
                break;

              case 1:
                triangle.p1 = p;
                break;

              case 2:
                triangle.p2 = p;
                break;
            }
          }
        }
        triangles.push_back(triangle);
      }
    }

    node = new Mesh(triangles, T);

  } else if (a -> getName() == Identifier::XML::Rotation::name) {

      /* Processing a rotation node.  A rotation node should have
       * exactly one child, namely the subtree which the
       * transformation applies to.  If a rotation node has more than
       * one child, all but the first are silently ignored.  In the
       * code below we read in the rotation axis, <code>a_x</code>,
       * <code>a_y</code>, <code>a_z</code>, as well as the rotation
       * angle, <code>angle</code>.  We construct the new
       * transformation matrix by first transforming the system so
       * that the rotation axis is aligned with the
       * <code>z</code>-axis.  Then we rotate by <code>angle</code>
       * around the <code>z</code>-axis, and back-transform the
       * system, so that the rotation axis is again in its original
       * position.  Finally we recursively read the child with
       * the new transformation matrix.
       */
    double conversion = 1.0;

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Rotation::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::radian());
    } catch (...) {
      ;
    }

    const double angle =
      fmod(a -> getDoubleAttribute(
             Identifier::XML::Rotation::Attributes::angle)
        * conversion + 2.0 * M_PI, 2.0 * M_PI);

    const double a_x =
      a -> getDoubleAttribute(Identifier::XML::Rotation::Attributes::x);
    const double a_y =
      a -> getDoubleAttribute(Identifier::XML::Rotation::Attributes::y);
    const double a_z =
      a -> getDoubleAttribute(Identifier::XML::Rotation::Attributes::z);

    const double rho2_xy  = a_x * a_x + a_y * a_y;
    const double rho2_xyz = rho2_xy   + a_z * a_z;
    const double angle_y  = rho2_xyz > 0.0 ? acos(a_z / sqrt(rho2_xyz)) : 0.0;
    const double angle_z  = rho2_xy  > 0.0 ? acos(a_x / sqrt(rho2_xy))  : 0.0;

    Matrix R(4, 4);

    R       = Matrix::eye(4);
    R(0, 0) = +1.0 * cos(angle_y) * cos(angle_z);
    R(0, 1) = +1.0 * sin(angle_z);
    R(0, 2) = -1.0 * sin(angle_y) * cos(angle_z);
    R(1, 0) = -1.0 * cos(angle_y) * sin(angle_z);
    R(1, 1) = +1.0 * cos(angle_z);
    R(1, 2) = +1.0 * sin(angle_y) * sin(angle_z);
    R(2, 0) = +1.0 * sin(angle_y);
    R(2, 2) = +1.0 * cos(angle_y);
    T       = T * R;

    R       = Matrix::eye(4);
    R(0, 0) = +1.0 * cos(angle);
    R(0, 1) = -1.0 * sin(angle);
    R(1, 0) = +1.0 * sin(angle);
    R(1, 1) = +1.0 * cos(angle);
    T       = T * R;

    R       = Matrix::eye(4);
    R(0, 0) = +1.0 * cos(angle_y) * cos(angle_z);
    R(0, 1) = -1.0 * cos(angle_y) * sin(angle_z);
    R(0, 2) = +1.0 * sin(angle_y);
    R(1, 0) = +1.0 * sin(angle_z);
    R(1, 1) = +1.0 * cos(angle_z);
    R(2, 0) = -1.0 * sin(angle_y) * cos(angle_z);
    R(2, 1) = +1.0 * sin(angle_y) * sin(angle_z);
    R(2, 2) = +1.0 * cos(angle_y);
    T       = T * R;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation *child = a -> getChild(i);
      if (child) {
        node = fromXML(child, ut, size, T);
        break;
      }
    }

  } else if (a -> getName() == Identifier::XML::Scale::name) {

      /* Processing a scale node.  A scale node should have exactly
       * one child, namely the subtree which the transformation
       * applies to.  If a scale node has more than one child, all but
       * the first are silently ignored.  In the code below the scale
       * vector is read into the scale matrix, <code>S</code>.  Then
       * the compound transformation matrix is computed as the product
       * of the current transformation matrix and the new scale
       * matrix.  Finally we recursively read the child with the new
       * transformation matrix.
       */
    Matrix S(Matrix::eye(4));
    S(0, 0) = a -> getDoubleAttribute(Identifier::XML::Scale::Attributes::x);
    S(1, 1) = a -> getDoubleAttribute(Identifier::XML::Scale::Attributes::y);
    S(2, 2) = a -> getDoubleAttribute(Identifier::XML::Scale::Attributes::z);

    std::cout << "parsed scale:" << S << std::endl;
    

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);
      if (child) {
        node = fromXML(child, ut, size, T * S);
        break;
      }
    }

  } else if (a -> getName() == Identifier::XML::Shear::name) {

      /* Processing a shear node.  A shear node should have exactly
       * one child, namely the subtree which the transformation
       * applies to.  If a shear node has more than one child, all but
       * the first are silently ignored.  In the code below we read in
       * the shear axis, <code>s_x</code>, <code>s_y</code>,
       * <code>s_z</code>, as well as the shear angle,
       * <code>angle</code>.  We construct the new transformation
       * matrix by first transforming the system so that the shear
       * axis is aligned with the <code>z</code>-axis.  Then we shear
       * by <code>angle</code> around the <code>z</code>-axis, and
       * back-transform the system, so that the shear axis is again in
       * its original position.  Finally we recursively read the child
       * with the new transformation matrix.
       */
    double conversion = 1.0;

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Shear::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::radian());
    } catch (...) {
      ;
    }

    const double angle =
      fmod(a -> getDoubleAttribute(
             Identifier::XML::Rotation::Attributes::angle)
        * conversion + 2.0 * M_PI, 2.0 * M_PI);

    const double s_x =
      a -> getDoubleAttribute(Identifier::XML::Shear::Attributes::x);
    const double s_y =
      a -> getDoubleAttribute(Identifier::XML::Shear::Attributes::y);
    const double s_z =
      a -> getDoubleAttribute(Identifier::XML::Shear::Attributes::z);

    const double rho2_xy  = s_x * s_x + s_y * s_y;
    const double rho2_xyz = rho2_xy   + s_z * s_z;
    const double angle_y  = rho2_xyz > 0.0 ? acos(s_z / sqrt(rho2_xyz)) : 0.0;
    const double angle_z  = rho2_xy  > 0.0 ? acos(s_x / sqrt(rho2_xy))  : 0.0;

    Matrix R(4, 4);

    R       = Matrix::eye(4);
    R(0, 0) = +1.0 * cos(angle_y) * cos(angle_z);
    R(0, 1) = +1.0 * sin(angle_z);
    R(0, 2) = -1.0 * sin(angle_y) * cos(angle_z);
    R(1, 0) = -1.0 * cos(angle_y) * sin(angle_z);
    R(1, 1) = +1.0 * cos(angle_z);
    R(1, 2) = +1.0 * sin(angle_y) * sin(angle_z);
    R(2, 0) = +1.0 * sin(angle_y);
    R(2, 2) = +1.0 * cos(angle_y);
    T       = T * R;

    R       = Matrix::eye(4);
    R(2, 1) = fabs(sin(angle)) > std::numeric_limits<double>::min()
      ? cos(angle) / sin(angle) : 0.0;
    T       = T * R;

    R       = Matrix::eye(4);
    R(0, 0) = +1.0 * cos(angle_y) * cos(angle_z);
    R(0, 1) = -1.0 * cos(angle_y) * sin(angle_z);
    R(0, 2) = +1.0 * sin(angle_y);
    R(1, 0) = +1.0 * sin(angle_z);
    R(1, 1) = +1.0 * cos(angle_z);
    R(2, 0) = -1.0 * sin(angle_y) * cos(angle_z);
    R(2, 1) = +1.0 * sin(angle_y) * sin(angle_z);
    R(2, 2) = +1.0 * cos(angle_y);
    T       = T * R;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation *child = a -> getChild(i);
      if (child) {
        node = fromXML(child, ut, size, T);
        break;
      }
    }

  } else if (a -> getName() == Identifier::XML::Sphere::name) {

      /* Constructing a new sphere node.  A sphere node can have no
       * children, nor can it have periodic boundary conditions.  We
       * just need to read the radius.
       */
    double conversion = 1.0;

    const double radius =
      a -> getDoubleAttribute(Identifier::XML::Sphere::Attributes::radius);

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Sphere::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::metre());
    } catch (...) {
      conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
    }

    node = new Sphere(
      radius * conversion / size,
      T);

  } else if (a -> getName() == Identifier::XML::Translation::name) {

      /* Processing a translation node.  A translation node should
       * have exactly one child, namely the subtree which the
       * transformation applies to.  If a translation node has more
       * than one child, all but the first are silently ignored.  In
       * the code below the translation vector is read into the
       * translation matrix, <code>D</code>.  Then the compound
       * transformation matrix is computed as the product of the
       * current transformation matrix and the new translation matrix.
       * Finally we recursively read the child with the new
       * transformation matrix.
       */
    double conversion = 1.0;

    try {
      const std::string unit_str = a -> getStringAttribute(
        Identifier::XML::Translation::Attributes::units);
      const Unit*       unit_ptr = ut -> getUnit(unit_str);

      conversion = unit_ptr -> toUnit(Unit::metre());
    } catch (...) {
      conversion = ut -> getUnit("length") -> toUnit(Unit::metre());
    }

    Matrix D(Matrix::eye(4));
    D(0, 3) = a -> getDoubleAttribute(
      Identifier::XML::Translation::Attributes::x)
      * conversion / size;
    D(1, 3) = a -> getDoubleAttribute(
      Identifier::XML::Translation::Attributes::y)
      * conversion / size;
    D(2, 3) = a -> getDoubleAttribute(
      Identifier::XML::Translation::Attributes::z)
      * conversion / size;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);
      if (child) {
        node = fromXML(child, ut, size, T * D);
        break;
      }
    }

  } else if (a -> getName() == Identifier::XML::Union::name) {

      /* Constructing a new union node.  A union element can have any
       * number of children, even though the operation makes no sense
       * if the number of children is less than two.  We read in the
       * children recursively one after the other.  The order is in
       * fact irrelevant.
       */
    std::list<Node*> operands;

    for (size_t i = 0; i < a -> getNumChildren(); i++) {
      const Annotation* child = a -> getChild(i);
      if (child)
        operands.push_back(fromXML(child, ut, size, T));
    }

    node = new Union(operands);
  }

  return node;
}


  /* First, compute the bounding boxes for each CSG tree.  The global
   * bounding box is the bounding box that completely contains the
   * bounding boxes for all the trees.  The global bounding box
   * defines the extrema of the union of the geometries in the global
   * coordinate system.  In what follows, all space outside the global
   * bounding box can be safely ignored.  Finally, the geometries are
   * realised by the templated build() function.
   */
void
MesoRD::CSG::realise(
  const std::map<std::string, Node*>&              forest,
  const Subvolume_t                                type,
  std::map<std::string, std::vector<SubVolume*> >& geometries,
  ProgressBar*                                     progressBar,
  Random::Random*                                  prng)
{
  Vector                                       global_max(4);
  Vector                                       global_min(4);
  std::map<std::string, Vector>                all_max;
  std::map<std::string, Vector>                all_min;
  std::map<std::string, Node*>::const_iterator i;

  if (progressBar)
    progressBar -> setFraction(0.0);

  global_max(0) = global_max(1) = global_max(2) =
    -1.0 * std::numeric_limits<double>::max();
  global_min(0) = global_min(1) = global_min(2) =
    +1.0 * std::numeric_limits<double>::max();

  for (i = forest.begin(); i != forest.end(); i++) {
    const std::string id   = i -> first;
    const Node*       node = i -> second;

    if (node) {
      Vector max(4);
      Vector min(4);

      node -> getBoundingBox(max, min, Matrix::eye(4));

      all_min[id] = min;
      all_max[id] = max;

      for (size_t j = 0; j < 3; j++) {
        global_max(static_cast<unsigned int>(j)) =
	  MesoRD::max(ceil(global_max(static_cast<unsigned int>(j))),
		      max(static_cast<unsigned int>(j)));
        global_min(static_cast<unsigned int>(j)) =
	  MesoRD::min(floor(global_min(static_cast<unsigned int>(j))),
		      min(static_cast<unsigned int>(j)));
      }
    }
  }

  switch (type) {
    case cubic:
      build<MesoRD::CubicSubVolume>(
        global_max,
        global_min,
        all_max,
        all_min,
        forest,
        geometries,
        progressBar,
        prng);
      break;

    case octahedric:
      build<MesoRD::OctahedricSubvolume>(
        global_max,
        global_min,
        all_max,
        all_min,
        forest,
        geometries,
        progressBar,
        prng);
      break;
  }
}



  /* Dump the entire forest in POV-Ray format using the individual
   * node's putPOVRay() function.  But first some prerequisites.
   */
std::ostream&
MesoRD::CSG::putPOVRay(
  const std::map<std::string, Node*>& forest, std::ostream& os)
{
    /* Determine global bounding box.  This almost mirrors some code
     * from the CSG building phase.  This is then used to determine
     * the distance along the <code>z</code>-axis from the origin a
     * camera point must have, so that a ninety degree frustum at the
     * camera point, aimed at the origin, will contain the entire
     * geometry.
     */
  double z_max  = -1.0 * std::numeric_limits<double>::max();
  double xy_max = -1.0 * std::numeric_limits<double>::max();
  std::map<std::string, Node*>::const_iterator i;

  for (i = forest.begin(); i != forest.end(); i++) {
    const Node* node = i -> second;

    if (node) {
      MesoRD::Vector max(4);
      MesoRD::Vector min(4);

      node -> getBoundingBox(max, min, MesoRD::Matrix::eye(4));

      xy_max = MesoRD::max(xy_max,
        MesoRD::max(fabs(ceil(max(0))), fabs(floor(min(0)))));
      xy_max = MesoRD::max(xy_max,
        MesoRD::max(fabs(ceil(max(1))), fabs(floor(min(1)))));
      z_max = MesoRD::max(z_max,
        MesoRD::max(fabs(ceil(max(2))), fabs(floor(min(2)))));
    }
  }

  const double dist = z_max + double(1.1) * xy_max;


    /* Emit the preamble, with background, camera and lighting
     * information.
     */
  os << "// ***********************"                            << std::endl
     << "// * Generated by MesoRD *"                            << std::endl
     << "// ***********************"                            << std::endl
     << "// Run e.g."                                           << std::endl
     << "//   povray <filename> +H500 +W500 +FT +A0.1 +D +P -V" << std::endl
     << "// to generate image."                                 << std::endl
     << std::endl

     << "// ************************************"               << std::endl
     << "// * CAMERA SETTINGS, BACKGROUND, ETC *"               << std::endl
     << "// ************************************"               << std::endl
     << "#declare HEIGHT       = 500;"                          << std::endl
     << "#declare WIDTH        = 500;"                          << std::endl
     << "#declare ASPECT_RATIO = WIDTH / HEIGHT;"               << std::endl
     << "#declare OPERATIONS   = true;"                         << std::endl
     << "#declare USE_MERGE    = false;"                        << std::endl
     << std::endl

     << "camera"                                                << std::endl
     << "{"                                                     << std::endl
     << "  location  < 0,            0,  "
     << dist << ">"                                             << std::endl
     << "  look_at   < 0,            0,  1>"                    << std::endl
     << "  direction < 0,            0,  1>"                    << std::endl
     << "  right     <-ASPECT_RATIO, 0,  0>"                    << std::endl
     << "  angle     90"                                        << std::endl
     << "}"                                                     << std::endl
     << std::endl

     << "background"                                            << std::endl
     << "{"                                                     << std::endl
     << "  color rgb <1.0, 1.0, 1.0>"                           << std::endl
     << "}"                                                     << std::endl
     <<std::endl

     << "light_source"                                          << std::endl
     << "{"                                                     << std::endl
     << "  <0, 0, " << dist << ">"                              << std::endl
     << "  rgb <1.0, 1.0, 1.0>"                                 << std::endl
     << "}"                                                     << std::endl
     <<std::endl                                                << std::endl

     << "// *******************************"                    << std::endl
     << "// * MACROS COMMON TO ALL SHAPES *"                    << std::endl
     << "// *******************************"                    << std::endl
     << "#default {"                                            << std::endl
     << "  finish "                                             << std::endl
     << "  {"                                                   << std::endl
     << "    ambient    0"                                      << std::endl
     << "    brilliance 5"                                      << std::endl
     << "    diffuse    0.1"                                    << std::endl
     << "    roughness  0.002"                                  << std::endl
     << "    specular   0.2"                                    << std::endl
     << "    reflection"                                        << std::endl
     << "    {"                                                 << std::endl
     << "      0.1, 1.0"                                        << std::endl
     << "      fresnel on"                                      << std::endl
     << "    }"                                                 << std::endl
     << "    conserve_energy"                                   << std::endl
     << "  }"                                                   << std::endl
     << "}"                                                     << std::endl
     << std::endl                                               << std::endl;


    /* Emit all the compartments.  The colours are cycled through one
     * by one.
     *
     * @todo It would probably be nicer if one could specify the
     * colour of the compartment in the model.
     */
  size_t count = 0;
  for (i = forest.begin(); i != forest.end(); i++) {
    std::string              asterix("");
    const std::string        id   = i -> first;
    const MesoRD::CSG::Node* node = i -> second;

    for (size_t j = 0; j < id.length(); j++)
      asterix += '*';

    os << "// ****************" << asterix    << std::endl
       << "// * COMPARTMENT "   << id << " *" << std::endl
       << "// ****************" << asterix    << std::endl;

    switch (count++ % 3) {
      case 0:
        node -> putPOVRay((const float)1.00, (const float)0.92, (const float)0.96, (const float)0.90, (const float)1.0, os);
        break;

      case 1:
        node -> putPOVRay((const float)0.96, (const float)1.00, (const float)0.92, (const float)0.90, (const float)1.0, os);
        break;

      case 2:
        node -> putPOVRay((const float)0.92, (const float)0.96, (const float)1.00, (const float)0.90, (const float)1.0, os);
        break;
    }

    os << std::endl << std::endl;
  }

  return os;
}


  /* Resolve tree inclusion by means of the Compartment node.  This
   * function traverses @p tree recursively.  If we find a
   * SetOperation node, we recurse down all its children.  Should we
   * stumble upon a Compartment node, we check if we have passed that
   * tree corresponding to the compartment already.  This can be
   * checked by looking through the @p trace.  If the compartment is
   * in the trace, there is a cycle in the inclusion graph, so we
   * throw an exception.  Otherwise, we add the identifier of the
   * referred compartment to the trace and start recursing down the
   * referred compartment.  For this to work, the trace must be primed
   * with the identifier of the first tree when resolve_tree() is
   * first called.  Obviously, if the referred compartment is not
   * found in the forest, there is another kind of error.  Note that
   * the trace always traces a single path through the inclusion
   * graph.  This is accomplished by copying the trace on each
   * recursion, instead of passing it by reference.
   */
namespace
{
  void
  resolve_tree(
    const std::map<std::string, MesoRD::CSG::Node*>& forest,
    std::list<std::string>                          trace,
    MesoRD::CSG::Node*                               tree)
    throw(MesoRD::CSG::CyclicInclusion, MesoRD::CSG::UndefinedCompartment)
  {
    MesoRD::CSG::Compartment*  c;
    MesoRD::CSG::SetOperation* o;

    if ((o = dynamic_cast<MesoRD::CSG::SetOperation*>(tree))) {
      std::list<MesoRD::CSG::Node*> operands = o -> getOperands();
      std::list<MesoRD::CSG::Node*>::iterator i;

      for (i = operands.begin(); i != operands.end(); i++)
        resolve_tree(forest, trace, *i);

    } else if ((c = dynamic_cast<MesoRD::CSG::Compartment*>(tree))) {
      const std::string id = c -> getId();

      if (std::find(trace.begin(), trace.end(), id) != trace.end()) {
        throw MesoRD::CSG::CyclicInclusion(id.c_str());

      } else {
        std::map<std::string, MesoRD::CSG::Node*>::const_iterator i;

        if ((i = forest.find(id)) != forest.end()) {
          trace.push_back(i -> first);
          resolve_tree(forest, trace, i -> second);
          c -> setCompartment(i -> second);

        } else {
          throw MesoRD::CSG::UndefinedCompartment(id.c_str());
        }
      }
    }
  }
}


  /* Resolve all trees iteratively.  For each tree, prime the trace
   * with the identifier of the tree before starting to recurse down
   * it.
   */
void
MesoRD::CSG::resolve(std::map<std::string, Node*>& forest)
  throw(CyclicInclusion, UndefinedCompartment)
{
  std::map<std::string, Node*>::iterator i;

  for (i = forest.begin(); i != forest.end(); i++) {
    std::list<std::string> trace;

    trace.push_back(i -> first);
    resolve_tree(forest, trace, i -> second);
  }
}
