/* The CSG class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSG.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef MESORD_CSG_CSG_H
#define MESORD_CSG_CSG_H 1

#include <map>
#include <string>

#include "CSGNode.hpp"
#include "../definition.hpp"
#include "../Annotation.hpp"
#include "../SubVolume.hpp"
#include "../linalg/Matrix.hpp"
#include "../ui/ProgressBar.hpp"


namespace MesoRD
{

    /**
     * @brief The <i>Constructive Solid Geometry</i>, or <i>%CSG</i>,
     * module.
     *
     * Geometries in MesoRD are built using <i>Constructive Solid
     * Geometry</i>, or <i>%CSG</i>.  The complete MesoRD::System
     * consists of several MesoRD::Compartment instances, and every
     * MesoRD::Compartment must have a %CSG tree.  By artistic
     * application of the English language, the root nodes of all the
     * system's compartments are said to make up the %CSG
     * <i>forest</i>.  Because a forest is a <em>union</em> of trees,
     * and <code>union</code> is a valid %CSG operation, forests as
     * well as trees are both %CSG <i>procedures</i>.  The theory of
     * constructive solid geometry procedures is discussed more in
     * depth elsewhere.
     *
     * A single %CSG tree is built from an XML definition using the
     * fromXML() function.  Typically, a tree or a forest is of little
     * interest by itself; what is really needed is a
     * <em>realisation</em> of the %CSG procedure in terms of
     * <em>subvolumes</em>.  Such a realisation is brought about by
     * the realise() function.
     *
     * This module, defined by the CSG namespace, includes some
     * functions for high-level management of geometry building.  The
     * low-level management is implemented in the CSG::Node classes.
     */
  namespace CSG
  {

      /**
       * @brief Class for reporting errors caused by cycles in the
       * compartment inclusion graph.
       *
       * It is possible to define geometries where for instance
       * MesoRD::Compartment <code>foo</code> includes
       * MesoRD::Compartment <code>bar</code>.  This is quite in
       * order.  If, however, MesoRD::Compartment <code>bar</code>
       * also includes MesoRD::Compartment <code>foo</code>, there is
       * a cycle in the compartment inclusion graph.  This cannot be
       * tolerated, so a CyclicInclusion exception will be thrown.
       */
    class CyclicInclusion : public std::runtime_error
    {
    private:
      const char* id;

    public:

        /**
         * @brief The CyclicInclusion value constructor.
         *
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param id A textual identifier of the MesoRD::Compartment
         *           which caused the cyclic inclusion problem.
         */
      inline
      CyclicInclusion(const char* id)
        : std::runtime_error("CyclicInclusion"),
          id(id)
        {}


        /**
         * @brief Get the identifier of the MesoRD::Compartment that
         * caused the cyclic inclusion problem.
         *
         * This function returns a pointer to the string holding the
         * identifier of the MesoRD::Compartment that caused the
         * cyclic inclusion problem.  This may be the
         * <code>null</code> pointer.
         *
         * @return A pointer to the identifier of the
         *         MesoRD::Compartment that caused the cyclic
         *         inclusion problem.
         */
      inline const char*
      getId() const
        {
          return id;
        }
    };


      /**
       * @brief Class for reporting errors caused by inclusion of
       * undefined compartments.
       *
       * It is possible to define geometries where for instance
       * MesoRD::Compartment <code>foo</code> includes
       * MesoRD::Compartment <code>bar</code>.  This is quite in
       * order.  If, however, MesoRD::Compartment <code>bar</code> is
       * nowhere defined, there is an unrecoverable problem.  An
       * UndefinedCompartment exception will be thrown.
       */
    class UndefinedCompartment : public std::runtime_error
    {
    private:
      const char* id;

    public:

        /**
         * @brief The UndefinedCompartment value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param id A pointer to the identifier of the undefined
         *           MesoRD::Compartment, the inclusion of which was
         *           attempted.
         */
      inline
      UndefinedCompartment(const char* id)
        : std::runtime_error("UndefinedCompartment"),
          id(id)
        {}


        /**
         * @brief Get the identifier of the undefined
         * MesoRD::Compartment, the inclusion of which was attempted.
         *
         * This function returns a pointer to the string holding the
         * identifier of the undefined MesoRD::Compartment, the
         * inclusion of which was attempted.  This may be the
         * <code>null</code> pointer.
         *
         * @return A pointer to the undefined MesoRD::Compartment, the
         *         inclusion of which was attempted.
         */
      inline const char*
      getId() const
        {
          return id;
        }
    };


      /**
       * @brief Read and parse the geometry from an XML representation
       * of the %CSG tree.
       *
       * This function will attempt to recursively read and parse the
       * information contained in an XML %CSG tree.  The function will
       * in fact rebuild the entire tree using subclasses of the
       * CSG::Node class instead of XML nodes.  During rebuilding, the
       * tree is also restructured a bit:
       *
       * <ul>
       *
       *   <li>All transformation nodes are pruned.  Any
       *   transformation information is stored directly in the
       *   terminal, primitive nodes instead.</li>
       *
       *   <li>The new tree is built in units of subvolume size.
       *   Thus, a @f$42 \times 42 \times 42@f$ cubic kilometre %CSG
       *   box primitive becomes @f$6 \times 6 \times 6@f$ when using
       *   subvolumes of size @f$7@f$ kilometres.  A complete
       *   MesoRD::UnitTable is needed in order to resolve the units
       *   in which the tree is defined.</li>
       *
       * </ul>
       *
       * The function returns a pointer to the parsed root node of the
       * new %CSG tree.
       *
       * @attention When first calling this function, the
       * transformation should be set to the identity transformation,
       * <i>i.e.</i> the four-by-four identity matrix.
       *
       * @todo Perhaps the @f$4 \times 4@f$ identity matrix should be
       * a default value for @p T.
       *
       * @param a    The root node of the XML %CSG tree definition.
       * @param ut   The table of unit definitions.
       * @param size The size of a subvolume in metres.
       * @param T    The transformation matrix of the node.
       * @return     The root node of the parsed tree.
       */
    Node*
    fromXML(
      const Annotation* a, const UnitTable* ut, const double size, Matrix T);


      /**
       * @brief Put the forest in POV-Ray format onto an output
       * stream.
       *
       * This function accomplishes the same thing as C++'s <i>put
       * to</i>-operator, <code>&lt;&lt;</code>, except that a node is
       * written in POV-Ray format.  The reason for its existence is
       * purely technical; it makes the implementation a little nicer.
       * Internally, this function just writes a POV-Ray preamble and
       * then calls Node::putPOVRay() for each node in the forest.
       * The preamble ensures that file is a valid, stand-alone
       * POV-Ray file.
       *
       * @param forest The forest which to output.
       * @param os     The output stream to write to.
       * @return       The output stream once written to.
       */
    std::ostream&
    putPOVRay(const std::map<std::string, Node*>& forest, std::ostream& os);


      /**
       * @brief Put the node in XML format onto an output stream.
       *
       * This function accomplishes the same thing as C++'s <i>put
       * to</i>-operator, <code>&lt;&lt;</code>, except that a node is
       * written in XML format.  The reason for its existence is
       * purely technical; it makes the implementation a little nicer.
       * Internally, this function just calls Node::putXML() for each
       * node in the forest.
       *
       * @param forest The forest which to output.
       * @param os     The output stream to write to.
       * @return       The output stream once written to.
       */
    inline std::ostream&
    putXML(const std::map<std::string, Node*>& forest, std::ostream& os)
    {
      std::map<std::string, Node*>::const_iterator i;
      for (i = forest.begin(); i != forest.end(); i++) {
        if (i -> second)
          i -> second -> putXML(os);
      }
      return os;
    }


      /**
       * @brief Realise the geometry as defined by a set of %CSG
       * trees.
       *
       * To build the system geometry from a forest of %CSG trees, one
       * needs to know what kind, or @p type, of subvolumes should be
       * used.  One does <em>not</em> need to know how big the
       * subvolumes are, because the trees, once passed to this
       * function, should be scaled such that they are in units of the
       * size of a single subvolume.  The forest, that is the union of
       * the trees, is supplied in an STL <code>map</code>, keyed on
       * some textual identifier of the trees.  An optional progress
       * bar can be passed along if progress monitoring is desired.
       *
       * The realise() function fills the volume defined by the
       * forest's %CSG procedure.  This volume must, by definition, be
       * included in the procedure's bounding box.  Internally, the
       * space inside the bounding box is interpreted in terms of an
       * integer, grid coordinate system.  The integer grid coordinate
       * system is related to the scaled coordinate system of the
       * individual trees by a scale factor only.  For every grid
       * point which is contained in any of the trees of the %CSG
       * forest, a subvolume is constructed.  The generated subvolume
       * is assigned to one, and only one, tree -- the first tree in
       * the forest that contains the grid point.  One should
       * <em>not</em> think one knows which tree is the first in the
       * forest.
       *
       * The realised geometries are <em>not</em>
       * <code>return</code>:ed, but are similarly stored in the
       * supplied STL <code>map</code>, @p geometries.  The reason is
       * that the copying that would be caused by
       * <code>return</code>:ing them could lead to significant
       * overhead.  The built geometries will be keyed on the same
       * identifiers that were used to key the trees: all subvolumes
       * in the vector at <code>geometries["foo"]</code> arose from
       * the tree in <code>forest["foo"]</code>.
       *
       * @warning It is <em>very important</em> to note that the
       * subvolumes constructed by this function will not have their
       * sizes set correctly!  One must make sure to set up those
       * sizes before attempting to use the geometries in simulation.
       * This function cannot be "fixed" to set up the size of the
       * subvolumes, because it internally works with a coordinate
       * system scaled to subvolume size -- the absolute subvolume
       * size is completely unknown to realise().
       *
       * @warning Furthermore, once subvolumes come out of this
       * function, they will have <em>no idea</em> of which
       * MesoRD::Compartment they belong to!  Knowing its
       * MesoRD::Compartment is important for a MesoRD::SubVolume,
       * because it allows retrieval of the right reaction and
       * diffusion constants.  Normally, a MesoRD::SubVolume knows its
       * MesoRD::Compartment, because it holds some sort of reference
       * to it.  Since realise() only deals with instances of
       * MesoRD::Compartment by some textual identifier through keys
       * in the @p forest STL <code>map</code>, no such references can
       * be resolved by realise().  Therefore, one must take the
       * geometries tree-wise as returned by realise() and make sure
       * every MesoRD::SubVolume knows its containing
       * MesoRD::Compartment.
       *
       * @param forest      The %CSG trees for each compartment.
       * @param type        The type of subvolume which should be used
       *                    for building.
       * @param geometries  A map which will hold the built geometries.
       * @param progressBar An optional progress bar.
       * @param prng        Pointer to the pseudo-random number
       *                    generator.
       */
    void
    realise(
      const std::map<std::string, Node*>&              forest,
      const Subvolume_t                                type,
      std::map<std::string, std::vector<SubVolume*> >& geometries,
      ProgressBar*                                     progressBar,
      Random::Random*                                  prng);


      /**
       * @brief Resolve %CSG tree references.
       *
       * %CSG trees can contain CSG::Compartment nodes, which prompt
       * the inclusion of another %CSG tree.  The fromXML() function
       * builds the %CSG forest one tree at the time, so while
       * building one tree it is generally impossible to obtain a
       * reference to another.  Therefore, the build process leaves
       * any references to included trees unresolved, and only stores
       * an identifier to the tree which should have been included.
       * The resolve() function addresses this problem.  Calling it on
       * the forest will resolve any CSG::Compartment inclusions left
       * dangling by the first building pass.
       *
       * There are two things that can go wrong:
       *
       * <ul>
       *
       *   <li>If one references non-existent tree, this function will
       *   throw an UndefinedCompartment exception.</li>
       *
       *   <li>If there are cycles in the inclusion graph, the
       *   function throws a CyclicInclusion exception.</li>
       *
       * </ul>
       *
       * @param forest The forest to resolve.
       */
    void
    resolve(std::map<std::string, Node*>& forest)
      throw(CyclicInclusion, UndefinedCompartment);
  }
}
#endif /* ifndef MESORD_CSG_CSG_H */
