/* The CSGNode class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSGConnectivity.cpp 555 2008-06-08 18:07:05Z davidfange $
 */
#include <cmath>

#include "CSGNode.hpp"
#include "../linalg/Vector.hpp"



  /* We consider one dimension at a time.  Only dimensions which have
   * non-zero displacement components <em>and</em> periodic boundary
   * conditions are of interest.  If it is found that the point itself
   * lies on a boundary <em>and</em> that the point with the
   * directional displacement is an exterior point, the displacement
   * is wrapped.  Wrapping means subtracting the side length from the
   * displacement vector in the dimension we are considering.  If we
   * have found the new neighbour, the wrapped displacement vector has
   * to point to a boundary point.
   */
MesoRD::Vector
MesoRD::CSG::Box::neighbour(const Vector& p, const Vector& d) const
{
  const Vector q = p + d;

  if (contains(p).type == Containment::boundary
    && contains(q).type == Containment::exterior) {
    Vector n = q;

    for (size_t i = 0; i < 3; i++) {
      if (pbc[i] && fabs(d(static_cast<unsigned int>(i)))
	  > std::numeric_limits<double>::min()) {
        const double sign =
	  d(static_cast<unsigned int>(i)) > 0.0 ? -1.0 : +1.0;
        n(static_cast<unsigned int>(i)) += side[i] * sign;
      }
    }

    if (contains(n).type == Containment::boundary)
      return n;
  }

  return q;
}


  /* The key difference between cones and cylinders is that cylinders
   * can have periodic boundary conditions.  The periodic boundary
   * condition must however always be along the <code>y</code>-axis.
   * This code works just like the code for the box, although only one
   * dimension needs to be considered.
   */
MesoRD::Vector
MesoRD::CSG::Cylinder::neighbour(const Vector& p, const Vector& d) const
{
  const Vector q = p + d;

  if (contains(p).type == Containment::boundary
    && contains(q).type == Containment::exterior) {
    Vector n = q;

    if (pbc && fabs(d(1)) > std::numeric_limits<double>::min()) {
      const double sign = d(1) > 0.0 ? -1.0 : +1.0;
      n(1) += height * sign;
    }

    if (contains(n).type == Containment::boundary)
      return n;
  }

  return q;
}
