/* The CSGNode class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSGGeometry.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <algorithm>
#include <cmath>
#include <limits>

#include "CSGNode.hpp"
#include "../mesord.hpp"
#include "../linalg/Matrix.hpp"
#include "../linalg/Vector.hpp"


  /*****************
   * FILE CONTENTS *
   *****************/

  /* This file contains the definitions for the following functions of
   * all classes derived from the Node base class:
   *
   * <ul>
   *
   *   <li>The contains() function.  In general for these functions, q
   *   denotes the point p when transformed into the coordinate system
   *   of the primitive.</li>
   *
   *   <li>The getBoundingBox() function.</li>
   *
   * The functions are commented along the way.
   */



  /*************************
   * THE CONTAINS FUNCTION *
   *************************/

  /* Checks if a point lies inside the box.  A point is inside the box
   * if, and only if, its distance from the origin along each axis is
   * less than half the side length along that axis.  A point is a
   * boundary point if it lies inside the box, and lies on any face
   * for which periodic boundary conditions are active.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Box::contains(const Vector& p) const
{
  Vector q(4);
  invert(p, q);

  if (fabs(q(0))  > 0.5 * side[0]
    || fabs(q(1)) > 0.5 * side[1]
    || fabs(q(2)) > 0.5 * side[2])
    return Containment(false);

  for (size_t i = 0; i < 3; i++) {
    if (pbc[i] && 0.5 * side[i] - fabs(q(static_cast<unsigned int>(i))) < 1)
      return Containment(this);
  }

  return Containment(true);
}


  /* Checks if a point lies inside the compartment.  We just hand all
   * the work over to the referred compartment after transforming the
   * point vector.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Compartment::contains(const Vector& p) const
{
  if (compartment) {
    Vector q(4);
    invert(p, q);
    return compartment -> contains(q);
  }
  return Containment(false);
}


  /* Checks if a point lies inside the cone.  A point is inside if and
   * only if the <code>y</code>-value is within the height of the cone
   * and the distance from the <code>y</code>-axis is within the
   * radius of the cone at that point.  The radius of the cone
   * decreases linearly from <code>radius</code> at the base, to zero
   * at the cap.  Since we do not care about boundary points on a
   * cone, this function will only return exterior or interior.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Cone::contains(const Vector& p) const
{
  Vector q(4);
  invert(p, q);
  const double y = q(1);

  if (y >= -0.5 * height && y <= +0.5 * height) {
    const double d = q(0) * q(0) + q(2) * q(2);
    const double r = radius * (1.0 - (y / height + 0.5));

    if (d <= r * r)
      return Containment(true);
  }

  return Containment(false);
}


  /* Checks if a point lies inside the cylinder.  A point is inside if
   * and only if the <code>y</code>-value is within the height of the
   * cylinder and the distance from the <code>y</code>-axis is within
   * <code>radius</code>.  If periodic boundary conditions are active,
   * any point on either the base or the cap will be flagged as a
   * boundary point.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Cylinder::contains(const Vector& p) const
{
  Vector q(4);
  invert(p, q);
  const double y = q(1);

  if (y >= -0.5 * height && y <= +0.5 * height) {
    const double d = q(0) * q(0) + q(2) * q(2);

    if (d <= radius * radius) {
      if (pbc && 0.5 * height - fabs(y) < 1.0)
        return Containment(this);
      else
        return Containment(true);
    }
  }

  return Containment(false);
}


  /* Checks if a point lies inside the difference.  A point is inside
   * if and only if it is inside the volume on the left hand side
   * <em>and</em> outside the volume on the right hand side.  Thus,
   * the node that contains the point is always the left node.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Difference::contains(const Vector& p) const
{
  Node* left  = operands.front();
  Node* right = operands.back();

  if (left && right && right -> contains(p).type == Containment::exterior)
    return left -> contains(p);
  return Containment(false);
}


  /* Checks if a point lies inside the intersection.  A point is
   * inside if and only if it is inside <em>all</em> the volumes in
   * the list.  If at least one of the operands flags the point as a
   * boundary point, the intersection of the operands will also flag
   * the point as a boundary point.  Otherwise, the point is an
   * interior point.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Intersection::contains(const Vector& p) const
{
  Containment containment(false);
  std::list<Node*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (*i) {
      switch ((*i) -> contains(p).type) {

        case Containment::boundary:
          if (containment.type != Containment::boundary)
            containment = this;
          break;

        case Containment::exterior:
          return Containment(false);

        case Containment::interior:
          if (containment.type != Containment::boundary)
            containment = true;
          break;
      }
    }
  }

  return containment;
}


  /* Checks if a point lies inside the mesh.  For every axis, we keep
   * a list of intersections with the primitive mesh.  For instance,
   * for the <code>x</code>-axis, there is a list called
   * <code>x_intersections</code>, which for every two-dimensional
   * coordinate in the <code>yz</code>-plane lists the intervals of
   * <code>x</code>-values that lie inside the mesh.  The first part
   * of the function below is concerned with determining these lists.
   * In the second part, we just check whether the point is found in
   * <em>any</em> of the lists.  If it is, then the point lies inside
   * the mesh.  A breakdown of the first part follows:
   *
   * <ul>
   *
   *   <li>Computing the intersection of a triangle and a ray parallel
   *   to any of the coordinate axes essentially involves computing
   *   the * inverse of a two-by-two matrix.  If the inverse exists,
   *   we know * that the ray intersects the <em>plane</em> of the
   *   triangle.  In * cases of intersection, the <code>a</code>- and
   *   * <code>b</code>-values define the intersection point in the *
   *   coordinate system defined by the two vectors of the triangle, *
   *   <code>u</code> = <code>p1</code> - <code>p0</code> and *
   *   <code>v</code> = <code>p2</code> - <code>p0</code>.  If *
   *   <code>a</code> and <code>b</code> lie within the triangle
   *   spanned * by the unit distance along the latter vectors, the
   *   point is not * only in the plane, but also inside the
   *   triangle.</li>
   *
   *   <li>If the angle the intersecting ray makes with the normal of
   *   triangle is sufficiently negative, we know that we are facing
   *   the * triangle.  Requiring the angle to be more than less than
   *   zero, is * a good thing, because it is difficult to tell
   *   whether we see the * front or the back of a triangle in cases
   *   where the angle the * intersecting ray makes with the normal is
   *   close to zero.</li>
   *
   *   <li>A new segment is started as soon as we intersect the front
   *   of a triangle.  The segment is ended as soon as we see another
   *   * intersection.  Because while we are testing triangles, *
   *   intersections may pop up in any order, we store front *
   *   intersections and back intersections in lists.  The lists are *
   *   post-processed in a second pass to find the containing
   *   segments.  * Because we want to be as conservative as possible,
   *   an * intersection point is always treated as seeing the
   *   backside of a * triangle.</li>
   *
   * </ul>
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Mesh::contains(const Vector& p) const
{
  Vector q(4);
  invert(p, q);

  std::pair<int, int> xy(
    static_cast<int>(lrint(q(0))), static_cast<int>(lrint(q(1))));
  std::pair<int, int> xz(
    static_cast<int>(lrint(q(0))), static_cast<int>(lrint(q(2))));
  std::pair<int, int> yz(
    static_cast<int>(lrint(q(1))), static_cast<int>(lrint(q(2))));

  const bool missing_xy = z_intersections.find(xy) == z_intersections.end();
  const bool missing_xz = y_intersections.find(xz) == y_intersections.end();
  const bool missing_yz = x_intersections.find(yz) == x_intersections.end();

    /* The cosine of the largest angle the ray is allowed to make with
     * the normal, in order to still count as intersecting the front
     * face of the triangle.
     */
  const double angle = 0.5;

    /* Compute ray-triangle intersections and store front and back
     * intersection points.
     */
  if (missing_xy || missing_xz || missing_yz) {
    std::list<double> x_b;
    std::list<double> x_f;
    std::list<double> y_b;
    std::list<double> y_f;
    std::list<double> z_b;
    std::list<double> z_f;

    std::list<Triangle>::const_iterator i;

    for (i = triangles.begin(); i != triangles.end(); i++) {
      const double u_x = i -> p1(0) - i -> p0(0);
      const double u_y = i -> p1(1) - i -> p0(1);
      const double u_z = i -> p1(2) - i -> p0(2);

      const double v_x = i -> p2(0) - i -> p0(0);
      const double v_y = i -> p2(1) - i -> p0(1);
      const double v_z = i -> p2(2) - i -> p0(2);

      if (missing_xy) {
        const double d = u_x * v_y - u_y * v_x;

        if (fabs(d) > std::numeric_limits<double>::min()) {
          const double c_x = xy.first  - i -> p0(0);
          const double c_y = xy.second - i -> p0(1);
          const double c_z = min(2)    - i -> p0(2);

          const double a   = (v_y * c_x - v_x * c_y) / d;
          const double b   = (u_x * c_y - u_y * c_x) / d;

          if (a >= 0.0 && a <= 1.0 && b >= 0.0 && b <= 1.0 && a + b <= 1.0) {
            const double n_x = u_y * v_z - u_z * v_y;
            const double n_y = u_z * v_x - u_x * v_z;
            const double n_z = u_x * v_y - u_y * v_x;
            const double n_n = sqrt(n_x * n_x + n_y * n_y + n_z * n_z);
            const double t   = a * u_z + b * v_z - c_z;

            if (n_z < -1.0 * angle * n_n)
              z_f.push_back(min(2) + t);
            z_b.push_back(min(2) + t);
          }
        }
      }

      if (missing_xz) {
        const double d = u_x * v_z - u_z * v_x;

        if (fabs(d) > std::numeric_limits<double>::min()) {
          const double c_x = xz.first  - i -> p0(0);
          const double c_y = min(1)    - i -> p0(1);
          const double c_z = xz.second - i -> p0(2);

          const double a   = (v_z * c_x - v_x * c_z) / d;
          const double b   = (u_x * c_z - u_z * c_x) / d;

          if (a >= 0.0 && a <= 1.0 && b >= 0.0 && b <= 1.0 && a + b <= 1.0) {
            const double n_x = u_y * v_z - u_z * v_y;
            const double n_y = u_z * v_x - u_x * v_z;
            const double n_z = u_x * v_y - u_y * v_x;
            const double n_n = sqrt(n_x * n_x + n_y * n_y + n_z * n_z);
            const double t   = a * u_y + b * v_y - c_y;

            if (n_y < -1.0 * angle * n_n)
              y_f.push_back(min(1) + t);
            y_b.push_back(min(1) + t);
          }
        }
      }

      if (missing_yz) {
        const double d = u_y * v_z - u_z * v_y;

        if (fabs(d) > std::numeric_limits<double>::min()) {
          const double c_x = min(0)    - i -> p0(0);
          const double c_y = yz.first  - i -> p0(1);
          const double c_z = yz.second - i -> p0(2);

          const double a   = (v_z * c_y - v_y * c_z) / d;
          const double b   = (u_y * c_z - u_z * c_y) / d;

          if (a >= 0.0 && a <= 1.0 && b >= 0.0 && b <= 1.0 && a + b <= 1.0) {
            const double n_x = u_y * v_z - u_z * v_y;

            const double n_y = u_z * v_x - u_x * v_z;
            const double n_z = u_x * v_y - u_y * v_x;
            const double n_n = sqrt(n_x * n_x + n_y * n_y + n_z * n_z);
            const double t   = a * u_x + b * v_x - c_x;

            if (n_x < -1.0 * angle * n_n)
              x_f.push_back(min(0) + t);
            x_b.push_back(min(0) + t);
          }
        }
      }
    }

      /* Transform the lists of front and back intersection points to
       * segments, which indicate containment inside the mesh.
       */
    if (missing_xy) {
      std::list<std::pair<double, double> > z_segments;

      while (!z_f.empty() && !z_b.empty()) {
        double last_back;
        const double last_front = z_f.front();

        do {
          last_back = z_b.front();
          z_b.pop_front();
        } while (!z_b.empty() && last_front >= last_back);

        do {
          z_f.pop_front();
        } while (!z_f.empty() && z_f.front() < last_back);

        if (last_front < last_back)
          z_segments.push_back(
            std::pair<double, double>(last_front, last_back));
      }

      z_intersections[xy] = z_segments;
    }

    if (missing_xz) {
      std::list<std::pair<double, double> > y_segments;

      while (!y_f.empty() && !y_b.empty()) {
        double last_back;
        const double last_front = y_f.front();

        do {
          last_back = y_b.front();
          y_b.pop_front();
        } while (!y_b.empty() && last_front >= last_back);

        do {
          y_f.pop_front();
        } while (!y_f.empty() && y_f.front() < last_back);

        if (last_front < last_back)
          y_segments.push_back(
            std::pair<double, double>(last_front, last_back));
      }

      y_intersections[xz] = y_segments;
    }

    if (missing_yz) {
      std::list<std::pair<double, double> > x_segments;

      while (!x_f.empty() && !x_b.empty()) {
        double last_back;
        const double last_front = x_f.front();

        do {
          last_back = x_b.front();
          x_b.pop_front();
        } while (!x_b.empty() && last_front >= last_back);

        do {
          x_f.pop_front();
        } while (!x_f.empty() && x_f.front() < last_back);

        if (last_front < last_back)
          x_segments.push_back(
            std::pair<double, double>(last_front, last_back));
      }

      x_intersections[yz] = x_segments;
    }
  }

    /* Second part begins here.  We check all the lists and report the
     * point as being contained in the mesh if it lies within any of
     * the segments in the <code>x</code>-, <code>y</code>- or
     * <code>z</code>-axis.
     */
  bool in_x = false;
  bool in_y = false;
  bool in_z = false;

  std::map<std::pair<int, int>,
    std::list<std::pair<double, double> > >::const_iterator i;

  if ((i = z_intersections.find(xy)) != z_intersections.end()) {
    std::list<std::pair<double, double> >::const_iterator j;

    for (j = i -> second.begin(); j != i -> second.end(); j++) {
      if (q(2) >= j -> first && q(2) <= j -> second) {
        in_z = true;
        break;
      }
    }
  }

  if ((i = y_intersections.find(xz)) != y_intersections.end()) {
    std::list<std::pair<double, double> >::const_iterator j;

    for (j = i -> second.begin(); j != i -> second.end(); j++) {
      if (q(1) >= j -> first && q(1) <= j -> second) {
        in_y = true;
        break;
      }
    }
  }

  if ((i = x_intersections.find(yz)) != x_intersections.end()) {
    std::list<std::pair<double, double> >::const_iterator j;

    for (j = i -> second.begin(); j != i -> second.end(); j++) {
      if (q(0) >= j -> first && q(0) <= j -> second) {
        in_x = true;
        break;
      }
    }
  }

  return in_x || in_y || in_z ? Containment(true) : Containment(false);
}


  /* Checks if a point lies inside the sphere.  A point is inside if
   * and only if the distance from the origin to the point is within
   * the radius of the sphere.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Sphere::contains(const Vector& p) const
{
  Vector q(4);
  invert(p, q);
  const double r2 = q(0) * q(0) + q(1) * q(1) + q(2) * q(2);

  if (r2 <= radius * radius)
    return Containment(true);
  return Containment(false);
}


  /* Checks if a point lies inside the union.  A point is inside if
   * and only if it is inside <em>any</em> of the volumes in the
   * list.  If at least one of the operands flags the point as a
   * boundary point, the union of the operands will also flag the
   * point as a boundary point.  Otherwise, the point is an interior
   * point.
   */
const MesoRD::CSG::Node::Containment
MesoRD::CSG::Union::contains(const Vector& p) const
{
  Containment containment(false);
  std::list<Node*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (*i) {
      switch ((*i) -> contains(p).type) {

        case Containment::boundary:
          if (containment.type != Containment::boundary)
            containment = this;
          break;

        case Containment::exterior:
          break;

        case Containment::interior:
          if (containment.type != Containment::boundary)
            containment = true;
          break;
      }
    }
  }

  return containment;
}



  /*******************************
   * THE GETBOUNDINGBOX FUNCTION *
   *******************************/

  /* Compute the bounding box of a box.  In the outer for-loop, we
   * find the the eight corners of the box.  These can be computed as
   * all possible combinations of either -0.5 or +0.5 times the side
   * length.  After transformation of the corner, we find the maximum
   * and minimum values along all of the axes.  The computations for
   * the three axes is done in the inner for-loop.
   */
void
MesoRD::CSG::Box::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  Matrix V = U * T;
  Vector corner(4);

  max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();

  for (size_t i = 0; i < 8; i++) {
    corner(0) = (i / 4 % 2 == 0 ? -0.5 : +0.5) * side[0];
    corner(1) = (i / 2 % 2 == 0 ? -0.5 : +0.5) * side[1];
    corner(2) = (i / 1 % 2 == 0 ? -0.5 : +0.5) * side[2];
    corner(3) = 1.0;
    corner    = V * corner;

    for (size_t j = 0; j < 3; j++) {
      max(static_cast<unsigned int>(j)) =
	MesoRD::max(max(static_cast<unsigned int>(j)),
		    corner(static_cast<unsigned int>(j)));
      min(static_cast<unsigned int>(j)) =
	MesoRD::min(min(static_cast<unsigned int>(j)),
		    corner(static_cast<unsigned int>(j)));
    }
  }
}


  /* Compute the bounding box of a compartment.  We just hand all the
   * work over to the referred compartment and pass along the local
   * transformation matrix for pre-multiplication.
   *
   * @todo This function should probably be inlined.
   */
void
MesoRD::CSG::Compartment::getBoundingBox(
  Vector& max, Vector& min, const Matrix&) const
{
  compartment -> getBoundingBox(max, min, T);
}


  /* Compute the bounding box of a cone.  The cone extends its radius
   * in the <code>x</code>- and <code>y</code>-direction and spans its
   * height on the <code>z</code>-axis, so that equally big regions
   * are covered on both sides of the <code>xy</code>-plane.  In this
   * function, we transform the cap and the equation for the circle,
   * which defines the base of the cone.  We then look for extrema on
   * the base.  Extrema may occur where either the sines or cosines
   * take extreme values, or where the differentiated base transform
   * is zero.  The bounding box is defined by whatever box contains
   * the transformed cap and the extrema of the transformed base
   * circle.
   */
void
MesoRD::CSG::Cone::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  const Matrix V   = U * T;
  const double y   = -0.5 * height;
  Vector       cap(4);

  cap(0) = 0.0;
  cap(1) = +0.5 * height;
  cap(2) = 0.0;
  cap(3) = 1.0;
  cap    = V * cap;

  for (size_t i = 0; i < 3; i++) {
    double a = 0;
    if (fabs(V(static_cast<unsigned int>(i), 2))
	< std::numeric_limits<double>::min()) {
      if (V(static_cast<unsigned int>(i), 0) < 0) {
        a = -M_PI / 2;
      } else {
        if (V(static_cast<unsigned int>(i), 0) > 0)
          a = M_PI /2;
        else
          a = 0;
      }
    } else {
      a = atan(V(static_cast<unsigned int>(i), 0)
	       / V(static_cast<unsigned int>(i), 2));
    }

    max(static_cast<unsigned int>(i)) =
      min(static_cast<unsigned int>(i)) =
      cap(static_cast<unsigned int>(i));
    for (size_t j = 0; j < 6; j++) {
      double c;
      double phi;

      switch (j) {

        case 0:
          phi = 0.0 * M_PI;
          break;

        case 1:
          phi = 0.5 * M_PI;
          break;

        case 2:
          phi = 1.0 * M_PI;
          break;

        case 3:
          phi = 1.5 * M_PI;
          break;

        case 4:
          phi = fmod(a + 2.0 * M_PI, 2.0 * M_PI);
          break;

        case 5:
          phi = fmod(a + 3.0 * M_PI, 2.0 * M_PI);
          break;
      }

      c = V(static_cast<unsigned int>(i), 0) * radius * cos(phi)
        + V(static_cast<unsigned int>(i), 1) * y
        + V(static_cast<unsigned int>(i), 2) * radius * sin(phi)
        + V(static_cast<unsigned int>(i), 3);

      max(static_cast<unsigned int>(i)) =
	MesoRD::max(max(static_cast<unsigned int>(i)), c);
      min(static_cast<unsigned int>(i)) =
	MesoRD::min(min(static_cast<unsigned int>(i)), c);
    }
  }
}


  /* Compute the bounding box of a cylinder.  The cylinder extends its
   * radius in the <code>x</code>- and <code>y</code>-direction and
   * spans its height on the <code>z</code>-axis, so that equally big
   * regions are covered on both sides of the <code>xy</code>-plane.
   * In this function, we transform the equations for the circles,
   * which define the base and cap of the cylinder.  We then look for
   * extrema.  Extrema may occur where either the sines or cosines
   * take extreme values, or where the differentiated circle
   * transforms are zero.  The bounding box is defined by whatever box
   * contains the transformed base and cap circles.
   */
void
MesoRD::CSG::Cylinder::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  const Matrix V = U * T;

  max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();

  for (size_t i = 0; i < 2; i++) {
    const double y = (i % 2 == 0 ? -0.5 : +0.5) * height;

    for (size_t j = 0; j < 3; j++) {
      
      double a = 0;
      if (fabs(V(static_cast<unsigned int>(i), 2))
	  < std::numeric_limits<double>::min()) {
        if (V(static_cast<unsigned int>(i), 0) < 0) {
          a = -M_PI / 2;
        } else {
          if (V(static_cast<unsigned int>(i), 0) > 0)
            a = M_PI /2;
          else
            a = 0;
        }
      }
      else {
        a = atan(V(static_cast<unsigned int>(i), 0)
		 / V(static_cast<unsigned int>(i), 2));
      }

      for (size_t k = 0; k < 6; k++) {
        double c;
        double phi;

        switch (k) {

          case 0:
            phi = 0.0 * M_PI;
            break;

          case 1:
            phi = 0.5 * M_PI;
            break;

          case 2:
            phi = 1.0 * M_PI;
            break;

          case 3:
            phi = 1.5 * M_PI;
            break;

          case 4:
            phi = fmod(a + 2.0 * M_PI, 2.0 * M_PI);
            break;

          case 5:
            phi = fmod(a + 3.0 * M_PI, 2.0 * M_PI);
            break;
        }

        c = V(static_cast<unsigned int>(j), 0) * radius * cos(phi)
          + V(static_cast<unsigned int>(j), 1) * y
          + V(static_cast<unsigned int>(j), 2) * radius * sin(phi)
          + V(static_cast<unsigned int>(j), 3);

        max(static_cast<unsigned int>(j)) =
	  MesoRD::max(max(static_cast<unsigned int>(j)), c);
        min(static_cast<unsigned int>(j)) =
	  MesoRD::min(min(static_cast<unsigned int>(j)), c);
      }
    }
  }
}


  /* Compute the bounding box of a difference operation.  As discussed
   * in the documentation, the bounding box of a difference operation
   * is always the bounding box of the left operand.
   */
void
MesoRD::CSG::Difference::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  Node* left = operands.front();
  if (left)
    left -> getBoundingBox(max, min, U);
}


  /* Compute the bounding box of an intersection operation.  The
   * bounding box of an intersection is the smallest bounding box we
   * can possibly construct from all the operands.  This is very
   * related to computing the bounding box for the union set
   * operation.
   */
void
MesoRD::CSG::Intersection::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  std::list<Node*>::const_iterator i;

  max(0) = max(1) = max(2) = +1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = -1.0 * std::numeric_limits<double>::max();

  for (i = operands.begin(); i != operands.end(); i++) {
    if (*i) {
      Vector local_max(3);
      Vector local_min(3);
      (*i) -> getBoundingBox(local_max, local_min, U);

      for (size_t j = 0; j < 3; j++) {
        max(static_cast<unsigned int>(j)) =
	  MesoRD::min(max(static_cast<unsigned int>(j)),
		      local_max(static_cast<unsigned int>(j)));
        min(static_cast<unsigned int>(j)) =
	  MesoRD::max(min(static_cast<unsigned int>(j)),
		      local_min(static_cast<unsigned int>(j)));
      }
    }
  }
}


  /* Compute the bounding box of a mesh.  In the outer for-loop, we
   * iterate through all triangles that make up the mesh.  Each vertex
   * of every triangle is transformed according to the transformation
   * matrix of the node.  In the inner for loop, we check all three
   * transformed vertices for their maximum and minimum values
   * respectively.
   */
void
MesoRD::CSG::Mesh::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  Matrix V = U * T;
  std::list<Triangle>::const_iterator i;

  max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();

  for (i = triangles.begin(); i != triangles.end(); i++) {
    Vector v0 = V * i -> p0;
    Vector v1 = V * i -> p1;
    Vector v2 = V * i -> p2;

    for (size_t j = 0; j < 3; j++) {
      max(static_cast<unsigned int>(j)) =
	MesoRD::max(max(static_cast<unsigned int>(j)),
		    v0(static_cast<unsigned int>(j)));
      max(static_cast<unsigned int>(j)) =
	MesoRD::max(max(static_cast<unsigned int>(j)),
		    v1(static_cast<unsigned int>(j)));
      max(static_cast<unsigned int>(j)) =
	MesoRD::max(max(static_cast<unsigned int>(j)),
		    v2(static_cast<unsigned int>(j)));

      min(static_cast<unsigned int>(j)) =
	MesoRD::min(min(static_cast<unsigned int>(j)),
		    v0(static_cast<unsigned int>(j)));
      min(static_cast<unsigned int>(j)) =
	MesoRD::min(min(static_cast<unsigned int>(j)),
		    v1(static_cast<unsigned int>(j)));
      min(static_cast<unsigned int>(j)) =
	MesoRD::min(min(static_cast<unsigned int>(j)),
		    v2(static_cast<unsigned int>(j)));
    }
  }
}


  /* Compute the bounding box of a sphere.  We transform the equation
   * of the sphere and look for extrema.  Extrema may occur where
   * either the sines or cosines take extreme values, or where the
   * differentiated sphere transform is zero.  We handle each
   * coordinate separately in the outer <code>for</code>-loop.  The
   * next <code>for</code>-loop checks the "singular" values for the
   * <code>phi</code> angle, and the innermost <code>for</code>-loop
   * handles the <code>theta</code>-angle as well as combinations
   * which yield a derivative of zero.  The bounding box is defined by
   * whatever box contains the transformed sphere.
   */
void
MesoRD::CSG::Sphere::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  const Matrix V = U * T;

  max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();

  for (size_t i = 0; i < 3; i++) {
    double a = 0;
    if (fabs(V(static_cast<unsigned int>(i), 1))
	< std::numeric_limits<double>::min()) {
      if (V(static_cast<unsigned int>(i), 0) < 0) {
        a = - M_PI / 2;
      } else {
        if (V(static_cast<unsigned int>(i),0) > 0)
          a = M_PI /2;
        else
          a = 0;
      }
    } else {
      a = atan(V(static_cast<unsigned int>(i), 0)
	       / V(static_cast<unsigned int>(i), 1));
    }

    for (size_t j = 0; j < 4; j++) {
      double phi;

      switch (j) {

        case 0:
          phi = 0.0 * M_PI;
          break;

        case 1:
          phi = 0.5 * M_PI;
          break;

        case 2:
          phi = 1.0 * M_PI;
          break;

        case 3:
          phi = 1.5 * M_PI;
          break;
      }

      for (size_t k = 0; k < 5; k++) {
        double c;
        double d;
        double theta;

        switch (k) {

          case 0:
            theta = 0.0 * M_PI;
            break;

          case 1:
            theta = 0.5 * M_PI;
            break;

          case 2:
            theta = 1.0 * M_PI;
            break;

          case 3:
            if (j != 0)
              continue;
            phi = fmod(a + 2.0 * M_PI, 2.0 * M_PI);
            d   = V(static_cast<unsigned int>(i), 0)
	      * cos(phi) * V(static_cast<unsigned int>(i), 1) * sin(phi);
            if (fabs(d) < std::numeric_limits<double>::min()) {
              if (V(static_cast<unsigned int>(i), 2) < 0) {
                theta = fmod(-M_PI / 2.0 + 2.0 * M_PI, 2.0 * M_PI);
              } else {
                if (V(static_cast<unsigned int>(i), 2) > 0)
                  theta = fmod(M_PI / 2.0 + 2.0 * M_PI, 2.0 * M_PI);
                else
                  theta = fmod(0 + 2.0 * M_PI, 2.0 * M_PI);
              }
            } else {
              theta = fmod(atan(V(static_cast<unsigned int>(i), 2)
				/ (V(static_cast<unsigned int>(i), 0)
				   * cos(phi)
				   * V(static_cast<unsigned int>(i), 1)
				   * sin(phi)))
                + 2.0 * M_PI, 2.0 * M_PI);              
            }
            break;

          case 4:
            if (j != 0)
              continue;
            phi = fmod(a + 3.0 * M_PI, 2.0 * M_PI);
            d   = V(static_cast<unsigned int>(i), 0)
	      * cos(phi) * V(static_cast<unsigned int>(i), 1) * sin(phi);
            if (fabs(d) < std::numeric_limits<double>::min()) {
              if (V(static_cast<unsigned int>(i), 2) < 0) {
                theta = fmod(-M_PI / 2.0 +  2.0 * M_PI, 2.0 * M_PI);
              } else {
                if (V(static_cast<unsigned int>(i), 2) > 0)
                  theta = fmod(M_PI / 2.0 + 2.0 * M_PI, 2.0 * M_PI);
                else
                  theta = fmod(0 + 2.0 * M_PI, 2.0 * M_PI);
              }
            } else {
              theta = fmod(atan(V(static_cast<unsigned int>(i), 2)
				/ (V(static_cast<unsigned int>(i), 0)
				   * cos(phi)
				   * V(static_cast<unsigned int>(i), 1)
				   * sin(phi)))
                + 2.0 * M_PI, 2.0 * M_PI);              
            }
            break;
        }

        c = V(static_cast<unsigned int>(i), 0) * radius * sin(theta) * cos(phi)
          + V(static_cast<unsigned int>(i), 1) * radius * sin(theta) * sin(phi)
          + V(static_cast<unsigned int>(i), 2) * radius * cos(theta)
          + V(static_cast<unsigned int>(i), 3);

        max(static_cast<unsigned int>(i)) =
	  MesoRD::max(max(static_cast<unsigned int>(i)), c);
        min(static_cast<unsigned int>(i)) =
	  MesoRD::min(min(static_cast<unsigned int>(i)), c);
      }
    }
  }
}


  /* Compute the bounding box of a union operation.  The bounding box
   * of a union is the largest bounding box we can possibly construct
   * from all the operands.  This is very related to computing the
   * bounding box for the intersection set operation.
   */
void
MesoRD::CSG::Union::getBoundingBox(
  Vector& max, Vector& min, const Matrix& U) const
{
  std::list<Node*>::const_iterator i;

  max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
  min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();

  for (i = operands.begin(); i != operands.end(); i++) {
    if (*i) {
      Vector local_max(3);
      Vector local_min(3);
      (*i) -> getBoundingBox(local_max, local_min, U);

      for (size_t j = 0; j < 3; j++) {
        max(static_cast<unsigned int>(j)) =
	  MesoRD::max(max(static_cast<unsigned int>(j)),
		      local_max(static_cast<unsigned int>(j)));
        min(static_cast<unsigned int>(j)) =
	  MesoRD::min(min(static_cast<unsigned int>(j)),
		      local_min(static_cast<unsigned int>(j)));
      }
    }
  }
}
