/* The CSGNode class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSGNode.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <iomanip>

#include "CSGNode.hpp"
#include "../definition.hpp"
#include "../ui/OPrefixStream.hpp"



  /***************
   * DESTRUCTORS *
   ***************/

  /* For technical reasons one should always provide a non-inline
   * implementation for virtual destructors, even if it is trivial.
   * This will make sure that the compiler generates a vtable.
   * Without a vtable, there will be unresolved symbols when it comes
   * to linking.
   */
MesoRD::CSG::Node::~Node()
{}


  /* For technical reasons one should always provide a non-inline
   * implementation for virtual destructors, even if it is trivial.
   */
MesoRD::CSG::Primitive::~Primitive()
{}


  /* Since the SetOperation nodes do have children, the destructor of
   * the SetOperation base makes sure to destruct those as well.  This
   * will free the memory allocated to the CSG tree recursively, and
   * will prevent leaking memory on the heap on systems that do not do
   * garbage collection.
   */
MesoRD::CSG::SetOperation::~SetOperation()
{
  std::list<Node*>::iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (*i)
      delete (*i);
  }
}



  /**********
   * INVERT *
   **********/

  /* For every one of the <code>x</code>-, <code>y</code>- and
   * <code>z</code>-axes, check whether the axis unit vector is equal
   * to <code>d</code> = <code>p</code> - po</code>, the difference
   * between the previous <code>p</code> and the new one.  If equality
   * is found, add, or subtract, depending on the sign of the relevant
   * component in <code>d</code>, the relevant column from
   * <code>Tinv</code>.  The (<code>po</code>, <code>qo</code>) pair
   * must always be valid, that is, at all points must <code>qo</code>
   * = <code>Tinv</code> * <code>po</code> hold.
   */
void
MesoRD::CSG::Primitive::invert(const Vector& p, Vector& q) const
{
  const Vector d(p - po);

  for (size_t i = 0; i < 3; i++) {
    const double d0 = d(static_cast<unsigned int>((i + 0) % 3));
    const double d1 = d(static_cast<unsigned int>((i + 1) % 3));
    const double d2 = d(static_cast<unsigned int>((i + 2) % 3));

    if (fabs(d0 - 1.0) < std::numeric_limits<double>::min()
      && fabs(d1)      < std::numeric_limits<double>::min()
      && fabs(d2)      < std::numeric_limits<double>::min()) {
      const double sign = d0 > 0.0 ? +1.0 : -1.0;

      for (size_t j = 0; j < p.getDimension(); j++) {
        q(static_cast<unsigned int>(j)) = qo(static_cast<unsigned int>(j))
	  + sign * Tinv(static_cast<unsigned int>(j),
			static_cast<unsigned int>(i));
      }

      po = p;
      qo = q;
      return;
    }
  }

  po = p;
  qo = q = Tinv * p;
}



  /*********************
   * PUTPOVRAY METHODS *
   *********************/

namespace
{
    /* This is the amount by which the left operands in a difference
     * operation are scaled.  It is important that the left operands
     * are slightly larger, otherwise rounding error may produce
     * unintended edges around differences.
     */
  const double difference_factor = double(1) + pow(double(10), double(-5));
}

  /**
   * @brief Write a transformation matrix in POV-Ray format.
   *
   * Emits the transformation matrix for a node as POV-Ray matrix.
   * The scale factor is used to bring about an isotropic scaling.
   * This is useful for differences, where it is important that the
   * right operand are slightly smaller than the left operands.
   *
   * @param T     The transformation matrix.
   * @param scale The isotropic scale factor.
   * @param os    The stream onto which to write.
   */
namespace
{
  inline std::ostream&
  putPOVRayTransform(
    const MesoRD::Matrix& T,
    const double          scale,
    std::ostream&         os)
  {
    os << "transform"               << std::endl
       << "{"                       << std::endl
       << "  matrix <"
       << scale * T(0, 0)   << ", "
       << scale * T(1, 0)   << ", "
       << scale * T(2, 0)   << ","  << std::endl
       << "          "
       << scale * T(0, 1)   << ", "
       << scale * T(1, 1)   << ", "
       << scale * T(2, 1)   << ","  << std::endl
       << "          "
       << scale * T(0, 2)   << ", "
       << scale * T(1, 2)   << ", "
       << scale * T(2, 2)   << ","  << std::endl
       << "          "
       << T(0, 3) / T(3, 3) << ", "
       << T(1, 3) / T(3, 3) << ", "
       << T(2, 3) / T(3, 3) << ">"  << std::endl
       << "}"                       << std::endl;
    return os;
  }
}


  /**
   * @brief Write pigment information in POV-Ray format.
   *
   * Emits the pigmentation information for a node.  In this case, the
   * pigmentation is the colour and the transparency.  All these
   * values should be between zero and one inclusive.
   *
   * @param r  The red value, between zero and one inclusive.
   * @param g  The green value, between zero and one inclusive.
   * @param b  The blue value, between zero and one inclusive.
   * @param f  The transparency value, between zero and one inclusive.
   * @param os The stream onto which to write.
   */
namespace
{
  inline std::ostream&
  putPOVRayPigment(
    const float   r,
    const float   g,
    const float   b,
    const float   f,
    std::ostream& os)
  {
    os << "pigment"  << std::endl
       << "{"        << std::endl
       << "  rgbf <"
       << r << ", "
       << g << ", "
       << b << ", "
       << f << ">"   << std::endl
       << "}"        << std::endl;
    return os;
  }
}


  /* A simple box can be defined by listing two corners of the box,
   * such as
   *
   * 
   * box
   * {
   *   <Corner_1>, <Corner_2>
   * }
   *
   * where <Corner_1> and <Corner_2> are vectors defining the
   * <code>x</code>, <code>y</code> and <code>z</code> coordinates of
   * the opposite corners of the box.
   */
std::ostream&
MesoRD::CSG::Box::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "box" << std::endl
              << "{"   << std::endl;
  opfs_level1 << "<"
              << double(-0.5) * side[0] << ", "
              << double(-0.5) * side[1] << ", "
              << double(-0.5) * side[2] << ">," << std::endl
              << "<"
              << double(+0.5) * side[0] << ", "
              << double(+0.5) * side[1] << ", "
              << double(+0.5) * side[2] << ">"  << std::endl;

  putPOVRayTransform(T, scale, opfs_level1);
  putPOVRayPigment(r, g, b, f, opfs_level1);

  opfs_level0 << "}" << std::endl;

  return os;
}


  /* Just output the compartment, if it exists.
   */
std::ostream&
MesoRD::CSG::Compartment::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  if (compartment)
    return compartment -> putPOVRay(r, g, b, f, scale, os);
  return os;
}


  /* The cone statement creates a finite length cone or a frustum (a
   * cone with the point cut off).  The syntax is:
   *
   * cone
   * {
   *   <Base_Point>, Base_Radius, <Cap_Point>, Cap_Radius
   * }
   *
   * where <Base_Point> and <Cap_Point> are vectors defining the
   * <code>x</code>, <code>y</code>, <code>z</code> coordinates of the
   * centre of the cone's base and cap and Base_Radius and Cap_Radius
   * are values for the corresponding radii.  In MesoRD cones always
   * have Base_Radius equal to Cap_Radius, as periodic boundary
   * conditions would not make sense otherwise.  The ends of a cone
   * are closed by flat discs that are parallel to each other and
   * perpendicular to the length of the cone.
   */
std::ostream&
MesoRD::CSG::Cone::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "cone" << std::endl
              << "{"    << std::endl;
  opfs_level1 << "<0, "
              << double(-0.5) * height << ", 0>," << std::endl
              << radius                << ","     << std::endl
              << "<0, "
              << double(+0.5) * height << ", 0>"  << std::endl
              << "0"                              << std::endl;

  putPOVRayTransform(T, scale, opfs_level1);
  putPOVRayPigment(r, g, b, f, opfs_level1);

  opfs_level0 << "}" << std::endl;

  return os;
}


  /* The cylinder statement creates a finite length cylinder with
   * parallel end caps.  The syntax is:
   *
   * cylinder
   * {
   *   <Base_Point>, <Cap_Point>, Radius
   * }
   *
   * where <Base_Point> and <Cap_Point> are vectors defining the
   * <code>x</code<, <code>y</code>, <code>z</code> coordinates of the
   * cylinder's base and cap and Radius is a float value for the
   * radius.  The ends of a cylinder are closed by flat discs that are
   * parallel to each other and perpendicular to the length of the
   * cylinder.
   */
std::ostream&
MesoRD::CSG::Cylinder::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "cone" << std::endl
              << "{"    << std::endl;
  opfs_level1 << "<0, "
              << double(-0.5) * height << ", 0>," << std::endl
              << radius                << ","     << std::endl
              << "<0, "
              << double(+0.5) * height << ", 0>"  << std::endl
              << radius                           << std::endl;

  putPOVRayTransform(T, scale, opfs_level1);
  putPOVRayPigment(r, g, b, f, opfs_level1);

  opfs_level0 << "}" << std::endl;

  return os;
}


  /* The mesh object can be used to efficiently store large numbers of
   * triangles.  Its syntax is:
   *
   * mesh
   * {
   *   triangle
   *   {
   *     <Corner_1>, <Corner_2>, <Corner_3>
   *   }
   *   ...
   * }
   *
   * where <Corner_1>, <corner_2> and <Corner_3> are vectors defining
   * the <code>x</code<, <code>y</code>, <code>z</code> coordinates of
   * the vertices of the individual triangles.  Well-behaved, or
   * completely closed, triangle meshes can be used in POV-Ray CSG
   * operations such as difference and intersect, because they do have
   * a defined "inside".  If meshes have any holes in them, this might
   * work, but the results are not guaranteed.
   */
std::ostream&
MesoRD::CSG::Mesh::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "mesh" << std::endl
              << "{"    << std::endl;

  std::list<Mesh::Triangle>::const_iterator i;
  for (i = triangles.begin(); i != triangles.end(); i++) {
    opfs_level1 << "triangle"         << std::endl
                << "{"                << std::endl
                << "  <"
                << i -> p0(0) << ", "
                << i -> p0(1) << ", "
                << i -> p0(2) << ">," << std::endl
                << "  <"
                << i -> p1(0) << ", "
                << i -> p1(1) << ", "
                << i -> p1(2) << ">," << std::endl
                << "  <"
                << i -> p2(0) << ", "
                << i -> p2(1) << ", "
                << i -> p2(2) << ">"  << std::endl
                << "}"                << std::endl;
  }

  putPOVRayTransform(T, scale, opfs_level1);
  putPOVRayPigment(r, g, b, f, opfs_level1);

  opfs_level0 << "}" << std::endl;

  return os;
}


  /* The syntax of the sphere object is:
   *
   * sphere
   * {
   *   <Centre>, Radius
   * }
   *
   * Where <Centre> is a vector specifying the <code>x</code>,
   * <code>y</code>, <code>z</code> coordinates of the centre of the
   * sphere and Radius is a value specifying the radius.
   */
std::ostream&
MesoRD::CSG::Sphere::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "sphere"     << std::endl
              << "{"          << std::endl;
  opfs_level1 << "<0, 0, 0>," << std::endl
              << radius       << std::endl;

  putPOVRayTransform(T, scale, opfs_level1);
  putPOVRayPigment(r, g, b, f, opfs_level1);

  opfs_level0 << "}" << std::endl;

  return os;
}


  /* The CSG difference operation takes the intersection between the
   * first object and the inverse of all subsequent objects.  The
   * syntax of the difference operation is:
   *
   * difference
   * {
   *   SOLID_OBJECTS...
   * }
   */
std::ostream&
MesoRD::CSG::Difference::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  if (operands.size() > 0) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "#if (OPERATIONS)" << std::endl
                << "difference"       << std::endl
                << "{"                << std::endl
                << "#end"             << std::endl;

    std::list<Node*>::const_iterator i;
    for (i = operands.begin(); i != operands.end(); i++) {
      if (*i) {
        if (i == operands.begin())
          (*i) -> putPOVRay(r, g, b, f, scale, opfs_level1);
        else {
          const float darkest = MesoRD::min(r, MesoRD::min(g, b));
          (*i) -> putPOVRay(darkest, darkest, darkest, f,
            difference_factor * scale, opfs_level1);
        }
      }
    }

    opfs_level0 << "#if (OPERATIONS)" << std::endl
                << "}"                << std::endl
                << "#end"             << std::endl;
  }

  return os;
}


  /* The intersection object creates a shape containing only those
   * areas where all components overlap.  The syntax of the
   * intersection operation is:
   *
   * intersection
   * {
   *   SOLID_OBJECTS...
   * }
   */
std::ostream&
MesoRD::CSG::Intersection::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  if (operands.size() > 0) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "#if (OPERATIONS)" << std::endl
                << "intersection"     << std::endl
                << "{"                << std::endl
                << "#end"             << std::endl;

    std::list<Node*>::const_iterator i;
    for (i = operands.begin(); i != operands.end(); i++) {
      if (*i)
        (*i) -> putPOVRay(r, g, b, f, scale, opfs_level1);
    }

    opfs_level0 << "#if (OPERATIONS)" << std::endl
                << "}"                << std::endl
                << "#end"             << std::endl;
  }

  return os;
}


  /* Unions are simply glue used to bind two or more shapes into a
   * single entity that can be manipulated as a single object.  One
   * should be aware that the surfaces inside the union will not be
   * removed.  This may be a problem for transparent unions.  If those
   * surfaces are to be removed, one will have to use the merge
   * operation instead.  The syntax of the union operation is:
   *
   * union
   * {
   *   SOLID_OBJECTS...
   * }
   *
   * The syntax of the merge operation is, for these purposes,
   * identical.
   */
std::ostream&
MesoRD::CSG::Union::putPOVRay(
  const float   r,
  const float   g,
  const float   b,
  const float   f,
  const double  scale,
  std::ostream& os) const
{
  if (operands.size() > 0) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "#if (OPERATIONS)"                       << std::endl
                << "#if (USE_MERGE) merge #else union #end" << std::endl
                << "{"                                      << std::endl
                << "#end"                                   << std::endl;

    std::list<Node*>::const_iterator i;
    for (i = operands.begin(); i != operands.end(); i++) {
      if (*i)
        (*i) -> putPOVRay(r, g, b, f, scale, opfs_level1);
    }

    opfs_level0 << "#if (OPERATIONS)" << std::endl
                << "}"                << std::endl
                << "#end"             << std::endl;
  }

  return os;
}


  /******************
   * PUTXML METHODS *
   ******************/

std::ostream&
MesoRD::CSG::Box::putXML(std::ostream& os) const
{
  std::string leader1 = "";
  std::string leader2 = "";

  for (size_t i = 0; i < Identifier::XML::Box::name.length() + 2; i++)
    leader1 += " ";

  for (size_t j = 0; j < Identifier::XML::PBC::name.length() + 2; j++)
    leader2 += " ";

  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  OPrefixStream opfs_level2(opfs_level0.rdbuf(), leader1.c_str());
  OPrefixStream opfs_level3(opfs_level1.rdbuf(), leader2.c_str());

  opfs_level0 << "<"   << Identifier::XML::Box::name
              << " "   << Identifier::XML::Box::Attributes::x
              << "=\"" << side[0]
              << "\""  << std::endl;

  opfs_level2 << Identifier::XML::Box::Attributes::y
              << "=\"" << side[1]
              << "\""  << std::endl
              << Identifier::XML::Box::Attributes::z
              << "=\"" << side[2]
              << "\">" << std::endl;

  opfs_level1 << "<"   << Identifier::XML::PBC::name
              << " "   << Identifier::XML::PBC::Attributes::x
              << "=\"" << (pbc[0] ? "true" : "false")
              << "\""  << std::endl;

  opfs_level3 << Identifier::XML::PBC::Attributes::y
              << "=\""  << (pbc[1] ? "true" : "false")
              << "\""   << std::endl
              << Identifier::XML::PBC::Attributes::z
              << "=\""  << (pbc[2] ? "true" : "false")
              << "\"/>" << std::endl;

  opfs_level0 << "</" << Identifier::XML::Box::name
              << ">"  << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Compartment::putXML(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(), "");

  opfs_level0 << "<"    << Identifier::XML::Compartment::name
              << " "    << Identifier::XML::Compartment::id
              << "=\""  << id
              << "\"/>" << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Cone::putXML(std::ostream& os) const
{
  std::string leader = "";

  for (size_t i = 0; i < Identifier::XML::Cone::name.length() + 2; i++)
    leader += " ";

  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), leader.c_str());

  opfs_level0 << "<"   << Identifier::XML::Cone::name
              << " "   << Identifier::XML::Cone::Attributes::height
              << "=\"" << height
              << "\""  << std::endl;

  opfs_level1 << Identifier::XML::Cone::Attributes::radius
              << "=\"" << radius << "\"/>" << std::endl;

  std::cout << std::endl << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Cylinder::putXML(std::ostream& os) const
{
  std::string leader = "";

  for (size_t i = 0; i < Identifier::XML::Cylinder::name.length() + 2; i++)
    leader += " ";

  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  OPrefixStream opfs_level2(opfs_level0.rdbuf(), leader.c_str());

  opfs_level0 << "<"    << Identifier::XML::Cylinder::name
              << " "    << Identifier::XML::Cylinder::Attributes::height
              << "=\""  << height
              << "\""   << std::endl;

  opfs_level2 << Identifier::XML::Cylinder::Attributes::radius
              << "=\""  << radius << "\">" << std::endl;

  opfs_level1 << "<"    << Identifier::XML::PBC::name
              << " "    << Identifier::XML::PBC::Attributes::z
              << "=\""  << (pbc ? "true" : "false")
              << "\"/>" << std::endl;

  opfs_level0 << "</"   << Identifier::XML::Cylinder::name
              << ">"    << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Mesh::putXML(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");

  std::list<Mesh::Triangle>::const_iterator i;

  opfs_level0 << "<" << Identifier::XML::Mesh::name << ">" << std::endl;
  for (i = triangles.begin(); i != triangles.end(); i++) {
    opfs_level1 << "<" << Identifier::XML::Mesh::Triangle::name
                << ">" << std::endl;

    opfs_level2 << "<" << Identifier::XML::Mesh::Triangle::Vertex::name
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::x
                << "=\""  << i -> p0(0) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::y
                << "=\""  << i -> p0(1) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::z
                << "=\""  << i -> p0(2) << "\""
                << "/>" << std::endl;

    opfs_level2 << "<" << Identifier::XML::Mesh::Triangle::Vertex::name
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::x
                << "=\""  << i -> p1(0) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::y
                << "=\""  << i -> p1(1) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::z
                << "=\""  << i -> p1(2) << "\""
                << "/>" << std::endl;

    opfs_level2 << "<" << Identifier::XML::Mesh::Triangle::Vertex::name
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::x
                << "=\""  << i -> p2(0) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::y
                << "=\""  << i -> p2(1) << "\""
                << " "
                << Identifier::XML::Mesh::Triangle::Vertex::Attributes::z
                << "=\""  << i -> p2(2) << "\""
                << "/>" << std::endl;

    opfs_level1 << "</" << Identifier::XML::Mesh::Triangle::name
                << ">" << std::endl;
  }
  opfs_level0 << "</" << Identifier::XML::Mesh::name << ">" << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Sphere::putXML(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(), "");

  opfs_level0 << "<"    << Identifier::XML::Sphere::name
              << " "    << Identifier::XML::Sphere::Attributes::radius
              << "=\""  << radius
              << "\"/>" << std::endl;

  return os;
}


std::ostream&
MesoRD::CSG::Difference::putXML(std::ostream& os) const
{
  Node* left  = operands.front();
  Node* right = operands.back();

  if (left && right) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "<"   << Identifier::XML::Difference::name
                << ">"   << std::endl;

    left  -> putXML(opfs_level1);
    right -> putXML(opfs_level1);

    opfs_level0 << "</"  << Identifier::XML::Difference::name
                << ">"   << std::endl;
  }

  return os;
}


std::ostream&
MesoRD::CSG::Intersection::putXML(std::ostream& os) const
{
  if (operands.size() > 0) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "<" << Identifier::XML::Intersection::name
                << ">" << std::endl;

    std::list<Node*>::const_iterator i;
    for (i = operands.begin(); i != operands.end(); i++) {
      if (*i)
        (*i) -> putXML(opfs_level1);
    }

    opfs_level0 << "</" << Identifier::XML::Intersection::name
                << ">"  << std::endl;
  }

  return os;
}


std::ostream&
MesoRD::CSG::Union::putXML(std::ostream& os) const
{
  if (operands.size() > 0) {
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

    opfs_level0 << "<" << Identifier::XML::Union::name
                << ">" << std::endl;

    std::list<Node *>::const_iterator i;
    for (i = operands.begin(); i != operands.end(); i++) {
      if (*i)
        (*i) -> putXML(opfs_level1);
    }

    opfs_level0 << "</" << Identifier::XML::Union::name
                << ">"  << std::endl;
  }

  return os;
}
