/* The CSGNode class.
 *
 * Copyright 2004-2007 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSGNode.hpp 541 2008-04-12 18:52:41Z davidfange $
 */

#ifndef MESORD_CSG_CSGNODE_H
#define MESORD_CSG_CSGNODE_H 1

#include <limits>
#include <list>
#include <map>
#include <stdexcept>
#include <valarray>

#include "../mesord.hpp"
#include "../linalg/Matrix.hpp"
#include "../linalg/Vector.hpp"

namespace MesoRD
{
  namespace CSG
  {

      /***************************
       * THE CSG NODE BASE CLASS *
       ***************************/

      /**
       * @brief Base class for constructing solid geometries from a
       * fixed set of geometric primitives and operations.
       *
       * Constructive Solid Geometry galore! This class defines an
       * interface for the set operation and terminal primitive nodes
       * of %CSG trees.  Perhaps, a note on where all this is coming
       * from might not be amiss: the predefined set of primitives and
       * transformations, including the parameters required for their
       * definition, is taken from <i>VRML</i>.  Note that VRML is the
       * <i>Virtual Reality <em>Modelling</em> Language</i>, not
       * <i>Markup Language</i>.  Because MesoRD's %CSG trees are
       * defined using XML, some changes had to be introduced in order
       * to transform it to an extended markup language.  Other
       * sources of inspiration are <i>POV-Ray: The Persistence of
       * Vision Ray Tracer</i> and <i>GDML: Geometry Description
       * Markup Language</i>.  Other things that are bad to forget:
       *
       * <ul>
       *
       *   <li>Transformations do <em>not</em> have their own nodes.
       *   Instead, each terminal node, or equivalently, each node
       *   corresponding to a <em>geometric primitive</em>, stores its
       *   own transformation matrix, that applies to it, and only
       *   it.</li>
       *
       *   <li>There is a special CSG::Compartment node which will
       *   include the definition of a %CSG tree defined elsewhere.
       *   The CSG::Compartment nodes allow linkages to be built
       *   between different %CSG trees.</li>
       *
       *   <li>The tree contains no information about the unit in
       *   which box lengths, radii, <i>etc.</i> are given.  This
       *   effectively relieves the nodes from the entire unit mess.
       *   Instead, one is supposed to construct the %CSG nodes in
       *   units of subvolume size.  For instance, if one wants to
       *   build a cylinder three subvolumes high, one should define
       *   the CSG::Cylinder node with <code>height = 3</code>.</li>
       *
       *
       *   <li>In general, the validity of input is not checked here.
       *   If it is felt that users should be informed of silly
       *   values, such as radii of <code>-69</code>, by other means
       *   than a program crash, it should be checked before entering
       *   the %CSG module.</li>
       *
       * </ul>
       */
    class Node {
    public:

        /**
         * @brief The CSG::Node::Containment struct gives information
         * about the containment relationship between points and a
         * %CSG subtree.
         *
         * The realisation process of a %CSG tree essentially involves
         * asking the %CSG procedure whether or not it contains a
         * point at a certain coordinate in space.  This struct wraps
         * up the answer to such a question: a point is either an
         * CSG::Node::Containment::exterior,
         * CSG::Node::Containment::interior or
         * CSG::Node::Containment::boundary point.  If the point is a
         * CSG::Node::Containment::boundary point, the struct carries
         * information about which node in the tree flagged it as a
         * boundary point.  This information is needed in order to
         * resolve the neighbour relationships of boundary points when
         * periodic boundary conditions are active.
         */
      struct Containment
      {

          /**
           * @brief The different types of containment relationships a
           * point can have with a %CSG tree.
           */
        enum type {

            /**
             * @brief The point is a boundary point.
             */
          boundary,

            /**
             * @brief The point is an exterior point.
             */
          exterior,

            /**
             * @brief The point is an interior point.
             */
          interior
        } type;


          /**
           * @brief A pointer to the node which, in case of a boundary
           * point, flagged the coordinate as such.
           *
           * @warning The node pointer is <em>not</em> to be used for
           * anything but boundary points.  It may be
           * <code>null</code>.
           */
        const Node* node;


          /**
           * @brief Construct a new interior/exterior
           * CSG::Node::Containment struct.
           *
           * This constructor should be used when it is found that a
           * point is either interior or exterior.  In either case,
           * the node which determines the containment, is not
           * interesting and needs not to be saved.
           *
           * @param contains True if the point is interior, false if
           *                 the point is exterior.
           */
        inline
        Containment(const bool contains)
          : type(contains ? interior : exterior), node(0)
          {}


          /**
           * @brief Construct a new boundary CSG::Node::Containment
           * struct.
           *
           * This constructor should be used when it is found that a
           * point is a boundary point.  In such cases, the node which
           * determines the containment <em>is</em> interesting, and
           * needs to be saved for future reference.
           *
           * @param node The node which returns the struct.
           */
        inline
        Containment(const Node* node)
          : type(boundary), node(node)
          {}
      };


    public:

        /**
         * @brief Constructor for the Node base class.
         *
         * The node base class carries no data and needs no fancy
         * constructor.
         */
      inline
      Node() 
        {}


        /**
         * @brief Destructor for the Node base class.
         *
         * Does nothing at all.  For technical reasons, destructors of
         * abstract classes should not be defined inline.
         */
      virtual
      ~Node();


        /**
         * @brief Figure out whether a point in Cartesian space is
         * contained in the subtree.
         *
         * This function will figure out if a point in Cartesian space
         * is contained in the subtree rooted at the current node.  A
         * point can be either interior, exterior or a boundary point.
         *
         * @param p The point in three-dimensional space to check.
         * @return  Information about the containment of the point @p
         *          p.
         */
      virtual const Containment
      contains(const Vector& p) const = 0;


        /**
         * @brief Determine the bounding box of the subtree.
         *
         * This function will figure out the bounding box of the
         * geometry defined by the subtree rooted at the current node.
         * The bounding box is the smallest box, the sides of which
         * are aligned with the global <code>x</code>, <code>y</code>-
         * and <code>z</code>-axes, which entirely contains the
         * geometry.  The transformation matrix of the node is
         * multiplied from the left by @p U.  The <i>raison d'etre</i>
         * of the @p U matrix is that while determining the bounding
         * box of a CSG::Compartment node, one needs to consider the
         * transformations of the <em>including</em> as well as the
         * <em>included</em> node.
         *
         * @param max The coordinate of the top, left and front
         *            corner.
         * @param min The coordinate of the bottom, right and back
         *            corner.
         * @param U   An additional transformation matrix, applied
         *            <em>before</em> the ordinary transformation
         *            matrix of the node.
         */
      virtual void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const = 0;


        /**
         * @brief Find the location of the neighbour at @p p in
         * direction @p d.
         *
         * This function will return a vector, in homogeneous
         * coordinates, that points to the neighbour of a point @p p
         * in direction @p d.  The length of the direction vector, @p
         * d, should be equal to the centre-to-centre distance between
         * subvolumes in the sought direction.  Trivially, for
         * geometries without periodic boundary conditions and with @p
         * d defined as above, the neighbour of a subvolume at @p p in
         * the direction @p d lies at @p p + @p d.  The complications
         * caused by periodic boundary conditions make this function
         * non-trivial.
         *
         * The principle for handling periodic boundary conditions is
         * this: if it is found that the point @p p itself lies on the
         * primitive boundary, but the point <em>including</em> the
         * directional displacement @p d does not, the displacement is
         * wrapped around the length of the primitive.  Calling this
         * function on anything but boundary points is not advisable.
         *
         * According to nice C++ coding practise as advertised by
         * Stroustrup, one should use call-by-pointer because the
         * argument vectors are constant.  However, because
         * neighbour() needs to do arithmetic on the argument types,
         * call-by-reference is more convenient.  Thus the departure
         * from nice practise.
         *
         * @param p The point to test for neighbours.
         * @param d The direction in which to search for neighbours.
         * @return  The point where the neighbour should be, if there
         *          is one.
         */
      virtual Vector
      neighbour(const Vector& p, const Vector& d) const
        {
          if (contains(p + d).type != Containment::exterior)
            return p + d;
          return p;
        }


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * This function accomplishes the same thing as C++'s <i>put
         * to</i>-operator, <code>&lt;&lt;</code>, except that a node
         * is written in POV-Ray format.  The node is coloured
         * according to the supplied @p r, @p g and @p b values with
         * transparency @p f.  A scale is needed to make the entire
         * compound object slightly bigger or smaller -- rounding
         * errors would otherwise produce strange effects with
         * POV-Ray's %CSG output.
         *
         * @param r     The red component of the node's colour, in the
         *              range from zero to one inclusive.
         * @param g     The green component of the node's colour, in
         *              the range from zero to one inclusive.
         * @param b     The blue component of the node's colour, in
         *              the range from zero to one inclusive.
         * @param f     The transparency of the node, in the range
         *              from zero to one inclusive.
         * @param scale An isotropic scaling factor applied to the
         *              tree rooted at the current node.
         * @param os    The output stream to write to.
         * @return      The output stream once written to.
         */
      virtual std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const = 0;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * This function accomplishes the same thing as C++'s <i>put
         * to</i>-operator, <code>&lt;&lt;</code>, except that a node
         * is written in XML format.  The reason for its existence is
         * purely technical; it makes the implementation a little
         * nicer.
         *
         * @param os The output stream to write to.
         * @return   The output stream once written to.
         */
      virtual std::ostream&
      putXML(std::ostream& os) const = 0;
    };



      /*****************************
       * THE CSG PRIMITIVE CLASSES *
       *****************************/

      /**
       * @brief Base class for primitive, terminal, geometry nodes.
       *
       * The terminal nodes in a %CSG tree must all define geometric
       * primitives.  Because they are terminal nodes, they cannot
       * have any children.  Because they define a concrete geometric
       * object, they require a transformation matrix, <code>T</code>,
       * which encodes the mapping from the global coordinate system
       * to the local coordinate system in which the primitive is
       * defined.  This class implements the functions that are common
       * to all geometric primitives: transformation matrix handling
       * and transformation of vectors from the global coordinate
       * system to the local ditto.
       */
    class Primitive : public Node 
    {
    protected:

        /**
         * @brief The transformation matrix that should be applied to
         * this node only.
         *
         * Note that the transformation matrix does not apply to any
         * other nodes, as those will have their own transformation
         * matrices.  The transformation matrix, <code>T</code>, is
         * used by for instance getBoundingBox().
         */
      Matrix T;

        /**
         * @brief The inverse of the transformation matrix.
         *
         * The reason for storing this matrix is to avoid expensive,
         * repeated inversion of <code>T</code>.  The inverse
         * transformation matrix, <code>Tinv</code>, is used by for
         * instance contains().  One must always make sure that
         * <code>T</code> and <code>Tinv</code> are in sync, that is
         * one must always have <code>T * Tinv = 1</code>.
         */
      Matrix Tinv;

    private:
      mutable Vector po;
      mutable Vector qo;

    public:

        /**
         * @brief Constructor for the Primitive base class.
         *
         * This constructor will initialise the transformation matrix,
         * <code>T</code>, and its inverse, <code>Tinv</code> from the
         * input data given.  It will also set the <code>po</code> and
         * <code>qo</code> vectors so that <code>qo = Tinv *
         * po</code>.  The <code>po</code> and <code>qo</code> members
         * are used by invert().
         *
         * @param T The transformation matrix of the node.
         */
      inline
      Primitive(const Matrix& T)
        : T(T), Tinv(T.inv()), po(4), qo(4)
        {
	  po(0) = 0;
	  po(1) = 0;
	  po(2) = 0;
	  po(3) = 1.0;
	  qo(0) = Tinv(0, 3);
	  qo(1) = Tinv(1, 3);
	  qo(2) = Tinv(2, 3);
	  qo(3) = Tinv(3, 3);
        }


        /**
         * @brief Destructor for the Primitive base class.
         *
         * Does nothing at all.  For technical reasons, destructors of
         * abstract classes should not be defined inline.
         */
      virtual
      ~Primitive();


        /**
         * @brief Transform @p p from the global coordinate system to
         * @p q in the local coordinate system.
         *
         * Generally, @p q is computed from @p p by multiplication
         * with the inverse transformation matrix, <code>q = Tinv *
         * p</code>.  However, because one will typically need to
         * transform a whole bunch of @p p:s, where the difference
         * between one @p p and the previous is exactly unity times
         * any one of the coordinate axes, things can be improved a
         * little.
         *
         * The idea is to store the previously computed @p p and @p q
         * as <code>po</code> and <code>qo</code>.  If the next @p p
         * to transform differs from the previous by unity in exactly
         * one of the <code>x</code>-, <code>y</code>- or
         * <code>z</code>-axes, one can just take the previously
         * computed @p q and add, or subtract, the appropriate column
         * from <code>Tinv</code>.  Note that @p p and @p q must be
         * the right sizes.  In particular, one must have
         * <code>size(q) &gt;= size(p) &gt;= 3</code>.
         *
         * @param p The untransformed vector, of size &gt;= 3.
         * @param q The transformed vector, the size of which must be
         *          bigger or equal to the size of @p q.
         */
      void
      invert(const Vector& p, Vector& q) const;
    };



      /**
       * @brief The Box class.
       *
       * The box is one of the geometric primitives in constructive
       * solid geometry.  An untransformed box is centred on the
       * origin, so that it spans the <code>x</code>-axis from <code>x
       * = -size_x / 2</code> to <code>x = +size_x / 2</code>, and
       * similarly for the <code>y</code>- and <code>z</code>-axes.
       * Boxes can have periodic boundary conditions.
       */
    class Box : public Primitive
    {
    private:
      std::valarray<bool>   pbc;
      std::valarray<double> side;

    public:

        /**
         * @brief Box class value constructor.
         *
         * Initialises the variables and nothing more.  Negative side
         * lengths make no sense, but that is neither checked nor
         * enforced here.
         *
         * @param x     The Cartesian <code>x</code>-coordinate of the
         *              node.
         * @param y     The Cartesian <code>y</code>-coordinate of the
         *              node.
         * @param z     The Cartesian <code>z</code>-coordinate of the
         *              node.
         * @param pbc_x Whether PBC:s apply in the
         *              <code>x</code>-direction or not.
         * @param pbc_y Whether PBC:s apply in the
         *              <code>y</code>-direction or not.
         * @param pbc_z Whether PBC:s apply in the
         *              <code>z</code>-direction or not.
         * @param T     The transformation matrix of the node.
         */
      inline
      Box(const double x, const double y, const double z,
        const bool pbc_x, const bool pbc_y, const bool pbc_z, const Matrix& T)
        : Primitive(T), pbc(3), side(3)
        {
          pbc[0]  = pbc_x;
          pbc[1]  = pbc_y;
          pbc[2]  = pbc_z;
          side[0] = x;
          side[1] = y;
          side[2] = z;
        }


        /**
         * @brief Box class destructor.
         */
      inline
      ~Box()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Find the location of the neighbour at @p p in
         * direction @p d.
         *
         * @copydoc Node::neighbour()
         */
      Vector
      neighbour(const Vector& p, const Vector& d) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Compartment class.
       *
       * The Compartment class offers a way to include %CSG trees
       * defined elsewhere into the building of a new %CSG tree.  This
       * means the CSG::Compartment node is not a <em>proper</em>
       * terminal node in the %CSG tree, as it attaches another %CSG
       * tree to what would otherwise have been a terminal.
       *
       * @warning The transformation matrix at the point of inclusion
       * can play tricks on the model builder.  For example if one
       * rotates the union of a box and an included %CSG tree, the
       * included tree will also be rotated.
       */
    class Compartment : public Primitive
    {
    private:
      const Node* compartment;
      std::string id;

    public:

        /**
         * @brief Compartment class value constructor.
         *
         * A CSG::Compartment node is really just a reference to
         * another %CSG tree.  However, because all %CSG trees cannot
         * be built simultaneously, CSG::Compartment nodes will have
         * to be built in two passes.  In the first pass, the
         * constructor is called with a textual identifier of the tree
         * that is to be included.  In the second pass, getId() and
         * setCompartment() will have to be used to match the
         * identifier to a pointer to the tree to include.
         *
         * @param id The identifier of the tree to include.
         * @param T  The transformation matrix of the node.
         */
      inline
      Compartment(const std::string& id, const Matrix& T)
        : Primitive(T), id(id)
        {}


        /**
         * @brief Compartment class destructor.
         *
         * One does <em>not</em> want to destruct the compartment one
         * is referring to here.  That could cause some really strange
         * problems for whomever keeps track of all the %CSG forest.
         */
      inline
      ~Compartment()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


       /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Get the identifier of the tree to include.
         *
         * This function will return a string identifying the tree
         * which is to be included.
         *
         * @return The identifier of the tree to include.
         */
      inline const std::string&
      getId() const
        {
          return id;
        }


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;


        /**
         * @brief Set the tree to include.
         *
         * This function will set up the reference to the tree to
         * include.  Only once this function has been called with a
         * valid argument, is the instance to be considered fully
         * initialised.
         *
         * @param tree A pointer to the root of the tree corresponding
         *             to the compartment to include.
         */
      inline void
      setCompartment(const Node* tree)
        {
          compartment = tree;
        }
    };



      /**
       * @brief The Cone class.
       *
       * The cone is one of the geometric primitives in constructive
       * solid geometry.  An untransformed cone is oriented along the
       * <code>y</code>-axis, extending half of its height along the
       * positive <code>y</code>-axis, and half of its height along
       * the negative <code>y</code>-axis.  The base of the cone lies
       * on the negative <code>y</code>-axis, at <code>y = -height /
       * 2</code>.
       */
    class Cone : public Primitive
    {
    private:
      double height;
      double radius;

    public:

        /**
         * @brief Cone class value constructor.
         *
         * Initialises the variables and nothing more.  Negative
         * heights or radii make no sense, but that is neither
         * checked nor enforced here.
         *
         * @param height The height of the cone.
         * @param radius The radius of the cone.
         * @param T      The transformation matrix of the node.
         */
      inline
      Cone(const double height, const double radius,
        const Matrix& T)
        : Primitive(T), height(height), radius(radius)
        {}


        /**
         * @brief Cone class destructor.
         */
      inline
      ~Cone()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Cylinder class.
       *
       * The cylinder is one of the geometric primitives in
       * constructive solid geometry.  An untransformed cylinder is
       * oriented along the <code>y</code>-axis, extending half of its
       * height along the positive <code>y</code>-axis, and half of
       * its height along the negative <code>y</code>-axis.  Cylinders
       * can have periodic boundary conditions.
       *
       * Cylinders are really special cases of generalised cones, in
       * the sense that the radius is not dependant on the
       * <code>y</code>-coordinate.  This means that the radius at the
       * base is always equal to the radius at the cap.  The fact that
       * the areas of the base and the cap are equal is also what
       * makes periodic boundary conditions possible for cylinders.
       */
    class Cylinder : public Primitive
    {
    private:
      bool   pbc;
      double height;
      double radius;

    public:

        /**
         * @brief Cylinder class value constructor.
         *
         * Initialises the variables and nothing more.  Negative
         * heights or radii make no sense, but that is neither
         * checked nor enforced here.
         *
         * @param height The height of the cylinder.
         * @param radius The radius of the cylinder.
         * @param pbc    Whether PBC applies in the
         *               <code>z</code>-direction or not.
         * @param T      The transformation matrix of the node.
         */
      inline
      Cylinder(const double height, const double radius,
        const bool pbc, const Matrix& T)
        : Primitive(T), pbc(pbc), height(height), radius(radius)
        {}


        /**
         * @brief Cylinder class destructor.
         */
      inline
      ~Cylinder()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Find the location of the neighbour at @p p in
         * direction @p d.
         *
         * @copydoc Node::neighbour()
         */
      Vector
      neighbour(const Vector& p, const Vector& d) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Mesh class.
       *
       * The mesh is one of the geometric primitives in constructive
       * solid geometry.  A mesh is in principle a list of
       * <em>sided</em> triangles -- sided meaning that the triangles
       * have a <i>front face</i> and a <i>back face</i>.  The
       * sidedness of a triangle is determined by computing the normal
       * of the triangle, @f$ \mathbf{n} = \left( \mathbf{p}_{1} -
       * \mathbf{p}_{0} \right) \times \left( \mathbf{p}_{2} -
       * \mathbf{p}_{0} \right) @f$, where @f$ \mathbf{p}_{0} @f$, @f$
       * \mathbf{p}_{1} @f$ and @f$ \mathbf{p}_{2} @f$ are the
       * <em>ordered</em> vertices of the triangle.
       *
       * By convention, the normal is taken to point away from the
       * front face of the triangle.  Equivalently, the normal is
       * assumed to point <em>out of</em> the solid enclosed by all
       * triangles.  Another way to state the above is this: when
       * looking at the triangle's front face, that is when inspecting
       * the exterior surface of the solid, the vertices of the
       * triangle should be listed in counterclockwise order.  The
       * volume behind the list of triangles <em>should</em> be
       * closed.  Interesting surprises await those who define
       * unclosed meshes.
       */
    class Mesh : public Primitive
    {
    public:

        /**
         * @brief The triangle struct defines a triangle from three
         * vertices.
         *
         * The vertices should be given in counterclockwise order.
         * That is, when viewing the front side of the triangle, the
         * counterclockwise order of the vertices should be
         * <code>p0</code>, <code>p1</code>, <code>p2</code>, or a
         * cyclic permutation thereof.
         */
      typedef struct Triangle {

          /**
           * @brief The first vertex in the triangle.
           */
        Vector p0;

          /**
           * @brief The second vertex in the triangle.
           */
        Vector p1;

          /**
           * @brief The third vertex in the triangle.
           */
        Vector p2;
      } Triangle;

    private:
      int                 lcd;
      std::list<Triangle> triangles;
      Vector              max;
      Vector              min;

      mutable std::map<std::pair<int, int>, std::list<
        std::pair<double, double> > > x_intersections;
      mutable std::map<std::pair<int, int>, std::list<
        std::pair<double, double> > > y_intersections;
      mutable std::map<std::pair<int, int>, std::list<
        std::pair<double, double> > > z_intersections;

    public:

        /**
         * @brief Mesh class value constructor.
         *
         * The triangles should adhere to the convention used here:
         * the vertices are given in counterclockwise order when
         * viewing the triangle from the front.  Furthermore, the
         * entire list of triangle should bound a closed volume.
         *
         * The private <code>lcd</code> member attribute is a strange
         * thing: it gives the least common divisor of the
         * inter-subvolume distances in grid coordinates.  For
         * instance, when using cubic subvolumes, the centre of a
         * prospective subvolume along any coordinate axis will always
         * be at <code>i + 0.5</code>, where <code>i</code> is an
         * integer.  For octahedral subvolumes, centres lie either on
         * integer or half-integer values.  An <code>lcd</code> of
         * <code>&gt;= 1</code> ensures that cubic subvolumes are
         * always an integer number of grid coordinates apart.  For
         * octahedral subvolumes, <code>lcd &gt;= 2</code> does the
         * trick.
         *
         * Because integer arithmetic is faster than floating point
         * ditto, the triangle coordinates are multiplied by
         * <code>lcd</code> prior to storage.  The <code>T</code> and
         * <code>Tinv</code> matrices are adjusted accordingly.  Once
         * this is done, one can safely round any coordinate
         * components tested in contains() to its nearest integer
         * values.
         *
         * @todo The <code>lcd</code> business should be cleared up,
         * once meshes have passed more extensive tests.  If
         * <code>lcd</code> is kept, it should probably be made a
         * <code>private const</code>.
         *
         * @param t A list of triangles, according to the definition.
         * @param T The transformation matrix of the node.
         */
      inline
      Mesh(const std::list<Triangle>& t, const Matrix& T)
        : Primitive(T), lcd(2), triangles(t), max(3), min(3)
        {
          Matrix S  = Matrix::eye(4);
          S(0, 0)   = S(1, 1) = S(2, 2) = 1.0 / static_cast<double>(lcd);
          this -> T = T * S;
          S(0, 0)   = S(1, 1) = S(2, 2) = 1.0 * static_cast<double>(lcd);
          Tinv      = S * Tinv;
          std::list<Triangle>::iterator i;
          max(0) = max(1) = max(2) = -1.0 * std::numeric_limits<double>::max();
          min(0) = min(1) = min(2) = +1.0 * std::numeric_limits<double>::max();
          for (i = triangles.begin(); i != triangles.end(); i++) {
            for (size_t j = 0; j < 3; j++) {
              i -> p0(static_cast<unsigned int>(j)) *= lcd;
              i -> p1(static_cast<unsigned int>(j)) *= lcd;
              i -> p2(static_cast<unsigned int>(j)) *= lcd;

              max(static_cast<unsigned int>(j)) =
		MesoRD::max(max(static_cast<unsigned int>(j)),
			    i -> p0(static_cast<unsigned int>(j)));
              max(static_cast<unsigned int>(j)) =
		MesoRD::max(max(static_cast<unsigned int>(j)),
			    i -> p1(static_cast<unsigned int>(j)));
              max(static_cast<unsigned int>(j)) =
		MesoRD::max(max(static_cast<unsigned int>(j)),
			    i -> p2(static_cast<unsigned int>(j)));
              min(static_cast<unsigned int>(j)) =
		MesoRD::min(min(static_cast<unsigned int>(j)),
			    i -> p0(static_cast<unsigned int>(j)));
              min(static_cast<unsigned int>(j)) =
		MesoRD::min(min(static_cast<unsigned int>(j)),
			    i -> p1(static_cast<unsigned int>(j)));
              min(static_cast<unsigned int>(j)) =
		MesoRD::min(min(static_cast<unsigned int>(j)),
			    i -> p2(static_cast<unsigned int>(j)));
            }
          }
        }


        /**
         * @brief Mesh class destructor.
         */
      inline
      ~Mesh()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Sphere class.
       *
       * The sphere is one of the geometric primitives in constructive
       * solid geometry.  An untransformed sphere is centred on the
       * origin, and extends its radius in all directions
       */
    class Sphere : public Primitive
    {
    private:
      double radius;

    public:

        /**
         * @brief Sphere class value constructor.
         *
         * Initialises the variables and nothing more.  Negative
         * radii make no sense, but that is neither checked nor
         * enforced here.
         *
         * @param radius The radius of the sphere.
         * @param T      The transformation matrix of the node.
         */
      inline
      Sphere(const double radius, const Matrix& T)
        : Primitive(T), radius(radius)
        {}


        /**
         * @brief Sphere class destructor.
         */
      inline
      ~Sphere()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * primitive.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the primitive.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /*********************************
       * THE CSG SET OPERATION CLASSES *
       *********************************/

      /**
       * @brief Base class for non-terminal, set-operation nodes.
       *
       * The non-terminal nodes in a %CSG tree are set operations.
       * They <em>could</em> have been transformation nodes as well,
       * but in this particular %CSG implementation, the
       * transformation information is stored in the terminal geometry
       * nodes instead.  Because set-operation nodes are non-terminal,
       * they <em>will</em> have children.  This class implements
       * common functionality for set operation nodes, which almost
       * exclusively boils down to daycare, or child management.
       *
       * In general, set operations do not commute: for instance
       * <code>difference(x, y)</code> is not the same as
       * <code>difference(y, x)</code>.  This means that the operands
       * of set operations are ordered.  The ordering of operands is
       * relevant in containment testing as well, because the
       * evaluation will stop as soon as the sought answer is known.
       * If, for example, one is testing for containment in a union,
       * tests stop as soon as a containment is found.  This is
       * analogous to C/C++ statements such as <code>if (a || b) {
       * ... }</code>: if <code>a</code> is <code>true</code>, the
       * value of <code>b</code> need not be evaluated.
       */
    class SetOperation : public Node 
    {
    protected:

        /**
         * @brief The list of children of a set operation, stored in a
         * STL <code>list</code> container.
         *
         * Because each class that inherits from SetOperation is an
         * <i>operation</i>, children are referred to as
         * <i>operands</i>.  Famous quote: <i>"Luke, I am your
         * operator"</i>.
         */
      std::list<Node*> operands;

    public:

        /**
         * @brief Constructor for the SetOperation base class.
         *
         * This constructor will initialise the operands from the
         * input data given.
         *
         * @param operands The operands of the node.
         */
      inline
      SetOperation(const std::list<Node*>& operands)
        : operands(operands)
        {}


        /**
         * @brief Destructor of the SetOperation base class.
         *
         * Since the SetOperation nodes <em>do</em> have children, the
         * destructor of the SetOperation base makes sure to destruct
         * those as well.  This will free the memory allocated to the
         * %CSG tree rooted here recursively, and will prevent leaking
         * memory on the heap on systems that do not do garbage
         * collection.  For technical reasons, destructors of abstract
         * classes should not be defined inline.
         */
      virtual
      ~SetOperation();


        /**
         * @brief Get a list of the operands of the set operation.
         *
         * In general, a set operation may have any number of
         * operands.  This function will return a constant reference
         * to a list of all operands of a set operation.
         *
         * @return A reference to the operands of the set operation.
         */
      inline const std::list<Node*>&
      getOperands() const
        {
          return operands;
        }
    };



      /**
       * @brief The Difference class.
       *
       * The difference operation is one of the set operations in
       * constructive solid geometry.  A difference operation always
       * has exactly two operands: the result of the difference
       * operation is the left operand minus the right operand.
       *
       * @note Some %CSG implementation allow difference operations
       * with an arbitrary number of operands.  This is <em>not</em>
       * the case here.
       */
    class Difference : public SetOperation
    {
    public:

        /**
         * @brief Difference class value constructor.
         *
         * Initialises the operands.  A difference operation always
         * has exactly two operands, @p left and @p right.  The result
         * of a difference operation is @p left minus @p right.
         *
         * @param left  The left operand of the difference operation.
         * @param right The right operand of the difference operation.
         */
      inline
      Difference(Node* left, Node* right)
        : SetOperation(std::list<Node*>(0))
        {
          operands.push_front(left);
          operands.push_back(right);
        }


        /**
         * @brief Difference class destructor.
         */
      inline
      ~Difference()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * result of the set operation.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the set operation.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Intersection class.
       *
       * The intersection operation is one of the set operations in
       * constructive solid geometry.  An intersection operation can
       * have any number of operands: the result of the intersection
       * operation is the intersection of all operands.
       */
    class Intersection : public SetOperation
    {
    public:

        /**
         * @brief Intersection class value constructor.
         *
         * Initialises the operands.  An intersection operation can
         * have any number of operands.  The result of the
         * intersection operation is the intersection of all operands.
         *
         * @param operands The operands of the intersection operation.
         */
      inline
      Intersection(const std::list<Node*>& operands)
        : SetOperation(operands)
        {}


        /**
         * @brief Intersection class destructor.
         */
      inline
      ~Intersection()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * result of the set operation.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector &p) const;


        /**
         * @brief Determine the bounding box of the set operation.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };



      /**
       * @brief The Union class.
       *
       * The union operation is one of the set operations in
       * constructive solid geometry.  A union operation can have any
       * number of operands: the result of the union operation is the
       * union of all operands.
       */
    class Union : public SetOperation
    {
    public:

        /**
         * @brief Union class value constructor.
         *
         * Initialises the operands.  A union operation can have any
         * number of operands.  The result of the union operation is
         * the union of all operands.
         *
         * @param operands The operands of the union operation.
         */
      inline
      Union(const std::list<Node*>& operands)
        : SetOperation(operands)
        {}


        /**
         * @brief Union class destructor.
         */
      inline
      ~Union()
        {}


        /**
         * @brief Determine whether a point is contained in the
         * result of the set operation.
         *
         * @copydoc Node::contains()
         */
      const Containment
      contains(const Vector& p) const;


        /**
         * @brief Determine the bounding box of the set operation.
         *
         * @copydoc Node::getBoundingBox()
         */
      void
      getBoundingBox(Vector& max, Vector& min, const Matrix& U) const;


        /**
         * @brief Put the node in POV-Ray format onto an output
         * stream.
         *
         * @copydoc Node::putPOVRay()
         */
      std::ostream&
      putPOVRay(const float r, const float g, const float b, const float f,
        const double scale, std::ostream& os) const;


        /**
         * @brief Put the node in XML format onto an output stream.
         *
         * @copydoc Node::putXML()
         */
      std::ostream&
      putXML(std::ostream& os) const;
    };
  }
}
#endif
