/* The CSGPoint class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: CSGPoint.hpp 540 2008-04-12 14:30:36Z davidfange $
 */

#ifndef MESORD_CSG_CSGPOINT_H
#define MESORD_CSG_CSGPOINT_H 1

#include <math.h>

namespace MesoRD {

    /**
     * Class for handling points in Cartesian space with sloppy
     * equality.
     *
     * This class provides methods to handle points in
     * three-dimensional Cartesian space.  What this class does that
     * makes it different from the average vector class, is that it
     * provides a sort of sloppiness on the equality comparison.
     * Actually, it is not just the equality comparison that is
     * sloppy, but several inequalities as well.  The inequalities are
     * necessary to use the class in sorted containers, such as
     * standard C++ maps.
     *
     * The sloppiness works as follows.  Two points need not be
     * exactly equal to be regarded as equal by this class.  They can
     * differ by a statically set amount in each of the three
     * Cartesian dimensions.  This two instances of the CSGPoint class
     * are regarded equal if they lie in a box with fixed sides,
     * centred on one of the points.
     */
  class CSGPoint {
  private:
    float        x;
    float        y;
    float        z;

    static float sloppy_x;
    static float sloppy_y;
    static float sloppy_z;
    
  public:

      /**
       * Default constructor for the CSGPoint class.
       *
       * This constructor will initialise the point to its default
       * value and set the sloppiness to zero in all directions.
       */
    inline CSGPoint()
      : x(0), y(0), z(0)
      { this -> sloppy_x = 0; this -> sloppy_y = 0; this -> sloppy_z = 0; }

      /**
       * Value constructor for the CSGPoint class.
       *
       * This constructor will initialise the coordinates and the
       * sloppiness according to the parameters provided.
       *
       * @param x  The Cartesian x-coordinate of the point.
       * @param y  The Cartesian x-coordinate of the point.
       * @param z  The Cartesian x-coordinate of the point.
       * @param sx The sloppiness along the x-axis.
       * @param sy The sloppiness along the y-axis.
       * @param sz The sloppiness along the z-axis.
       */
    inline CSGPoint(
      float x,
      float y,
      float z,
      float sx,
      float sy,
      float sz)
      : x(x), y(y), z(z)
      { sloppy_x = sx; sloppy_y = sy; sloppy_z = sz; }
    
      /**
       * Copy constructor for the CSGPoint class.
       *
       * This constructor will initialise the coordinates and the
       * sloppiness from the CSGPoint provided.
       *
       * @param c The CSGPoint to copy data from.
       */
    inline CSGPoint(const CSGPoint& c)
      : x(c.x), y(c.y), z(c.z)
      { this -> sloppy_x = c.sloppy_x; this -> sloppy_y = c.sloppy_y;
        this -> sloppy_z = c.sloppy_z; }

      /**
       * Destructor for the CSGPoint class.
       *
       * Does nothing at all.
       */
    inline ~CSGPoint()
      {}

      /**
       * Less-than operator for the CSGPoint class.
       *
       * The less-than operator implies a certain ordering among the
       * Cartesian dimensions.  First the x-value is checked for
       * containment in the sloppy region, then the y-value and
       * finally the z-value.  If any of them is outside, the two
       * points are not equal, and the sign of the difference
       * determines if it is greater or less.
       *
       * @param c The CSGPoint to compare to.
       * @return  True if the left instance is less than the right.
       */
    bool operator<  (const CSGPoint& c) const;

      /**
       * Greater-than operator for the CSGPoint class.
       *
       * The greater-than operator implies a certain ordering among
       * the Cartesian dimensions.  First the x-value is checked for
       * containment in the sloppy region, then the y-value and
       * finally the z-value.  If any of them is outside, the two
       * points are not equal, and the sign of the difference
       * determines if it is greater or less.
       *
       * @param c The CSGPoint to compare to.
       * @return  True if the left instance is greater than the right.
       */
    bool operator>  (const CSGPoint& c) const;

      /**
       * Equality operator for the CSGPoint class.
       *
       * This method perform the equality check for two instances of
       * the CSGPoint class.  The two instances are equal if the
       * modulus of the distance in each dimension is less or equal to
       * the sloppy-value in that particular dimension.
       *
       * @param c The CSGPoint to compare to.
       * @return  True if the two instances are equal, false otherwise.
       */
    inline bool operator== (const CSGPoint& c) const
      { return ((fabs(x - c.x) <= sloppy_x) && (fabs(y - c.y) <= sloppy_y)
        && (fabs(z - c.z) <= sloppy_z)); }
    
      /**
       * Inequality operator for the CSGPoint class.
       *
       * This method perform the inequality check for two instances of
       * the CSGPoint class.  The two instances are unequal if the
       * modulus of the distance in any dimension is greater than the
       * sloppy-value in that particular dimension.
       *
       * @param c The CSGPoint to compare to.
       * @return  True if the two instances are unequal, false otherwise.
       */
    inline bool operator!= (const CSGPoint& c) const
      { return ((fabs(x - c.x) > sloppy_x) || (fabs(y - c.y) > sloppy_y)
        || (fabs(z - c.z) > sloppy_z)); }
  };
}

#endif
