/* The BDF2 class.
 *
 * Copyright 2004 Johan Elf,Johan Hattne
 * 
 * This file is part of MesoRD.  
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: BDF2.cpp 541 2008-04-12 18:52:41Z davidfange $
 */


#include "BDF2.hpp"
#include "../Derivative.hpp"

#include <string>
#include <stdio.h>
#include <iostream>

// from the IML++ package...
#include "../iml/cg.h"
// from the sparselib++ package...
#include "../sparselib/comprow_double.hpp"
#include "../sparselib/ilupre_double.hpp"
// from the mv++ package...
#include "../mv/mvvd.hpp"
#include "../mv/mvblasd.hpp"


MesoRD::BDF2::BDF2() 
  : Integration()
{
  imConvTolerance = 0.0;
  imMaxIterations = 0;
  preconditioner = 0;
}


MesoRD::BDF2::BDF2(Derivative* derivative, 
  CompRow_Mat_double* diffusionMatrix, 
  double timeStep, std::ostream* outstream_log,
  double imConvTolerance, size_t imMaxIterations,
  MV_Vector_double& currentState) 
  : Integration(derivative, diffusionMatrix, timeStep, outstream_log)
{
  this -> imConvTolerance = imConvTolerance;
  this -> imMaxIterations = imMaxIterations;

  int i;
    /* multiply each elemnt with the timestep */
  for(i = 0; i < this -> diffusionMatrix -> NumNonzeros(); i++){
    this -> diffusionMatrix -> val(i) *= timeStep;
  }
  
    /* allocating space for temp storage vectors.  */
  oldState = MV_Vector_double(this -> diffusionMatrix -> dim(0), (double)0);
  tempState = MV_Vector_double(this -> diffusionMatrix -> dim(0), (double)0);
  
  
  startBDF2(currentState);
  

    
    /* taking (3/2) - diffusionMatrix.  */
  for(i = 0; i < this -> diffusionMatrix -> dim(0); i++){
    for(int j = this -> diffusionMatrix -> row_ptr(i); 
	j < this -> diffusionMatrix -> row_ptr(i+1); j++){
      if(this -> diffusionMatrix -> col_ind(j) == i){
	this -> diffusionMatrix -> set(i,i) = 
	  (3.0/2.0) - (*(this -> diffusionMatrix))(i,i);
      }
      else{
	this -> diffusionMatrix -> 
	  set(i, this -> diffusionMatrix -> col_ind(j)) = 
	  -(*(this -> diffusionMatrix))(i, 
            this -> diffusionMatrix-> col_ind(j));
      }
    } 
  }  
  
  this -> preconditioner = 
    new CompRow_ILUPreconditioner_double(*(this -> diffusionMatrix));
}

MesoRD::BDF2::~BDF2()
{
  delete preconditioner;
}



double
MesoRD::BDF2::updateState(MV_Vector_double& currentState)
{
  
  double localTolerance = imConvTolerance;
  int MaxIterations = static_cast<int>(imMaxIterations);
  
  rxnDerivativeVector*=0;
  tempState=(2*currentState)-oldState;
  derivative -> getReactionDerivative(
  tempState, rxnDerivativeVector);
  rxnDerivativeVector *= timeStep;
  
  tempState = currentState;

  if(CG(*diffusionMatrix, currentState, 
    (2.0*currentState) - ((1.0/2.0)*oldState) + rxnDerivativeVector, 
	*preconditioner, MaxIterations, localTolerance))
  {
    (*outstream_log) << "Convergence not reached in CG." << std::endl <<
      "Residual:" << localTolerance << std::endl <<
      "Nr Iterations:" << MaxIterations << std::endl;
  }
  oldState = tempState;
  
  return timeStep;
  
}



void
MesoRD::BDF2::startBDF2(MV_Vector_double& currentState)
{

  /* to start the bdf method, one old state is needed.  
   * Here this state will be created using a somewhat modified Euler-backwards
   */
  int i;
  
  /** 
   * Taking I - diffusionMatrix and storing it in a new matrix since the
   * operations on this will be local.
   */  
  CompRow_Mat_double localDiffusionMatrix = 
    CompRow_Mat_double(*diffusionMatrix);
  /* taking I - diffusionMatrix */
  for(i = 0; i < localDiffusionMatrix.dim(0); i++){
    for(int j = localDiffusionMatrix.row_ptr(i); 
	j < localDiffusionMatrix.row_ptr(i+1); j++){
      if(localDiffusionMatrix.col_ind(j) == i){
	localDiffusionMatrix.set(i,i) = 
	  1 - localDiffusionMatrix(i,i);
      }
      else{
	localDiffusionMatrix.set(i, localDiffusionMatrix.col_ind(j)) = 
	  -localDiffusionMatrix(i, localDiffusionMatrix.col_ind(j));
      }
    } 
  }

  CompRow_ILUPreconditioner_double localPreconditioner =
    CompRow_ILUPreconditioner_double(localDiffusionMatrix);
  
  oldState = currentState;

  rxnDerivativeVector*=0;
  tempState=2*currentState-oldState;
  derivative -> getReactionDerivative(
  tempState, rxnDerivativeVector);
  rxnDerivativeVector *= timeStep;
  
  
  double localTolerance = imConvTolerance;
  int MaxIterations = static_cast<int>(imMaxIterations);

  /* making one euler-backwards step.  */
  
  if(CG(localDiffusionMatrix, currentState, 
       currentState + rxnDerivativeVector, 
       localPreconditioner, MaxIterations, localTolerance))
  {
    (*outstream_log) << "Convergence not reached in CG." << std::endl <<
      "Residual:" << localTolerance << std::endl <<
      "Nr Iterations:" << MaxIterations << std::endl;
  }
  
}
