/* The EulerBackward class.
 *
 * Copyright 2004 Johan Elf,Johan Hattne
 * 
 * This file is part of MesoRD.  
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: EulerBackward.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef __EulerBackward_H
#define __EulerBackward_H

#include "Integration.hpp"

#include "../Derivative.hpp"
#include "../linalg/Vector.hpp"

#include "../sparselib/comprow_double.hpp"
#include "../sparselib/ilupre_double.hpp"
#include "../mv/mvvd.hpp"

namespace MesoRD
{
  /** 
   *  Class that will perform the timestepping when doing determinsitic 
   *  simulation.   
   */
  class EulerBackward : public Integration
  { 
  private:
    
    /** 
     * The maximun allowed residual before exiting the iterative method.  
     */
    double imConvTolerance;
    
    /** The maximum number of iterations that the iterativmethod will perform
     *  before bailing out.  An error message will be printed to ostream_log
     *  if the method did not converge to imConvTolerance in imMaxIterations.  
     *
     */
    size_t imMaxIterations;

      /**
       * Preconditioner for the conjugated gradient method.  
       */
    CompRow_ILUPreconditioner_double* preconditioner;

    /** 
     * Vector used to store old derivatives for the reactions for 
     *  the multistep method BDF2.   
     */
    MV_Vector_double oldState;
    MV_Vector_double tempState;
    

  public:
        /**
     * Defualt constructor for the Integration class.  
     */
    EulerBackward();
    
    /**
     * Value constructor for the EulerBackward class.  
     *
     * @param derivative A derivative object from which the system can 
     *                   be queried for reaction derivatives.  
     * @param diffusionMatrix Matrix specifying the diffusion in all 
     *                        directions.  
     * @param timeStep The timestep used.  
     * @param integrationMethod The name fo the method used to calculated the 
     *                          new currentState.  
     * @param imConvTolerance The convergence tolerance used in the 
     *                        iterative methods used
     *                        to solve the equations systems from the 
     *                        implicit methods.  
     * @param imMaxIterations The maximum number of iterations that the
     *                        iterative solver is allowed to do before bailing
     *                        out.  
     * @param outstream_log   Stream to which logging messages can be written
     *                        If the iterative solver does not converge an
     *                        message will be printed here.  
     */
    EulerBackward(Derivative* derivative, 
      CompRow_Mat_double* diffusionMatrix,  
      double timeStep, std::ostream* outstream_log,
      double imConvTolerance, size_t imMaxIterations,
      MV_Vector_double& currentState);

    /** 
     * Destructor the the Integration class.  
     */
    ~EulerBackward();
    
    /** Method used to update the current state vector. The method used
     *  is set in the Integration constructor.  
     */
    double updateState(MV_Vector_double& currentState);
    
  };
}

#endif
