/* The Integration class.
 *
 * Copyright 2004 Johan Elf,Johan Hattne
 * 
 * This file is part of MesoRD.  
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Integration.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef __Integration_H
#define __Integration_H


#include "../Derivative.hpp"

#include "../sparselib/comprow_double.hpp"
#include "../mv/mvvd.hpp"

namespace MesoRD
{

  /** 
   * Base class for the Integration classes.  
   * 
   */
  class Integration
  {
  public:
    
    typedef enum IntegrationMethod {
      Euler,
      EulerBackward,
      EulerAdaptive,
      Trapezoid,
      RungeKutta4,
      BDF2
    } IntegraionMethod;
    
  private: 
    
    static const std::string IntegrationMethods[6];
    
    static const int NrOfIntMethods;    
    
    static IntegrationMethod intMethod;
      
  
  protected:
    
    /** 
     * The timestep used 
     */
    double timeStep;
    
    Derivative*           derivative;
    CompRow_Mat_double*   diffusionMatrix;

    MV_Vector_double      rxnDerivativeVector;
    
    std::ostream*         outstream_log;
    
    
  public:
    
      /**
       * Defualt constructor for the Integration class.  
       */
    Integration();
    
    /**
     * Value constructor for the Integration class.
     *
     * Sets the derivative object, the diffusion matrix and the timestep.    
     *
     * @param derivative A derivative object from which the system can 
     *                   be queried for reaction derivatives.  
     * @param diffusionMatrix Matrix specifying the diffusion in all 
     *                        directions.  
     * @param timeStep The timestep used.  
     */
    Integration(Derivative* derivative, CompRow_Mat_double* diffusionMatrix,  
      double timeStep, std::ostream* outstream_log);

    /** 
     * Destructor the the Integration class.  
     */
    virtual ~Integration();
    
    /** 
     * Method used to update the current state vector.
     *
     * This function is overloaded by all the methods that implement
     * the integration class. This function will update the current state
     * according to the scheme of integration that is implemented in 
     * each of the classes.
     *
     * @param currentState The current state of the system.
     */
    virtual double updateState(MV_Vector_double& currentState) = 0;
    
    
      /******************
       * STATIC METHODS *
       ******************/

      /**
       * Gets the integration to use during simulation.
       *
       * This will return the integration-method that is used in
       * the current simulation.
       *
       * @return The currently used integration-method.
       */
    static inline IntegrationMethod getIntegrationMethod()
      { return intMethod; }
    
      /**
       * Set the integration-method that will be used for the rest
       * of the simulation. Used when the simulation is started.
       *
       * @param intMethod The method to be used.
       */
    static inline void setIntegrationMethod(IntegrationMethod intMethod)
      { Integration::intMethod = intMethod; }
  
      /**
       * Set the integration-method that will be used for the rest
       * of the simulation. Used when the simulation is started.
       *
       * @param intMethod The method to be used.
       */
    static bool setIntegrationMethod(std::string intMethod);
    
      /**
       * Get a list of currently implemented integration-methods.
       *
       * Will return a pointer to first element in array of 
       * strings that contain all the integration-methods that
       * has been implemented.
       */
    static inline const std::string* getIntegrationMethods()
      { return &IntegrationMethods[0]; }
    
        
  
    
  };
}

#endif





