/* The Trapez class.
 *
 * Copyright 2004 Johan Elf,Johan Hattne
 * 
 * This file is part of MesoRD.  
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Trapez.cpp 541 2008-04-12 18:52:41Z davidfange $
 */


#include "Trapez.hpp"
#include "../Derivative.hpp"

#include <string>
#include <stdio.h>
#include <iostream>

// from the IML++ package...
#include "../iml/cg.h"
// from the sparselib++ package...
#include "../sparselib/comprow_double.hpp"
#include "../sparselib/ilupre_double.hpp"
// from the mv++ package...
#include "../mv/mvvd.hpp"
#include "../mv/mvblasd.hpp"


MesoRD::Trapez::Trapez() 
  : Integration()
{
  imConvTolerance = 0.0;
  imMaxIterations = 0;
  preconditioner = 0;
  trapezoidDiffusionMatrix = 0; 
} 


MesoRD::Trapez::Trapez(Derivative* derivative, 
  CompRow_Mat_double* diffusionMatrix, 
  double timeStep, std::ostream* outstream_log,
  double imConvTolerance, size_t imMaxIterations,
  MV_Vector_double& currentState) 
  : Integration(derivative, diffusionMatrix, timeStep, outstream_log)
{
  this -> imConvTolerance = imConvTolerance;
  this -> imMaxIterations = imMaxIterations;
  
  tempState = 
    MV_Vector_double(this -> diffusionMatrix -> dim(0), (double)0);
  
  oldState = 
    MV_Vector_double(this -> diffusionMatrix -> dim(0), (double)0);
  
  rnxDerivativeVector2 = 
    MV_Vector_double(this -> diffusionMatrix -> dim(0), (double)0);
  
  int i;
    /* multiply each elemnt with the timestep */
  for(i = 0; i < this -> diffusionMatrix -> NumNonzeros(); i++){
    this -> diffusionMatrix -> val(i) *= timeStep;
  }
  
  trapezoidDiffusionMatrix = new CompRow_Mat_double(*diffusionMatrix);
  
    /* taking 2 - diffusionMatrix */
  for(i = 0; i < this -> diffusionMatrix -> dim(0); i++){
    for(int j = this -> diffusionMatrix -> row_ptr(i); 
        j < this -> diffusionMatrix -> row_ptr(i+1); j++){
      if(this -> diffusionMatrix -> col_ind(j) == i){
        this -> diffusionMatrix -> set(i,i) = 
          2 - (*(this -> diffusionMatrix))(i,i);
      }
      else{
        this -> diffusionMatrix -> 
          set(i, this -> diffusionMatrix -> col_ind(j)) = 
          -(*(this -> diffusionMatrix))(i, 
            this -> diffusionMatrix-> col_ind(j));
      }
    } 
  }

    /* taking 2 + diffusionMatrix */
  for(i = 0; i < trapezoidDiffusionMatrix -> dim(0); i++){
    trapezoidDiffusionMatrix -> set(i,i) = 
      2 + (*trapezoidDiffusionMatrix)(i,i);
  }

  this -> preconditioner = 
    new CompRow_ILUPreconditioner_double(*(this -> diffusionMatrix));
  
  oldState = currentState;

}

MesoRD::Trapez::~Trapez()
{
  delete preconditioner;
}



double
MesoRD::Trapez::updateState(MV_Vector_double& currentState)
{
  double localTolerance = imConvTolerance;
  int MaxIterations = static_cast<int>(imMaxIterations);
  
  rxnDerivativeVector*=0;
  derivative -> getReactionDerivative(currentState, rxnDerivativeVector);
  rxnDerivativeVector *= timeStep;

  rnxDerivativeVector2*=0;
  tempState = 2*currentState - oldState;
  derivative -> getReactionDerivative(tempState, rnxDerivativeVector2);
  rnxDerivativeVector2 *= timeStep;

  oldState = currentState;
  
    /* use conjucated gradient to calculate the next state.  
     */
  if(CG(*diffusionMatrix, currentState, 
       (*trapezoidDiffusionMatrix)*currentState + rxnDerivativeVector +
       rnxDerivativeVector2, 
       *preconditioner, MaxIterations, localTolerance))
  {
    (*outstream_log) << "Convergence not reached in CG." << std::endl <<
      "Residual:" << localTolerance << std::endl <<
      "Nr Iterations:" << MaxIterations << std::endl;
  }

  return timeStep;
  
}
