/* The Exception class.
 *
 * Copyright 2001-2004 Johan Hattne
 *
 * This file is part of libsim.
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Exception.hpp 540 2008-04-12 14:30:36Z davidfange $
 */

/*
 * The SimExcept class. Contains one subclass for each error that the
 * library can report.
 */


#ifndef __Exception_H
#define __Exception_H

#include <stdexcept>



  /****************
   * PARENT CLASS *
   ****************/

  /* Parent of all libsim exceptions.
   */
namespace MesoRD {
  class SimException : public std::runtime_error {
  public:
    SimException(const char *what_arg)
      : runtime_error(what_arg) {}
  };
}


  /********************
   * CHILDREN CLASSES *
   ********************/

  /* General audio error.
   */
namespace MesoRD {
  class AudioError : public SimException {
  public:
    AudioError(const char *what_arg = "Audio Error")
      : SimException(what_arg) {}
  };
}


  /* General compression error.
   */
namespace MesoRD {
  class CompressionError : public SimException {
  public:
    CompressionError(const char *what_arg = "Compression Error")
      : SimException(what_arg) {}
  };
}


  /* General decompression error.
   */
namespace MesoRD {
  class DecompressionError : public SimException {
  public:
    DecompressionError(const char *what_arg = "Decompression Error")
      : SimException(what_arg) {}
  };
}


  /* Dimensional mismatch for FFT.
   */
namespace MesoRD {
  class DimensionMismatch : public SimException {
  public:
    DimensionMismatch(const char *what_arg = "Dimension Mismatch")
      : SimException(what_arg) {}
  };
}


  /* Empty intersection between features.
   */
namespace MesoRD {
  class EmptyIntersection : public SimException {
  public:
    EmptyIntersection(const char *what_arg = "Empty Intersection")
      : SimException(what_arg) {}
  };
}


  /* Premature EOF condition.
   */
namespace MesoRD {
  class EndOfFile : public SimException {
  public:
    EndOfFile(const char *what_arg = "End of File")
      : SimException(what_arg) {}
  };
}


  /* General file error condition.
   */
namespace MesoRD {
  class FileError : public SimException {
  public:
    FileError(const char *what_arg = "File Error")
      : SimException(what_arg) {}
  };
}


  /* General error. Reset the machine and divorce your wife.
   */
namespace MesoRD {
  class GeneralError : public SimException {
  public:
    GeneralError(const char *what_arg = "General Error")
      : SimException(what_arg) {}
  };
}


  /* Invalid value.
   */
namespace MesoRD {
  class InvalidValue : public SimException {
  public:
    InvalidValue(const char *what_arg = "Invalid Value")
      : SimException(what_arg) {}
  };
}


  /* Feature not (yet) implemented.
   */
namespace MesoRD {
  class NotImplemented : public SimException {
  public:
    NotImplemented(const char *what_arg = "Not Implemented")
      : SimException(what_arg) {}
  };
}


  /* Matrix is not invertible.
   */
namespace MesoRD {
  class NotInvertible : public SimException {
  public:
    NotInvertible(const char *what_arg = "Matrix Not Invertible")
      : SimException(what_arg) {}
  };
}


  /* Matrix is not positive definite.
   */
namespace MesoRD {
  class NotPositiveDefinite : public SimException {
  public:
    NotPositiveDefinite(const char *what_arg = "Matrix Not Positive Definte")
      : SimException(what_arg) {}
  };
}


  /* Matrix is not symmetric.
   */
namespace MesoRD {
  class NotSymmetric : public SimException {
  public:
    NotSymmetric(const char *what_arg = "Matrix Not Symmetric")
      : SimException(what_arg) {}
  };
}


  /* No support for bzip2 compression.
   */
namespace MesoRD {
  class NoBZ2LIB : public SimException {
  public:
    NoBZ2LIB(const char *what_arg = "No bzip2 Support")
      : SimException(what_arg) {}
  };
}


  /* No convergence obtained.
   */
namespace MesoRD {
  class NoConvergence : public SimException {
  public:
    NoConvergence(const char *what_arg = "No convergence obtained")
      : SimException(what_arg) {}
  };
}


  /* No data (empty).
   */
namespace MesoRD {
  class NoData : public SimException {
  public:
    NoData(const char *what_arg = "No Data")
      : SimException(what_arg) {}
  };
}


  /* No support for zlib compression.
   */
namespace MesoRD {
  class NoZLIB : public SimException {
  public:
    NoZLIB(const char *what_arg = "No zlib Support")
      : SimException(what_arg) {}
  };
}


  /* Out of Core (no more memory).
   */
namespace MesoRD {
  class OutOfCore : public SimException {
  public:
    OutOfCore(const char *what_arg = "Out of Core")
      : SimException(what_arg) {}
  };
}


  /* Out of Range.
   */
namespace MesoRD {
  class OutOfRange : public SimException {
  public:
  OutOfRange(const char *what_arg = "Out of Range")
    : SimException(what_arg) {}
  };
}


  /* PGM image format error.
   */
namespace MesoRD {
  class PGMFormatError : public SimException {
  public:
    PGMFormatError(const char *what_arg = "PGM Format Error")
      : SimException(what_arg) {}
  };
}


  /* General read error.
   */
namespace MesoRD {
  class ReadError : public SimException {
  public:
    ReadError(const char *what_arg = "Read Error")
      : SimException(what_arg) {}
  };
}


  /* Unsupported image format.
   */
namespace MesoRD {
  class UnsupportedImageFormat : public SimException {
  public:
    UnsupportedImageFormat(const char *what_arg = "Unsupported Image Format")
      : SimException(what_arg) {}
  };
}


  /* General write error.
   */
namespace MesoRD {
  class WriteError : public SimException {
  public:
    WriteError(const char *what_arg = "Write Error")
      : SimException(what_arg) {}
  };
}

#endif
