/* The Matrix class.
 *
 * Copyright 2001-2004 Johan Hattne
 *
 * This file is part of libsim.
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Matrix.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

/*
 * The matrix class. Contains some basic linear algebra functions. The
 * calling syntax for many methods is modelled after their MATLAB
 * equivalents. Generally we use the index "i" for rows and the index
 * "j" for columns.
 */

//#include <limits.h> // not on windows
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <limits>
//#include <strings.h> // Not on windows
//#include <sys/time.h>

#include "Matrix.hpp"
#include "Vector.hpp"


  /* How many iterations to do at most when finding eigenvectors using
   * inverse iterations. Press et al. suggest to use a number between
   * 5 and 10.
   */
#define SIM_MAX_INVERSE_ITERATIONS 10

  /* How small the absolute value of the smallest eigenvalue can be in
   * order not to be regarded as zero.
   */
#define SIM_MIN_EIGENVALUE 1E-15



  /**
   * C-Style defines.
   */
#if(LONG_MAX <= RAND_MAX)
#  define RAND_WRAP LONG_MAX
#else
#  define RAND_WRAP (RAND_MAX + 1)
#endif

/**
 * Defines the useful macros SIM_MAX, SIM_MIN and SIM_RINT, if they
 * are not already defined.
 */
#ifndef SIM_MAX
#  define SIM_MAX(a, b) \
     (a > b ? a : b)
#endif

#ifndef SIM_MIN
#  define SIM_MIN(a, b) \
     (a < b ? a : b)
#endif

#ifndef SIM_RINT
#  define SIM_RINT(a) \
     (fmod(a, 1.0) >= 0.5 ? ceil(a) : floor(a))
#endif

//FIXME!! lrint does not exists in the VC++ world, can this be a solution??
#ifdef WIN32
#ifndef lrint
#	define lrint(a)                         \
		(long) SIM_RINT(a)
#endif
#endif


  /* The namespace to use.
   */
using namespace MesoRD;

#define rint(x) (x < 0 ? ceil((x) - 0.5) : floor((x) + 0.5))


  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /**
   * Default constructor. Intialise the dimensions to zero and set the
   * data pointer to NULL. One should not ever forget to release
   * matrices and vectors allocated, otherwise there is leakage of
   * memory.
   */
Matrix::Matrix() {
  cols = 0;
  rows = 0;
  data = NULL;
}


  /**
   * Copy constructor. Do deep copy. This will allow using this
   * constructor for cloning matrices.
   */
Matrix::Matrix(const Matrix &M) {
  cols = M.cols;
  rows = M.rows;
  
  data = new double *[rows];
  for (unsigned int i = 0; i < rows; i++) {
    data[i] = new double[cols];
    for (unsigned int j = 0; j < cols; j++) {
      data[i][j] = M.data[i][j];    
	}
//	  bcopy(M.data[i], data[i], cols * sizeof(double));
  }
}


  /**
   * Value constructor. Just allocate space for a square matrix of the
   * dimension indicated and initialse all elements to zero.
   */
Matrix::Matrix(unsigned int d) {
  cols = d;
  rows = d;
  
  data = new double *[rows];
  for (unsigned int i = 0; i < rows; i++) {
    data[i] = new double[cols];
    for (unsigned int j = 0; j < cols; j++) {
		data[i][j] = 0.0;
	}
//    bzero(data[i], cols * sizeof(double));
  }
}


  /**
   * Value constructor. Just allocate space for a matrix of the
   * dimensions indicated and initialse all elements to zero.
   */
Matrix::Matrix(unsigned int m, unsigned int n) {
  rows = m;
  cols = n;

  data = new double *[rows];
  for (unsigned int i = 0; i < rows; i++) {
    data[i] = new double[cols];
    for (unsigned int j = 0; j < cols; j++) {
	  data[i][j] = 0.0;
	}
//    bzero(data[i], cols * sizeof(double));
  }
}


  /**
   * Destructor. Free the data in the matrix if there is any.
   */
Matrix::~Matrix() {
  release();
}



  /******************************
   * GENERAL MANAGEMENT METHODS *
   ******************************/

  /**
   * Free the data if there is any. Then set the dimensions and the
   * pointers to zero and NULL respectively to avoid confusion. This
   * makes it possible to check before trying to free the same chunk
   * of memory again.
   */
void Matrix::release() {
  if (data != NULL) {
    for (unsigned int i = 0; i < rows; i++) {
      if (data[i] != NULL)
        delete data[i];
    }
    delete data;
  }
  
  cols = 0;
  rows = 0;
  data = NULL;
}


  /**
   * Return the unit matrix of dimension "d" -- ones on the diagonal
   * and zeroes elsewhere. The unit matrix is always square.
   */
Matrix Matrix::eye(unsigned int d) {
  Matrix M(d);

  for (unsigned int i = 0; i < d; i++) {
    for (unsigned int j = 0; j < d; j++) {
      if (i == j)
        M.data[i][j] = 1.0;
      else
        M.data[i][j] = 0.0;
    }
  }

  return M;
}


  /**
   * Returns a square matrix of order equal to the number of elements
   * in the input vector. The elements of the input vector are put on
   * the main diagonal.
   */
Matrix Matrix::diag(Vector d) {
  Matrix M(d.getDimension());
  
  for (unsigned int i = 0; i < d.getDimension(); i++) {
    for (unsigned int j = 0; j < d.getDimension(); j++) {
      if (i == j)
        M.data[i][j] = d(i);
      else
        M.data[i][j] = 0.0;
    }
  }
  
  return M;
}


  /**
   * Return a "d" by "d" matrix of ones.
   */
Matrix Matrix::ones(unsigned int d) {
  Matrix M(d, d);
  
  for (unsigned int i = 0; i < d; i++)
    for (unsigned int j = 0; j < d; j++)
      M(i, j) = 1.0;
  
  return M;
}


  /**
   * Return a "m" by "n" matrix of ones.
   */
Matrix Matrix::ones(unsigned int m, unsigned int n) {
  Matrix M(m, n);
  
  for (unsigned int i = 0; i < m; i++)
    for (unsigned int j = 0; j < n; j++)
      M(i, j) = 1.0;
  
  return M;
}


  /**
   * Return a "d" by "d" matrix with random entries, chosen on the
   * interval (0.0, 1.0).
   */
Matrix Matrix::rand(unsigned int d) {
  return rand(d, d);
}


  /**
   * Return a "m" by "n" matrix with random entries, chosen on the
   * interval (0.0, 1.0).
   *
   * Some operating systems don't provide a random() function.  Poor Windows users.
   */
Matrix Matrix::rand(unsigned int m, unsigned int n) {
  Matrix M(m, n);
  
/*
  struct timeval  tv;
  struct timezone tz;

  if (gettimeofday(&tv, &tz) != 0)
    return -1;

  srandom(tv.tv_usec);
  return 0;
*/
  for (unsigned int i = 0; i < m; i++)
    for (unsigned int j = 0; j < n; j++)
//      M.data[i][j] = (1.0 * (random() % RAND_WRAP)) / (1.0 * RAND_MAX);
        M.data[i][j] = (1.0 * ::rand()) / (1.0 * RAND_MAX);
  
  return M;
}


  /**
   * Return a "d" by "d" matrix of zeros.
   */
Matrix Matrix::zeros(unsigned int d) {
  Matrix M(d, d);
  
  for (unsigned int i = 0; i < d; i++)
    for (unsigned int j = 0; j < d; j++)
      M(i, j) = 0.0;
  
  return M;
}


  /**
   * Return a "m" by "n" matrix of zeros.
   */
Matrix Matrix::zeros(unsigned int m, unsigned int n) {
  Matrix M(m, n);
  
  for (unsigned int i = 0; i < m; i++)
    for (unsigned int j = 0; j < n; j++)
      M(i, j) = 0.0;
  
  return M;
}

           
  

  /******************
   * GETTER METHODS *
   ******************/

  /**
   * Get the number of columns in the matrix.
   */
unsigned int Matrix::getCols() const {
  return cols;
}


  /**
   * Get the number of rows in the matrix.
   */
unsigned int Matrix::getRows() const {
  return rows;
}


  /**
   * Extract column "j" from the matric. The column is always
   * extracted as a column.
   */
Vector Matrix::getCol(unsigned int j) const throw(OutOfRange) {
  if (j >= cols)
    throw OutOfRange("getCol() out of range");
  
  Vector v(rows);
  v.setColumn(true);
  for (unsigned int i = 0; i < rows; i++)
    v(i) = data[i][j];
  
  return v;
}


  /**
   * Extract row "i" from the matrix. The row is always extracted as
   * column vector.
   */
Vector Matrix::getRow(unsigned int i) const throw(OutOfRange) {
  if (i >= rows)
    throw OutOfRange("getRow() out of range");

  Vector v(cols);
  v.setColumn(false);
  for (unsigned int j = 0; j < cols; j++)
    v(j) = data[i][j];
  
  return v;
}



  /******************
   * SETTER METHODS *
   ******************/

  /**
   * Set column "j" in the matrix to the contents of vector "v". The
   * vector should be a column vector but since there is no ambiguity,
   * a row vector will do as well.
   */
void Matrix::setCol(unsigned int j, const Vector v)
  throw(DimensionMismatch, OutOfRange) {

  if (rows != v.getDimension())
    throw DimensionMismatch("setCol()");

  if (j >= cols)
    throw OutOfRange("setCol()");
  
  for (unsigned int i = 0; i < rows; i++)
    data[i][j] = v(i);

  return;
}


  /**
   * Set row "i" in the matrix to the contents of vector "v". The
   * vector should be a row vector but since there is no ambiguity a
   * column vector will do as well.
   */
void Matrix::setRow(unsigned int i, const Vector v)
  throw(DimensionMismatch, OutOfRange) {

  if (cols != v.getDimension())
    throw DimensionMismatch("setRow()");

  if (i >= rows)
    throw OutOfRange("setRow()");
  
  for (unsigned int j = 0; j < cols; j++)
    data[i][j] = v(j);

  return;
}
 


  /************************
   * MATRIX MINOR METHODS *
   ************************/

  /**
   * Get the minor of the matrix. The (i, j)-minor is just the matrix
   * but without row "i" and column "j". The minor is used in
   * determinant computation.
   */
Matrix Matrix::getMinor(unsigned int i, unsigned int j) const
  throw(OutOfRange) {

  if ((i >= rows) || (j >= cols))
    throw OutOfRange("getMinor()");

  unsigned int n_i, m_i;
  unsigned int n_j, m_j;
  Matrix M(rows - 1, cols - 1);

  for (n_i = m_i = 0; n_i < (rows - 1); n_i++, m_i++) {
    if (n_i == i)
      m_i++;
    for (n_j = m_j = 0; n_j < (cols - 1); n_j++, m_j++) {
      if (n_j == j)
	m_j++;

      M.data[n_i][n_j] = data[m_i][m_j];
    }
  }

  return M;
}



  /************************
   * SUBMATRIX OPERATIONS *
   ************************/

  /**
   * Get the sub-matrix defined by the upper left corner (i_1, j_1)
   * and the lower right corner (i_2, j_2). If the sub-matrix is
   * bigger than the original matrix, the overshooting space is set to
   * the default value.
   */
Matrix Matrix::getSubMatrix(
  unsigned int i1, unsigned int j1, unsigned int i2, unsigned int j2)
  const throw(OutOfRange) {

  if ((i1 < rows) && (i2 < rows) && (j1 < cols) && (j2 < cols)) {
    unsigned int sub_rows = SIM_MIN((i2 - i1) + 1, (rows - i1) + 1);
    unsigned int sub_cols = SIM_MIN((j2 - j1) + 1, (cols - j1) + 1);

    Matrix M(sub_rows, sub_cols);
      
    for (unsigned int i = 0; i < sub_rows; i++) {
  	  for (unsigned int j = 0; j < sub_cols; j++) {
		M.data[i][j] = data[i + i1][j + j1];
	  }
	}
/*
		bcopy(
        &(data[i + i1][j1]),
        &(M.data[i][0]),
        sub_cols * sizeof(double));
 */   
    return M;
  }
  throw OutOfRange("Matrix::getSubMatrix()");
}


  /**
   * The source matrix "M" can be any dimension. As much as will fit
   * is replaced in the destination matrix, that is "this" matrix. The
   * rest is silently discarded. The method replaces the destination
   * image starting at the indices (i, j).
   */
void Matrix::setSubMatrix(unsigned int i, unsigned int j, const Matrix M)
  throw(OutOfRange) {

  if ((i < rows) && (j < cols)) {
    unsigned int sub_rows = M.rows;
    unsigned int sub_cols = M.cols;

    unsigned int rep_rows = SIM_MIN(rows - i, sub_rows);
    unsigned int rep_cols = SIM_MIN(cols - j, sub_cols);

    for (unsigned int rep_i = 0; rep_i < rep_rows; rep_i++) {
      for (unsigned int rep_j = 0; rep_j < rep_cols; rep_j++)
        data[rep_i + i][rep_j + j] = M.data[rep_i][rep_j];
    }

    return;
  }
  throw OutOfRange("Matrix::setSubMatrix()");
}

  



  /****************
   * MATRIX NORMS *
   ****************/

  /**
   * For matrices (as opposed to vectors), the Euclidian norm is the
   * square root of the maximal eigenvalue of the matrix multiplied by
   * its transpose.
   */
double Matrix::norm2() const {
  double norm;
  Matrix M(*this);
  Vector v;
  
  M = M.transpose() * M;
  v = M.eigenvalues();
    
  norm = v(0);
  for (unsigned int i = 1; i < cols; i++)
    if (norm < v(i))
      norm = v(i);

  return ::sqrt(norm);
}


  /**
   * Maximum norm for matrices. See section 16.2 of Mathematics
   * Handbook for the definition of this norm. It is the maximal sum
   * of the absolute value of the rows.
   */
double Matrix::normMax() const {
  double norm, norm_row;
  
  norm = 0;
  for (unsigned int i = 0; i < rows; i++) {
    norm_row = 0;
    for (unsigned int j = 0; j < cols; j++)
      norm_row += fabs(data[i][j]);

    if (norm_row > norm)
      norm = norm_row;
  }
  
  return norm;
}



  /*******************************************
   * FORWARD & BACKWARD SUBSTITUTION METHODS *
   *******************************************/

  /**
   * Forward substitution. The matrix must be a lower triangular
   * matrix. The elements on the diagonal must be one. This is not
   * checked. 
   */
Vector Matrix::forward(const Vector b, const Vector p)
  const throw(DimensionMismatch) {

  if (rows != cols)
    throw DimensionMismatch("forward(): Not Square Matrix");

  if (rows != b.getDimension())
    throw DimensionMismatch("forward(): Input Vector b");

  if (rows != p.getDimension())
    throw DimensionMismatch("forward(): Permutation Vector p");
  
  Vector y(rows);
  for (unsigned int i = 0; i < rows; i++) {
    y(i) = b(static_cast<unsigned int>(lrint(p(i))));
    for (unsigned int j = 0; j < i; j++)
      y(i) -= data[i][j] * y(j);
  }

  return y;
}


  /**
   * Backward substitution. The matrix must be an upper triangular
   * matrix. The elements on the diagonal do not have to be
   * one. Neither of theses requirements are checked. There is a
   * strange comparison in the outer for loop, but this has to do with
   * the signedness of the index i.
   */
Vector Matrix::backward(const Vector y) const throw(DimensionMismatch) {
  if (rows != cols)
    throw DimensionMismatch("backward(): Not Square Matrix");

  if (rows != y.getDimension())
    throw DimensionMismatch("backward(): Input Vector y");

  Vector x(rows);
  for (unsigned int i = (rows - 1); i < rows; i--) {
    x(i) = y(i);
    for (unsigned int j = (i + 1); j < cols; j++)
      x(i) -= data[i][j] * x(j);
    x(i) /= data[i][i];
  }

  return x;
}


  /**
   * Solve an equation of the type M x y = b for the vector y. This is
   * done by using LUP-decomposition and using forward and backward
   * substitution instead of computing the inverse first.
   */
Vector Matrix::solve(const Vector b)
  const throw(DimensionMismatch, NotInvertible) {

  Matrix M(*this);
  Vector p, x, y;

  try {
    p = M.LUP();
    x = M.forward(b, p);
    y = M.backward(x);
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }

  return y;
}


  /**
   * Solve an equation of the type M x Y = B for the matrix Y. This is
   * done by using LUP-decomposition and using forward and backward
   * substitution instead of computing the inverse first.
   */
Matrix Matrix::solve(const Matrix B)
  const throw(DimensionMismatch, NotInvertible) {

  Matrix M(*this);
  Matrix Y(cols);
  Vector p, b, x, y;

  try {
    p = M.LUP();
    for (unsigned int j = 0; j < cols; j++) {
      b = B.getCol(j);

      x = M.forward(b, p);
      y = M.backward(x);

      Y.setCol(j, y);
    }
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }

  return Y;
}



  /*************************
   * MATRIX DECOMPOSITIONS *
   *************************/

  /**
   * Compute the Cholesky factorisation of the matrix. The matrix must
   * be positive definite and symmetric. The method only uses the
   * diagonal and upper triangle of the input matrix.  The lower
   * triangular is assumed to be the transpose of the upper.  If the
   * input matrix is positive definite, then chol() produces an upper
   * triangular matrix "R" so that "R' x R" equals the input matrix.
   * If the input matrix is not positive definite, an exception is
   * thrown.
   *
   * The brain-damaged windows compilers do not allow declaration of
   * loop indices the way we want.
   *
   * These windows compilers really suck when it comes to compiling things.
   */
Matrix Matrix::chol() const throw(NotPositiveDefinite, NotSymmetric) {
  unsigned int i;
  unsigned int j;
  unsigned int k;
  Matrix M(*this);

  if (cols != rows)
    throw NotSymmetric("Matrix::choleksy()");
  
  for (j = 0; j < cols; j++) {
    for (k = 0; k < j; k++)
      for (i = j; i < cols; i++)
        M.data[i][j] = M.data[i][j] - M.data[i][k] * M.data[j][k];
  
    if (M.data[j][j] <= 0)
      throw NotPositiveDefinite("Matrix::cholesky()");
    else
      M.data[j][j] = ::sqrt(M.data[j][j]);

    for (k = 0; k < cols; k++)
      if (k > j)
        M.data[k][j] = M.data[k][j] / M.data[j][j];
      else if (k < j)
        M.data[k][j] = 0;
  }

  return M;
}


  /**
   * Compute the Cholesky factorisation of the matrix. The matrix must
   * be positive definite and symmetric. Returns two matrices so that
   * "L * U" equals the input matrix.
   */
void Matrix::chol(Matrix *L, Matrix *U)
  const throw (NotPositiveDefinite, NotSymmetric) {

  try {
    (*U) = chol();
    (*L) = U -> transpose();
  } catch (NotPositiveDefinite &e) {
    throw NotPositiveDefinite(e.what());
  } catch (NotSymmetric &e) {
    throw NotSymmetric(e.what());
  }
}


  /**
   * LU-decomposition of the matrix "A" into the input matrices "L",
   * "U" and "P". This is described more in section 16.2 of
   * Mathematics Handbook, and also in the libsim documentation. The
   * matrix has to be square and the output matrices will have to be
   * square and of the same dimension.
   */
Vector Matrix::LUP() throw (DimensionMismatch, NotInvertible) {
  double       pivot, temp;
  unsigned int i, j, k, l;
  Vector       p(rows);

  if (cols != rows)
    throw DimensionMismatch("LU decomposition matrix not square");

    /* Initialise the permutation vector to the unit permutation, that
     * is row "i" has a one in column "i".
     */
  for (i = 0; i < rows; i++)
    p(i) = i;

    /* Main loop. Loop through the rows starting at the first and
     * ending at the second last from the bottom.
     */
  for (k = 0; k < (rows - 1); k++) {

      /* Find the pivot element, that is the greatest element in the
       * column starting at the diagonal element and going down the
       * column to bottom element. The row with the pivot element is
       * column "l".
       */
    l     = 0;
    pivot = 0;
    for (i = k; i < rows; i++) {
      if (fabs(data[i][k]) > pivot) {
        pivot = fabs(data[i][k]);
        l     = i;
      }
    }
    
      /* If the pivot element is zero then the matrix is not
       * invertible. Swap the pivot row with the row currently
       * considered. This is done by exhanging elements in the
       * permutation vector. Then we also need to swap the rows in the
       * matrix we are working on.
       */
    if (fabs(pivot) < std::numeric_limits<double>::min())
      throw NotInvertible("LU Not invertible");
    temp = p(k);
    p(k) = p(l);
    p(l) = temp;

    for (j = 0; j < cols; j++) {
      temp       = data[k][j];
      data[k][j] = data[l][j];
      data[l][j] = temp;
    }
    
      /* This is the LU decomposition step itself. This is discussed
       * further in the documentation.
       */
    for (i = (k + 1); i < rows; i++) {
      data[i][k] = data[i][k] / data[k][k];
      for (j = (k + 1); j < cols; j++)
        data[i][j] -= data[i][k] * data[k][j];
    }
  }
  return p;
}


  /**
   * LUP-decomposition with explicit and separated matrices. This is
   * just normal LUP-decomposition but with the added overhead of
   * separate matrices.
   */
void Matrix::LUP(Matrix *L, Matrix *U, Matrix *P)
  const throw (DimensionMismatch, NotInvertible) {

  if (cols != rows)
    throw DimensionMismatch("LU decomposition matrix not square");
  
  if ((L -> rows != rows) || (L -> cols != cols))
    throw DimensionMismatch("LU wrong dimension of L");

  if ((U -> rows != rows) || (U -> cols != cols))
    throw DimensionMismatch("LU wrong dimension of U");

  if ((P -> rows != rows) || (P -> cols != cols))
    throw DimensionMismatch("LU wrong dimension of P");


    /* Initialise the lower triangulaar matrix to the original matrix
     * we will want to decompose.
     */
  Vector p;
  (*L) = (*this);

    /* Do the decomposition. Convert the permutation vector "p" to the
     * permutation matrix "P". The matrix has a one in row "i" in the
     * column indicated by p(i). Also fix the "L" matrix so that it
     * has ones on the diagonal and zeros in the upper part. Write the
     * part that is not in the "L" matrix to the "U" matrix.
     */
  p = L -> LUP();
  for (unsigned int i = 0; i < rows; i++) {
    for (unsigned int j = 0; j < cols; j++) {
      if (fabs(p(i) - j) < std::numeric_limits<double>::epsilon())
        P -> data[i][j] = 1.0;
      else
        P -> data[i][j] = 0.0;
      
      if (i == j) {
        U -> data[i][j] = L -> data[i][j];
        L -> data[i][j] = 1.0;
      } else if (i < j) {
        U -> data[i][j] = L -> data[i][j];
        L -> data[i][j] = 0.0;
      }
    }
  }
}


  /**
   * QR-decomposition of the matrix a into the input matrices "Q" and
   * "R". "R" is an upper triangular matrix of the same dimension as
   * A. "Q" is a unitary matrix so that A=QR. This is described more
   * in section 16.2 of Mathematics Handbook, and also in the libsim
   * documentation. This code is a translation and modification of a
   * MATLAB function originally written by Jonas Soederberg and Johan
   * Hattne.
   */
void Matrix::QR(Matrix *Q, Matrix *R)
  const throw (DimensionMismatch) {
  
  if ((Q -> rows != rows) || (Q -> cols != rows))
    throw DimensionMismatch("QR wrong dimension of Q");

  if ((R -> rows != rows) || (R -> cols != cols))
    throw DimensionMismatch("QR wrong dimension of R");
  
  Matrix A_temp(*this);
  Matrix R_temp(rows);

  *Q     = eye(rows);
  R_temp = eye(rows);

    /* The loop. Compute the vectors a and e.
     */
  for (unsigned int i = 0; i < (SIM_MIN(rows, cols) - 1); i++) {
    Vector a = A_temp.getCol(0);
    Vector e(rows - i);
    Vector v;

      /* Set the first element of the e vector to the Euclidian norm
       * of the a vector. The v vector will be a normalised linear
       * combination of the a and e vectors. The combination is made
       * such that cancellation is avoided. This is also known as the
       * Householder vector.
       */
    e(0) = a.norm2();
    if (a(0) * e(0) > 1)
      v = a + e;
    else
      v = a - e;
    v.normalise();

      /* H is a temporary matrix that will lead to the Householder
       * transformation. H was originally called Phatt, which is also
       * an island in Monkey Island (TM). LOL.
       */
    Matrix H(rows - i);
    H = eye(rows - i) - (v % v.transpose()) * 2;

      /* A_temp is the transformed matrix. The lower right part of it
       * will be used in the next iteration.
       */
    A_temp = H * A_temp;
    A_temp = A_temp.getSubMatrix(
      (unsigned)1,                (unsigned)1,
      (unsigned)((rows - i) - 1), (unsigned)((cols - i) - 1));

      /* H_temp is the Householder transformation.
       */
    Matrix H_temp;
    H_temp = eye(rows);
    H_temp.setSubMatrix(i, i, H);

      /* Q is built up as we iterate through the loop and consists of
       * H_temp(1) x H_temp(2) x ... x H_temp(k). R_temp is
       * constructed similarly: H_temp(k) x H_temp(k-1) x ... x
       * H_temp(1).
       */
    (*Q)   = (*Q)   * H_temp;
    R_temp = H_temp * R_temp;
  }

    /* Final computation of R.
     */
  (*R) = R_temp * (*this);
}


  
  /******************
   * MATRIX ALGEBRA *
   ******************/
  
  /**
   * Returns the absolute value of the elements in the matrix.
   */
Matrix Matrix::abs() const {
  Matrix M(rows, cols);
  
  
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      M.data[i][j] = fabs(data[i][j]);
  
  return M;
}


  /**
   * Recursive determinant computation. Do development by the first
   * column. This is described in section 4.2 of Mathematics Handbook.
   */
double Matrix::determinant() const throw(DimensionMismatch) {
  double det;
  Matrix M;

  if (rows != cols) {
    throw DimensionMismatch("Matrix Determinant");
  
  } else if (rows == 1) {
    det = data[0][0];

  } else if (rows == 2) {
    det = data[0][0] * data[1][1] - data[0][1] * data[1][0];
    
  } else {
    det = 0;
    for (unsigned int i = 0; i < rows; i++) {
      M = getMinor(i, 0);
      if ((i % 2) == 0)
        det += data[i][0] * M.determinant();
      else
        det -= data[i][0] * M.determinant();
    }
  }
  
  return det;
}   


  /**
   * Returns the main diagonal of the input matrix.
   */
Vector Matrix::diag() const {
  unsigned int d = SIM_MIN(cols, rows);
  Vector       v(d);

  for (unsigned int i = 0; i < d; i++)
    v(i) = data[i][i];
    
  return v;
}


  /**
   * produces a vector "d" of eigenvalues and a full matrix "V" whose
   * columns are the corresponding eigenvectors so that if the input
   * matrix is denoted "X" then "X x V = V x diag(D)". The method
   * works iteratively method for finding eigenvalues using the
   * QR-decomposition. See discussion in the report. Iterate the
   * epsilon value does no longer decrease. The eigenvalues are then
   * the diagonal elements of the A matrix.
   */
void Matrix::eig(Matrix *V, Vector *d) const throw(DimensionMismatch) {
  double epsilon, epsilon_previous, sum;
  Matrix A(*this);
  Matrix Q(rows);
  Matrix R(rows, cols);

  (*V)    = eye(rows);
  epsilon = std::numeric_limits<double>::max();
  do {
    epsilon_previous = epsilon;

    try {
      A.QR(&Q, &R);
      A    = R * Q;
      (*V) = Q * (*V);
      sum  = 0;
      for (unsigned int i = 1; i < A.rows; i++)
        for (unsigned int j = 0; j < i; j++)
          sum = sum + A(i, j) * A(i, j);
      epsilon = ::sqrt(sum);
    } catch (DimensionMismatch& e) {
      throw DimensionMismatch(e.what());
    }
  } while (epsilon < epsilon_previous);

  (*d) = Vector::zeros(rows);
  for (unsigned int i = 0; i < rows; i++)
    (*d)(i) =  A(i, i);
}


  /**
   * Compute the two eigenvalues of a square matrix. One- and
   * two-dimensional matrices are handles separately. The eigenvalues
   * are given in no particular order. The general method sets some
   * initial values and iterates until the sum of the squares of the
   * subdiagonal elements is less than some predefined constant.
   */
Vector Matrix::eigenvalues() const throw(DimensionMismatch) {
  if (rows != cols)
    throw DimensionMismatch("eigenvalue(): Not Square Matrix");
  
    /* Handle the one- and two-dimensional case separately. The
     * derivation of the expressions for the eigenvalues in these
     * cases are given in the libsim documentation.
     */
  if (rows == 1) {
    Vector v(1);
    v(0) = data[0][0];
    return v;
       
  } else if (rows == 2) {
    double a = data[0][0] + data[1][1];
    double b = data[0][0] * data[1][1] - data[0][1] * data[1][0];
    double c = ::sqrt(a * a - b);
    
    Vector v(2);
    v(0) = (a + c) / 2;
    v(1) = (a - c) / 2;
    return v;
  }

    /* Iterative method for finding eigenvalues using the
     * QR-decomposition. See discussion in the report. Iterate the
     * epsilon value does no longer decrease. The eigenvalues are then
     * the diagonal elements of the A matrix.
     */
  double epsilon, epsilon_previous, sum;
  Matrix A(*this);
  Matrix Q(rows);
  Matrix R(rows, cols);

  epsilon = std::numeric_limits<double>::max();
  do {
    epsilon_previous = epsilon;

    try {
      A.QR(&Q, &R);
      A   = R * Q;
      sum = 0;
      for (unsigned int i = 1; i < A.rows; i++)
        for (unsigned int j = 0; j < i; j++)
          sum = sum + A(i, j) * A(i, j);
      epsilon = ::sqrt(sum);
    } catch (DimensionMismatch& e) {
      throw DimensionMismatch(e.what());
    }
  } while (epsilon < epsilon_previous);

  Vector v(rows);
  for (unsigned int i = 0; i < rows; i++)
    v(i) =  A(i, i);
  return v;
}


  /**
   * Compute the eigenvector corresponding to the argument
   * eigenvalue. This is done by inverse iteration as described in the
   * documentation. If the LU-decomposition says that the matrix is
   * not invertible, then we check if we have a zero element on the
   * diagonal.
   */
Vector Matrix::eigenvectors(const double eigenvalue) const
  throw(DimensionMismatch, NotInvertible)  {

  if (cols != rows)
    throw DimensionMismatch("Matrix::eigenvectors()");

  double       epsilon, epsilon_previous;
  unsigned int iterations;
  Matrix       M(*this);
  Vector       b, x, y, p;

  for (unsigned int i = 0; i < cols; i++)
    M.data[i][i] -= eigenvalue;

  try {
    p          = M.LUP();
    y          = (Vector::rand(cols)).normalise();
    epsilon    = std::numeric_limits<double>::max();
    iterations = 0;
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    for (unsigned int i = 0; i < M.rows; i++) {
      if (fabs(M.data[i][i]) < std::numeric_limits<double>::min()) {
        y    = Vector::zeros(M.cols);
        y(i) = 1;
        return y;
      }
    }

    throw NotInvertible(e.what());
  }

  do {
    try {
      epsilon_previous = epsilon;
      iterations++;

      b = y;
      x = M.forward(b, p);
      y = M.backward(x);
      y.normalise();

      epsilon = (b - y).norm2();
    } catch (DimensionMismatch &e) {
      throw DimensionMismatch(e.what());
    } catch (NotInvertible &e) {
      throw NotInvertible(e.what());
    }
  } while (
    (epsilon    < epsilon_previous) &&
    (iterations < SIM_MAX_INVERSE_ITERATIONS));

  return y;
}


  /**
   * Compute the eigenvectors corresponding to the eigenvalues as
   * given in the argument vector.
   */
Matrix Matrix::eigenvectors(const Vector eigenvalues) const
  throw(DimensionMismatch, NotInvertible)  {

  Matrix M(rows, eigenvalues.getDimension());
  
  try {
    for (unsigned int i = 0; i < eigenvalues.getDimension(); i++)
      M.setCol(i, eigenvectors(eigenvalues(i)));
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }
  
  return M;
}


  /**
   * Flip matrix in left/right direction.
   */
Matrix Matrix::flipLR() const {
  Matrix M(rows, cols);
  
  for (unsigned int j = 0; j < cols; j++)
    for (unsigned int i = 0; i < rows; i++)
      M.data[i][j] = data[i][(cols - 1) - j];

  return M;
}

    
  /**
   * Flip matrix in up/down direction.
   */
Matrix Matrix::flipUD() const {
  Matrix M(rows, cols);
  
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      M.data[i][j] = data[(rows - 1) - i][j];

  return M;
}
  

  /**
   * Compute the inverse of a matrix. This will require a square
   * matrix. The one- and two-dimensional cases are handled separtely.
   */
Matrix Matrix::inv() const throw(DimensionMismatch, NotInvertible) {
  if (rows != cols)
    throw DimensionMismatch("determinant(): Not Square Matrix");

    /* Handle the one- and two-dimensional case separately.
     */
  if (cols == 1) {
    Matrix M(1);
    if (fabs(data[0][0]) < std::numeric_limits<double>::min())
      throw NotInvertible("determinant(): 1-dim not invertible");
        
    M.data[0][0] = 1.0 / data[0][0];
    return M;

  } if (cols == 2) {
    Matrix M(2);
    double det = determinant();
    
    if (fabs(det) < std::numeric_limits<double>::min())
      throw NotInvertible("determinant(): 2-dim not invertible");
      
    M.data[0][0] =  1.0 * data[1][1] / det;
    M.data[0][1] = -1.0 * data[0][1] / det;
    M.data[1][0] = -1.0 * data[1][0] / det;
    M.data[1][1] =  1.0 * data[0][0] / det;
    return M;

  }
  
    /* The general case. Do a LUP decomposition and then solve each
     * system of equations in turn. At each equation the b vector
     * should be the unit vector with a one in the "j"th position.
     */
  Matrix I(rows);
  Matrix M(*this);
  Vector b(rows);
  Vector p, x, y;

  try {
    p = M.LUP();

    for (unsigned int j = 0; j < cols; j++) {
      for (unsigned int i = 0; i < rows; i++) {
        if (i == j)
          b(i) = 1.0;
        else
          b(i) = 0.0;
      }
      
      y = M.forward(b, p);
      x = M.backward(y);

      I.setCol(j, x);
    }
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }
    
  return I;
}


  /**
   * Returns an orthonormal basis for the range of the matrix, that is
   * if the return values of this method is denoted by "Q", then "Q' *
   * Q = I".  The columns of "Q" span the same space as the columns of
   * the matrix to orthogonalise. The number of columns of "Q" is the
   * rank of the matrix.
   */
Matrix Matrix::orth() const throw(DimensionMismatch) {
  Matrix M(*this);
  Vector a, q;

  try {
    if (cols > 0)
      M.setCol(0, (M.getCol(0)).normalise());

    for (unsigned int j = 0; j < M.cols; j++) {
      q = (M.getCol(j)).normalise();
    
      for (unsigned int k = j + 1; k < M.cols; k++) {
        a = M.getCol(k);
        M.setCol(k, (a - q * (q * a)).normalise());
      }
    }
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  }

  return M;
}


  /**
   * Returns a matrix where every element is raised to the power of
   * the argument.
   */
Matrix Matrix::pow(double p) const {
  Matrix M(rows, cols);
  
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      M.data[i][j] = ::pow(data[i][j], p);
  
  return M;
}
  

  /**
   * Transform the matrix to reduced row echelon form. This is
   * discussed further in section 4.1 of Mathematics Handbook. First
   * find the row with the maximal entry on the diagonal. Write a
   * scaled version of this row to the "N" matrix and delete it from
   * the "M" matrix. Subtract a scaled version of the row from all the
   * remaining rows in the "M" matrix.
   */
Matrix Matrix::rref(double tolerance) const {
  Matrix       M(*this), N(rows, cols);
  double       max_data, temp;
  unsigned int max_index, i, j, k, l;

  l = 0;
  for (k = 0; k <  SIM_MIN(cols, rows); k++) {
    max_data  = 0;
    max_index = 0;
    for (i = l; i < rows; i++) {
      if (fabs(M.data[i][k]) > fabs(max_data)) {
        max_data  = M.data[i][k];
        max_index = i;
      }
    }

    if (fabs(max_data) > tolerance) {
      for (j = k; j < cols; j++) {
        temp                 = M.data[max_index][j] / max_data;
        M.data[max_index][j] = M.data[l][j];
        M.data[l][j]         = temp;
      }
      
      for (i = l + 1; i < rows; i++) {
        for (j = k; j < cols; j++)
          M.data[i][j] -= M.data[l][j] * M.data[i][j];
      }

      l++;
    } else {
      for (i = 0; i < rows; i++)
        M.data[i][k] = 0;
    }
  }

  for (k = SIM_MIN(cols, rows) - 1;
       k < SIM_MIN(cols, rows);
       k--) {

    if (fabs(M.data[k][k]) >= std::numeric_limits<double>::min()) {
      for (i = 0; i < k; i++)
        M.data[i][k] = 0;
    }
  }

  return M;
}


  /**
   * Sorts each column vector of the matrix in ascending order using
   * the system's quicksort implementation. When more than one element
   * has the same value, the order of the elements are preserved in
   * the sorted result and the indexes of equal elements will be
   * ascending in any index matrix.
   */
extern "C" 
{
  static int
  sim_matrix_compare(const void *i, const void *j)
  {
    if (*((double *)i) > *((double *)j))
      return 1;
    else if (*((double *)i) < *((double *)j))
      return -1;
    return 0;
  }
}

Matrix Matrix::sort() const {
  unsigned int  i, j;
  Matrix        M(rows, cols);
  double       *base;
  
  base = new double[rows];
  for (j = 0; j < cols; j++) {
    for (i = 0; i < rows; i++)
      base[i] = data[i][j];

    qsort(base, rows, sizeof(double), sim_matrix_compare);

    for (i = 0; i < rows; i++)
      M.data[i][j] = base[i];
  }
  delete[] base;

  return M;
}


  /**
   * Return the matrix where each element is the square root of the
   * original corresponding element.
   */
Matrix Matrix::sqrt() const {
  Matrix M(rows, cols);
  
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      if (data[i][j] > 0)
        M.data[i][j] = ::sqrt(data[i][j]);
      else
        M.data[i][j] = 0;
      
  return M;
}


  /**
   * Matrix square root. Returns the principal square root of the
   * matrix, that is if the return values of this method is denoted by
   * "X", then "X x X" equals the input matrix. Furthermore, "X" is
   * the unique square root for which every eigenvalue is
   * nonnegative. If the matrix is singular then it may not have a
   * square root. Exceptions are thrown on errors.
   */
Matrix Matrix::sqrtm() const
  throw(DimensionMismatch, NotInvertible, NotPositiveDefinite, NotSymmetric) {

  Matrix M(rows, cols);
  Matrix G, U, V;
  Vector s;
  
  try {
    G = chol();
    G.svd(&U, &s, &V);
    M = U * diag(s) * (U.transpose());
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  } catch (NotPositiveDefinite &e) {
    throw NotPositiveDefinite(e.what());
  } catch (NotSymmetric &e) {
    throw NotSymmetric(e.what());
  }
       
  return M;
}


  /**
   * Singular value decomposition. Returns two matrices "U" and "V"
   * and a vector "s". "s" is of the same dimension as the input
   * matrix, with elements. "U" and "V" are unitary matrices so that
   * "U x diag(S) x V'" equals the input matrix. Also see section 4.5
   * in Mathematics Handbook.
   */
void Matrix::svd(Matrix *U, Vector *s, Matrix *V)
  const throw(DimensionMismatch, NotInvertible) {

  double mu;
  Matrix ATA;

  try {
    ATA = ((*this).transpose()) * (*this);
    ATA.eig(V, s);

    (*U) = Matrix(V -> rows, s -> getDimension());
    for (unsigned int i = 0; i < s -> getDimension(); i++) {
      mu = (*s)(i) = ::sqrt((*s)(i));
      if (fabs(mu) >= std::numeric_limits<double>::min())
        U -> setCol(i, ((*this) * (V -> getCol(i))) / mu);
    }

  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }
}


  /**
   * Transpose a matrix. Exchanage the rows and the columns of the
   * matrix. This operation can certainly not be performed inplace, as
   * the dimensions may be different.
   */
Matrix Matrix::transpose() const {
  Matrix M(cols, rows);
  
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      M.data[j][i] = data[i][j];

  return M;
}


  
  /*********************
   * MATRIX STATISTICS *
   *********************/

  /**
   * For matrices where each row is an observation and each column a
   * variable, this method returns the covariance matrix. The mean is
   * removed from each column before calculating the result. This
   * method normalises by (N - 1) where N is the number of
   * observations. This makes cov() the best unbiased estimate of the
   * covariance matrix if the observations are from a normal
   * distribution.
   */
Matrix Matrix::cov() const {
  unsigned int i, j, k;
  double       covariance, mean;
  Matrix       M(cols, cols);
  Matrix       N(rows, cols);

  for (j = 0; j < cols; j++) {
    mean = 0;
    for (i = 0; i < rows; i++)
      mean += data[i][j];

    mean /= rows;
    for (i = 0; i < rows; i++)
      N.data[i][j] = data[i][j] - mean;
  }

  for (i = 0; i < cols; i++) {
    for (j = i; j < cols; j++) {
      covariance = 0;
      for (k = 0; k < rows; k++)
        covariance += N.data[k][i] * N.data[k][j];
      
      covariance   /= (rows - 1);
      M.data[i][j]  = covariance;
      M.data[j][i]  = covariance;
    }
  }

  return M;
}

  
  /**
   * Return a vector containing the maximum element from each column.
   */
Vector Matrix::max() const {
  Vector v(cols);
  
  for (unsigned int j = 0; j < cols; j++) {
    v(j) = -1.0 * std::numeric_limits<double>::max();
    
    for (unsigned int i= 0; i < rows; i++)
      if (data[i][j] > v(j))
        v(j) = data[i][j];
  }
  
  return v;
}


  /**
   * Return the vector containing the mean value of each column.
   */
Vector Matrix::mean() const {
  Vector v(cols);
  
  for (unsigned int j = 0; j < cols; j++) {
    v(j) = 0;
    for (unsigned int i = 0; i < rows; i++)
      v(j) += data[i][j];
    v(j) /= rows;
  }
  
  return v;
}


  /**
   * Return a vector containing the minimum element from each column.
   */
Vector Matrix::min() const {
  Vector v(cols);
  
  for (unsigned int j = 0; j < cols; j++) {
    v(j) = +1.0 * std::numeric_limits<double>::max();
    
    for (unsigned int i= 0; i < rows; i++)
      if (data[i][j] < v(j))
        v(j) = data[i][j];
  }
  
  return v;
}


  /**
   * Return a vector containing the standard deviation of each
   * column. Normalised by (N-1) where "N" is the length of the column
   * vectors. This makes std() the best unbiased estimate of the
   * variance if the vector is a sample from a normal distribution.
   */
Vector Matrix::std() const {
  unsigned int i, j;
  double       mu;
  double       sigma;
  Vector       v(cols);

  if (rows < 1)
    return v;
  
  for (j = 0; j < cols; j++) {
    mu = 0;
    for (i = 0; i < rows; i++)
      mu += data[i][j];
  
    mu    = mu / rows;
    sigma = 0;
    for (i = 0; i < rows; i++)
      sigma += (data[i][j] - mu) * (data[i][j] - mu);
    
    v(j) = ::sqrt(sigma / (rows - 1));
  }
  
  return v;
}


  /**
   * Return a vector containing the sum over each column.
   */
Vector Matrix::sum() const {
  Vector v(cols);

  for (unsigned int j = 0; j < cols; j++) {
    v(j) = 0;

    for (unsigned int i = 0; i < rows; i++)
      v(j) += data[i][j];
  }

  return v;
}

 
 /********************
  * MATRIX OPERATORS *
  *******************/

  /**
   * Unary minus.
   */
Matrix Matrix::operator- () const {
  if(*this == 0) {
    return 0;

  } else {
    Matrix N(*this);
    for (unsigned int i = 0; i < rows; i++)
      for (unsigned int j = 0; j < cols; j++)
        N.data[i][j] *= -1.0;
    return N;
  }
}


  /**
   * Binary minus with a real number. This is just subtraction of all
   * the matrix components.
   */
Matrix Matrix::operator- (const double d) const {
  Matrix N(*this);

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] -= d;
  return N;
}


  /**
   * Binary minus with another matrix. This is matrix subtraction --
   * the dimensions must match.
   */
Matrix Matrix::operator- (const Matrix &M) const throw(DimensionMismatch) {
  if ((cols != M.cols) || (rows != M.rows))
    throw DimensionMismatch("Matrix-Matrix Subtraction");

  Matrix N(*this);
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] -= M.data[i][j];
  return N;
}


  /**
   * Decrement operator for reals.
   */
const Matrix& Matrix::operator-= (const double d) {
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] -= d;
  return *this;
}


  /**
   * Decrement operator for matrices. The dimensions must match.
   */
const Matrix& Matrix::operator-= (const Matrix &M) throw(DimensionMismatch) {
  if ((cols != M.cols) || (rows != M.rows))
    throw DimensionMismatch("Matrix-Matrix Subtraction");

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] -= M.data[i][j];
  return *this;
}


  /**
   * Binary plus with a real number. This is just addition of all the
   * matrix components.
   */
Matrix Matrix::operator+ (const double d) const {
  Matrix N(*this);

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] += d;
  return N;
}


  /**
   * Binary plus with another matrix. This is matrix addition --
   * the dimensions must match.
   */
Matrix Matrix::operator+ (const Matrix &M) const throw(DimensionMismatch) {
  if ((cols != M.cols) || (rows != M.rows))
    throw DimensionMismatch("Matrix-Matrix Subtraction");

  Matrix N(*this);
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] += M.data[i][j];
  return N;
}


  /**
   * Increment operator for reals.
   */
const Matrix& Matrix::operator+= (const double d) {
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] += d;
  return *this;
}


  /**
   * Increment operator for matrices. The dimensions must match.
   */
const Matrix& Matrix::operator+= (const Matrix &M) throw(DimensionMismatch) {
  if ((cols != M.cols) || (rows != M.rows))
    throw DimensionMismatch("Matrix-Matrix Subtraction");

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] += M.data[i][j];
  return *this;
}


  /**
   * Binary multiplication with a real number. This is scaling. Scale
   * all the components of the matrix.
   */
Matrix Matrix::operator* (const double d) const {
  Matrix N(*this);

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] *= d;
  return N;
}


  /**
   * Binary multiplication. This is the matrix product. The dimensions
   * must match. The element at position (i, j) is the scalar vector
   * product of the i:th row of the left matrix with the j:th column
   * of the right matrix.
   */
Matrix Matrix::operator* (const Matrix &M) const throw(DimensionMismatch) {
  if (cols != M.rows)
    throw DimensionMismatch("Matrix-Matrix Product");

  Matrix N(rows, M.cols);
  Vector v, w;

  for (unsigned int i = 0; i < N.rows; i++) {
    for (unsigned int j = 0; j < N.cols; j++) {
      
      N.data[i][j] = 0;
      for (unsigned int k = 0; k < cols; k++)
        N.data[i][j] += data[i][k] * M.data[k][j];
      
    }
  }
  
  return N;
}


  /**
   * Binary multiplication with a vector. Even though this is only
   * allowed for column vectors, it introduces no ambiguity to allow
   * the operation for row vectors as well. The dimensions must still
   * match, though. The product is always a column vector.
   */
Vector Matrix::operator* (const Vector &v) const throw(DimensionMismatch) {
  if (cols != v.getDimension())
    throw DimensionMismatch("Matrix-Vector Product");

  Vector w(rows);
  for (unsigned int i = 0; i < rows; i++) {
    w(i) = 0;
    for (unsigned int j = 0; j < cols; j++)
      w(i) += data[i][j] * v(j);
  }

  w.setColumn(true);
  return w;
}


  /**
   * Factor operator with a real number. This is scaling. Scale all
   * the components of the matrix.
   */
const Matrix& Matrix::operator*= (const double d) {
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] *= d;
  return *this;
}


  /**
   * Factor operator with another matrix.
   */
const Matrix& Matrix::operator*= (const Matrix &M) throw(DimensionMismatch) {
  try {
    *this = (*this) * M;
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  }
  
  return *this;
}


  /**
   * Binary division with a real number. This is scaling. Scale
   * all the components of the matrix.
   */
Matrix Matrix::operator/ (const double d) const {
  Matrix N(*this);

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      N.data[i][j] /= d;
  return N;
}


  /**
   * Binary division. This is the matrix product from the right with
   * the inverse. The dimensions must match.
   */
Matrix Matrix::operator/  (const Matrix &M)
  const throw(DimensionMismatch, NotInvertible) {

  try {
    return (*this) * M.inv();
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }
}

  
  /**
   * Factor operator with a real number. This is scaling. Scale all
   * the components of the matrix.
   */
const Matrix& Matrix::operator/= (const double d) {
  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      data[i][j] /= d;
  return *this;
}


  /**
   * Factor operator with another matrix.
   */
const Matrix& Matrix::operator/= (const Matrix &M)
  throw(DimensionMismatch, NotInvertible) {

  try {
    *this = (*this) / M;
  } catch (DimensionMismatch &e) {
    throw DimensionMismatch(e.what());
  } catch (NotInvertible &e) {
    throw NotInvertible(e.what());
  }

  return *this;
}


  /**
   * Element accessor operator.
   */
double Matrix::operator() (unsigned int i, unsigned int j) const
  throw(OutOfRange) {

  if((i >= rows) || (j >= cols))
    throw OutOfRange("Matrix () Operator");

  return data[i][j];
}


  /**
   * Element accessor operator. This is for element assignment using
   * the accessor operator.
   */
double &Matrix::operator() (unsigned int i, unsigned int j)
  throw(OutOfRange) {

  if((i >= rows) || (j >= cols))
    throw OutOfRange("Matrix () Operator");

  return data[i][j];
}


  /**
   * Assignment operator. Reallocate the memory for the matrix only if
   * the dimensions do not match.
   */
const Matrix& Matrix::operator= (const Matrix &M) {
  if (this != &M) {
    if ((cols != M.cols) || (rows != M.rows)) {
      release();
      cols = M.cols;
      rows = M.rows;
      data = new double *[rows];
      for (unsigned int i = 0; i < rows; i++)
	data[i] = new double[cols];
    }

    for (unsigned int i = 0; i < rows; i++)
      for (unsigned int j = 0; j < cols; j++)
        data[i][j] = M.data[i][j];
  }

  return *this;
}


  /**
   * Equality operator. A matrix is equal to a real number if it is
   * one-dimensional and its single component is equal to the number.
   */
bool Matrix::operator== (const double d) const {
  return cols == 1 && rows == 1
    && fabs(data[0][0] - d) < std::numeric_limits<double>::epsilon();
}


  /**
   * Equality operator. Two matrices are equal if and only if their
   * dimensions match and all their components are equal.
   */
bool Matrix::operator== (const Matrix &M) const {
  if ((cols != M.cols) || (rows != M.rows))
    return false;

  for (unsigned int i = 0; i < rows; i++)
    for (unsigned int j = 0; j < cols; j++)
      if (fabs(data[i][j] - M.data[i][j])
        >= std::numeric_limits<double>::epsilon())
        return false;
  
  return true;
}


  /**
   * Inequality operator.
   */
bool Matrix::operator!= (const double d) const {
  return !(*this == d);
}


  /**
   * Inequality operator for matrices.
   */
bool Matrix::operator!= (const Matrix &M) const {
  return !(*this == M);
}


  /**
   * Output operator. Sort of looks like MATLAB output when it is
   * done.
   */
namespace MesoRD {
  std::ostream &operator<< (std::ostream &os, const Matrix &M) {
    os << "[";
    for (unsigned int i = 0; i < M.rows; i++) {
      for (unsigned int j = 0; j < M.cols; j++) {
        os << M.data[i][j];
        if (j != (M.cols - 1))
          os << " ";
      }

      if (i != (M.rows - 1))
        os << "; ";
    }
    os << "]";
    
    return os;
  }
}
