/* The Matrix class.
 *
 * Copyright 2001-2004 Johan Hattne
 *
 * This file is part of libsim.
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Matrix.hpp 552 2008-05-08 09:32:53Z davidfange $
 */

/*
 * The matrix class. This is restricted to real matrices of any
 * dimension. Note that a vector, defined in the Vector class, is just
 * a special case of a matrix where either the number of rows or the
 * number of columns is equal to one. General convention: Matrices
 * should be held in uppercase variable.
 */

#ifndef __Matrix_H
#define __Matrix_H

#include <iostream>

#include "Exception.hpp"

#ifdef WIN32
#pragma warning(disable: 4290)
#endif
  /* We will need access to the Vector class, but we don't want to
   * include the header file and cause circular inclusion problems.
   */

namespace MesoRD {
  class Vector;

  class Matrix {
  private:
    unsigned int   cols;
    unsigned int   rows;
    double       **data;


  public:
 

      /* Constructor and destructor.
       */
    Matrix();
    Matrix(const Matrix &M);
    Matrix(unsigned int d);
    Matrix(unsigned int m, unsigned int n);
    ~Matrix();


      /* General management methods.
       */
    void          release();
    static Matrix eye(unsigned int d);
    static Matrix diag(Vector d);
    static Matrix ones(unsigned int d);
    static Matrix ones(unsigned int m, unsigned int n);
    static Matrix rand(unsigned int d);
    static Matrix rand(unsigned int m, unsigned int n);
    static Matrix zeros(unsigned int d);
    static Matrix zeros(unsigned int m, unsigned int n);
    

      /* Getter Methods.
       */
    unsigned int getCols()              const;
    unsigned int getRows()              const;
    Vector       getCol(unsigned int j) const throw(OutOfRange);
    Vector       getRow(unsigned int i) const throw(OutOfRange);
    

    /* Setter Methods.
     */
    void setCol(unsigned int j, const Vector v)
      throw(DimensionMismatch, OutOfRange);
    void setRow(unsigned int i, const Vector v)
      throw(DimensionMismatch, OutOfRange);


      /* Matrix Minor Methods.
       */
    Matrix getMinor(unsigned int i, unsigned int j) const
      throw(OutOfRange);
    

      /* Submatrix Operations.
       */
    Matrix getSubMatrix(
      unsigned int i1, unsigned int j1, unsigned int i2, unsigned int j2)
      const throw(OutOfRange);
    void   setSubMatrix(unsigned int i, unsigned int j, const Matrix M)
      throw(OutOfRange);

  
      /* Matrix Norms.
       */
    double norm2()   const;
    double normMax() const;
    

      /* Forward & Backward Substitution Methods.
       */
    Vector forward(const Vector b, const Vector p)
      const throw(DimensionMismatch);
    Vector backward(const Vector y)
      const throw(DimensionMismatch);
    Vector solve(const Vector b)
      const throw(DimensionMismatch, NotInvertible);
    Matrix solve(const Matrix B)
      const throw(DimensionMismatch, NotInvertible);


      /* Matrix Decompositions.
       */
    Matrix chol()
      const throw (NotPositiveDefinite, NotSymmetric);
    void   chol(Matrix *L, Matrix *U)
      const throw (NotPositiveDefinite, NotSymmetric);
    Vector LUP()
      throw (DimensionMismatch, NotInvertible);
    void   LUP(Matrix *L, Matrix *U, Matrix *P)
      const throw (DimensionMismatch, NotInvertible);
    void   QR(Matrix *Q, Matrix *R)
      const throw (DimensionMismatch);


      /* Matrix Algebra.
       */
    Matrix abs()
      const;
    double determinant()
      const throw(DimensionMismatch);
    Vector diag()
      const;
    void   eig(Matrix *E, Vector *d)
      const throw(DimensionMismatch);
    Vector eigenvalues()
      const throw(DimensionMismatch);
    Vector eigenvectors(const double eigenvalue)
      const throw(DimensionMismatch, NotInvertible);
    Matrix eigenvectors(const Vector eigenvalues)
      const throw(DimensionMismatch, NotInvertible);
    Matrix flipLR()
      const;
    Matrix flipUD()
      const;
    Matrix inv()
      const throw(DimensionMismatch, NotInvertible);
    Matrix orth()
      const throw(DimensionMismatch);
    Matrix pow(double p)
      const;
    Matrix rref(double tolerance)
      const;
    Matrix sort()
      const;
    Matrix sqrt()
      const;
    Matrix sqrtm()
      const
      throw(DimensionMismatch, NotInvertible,
        NotPositiveDefinite, NotSymmetric);
    void   svd(Matrix *U, Vector *s, Matrix *V)
      const throw(DimensionMismatch, NotInvertible);
    Matrix transpose()
      const;
  

      /* Matrix Statistics.
       */
    Matrix cov()  const;
    Vector max()  const;
    Vector mean() const;
    Vector min()  const;
    Vector std()  const;
    Vector sum()  const;
    
  
      /* Operators.
       */
    Matrix        operator-  ()                               const;
    Matrix        operator-  (const double  d)                const;
    Matrix        operator-  (const Matrix &M)                const
      throw(DimensionMismatch);
    const Matrix& operator-= (const double  d);
  const Matrix& operator-= (const Matrix &M)
    throw(DimensionMismatch);
    Matrix        operator+  (const double  d)                const;
    Matrix        operator+  (const Matrix &M)                const
      throw(DimensionMismatch);
    const Matrix& operator+= (const double  d);
    const Matrix& operator+= (const Matrix &M)
      throw(DimensionMismatch);
    Matrix        operator*  (const double  d)                const;
    Matrix        operator*  (const Matrix &M)                const
    throw(DimensionMismatch);
    Vector        operator*  (const Vector &v)                const
      throw(DimensionMismatch);
    const Matrix& operator*= (const double  d);
    const Matrix& operator*= (const Matrix &M)
      throw(DimensionMismatch);
    Matrix        operator/  (const double  d)                const;
    Matrix        operator/  (const Matrix &M)                const
      throw(DimensionMismatch, NotInvertible);
    const Matrix& operator/= (const double  d);
    const Matrix& operator/= (const Matrix &M)
      throw(DimensionMismatch, NotInvertible);
    double        operator() (unsigned int i, unsigned int j) const
      throw(OutOfRange);
    double&       operator() (unsigned int i, unsigned int j)
      throw(OutOfRange);
    const Matrix& operator=  (const Matrix& v);
    bool          operator== (const double  d)                const;
    bool          operator== (const Matrix &M)                const;
    bool          operator!= (const double  d)                const;
    bool          operator!= (const Matrix &M)                const;
    
    friend std::ostream &operator<< (std::ostream &os, const Matrix &M);
  };
}

#endif
