/* The Vector class.
 *
 * Copyright 2001-2004 Johan Hattne
 *
 * This file is part of libsim.
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Vector.cpp 541 2008-04-12 18:52:41Z davidfange $
 */

/*
 * The Vector class. Contains some basic linear algebra functions. In
 * general all operations are permitted on row and column vectors
 * alike. There is no ambiguity in this relaxation of "real" linear
 * algebra.
 */

//#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
//#include <strings.h> // Not on windows

#include <cstdarg>
#include <limits>

#include "Matrix.hpp"
#include "Vector.hpp"



  /**
   * C-Style defines.
   */
#if(LONG_MAX <= RAND_MAX)
#  define RAND_WRAP LONG_MAX
#else
#  define RAND_WRAP (RAND_MAX + 1)
#endif


 /**
 * Defines the useful macros SIM_MAX, SIM_MIN and SIM_RINT, if they
 * are not already defined.
 */
#ifndef SIM_MAX
#  define SIM_MAX(a, b) \
     (a > b ? a : b)
#endif

#ifndef SIM_MIN
#  define SIM_MIN(a, b) \
     (a < b ? a : b)
#endif

#ifdef SIM_RINT
#  define SIM_RINT(a) \
     (fmod(a, 1.0) >= 0.5 ? ceil(a) : floor(a))
#endif

 /* The namespace to use.
   */
using namespace MesoRD;



  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /**
   * Default constructor. Initialise the dimension to zero and set the
   * data pointer to NULL. Vectors are column vectors by default. One
   * should not ever forget to release any vectors allocated,
   * otherwise there is leakage of memory. The destructor should take
   * care of this.
   */
Vector::Vector() {
  dimension = 0;
  data      = NULL;
  column    = true;
}


  /**
   * Copy constructor. Do deep copy. This will allow using this
   * constructor for cloning vectors.
   */
Vector::Vector(const Vector &v) {
  dimension = v.dimension;
  column    = v.column;
  
  data = new double[dimension];
  for (unsigned int i = 0; i < dimension; i++)
    data[i] = v.data[i];
  
//  bcopy(v.data, data, dimension * sizeof(double));
}


  /**
   * Value constructor. Vectors are column vectors by default. Just
   * allocate space for a vector of the dimension indicated and
   * initialise all elements to zero.
   */
Vector::Vector(size_t l) {
  dimension = static_cast<unsigned int>(l);
  column    = true;
  
  data = new double[dimension];
  for (unsigned int i = 0; i < dimension; i++)
    data[i] = 0.0;
//  bzero(data, dimension * sizeof(double));
}


  /**
   * Destructor. Free the data in the vector if there is any.
   */
Vector::~Vector() {
  release();
}



  /******************************
   * GENERAL MANAGEMENT METHODS *
   ******************************/

  /**
   * Free the data if there is any. Then set the dimension and the
   * pointer to zero and NULL respectively to avoid confusion. This
   * makes it possible to check before trying to free the same chunk
   * of memory again.
   */
void Vector::release() {
  if (data != NULL)
    delete data;
  
  dimension = 0;
  data      = NULL;
}

  
  /**
   * Return a "l"-dimensional vector of ones.
   */
Vector Vector::ones(size_t l) {
  Vector v(l);
  
  for (unsigned int i = 0; i < l; i++)
    v(i) = 1.0;
  
  return v;
}


  /**
   * Return a "l"-dimensional vector with random entries, chosen on
   * the interval (0.0, 1.0).
   */
Vector Vector::rand(size_t l) {
  Vector v(l);
  
/*
  struct timeval  tv;
  struct timezone tz;

  if (gettimeofday(&tv, &tz) != 0)
    return -1;

  srandom(tv.tv_usec);
  return 0;
*/
  for (unsigned int i = 0; i < l; i++)
//    v.data[i] = (1.0 * (random() % RAND_WRAP)) / (1.0 * RAND_MAX); // Poor windows users
    v.data[i] = (1.0 * ::rand()) / (1.0 * RAND_MAX);
  
  return v;
}

  /**
   * Return a "l"-dimensional vector of zeros.
   */
Vector Vector::zeros(size_t l) {
  Vector v(l);
  
  for (unsigned int i = 0; i < l; i++)
    v(i) = 0.0;
  
  return v;
}



  /******************
   * GETTER METHODS *
   ******************/

  /**
   * Get the dimension, or the dimension, of the vector. This is the
   * same as the number of elements in the vector.
   */
unsigned int Vector::getDimension() const {
  return dimension;
}


  /**
   * Return true if the vector is a column vector. If the vector is a
   * row vector return false.
   */
bool Vector::isColumn() const {
  return column;
}



  /******************
   * SETTER METHODS *
   ******************/

  /**
   * Set the vector to be a column vector (input is "true") or a row
   * vector (input is "false").
   */
void Vector::setColumn(bool c) {
  column = c;
}



  /************************
   * SUBVECTOR OPERATIONS *
   ************************/

  /**
   * Get the sub-vector defined by the left index "i1" and the lower
   * right index "i2". If the sub-vector is bigger than the original
   * vector, the overshooting space is set to the default value.
   */
Vector Vector::getSubVector(unsigned int i1, unsigned int i2)
  const throw(OutOfRange) {

  if ((i1 < dimension) && (i2 < dimension)) {
    unsigned int sub_dimension = SIM_MIN((i2 - i1) + 1, dimension - i1);

    Vector v(sub_dimension);
    for (unsigned int i = 0; i < sub_dimension; i++)
		v.data[i] = data[i1 + i];
 //	bcopy(&(data[i1]), &(v.data[0]), sub_dimension * sizeof(double));
    
    return v;
  }
  throw OutOfRange("Vector::getSubVector()");
}


  /**
   * The source matrix "M" can be any dimension. As much as will fit
   * is replaced in the destination matrix, that is "this" matrix. The
   * rest is silently discarded. The method replaces the destination
   * image starting at the indices (i, j).
   */
void Vector::setSubVector(unsigned int i, const Vector v)
  throw(OutOfRange) {
  

  if (i < dimension) {
    unsigned int sub_dimension = v.dimension;
    unsigned int rep_dimension = SIM_MIN(dimension - i, sub_dimension);
    
	for (unsigned int j = 0; j < rep_dimension; j++)
		data[i + j] = v.data[j];
 //   bcopy(&(v.data[0]), &(data[i]), rep_dimension * sizeof(double));
  }
  else{  
    throw OutOfRange("Vector::setSubVector()");
  }
  
}



  /****************
   * VECTOR NORMS *
   ****************/

  /**
   * Euclidean norm. This is just a p-norm for the special case where
   * p=2. Since it is the most common norm needed here, this method
   * may be useful -- it is faster than invoking normp(v, 2).
   */
double Vector::norm2() const {
  double norm;

  norm = 0;
  for (unsigned int i = 0; i < dimension; i++)
    norm += data[i] * data[i];

  return ::sqrt(norm);
}


  /**
   * Calculate the p-norm of a vector. See Mathematics Handbook,
   * section 16.2 for the definition of vector and matrix norms.
   */
double Vector::normp(double p) const {
  double norm;

  if (fabs(p) < std::numeric_limits<double>::min())
    return 0;

  norm = 0;
  for (unsigned i = 0; i < dimension; i++)
    norm += ::pow(data[i], p);
  return ::pow(norm, 1.0 / p);
}

    
  /**
   * Maximum vector norm. See section 16.2 of Mathematics Handbook for
   * the definition of this norm.
   */
double Vector::normMax() const {
  double norm;
  
  norm = 0;
  for (unsigned int i = 0; i < dimension; i++) {
    if (fabs(data[i]) > norm)
      norm = fabs(data[i]);
  }
  
  return norm;
}


double Vector::norm1() const {
  double norm;

  norm = 0;
  for (unsigned int i = 0; i < dimension; i++)
    norm += data[i];

  return norm;
}


double Vector::norm() const{
  return this -> norm2();
}

  /**
   * Normalise a vector. The normalisation is Euclidean. After the
   * function is called the vector will extend over an Euclidean
   * distance of one provided it is non-zero to start with. Both row
   * and column vectors can be normalised. If the dimension of the
   * vector is zero, normalisation yields the zero vector.
   */
const Vector& Vector::normalise() {
  double norm = norm2();

  if (fabs(norm) >= std::numeric_limits<double>::min())
    *this /= norm;
  else
    *this *= 0;

  return *this;
}

  

  /******************
   * VECTOR ALGEBRA *
   ******************/

  /**
   * Returns the absolute value of the elements in the vector.
   */
Vector Vector::abs() const {
  Vector v(dimension);
  
  for (unsigned int i = 0; i < dimension; i++)
    v.data[i] = fabs(data[i]);
  
  return v;
}
  

  /**
   * Flip vector in left/right direction.
   */
Vector Vector::flipLR() const {
  Vector v(dimension);
  
  for (unsigned int i = 0; i < dimension; i++)
      v.data[i] = data[(dimension - 1) - i];

  return v;
}

    
  /**
   * Flip Matrix in up/down direction.
   */
Vector Vector::flipUD() const {
  Vector v(dimension);
  
  for (unsigned int i = 0; i < dimension; i++)
      v.data[i] = data[(dimension - 1) - i];

  return v;
}

  
  /**
   * Returns a vector where every element is raised to the power of
   * the argument.
   */
Vector Vector::pow(double p) const {
  Vector v(dimension);
  
  for (unsigned int i = 0; i < dimension; i++)
    v.data[i] = ::pow(data[i], p);
  
  return v;
}


  /**
   * Sorts the elements of the vector in ascending order using the
   * system's quicksort implementation. When more than one element has
   * the same value, the order of the elements are preserved in the
   * sorted result and the indexes of equal elements will be ascending
   * in any index matrix.
   */
extern "C" 
{
  static int
  sim_vector_compare(const void *i, const void *j)
  {
    if (*((double *)i) > *((double *)j))
      return 1;
    else if (*((double *)i) < *((double *)j))
      return -1;
    return 0;
  }
}


Vector Vector::sort() const {
  Vector v(*this);

  qsort(v.data, dimension, sizeof(double), sim_vector_compare);
  
  return v;
}


  /**
   * Return the vector where each element is the square root of the
   * original corresponding element.
   */
Vector Vector::sqrt() const {
  Vector v(dimension);
  
  for (unsigned int i = 0; i < dimension; i++)
    if (data[i] > 0)
      v.data[i] = ::sqrt(data[i]);
    else
      v.data[i] = 0;
      
  return v;
}

  
  /**
   * Transpose a vector. If the vector was a column vector it becomes
   * a row vector and vice versa. Note that this method does not
   * change the vector itself.
   */
Vector Vector::transpose() const {
  Vector w(*this);
  w.column = !column;
  return w;
}


  
  /*********************
   * VECTOR STATISTICS *
   *********************/

  /**
   * Get the variance of the elements in the vector.
   */
double Vector::cov() const {
  double e1, e2;

  if (dimension == 0)
    return 0;

  e1 = e2 = 0;
  for (unsigned int i = 0; i < dimension; i++) {
    e1 += data[i] * data[i];
    e2 += data[i];
  }
  
  return ((e1 / dimension) - (e2 / dimension) * (e2 / dimension));
}
  

  /**
   * Return the largest component of the vector.
   */
double Vector::max() const {
	double max = -1.0 * std::numeric_limits<double>::max();

  for (unsigned int i = 0; i < dimension; i++)
    if (data[i] > max)
      max = data[i];
  
  return max;
}


  /**
   * Get the mean value of the elements in the vector.
   */
double Vector::mean() const {
  double m;

  if (dimension == 0)
    return 0;

  m = 0;
  for (unsigned int i = 0; i < dimension; i++)
    m += data[i];
  
  return m / dimension;
}


  /**
   * Return the smallest component of the vector.
   */
double Vector::min() const {
	double min = +1.0 * std::numeric_limits<double>::max();

  for (unsigned int i = 0; i < dimension; i++)
    if (data[i] < min)
      min = data[i];
  
  return min;
}


  /**
   * Return the standard deviation of the vector. Normalised by (N-1)
   * where "N" is the vector dimension. This makes std() the best
   * unbiased estimate of the variance if the vector is a sample from
   * a normal distribution.
   */
double Vector::std() const {
  unsigned int i;
  double       mu;
  double       sigma;

  if (dimension < 1)
    return 0;
  
  mu = 0;
  for (i = 0; i < dimension; i++)
    mu += data[i];
  
  mu    = mu / dimension;
  sigma = 0;
  for (i = 0; i < dimension; i++)
    sigma += (data[i] - mu) * (data[i] - mu);
    
  return ::sqrt(sigma / (dimension - 1));
}


  /**
   * Return the sum of all the elements in the vector.
   */
double Vector::sum() const {
  double sum = 0;

  for (unsigned int i = 0; i < dimension; i++)
    sum += data[i];
  
  return sum;
}


bool
Vector::elementWiseLessThan(const Vector &v) const
{
  for(unsigned int i = 0; i < dimension; i++)
    {
      if(data[i] >= v.data[i])
	{
	  return false;
	}
    }
  return true;
}

 /********************
  * VECTOR OPERATORS *
  ********************/

  /**
   * Unary minus.
   */
Vector Vector::operator- () const {
  if(*this == 0) {
    return 0;

  } else {
    Vector w(*this);
    for (unsigned int i = 0; i < dimension; i++)
      w.data[i] *= -1.0;
    return w;
  }
}


  /**
   * Binary minus with a real number. This is just subtraction of all
   * the vector components.
   */
Vector Vector::operator- (const double d) const {
  Vector w(*this);

  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] -= d;
  return w;
}


  /**
   * Binary minus with another vector. This is vector subtraction --
   * the dimensions must match.
   */
Vector Vector::operator- (const Vector &v) const throw(DimensionMismatch) {
  if (dimension != v.dimension)
    throw DimensionMismatch("Vector-Vector Subtraction");

  Vector w(*this);
  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] -= v.data[i];
  return w;
}


  /**
   * Decrement operator for reals.
   */
const Vector& Vector::operator-= (const double d) {
  for (unsigned int i = 0; i < dimension; i++)
    data[i] -= d;
  return *this;
}


  /**
   * Decrement operator for vectors. The dimensions must match.
   */
const Vector& Vector::operator-= (const Vector &v) throw(DimensionMismatch) {
  if (dimension != v.dimension)
    throw DimensionMismatch("Vector -=");

  for (unsigned int i = 0; i < dimension; i++)
    data[i] -= v.data[i];
  return *this;
}


  /**
   * Binary plus with a real number. This is just addition of all the
   * vector components.
   */
Vector Vector::operator+ (const double d) const {
  Vector w(*this);

  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] += d;
  return w;
}


  /**
   * Binary plus with another vector. This is vector addition -- the
   * dimensions must match.
   */
Vector Vector::operator+ (const Vector &v) const throw(DimensionMismatch) {
  if (dimension != v.dimension)
    throw DimensionMismatch("Vector-Vector Addition");

  Vector w(*this);
  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] += v.data[i];
  return w;
}


  /**
   * Increment operator for reals.
   */
const Vector& Vector::operator+= (const double d) {
  for (unsigned int i = 0; i < dimension; i++)
    data[i] += d;
  return *this;
}


  /**
   * Increment operator for vectors. The dimensions must match.
   */
const Vector& Vector::operator+= (const Vector &v) throw(DimensionMismatch) {
  if (dimension != v.dimension)
    throw DimensionMismatch("Vector +=");

  for (unsigned int i = 0; i < dimension; i++)
    data[i] += v.data[i];
  return *this;
}


  /**
   * Binary multiplication with a real number. This is scaling. Scale
   * all the components of the vector.
   */
Vector Vector::operator* (const double d) const {
  Vector w(*this);

  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] *= d;
  return w;
}


  /**
   * Binary multiplication with a matrix. Even though this is only
   * allowed for row vectors, it introduces no ambiguity to allow the
   * operation for column vectors as well. The dimensions must still
   * match, though. The product is always a row vector.
   */
Vector Vector::operator* (const Matrix &M) const throw(DimensionMismatch) {
  if (dimension != M.getRows())
    throw DimensionMismatch("Vector Scalar Product");

  Vector w(M.getCols());
  for (unsigned int j = 0; j < M.getCols(); j++) {
    w(j) = 0;
    for (unsigned int i = 0; i < dimension; i++)
      w(j) += data[i] * M(i, j);
  }

  w.setColumn(false);
  return w;
}


  /**
   * Binary multiplication. This is the scalar product. The scalar
   * product is defined for all combinations of row and column vectors
   * as long as the dimension matches.
   */
double Vector::operator* (const Vector &v) const throw(DimensionMismatch) {
  if (dimension != v.dimension)
    throw DimensionMismatch("Vector Scalar Product");

  double scalar = 0;
  for (unsigned int i = 0; i < dimension; i++)
    scalar += data[i] * v.data[i];

  return scalar;
}


  /**
   * Factor operator with a real number. This is scaling. Scale all
   * the components of the vector.
   */
const Vector& Vector::operator*= (const double d) {
  for (unsigned int i = 0; i < dimension; i++)
    data[i] *= d;
  return *this;
}


  /**
   * Binary multiplication. This is the vector product. The dimensions
   * must match. Furthermore, the vector product only has a well
   * defined meaning in three dimensions. As long as these criteria
   * are met, the vector product is defined for all combinations of
   * row and column vectors. The result is always a column vector.
   */
Vector Vector::operator^ (const Vector &v) const throw(DimensionMismatch) {
  if ((dimension != 3) || (v.dimension != 3))
    throw DimensionMismatch("Vector product");

  Vector w(3);
  w.data[0] = data[1] * v.data[2] - data[2] * v.data[1];
  w.data[1] = data[2] * v.data[0] - data[0] * v.data[2];
  w.data[2] = data[0] * v.data[1] - data[1] * v.data[0];
  w.setColumn(true);
  return w;
}


  /**
   * Binary multiplication. This is the outer vector product. The
   * dimensions must match. The result is matrix. The outer vector
   * product is defined for all combinations of row and column
   * vectors.
   */
Matrix Vector::operator% (const Vector &v) const throw(DimensionMismatch) {
  Matrix M(dimension, v.dimension);

  for (unsigned int i = 0; i < dimension; i++)
    for (unsigned int j = 0; j < v.dimension; j++)
      M(i, j) = data[i] * v.data[j];
  
  return M;
}


  /**
   * Binary division with a real number. This is scaling. Scale
   * all the components of the vector.
   */
Vector Vector::operator/ (const double d) const {
  Vector w(*this);

  for (unsigned int i = 0; i < dimension; i++)
    w.data[i] /= d;
  return w;
}


  /**
   * Factor operator with a real number. This is scaling. Scale all
   * the components of the vector.
   */
const Vector& Vector::operator/= (const double d) {
  for (unsigned int i = 0; i < dimension; i++)
    data[i] /= d;
  return *this;
}


  /**
   * Element accessor operator.
   */
double Vector::operator() (unsigned int i) const throw(OutOfRange) {
  //  if(i >= dimension)
  //  throw OutOfRange("Vector () Operator");

  return data[i];
}


  /**
   * Element accessor operator. This is for element assignment using
   * the accessor operator.
   */
double &Vector::operator() (unsigned int i) throw(OutOfRange) {
  //if(i >= dimension)
  //  throw OutOfRange("Vector () Operator");

  return data[i];
}


  /**
   * Assignment operator. Reallocate space for the vector only if the
   * dimensions are different.
   */
const Vector& Vector::operator= (const Vector &v) {
  if (this != &v) {
    if (dimension != v.dimension) {
      release();
      dimension = v.dimension;
      data      = new double[dimension];
    }

    column = v.column;
    for (unsigned int i = 0; i < dimension; i++)
      data[i] = v.data[i];
  }
  return *this;
}

const Vector& Vector::operator= (const double d) 
{
  for(unsigned int i = 0; i < dimension; i++)
    data[i] = d;
  return *this;
}



  /**
   * Less-than operator. A vector is less than another vector if it
   * smaller in a two-norm sense.
   */
bool
Vector::operator< (const Vector& v) const
{
  double       left  = 0;
  double       right = 0;
  unsigned int i;
  
  for (i = 0; i < dimension; i++)
    left += data[i] * data[i];
  
  for (i = 0; i < v.dimension; i++)
    right += v.data[i] * v.data[i];
  
  return (left < right);
}


  /**
   * Greater-than operator. A vector is greater than another vector if
   * it bigger in a two-norm sense.
   */
bool
Vector::operator> (const Vector& v) const
{
  double       left  = 0;
  double       right = 0;
  unsigned int i;
  
  for (i = 0; i < dimension; i++)
    left += data[i] * data[i];
  
  for (i = 0; i < v.dimension; i++)
    right += v.data[i] * v.data[i];
  
  return (left > right);
}


  /**
   * Equality operator. A vector is equal to a real number if it is
   * one-dimensional and its single component is equal to the number.
   */
bool Vector::operator== (const double d) const {
  return dimension == 1
    && fabs(data[0] - d) < std::numeric_limits<double>::epsilon();
}


  /**
   * Equality operator. Two vectors are equal if and only if their
   * dimensions match and all their components are equal. A column
   * vector can be equal to a row vector.
   */
bool Vector::operator== (const Vector &v) const {
  if (dimension != v.dimension)
    return false;

  for (unsigned int i = 0; i < dimension; i++)
    if (fabs(data[i] - v.data[i]) >= std::numeric_limits<double>::epsilon())
      return false;
  
  return true;
}


  /**
   * Inequality operator.
   */
bool Vector::operator!= (const double d) const {
  return !(*this == d);
}


  /**
   * Inequality operator for vectors.
   */
bool Vector::operator!= (const Vector &v) const {
  return !(*this == v);
}


  /**
   * Output operator. Sort of looks like MATLAB output when it is
   * done.
   */
namespace MesoRD {
  std::ostream &operator<< (std::ostream &os, const Vector &v) {
    os << "[";
    for (unsigned int i = 0; i < v.dimension; i++) {
      os << v.data[i];
      if (i != (v.dimension - 1)) {
        if (v.column)
          os << "; ";
        else
          os << " ";
      }
    }
    os << "]";
    
    return os;
  }
}

    
