/* The Vector class.
 *
 * Copyright 2001-2004 Johan Hattne
 *
 * This file is part of libsim.
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Vector.hpp 541 2008-04-12 18:52:41Z davidfange $
 */

/*
 * The Vector class. This is restricted to real and vectors of any
 * dimension. Note that a vector is really just a special case of a
 * matrix where either the number of rows or the number of columns is
 * equal to one. General convention: Vectors should be held in
 * lowercase variables.
 */

#ifndef __Vector_H
#define __Vector_H

#include <iostream>

#include "Exception.hpp"


  /* We will need access to the Matrix class, but we don't want to
   * include the header file and cause circular inclusion problems.
   */

namespace MesoRD {
  class Matrix;
  
  class Vector {
  private:
    unsigned int  dimension;
    double       *data;
    bool          column;
  
  public:
 
    
      /* Constructor and destructor.
       */
    Vector();
    Vector(const Vector &v);
    Vector(size_t l);
    ~Vector();


      /* General management methods.
       */
    void          release();
    static Vector ones(size_t l);
    static Vector rand(size_t l);
    static Vector zeros(size_t l);


      /* Getter Methods.
       */
    unsigned int getDimension() const;
    bool         isColumn()     const;
    

      /* Setter Methods.
       */
    void setColumn(bool c);
    

      /* Subvector Operations.
       */
    Vector getSubVector(unsigned int i1, unsigned int i2)
      const throw(OutOfRange);
    void   setSubVector(unsigned int i, const Vector v)
      throw(OutOfRange);
    
    
      /* Vector Norms.
       */
    double        norm2()         const;
    double        normp(double p) const;
    double        normMax()       const;
    double        norm1()         const;
    double        norm()          const; 
    const Vector& normalise();
    
    
      /* Vector Algebra.
       */
    Vector abs()         const;
    Vector flipLR()      const;
    Vector flipUD()      const;
    Vector pow(double p) const;
    Vector sort()        const;
    Vector sqrt()        const;
    Vector transpose()   const;
    

      /* Vector Statistics.
       */
    double cov()  const;
    double max()  const;
    double mean() const;
    double min()  const;
    double std()  const;
    double sum()  const;

    /* Extendend operatorfunction */
    bool elementWiseLessThan(const Vector &v) const;

  
      /* Vector Operators.
       */
    Vector        operator-  ()                const;
    Vector        operator-  (const double  d) const;
    Vector        operator-  (const Vector &v) const throw(DimensionMismatch);
    const Vector& operator-= (const double  d);
    const Vector& operator-= (const Vector &v)       throw(DimensionMismatch);
    Vector        operator+  (const double  d) const;
    Vector        operator+  (const Vector &v) const throw(DimensionMismatch);
    const Vector& operator+= (const double  d);
    const Vector& operator+= (const Vector &v)       throw(DimensionMismatch);
    Vector        operator*  (const double  d) const;
    Vector        operator*  (const Matrix &M) const throw(DimensionMismatch);
    double        operator*  (const Vector &v) const throw(DimensionMismatch);
    const Vector& operator*= (const double  d);
    Vector        operator^  (const Vector &v) const throw(DimensionMismatch);
    Matrix        operator%  (const Vector &v) const throw(DimensionMismatch);
    Vector        operator/  (const double  d) const;
    const Vector& operator/= (const double  d);
    double        operator() (const unsigned int  i) const throw(OutOfRange);
    double&       operator() (const unsigned int  i)       throw(OutOfRange);
    const Vector& operator=  (const Vector& v);
    const Vector& operator=  (const double d); 
    bool          operator<  (const Vector& v) const;
    bool          operator>  (const Vector& v) const;
    bool          operator== (const double  d) const;
    bool          operator== (const Vector &v) const;
    bool          operator!= (const double  d) const;
    bool          operator!= (const Vector &v) const;

    friend std::ostream &operator<< (std::ostream &os, const Vector &c);
  };
}

#endif
