/* The MesoRD main program.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: mesord.cpp 561 2008-11-25 20:10:53Z davidfange $
 */
#include <math.h>
#include <stdio.h>
#include <time.h>

#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <typeinfo>
#include <sstream>

#ifdef WIN32
#  include "fcntl.h" // For debug console
#  include "io.h"    // For debug console
#  include <valarray> // Required before windows.h on windows
#  define NOMINMAX
#  include "windows.h"
#  include "../woe32/resources/resource.h"
#  include "ui/WoeGetOpt.hpp"
#  include "ui/WoeProgressBar.hpp"
#else
#  include "ui/TextProgressBar.hpp"  // The TextProgressBar is unix specific at the moment
#endif
#  include "ui/TextGetOpt.hpp"

#include "mesord.hpp"

#include "CompartmentTable.hpp"
#include "Compartment.hpp"
#include "System.hpp"
#include "Thread.hpp"
#include "SubVolume.hpp"
#include "Species.hpp"

#ifdef DETERMINISTIC
#include "integration/Integration.hpp"
#endif

#ifdef HAVE_VIEWER
#  include "Visual.hpp"
#endif



class SubVolme;
class Species;




  /* Function prototypes and global variables.
   */
namespace
{
void stardate(int* issue, unsigned long* units, unsigned short* frac);

  /**
   * @brief Timer struct for timing events to second resolution.
   *
   * The implemention assumes that this code will run after January
   * 1st, 1970, because that is when <code>time(3)</code> returns
   * zero.  Because this text was written in 2007, there should be no
   * problem fulfilling that requirement, unless the clock is broken.
   *
   * @note The timer API presented here is modelled after
   * <code>glib</code>'s <code>g_timer</code>, with the exception that
   * there is no <code>reset()</code> function, and that the
   * <code>elapsed()</code> function breaks the elapsed time into
   * days, hours, minutes and seconds.
   */
  struct timer
  {
    bool   running;
    time_t time_start;
    time_t time_stop;

    /**
     * @brief Constructor for the timer struct.
     *
     * Initialises the start and stop time of the timer to zero.  The
     * timer is started by default.  This can be changed by setting
     * the @p run argument to <code>false</code>.
     *
     * @note As there are no extra resources allocated to timers,
     * there is no need for a destructor.
     *
     * @param run <code>true</code> if timing should commence on
     *            initialisation, <code>false</code> otherwise.
     */
    inline
    timer(const bool run = true)
      : running(false), time_start(0), time_stop(0)
    {
      if (run)
	start();
    }

    /**
     * @brief Resume the timer.
     *
     * Resumes a timer that has previously been stopped with stop().
     * This function has no effect if the timer is running.
     *
     * @note This function is commented out, because it is currently
     * not used.
     */
/*
    inline void
    continue()
    {
      running = true;
    }
*/


    /**
     * @brief See how much time has elapsed since the timer was
     * started.
     *
     * If the timer has been started but not stopped, this function
     * obtains the time since the timer was started.  If the timer has
     * been stopped, this function obtains the elapsed time between
     * the time it was started and the time it was stopped.
     *
     * @note This function does <em>not</em> stop the timer.
     *
     * @param[out] days    The number of days elapsed since the timer
     *                     was started.
     * @param[out] hours   The number of hours elapsed since the timer
     *                     was started.
     * @param[out] minutes The number of minutes elapsed since the
     *                     timer was started.
     * @param[out] seconds The number of seconds elapsed since the
     *                     timer was started.
     */
    inline void
    elapsed(size_t& days,
	    size_t& hours,
	    size_t& minutes,
	    size_t& seconds) const
    {
      time_t d = running ? time(NULL) - time_start : time_stop - time_start;
  
      days     = static_cast<size_t>(lrint(floor(1.0 * d / (24 * 60 * 60))));
      d       -= days * (24 * 60 * 60);

      hours    = static_cast<size_t>(lrint(floor(1.0 * d / (60 * 60))));
      d       -= hours * (60 * 60);

      minutes  = static_cast<size_t>(lrint(floor(1.0 * d / 60)));
      d       -= minutes * 60;

      seconds  = static_cast<size_t>(d);
    }


    /**
     * @brief Start the timer.
     *
     * Marks a start time, so that future calls to elapsed() will
     * report the time since start() was called.  This function is, by
     * default, called automatically on initialisation, so there is no
     * need to call start() immediately after creating a timer.
     *
     * @note If the timer is already running, it is restarted.  The
     * obsoletes any kind of <code>reset()</code> function.
     */
    inline void
    start()
    {
      time_start = time(NULL);
      running    = true;
    }


    /**
     * @brief Stop the timer.
     *
     * Marks an end time, so calls to elapsed() will return the
     * difference between this end time and the start time.  This
     * function has no effect if the timer is not running.
     *
     * @note This function is commented out, because it is currently
     * not used.
     */
/*
    inline void
    stop()
    {
      if (running) {
	time_stop = time(NULL);
	running   = false;
      }
    }
*/
  };
}


#ifdef WIN32
const char MESORD_WINDOWS_CLASSNAME[] = "MesoRDWindowClass";
std::string getModuleVersion();
#endif



  /*****************
   * MAIN FUNCTION *
   *****************/

  /* This is the MesoRD main function.  Basically it parses the
   * command line options and starts threads for the simulator and the
   * viewer.
   *
   */
int
main(int argc, char* argv[])
{

#ifdef WIN32
  STARTUPINFO si;
  GetStartupInfo(&si);
  HINSTANCE hInstance = (HINSTANCE)GetModuleHandle(NULL);
  
  WNDCLASSEX wndClass;
  memset(&wndClass,	0, sizeof(wndClass));
  wndClass.cbClsExtra    = 0;
  wndClass.cbSize        = sizeof(WNDCLASSEX);
  wndClass.cbWndExtra    = 0;
  wndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wndClass.hIcon         = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON1)); //LoadIcon(NULL, IDI_APPLICATION); // 32x32 icon for Alt-Tabbing
  wndClass.hIconSm       = (HICON)LoadImage(GetModuleHandle(NULL),
    MAKEINTRESOURCE(IDI_ICON2), IMAGE_ICON, 16, 16, 0); // 16x16 icon for taskbar
  wndClass.hInstance     = hInstance;
  wndClass.hbrBackground = (HBRUSH)(COLOR_APPWORKSPACE + 1);
    //wndClass.lpfnWndProc   = WndProc;
  wndClass.lpszClassName = MESORD_WINDOWS_CLASSNAME;
  wndClass.lpszMenuName  = NULL;
  wndClass.style         = CS_HREDRAW | CS_VREDRAW;
  
  if (!RegisterClassEx(&wndClass)) {
    MessageBox(NULL, "Window Registration Failed!", "Error!",
      MB_ICONEXCLAMATION | MB_OK);
    return FALSE;
  }

#endif

  
  int                      issue;
  std::ostream*            ostream_log;
  std::string              filename_model;
  time_t                   time_local;
  unsigned int             errors;
  size_t                   days    = 0;
  size_t                   hours   = 0;
  size_t                   minutes = 0;
  size_t                   seconds = 0;
  unsigned int             warnings;
  unsigned long            units;
  unsigned short           frac;

  struct timer             timer;

  MesoRD::System           system;
  MesoRD::ThreadProgress   threadProgress;

#ifdef HAVE_VIEWER
  MesoRD::ThreadVisualiser threadVisualiser;
#endif

    /* On all systems we start out by running in regular command prompt mode */
  MesoRD::GetOpt* getOpt;
  getOpt = new MesoRD::TextGetOpt(argc, argv);
  
    /* On the windows platform, open the GUI if asked for.  
     * Use the parsed values from the command promt, if they are supplied.  
     */
#ifdef WIN32
  if (getOpt -> getOptionViewer() && !getOpt -> getOptionHelp()) {
    MesoRD::GetOpt* textGetOpt = getOpt;
    getOpt = new MesoRD::WoeGetOpt(hInstance, NULL, si.wShowWindow, textGetOpt);
    delete textGetOpt;
  }	
#endif

  
    /* Initialise the window system.
     */
    /*
      Gtk::Main kit(argc, argv);
      Gtk::Main::run(progressBar); //Shows the window and returns when it is closed.
      return 0;
      }
    */
  
  errors      = 0;
  warnings    = 0;


    /* Write the messages from MesoRD to file if so requested.
     * Otherwise write the messages to stanard output.
     * FIXME!! Let the user decide the filename.  
     */
  std::ofstream* outfile_log = 0;
  if(getOpt -> getOptionLog()){
    outfile_log = new std::ofstream("output.log", std::ios::trunc);
    ostream_log = outfile_log;
  } else{
    ostream_log = &std::cout;
  }
  

    /* Pour les utilisateurs stupides.  Also pop up some versioning
     * information if requested.
     */
  if (getOpt -> getOptionHelp() || getOpt -> getOptionUsage()) {
    std::cout << "If there were a manual, this is where we would say RTFM!"
              << std::endl;

    getOpt -> displayHelp();
    return 0;
  }

#ifdef DETERMINISTIC
  if(getOpt -> getOptionDeterministic()){
    if(!(MesoRD::Integration::setIntegrationMethod(
           getOpt -> getOptionIntegrationMethod()))){
      std::cout << "Unable to find integration method " 
                << getOpt -> getOptionIntegrationMethod()
                << std::endl
                << " bailing out..."
                << std::endl;
      return 0;
    }
  }
#endif //DETERMINISTIC



  if (getOpt -> getOptionVersion()) {

    std::cout << "This is MesoRD @ pre-alpha"   << std::endl
#ifdef VERSION
              << "  Major Version: " << VERSION << std::endl
#endif
#ifdef WIN32
              << "Windows version "<< getModuleVersion() << std::endl
#endif
              << "  CVS $Revision: 561 $"      << std::endl;
    return 0;
  }
    

    /* Logfile preamble. "Captain's log"-kind-of-stuff.  Then
     * summarise whatever options we have collected from GetOpt().
     */
  stardate(&issue, &units, &frac);
  (*ostream_log) << "Captain's log, stardate "
                 << "["               << issue        << "]"
                 << units             << "."
                 << std::setfill('0') << std::setw(2)
                 << frac              << "." 
                 << std::endl         << *getOpt       << std::endl;



    /* Read the SBML document using libsbml.  Then set the options that
     * are relevant for the system.  If there is no SBML file to work
     * on, or if the user has chosen not to simulate the system, this
     * is where we break.
     */
  filename_model = getOpt -> getOptionFilename();


    /* Telling the system (or the parts that need to know) if we are
     * doing a deterministic simulation or not. (using doubles instead
     * of ints...)
     */
  MesoRD::SubVolume::setUseDoubleInSpecies(getOpt -> getOptionDeterministic());
  MesoRD::Species::setUseDoubleInAmount(getOpt -> getOptionDeterministic());

  if (filename_model.length() > 0) {
    
      system.setOptionLogfile(ostream_log);
      system.setOptionOutputSparse(getOpt -> getOptionSparse());
      
      try {
        system.fromSBML(
          getOpt -> getOptionFilename(),
          getOpt -> getOptionSize(),
          getOpt -> getOptionUnit());
      } catch (std::runtime_error& e) {
        std::stringstream message;
        
        message << "Error: "//<< "Cought Exception " 
            //<< typeid(e).name() 
            //<< std::endl 
                << e.what() 
                << std::endl;
        
        message << std::endl
                << "Something is exceptionally wrong! " 
                << "Thank you, come again!"
                << std::endl;
        
#ifdef WIN32
        MessageBox(NULL, (message.str()).c_str(), "Error!", MB_ICONEXCLAMATION | MB_OK);
#else
        (*ostream_log) << message.str();
#endif
        return -1;
      }
       
      
      system.setOptionCheckpointIterations(
        getOpt -> getOptionCheckpointIterations());
      system.setOptionCheckpointTime(
        getOpt -> getOptionCheckpointTime());
      system.setOptionMaxIterations(
        getOpt -> getOptionMaxIterations(),
        getOpt -> getOptionMaxIterationsScale());
      system.setOptionMaxTime(
        getOpt -> getOptionMaxTime());
      system.setOptionDeterministicSimulation(
        getOpt -> getOptionDeterministic());
      system.setOptionTimeStep(
        getOpt -> getOptionTimeStep());
      system.setOptionIntegrationMethod(
        getOpt -> getOptionIntegrationMethod());
      system.setOptionItMetConvergenceTolerance(
	getOpt -> getOptionItMetConvergenceTolerance());
      system.setOptionItMetMaxIterations(
	getOpt -> getOptionItMetMaxIterations());
      system.setOptionReadOldState(
        getOpt -> getOptionReadOldState());
      system.setOptionReadOldStateTime(
        getOpt -> getOptionReadOldStateTime());
      system.setOptionOutfilePrefix(
        getOpt -> getOptionOutputDir());
      system.setOptionFullStateCheckPoint(
        getOpt -> getOptionFullStateCheckPoint());    
    
  } else {
    (*ostream_log) << "No input file specified!" << std::endl;
    return 1;
  }

  if (!getOpt -> getOptionSimulate()) {
    (*ostream_log) << "Simulation turned off. Exiting." << std::endl;
    return 0;
  }
  

    /* Create the thread for the viewer.  The thread will run just one
     * function: the thread function.  A pointer to this function is
     * given as the third argument
     */
#ifdef HAVE_VIEWER
  if (getOpt -> getOptionViewer()) {
#ifdef HAVE_PTHREAD
    threadVisualiser.createVisualiser(getOpt, &system);
#endif
#ifdef WIN32
    threadVisualiser.createVisualiser(
      wndClass.hInstance, NULL, si.wShowWindow, getOpt, &system);
#endif
  }
#endif

    /* Exit if errors. The GNU C++ compiler has changed my life in
     * more ways than one.
     */
  if ((errors > 0) 
    || (getOpt -> getOptionPedantic() 
      && (warnings > 0))) {
    (*ostream_log) << "Confused by earlier errors, bailing out!" 
                   << std::endl;

#ifdef HAVE_VIEWER
    threadVisualiser.cancel();
#endif
    return 1;
  }
  

    /*****************
     * THE SIMULATOR *
     *****************/

    /* Den stora yttre loopen fr att gra typ MonteCarlo liksom va?
     * Se till att distribuera hr!
     */
  if ((errors == 0) && getOpt -> getOptionSimulate()) {
    
      /* Do some intial setup for the simulation.  Write console
       * output headers.  Open files for geomery and species
       * composision.
       */
    try {
      system.simulationInit();
    }  catch (std::runtime_error e) {
      std::stringstream message;
      
      message << "Error: "//"Cought Exception " 
          //<< typeid(e).name() 
          //  << std::endl 
              << e.what() 
              << std::endl;
      
      message << std::endl
              << "Something is exceptionally wrong! " 
              << "Thank you, come again!"
              << std::endl;
      
#ifdef WIN32
      MessageBox(NULL, (message.str()).c_str(), "Error!",
        MB_ICONEXCLAMATION | MB_OK);
#else
      std::cerr << message.str();
#endif
      return -1;
    }

    if (getOpt -> getOptionProgressBars()) {
#ifdef HAVE_PTHREAD
      threadProgress.createProgress(false, &system);
#endif
#ifdef WIN32
      threadProgress.createProgress(hInstance, NULL, si.wShowWindow, &system);
#endif
    }
    
    
    timer.start();
    
#ifdef DETERMINISTIC
    if(getOpt -> getOptionDeterministic()){
      system.simulateDeterministic();
    } else {
      system.simulate();
    }
#else
    system.simulate();
#endif //DETERMINISTIC
    
    timer.elapsed(days, hours, minutes, seconds);
      
    time(&time_local);
    (*ostream_log) << "The time after Monte Carlo loop "
                   << " is:   " << ctime(&time_local)    << std::endl;

    threadProgress.cancel();
  }


    /* Wait for the viewer thread to exit, provided it is started.
     * Then exit the program.  Thank you, please call again.
     */
#ifdef HAVE_VIEWER
  threadVisualiser.cancel();
#endif

    /* Close the output files when the simulation is done
     */
  system.simulationFinalise();

  (*ostream_log) << "Run terminated normally!" << std::endl
                 << "Wall time used: "
                 << days    << " day"    << ((days    != 1) ? "s" : "") << ", "
                 << hours   << " hour"   << ((hours   != 1) ? "s" : "") << ", "
                 << minutes << " minute" << ((minutes != 1) ? "s" : "") << ", "
                 << seconds << " second" << ((seconds != 1) ? "s" : "") << "."
                 << std::endl;

  if (outfile_log && getOpt -> getOptionLog()) {
    outfile_log -> close();
    delete outfile_log;
  }
  return 0;
}



  /***************
   * GEEK-O-RAMA *
   ***************/

  /* This will only work until stardate zero.  The program should be
   * revised until then.
   *
   * @bug This function is horribly broken!
   */
namespace
{
void
stardate(int* issue, unsigned long* units, unsigned short* frac) 
{
  unsigned long residual_seconds;
//  long long     zero_seconds; // ISO C++ does not support "long long"
  long          zero_seconds;
  
  //add explicit cast to make VC++ compiler happy
  zero_seconds = (long) time(NULL);
  for (unsigned int i = 0; i < 100; i++)
    zero_seconds -= 60600096;
  
  *issue            =
    static_cast<int>(lrint(floor(static_cast<double>(zero_seconds)
				 / (2000 * 86400))));
  residual_seconds  = static_cast<unsigned long>(
    zero_seconds - abs(*issue) * (2000 * 86400));
  *units            = static_cast<unsigned long>(lrint(
    1.0 * residual_seconds / (86400 / 5)));
  residual_seconds -= (*units) * (86400 / 5);
  *frac             = static_cast<unsigned short>(lrint(
    1.0 * residual_seconds * 100 / (86400 / 5)));
}
}


#ifdef WIN32

/*
 * This routine will extract the version string from the
 * string version resource in the RC file for the current module.
 * Note that you must add version.lib to your project to
 * link to the Win32 versioning API calls. The actual call
 * VerQueryValue() uses a value of 040904B0 for the language
 * and character set. This value is equivalent to English
 * language text encoded using Unicode.000004B0 is language neutral
*/
std::string
getModuleVersion()
{
    char file_name[MAX_PATH];
    GetModuleFileName(::GetModuleHandle(NULL), file_name, MAX_PATH);

    DWORD dwDummyHandle;
    DWORD len = GetFileVersionInfoSize(file_name, &dwDummyHandle);

    BYTE* buf = new BYTE[len];

//    std::vector<BYTE> buf( len );

//    ::GetFileVersionInfo( file_name, 0, len, buf.begin());
    ::GetFileVersionInfo(file_name, 0, len, buf);
    char *version;
    unsigned int ver_length;

//    ::VerQueryValue( buf.begin(), 
    ::VerQueryValue(buf, 
//                     "\\StringFileInfo\\040904B0\\FileVersion",
                     "\\StringFileInfo\\000004B0\\FileVersion",
                     (void **) &version,
                     &ver_length );
    delete buf;
    return std::string(version, ver_length);
}
#endif
