/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2008, Johan Elf, David Fange, Johan Hattne
 *
 * Permission to use, copy, modify, and/or distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear
 * in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: Device.hpp 546 2008-05-02 00:15:26Z hattne $
 */
#ifndef MESORD_RANDOM_DEVICE_HPP
#define MESORD_RANDOM_DEVICE_HPP 1

  /* For compilation on Windows using the Microsoft Visual C++
   * compiler, replace calls to deprecated functions with calls to the
   * new, secure versions of those functions by defining
   * _CRT_SECURE_CPP_OVERLOAD STANDARD_NAMES as 1.  Turn off warnings
   * about any remaining deprecated functions by defining
   * _CRT_SECURE_NO_WARNINGS as 1.  These preprocessor directives are
   * needed before the inclusion of any STL headers in order to be
   * useful.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
#    define _CRT_SECURE_NO_WARNINGS                 1
#  endif
#endif

#include "Random.hpp"

namespace MesoRD
{
  namespace Random
  {
      /**
       * @brief Obtains pseudo-random numbers from random data source
       * devices.
       *
       * Some systems provide random data source devices, such as
       * <code>/dev/random</code> or <code>/dev/urandom</code>.  This
       * class can make use of these sources to obtain pseudo-random
       * numbers.
       */
    class Device : public Random {
    private:
        /**
         * @brief FILE pointer to the random data source device.
         */
      FILE* device;

    public:
        /**
         * @brief Default constructor for the Device class.
         *
         * Opens the device whose name is the string pointed to by @p
         * path.
         *
         * @param path The path to the random data source device.
         */
      inline
      Device(const char* path)
        : device(NULL)
        {
          if (path != NULL)
            device = fopen(path, "rb");
        }


        /**
         * @brief Destructor for the Device class.
         *
         * Closes the random data source device, if opened.
         */
      virtual inline
      ~Device()
        {
          if (device != NULL)
            fclose(device);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the closed interval @f$\left[ 0,\; 1 \right]@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number by dividing the unsigned long integer
         * polled from the random data source by its maximum value.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right]@f$, or
         *         <code>NaN</code> for errors.
         */
      inline double
      randomClosed()
        {
          unsigned long n;
          if (fread(&n, sizeof(unsigned long), 1, device) < 1)
            return std::numeric_limits<double>::quiet_NaN();
          return static_cast<double>(n)
            * (1.0 / static_cast<double>(
                 std::numeric_limits<unsigned long>::max()));
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the open interval @f$\left( 0,\; 1 \right)@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number by dividing the unsigned long integer
         * polled from the random data source <code>+ 0.5</code> by
         * its maximum value <code>+ 1.0</code>.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left( 0,\; 1 \right)@f$, or
         *         <code>NaN</code> for errors.
         */
      inline double
      randomOpen()
        {
          unsigned long n;
          if (fread(&n, sizeof(unsigned long), 1, device) < 1)
            return std::numeric_limits<double>::quiet_NaN();
          return (static_cast<double>(n) + 0.5)
            * (1.0 / (static_cast<double>(
                        std::numeric_limits<unsigned long>::max()) + 1.0));
        }
    };
  }
}


  /* Restore changed preprocessor directives.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    undef _CRT_SECURE_NO_WARNINGS
#    undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
#  endif
#endif

#endif /* ifndef MESORD_RANDOM_DEVICE_HPP */
