/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2008, Johan Elf, David Fange, Johan Hattne
 * Copyright (c) 2003-2007, Jasper Bedaux
 * Copyright (c) 2000-2003, Richard J. Wagner
 * Copyright (c) 1997-2002, Makoto Matsumoto and Takuji Nishimura.
 *                          All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 *
 *   3. The names of its contributors may not be used to endorse or
 *      promote products derived from this software without specific
 *      prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: MersenneTwister.cpp 560 2008-11-09 11:12:14Z davidfange $
 */

  /* For compilation on Windows using the Microsoft Visual C++
   * compiler, replace calls to deprecated functions with calls to the
   * new, secure versions of those functions by defining
   * _CRT_SECURE_CPP_OVERLOAD STANDARD_NAMES as 1.  Turn off warnings
   * about any remaining deprecated functions by defining
   * _CRT_SECURE_NO_WARNINGS as 1.  These preprocessor directives are
   * needed before the inclusion of any STL headers in order to be
   * useful.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
#    define _CRT_SECURE_NO_WARNINGS                 1
#    pragma warning(push)
#    pragma warning( disable:4996)
#  endif
#endif

#include "MersenneTwister.hpp"

#include <cstdio>
#include <ctime>
#include <limits>




  /* First, attempt to read an array of N 32-bit integers from
   * /dev/urandom into init_keys[].  If reading /dev/urandom was not
   * successful, seed with a hash of time(3) and clock(3) instead.
   */
MesoRD::Random::MersenneTwister::MersenneTwister()
{
  uint32_t init_keys[N];
  size_t   i;
  size_t   ret;
  FILE*    urandom;
  clock_t  c;
  time_t   t;

  urandom = fopen("/dev/urandom", "rb");
  if (urandom) {
    for (i = 0, ret = 1; i < N && ret >= 1; i++)
      ret = fread(init_keys + i, sizeof(uint32_t), 1, urandom);
    fclose(urandom);

    if (ret >= 1) {
      seed(init_keys, N);
      return;
    }
  }

  c = clock();
  t = time(0);
  seed(hash_clock_time<uint32_t>(c, t));
}


  /* This function leaves mti at N, so genstate() will be called
   * before the first random number is polled.
   */
void
MesoRD::Random::MersenneTwister::seed(uint32_t key)
{
  mt[0]= key & 0xffffffff;
  for (mti = 1; mti < N; mti++) {
      /* See Donald E. Knuth, "The Art of Computer Programming",
       * Vol. 2, 3rd Ed., p. 106 for multiplier.  In the previous
       * versions, the most significant bits of the seed affect only
       * the most significant bits of the state vector mt[].  Modified
       * by Makoto Matsumoto on January 9th, 2002.
       */
    mt[mti] = (1812433253 * (mt[mti - 1] ^ (mt[mti - 1] >> 30))
      + static_cast<uint32_t>(mti)); 

      /* For >32-bit machines.
       */
    mt[mti] &= 0xffffffff;
  }
}


  /* The call to seed() sets mti to N, so genstate() will be called
   * before the first random number is polled.  Incorporates a slight
   * change for C++ from February 2nd, 2004.
   */
void
MesoRD::Random::MersenneTwister::seed(
  const uint32_t* const init_keys,
  size_t                len)
{
  size_t i;
  size_t j;
  size_t k;

  seed(19650218);

  i = 1;
  j = 0;
  for (k = (N > len ? N : len); k; k--) {
      /* Non-linear.
       */
    mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1664525))
      + init_keys[j] + static_cast<uint32_t>(j);

      /* For >32-bit machines.
       */
    mt[i] &= 0xffffffff;
    if (++i >= N) {
      mt[0] = mt[N - 1];
      i = 1;
    }
    j = (j + 1) % len;
  }
  for (k = N - 1; k; k--) {
      /* Non-linear.
       */
    mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1566083941))
      - static_cast<uint32_t>(i);

      /* For >32-bit machines.
       */
    mt[i] &= 0xffffffff;
    if (++i >= N) {
      mt[0] = mt[N - 1];
      i = 1;
    }
  }

    /* Most significant bit is 1, assuring non-zero initial vector.
     */ 
  mt[0] = 0x80000000;
}

/* Restore changed preprocessor directives.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    undef _CRT_SECURE_NO_WARNINGS
#    undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
#    pragma warning(pop)
#  endif
#endif