/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2008, Johan Elf, David Fange, Johan Hattne
 * Copyright (c) 2003-2007, Jasper Bedaux
 * Copyright (c) 2000-2003, Richard J. Wagner
 * Copyright (c) 1997-2002, Makoto Matsumoto and Takuji Nishimura.
 *                          All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 *
 *   3. The names of its contributors may not be used to endorse or
 *      promote products derived from this software without specific
 *      prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: MersenneTwister.hpp 560 2008-11-09 11:12:14Z davidfange $
 */
#ifndef MESORD_RANDOM_MERSENNETWISTER_HPP
#define MESORD_RANDOM_MERSENNETWISTER_HPP 1

#include "Random.hpp"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif

  /* Fixing the inttypes.h incompatibility on Windows.
   */
#ifdef WIN32
typedef unsigned __int32 uint32_t;
typedef          __int32  int32_t;
#endif

#include <cstddef>



namespace MesoRD
{
  namespace Random
  {
      /**
       * @brief Pseudo-random number generator of type Mersenne
       * Twister.
       *
       * This is an implementation of Takuji Nishimura and Makoto
       * Matsumoto's Mersenne Twister pseudo-random number generator,
       * with period @f$2^{19937} - 1@f$.  Just like the base version
       * from January 26<sup>th</sup>, 2002, it can be initialised
       * from a seed or from an array.  The initialisation scheme for
       * previous versions of the Mersenne Twister (<i>e.g.</i>
       * versions from October 28<sup>th</sup>, 1999 or earlier) had a
       * minor problem: the most significant bit of the seed was not
       * well reflected to the state vector of the Mersenne Twister.
       *
       * This implementation is an adaption of the original C-program
       * for MT19937 written by Takuji Nishimura and Makoto Matsumoto.
       * The original authors welcome any feedback at
       * http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html.
       *
       * <b>Reference:</b> Matsumoto, M. & Nishimura, T. <i>Mersenne
       * Twister: A 623-Dimensionally Equidistributed Uniform
       * Pseudo-Random Number Generator</i>, <i>ACM Transactions on
       * Modeling and Computer Simulation</i>, <b>8(1)</b>, 3--30,
       * (January 1998).
       *
       * @note This generator is not cryptographically secure, as the
       * generator state can be learnt after observing @f$N@f$
       * consecutive values.  One would need to use a one-way
       * function, or hash function, to obtain secure pseudo-random
       * numbers from a sequence of pseudo-random numbers from the
       * Mersenne Twister.
       *
       * @note This generator is not thread safe, unless
       * auto-initialisation is avoided and each thread has its own
       * MersenneTwister object.
       */
    class MersenneTwister : public Random
    {
    private:
        /**
         * @brief Configurable constants for the Mersenne Twister.
         *
         * Because an enumeration is used, all values should fit into
         * a, possibly unsigned, <code>int</code>.
         */
      enum constants {
          /**
           * @brief Period parameter, such that @f$M \geq 1@f$.
           */
        M = 397,

          /**
           * @brief Length of state vector, MersenneTwister::mt, such
           * that @f$N \geq M \geq 1@f$.
           */
        N = 624,

          /**
           * @brief Constant vector @f$\mathbf{a}@f$.
           */
        MATRIX_A = 0x9908b0df,

          /**
           * @brief Least significant @f$r@f$ bits, where @f$r =
           * 31@f$.
           */
        LOWER_MASK = 0x7fffffff,

          /**
           * @brief Most significant @f$w - r@f$ bits, where @f$w =
           * 32@f$ and @f$r = 31@f$.
           */
        UPPER_MASK = 0x80000000
      };

        /**
         * @brief The state vector, which holds the internal state.
         *
         * The state vector is initialised by the seed() functions,
         * which are called by the constructors.
         */
      uint32_t mt[N];

        /**
         * @brief Index into the state vector, MersenneTwister::mt.
         */
      size_t mti;


        /**
         * @brief Generates a new state vector with @f$N@f$ elements.
         *
         * This modification is due to Jasper Bedaux.  The genstate()
         * function is called <code>reload()</code> by Richard
         * J. Wagner.
         */
      inline void
      genState()
        {
          size_t kk;

          for (kk = 0; kk < N - M; kk++)
            mt[kk] = twist(mt[kk + M], mt[kk], mt[kk + 1]);

          for (kk = N - M; kk < N - 1; kk++) {
            mt[kk] = twist(
              mt[kk + static_cast<size_t>(M) - static_cast<size_t>(N)],
              mt[kk],
              mt[kk + 1]);
          }

          mt[N - 1] = twist(mt[M - 1], mt[N - 1], mt[0]);
          mti       = 0;
        }


        /**
         * @brief Mixes the upper bits from @p u with the lower bits
         * from @p v.
         *
         * This modification is due to Richard J. Wagner.
         *
         * @param u The upper bits for the mix.
         * @param v The lower bits for the mix.
         * @return  The mixed bits.
         */
      inline uint32_t
      mixBits(uint32_t u, uint32_t v) const
        {
          return (u & static_cast<uint32_t>(UPPER_MASK))
            |    (v & static_cast<uint32_t>(LOWER_MASK));
        }


        /**
         * @brief Initialises the pseudo-random number generator with
         * a seed.
         *
         * Initialises the state vector, MersenneTwister::mt, with a
         * 32-bit integer seed, @p init_key.
         *
         * @param init_key The 32-bit integer seed.
         */
      void
      seed(uint32_t init_key);


        /**
         * @brief Initialises the pseudo-random number generator with
         * an array of seeds.
         *
         * Initialises the state vector, MersenneTwister::mt, with an
         * array of @p len 32-bit integers, @p init_key.  Each of the
         * @f$2^{19937} - 1@f$ possible initial states can be accessed
         * by using an array of 624 32-bit integers, because @f$\lceil
         * 19937 / 624 \rceil = 32@f$.
         *
         * @param init_keys The array of initialising keys.
         * @param len       The length of @p init_keys.
         */
      void
      seed(const uint32_t* const init_keys, size_t len);


        /**
         * @brief Twists three 32-bit integers into a single 32-bit
         * integer.
         *
         * This is used during the generation of a new state.  The
         * modifications are due to Jasper Bedaux and Richard
         * J. Wagner
         *
         * @param u The first 32-bit integer to twist.
         * @param v The second 32-bit integer to twist.
         * @param w The third 32-bit integer to twist.
         * @return  The twisted 32-bit integer.
         */
      inline uint32_t
      twist(uint32_t u, uint32_t v, uint32_t w) const
        {
          return u
            ^ (mixBits(v, w) >> 1)
            ^ ((w & 0x1 != 0) ? static_cast<uint32_t>(MATRIX_A) : 0);
        }


    public:
        /**
         * @brief Default constructor for the MersenneTwister class.
         *
         * Constructs a @f$N@f$-element seed for the pseudo-random
         * number generator by reading 32-bit integers from
         * <code>/dev/urandom</code>.  If <code>/dev/urandom</code> is
         * not available, a hash of <code>clock(3)</code> and
         * <code>time(3)</code> is used instead.
         *
         * @note Based on code by Lawrence Kirby and Richard
         * J. Wagner.
         */
      MersenneTwister();


        /**
         * @brief Constructs a MersenneTwister instance given a 32-bit
         * integer seed.
         *
         * Sets its argument @p seed as the seed for a new sequence of
         * pseudo-random numbers to be returned by the MersenneTwister
         * instance.  The sequence is repeatable by initialising with
         * the same seed value.
         *
         * @param init_key The seed for the pseudo-random number
         *                 generator.
         */
      inline
      MersenneTwister(uint32_t init_key)
        {
          seed(init_key);
        }


        /**
         * @brief Constructs a MersenneTwister instance given an array
         * of integer seeds.
         *
         * All bits in the array contribute to the generated sequence
         * of pseudo-random numbers.  This is useful if one wants a
         * larger seed space than a single 32-bit integer would
         * permit.
         *
         * @param init_keys The array for seeding.
         * @param len       The length of @p init_keys.
         */
      inline
      MersenneTwister(const uint32_t* const init_keys, size_t len)
        {
          seed(init_keys, len);
        }


        /**
         * @brief Generates a pseudo-random, 32-bit integer number in
         * the closed interval @f$\left[ 0,\; \mathtt{0xffffffff}
         * \right]@f$.
         *
         * Generates a uniform, positive, 32-bit, pseudo-random
         * integer by polling a 32-bit number from the state vector.
         * All other generator functions are simple transformations of
         * the return value of genrand_int32().  The genstate()
         * function is split off, because it is only called once every
         * @f$N@f$ calls.
         *
         * @return A pseudo-random, integer number in the interval
         *         @f$\left[ 0,\; \mathtt{0xffffffff} \right]@f$.
         */
      inline uint32_t
      genrand_int32()
        {
          uint32_t y;

            /* Generate N words at one time.
             */
          if (mti >= N)
            genState();
          y = mt[mti++];

            /* Tempering.
             */
          y ^= (y >> 11);
          y ^= (y <<  7) & 0x9d2c5680;
          y ^= (y << 15) & 0xefc60000;
          y ^= (y >> 18);
          return y;
        }


        /**
         * @brief Generates a pseudo-random, 31-bit integer number in
         * the closed interval @f$\left[ 0,\; \mathtt{0x7fffffff}
         * \right]@f$.
         *
         * Generates a uniform, positive, 31-bit, pseudo-random
         * integer number by shifting out the least significant bit
         * returned by genrand_int32().
         *
         * @return A pseudo-random, integer number in the interval
         *         @f$\left[ 0,\; \mathtt{0x7fffffff} \right]@f$.
         */
      inline int32_t
      genrand_int31()
        {
          return static_cast<int32_t>(genrand_int32() >> 1);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the closed interval @f$\left[ 0,\; 1 \right]@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number to 32-bit resolution by dividing the
         * return value of genrand_int32() by @f$2^{32} - 1@f$.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right]@f$.
         */
      inline double
      genrand_real1()
        {
          return static_cast<double>(genrand_int32()) * (1.0 / 4294967295.0);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the half open interval @f$\left[ 0,\; 1 \right)@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number to 32-bit resolution by dividing the
         * return value of genrand_int32() by @f$2^{32}@f$.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right)@f$.
         */
      inline double
      genrand_real2()
        {
          return static_cast<double>(genrand_int32()) * (1.0 / 4294967296.0);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the open interval @f$\left( 0,\; 1 \right)@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number to 32-bit resolution by dividing the
         * value of <code>genrand_int32() + 0.5</code> by
         * @f$2^{32}@f$.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left( 0,\; 1 \right)@f$.
         */
      inline double
      genrand_real3()
        {
          return (static_cast<double>(genrand_int32()) + 0.5)
            * (1.0 / 4294967296.0);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the half open interval @f$\left[ 0,\; 1 \right)@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number to 53-bit resolution, which is the
         * width of the significand in IEEE double precision floating
         * point numbers.  Two calls to genrand_int32() are done; one
         * for the most significant 27 bits, and another for the least
         * significant 26.
         *
         * @note The genrand_res53() function is due to Isaku Wada,
         * added on January 9<sup>th</sup>, 2002.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right)@f$.
         */
      inline double
      genrand_res53()
        {
          uint32_t ls;
          uint32_t ms;

          ls = genrand_int32();
          ms = genrand_int32();
          return (
            + static_cast<double>(ms >> 5) * 67108864.0
            + static_cast<double>(ls >> 6))
            * (1.0 / 9007199254740992.0);
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the closed interval @f$\left[ 0,\; 1 \right]@f$.
         *
         * @see genrand_real1()
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right]@f$.
         */
      inline double
      randomClosed()
        {
          return genrand_real1();
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the open interval @f$\left( 0,\; 1 \right)@f$.
         *
         * @see genrand_real3()
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left( 0,\; 1 \right)@f$.
         */
      inline double
      randomOpen()
        {
          return genrand_real3();
        }
    };
  }
}
#endif /* ifndef MESORD_RANDOM_MERSENNETWISTER_HPP */
