/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2008, Johan Elf, David Fange, Johan Hattne
 * Copyright (c) 2000-2003, Richard J. Wagner
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 *
 *   3. The names of its contributors may not be used to endorse or
 *      promote products derived from this software without specific
 *      prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: Random.hpp 560 2008-11-09 11:12:14Z davidfange $
 */
#ifndef MESORD_RANDOM_RANDOM_HPP
#define MESORD_RANDOM_RANDOM_HPP 1

  /* For compilation on Windows using the Microsoft Visual C++
   * compiler, replace calls to deprecated functions with calls to the
   * new, secure versions of those functions by defining
   * _CRT_SECURE_CPP_OVERLOAD STANDARD_NAMES as 1.  Turn off warnings
   * about any remaining deprecated functions by defining
   * _CRT_SECURE_NO_WARNINGS as 1.  These preprocessor directives are
   * needed before the inclusion of any STL headers in order to be
   * useful.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    ifndef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
#      define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
#    endif
#    define _CRT_SECURE_NO_WARNINGS                 1
#    pragma warning(push)
#    pragma warning(disable:4996)
#    pragma warning(disable:4005)
#  endif
#endif


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <limits>

namespace MesoRD
{
    /**
     * @brief The %Random module for pseudo-random number generation.
     *
     * @todo Insert documentation from the reports here?
     */
  namespace Random
  {
      /**
       * @brief System-dependent pseudo-random number generator.
       *
       * This class uses whatever algorithm is appropriate for the
       * system to produce pseudo-random numbers.  There are at least
       * two different ways to do this.
       *
       * <ul>
       *
       *   <li>The traditional <code>rand(3)</code> function returns a
       *   pseudo-random number in the interval @f$\left[ 0,\;
       *   \mathtt{RAND\_MAX} \right]@f$.  Due to its inclusion in ISO
       *   C89, the <code>rand(3)</code> interface is generally
       *   available.</li>
       *
       *   <li>The <code>random(3)</code> function returns a random
       *   number in the range @f$\left[ 0,\; 2^{31} - 1 \right]@f$.
       *   The period of <code>random(3)</code> is approximately @f$16
       *   \cdot \left( 2^{31} - 1 \right)@f$, which is much larger
       *   than the period of <code>rand(3)</code>.  The downside is
       *   that <code>random(3)</code> performs at about @f$2 / 3@f$
       *   the speed of <code>rand(3)</code>.</li>
       *
       * </ul>
       *
       * The notion of intervals is important in this class, as well
       * as all its subclasses.  An open interval is an interval that
       * does not include its end points.  The open interval
       * @f$\left\{ x : a < x <b \right\}@f$ is denoted @f$\left( a,\;
       * b \right)@f$.  A closed interval is an interval that includes
       * all of its limit points.  The closed interval @f$\left\{ x :
       * a \leq x \leq b \right\}@f$ is denoted @f$\left[ a ,\; b
       * \right]@f$.
       */
    class Random {
    protected:
        /**
         * @brief The static variable <code>differ</code> ensures that
         * the return value of hash_clock_time() always changes.
         *
         * Using Random::differ, the return value of hash_clock_time()
         * <em>will</em> change, even if successive calls are made
         * faster than the time resolution of <code>clock_t</code> and
         * <code>time_t</code>.
         */
      size_t differ;

        /**
         * @brief Hashes a <code>clock_t</code> and a
         * <code>time_t</code>.
         *
         * The return value of hash_clock_time() may be used to seed
         * the various pseudo-random number generators.
         *
         * @note Based on code by Lawrence Kirby and Richard
         * J. Wagner.
         *
         * @param c The <code>clock_t</code> contribution.
         * @param t The <code>time_t</code> contribution.
         * @return  A hash value, based on @p c and @p t.
         */
      template<class T>
      inline T
      hash_clock_time(clock_t c, time_t t)
        {
          size_t               i;
          const unsigned char* p;
          T                    h1;
          T                    h2;

          h1 = 0;
          h2 = 0;
          p  = reinterpret_cast<const unsigned char *>(&t);
          for (i = 0; i < sizeof(t) / sizeof(unsigned char); i++) {
            h1 *= std::numeric_limits<unsigned char>::max() + 2;
            h1 += p[i];
          }

          p = reinterpret_cast<const unsigned char *>(&c);
          for (i = 0; i < sizeof(c) / sizeof(unsigned char); i++) {
            h2 *= std::numeric_limits<unsigned char>::max() + 2;
            h2 += p[i];
          }

          return (h1 + static_cast<T>(differ++)) ^ h2;
        }

    public:
        /**
         * @brief Default constructor for the Random class.
         *
         * Attempts to read an unsigned integer from
         * <code>/dev/urandom</code>.  If <code>/dev/urandom</code>
         * cannot be read, the pseudo-random number generator is
         * seeded with a hash of <code>time(3)</code> and
         * <code>clock(3)</code> instead.
         */
      inline
      Random()
        : differ(0)
        {
          unsigned int seed;
          FILE*        urandom;
          clock_t      c;
          time_t       t;

          urandom = fopen("/dev/urandom", "rb");
          if ( urandom == NULL
            || fread(&seed, sizeof(unsigned int), 1, urandom) < 1) {
            c    = clock();
            t    = time(0);
            seed = hash_clock_time<unsigned int>(c, t);
          }
          if (urandom != NULL)
            fclose(urandom);

#ifdef HAVE_RANDOM
          srandom(seed);
#else
          srand(seed);
#endif
        }


        /**
         * @brief Constructs a Random instance given an integer seed.
         *
         * Sets its argument @p seed as the seed for a new sequence of
         * pseudo-random numbers to be returned by the Random
         * instance.  The sequence is repeatable by initialising with
         * the same seed value.
         *
         * @param seed The seed for the pseudo-random number
         *             generator.
         */
      inline
      Random(unsigned int seed)
        : differ(0)
        {
#ifdef HAVE_RANDOM
          srandom(seed);
#else
          srand(seed);
#endif
        }


        /**
         * @brief Destructor for the Random class.
         */
      virtual inline
      ~Random()
        {}


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the closed interval @f$\left[ 0,\; 1 \right]@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number by dividing the return value of
         * <code>rand(3)</code>, or <code>random(3)</code>, by its
         * respective upper limit.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left[ 0,\; 1 \right]@f$.
         */
      virtual inline double
      randomClosed()
        {
#ifdef HAVE_RANDOM
          return static_cast<double>(random())
            * (1.0 / 2147483647.0);
#else
          return static_cast<double>(rand())
            * (1.0 / RAND_MAX);
#endif
        }


        /**
         * @brief Generates a pseudo-random, double precision number
         * in the open interval @f$\left( 0,\; 1 \right)@f$.
         *
         * Generates a uniform, positive, double precision,
         * pseudo-random number by dividing the return value of
         * <code>rand(3) + 0.5</code>, or <code>random(3) +
         * 0.5</code>, by its respective upper limit <code>+ 1</code>.
         *
         * @return A pseudo-random, double precision number in the
         *         interval @f$\left( 0,\; 1 \right)@f$.
         */
      virtual inline double
      randomOpen()
        {
#ifdef HAVE_RANDOM
          return (static_cast<double>(random()) + 0.5)
            * (1.0 / 2147483648.0);
#else
          return (static_cast<double>(rand()) + 0.5)
            * (1.0 / (RAND_MAX + 1));
#endif
        }
    };
  }
}


  /* Restore changed preprocessor directives.
   */
#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    undef _CRT_SECURE_NO_WARNINGS
#    undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
#    pragma warning(pop)
#  endif
#endif

#endif /* ifndef MESORD_RANDOM_RANDOM_HPP */
