/* The LinearElement class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: LinearElement.cpp 541 2008-04-12 18:52:41Z davidfange $
 */

#include <math.h>

#include "LinearElement.hpp"



  /**************************
   * THE LINEARFACTOR CLASS *
   **************************/

  /* Evaluate the factor.  This is done by using up all the data
   * gathered and multiplying it together.
   */
double
MesoRD::LinearFactor::evaluate(
  const std::valarray<unsigned int>& species) const
{
  const LinearFactor* current;
  double              value = 1;

  for (current = this; current; current = current -> next) {
    double b =  ((double)species[current -> index]) + ((double)(current -> offset));
    int    e = current -> exponent;

    value *= (current -> factor) * pow(b, e);
  }
	
  return value;
}


double
MesoRD::LinearFactor::evaluate(
  const std::valarray<double>& species) const
{
  const LinearFactor* current;
  double              value = 1;

  for (current = this; current; current = current -> next) {
    double b = species[current -> index] + (current -> offset);

    int    e = current -> exponent;

    value *= (current -> factor) * pow(b, e);
  }

  return value;
}


  /* Output the factor in our own peculiar format.
   */
std::ostream&
MesoRD::LinearFactor::putTo(std::ostream& os) const 
{
  os << factor << " * ({"
     << index  << "} + ("
     << offset << ")) ^"
     << exponent;

  if (next) {
    os << " * ";
    next -> putTo(os);
  }

  return os;
}



  /************************
   * THE LINEARTERM CLASS *
   ************************/

  /* Evaluate the term.  This is done by using up all the data
   * gathered and multiplying it together.
   */
double
MesoRD::LinearTerm::evaluate(const std::valarray<unsigned int>& species) const
{
  const LinearTerm* current;
  double            value = 0;

  for (current = this; current; current = current -> next) {
    value += (current -> coefficient)
      * (current -> factor) -> evaluate(species);
  }

  return value;
}


double
MesoRD::LinearTerm::evaluate(const std::valarray<double>& species) const
{
  const LinearTerm* current;
  double            value = 0;

  for (current = this; current; current = current -> next) {
    value += (current -> coefficient)
      * (current -> factor) -> evaluate(species);
  }

  return value;
}


  /* Output the term in our own peculiar format.
   */
std::ostream&
MesoRD::LinearTerm::putTo(std::ostream& os) const
{
  os << "[" << coefficient << "] --> " << *factor << std::endl;

  if (next)
    next -> putTo(os);

  return os;
}




