/* The LinearElement class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: LinearElement.hpp 540 2008-04-12 14:30:36Z davidfange $
 */
#ifndef __LinearElement_H
#define __LinearElement_H

#include <iostream>
#include <valarray>

namespace MesoRD {

    /**
     * The LinearElement class.
     *
     * Probably the most commonly used expressions for reactions are
     * of the type "k * (X + a) * (Y + b) * ...".  The purpose of this
     * class is to provide a more efficient way to evaluate these
     * kinds of expressions.  In particular, if an expression can be
     * "linearised", we do not have to traverse a tree just to
     * evaluate it.  We allow for a little more general expressions
     * than that, since we can have sums of products.  The products in
     * turn are composed of terms which have the form "k * (X +
     * a)^(e)".
     */
  class LinearElement {
  protected:

      /**
       * Puts the linearisation in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).  The reason for its existence is purely
       * technical; it makes the implementation a little nicer.
       * Callers should not use this method, but make use of the
       * operator instead.
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    virtual std::ostream& putTo(std::ostream& os) const = 0;

  public:

      /**
       * Constructor for the LinearElement base class.
       *
       * Does nothing.
       */
    inline LinearElement()
      {};

      /**
       * Destructor for the LinearElement base class.
       *
       * Does nothing.  But it does it virtually.
       */
    virtual ~LinearElement()
      {};

      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    virtual double evaluate(
      const std::valarray<unsigned int>& species) const = 0;

      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    virtual double evaluate(
      const std::valarray<double>& species) const = 0;


      /**
       * Put a linearisation to an output stream.
       *
       * This method will put a linearisation onto an output stream.
       *
       * @param os The output stream on which to write.
       * @param e  The element in question.
       * @return   The output stream, once written to.
       */
    friend inline std::ostream& operator<<(
      std::ostream& os, const LinearElement& e) { return e.putTo(os); }
  };



    /**************************
     * THE LINEARFACTOR CLASS *
     **************************/

    /**
     * The LinearFactor class.
     *
     * As was noted previously, a linearisation consists of factors
     * and terms.  A factor is of the form "k *(X + a)^(e)", where "k"
     * is a factor, "X" is a concentration which is looked up in the
     * supplied species vector on evaluation, "a" is an offset and "e"
     * is an exponent.  Every factor also has a pointer to the next
     * factor in the product.
     */
  class LinearFactor : public LinearElement {
  private:
    int           exponent;
    double        factor;
    off_t         index;
    int           offset;
    LinearFactor* next;

  protected:

      /**
       * Puts the linearisation in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).  The reason for its existence is purely
       * technical; it makes the implementation a little nicer.
       * Callers should not use this method, but make use of the
       * operator instead.
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * Default constructor for the LinearFactor class.
       *
       * Constructs a new factor using default values.  A factor
       * constructed with this constructor is useless until it has
       * been properly set up.
       */
    LinearFactor()
      : exponent(0), factor(0), index(-1), offset(0), next(0) {};

      /**
       * Value constructor for the LinearFactor class.
       *
       * This constructor will create a new instance of the
       * LinearFactor class using the supplied values.
       *
       * @param exponent The exponent of the new factor.
       * @param factor   The multiplicative factor of the new factor.
       * @param index    The index of the species to use in the new factor.
       * @param offset   The offset of the new factor.
       */
    LinearFactor(int exponent, double factor, off_t index, int offset)
      : exponent(exponent), factor(factor), index(index),
        offset(offset), next(0) {};

      /**
       * Destructor for the LinearFactor class.
       *
       * Does nothing.
       */
    ~LinearFactor()
      {};

      /**
       * Append a factor to the chain of the factor.
       *
       * When evaluating the factor product, all factors in the chain
       * are evaluated.  The result is the product of all the factors.
       *
       * @param f The new factor to append to the end of the chain.
       */
    inline void appendNext(LinearFactor* f)
      { if (next) next -> appendNext(f); else setNext(f); }

      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression, given the
       * supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression, given the
       * supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Get the exponent of the factor.
       *
       * This method will return the exponent of the factor.
       *
       * @return The exponent of the factor.
       */
    inline int getExponent() const
      { return exponent; }

      /**
       * Get the factor of the factor.
       *
       * Hmm....  This may be a little confusing.  Maybe we should
       * have called the multiplicative factor "coefficient" or at
       * least something other than "factor".
       *
       * @return The multiplicative factor of the factor.
       */
    inline double getFactor() const
      { return factor; }

      /**
       * Get the offset of the factor.
       *
       * This method will return the offset of the factor.
       *
       * @return The offset of the factor.
       */
    inline int getOffset() const
      { return offset; }

      /**
       * Set the exponent of the factor.
       *
       * Calling this method will change the exponent of the factor.
       *
       * @param e The new exponent of the factor.
       */
    inline void setExponent(int e)
      { exponent = e; }

      /**
       * Set the multiplicative factor of the factor.
       *
       * This method will change the multiplicative factor of the
       * factor.
       *
       * @param f The new multiplicative factor of the factor.
       */
    inline void setFactor(double f)
      { factor = f; }

      /**
       * Set the offset of the factor.
       *
       * This method will change the offset of the factor.
       *
       * @param o The new offset of the factor.
       */
    inline void setOffset(int o)
      { offset = o; }

      /**
       * Set the pointer to the next factor in the chain.
       *
       * This method will update the pointer to the next factor in the
       * chain that composes the product.
       *
       * @param f A pointer to the next factor in the chain.
       */
    inline void setNext(LinearFactor* f)
      { next = f; }

      /**
       * Put a linearisation factor to an output stream.
       *
       * This method will put a linearisation onto an output stream.
       *
       * @param os The output stream on which to write.
       * @param f  The factor in question.
       * @return   The output stream, once written to.
       */
    friend inline std::ostream& operator<<(
      std::ostream& os, const LinearFactor& f)
      { return f.putTo(os); }
  };



    /************************
     * THE LINEARTERM CLASS *
     ************************/

    /**
     * The LinearTerm class.
     *
     * As was noted previously, a linearisation consists of factors
     * and terms.  A term is just a sum of products of factors.  It
     * has the general form "c1 * F1 + c2 * F2 + ...", where "c1",
     * "c2" are coefficients, and "F1", "F2" are instances of the
     * LinearFactor class.
     */
  class LinearTerm : public LinearElement {
  private:
    double        coefficient;
    LinearFactor* factor;
    LinearTerm*   next;
    
  protected:

      /**
       * Puts the linearisation in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).  The reason for its existence is purely
       * technical; it makes the implementation a little nicer.
       * Callers should not use this method, but make use of the
       * operator instead.
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
      */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * Default constructor for the LinearTerm class.
       *
       * Constructs a new factor using default values.  A term
       * constructed with this constructor is useless until it has
       * been properly set up.
       */
    LinearTerm()
      : coefficient(0), factor(0), next(0) {};

      /**
       * Value constructor for the LinearTerm class.
       *
       * This constructor will create a new instance of the LinearTerm
       * class using the supplied values.
       *
       * @param coefficient The coefficient of the new term.
       * @param factor      The factor of the new term.
       */
    LinearTerm(double coefficient, LinearFactor* factor)
      : coefficient(coefficient), factor(factor), next(0) {};
    
      /**
       * Destructor for the LinearTerm class.
       *
       * Does nothing.
       */
    ~LinearTerm()
      {};

      /**
       * Append a factor to the factor-chain of the term.
       *
       * When evaluating the term sum, all factor-chains in the term
       * are evaluated.  The result is the sum of all the chains.
       *
       * @param f The new factor to append to the end of the chain.
       */
    inline void appendFactor(LinearFactor* f)
      { if (factor) factor -> appendNext(f); else setFactor(f); }

      /**
       * Append a term to the chain of terms.
       *
       * When evaluating the term sum, all factor-chains in the term
       * are evaluated.  The result is the sum of all the chains.
       *
       * @param t The new term to append to the end of the chain.
       */
    inline void appendNext(LinearTerm* t)
      { if (next) next -> appendNext(t); else setNext(t); }

      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression, given the
       * supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Get the coefficient for the LinearTerm.
       *
       * @return The coefficient for the LinearTerm.
       */
    inline double getCoefficient()
      { return coefficient; }
    
      /**
       * Get the subsequent LinearTerm.
       * 
       * This will return the LinearTerm that follows this LinearTerm
       * allowing for traversal of all the LinearTerms.
       *
       * @return The subsequent LinearTerm.
       */
    inline LinearTerm* getNext()
      { return next; }
    

      /**
       * Evaluate the linearisation to given a species composition.
       *
       * Calling this method will evaluate the expression, given the
       * supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Set the coefficient for the LinearTerm.
       * 
       * This will set the coefficient for the LinearTerm.
       *
       * @param coefficient The coefficient to be set.
       */
    inline void setCoefficient(double coefficient)
      { this -> coefficient = coefficient; }
    

      /**
       * Set the pointer to the first factor in the chain.
       *
       * This method will update the pointer to the first factor in
       * the chain that composes the product.
       *
       * @param f A pointer to the first factor in the chain.
       */
    inline void setFactor(LinearFactor* f)
      { factor = f; }

      /**
       * Set the pointer to the next term in the chain.
       *
       * This method will update the pointer to the next term in the
       * chain that composes the sum.
       *
       * @param t A pointer to the next term in the chain.
       */
    inline void setNext(LinearTerm* t) 
      { next = t;}

      /**
       * Put a linearisation term to an output stream.
       *
       * This method will put a linearisation onto an output stream.
       *
       * @param os The output stream on which to write.
       * @param t  The term in question.
       * @return   The output stream, once written to.
       */
    friend inline std::ostream& operator<<(
      std::ostream& os, const LinearTerm& t)
      { return t.putTo(os); }
  };
}

#endif
