/* The Reaction class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Reaction.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <math.h>

#include <sstream>
#include <iostream>

#include "../mesord.hpp"
#include "../definition.hpp"

#include "../Annotation.hpp"
#include "../ParameterTable.hpp"
#include "../SpeciesTable.hpp"
#include "../ui/OPrefixStream.hpp"

#include "LinearElement.hpp"
#include "Reaction.hpp"
#include "ReactionNode.hpp"



  /****************************
   * CONSTRUCTOR & DESTRUCTOR *
   ****************************/

  /* Here we set the species_change vector to length zero.  The
   * subsequent call to fromSBML() will set the correct size.
   */
MesoRD::Reaction::Reaction(
  const ::Reaction* r, 
  const ParameterTable* p, 
  const SpeciesTable* s,
  std::ostream* o)
  throw(MesoRD::Reaction::ModelError)
  : species_change((int)0, 0),
    reactants((unsigned int)0, 0),
    products((unsigned int)0, 0)
{
  this -> id             = "";
  this -> name           = "";
  this -> linearisation  = 0;
  this -> math           = 0;
  this -> outstream_log  = o;
  this -> parameters     = p;
  this -> multi_sv_reaction_ass = false;
  this -> multi_sv_reaction_diss = false;
  
  
  if (r && s) {
    try {
      fromSBML(r, s);
    } catch (ModelError& e) {
      throw ModelError(e.what());
    }
  }
}


  /* Note that it is safe to copy valarrays as long as we know that
   * they are the same size.  If they are not, the result is
   * undefined.  The size requirement is ensured in the initial
   * initialisation.
   */
MesoRD::Reaction::Reaction(const Reaction& r)
  : species_change(r.species_change.size()),
    reactants(r.reactants.size()),
    products(r.products.size())
{
  this -> id             = r.id;
  this -> name           = name;
  this -> linearisation  = r.linearisation;
  this -> math           = r.math;
  this -> outstream_log  = 0;
  this -> parameters     = r.parameters;
  this -> multi_sv_reaction_diss = r.multi_sv_reaction_diss;
  this -> multi_sv_reaction_ass = r.multi_sv_reaction_ass;
  

  this -> reactants      = r.reactants;
  this -> products       = r.products;
  
  this -> species_change = r.species_change;
}


  /* Destructor.  Do-nothing.  Nothing much, at least.
   */
MesoRD::Reaction::~Reaction()
{
  if (linearisation)
    delete linearisation;
  if (math)
    delete math;
}



  /****************************************
   * CHANGE VECTOR MANIPULATION -- PUBLIC *
   ****************************************/

  /* There may be a better way to do it.  In any case, I do not think
   * we loose much doing it this way.
   */
void
MesoRD::Reaction::setSpeciesChangeVector(int change[], size_t species)
{
  species_change.resize(species);
  for (unsigned int i = 0; i < species_change.size(); i++)
    species_change[i] = change[i];
}



  /***********************
   * REACTION EVALUATION *
   ***********************/

  /* Figures out what compartment the reaction belongs to.  If all
   * species in the reaction are defined to be in the same
   * compartment, the reaction belongs to the compartment.  If no
   * suitable compartment could be found, return the empty string.
   */
std::string
MesoRD::Reaction::getCompartment(
  const ::Reaction* r, const SpeciesTable* s)
  throw( ModelError )
{
  unsigned int i;
  std::string  id = "";

  for (i = 0; i < r -> getNumProducts(); i++) {
    const SpeciesReference* sr    = r -> getProduct(i);
    off_t                   index = s -> getIndexById(sr -> getSpecies());
  
    if ((index >= 0) && (index < (off_t)(s -> getSpeciesNumber()))) {
      if (id.length() == 0) {
        id = s -> getCompartmentByIndex(index);
      } else if (id != (s -> getCompartmentByIndex(index))) {
        std::stringstream message;
        message << "All reactants and products not defined in the same " 
                << "compartment in reaction: " 
                << Reaction_getId(r);
        throw ModelError((message.str()).c_str());
      }
    }
  }

  for (i = 0; i < r -> getNumReactants(); i++) {
    const SpeciesReference* sr    = r -> getReactant(i);
    off_t                   index = s -> getIndexById(sr -> getSpecies());
    if ((index >= 0) && (index < (off_t)(s -> getSpeciesNumber()))) {
      if (id.length() == 0) {
        id = s -> getCompartmentByIndex(index);
      } else if (id != (s -> getCompartmentByIndex(index))) {
        std::stringstream message;
        message << "All reactants and products not defined in the same " 
                << "compartment in reaction: " 
                << Reaction_getId(r);
        throw ModelError((message.str()).c_str());
      }
    }
  }

  return id;
}

  
  /* Which method to use in order to evaluate the reaction depends on
   * whether it could be linearised or not.
   */
double
MesoRD::Reaction::getRate(const std::valarray<unsigned int>& species) const
{
  if (linearisation){
    
    double temp = linearisation -> evaluate(species);
    if(temp >= 0) {
      return temp;
    }
    else {
      return 0;
    }  
  }
  
  else if (math){
    double temp = math -> evaluate(species);
    if(temp >=0) {
      return temp;
    }
    else {
      return 0;
    }
  }
  
  return 0;
}


double
MesoRD::Reaction::getRate(const std::valarray<double>& species) const
{
  if (linearisation)
    return (linearisation -> evaluate(species));
  else if (math)
    return (math -> evaluate(species));

  return 0;
}


  /****************
   * SBML SUPPORT *
   ****************/

  /* Read the reaction from SBML.  Not very difficult and also
   * commented along the way.  We are repeating the code for reactants
   * and products here which is not very nice.
   */
void
MesoRD::Reaction::fromSBML(
  const Reaction_t* r, 
  const SpeciesTable* st)
  throw(MesoRD::Reaction::ModelError)
{
  double              stoichiometry;
  off_t               index;
  unsigned int        i;
  std::string         compartment_id;
  std::stringstream   message;
  

    /* Read the reaction id.
     */
  if (r -> isSetId())
    id = r -> getId();
  else
    id = "";
  
    /* Read the reaction name.
     */
  if (r -> isSetName())
    name = r -> getName();
  else
    name = "";
  
    /* Handle the reversibility of reactions.
     */
  if (r -> getReversible()) {
    if (outstream_log) {
      (*outstream_log) << "WARNING: Cannot handle reversible reactions."
                       << std::endl;
    }
  }

    /* Handle the inter subvolumes reactions, at the moment this is
     * obligatory in this version.   
     */
  if(const ::XMLNode* n = ((Reaction_t*) r) -> getAnnotation()) {
  
    Annotation   annotation( n );
    
    if (annotation.getNumChildren() > 2) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Some annotation children to the reaction: " 
                         << r -> getId() << " will be disregarded" 
                         << std::endl;
      }
    }
    
    for( unsigned int child_nr = 0; child_nr < annotation.getNumChildren(); child_nr++) {
      
      const Annotation *child = annotation.getChild(child_nr);
      
      if (child) {
        if (child -> getName() == Identifier::XML::InterSVAssnReactions::name) {
          try {
            
            multi_sv_reaction_ass = child -> getBoolAttribute(
              Identifier::XML::InterSVAssnReactions::Attributes::state);
            
          } catch(Annotation::InvalidFormat<double>& e) {
            message << e.what() 
                    << ", while parsing: " 
                    << e.getValue() << std::endl;
            throw ModelError((message.str()).c_str());
            
          }
        } else {
          if(child -> getName() == Identifier::XML::InterSVDissReactions::name) {
            try {
              
              multi_sv_reaction_diss = child -> getBoolAttribute(
                Identifier::XML::InterSVDissReactions::Attributes::state);
              
            } catch(Annotation::InvalidFormat<double>& e) {
              message << e.what() 
                      << ", while parsing: " 
                      << e.getValue() << std::endl;
              throw ModelError((message.str()).c_str()); 
            }
          } else {
            message << "The annotation in the reaction node requires " 
                    << "MesoRD:" 
                    << Identifier::XML::InterSVAssnReactions::name 
                    << " or "
                    << Identifier::XML::InterSVDissReactions::name
                    << " to be set" << std::endl;
            throw ModelError((message.str()).c_str());
          }
        }
      } else {
        message << "Less then two annotations were found for reaction: " 
                << r-> getId() << std::endl;
        throw ModelError((message.str()).c_str());
      }
    }
  }
  
  products.resize(r -> getNumProducts());
  
    /* Set up the species change vector for the products.  This needs
     * to consider the stoichiometry of the reaction.
     */
  species_change.resize(st -> getSpeciesNumber());
  for (i = 0; i < r -> getNumProducts(); i++) {
    const SpeciesReference* sr = r -> getProduct(i);
    if (sr -> isSetStoichiometryMath()) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Cannot handle math in stoichiometry."
                         << std::endl;
      }
    }

    index = st -> getIndexById(sr -> getSpecies());
    if ((index < 0) || (index >= (off_t)(st -> getSpeciesNumber()))) {
      message << "Undefined species "
              << sr -> getSpecies()
              << ".";
      throw ModelError((message.str()).c_str());
    }
    
    stoichiometry = sr -> getStoichiometry();
    if (fabs(stoichiometry - rint(stoichiometry))
      >= std::numeric_limits<double>::epsilon()) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Stoichiometry not an integer. Rounding."
                         << std::endl;
      }
    }
    
    products[i] = index;
    

    if (!(st -> isConstantByIndex(index)))
      species_change[index] += static_cast<int>(lrint(stoichiometry));
  }
  
  reactants.resize(r -> getNumReactants());
  
  
  
    /* Set up the species change vector for the reactants.  This needs
     * to consider the stoichiometry of the reaction.
     */
  for (i = 0; i < r -> getNumReactants(); i++) {
    const SpeciesReference* sr = r -> getReactant(i);
    if (sr -> isSetStoichiometryMath()) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Cannot handle math in stoichiometry."
                         << std::endl;
      }
    }

    index = st -> getIndexById(sr -> getSpecies());
    if ((index < 0) || (index >= (off_t)(st -> getSpeciesNumber()))) {
      message << "Undefined species "
              << sr -> getSpecies()
              << ".";
      throw ModelError((message.str()).c_str());
    }

    stoichiometry = sr -> getStoichiometry();
    if (fabs(stoichiometry - rint(stoichiometry))
      >= std::numeric_limits<double>::epsilon()) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Stoichiometry not an integer. Rounding."
                         << std::endl;
      }
    }

    reactants[i] = index;
    
    
    if (!(st -> isConstantByIndex(index)))
      species_change[index] -= static_cast<int>(lrint(stoichiometry));
    
  }

      /* Read the math for the reaction from the kinetic law
       * definition.  Do constant folding until we cannot fold
       * anymore.  We know we are done when the pointer to the
       * returned node is not changing anymore.
       */
  try {
    math = ReactionNode::fromSBML(
      (r -> getKineticLaw()) -> getMath(), parameters, st);    

    
    try {
      math = math -> toPerSecond(outstream_log);
    
    } catch(MesoRD::Unit::InvalidConversion) {
      if (outstream_log) {
        (*outstream_log) << "WARNING: Unknown unit in reaction. "
                         << "Assuming user knows what user does."
                         << std::endl;
      }
    }
    
    math = ReactionNode::refold(math);
    linearisation = math -> linearise();

  } catch (ReactionNode::InvalidUnit& e) {
    std::stringstream message;
    message << "Invalid unit " << e.what();
    throw ModelError((message.str()).c_str());

  } catch (ReactionNode::ModelError& e) {
    std::stringstream message;
    message << "Model Error " << e.what();
    throw ModelError((message.str()).c_str());

  } catch (ReactionNode::NotYetImplemented& e) {
    std::stringstream message;
    message << "Not yet Implemented " << e.what();
    throw ModelError((message.str()).c_str());

  } catch (ReactionNode::UndefinedIdentifier& e) {
    std::stringstream message;
    message << "Undefined identifier " << e.what();
    throw ModelError((message.str()).c_str());

  }
}



  /**********
   * OUTPUT *
   **********/

  /* Output operator.  Should be valid XML.  Note that since a
   * reaction does not carry a species table, the output will be a
   * little misleading.
   */
namespace MesoRD {
  std::ostream &operator<< (std::ostream &os, const Reaction &r) {
    bool          has_products  = false;
    bool          has_reactants = false;
    int           stoichiometry;
    OPrefixStream opfs_level0(os.rdbuf(),          "");
    OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
    OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");

    for (unsigned int i = 0; i < r.species_change.size(); i++) {
      if (r.species_change[i] < 0)
        has_reactants = true;
      else if (r.species_change[i] > 0)
        has_products = true;
    }
    
    if (r.linearisation) {
      opfs_level0 << "<!-- The reaction below has a linearisation:"
                  << std::endl
                  << "   - " << std::endl
                  << "   - " << *r.linearisation << std::endl
                  << "   - " << std::endl
                  << "  -->" << std::endl;
    }

    opfs_level0 << "<reaction id=\"" << r.id << "\"";
    if (r.name.length() > 0)
      opfs_level0 << " name=\"" << r.name << "\"";
    opfs_level0 << " reversible=\"false\">" << std::endl;

    if (has_reactants) {
      opfs_level1 << "<listOfReactants>" << std::endl;
      for (unsigned int i = 0; i < r.reactants.size(); i++) {
        opfs_level2 << "<speciesReference species=\""
                    << r.reactants[i] << "\""
                    << " stoichiometry=\"" << abs(r.species_change[i])
                    << "\"/>" << std::endl;
      }
      opfs_level1 << "</listOfReactants>" << std::endl;
    }

    if (has_products) {
      opfs_level1 << "<listOfProducts>" << std::endl;
      for (unsigned int i = 0; i < r.products.size(); i++) {
        opfs_level2 << "<speciesReference species=\""
                    << r.products[i] << "\""
                    << " stoichiometry=\"" << abs(r.species_change[i])
                    << "\"/>" << std::endl;
      }
      opfs_level1 << "</listOfProducts>" << std::endl;
    }
    
    if (r.math) {
      opfs_level1 << "<math xmlns=\"http://www.w3.org/1998/Math/MathML\">"
                  << std::endl;
      opfs_level2 << *r.math;
      opfs_level1 << "</math>" << std::endl;
    }

    opfs_level1 << "<annotation xmlns:" << Identifier::XML::xmlns
                << ">" << std::endl;
    
      opfs_level2 << "<"   << Identifier::XML::InterSVAssnReactions::name
                  << " "   << Identifier::XML::InterSVAssnReactions::Attributes::state
                  << "=\"" << r.multi_sv_reaction_ass
                  << "\""  << std::endl;

      opfs_level2 << "<"   << Identifier::XML::InterSVDissReactions::name
                  << " "   << Identifier::XML::InterSVDissReactions::Attributes::state
                  << "=\"" << r.multi_sv_reaction_diss
                  << "\""  << std::endl;

      opfs_level1 << "</annotation>" << std::endl;

    
    opfs_level0 << "</reaction>" << std::endl;

    return os;
  }
}
