/* The Reaction class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: Reaction.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_REACTION_HPP
#define MESORD_REACTION_HPP

#include <stdexcept>
#include <valarray>

#include <stdio.h>
#include <sbml/SBMLTypes.h>


namespace MesoRD {

  class LinearElement;
  class ParameterTable;
  class ReactionNode;
  class SpeciesTable;
  

    /**
     * Class for modelling and evaluation reactions.
     *
     * A reaction has a mathematical expression that depends on the
     * time-dependant species concentrations and parameters.  The
     * mathematical expression must be given using the MathML that is
     * supported in later versions of SBML.  A reaction also has a
     * change vector.  The change vector describes how the reaction
     * affects all the species.  The methods in this class use various
     * other class to read in and evaluate reactions.
     */
  class Reaction  {
  private:
    std::string           id;
    std::string           name;

    LinearElement*        linearisation;
    ReactionNode*         math;
    std::ostream*         outstream_log;
    const ParameterTable* parameters;
    std::valarray<int>    species_change;

    bool                  multi_sv_reaction_ass;
    bool                  multi_sv_reaction_diss;
    
    std::valarray<unsigned int> reactants;
    std::valarray<unsigned int> products;
    
    
  public:


      /**
       * Class for reporting errors during model building.
       *
       * If something went wrong during the model construction, this
       * exception will be thrown.
       */
    class ModelError : public std::runtime_error {
    public:

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Value constructor for the Reaction class.
       *
       * Constructs the reaction from the SBML definitions.  Most
       * importantly, this method will initialise the species change
       * vector.  The parameter table is needed in order to store and
       * possibly fold constant parameters.  The parameter table will
       * need to be consulted during simulation time for parameters
       * that are not constant.  The species table is needed in order
       * to map species names and id:s to species indices.
       *
       * @param r The SBML reaction definition.
       * @param p The parameter table.
       * @param s The species table.
       */
    Reaction(
      const ::Reaction*     r,
      const ParameterTable* p,
      const SpeciesTable*   s,
      std::ostream* o) throw(ModelError);

      /**
       * Copy constructor for the Reaction class.
       *
       * Constructs a new reaction as a copy of an existing reaction.
       * This method is not entirely safe as the parsed mathematical
       * expressions are not deep-copied.
       *
       * @param r The reaction which to copy.
       */
    Reaction(const Reaction& r);

      /**
       * Destructor for the Reaction class.
       *
       * Frees the parsed reaction and any possibly linearisation
       * created from the former.
       */
    ~Reaction();

      /**
       * Get the compartment of which the reaction is part.
       *
       * Returns the id compartment in which the reaction is defined.
       * If all species in the reaction are defined to belong to the
       * same compartment, the reaction belongs to that compartment,
       * too.  If there are ambiguities or other problems, this method
       * returns the empty string.
       *
       * @return The id of the compartment which is holding the reaction.
       */
    static std::string getCompartment(
      const ::Reaction* r, const SpeciesTable* s)
      throw( ModelError );
    
      /**
       * Get the multi-subvolume reaction flag.  
       *
       * If this flag is set to true, the reaction will participate in
       * reaction events that include species concentrations from two
       * neighbouring subvolumes.  The possibility for such reactions
       * were added as a spatial correction when then discretisation
       * approaches the size of the reaction radia of the reacting
       * molecules.
       *
       * @return The multi-subvolume reaction flag.  
       */
    inline bool getMultiSVReactionAss() const
      { return multi_sv_reaction_ass; }

          /**
       * Get the multi-subvolume reaction flag.  
       *
       * If this flag is set to true, the reaction will participate in
       * reaction events that include species concentrations from two
       * neighbouring subvolumes.  The possibility for such reactions
       * were added as a spatial correction when then discretisation
       * approaches the size of the reaction radia of the reacting
       * molecules.
       *
       * @return The multi-subvolume reaction flag.  
       */
    inline bool getMultiSVReactionDiss() const
      { return multi_sv_reaction_diss; }

    
      /**
       * Evaluate the reaction for the given species configuration.
       *
       * This method will return the reaction rate for the given
       * species configuration.  Each element of the input species
       * vector gives the number of molecules of the respective
       * species present.
       *
       * @param species The current state of the species.
       */
    double getRate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the reaction for the given species configuration.
       *
       * This method will return the reaction rate for the given
       * species configuration.  Each element of the input species
       * vector gives the number of molecules of the respective
       * species present.
       *
       * @param species The current state of the species.
       */
    double getRate(const std::valarray<double>& species) const;

      /*
       * Returns the reactants of each reaction.  
       *
       */
    std::valarray<unsigned int> getReactants()
      { return reactants; }
      
/*
 * Returns the reactants of each reaction.  
       *
       */
    std::valarray<unsigned int> getProducts()
      { return products; }


    
      /**
       * Set up the logging stream.
       *
       * If no logging stream has been set up before the parsing
       * begins, any error reporting and logging information will be
       * silently lost.  It is thus highly recommended to call this
       * method prior to calling fromSBML().  Note also that this
       * method will set up the logging stream for any member classes
       * that can make use of one.
       *
       * @param log The logging stream onto which to report errors, etc.
       */
    inline void setOptionLogfile(std::ostream* log)
      { outstream_log = log; }

      /**
       * Sets the species change vector from an array template.
       *
       * This does a deep copy of the supplied species change array to
       * the internal species change vector.  The elements of the
       * array describe how the reaction affects the species with the
       * corresponding index.
       *
       * @param change  The species change array.
       * @param species The size of the species change array.
       */
    void setSpeciesChangeVector(int change[], size_t species);
    
      /**
       * Get the species change vector of the reaction.
       *
       * The species change vector is a vector with the same size as
       * the number of species in the system.  Every element in the
       * species change vector is signed integer indicating how the
       * respective species will change upon execution of the
       * particular reaction.  For instance in a four-species system,
       * a reaction that decreases the first two species by one and
       * increases the fourth by two, would have a species change
       * vector of [-1 -1 0 +2].
       *
       * @return The species change vector of the reaction.
       */
    inline const std::valarray<int>& getSpeciesChangeVector() const
      { return species_change; }

      /**
       * Build a reaction from an SBML definition.
       *
       * The species table is needed in order to enable species name
       * and species id to species index mapping.  This method will
       * construct the species change vector from the stoichiometric
       * information in the SBML model.  It will also read in the
       * mathematics needed to evaluate the reaction constants.
       *
       * @param r  The SBML reaction to build from.
       * @param st The species table to consult for mapping.
       */ 
    void fromSBML(const ::Reaction* r, const SpeciesTable* st)
      throw(ModelError);
    
      /**
       * Put a reaction to an output stream.
       *
       * This method will put a reaction onto an output stream.  The
       * result should look something like XML.
       *
       * @param os The output stream on which to write.
       * @param r  The reaction in question.
       * @return   The output stream, once written to.
       */
    friend std::ostream& operator<< (std::ostream& os, const Reaction& r);
  };
}
#endif
