/* The ReactionNode class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionEvaluate.cpp 553 2008-05-09 13:28:22Z davidfange $
 */
#include <math.h>

#include "../mesord.hpp"

#include "ReactionNode.hpp"

#include "../ParameterTable.hpp"



  /********************
   * EVALUATE METHODS *
   ********************/
double
MesoRD::ReactionNodeAbs::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return fabs(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeAbs::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return fabs(argument -> evaluate(species));
  return 0;
}



double
MesoRD::ReactionNodeCeil::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return ceil(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCeil::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return ceil(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeExp::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return exp(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeExp::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return exp(argument -> evaluate(species));
  return 0;
}


  /* Evaluate the factorial function.  The factorial is defined only
   * for positive integers.  If the argument is not an integer, we
   * compute the factorial of the rounded integer instead.  If the
   * argument is negative, the result will always be one.  Returns
   * zero if there is no argument.
   */
double
MesoRD::ReactionNodeFactorial::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    double current   = rint(argument -> evaluate(species));
    double factorial = 1.0;

    while (current > 1)
      factorial *= current--;
    return factorial;
  }
  
  return 0;
}


double
MesoRD::ReactionNodeFactorial::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    double current   = rint(argument -> evaluate(species));
    double factorial = 1.0;

    while (current > 1)
      factorial *= current--;
    return factorial;
  }
  
  return 0;
}


double
MesoRD::ReactionNodeFloor::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return floor(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeFloor::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return floor(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeLn::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return log(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeLn::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return log(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeLog::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return log10(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeLog::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return log10(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodePower::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return pow(argument -> evaluate(species), power -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodePower::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return pow(argument -> evaluate(species), power -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeRoot::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument && degree)
    return pow(argument -> evaluate(species), 
      ((double(1.0))/(degree -> evaluate(species))));
  return 0;
}


double
MesoRD::ReactionNodeRoot::evaluate(
  const std::valarray<double>& species) const
{
  if (argument && degree)
    return pow(argument -> evaluate(species), 
      ((double(1.0))/(degree -> evaluate(species))));
  return 0;
}


double
MesoRD::ReactionNodeAnd::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) < std::numeric_limits<double>::min())
      return false;
  }
  return true;
}


double
MesoRD::ReactionNodeAnd::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) < std::numeric_limits<double>::min())
      return false;
  }
  return true;
}


double
MesoRD::ReactionNodeNot::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (operand) {
    return fabs(operand -> evaluate(species))
      < std::numeric_limits<double>::min();
  }
  return 0;
}


double
MesoRD::ReactionNodeNot::evaluate(
  const std::valarray<double>& species) const
{
  if (operand) {
    return fabs(operand -> evaluate(species))
      < std::numeric_limits<double>::min();
  }
  return 0;
}


double
MesoRD::ReactionNodeOr::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) >= std::numeric_limits<double>::min())
      return true;
  }
  return false;
}


double
MesoRD::ReactionNodeOr::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) >= std::numeric_limits<double>::min())
      return true;
  }
  return false;
}



double
MesoRD::ReactionNodeXor::evaluate(
  const std::valarray<unsigned int>& species) const
{
  int                                      n = 0;
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) >= std::numeric_limits<double>::min())
      n++;
  }
  return (n % 2);
}


double
MesoRD::ReactionNodeXor::evaluate(
  const std::valarray<double>& species) const
{
  int                                      n = 0;
  std::list<ReactionNode*>::const_iterator i;
  for (i = operands.begin(); i != operands.end(); i++) {
    if (fabs((*i) -> evaluate(species)) >= std::numeric_limits<double>::min())
      n++;
  }
  return (n % 2);
}


double
MesoRD::ReactionNodeAddition::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) + (right -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeAddition::evaluate(
  const std::valarray<double>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) + (right -> evaluate(species));
  return 0;
}



  /* We do not want to worry about division by zero here.  It should
   * be caught before we even get to this point.
   */
double
MesoRD::ReactionNodeDivision::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (left && right) {
    double denominator = right -> evaluate(species);
    
    if (fabs(denominator) >= std::numeric_limits<double>::min())
      return (left -> evaluate(species)) / denominator;
  }
  return 0;
}


double
MesoRD::ReactionNodeDivision::evaluate(
  const std::valarray<double>& species) const
{
  if (left && right) {
    double denominator = right -> evaluate(species);
    
    if (fabs(denominator) >= std::numeric_limits<double>::min())
      return (left -> evaluate(species)) / denominator;
  }
  return 0;
}


double
MesoRD::ReactionNodeMultiplication::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) * (right -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeMultiplication::evaluate(
  const std::valarray<double>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) * (right -> evaluate(species));
  return 0;
}



double
MesoRD::ReactionNodeSubtraction::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) - (right -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSubtraction::evaluate(
  const std::valarray<double>& species) const
{
  if (left && right)
    return (left -> evaluate(species)) - (right -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeEq::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (fabs((*i) -> evaluate(species) - (*j) -> evaluate(species))
        >= std::numeric_limits<double>::epsilon()) {
        return false;
      }
    }
  }

  return true;
}


double
MesoRD::ReactionNodeEq::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (fabs((*i) -> evaluate(species) - (*j) -> evaluate(species))
        >= std::numeric_limits<double>::epsilon()) {
        return false;
      }
    }
  }

  return true;
}


double
MesoRD::ReactionNodeGeq::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) < ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeGeq::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) < ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeGt::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) <= ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeGt::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) <= ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeLeq::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) > ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeLeq::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) > ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}



double
MesoRD::ReactionNodeLt::evaluate(
  const std::valarray<unsigned int>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      
      if (((*j) -> evaluate(species)) >= ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}


double
MesoRD::ReactionNodeLt::evaluate(
  const std::valarray<double>& species) const
{
  std::list<ReactionNode*>::const_iterator i;
  std::list<ReactionNode*>::const_iterator j;

  for (i = j = operands.begin(); i != operands.end(); j = i++) {
    if (i != j) {
      if (((*j) -> evaluate(species)) >= ((*i) -> evaluate(species)))
        return false;
    }
  }

  return true;
}



double
MesoRD::ReactionNodeNeq::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (left && right) {
    return fabs(left -> evaluate(species) - right -> evaluate(species))
      >= std::numeric_limits<double>::epsilon();
  }
  return 0;
}


double
MesoRD::ReactionNodeNeq::evaluate(
  const std::valarray<double>& species) const
{
  if (left && right) {
    return fabs(left -> evaluate(species) - right -> evaluate(species))
      >= std::numeric_limits<double>::epsilon();
  }
  return 0;
}



double
MesoRD::ReactionNodeArccos::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return acos(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeArccos::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return acos(argument -> evaluate(species));
  return 0;
}



double
MesoRD::ReactionNodeArccosh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(z + sqrt(z - 1) + sqrt(z + 1));
  }
  return 0;
}


double
MesoRD::ReactionNodeArccosh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(z + sqrt(z - 1) + sqrt(z + 1));
  }
  return 0;
}


double
MesoRD::ReactionNodeArccot::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return atan(1.0 / (argument -> evaluate(species)));
  return 0;
}



double
MesoRD::ReactionNodeArccot::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return atan(1.0 / (argument -> evaluate(species)));
  return 0;
}


  /* If we had complex numbers, we compute the true values of
   * arccoth(0), which is (0.5 * i * M_PI).
   */
double
MesoRD::ReactionNodeArccoth::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);

    if (fabs(z) >= std::numeric_limits<double>::min())
      return 0.5 * (log10(1.0 + 1.0 / z) - log10(1.0 - 1.0 / z)) / z;
    else
      return 0;
  }
  return 0;
}


double
MesoRD::ReactionNodeArccoth::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);

    if (fabs(z) >= std::numeric_limits<double>::min())
      return 0.5 * (log10(1.0 + 1.0 / z) - log10(1.0 - 1.0 / z)) / z;
    else
      return 0;
  }
  return 0;
}



  /* If we had complex numbers, we compute the true values of
   * arccsc(z), which is (-1.0 * i * log10(i / z + sqrt(1.0 - 1.0 / (z
   * * z)))).
   */
double
MesoRD::ReactionNodeArccsc::evaluate(
  const std::valarray<unsigned int>&) const
{
  if (argument) {
//    double z = argument -> evaluate(species);
    return 0;
  }
  return 0;
}


double
MesoRD::ReactionNodeArccsc::evaluate(
  const std::valarray<double>&) const
{
  if (argument) {
//    double z = argument -> evaluate(species);
    return 0;
  }
  return 0;
}


double
MesoRD::ReactionNodeArccsch::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(sqrt(1.0 + 1.0 / (z * z)) + 1.0 / z);
  }
  return 0;
}


double
MesoRD::ReactionNodeArccsch::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(sqrt(1.0 + 1.0 / (z * z)) + 1.0 / z);
  }
  return 0;
}


  /* If we had complex numbers, we compute the true values of
   * arcsec(z), which is (M_PI / 2.0 + i * log10(i / z + sqrt(1.0 -
   * 1.0 / (z * z)))).
   */
double
MesoRD::ReactionNodeArcsec::evaluate(
  const std::valarray<unsigned int>&) const
{
  if (argument) {
//    double z = argument -> evaluate(species);
    return 0;
  }
  return 0;
}


double
MesoRD::ReactionNodeArcsec::evaluate(
  const std::valarray<double>&) const
{
  if (argument) {
//    double z = argument -> evaluate(species);
    return 0;
  }
  return 0;
}


double
MesoRD::ReactionNodeArcsech::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(sqrt(1.0 / z - 1.0) * sqrt(1.0 + 1.0 / z) + 1.0 / z);
  }
  return 0;
}


double
MesoRD::ReactionNodeArcsech::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(sqrt(1.0 / z - 1.0) * sqrt(1.0 + 1.0 / z) + 1.0 / z);
  }
  return 0;
}


double
MesoRD::ReactionNodeArcsin::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return asin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeArcsin::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return asin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeArcsinh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(z + sqrt(z * z + 1.0));
  }
  return 0;
}


double
MesoRD::ReactionNodeArcsinh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return log10(z + sqrt(z * z + 1.0));
  }
  return 0;
}


double
MesoRD::ReactionNodeArctan::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return atan(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeArctan::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return atan(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeArctanh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (0.5 * (log10(1.0 + z) - log10(1.0 - z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeArctanh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (0.5 * (log10(1.0 + z) - log10(1.0 - z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeCos::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return cos(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCos::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return cos(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCosh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return cosh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCosh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return cosh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCot::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return 1.0 / tan(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCot::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return 1.0 / tan(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCoth::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return 1.0 / tanh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCoth::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return 1.0 / tanh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCsc::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return 1.0 / sin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCsc::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return 1.0 / sin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeCsch::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (2.0 / (exp(+1.0 * z) - exp(-1.0 * z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeCsch::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (2.0 / (exp(+1.0 * z) - exp(-1.0 * z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeSec::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return 1.0 / cos(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSec::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return 1.0 / cos(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSech::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (2.0 / (exp(+1.0 * z) + exp(-1.0 * z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeSech::evaluate(
  const std::valarray<double>& species) const
{
  if (argument) {
    const double z = argument -> evaluate(species);
    return (2.0 / (exp(+1.0 * z) + exp(-1.0 * z)));
  }
  return 0;
}


double
MesoRD::ReactionNodeSin::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return sin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSin::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return sin(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSinh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return sinh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeSinh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return sinh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeTan::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return tan(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeTan::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return tan(argument -> evaluate(species));
  return 0;
}



double
MesoRD::ReactionNodeTanh::evaluate(
  const std::valarray<unsigned int>& species) const
{
  if (argument)
    return tanh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeTanh::evaluate(
  const std::valarray<double>& species) const
{
  if (argument)
    return tanh(argument -> evaluate(species));
  return 0;
}


double
MesoRD::ReactionNodeConcentration::evaluate(
  const std::valarray<unsigned int>& species) const
{
  return static_cast<double>(species[index]);
  //  return 0;
}


double
MesoRD::ReactionNodeConcentration::evaluate(
  const std::valarray<double>& species) const
{
  if(species[index] >= 0){
    return static_cast<double>(species[index]);
  }
  else{
    return 0;
  }
}


double
MesoRD::ReactionNodeConstant::evaluate(
  const std::valarray<unsigned int>&) const
{
  return value;
}


double
MesoRD::ReactionNodeConstant::evaluate(
  const std::valarray<double>&) const
{
  return value;
}


  /* This is not implemented yet.
   */
double
MesoRD::ReactionNodeFunction::evaluate(
  const std::valarray<unsigned int>&) const
{
  return 0;
}


double
MesoRD::ReactionNodeFunction::evaluate(
  const std::valarray<double>&) const
{
  return 0;
}


  /* If the parameter is not defined, the exception should have been
   * thrown before we get to this point.
   */
double
MesoRD::ReactionNodeParameter::evaluate(
  const std::valarray<unsigned int>&) const
{
  if (parameterTable -> isDefined(id))
    return parameterTable -> getValue(id);
  return 0;
}


double
MesoRD::ReactionNodeParameter::evaluate(
  const std::valarray<double>&) const
{
  if (parameterTable -> isDefined(id))
    return parameterTable -> getValue(id);
  return 0;
}
