/* The ReactionNode class.
 *
 * Copyright 2004 Johan Elf, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionLinearise.cpp 541 2008-04-12 18:52:41Z davidfange $
 */
#include <math.h>

#include "../mesord.hpp"

#include "LinearElement.hpp"
#include "ReactionNode.hpp"



  /*************************
   * LINEARISATION METHODS *
   *************************/

  /* If the children are a concentration and a constant, then we may
   * generate a new factor and return it.  A note on notation may not
   * be amiss.  Most local variables are only a few characters in
   * length.  A leading "c" indicates concentration, "e" element, "k"
   * for constant, "f" for factor, "t" for term.  Trailing "l"
   * indicates left, "r" similarly indicates right.  Note that for
   * multiplication nodes, the constant will always be on the left
   * side.
   */
MesoRD::LinearElement*
MesoRD::ReactionNodeAddition::linearise()
{
  LinearFactor*              fl = isFactor(left -> linearise());
  LinearFactor*              fr = isFactor(right -> linearise());
  LinearTerm*                tl = isTerm(left -> linearise());
  LinearTerm*                tr = isTerm(right -> linearise());
  ReactionNodeConcentration* cl = isConcentration(left);
  ReactionNodeConcentration* cr = isConcentration(right);
  ReactionNodeConstant*      kl = isConstant(left);

  if (cl && cr) {
    fl = new LinearFactor(1, 1.0, cl -> getIndex(), 0);
    tl = new LinearTerm(1,fl);
    
    fr = new LinearFactor(1, 1.0, cr -> getIndex(), 0);
    tr = new LinearTerm(1,fr);
    
    tl -> appendNext(tr);
    return tl;

  } else if (kl && cr) {
    fl = new LinearFactor(
      1, 1.0, cr -> getIndex(), static_cast<int>(lrint(kl -> getValue())));
    return fl;
    
  } 
/* removing this faulty linearisation for now....
  else if (kl && fr) {
    fr -> setOffset((fr -> getOffset())
      + (int)rint(pow((kl -> getValue()) / (fr -> getFactor()),
        1.0 / (fr -> getExponent()))));
    return fr;

  } 
  else if (kl && tr) {
    fl = new LinearFactor(1, 1.0, -1, (off_t)rint(kl -> getValue()));
    tl = new LinearTerm();
    tl -> setFactor(fl);
    tl -> appendNext(tr);
    return tl;
    
  } 
*/
  else if (fl && fr) {
    tl = new LinearTerm(1, fl);

    tr = new LinearTerm(1, fr);

    tl -> appendNext(tr);
    return tl;
    
  } else if (fl && tr) {
    tl = new LinearTerm(1.0, fl);
    tl -> appendNext(tr);
    return tl;
    
  } else if (tl && fr) {
    tr = new LinearTerm(1.0, fr);
    tl -> appendNext(tr);
    return tl;

  } else if (tl && tr) {
    tl -> appendNext(tr);
    return tl;
  }

  return 0;
}


MesoRD::LinearElement*
MesoRD::ReactionNodeDivision::linearise()
{
  LinearFactor*              fl = isFactor(left -> linearise());
  LinearFactor*              fr = isFactor(right -> linearise());
  ReactionNodeConcentration* cl = isConcentration(left);
  ReactionNodeConcentration* cr = isConcentration(right);
  ReactionNodeConstant*      kl = isConstant(left);
  
  if (cl && cr) {
    fl = new LinearFactor(+1, 1.0, cl -> getIndex(), 0);
    fr = new LinearFactor(-1, 1.0, cr -> getIndex(), 0);
    
    fl -> appendNext(fr);
    return fl;

  } else if (kl && cr) {
    return new LinearFactor(
      -1, kl -> getValue(), cr -> getIndex(), 0);

  } else if (kl && fr) {
    fr -> setExponent(-1 * (fr -> getExponent()));
    fr -> setFactor((kl -> getValue()) * (fr -> getFactor()));
    return fr;

  } else if (fl && fr) {
    fr -> setExponent(-1 * (fr -> getExponent()));
    fl -> appendNext(fr);
    return fl;
  }

  return 0;
}


MesoRD::LinearElement*
MesoRD::ReactionNodeMultiplication::linearise()
{
  LinearFactor*              fl = isFactor(left -> linearise());
  LinearFactor*              fr = isFactor(right -> linearise());
  ReactionNodeConcentration* cl = isConcentration(left);
  ReactionNodeConcentration* cr = isConcentration(right);
  ReactionNodeConstant*      kl = isConstant(left);
  
  if (cl && cr) {
    fl = new LinearFactor(1, 1.0, cl -> getIndex(), 0);
    fr = new LinearFactor(1, 1.0, cr -> getIndex(), 0);
    
    fl -> appendNext(fr);
    return fl;

  } else if (kl && cr) {
    return new LinearFactor(
      1, kl -> getValue(), cr -> getIndex(), 0);

  } else if (kl && fr) {
    fr -> setFactor((kl -> getValue()) * (fr -> getFactor()));
    return fr;

  } else if (fl && fr) {
    fl -> appendNext(fr);
    return fl;
  } else if (fl && cr) {
    fr = new LinearFactor(1, 1.0, cr -> getIndex(), 0);
    
    fl -> appendNext(fr);
    return fl;
  }
  
  return 0;
}


MesoRD::LinearElement*
MesoRD::ReactionNodeSubtraction::linearise()
{
  LinearFactor*              fl = isFactor(left -> linearise());
  LinearFactor*              fr = isFactor(right -> linearise());
  LinearTerm*                tl = isTerm(left -> linearise());
  LinearTerm*                tr = isTerm(right -> linearise());
  ReactionNodeConcentration* cl = isConcentration(left);
  ReactionNodeConcentration* cr = isConcentration(right);
  ReactionNodeConstant*      kl = isConstant(left);

  if (cl && cr) {
    fl = new LinearFactor(1, +1.0, cl -> getIndex(), 0);
    tl = new LinearTerm(1,fl);
    
    fr = new LinearFactor(1, -1.0, cr -> getIndex(), 0);
    tr = new LinearTerm(1,fr);
    
    tl -> appendNext(tr);
    return tl;

  } else if (kl && cr) {
    fl = new LinearFactor(
      1, -1.0, cr -> getIndex(),
      static_cast<int>(lrint(-1.0 * kl -> getValue())));
    return fl;
    
  } 
/*
  else if (kl && fr) {
    if(fr -> getExponent() == 1) {
      fr -> setOffset((fr -> getOffset())
        - (int)rint(pow((kl -> getValue()) / (fr -> getFactor()),
          1.0 / (fr -> getExponent()))));
      fr -> setFactor(-1.0 * (fr->getFactor()));
    }
    return fr;

  }

  else if (kl && tr) {
    fl = new LinearFactor(1, 1.0, -1, (off_t)rint(-1.0 * (kl -> getValue())));
    tl = new LinearTerm(1,fl);
    tl -> appendNext(tr);
    return tl;
    
  }
*/ 
  else if (fl && fr) {
    tl = new LinearTerm(1, fl);

    fr -> setFactor(-1.0 *(fr -> getFactor()));
    tr = new LinearTerm(1,fr);

    tl -> appendNext(tr);
    return tl;
    
  } else if (fl && tr) {
    tl = new LinearTerm(1.0, fl);
    LinearTerm* current;
    for(current = tr; current != NULL; current = current -> getNext()) {
      current -> setCoefficient(current -> getCoefficient() * (-1));
    }
    tl -> appendNext(tr);
    return tl;
    
  } else if (tl && fr) {
    tr = new LinearTerm(-1.0, fr);
    tl -> appendNext(tr);
    return tl;

  } else if (tl && tr) {
    tl -> appendNext(tr);
    return tl;
  }

  return 0;
}

