/* The ReactionNode class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionNode.cpp 550 2008-05-07 14:24:19Z davidfange $
 */
#include <math.h>

#include <iostream>
#include <sstream>

#include "../mesord.hpp"
#include "../definition.hpp"

#include "LinearElement.hpp"
#include "ReactionNode.hpp"

#include "../ParameterTable.hpp"
#include "../SpeciesTable.hpp"
#include "../ui/OPrefixStream.hpp"



  /*****************
   * FILE CONTENTS *
   *****************/

  /* This file contains the definitions for all the methods of the
   * base class.  In addition to these, this file holds:
   *
   * [1] The fromSBML() method for every class that seriously
   *     implements it.
   *
   * [2] The getUnit() method for every class that seriously
   *     implements it.
   *
   * [3] The putTo() method for every class that seriously implements
   *     it.
   *
   * These methods are trivial most of the time.  Comments have been
   * omitted in order to make the file more manageable.
   */

  /********************************
   * THE REACTION NODE BASE CLASS *
   ********************************/

  /* Base class constructor.
   */
MesoRD::ReactionNode::ReactionNode(
  const ASTNode*, const ParameterTable* p, const SpeciesTable* s)
{
  parameterTable = p;
  speciesTable   = s;
}


  /* Base class destructor.
   */
MesoRD::ReactionNode::~ReactionNode()
{}


  /* Read in the reaction tree from SBML.  This is just one giant
   * switch statement, taking care of all the different kinds of nodes
   * there may be in the SBML AST.
   */
MesoRD::ReactionNode*
MesoRD::ReactionNode::fromSBML(
  const ASTNode *n, const ParameterTable* p, const SpeciesTable* s)
  throw(
    MesoRD::ReactionNode::ModelError,
    MesoRD::ReactionNode::NotYetImplemented,
    MesoRD::ReactionNode::UndefinedIdentifier)
{
  ReactionNode* node = 0;

  if (n) {
    switch (n -> getType()) {
      case AST_FUNCTION_ABS:
        node = new ReactionNodeAbs(n, p, s);
        break;

      case AST_FUNCTION_CEILING:
        node = new ReactionNodeCeil(n, p, s);
        break;

      case AST_FUNCTION_EXP:
        node = new ReactionNodeExp(n, p, s);
        break;

      case AST_FUNCTION_FACTORIAL:
        node = new ReactionNodeFactorial(n, p, s);
        break;

      case AST_FUNCTION_FLOOR:
        node = new ReactionNodeFloor(n, p, s);
        break;

      case AST_FUNCTION_LN:
        node = new ReactionNodeLn(n, p, s);
        break;

      case AST_FUNCTION_LOG:
        node = new ReactionNodeLog(n, p, s);
        break;

      case AST_POWER:
      case AST_FUNCTION_POWER:
        node = new ReactionNodePower(n, p, s);
        break;

      case AST_FUNCTION_ROOT:
        node = new ReactionNodeRoot(n, p, s);
        break;

      case AST_LOGICAL_AND:
        node = new ReactionNodeAnd(n, p, s);
        break;

      case AST_LOGICAL_NOT:
        node = new ReactionNodeNot(n, p, s);
        break;

      case AST_LOGICAL_OR:
        node = new ReactionNodeOr(n, p, s);
        break;

      case AST_LOGICAL_XOR:
        node = new ReactionNodeXor(n, p, s);
        break;

      case AST_DIVIDE:
        node = new ReactionNodeDivision(n, p, s);
        break;

      case AST_MINUS:
        node = new ReactionNodeSubtraction(n, p, s);
        break;

      case AST_PLUS:
        node = new ReactionNodeAddition(n, p, s);
        break;

      case AST_TIMES:
        node = new ReactionNodeMultiplication(n, p, s);
        break;

      case AST_RELATIONAL_EQ:
        node = new ReactionNodeEq(n, p, s);
        break;

      case AST_RELATIONAL_GEQ:
        node = new ReactionNodeGeq(n, p, s);
        break;

      case AST_RELATIONAL_GT:
        node = new ReactionNodeGt(n, p, s);
        break;

      case AST_RELATIONAL_LEQ:
        node = new ReactionNodeLeq(n, p, s);
        break;

      case AST_RELATIONAL_LT:
        node = new ReactionNodeLt(n, p, s);
        break;

      case AST_RELATIONAL_NEQ:
        node = new ReactionNodeNeq(n, p, s);
        break;

      case AST_FUNCTION_ARCCOS:
        node = new ReactionNodeArccos(n, p, s);
        break;

      case AST_FUNCTION_ARCCOSH:
        node = new ReactionNodeArccosh(n, p, s);
        break;

      case AST_FUNCTION_ARCCOT:
        node = new ReactionNodeArccot(n, p, s);
        break;

      case AST_FUNCTION_ARCCOTH:
        node = new ReactionNodeArccoth(n, p, s);
        break;

      case AST_FUNCTION_ARCCSC:
        node = new ReactionNodeArccsc(n, p, s);
        break;

      case AST_FUNCTION_ARCCSCH:
        node = new ReactionNodeArccsch(n, p, s);
        break;

      case AST_FUNCTION_ARCSEC:
        node = new ReactionNodeArcsec(n, p, s);
        break;

      case AST_FUNCTION_ARCSECH:
        node = new ReactionNodeArcsech(n, p, s);
        break;

      case AST_FUNCTION_ARCSIN:
        node = new ReactionNodeArcsin(n, p, s);
        break;

      case AST_FUNCTION_ARCSINH:
        node = new ReactionNodeArcsinh(n, p, s);
        break;

      case AST_FUNCTION_ARCTAN:
        node = new ReactionNodeArctan(n, p, s);
        break;

      case AST_FUNCTION_ARCTANH:
        node = new ReactionNodeArctanh(n, p, s);
        break;

      case AST_FUNCTION_COS:
        node = new ReactionNodeCos(n, p, s);
        break;

      case AST_FUNCTION_COSH:
        node = new ReactionNodeCosh(n, p, s);
        break;

      case AST_FUNCTION_COT:
        node = new ReactionNodeCot(n, p, s);
        break;

      case AST_FUNCTION_COTH:
        node = new ReactionNodeCoth(n, p, s);
        break;

      case AST_FUNCTION_CSC:
        node = new ReactionNodeCsc(n, p, s);
        break;

      case AST_FUNCTION_CSCH:
        node = new ReactionNodeCsch(n, p, s);
        break;

      case AST_FUNCTION_SEC:
        node = new ReactionNodeSec(n, p, s);
        break;

      case AST_FUNCTION_SECH:
        node = new ReactionNodeSech(n, p, s);
        break;

      case AST_FUNCTION_SIN:
        node = new ReactionNodeSin(n, p, s);
        break;

      case AST_FUNCTION_SINH:
        node = new ReactionNodeSinh(n, p, s);
        break;

      case AST_FUNCTION_TAN:
        node = new ReactionNodeTan(n, p, s);
        break;

      case AST_FUNCTION_TANH:
        node = new ReactionNodeTanh(n, p, s);
        break;

      case AST_CONSTANT_E:
      case AST_CONSTANT_FALSE:
      case AST_CONSTANT_PI:
      case AST_CONSTANT_TRUE:
      case AST_INTEGER:
      case AST_RATIONAL:
      case AST_REAL:
      case AST_REAL_E:
        try {
          node = new ReactionNodeConstant(n, p, s);
        } catch (ModelError& e) {
          throw ModelError(e.what());
        }
        break;

      case AST_NAME:
        
        if (!n -> isNumber() && !n -> isOperator()) {
          
          const std::string name = std::string(n -> getName());
          
          if (s && (s -> isDefinedById(name))) {
            const std::string volume = "volume_subVolume";
            
            node = new ReactionNodeMultiplication(
              new ReactionNodeConstant(
                1.0 / ((p -> getValue(volume)) * Constant::Avogadro),
                Unit::mole() / Unit::litre(),
                p,
                s),
              new ReactionNodeConcentration(n, p, s), p, s);
            
          } else if (p && (p -> isDefined(name))) {
            if ( p -> isConstant(name)) {
              node = new ReactionNodeConstant(
                p -> getValue(name),
                p -> getUnit(name),
                p,
                s);
            } else {
              try {
                node = new ReactionNodeParameter(n, p, s);
              } catch (ModelError& e) {
                throw ModelError(e.what());
              } catch (UndefinedIdentifier e) {
                throw UndefinedIdentifier(e.what());
              }
            }
          } else {
            std::stringstream message;
            message << "Identifier \""
                    << name
                    << "\" is not defined.";
            throw UndefinedIdentifier((message.str()).c_str());
          }
        } else {
          throw UndefinedIdentifier();
        }
        break;

      case AST_FUNCTION:
      case AST_FUNCTION_PIECEWISE:
      case AST_FUNCTION_DELAY:
      case AST_NAME_TIME:
      case AST_LAMBDA:
        throw NotYetImplemented();
        break;
      case AST_UNKNOWN:
        std::stringstream message;
        message << "Usage of operators outside of the SBML standard";
        throw ModelError((message.str()).c_str());
    }
  }
  
  return node;
}



  /*****************
   * NODE QUERYING *
   *****************/

  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeAddition*
MesoRD::ReactionNode::isAddition(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeAddition)))
    return static_cast<MesoRD::ReactionNodeAddition*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeConcentration*
MesoRD::ReactionNode::isConcentration(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeConcentration)))
    return static_cast<MesoRD::ReactionNodeConcentration*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeConstant*
MesoRD::ReactionNode::isConstant(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeConstant)))
    return static_cast<MesoRD::ReactionNodeConstant*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeDivision*
MesoRD::ReactionNode::isDivision(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeDivision)))
    return static_cast<MesoRD::ReactionNodeDivision*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::LinearFactor*
MesoRD::ReactionNode::isFactor(MesoRD::LinearElement* element)
{
  if (element && (typeid(*element) == typeid(MesoRD::LinearFactor)))
    return static_cast<MesoRD::LinearFactor*>(element);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeMultiplication*
MesoRD::ReactionNode::isMultiplication(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeMultiplication)))
    return static_cast<MesoRD::ReactionNodeMultiplication*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::ReactionNodeSubtraction*
MesoRD::ReactionNode::isSubtraction(MesoRD::ReactionNode* node)
{
  if (node && (typeid(*node) == typeid(MesoRD::ReactionNodeSubtraction)))
    return static_cast<MesoRD::ReactionNodeSubtraction*>(node);
  return 0;
}


  /* Attempt to cast the node.  Make sure to check that it is not null
   * before we begin.
   */
MesoRD::LinearTerm*
MesoRD::ReactionNode::isTerm(MesoRD::LinearElement* element)
{
  if (element && (typeid(*element) == typeid(MesoRD::LinearTerm)))
    return static_cast<MesoRD::LinearTerm*>(element);
  return 0;
}


  /* Do constant folding on a node until it does not change any more.
   */
MesoRD::ReactionNode*
MesoRD::ReactionNode::refold(ReactionNode* node) 
  throw( MesoRD::ReactionNode::InvalidUnit )
{
  ReactionNode* tmp;

  do {
    tmp  = node;
    try {
      node = node -> fold();
    } catch ( InvalidUnit i ) {
      throw InvalidUnit(i.what());
    }
  } while (node != tmp);

  return node;
}


  /* Multiply the entire tree by Avogadro's constant and the volume of
   * a sub volume.
   */
MesoRD::ReactionNode*
MesoRD::ReactionNode::toConcentration() 
{
  try {
    std::string   volume = "volume_subVolume";
    ReactionNode* tmp    = new ReactionNodeMultiplication(
      new ReactionNodeConstant(
        ((parameterTable -> getValue("volume_subVolume"))
          * Constant::Avogadro),
        (parameterTable -> getUnit("volume_subVolume")) * Unit::mole(),
        parameterTable,
        speciesTable),
      this, parameterTable, speciesTable);
    
    return tmp;
  } catch (ModelError& e) {
    throw ModelError(e.what());
  } catch (UndefinedIdentifier& e) {
    throw UndefinedIdentifier(e.what());
  }
}

/* checks the unit of the reaction tree, if it can be converted return
 * multiply the tree with the conversion multiplier. If it can be converted
 * to molar per second, muliply with the conversion factor and run the
 * toConcetration function.
 */
MesoRD::ReactionNode*
MesoRD::ReactionNode::toPerSecond(std::ostream* outstream_log)
  throw( MesoRD::Unit::InvalidConversion )
{
  try {
    double conv = (this -> getUnit()).toUnit(Unit::perSecond());
      if (outstream_log) {
        (*outstream_log) << "WARNING: Reaction is per second. "
        << "Assuming user knows what user does."
        << std::endl;
      }
    ReactionNode* tmp = 
      new ReactionNodeMultiplication(
        new ReactionNodeConstant(conv,Unit::dimensionless(),
          parameterTable,
          speciesTable),
        this, 
        parameterTable, 
        speciesTable);
    return tmp;
  } catch(MesoRD::Unit::InvalidConversion&) {
    ;
  }
  
  try {
    double conv = (this -> getUnit()).toUnit(Unit::molarPerSecond());

    ReactionNode* tmp = 
      new ReactionNodeMultiplication(
        new ReactionNodeConstant(conv,Unit::dimensionless(),
          parameterTable,
          speciesTable),
        toConcentration(), 
        parameterTable, 
        speciesTable);
    return tmp;

  } catch(MesoRD::Unit::InvalidConversion& e) {
    throw MesoRD::Unit::InvalidConversion(e);
  }
  return 0;
}

  
    
  /********************
   * FROMSBML METHODS *
   ********************/

MesoRD::ReactionNode*
MesoRD::ReactionNodeAbs::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ABS)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCeil::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_CEILING)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeExp::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_EXP)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeFactorial::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_FACTORIAL)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeFloor::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_FLOOR)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeLn::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_LN)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeLog::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_LOG)) {
    argument = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
    base     = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodePower::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_POWER)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(),  parameterTable, speciesTable);
    power    = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeRoot::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ROOT)) {
    argument = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
    degree   = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeAnd::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_LOGICAL_AND)) {
    for (unsigned int i = 0; i < ASTNode_getNumChildren(n); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeNot::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_LOGICAL_NOT))
    operand  = ReactionNode::fromSBML(
      n -> getLeftChild(),  parameterTable, speciesTable);
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeOr::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_LOGICAL_OR)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeXor::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_LOGICAL_XOR)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeAddition::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_PLUS)) {
    left = ReactionNode::fromSBML(
      ASTNode_getChild(n, 0),  parameterTable, speciesTable);

    ReactionNodeAddition* top = this;
    for (unsigned int i = 1; i < n -> getNumChildren(); i++) {
      ReactionNode* current = ReactionNode::fromSBML(
        n -> getChild(i),  parameterTable, speciesTable);

      if (i == (n -> getNumChildren() - 1)) {
        top -> setRightChild(current);
        break;

      } else {
        ReactionNodeAddition* top_new = new ReactionNodeAddition(
          current, 0, parameterTable, speciesTable);
        top -> setRightChild(top_new);
        top = top_new;
      }
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeDivision::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_DIVIDE)) {
    left  = ReactionNode::fromSBML(
      n -> getLeftChild(),  parameterTable, speciesTable);
    right = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeMultiplication::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_TIMES)) {
    left = ReactionNode::fromSBML(
      ASTNode_getChild(n, 0),  parameterTable, speciesTable);

    ReactionNodeMultiplication* top = this;
    for (unsigned int i = 1; i < n -> getNumChildren(); i++) {
      ReactionNode* current = ReactionNode::fromSBML(
        n -> getChild(i),  parameterTable, speciesTable);

      if (i == (n -> getNumChildren() - 1)) {
        top -> setRightChild(current);
        break;

      } else {
        ReactionNodeMultiplication* top_new = new ReactionNodeMultiplication(
          current, 0, parameterTable, speciesTable);
        top -> setRightChild(top_new);
        top = top_new;
      }
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeSubtraction::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_MINUS)) {
    left  = ReactionNode::fromSBML(
      n -> getLeftChild(),  parameterTable, speciesTable);
    right = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeEq::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_EQ)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeGeq::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_GEQ)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeGt::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_GT)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeLeq::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_LEQ)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeLt::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_LT)) {
    for (unsigned int i = 0; i < n -> getNumChildren(); i++) {
      operands.push_back(
        ReactionNode::fromSBML(
          n -> getChild(i), parameterTable, speciesTable));
    }
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeNeq::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_RELATIONAL_NEQ)) {
    left  = ReactionNode::fromSBML(
      n -> getLeftChild(),  parameterTable, speciesTable);
    right = ReactionNode::fromSBML(
      n -> getRightChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccos::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCOS)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccosh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCOSH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccot::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCOT)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccoth::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCOTH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccsc::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCSC)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArccsch::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCCSCH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArcsec::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCSEC)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArcsech::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCSECH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArcsin::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCSIN)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArcsinh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCSINH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArctan::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCTAN)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeArctanh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_ARCTANH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCos::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_COS)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCosh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_COSH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCot::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_COT)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCoth::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_COTH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCsc::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_CSC)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeCsch::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_CSCH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeSec::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_SEC)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeSech::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_SECH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeSin::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_SIN)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeSinh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_SINH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeTan::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_TAN)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeTanh::fromSBML(const ASTNode* n)
{
  if (n && (n -> getType() == AST_FUNCTION_TANH)) {
    argument = ReactionNode::fromSBML(
      n -> getLeftChild(), parameterTable, speciesTable);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeConcentration::fromSBML(const ASTNode* n)
{
  if (!n -> isNumber() && !n -> isOperator()) {
    const char* name = n -> getName();
    
    if (speciesTable && (speciesTable -> isDefinedById(name)))
      index = speciesTable -> getIndexById(name);
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeConstant::fromSBML(const ASTNode* n)
  throw(MesoRD::ReactionNode::ModelError)
{
  if ((n -> getType() == AST_RATIONAL)
    || (n -> getType() == AST_REAL)
    || (n -> getType() == AST_REAL_E)) {
    if (ASTNode_isReal(n)) {
      value = ASTNode_getReal(n);
    } else {
      throw ModelError();
    }
  } else if (n -> getType() == AST_INTEGER) {
    value = double(1.0) * ASTNode_getInteger(n);
  } else if (n -> getType() == AST_CONSTANT_E) {
    value = MesoRD::Constant::Exponentiale;
  } else if (n -> getType() == AST_CONSTANT_PI) {
    value = MesoRD::Constant::Pi;
  } else if (n -> getType() == AST_CONSTANT_FALSE) {
    value = 0;
  } else if (n -> getType() == AST_CONSTANT_TRUE) {
    value = 1;
  } else {
    throw ModelError();
  }
  return this;
}


MesoRD::ReactionNode*
MesoRD::ReactionNodeParameter::fromSBML(const ASTNode* n)
  throw(
    MesoRD::ReactionNode::ModelError,
    MesoRD::ReactionNode::UndefinedIdentifier)
{
  if ((n -> getType() == AST_NAME)
    && !ASTNode_isNumber(n)
    && !ASTNode_isOperator(n)) {
    id = std::string(ASTNode_getName(n));
    
    if (parameterTable) {
      if (parameterTable -> isDefined(id)) {
        if (parameterTable -> isConstant(id)) {
          throw ModelError();
        }
      } else {
        std::stringstream message;
        message << "Identifier \""
                << id
                << "\" is not defined.";
        throw UndefinedIdentifier((message.str()).c_str());
      }
    } else {
      throw ModelError();
    }
  } else {
    throw ModelError();
  }
  return this;
}



  /******************************
   * UNIT QUERY METHODS METHODS *
   ******************************/

MesoRD::Unit
MesoRD::ReactionNodeAbs::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) 
    return argument -> getUnit();
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCeil::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument)
    return argument -> getUnit();
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeExp::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if(u.isDimensionless())
      return u;
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeFactorial::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();

    if (u.isDimensionless())
      return u;
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeFloor::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument)
    return argument -> getUnit();
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeLn::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();

    if (u.isDimensionless())
      return u;
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeLog::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u_argument = argument -> getUnit();
    Unit u_base     = base -> getUnit();

    if (u_argument.isDimensionless() &&  u_base.isDimensionless())
      return u_argument;
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodePower::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit power_unit = power -> getUnit();
        
    if(power_unit != MesoRD::Unit::dimensionless())
      throw InvalidUnit();
    
    double  p = power -> evaluate(
      std::valarray<unsigned int>((unsigned int)0, 0));

    Unit u = argument -> getUnit();
    return (u ^ p);
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeRoot::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit degree_unit = degree -> getUnit();
    if(degree_unit != MesoRD::Unit::dimensionless())
      throw InvalidUnit();
    
    double  p = degree -> evaluate(
      std::valarray<unsigned int>((unsigned int)0, 0));
    
    Unit u = argument -> getUnit();
    return (u ^ (1/p));
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeAnd::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return Unit::dimensionless();
}


MesoRD::Unit
MesoRD::ReactionNodeNot::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return Unit::dimensionless();
}


MesoRD::Unit
MesoRD::ReactionNodeOr::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return Unit::dimensionless();
}


MesoRD::Unit
MesoRD::ReactionNodeXor::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return Unit::dimensionless();
}


MesoRD::Unit
MesoRD::ReactionNodeAddition::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (left && right) {
    try {
      return (left -> getUnit() + right -> getUnit());
    } catch(Unit::InvalidConversion&) {
      throw InvalidUnit();
    }
  }

  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeDivision::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (left && right) {
    try {
      return (left -> getUnit() / right -> getUnit());
    } catch(Unit::InvalidConversion&) {
      ;
    }
  }

  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeMultiplication::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (left && right) {
    try {
      return (left -> getUnit() * right -> getUnit());
    } catch(Unit::InvalidConversion&) {
      ;
    }
  }

  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeSubtraction::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (left && right) {
    try {
      return (left -> getUnit() - right -> getUnit());
    } catch(Unit::InvalidConversion&) {
      ;
    }
  }

  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeEq::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  Unit                                     u = (*operands.begin())
    -> getUnit();
  std::list<ReactionNode*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (((*i) -> getUnit()) != u)
      throw InvalidUnit();
  }
  
  return u;
}


MesoRD::Unit
MesoRD::ReactionNodeGeq::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  Unit                                     u = (*operands.begin())
    -> getUnit();
  std::list<ReactionNode*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (((*i) -> getUnit()) != u)
      throw InvalidUnit();
  }
  
  return u;
}


MesoRD::Unit
MesoRD::ReactionNodeGt::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  Unit                                     u = (*operands.begin())
    -> getUnit();
  std::list<ReactionNode*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (((*i) -> getUnit()) != u)
      throw InvalidUnit();
  }
  
  return u;
}


MesoRD::Unit
MesoRD::ReactionNodeLeq::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  Unit                                     u = (*operands.begin())
    -> getUnit();
  std::list<ReactionNode*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (((*i) -> getUnit()) != u)
      throw InvalidUnit();
  }
  
  return u;
}


MesoRD::Unit
MesoRD::ReactionNodeLt::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  Unit                                     u = (*operands.begin())
    -> getUnit();
  std::list<ReactionNode*>::const_iterator i;

  for (i = operands.begin(); i != operands.end(); i++) {
    if (((*i) -> getUnit()) != u)
      throw InvalidUnit();
  }
  
  return u;
}


MesoRD::Unit
MesoRD::ReactionNodeNeq::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (left && right) {
    if ((left -> getUnit()) == (right -> getUnit()))
      return (left -> getUnit());
  }
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccos::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccosh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccot::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccoth::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccsc::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArccsch::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArcsec::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArcsech::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArcsin::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArcsinh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArctan::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeArctanh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCos::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCosh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCot::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCoth::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCsc::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeCsch::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeSec::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeSech::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeSin::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeSinh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeTan::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeTanh::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  if (argument) {
    Unit u = argument -> getUnit();
    
    if (u.isDimensionless())
      return u;
  }
  
  throw InvalidUnit();
}


MesoRD::Unit
MesoRD::ReactionNodeConcentration::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return Unit::dimensionless();
}


MesoRD::Unit
MesoRD::ReactionNodeConstant::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return unit;
}


MesoRD::Unit
MesoRD::ReactionNodeParameter::getUnit() const
  throw(MesoRD::ReactionNode::InvalidUnit)
{
  return parameterTable -> getUnit(id);
}



  /*****************
   * PUTTO METHODS *
   *****************/

std::ostream&
MesoRD::ReactionNodeAbs::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<abs/>"   << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCeil::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<ceiling/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeExp::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<exp/>"   << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeFactorial::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"      << std::endl;
  opfs_level1 << "<factorial/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"     << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeFloor::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<floor/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeLn::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<ln/>"    << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeLog::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<log/>"     << std::endl;
  opfs_level1 << "<logbase>"  << std::endl;
  opfs_level2 << *base;
  opfs_level1 << "</logbase>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodePower::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<power/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level1 << *power;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeRoot::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  OPrefixStream opfs_level2(opfs_level1.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<root/>"   << std::endl;
  opfs_level1 << "<degree>"  << std::endl;
  opfs_level2 << *degree;
  opfs_level1 << "</degree>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeAnd::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<and/>"  << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeNot::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<not/>"   << std::endl;
  opfs_level1 << *operand;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeOr::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<or/>"  << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeXor::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<xor/>"  << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeAddition::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<plus/>"  << std::endl;
  opfs_level1 << *left;
  opfs_level1 << *right;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeDivision::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<divide/>" << std::endl;
  opfs_level1 << *left;
  opfs_level1 << *right;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeMultiplication::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<times/>" << std::endl;
  opfs_level1 << *left;
  opfs_level1 << *right;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeSubtraction::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<minus/>" << std::endl;
  opfs_level1 << *left;
  opfs_level1 << *right;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeEq::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<eq/>"   << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeGeq::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<geq/>"  << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeGt::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<gt/>"   << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeLeq::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<leq/>"  << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeLt::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");
  std::list<ReactionNode*>::const_iterator i;

  opfs_level0 << "<apply>" << std::endl;
  opfs_level1 << "<lt/>"   << std::endl;

  for (i = operands.begin(); i != operands.end(); i++)
    opfs_level1 << *(*i);

  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeNeq::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"  << std::endl;
  opfs_level1 << "<neq/>"   << std::endl;
  opfs_level1 << *left;
  opfs_level1 << *right;
  opfs_level0 << "</apply>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccos::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccos/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccosh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccosh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccot::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccot/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccoth::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccoth/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccsc::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccsc/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArccsch::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccsch/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArcsec::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arcsec/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArcsech::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arcsech/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArcsin::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arcsin/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArcsinh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arcsinh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArctan::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arctan/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeArctanh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arctanh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCos::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccos/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCosh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccosh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCot::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccot/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCoth::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccoth/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCsc::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arccsc/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeCsch::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arccsch/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeSec::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arcsec/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeSech::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arcsech/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeSin::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arcsin/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeSinh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arcsinh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeTan::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"   << std::endl;
  opfs_level1 << "<arctan/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"  << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeTanh::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(),          "");
  OPrefixStream opfs_level1(opfs_level0.rdbuf(), "  ");

  opfs_level0 << "<apply>"    << std::endl;
  opfs_level1 << "<arctanh/>" << std::endl;
  opfs_level1 << *argument;
  opfs_level0 << "</apply>"   << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeConcentration::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(), "");

  opfs_level0 << "<ci>"
              << speciesTable -> getId(index)
              << "</ci>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeConstant::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(), "");

  opfs_level0 << "<cn>"
              << value
              << "</cn>" << std::endl;

  return os;
}


std::ostream&
MesoRD::ReactionNodeParameter::putTo(std::ostream& os) const
{
  OPrefixStream opfs_level0(os.rdbuf(), "");

  opfs_level0 << "<ci>"
              << id
              << "</ci>" << std::endl;

  return os;
}
