/* The GetOpt class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: GetOpt.cpp 559 2008-09-10 19:46:33Z davidfange $
 */
#include <iomanip>
#include <iostream>

#include <stdio.h>
#include <sbml/SBMLTypes.h>

#include "../mesord.hpp"
#ifdef HAVE_VIEWER
#  include "../Visual.hpp"
#endif

#include "GetOpt.hpp"
#include "../SpeciesTable.hpp"
#include "../UnitTable.hpp"


  /* For technical reasons one should always provide a non-inline
   * implementation for virtual destructors, even if it is trivial.
   * This will make sure that the compiler generates a vtable.
   * Without a vtable, there will be unresolved symbols when it comes
   * to linking.
   */
MesoRD::GetOpt::~GetOpt()
{}


  /* If anything goes wrong with the SBML file at this point, we
   * simply ignore it.  Those mistakes will be caught elsewhere, when
   * the file is parsed for real.
   *
   * Should start looking at a replacement.  Entry point is
   * System::fromSBML().
   */
void
MesoRD::GetOpt::fromSBML()
  throw(MesoRD::Species::ModelError, MesoRD::System::ModelError)
{
  std::ofstream* outstream_log = 0;

  if (filename.length() > 0) {
	SBMLReader* reader = new SBMLReader();
	SBMLDocument* d = reader -> readSBML(filename);
    Model*        m = d -> getModel();
    unsigned int errors = d -> getNumErrors();

    if ((d -> getNumErrors() == 0)) {

        /* Suppressing some messages from the reading of species by
         * reading the units first.  Disregarding all error messages
         * at the moment.
         * FIXME!! Update manual to clearly state this fact.
         */
      
      unitTable.fromSBML(m -> getListOfUnitDefinitions());
      speciesTable.clear();
      for (size_t i = 0; i < m -> getNumSpecies(); i++) {
        Species* s;
        try {
          s = new Species(
            m -> getSpecies(static_cast<unsigned int>(i)), 0, 0, 0);
        } catch (MesoRD::Species::ModelError&) {
            //Errors are handeled elsewhere, ie do nothing.  
        }

        speciesTable.addSpecies(*s);
      }
    }
  }
}


  /* Read the times from file, convert them to doubles, and push them
   * onto a list.
   */
std::list<std::string>
MesoRD::GetOpt::getTimes(const std::string filename)
  throw (MesoRD::GetOpt::FileNotFoundError)
{
  double                 time;
  std::ifstream          time_stream(filename.c_str());
  std::list<std::string> times;

  if (!time_stream)
    throw FileNotFoundError(filename.c_str());

  while (time_stream >> time) {
    std::stringstream time_string;
    time_string << time;
    times.push_back(time_string.str());
  }

  time_stream.close();
  return times;
}


  /* Find the index of the species in the SpeciesTable.  Then find the
   * index of the desired colour in the predefined list of colours.
   * Finally set the colour in the <code>colours</code> vector at the
   * appropriate index to the colour values from the predefined list.
   */
#ifdef HAVE_VIEWER
void
MesoRD::GetOpt::setColour(
  const std::string& species, const std::string& colour)
{
  off_t i_species = speciesTable.getIndex(species);

  if (i_species >= 0) {
    off_t i_colour = -1;

    for (size_t i = 0; i < MESORD_VIEWER_COLOURS; i++) {
      if (MESORD_VIEWER_COLOUR[i].name == colour) {
        i_colour = static_cast<off_t>(i);
        break;
      }
    }

    if (i_colour >= 0) {
      colours[i_species] -> red   = MESORD_VIEWER_COLOUR[i_colour].red;
      colours[i_species] -> green = MESORD_VIEWER_COLOUR[i_colour].green;
      colours[i_species] -> blue  = MESORD_VIEWER_COLOUR[i_colour].blue;
      colours[i_species] -> name  = MESORD_VIEWER_COLOUR[i_colour].name;
    } else {
      colours[i_species] -> red   = -1;
      colours[i_species] -> green = -1;
      colours[i_species] -> blue  = -1;
      colours[i_species] -> name  = "";
    }
  }
}
#endif


  /* Clear the <code>colours</code> vector.  Then cycle through the
   * list of predefined colours, adding elements to
   * <code>colours</code> until there is one colour per species.
   */
#ifdef HAVE_VIEWER
void
MesoRD::GetOpt::setDefaultColours()
{
  colours.clear();

  for (size_t i = 0; i < speciesTable.getSpeciesNumber(); i++) {
    VisualColour* colour = new VisualColour;

/*#ifdef WIN32
    colour -> red   = -1;
    colour -> blue  = -1;
    colour -> green = -1;
    colour -> name  = "";
#else*/
    colour -> red   = MESORD_VIEWER_COLOUR[i % MESORD_VIEWER_COLOURS].red;
    colour -> blue  = MESORD_VIEWER_COLOUR[i % MESORD_VIEWER_COLOURS].green;
    colour -> green = MESORD_VIEWER_COLOUR[i % MESORD_VIEWER_COLOURS].blue;
    colour -> name  = MESORD_VIEWER_COLOUR[i % MESORD_VIEWER_COLOURS].name;
//#endif

    colours.push_back(colour);
  }
}
#endif


  /* Dump a formatted summary of the options to the stream
   * <code>os</code>.
   *
   * FIXME!! This output could be improved and shortened, by for
   * instance printing the concatenation of size and unit, and so on.
   *
   * FIXME!! The summary function should be make aware of compilation
   * flags, such as viewer and deterministic.
   */
namespace MesoRD {
  std::ostream& operator<< (std::ostream& os, const GetOpt& go) {
    const size_t width = 28;

    std::cout.fill(' ');
    os << std::right
       << "                  OPTION SUMMARY"                << std::endl
       << "  ---------------------------------------------" << std::endl;

    os << std::setw(width) << std::right << "checkpoint iterations : ";
    if (go.checkpoint_iterations > 0)
      os << go.checkpoint_iterations                        << std::endl;
    else
      os << "[disabled]"                                    << std::endl;

    os << std::setw(width) << "checkpoint time : ";
    if (go.checkpoint_time > 0)
      os << go.checkpoint_time << " second(s)"             << std::endl;
    else
      os << "[disabled]"                                    << std::endl;

    os << std::setw(width) << "filename : "
       << go.filename                                       << std::endl

       << std::setw(width) << "help : "
       << (go.help ? "yes" : "no")                          << std::endl

       << std::setw(width) << "max iterations : "
       << go.max_iterations_significand                     << std::endl

       << std::setw(width) << "max iterations scale : "
       << go.max_iterations_exponent                        << std::endl

       << std::setw(width) << "max time : "
       << go.max_time << " second(s)"                       << std::endl

       << std::setw(width) << "pedantic : "
       << (go.pedantic ? "yes" : "no")                      << std::endl

       << std::setw(width) << "progress bar : "
       << (go.progressBars ? "yes" : "no")                  << std::endl

       << std::setw(width) << "simulate : "
       << (go.simulate ? "yes" : "no")                      << std::endl

       << std::setw(width) << "size : "
       << go.size << " " << go.unit                            << std::endl

       << std::setw(width) << "species dimension : "
       << go.species_dimension                              << std::endl

       << std::setw(width) << "usage : "
       << (go.usage ? "yes" : "no")                         << std::endl

       << std::setw(width) << "version : "
       << (go.version ? "yes" : "no")                       << std::endl

       << std::setw(width) << "viewer : "
       << (go.viewer ? "yes" : "no")                        << std::endl

       << std::setw(width) << "old state : "
       << (go.readOldState ? "yes" : "no")                  << std::endl

       << std::setw(width) << "old state time : "
       << go.readOldStateTime                               << std::endl

       << std::setw(width) << "sparse : "
       << (go.sparse ? "yes" : "no")                        << std::endl

       << std::setw(width) << "deterministic : "
       << (go.deterministic ? "yes" : "no")                 << std::endl;

    if (go.deterministic) {
      os << std::endl
         << std::setw(width) << "convergence tolerance : "
         << go.itMetConvergenceTolerance                    << std::endl
         << std::setw(width) << "maximum integration its. : "
         << go.itMetMaxIterations                           << std::endl
         << std::setw(width) << "integration method : "
         << go.integrationMethod                            << std::endl;
    }
    
    return os;
  }
}
