/* The GetOpt class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: GetOpt.hpp 561 2008-11-25 20:10:53Z davidfange $
 */
#ifndef MESORD_UI_GETOPT_HPP
#define MESORD_UI_GETOPT_HPP 1


#include "../mesord.hpp"
#include "../SpeciesTable.hpp"
#include "../UnitTable.hpp"
#include "../System.hpp"


#include <iostream>
#include <list>
#include <string>
#include <vector>
#include <sys/stat.h>

#ifdef WIN32
#  include <io.h>
#  ifndef R_OK
#    define R_OK 04
#  endif
#  ifndef W_OK
#    define W_OK 02
#  endif
#endif

#ifdef _MSC_VER
#  if  _MSC_VER > 1000 && !defined(__INTEL_COMPILER)
#    define access(path,mode) _access(path,mode)

/* one would imaging that this conversion should be arranged
 * by the compiler if the _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
 * is set, but ones needs to be more explicit....
 */
#    define sscanf sscanf_s
#  endif
#endif

#ifdef HAVE_VIEWER
#  include "../Visual.hpp"
#endif

namespace MesoRD
{
    /**
     * @brief Abstract class for obtaining options from the user.
     *
     * This abstract class provides a framework for classes supposed
     * to obtain run-time options from the user.  In principle it is
     * just a bunch of getter and setter functions, one for each
     * supported option.  In addition, the GetOpt class attempts to
     * relieve inheriting classes of as much input validation a
     * possible.
     *
     * The general idea is that when validation fails, the setter
     * method returns <code>false</code>, and an error is logged.  The
     * value should then <em>not</em> be changed.  How to deal with
     * the failure of the setter function is up to the inheriting
     * class.
     *
     * There is a fair amount of string handling done in this class.
     * The reason is that the GetOpt class is concerned with user
     * interaction, and users, unlike machines, seem to prefer textual
     * strings over binary objects.
     *
     * @todo In general, when a validation fails, the setter method
     * should return false, and an error should be logged using the as
     * of yet non-existent messaging system.
     *
     * @todo Ensure that overflow is handled gracefully, for instance
     * in the iteration counters.
     */
  class GetOpt
  {
  public:

      /**
       * @brief Class for reporting non-accessible files.
       *
       * If a there is a file access problem, because the requested
       * file cannot be found, or because the user is lacking the
       * proper permissions, this exception will be thrown.  This
       * enables the program to abort or to offer recovery options
       * once the user has given the program problematic input.
       *
       * @todo This exception should probably be renamed to
       * <code>FileNotFound</code>.  Maybe it should even be moved
       * somewhere else, because this is a very general kind of
       * exception.  One could also implement functions to further
       * analyse the cause of the failure.
       */
    class FileNotFoundError : public std::runtime_error
    {
    private:
      const char* filename;

    public:
        /**
         * @brief The FileNotFoundError value constructor.
         * 
         * This constructor enables the thrower to set all the
         * information associated with the exception.  Note that any
         * of the arguments may be <code>null</code> pointers.
         * Invalid pointers on the other hand should be avoided, as
         * they have the potential to cause much confusion.
         *
         * @param filename The name of the file with access problems.
         */
      inline FileNotFoundError(const char* filename)
        : std::runtime_error("FileNotFound"), filename(filename)
        {}


        /**
         * @brief Get the name of the file which could not be
         * accessed.
         *
         * This function returns a pointer to the name of the file
         * which the program could not access.  This may be the
         * <code>null</code> pointer.
         *
         * @return A pointer to the name of the file with access
         *         problems.
         */
      inline const char*
      getFilename() const
        {
          return filename;
        }
    };

  protected:
      /**
       * @brief Parses a string into a <code>double</code> value.
       * 
       * Leading or trailing garbage is not permitted.
       *
       * @todo This is code duplication with respect to the Annotation
       * class.  That should be fixed.
       *
       * @param      str The string to parse.
       * @param[in]  val Space for the <code>double</code> value.
       * @param[out] val The <code>double</code> value, if parsing was
       *                 successful, untouched otherwise.
       * @return         <code>true</code> if parsing was successful,
       *                 <code>false</code> otherwise.
       */
    static inline bool
    strtodouble(const char* str, double& val)
      {
        double d = 0;
        int    n = 0;
        
        if (str && sscanf(str, "%lf%n", &d, &n) == 1 
          && (static_cast<size_t>(n) == strlen(str))) {
          val = d;
          return true;
        }
        return false;
      }


      /**
       * @brief Parses a string into a <code>long int</code> value.
       * 
       * @copydoc strtodouble()
       *
       * @todo This is code duplication with respect to the Annotation
       * class.  That should be fixed.
       *
       * @param      str The string to parse.
       * @param[in]  val Space for the <code>long int</code> value.
       * @param[out] val The <code>long int</code> value, if parsing
       *                 was successful, untouched otherwise.
       * @return         <code>true</code> if parsing was successful,
       *                 <code>false</code> otherwise.
       */
    static inline bool
    strtolongint(const char* str, long int& val)
      {
        int      n  = 0;
        long int li = 0;
        if (str && sscanf(str, "%li%n", &li, &n) == 1
          && static_cast<size_t>(n) == strlen(str)) {
          val = li;
          return true;
        }
        return false;
      }


      /**
       * @brief Parses a string into a <code>size_t</code> value.
       * 
       * @copydoc strtodouble()
       *
       * Anything less than zero results in @p val being set to zero.
       * Because ISO C++ does not support the <code>z</code> scanf(3)
       * length modifier, a little workaround is provided in order to
       * parse <code>size_t</code>.
       *
       * @todo This is code duplication with respect to the Annotation
       * class.  That should be fixed.
       *
       * @param      str The string to parse.
       * @param[in]  val Space for the <code>size_t</code> value.
       * @param[out] val The <code>size_t</code> value, if parsing was
       *                 successful, untouched otherwise.
       * @return         <code>true</code> if parsing was successful,
       *                 <code>false</code> otherwise.
       */
    static inline bool
    strtosizet(const char* str, size_t& val)
      {
        long int li = 0;
        if (strtolongint(str, li)) {
          val = li >= 0 ? li : 0;
          return true;
        }
        return false;
      }


  private:
      /**
       * @brief The number of iterations between checkpoints.
       *
       * A value of zero indicates that iteration-based checkpoints
       * are disabled.
       */
    size_t checkpoint_iterations;

      /**
       * @brief The number of simulation seconds between checkpoints.
       *
       * A value of zero indicates that time-based checkpoints are
       * disabled.
       */
    double checkpoint_time;

      /**
       * @brief All the colours one could imagine.
       *
       * @todo This member should probably be removed!
       */
#ifdef HAVE_VIEWER
    std::vector<VisualColour*> colours;
#endif

      /**
       * @brief Whether deterministic simulation is requested.
       */
    bool deterministic;

      /**
       * @brief The name of the file which holds the model.
       */
    std::string filename;

      /**
       * @brief Whether to do full state checkpointing or not.
       */
    bool fullStateCheckPoint;

      /**
       * @brief Whether to display the help message or not.
       */
    bool help;

      /**
       * @brief The name of the integration method.
       *
       * @todo This could probably be validated better, just like the
       * unit.
       */
    std::string integrationMethod;

      /**
       * @brief The convergence tolerance for integration.
       */
    double itMetConvergenceTolerance;

      /**
       * @brief The maximal number of iterations for integration.
       */
    size_t itMetMaxIterations;

      /**
       * @brief Whether to write output to a log file or not.
       */
    bool log;

      /**
       * @brief The maximal number of iterations.
       *
       * The actual maximal number of iterations is
       * @f$max\_iterations_{significand} \cdot
       * 10^{max\_iterations{exponent}}@f$.
       *
       * @note According to Goldberg (David Goldberg, <i>What Every
       * Computer Scientist Should Know About Floating-Point
       * Arithmetic</i>, <i>ACM Computing Surveys</i>, <b>23(1)</b>,
       * 5--48, (March 1991)), the term <em>significand</em> was
       * introduced by Forsythe and Moler in 1967 and "has generally
       * replaced the older term <em>mantissa</em>".
       */
    size_t max_iterations_significand;

      /**
       * @brief The scale of the maximal number of iterations.
       *
       * @copydoc GetOpt::max_iterations
       */
    size_t max_iterations_exponent;

      /**
       * @brief The maximal simulation duration.
       */
    double max_time;

      /**
       * @brief The directory where output files are written.
       */
    std::string outputdir;

      /**
       * @brief Whether to be pedantic or not.
       */
    bool pedantic;

      /**
       * @brief Whether to display progress bars or not.
       */
    bool progressBars;

      /**
       * @brief Whether to read the old state from file or not.
       */
    bool readOldState;

      /**
       * @brief The time of the state as read from file.
       */
    long int readOldStateTime;

      /**
       * @brief Whether to actually simulate something or not.
       */
    bool simulate;

      /**
       * @brief The size of a subvolume.
       *
       * The unit of this size is given in GetOpt::unit.
       */
    double size;

      /**
       * @brief Whether to use sparse output or not.
       */
    bool sparse;

      /**
       * @brief Number of molecules to plot per dimension.
       */
    size_t species_dimension;

      /**
       * @brief The table of species.
       *
       * @todo Perhaps the entire model should be parsed here?
       * Another option is to remove this member completely!
       */
    SpeciesTable speciesTable;

      /**
       * @brief Time step for deterministic simulation.
       */
    double timeStep;

      /**
       * @brief The unit of the subvolume size.
       *
       * The magnitude of the size is given in GetOpt::size.
       *
       * @todo This should perhaps be a proper Unit?
       */
    std::string unit;

      /**
       * @brief The table of units.
       *
       * @todo Perhaps the entire model should be parsed here?
       * Another option is to remove this member completely!
       */
    UnitTable unitTable;

      /**
       * @brief Whether to display usage information or not.
       */
    bool usage;

      /**
       * @brief Whether to display version information or not.
       */
    bool version;

      /**
       * @brief Whether to launch the graphical viewer or not.
       */
    bool viewer;


  public:
      /**
       * @brief The GetOpt class constructor.
       *
       * The constructor sets up what is hoped to be suitable defaults
       * for all supported options.  The default constructor is used
       * for some STL data members, as well as for the speciesTable
       * and unitTable data members.
       *
       * @param outstream_log A pointer the log file.  If this is
       *                      <code>null</code>, no log file will be
       *                      created.
       */
    inline
    GetOpt()
      : checkpoint_iterations(1000000000),
        checkpoint_time(1.0),
        deterministic(false),
        filename(""),
        fullStateCheckPoint(true),
        help(false),
        integrationMethod("BDF2"),
        itMetConvergenceTolerance(1e-6),
        itMetMaxIterations(100),
        log(false),
        max_iterations_significand(100000000),
        max_iterations_exponent(0),
        max_time(100.0),
        outputdir(""),
        pedantic(false),
        progressBars(true),
        readOldState(false),
        readOldStateTime(0),
        simulate(true),
        size(1.0e-7),
        sparse(true),
        species_dimension(2),
        timeStep(1.0e-3),
        unit("metre"),
        usage(false),
        version(false),
#ifdef HAVE_VIEWER
        viewer(true)
#else
        viewer(false)
#endif
      {}


      /**
       * @brief Destructor for the GetOpt class.
       *
       * Does nothing at all.  For technical reasons, destructors of
       * abstract classes should not be defined inline.
       */
    virtual ~GetOpt();


      /**
       * @brief Display help information to the user.
       *
       * This function exists because, according to the GNU Coding
       * Standards, all programs should support options to display
       * help and version information to the user.
       */
    virtual void
    displayHelp() const = 0;


      /**
       * @brief Display version information to the user.
       *
       * @copydoc displayHelp()
       */
    virtual void
    displayVersion() const = 0;


      /**
       * @brief Partially parse the model from the set file.
       *
       * This function uses the GetOpt::filename, so it is crucial
       * that the GetOpt::filename of this instance has been set up
       * correctly prior to this call.
       *
       * @todo This function should probably be removed!
       */
    void fromSBML()
      throw(Species::ModelError, System::ModelError);


      /**
       * @brief Get a textual list of all the species.
       *
       * This function extracts all the species from the species table
       * and puts their names into an STL <code>list</code>.  The list
       * is sorted once it comes out from here.
       *
       * @todo This function should probably be removed!
       *
       * @return A textual list of all the species in the
       *         SpeciesTable.
       */
    inline std::list<std::string>
    getAllSpecies() const
      {
        std::list<std::string> species;
        for (size_t i = 0; i < speciesTable.getSpeciesNumber(); i++)
          species.push_back(speciesTable.getKey(i));
        species.sort();
        return species;
      }


      /**
       * @brief Get a textual representation of the colour
       * corresponding to a certain species.
       *
       * This function returns the name of the colour corresponding to
       * a given species.  If no colour was found, the empty string is
       * returned.  The species is identified by its name in the
       * single argument.  If the program was compiled without support
       * for the viewer, this function always returns the empty
       * string.
       *
       * @todo This function should probably be removed!
       *
       * @param species The name of the species the colour of which is
       *                sought.
       * @return        A textual representation of the colour of the
       *                species, or the empty string on error.
       */
    std::string
    getColour(const std::string& species) const
      {
#ifdef HAVE_VIEWER
        if (species.length() > 0) {
          off_t index = speciesTable.getIndex(species);
          return colours[index] -> name;
        }
#endif
        return "";
      }


      /**
       * @brief Get all defined colours.
       *
       * This function returns an STL vector of all defined colours.
       * The getColours() function is only available if the program is
       * compiled with support for the viewer.
       *
       * @todo This function should probably be removed!  It is simple
       * enough for callers to do it themselves.
       *
       * @return A vector of all defined colours.
       */
#ifdef HAVE_VIEWER
    inline const std::vector<VisualColour*>&
    getColours() const
      {
        return colours;
      }
#endif


      /**
       * @brief Get the number of iterations between checkpoints.
       *
       * A value of zero indicates that iteration-based checkpoints
       * are disabled.
       *
       * @return The number of iterations between checkpoints.
       */
    inline size_t
    getOptionCheckpointIterations() const
      {
        return checkpoint_iterations;
      }


      /**
       * @brief Set the number of iterations between checkpoints.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionCheckpointIterations(const char* str)
      {
        if (!strtosizet(str, checkpoint_iterations)) {
          std::cerr << "Could not parse checkpoint iterations"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the number of iterations between checkpoints.
       *
       * @param ci The number of iterations between checkpoints.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionCheckpointIterations(const size_t ci)
      {
        checkpoint_iterations = ci;
        return true;
      }


      /**
       * @brief Get the number of simulation seconds between
       * checkpoints.
       *
       * A value of zero indicates that time-based checkpoints are
       * disabled.
       *
       * @return The number of seconds between checkpoints.
       */
    inline double
    getOptionCheckpointTime() const
      {
        return checkpoint_time;
      }


      /**
       * @brief Set the number of simulation seconds between
       * checkpoints.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionCheckpointTime(const char* str)
      {
        if (!strtodouble(str, checkpoint_time)) {
          std::cerr << "Could not parse checkpoint time"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the number of simulation seconds between
       * checkpoints.
       *
       * @param t The number of simulation seconds between
       *          checkpoints.
       * @return  This function always returns <code>true</code>.
       */
    inline bool
    setOptionCheckpointTime(const double t)
      {
        checkpoint_time = t;
        return true;
      }


      /**
       * @brief Check whether deterministic simulation is requested.
       *
       * @return <code>true</code> when asked to do a deterministic
       *         simulation, <code>false</code> otherwise.
       */
    inline bool
    getOptionDeterministic() const
      {
        return deterministic;
      }


      /**
       * @brief Set whether to do deterministic simulation.
       *
       * @param deterministic Whether to do deterministic simulation
       *                      or not.
       * @return              This function always returns
       *                      <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionDeterministic(const bool deterministic)
      {
        this -> deterministic = deterministic;
        return true;
      }


      /**
       * @brief Get the name of the file defining the model.
       *
       * @return The name of the file defining the model.
       */
    inline const std::string&
    getOptionFilename() const
      {
        return filename;
      }


      /**
       * @brief Set the name of the file defining the model.
       *
       * @warning This function calls access(2), which is a potential
       * security hole!
       *
       * @todo This function should probably build the model.  Once
       * the model is partially built, one can verify a lot of other
       * things, such as the unit and the colour assignment for each
       * species.
       *
       * @param filename The name of the file defining the model.
       * @return         <code>true</code> if the file is readable,
       *                 <code>false</code> otherwise.
       */
    inline bool
    setOptionFilename(const char* filename)
      {
        if ( access(filename, R_OK) == 0) {
          this -> filename = filename;
          return true;
        }
        return false;
      }


      /**
       * @brief Check whether full state checkpoints are requested.
       *
       * @return <code>true</code> when asked to do full state
       *         checkpoints, <code>false</code> otherwise.
       */
    inline bool
    getOptionFullStateCheckPoint() const
      {
        return fullStateCheckPoint;
      }


      /**
       * @brief Set whether to do deterministic simulation.
       *
       * @param fullStateCheckPoint Whether to do full state
       *                            checkpoints or not.
       * @return                    This function always returns
       *                            <code>true</code>, as it cannot
       *                            fail.
       */
    inline bool
    setOptionFullStateCheckPoint(const bool fullStateCheckPoint)
      {
        this -> fullStateCheckPoint = fullStateCheckPoint;
        return true;
      }


      /**
       * @brief Check whether help is requested.
       *
       * @return <code>true</code> when help is requested,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionHelp() const
      {
        return help;
      }


      /**
       * @brief Set whether to display help information.
       *
       * @param help Whether to display help information or not.
       * @return     This function always returns <code>true</code>,
       *             as it cannot fail.
       */
    inline bool
    setOptionHelp(const bool help)
      {
        this -> help = help;
        return true;
      }


      /**
       * @brief Get the name of the integration method.
       *
       * @return The name of the integration method.
       */
    inline std::string
    getOptionIntegrationMethod() const
      {
        return integrationMethod;
      }


      /**
       * @brief Set the name of the integration method.
       *
       * @todo It should be possible to do better validation in this
       * function.
       *
       * @param integrationMethod The name of the integration method.
       * @return                  This function always returns
       *                          <code>true</code>, as it cannot
       *                          fail.
       */
    inline bool
    setOptionIntegrationMethod(const char* integrationMethod)
      {
        this -> integrationMethod = integrationMethod;
        return true;
      }


      /**
       * @brief Get the integration convergence tolerance.
       *
       * @return The integration convergence tolerance.
       */
    inline double
    getOptionItMetConvergenceTolerance() const
      {
        return itMetConvergenceTolerance;
      }


      /**
       * @brief Set the integration convergence tolerance.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionItMetConvergenceTolerance(const char* str)
      {
        if (!strtodouble(str, itMetConvergenceTolerance)) {
          std::cerr << "Could not parse convergence tolerance"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the integration convergence tolerance.
       *
       * @param ct The integration convergence tolerance.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionItMetConvergenceTolerance(const double ct)
      {
        itMetConvergenceTolerance = ct;
        return true;
      }


      /**
       * @brief Get the maximal number of iterations for the
       * integration.
       *
       * @return The maximal number of iterations for the integration.
       */
    inline size_t
    getOptionItMetMaxIterations() const
      {
        return itMetMaxIterations;
      }


      /**
       * @brief Set the maximal number of iterations for the
       * integration.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionItMetMaxIterations(const char* str)
      {
        if (!strtosizet(str, itMetMaxIterations)) {
          std::cerr << "Could not parse iterations per integration"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the maximal number of iterations for the
       * integration.
       *
       * @param mi The maximal number of iterations for the
       *           integration.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionItMetMaxIterations(const size_t mi)
      {
        itMetMaxIterations = mi;
        return true;
      }


      /**
       * @brief Check whether to write to a log file or not.
       *
       * @return <code>true</code> when a log file should be written,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionLog() const
      {
        return log;
      }


      /**
       * @brief Set whether to write a log file.
       *
       * @todo There should also be an option to set the log file.
       * Then, this option could be entirely removed.
       *
       * @param log Whether to write a log file or not.
       * @return    This function always returns <code>true</code>, as
       *            it cannot fail.
       */
    inline bool
    setOptionLog(const bool log)
      {
        this -> log = log;
        return true;
      }


      /**
       * @brief Get the maximal number of iterations to perform.
       *
       * The actual maximal number of iterations is
       * @f$max_{iterations} \cdot 10^{max_{iterations\_scale}}@f$.
       *
       * @return The maximal number of iterations to perform.
       */
    inline size_t
    getOptionMaxIterations() const
      {
        return max_iterations_significand;
      }


      /**
       * @brief Set the maximal number of iterations.
       *
       * @copydoc getOptionMaxIterations()
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionMaxIterations(const char* str)
      {
        if (!strtosizet(str, max_iterations_significand)) {
          std::cerr << "Could not parse maximal iterations"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the maximal number of iterations.
       *
       * @copydoc getOptionMaxIterations()
       *
       * @param ct The maximal number of iterations.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionMaxIterations(const size_t mi)
      {
        max_iterations_significand = mi;
        return true;
      }


      /**
       * @brief Get scale of the maximal number of iterations to
       * perform.
       *
       * @copydoc getOptionMaxIterations()
       *
       * @return The scale of the maximal number of iterations to
       *         perform.
       */
    inline size_t
    getOptionMaxIterationsScale() const
      {
        return max_iterations_exponent;
      }


      /**
       * @brief Set the scale of the maximal number of iterations.
       *
       * @copydoc getOptionMaxIterations()
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionMaxIterationsScale(const char* str)
      {
        if (!strtosizet(str, max_iterations_exponent)) {
          std::cerr << "Could not parse iterations scale"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the scale of the maximal number of iterations.
       *
       * @copydoc getOptionMaxIterations()
       *
       * @param mis The scale of the maximal number of iterations.
       * @return    This function always returns <code>true</code>.
       */
    inline bool
    setOptionMaxIterationsScale(const size_t mis)
      {
        max_iterations_exponent = mis;
        return true;
      }


      /**
       * @brief Get the maximal duration of a simulation, in seconds.
       *
       * @return The maximal duration of a simulation in seconds.
       */
     inline double
     getOptionMaxTime() const
      {
        return max_time;
      }


      /**
       * @brief Set the maximal simulation duration, in seconds.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionMaxTime(const char* str)
      {
        if (!strtodouble(str, max_time)) {
          std::cerr << "Could not parse maximal simulation time"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the maximal simulation duration, in seconds.
       *
       * @param  t The maximal simulation duration, in seconds.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionMaxTime(const double t)
      {
        max_time = t;
        return true;
      }


      /**
       * @brief Get the directory where output files are written.
       *
       * @todo Using fixed output/input files is probably horrible
       * practise.  This should be fixed at some point.
       *
       * @return The directory where output files are written.
       */
    inline const std::string&
    getOptionOutputDir() const
      {
        return outputdir;
      }


      /**
       * @brief Set the directory where output files are written.
       *
       * @warning This function calls access(2), which is a potential
       * security hole!
       *
       * @param outputdir The directory where output files are
       *                  written.
       * @return          <code>true</code> if @p outputdir is a
       *                  writable directory, <code>false</code>
       *                  otherwise.
       */
    inline bool
    setOptionOutputDir(const char* outputdir)
      {
        struct stat sb;
        if (stat(outputdir, &sb)     == 0
          && ( sb.st_mode & S_IFMT )     == S_IFDIR
          && access(outputdir, W_OK) == 0) {
          this -> outputdir = outputdir;
          return true;
        }
        return false;
      }


      /**
       * @brief Check whether pedantry is requested.
       *
       * @todo Document what this pedantry is supposed to do.
       *
       * @return <code>true</code> when pedantry is requested,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionPedantic() const
      {
        return pedantic;
      }


      /**
       * @brief Set whether to be pedantic.
       *
       * @param pedantic Whether to be pedantic or not.
       * @return         This function always returns
       *                 <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionPedantic(const bool pedantic)
      {
        this -> pedantic = pedantic;
        return true;
      }


      /**
       * @brief Check whether progress bars are requested.
       *
       * @return <code>true</code> when progress bars are requested,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionProgressBars() const
      {
        return progressBars;
      }


      /**
       * @brief Set whether to use progress bars.
       *
       * @param progressBars Whether to use progress bars or not.
       * @return             This function always returns
       *                     <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionProgressBars(const bool progressBars)
      {
        this -> progressBars = progressBars;
        return true;
      }


      /**
       * @brief Check whether to read the old state or not.
       *
       * @return <code>true</code> when the old state should be read,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionReadOldState() const
      {
        return readOldState;
      }


      /**
       * @brief Set whether to read the old state or not.
       *
       * @param readOldState Whether to do read the old state or not.
       * @return             This function always returns
       *                     <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionReadOldState(const bool readOldState)
      {
        this -> readOldState = readOldState;
        return true;
      }


      /**
       * @brief Get the time of the old state that is to be read.
       *
       * @return The time of the old state that is to be read.
       */
    inline long int
    getOptionReadOldStateTime() const
      {
        return readOldStateTime;
      }


      /**
       * @brief Set the time of the old state.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionReadOldStateTime(const char* str)
      {
        if (!strtolongint(str, readOldStateTime)) {
          std::cerr << "Could not parse time of old state"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the time of the old state.
       *
       * @param  st The time of the old state.
       * @return    This function always returns <code>true</code>.
       */
    inline bool
    setOptionReadOldStateTime(const long int st)
      {
        readOldStateTime = st;
        return true;
      }


      /**
       * @brief Check whether to simulate or not.
       *
       * @return <code>true</code> when simulation should be
       *         performed, <code>false</code> otherwise.
       */
    inline bool
    getOptionSimulate() const
      {
        return simulate;
      }


      /**
       * @brief Set whether to simulate.
       *
       * @param simulate Whether to simulate or not.
       * @return         This function always returns
       *                 <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionSimulate(const bool simulate)
      {
        this -> simulate = simulate;
        return true;
      }


      /**
       * @brief Get the size of a subvolume.
       *
       * The unit of this size is obtained from getOptionUnit().
       *
       * @return The size of a subvolume.
       */
    inline double
    getOptionSize() const
      {
        return size;
      }


      /**
       * @brief Set size of a subvolume.
       *
       * The unit of this size is set using setOptionUnit().
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionSize(const char* str)
      {
        if (!strtodouble(str, size)) {
          std::cerr << "Could not parse subvolume size"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set size of a subvolume.
       *
       * @copydoc setOptionSize
       *
       * @param  s The size of a subvolume.
       * @return   This function always returns <code>true</code>.
       */
    inline bool
    setOptionSize(const double s)
      {
        size = s;
        return true;
      }


      /**
       * @brief Check whether sparse output is requested.
       *
       * @return <code>true</code> when sparse output is requested,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionSparse() const
      {
        return sparse;
      }


      /**
       * @brief Set whether to write sparse output.
       *
       * @param sparse Whether to write sparse output or not.
       * @return       This function always returns <code>true</code>,
       *               as it cannot fail.
       */
    inline bool
    setOptionSparse(const bool sparse)
      {
        this -> sparse = sparse;
        return true;
      }


      /**
       * @brief Get the number of molecules to plot per dimension.
       *
       * @return The number of molecules to plot per dimension.
       */
    inline size_t
    getOptionSpeciesDimension() const
      {
        return species_dimension;
      }


      /**
       * @brief Set the number of molecules to plot per dimension.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionSpeciesDimension(const char* str)
      {
        if (!strtosizet(str, species_dimension)) {
          std::cerr << "Could not parse species dimension"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the number of molecules to plot per dimension.
       *
       * @param  sd The number of molecules to plot per dimension.
       * @return    This function always returns <code>true</code>.
       */
    inline bool
    setOptionSpeciesDimension(const size_t sd)
      {
        species_dimension = sd;
        return true;
      }


      /**
       * @brief Get the time step for use in deterministic simulation.
       *
       * @return The time step for use in deterministic simulation.
       */
    inline double
    getOptionTimeStep() const
      {
        return timeStep;
      }


      /**
       * @brief Set the time step for use in deterministic simulation.
       *
       * @param str The string to parse.
       * @return    <code>true</code> if parsing was successful,
       *            <code>false</code> otherwise.
       */
    inline bool
    setOptionTimeStep(const char* str)
      {
        if (!strtodouble(str, timeStep)) {
          std::cerr << "Could not parse time step"
                    << std::endl;
          return false;
        }
        return true;
      }


      /**
       * @brief Set the time step for use in deterministic simulation.
       *
       * @param  ts The time step for use in deterministic simulation.
       * @return    This function always returns <code>true</code>.
       */
    inline bool
    setOptionTimeStep(const double ts)
      {
        timeStep = ts;
        return true;
      }


      /**
       * @brief Get the unit of the subvolume size.
       *
       * The size of this unit is obtained from getOptionSize().
       *
       * @return The unit of the subvolume size.
       */
    inline const std::string&
    getOptionUnit() const
      {
        return unit;
      }


      /**
       * @brief Set the unit of the subvolume size.
       *
       * The size of this unit is set using setOptionSize().
       *
       * @param unit The unit of the subvolume size.
       * @return     This function always returns <code>true</code>,
       *             as it cannot fail.
       */
    inline bool
    setOptionUnit(const char* unit)
      {
        this -> unit = unit;
        return true;
      }


      /**
       * @brief Check whether to print usage information or not.
       *
       * @return <code>true</code> when usage information should be
       *         printed, <code>false</code> otherwise.
       */
    inline bool
    getOptionUsage() const
      {
        return usage;
      }


      /**
       * @brief Set whether to print usage information.
       *
       * @param usage Whether to print usage information or not.
       * @return      This function always returns <code>true</code>,
       *              as it cannot fail.
       */
    inline bool
    setOptionUsage(const bool usage)
      {
        this -> usage = usage;
        return true;
      }


      /**
       * @brief Check whether to print version information or not.
       *
       * @return <code>true</code> when version information should be
       *         printed, <code>false</code> otherwise.
       */
    inline bool
    getOptionVersion() const
      {
        return version;
      }


      /**
       * @brief Set whether to print version information.
       *
       * @param version Whether to print version information or not.
       * @return        This function always returns
       *                <code>true</code>, as it cannot fail.
       */
    inline bool
    setOptionVersion(const bool version)
      {
        this -> version = version;
        return true;
      }


      /**
       * @brief Check whether to launch the viewer or not.
       *
       * @return <code>true</code> when the viewer should be launched,
       *         <code>false</code> otherwise.
       */
    inline bool
    getOptionViewer() const
      {
        return viewer;
      }


      /**
       * @brief Set whether to launch the viewer.
       *
       * @param viewer Whether to launch the viewer or not.
       * @return       This function always returns <code>true</code>,
       *               as it cannot fail.
       */
    inline bool
    setOptionViewer(const bool viewer)
      {
        this -> viewer = viewer;
        return true;
      }


      /**
       * @brief Get a list of all length units defined in the model.
       *
       * This function returns a list of textual representations of
       * all length units defined in the model.  The
       * <code>metre</code> unit is guaranteed to be in that list,
       * since it is the default SBML unit for length.
       *
       * @bug Whenever the file changes, someone should make sure to
       * call fromSBML()!
       *
       * @todo Maybe getLengthUnits() should be moved from UnitTable
       * to here?
       *
       * @return A list of all the length units, sorted
       *         alphabetically.
       */
    inline std::list<std::string>
    getPossibleUnits() const
      {
        std::list<std::string> units_length;
        units_length = unitTable.getLengthUnits();
        units_length.push_back(std::string("metre"));
        units_length.sort();
        return units_length;
      }


      /**
       * @brief Read the simulation times from file and store them in
       * an STL list.
       *
       * If there are problems accessing the file, a FileNotFoundError
       * exception will be thrown.  In such cases, it is up to the
       * caller to determine the reason for failure.
       *
       * @param filename The filename containing the simulation times.
       * @return         An ordered list of simulation times.
       */
    std::list<std::string>
    getTimes(const std::string filename = MESORD_TIMEFILE_NAME)
      throw(FileNotFoundError);


      /**
       * @brief Set the colour corresponding to a given species.
       *
       * Sets the colour in the <code>colours</code> vector
       * corresponding to @p species to the values from the predefined
       * list of colours corresponding to @p colour.  If @p species
       * was found, but @p colour was not, a colour with all negative
       * values will be set.  The setColour() function is only
       * available if the program is compiled with support for the
       * viewer.
       *
       * @todo This function should probably be removed!
       *
       * @param species The species, the colour of which should be
       *                updated.
       * @param colour  The new colour of @p species.
       */
#ifdef HAVE_VIEWER
    void
    setColour(const std::string& species, const std::string& colour);
#endif


      /**
       * @brief Set up a default colour scheme.
       *
       * Sets a default colour for all the species in the
       * GetOpt::colours vector.  The colour scheme is obtained by
       * cycling through the list of predefined colours.  The
       * setDefaultColours() function is only available if the program
       * is compiled with support for the viewer.
       *
       * @todo This function should probably be removed!
       */
#ifdef HAVE_VIEWER
    void
    setDefaultColours();
#endif


      /**
       * @brief Output operator for the GetOpt class.
       *
       * The output operator dumps a formatted summary of the options
       * and their current values to @p os.
       *
       * @param os The stream onto which to write.
       * @param go The instance of the GetOpt class to dump.
       * @return   The stream, once written to.
       */
    friend std::ostream&
    operator<<(std::ostream& os, const GetOpt& go);
  };
}
#endif
