/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2007, Johan Elf, David Fange, Johan Hattne
 *
 * Permission to use, copy, modify, and/or distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear
 * in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
  * $Id: ProgressBar.hpp 541 2008-04-12 18:52:41Z davidfange $
 */
#ifndef MESORD_UI_PROGRESSBAR_HPP
#define MESORD_UI_PROGRESSBAR_HPP 1

#include <string>
#include <valarray>

#include "../mesord.hpp"


namespace MesoRD
{

    /**
     * @brief Interface for progress bars.
     *
     * This class defines an interface for progress bars.  Progress
     * bars, or progress meters, are used to display the progress of
     * any process to the user.  Since one can imagine several
     * different flavours of progress bars, it makes sense to capture
     * the commonality between different meters in an abstract class.
     *
     * The process which has its progress displayed by a progress bar,
     * informs the meter of any progress by calling the setFraction()
     * function.  The fraction argument is a floating point value
     * between zero and one: zero indicates that the process is just
     * about starting, one means it has completed.  Progress bars also
     * allow the user to cancel whatever process, the progress of
     * which it displays.  The pending status is polled by calling
     * isAlive().
     *
     * @todo This entire class hierarchy, as well as most other things
     * in this directory, should probably be moved to a new
     * <code>MesoRD::UI</code> namespace.
     */
  class ProgressBar
  {
  protected:

      /**
       * @brief The current progress of the progress bar.
       *
       * The progress stored as a floating point number between zero
       * and one inclusive.  A value of zero means there has not been
       * any progress yet.  A value of one means the process is
       * completed.
       */
    double fraction;

      /**
       * @brief The time of the progress bar.
       *
       * @bug This should probably be removed as it is too specialised
       * towards simulation progress bars.
       */
    double time;

      /**
       * @brief The title of the progress bar.
       *
       * A progress bar has a title which holds information about the
       * process, the progress of which is monitored.
       */
    std::string title;

      /**
       * @brief The iteration number.
       *
       * @bug This should probably be removed as it is too specialised
       * towards simulation progress bars.
       */
    std::valarray<size_t> iteration;


  public:

      /**
       * @brief Constructor for the ProgressBar class.
       *
       * This constructor will initialise an instance of the
       * ProgressBar class.  Initially, the fraction is set to zero.
       * The title is set to the empty string by default.
       *
       * @param t The title of the progress bar.
       */
    inline ProgressBar(const std::string& t = "")
      : fraction(0), time(0), title(t), iteration(0)
      {}


      /**
       * @brief Destructor for the ProgressBar class.
       *
       * Does nothing at all.  For technical reasons, destructors of
       * abstract classes should not be defined inline.
       */
    virtual ~ProgressBar();


      /**
       * @brief Poll the status of the progress bar.
       *
       * This function is used to poll the status of the progress bar.
       * If the user has informed the progress bar that whatever
       * process that drives the bar should stop, isAlive() will
       * return <code>false</code>.  Otherwise, isAlive() returns
       * <code>true</code>.
       *
       * @bug Maybe <code>alive()</code> is a better name?
       */
    virtual bool
    isAlive() = 0;


      /**
       * @brief Run the progress bar.
       *
       * While a progress bar is alive, it can display some sort of
       * animation, to indicate that everything is going well -- or
       * going at all, for that matter.  Furthermore, a running
       * progress bar should continually update its display, so that
       * it always shows the approximately correct fraction to the
       * user.  Continuous update is important for another reason as
       * well: as long as the progress bar moves, the user should be
       * reasonably sure that the process is not hung.
       *
       * Calling this function will start this sort of animation
       * business.  The function will generally not return until the
       * progress has completed.  A call to run() should be matched by
       * a call to stop().
       */
    virtual void
    run() = 0;


      /**
       * @brief Set the fraction of the progress bar.
       *
       * This function will update the fraction of completion of the
       * progress bar.  The argument should be in the interval zero to
       * one, inclusive.  If necessary the argument will be truncated
       * to the correct interval, i.e. calling setFraction() with
       * argument <code>2.3</code> will set the fraction to
       * <code>1.0</code>.
       *
       * @param f The fraction of completion of the progress bar.
       */
    inline virtual void
    setFraction(const double f)
      {
        fraction = MesoRD::max(double(0), MesoRD::min(double(1), f));
      }


      /**
       * @brief Set the iteration of the progress bar.
       *
       * @bug This should probably be removed as it is too specialised
       * towards simulation progress bars.
       *
       * @param i The new iteration of the progress bar.
       */
    inline virtual void
    setIteration(const std::valarray<size_t>& i)
      {
        iteration = i;
      }


      /**
       * @brief Set the time of the progress bar.
       *
       * @bug This should probably be removed as it is too specialised
       * towards simulation progress bars.
       *
       * @param t The new time of the progress bar.
       */
    virtual void
    setTime(const double t)
      {
        time = MesoRD::max(double(0), t);
      }


      /**
       * @brief Set the title of the progress bar.
       *
       * This function will update the title of the progress bar.  The
       * title is supposed to inform the user about the process, of
       * which the progress bar is metering the progress.
       *
       * @param t The new title of the progress bar.
       */
    virtual void
    setTitle(const std::string& t)
      {
        title = t;
      }


      /**
       * @brief Run the progress bar.
       *
       * While a progress bar is alive, it can display some sort of
       * animation, to indicate that everything is going well.  This
       * sort of animation is started by run() and terminated by
       * calling stop().  Calling stop does not imply the process
       * using the progress bar has really finished.  The progress bar
       * itself will linger around until its destructor is called.
       */
    virtual void
    stop() = 0;
  };
}
#endif /* ifndef  MESORD_UI_PROGRESSBAR_HPP */
