/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2007, Johan Elf, David Fange, Johan Hattne
 *
 * Permission to use, copy, modify, and/or distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear
 * in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: SignalHandler.hpp 533 2007-10-24 09:55:48Z hattne $
 */
#ifndef MESORD_UI_SIGNALHANDLER_HPP
#define MESORD_UI_SIGNALHANDLER_HPP 1

#include <csignal>
#include <map>


namespace MesoRD
{
    /**
     * @brief Minimal interface for event handlers.
     *
     * This interface only contains a single virtual function to
     * handle a signal.  The function defines a callback which hooks
     * into SignalHandler, which, in turn, deals directly with the
     * operating system's signalling API.
     *
     * There are no globals or statics in this class, nor should there
     * be any in classes inheriting from EventHandler.  The state is
     * maintained locally within the object.  For portability, signal
     * handlers should only set variables of type
     * <code>sig_atomic_t</code> and then return.
     */
  class EventHandler
  {
  public:
      /**
       * @brief Destructor for the signal dispatcher.
       *
       * Does nothing at all.  For technical reasons, destructors of
       * abstract classes should not be defined inline.
       */
    virtual
    ~EventHandler();


      /**
       * @brief Hook function for the signal dispatcher.
       *
       * This function should be overridden by any concrete classes in
       * order to respond appropriately to received signals.  It is
       * called from SignalHandler::dispatch() in the SignalHandler
       * class.
       *
       * @param sig The received signal.
       */
    virtual int
    handleSignal(const int sig) = 0;
  };



    /**
     * @brief Handler interface to lower-level signalling API.
     *
     * This class encapsulates a collection of concrete event handlers
     * in an STL <code>map</code>.  It relieves other code from
     * dealing directly with the operating system's signalling API; by
     * deploying the adaptor pattern, the underlying C API is mapped
     * to the functions in this class.  This is generally believed to
     * be an improvement over the original C-style approach, because
     * no global variables need to be involved.
     *
     * Event handlers are registered in the <code>map</code> using the
     * registerHandler() function.  Registered handlers can similarly
     * be removed with the removeHandler() function.  Because a
     * typical operating system only keeps <em>one</em> table for
     * mapping signals to handlers, there need only be <em>one</em>
     * instance of the SignalHandler class.  Hence, SignalHandler is a
     * singleton.
     *
     * Signal handling is implemented using several design patterns
     * from Erich Gamma, Richard Helm, Ralph Johnson, and John
     * Vlissides, <i>Design Patterns: Elements of Reusable
     * Object-Oriented Software,</i> Addison-Wesley Professional
     * Computing Series, Addison-Wesley, first edition, 1995.  The
     * implementation here is largely due to Douglas C. Schmidt,
     * "Applying Design Patterns to Simplify Signal Handling," <i>C++
     * Report</i>, April 1998.  There may, however, still be issues
     * with multi-threaded applications, see Bradford Nichols, Dick
     * Buttlar, and Jacqueline Proulx Farrell, <i>Pthreads
     * Programming,</i> O'Reilly & Associates, first edition, 1996,
     * for more information.
     *
     * @todo This class hierarchy, as well as some other general
     * classes and structs like MesoRD::Annotation and
     * MesoRD::Visual::Quaternion, could possibly be moved to some
     * sort of <code>MesoRD::misc</code> namespace.
     */
  class SignalHandler
  {
  private:
    static std::map<int, EventHandler*> handlers;
    static SignalHandler*               ptr;


      /**
       * @brief Singleton constructor for the SignalHandler class.
       *
       * Making the constructor, copy constructor and assignment
       * operator private, ensures this class is a singleton.
       * Instead, the <code>ptr</code> data member provides a global
       * point of access for the single instance of this class.  The
       * <code>ptr</code> pointer must not be touched by anything but
       * the assignment operator -- and then, only as a return value,
       * which is guaranteed to keep the local copy intact.
       */
    SignalHandler()
      {}


      /**
       * @brief Singleton copy constructor for the SignalHandler
       * class.
       *
       * @copydoc SignalHandler::SignalHandler()
       *
       * The handler to copy from is completely ignored, since there
       * can only be one instance of the SignalHandler class.
       */
    SignalHandler(const SignalHandler&)
      {}


      /**
       * @brief Singleton assignment operator for the SignalHandler
       * class.
       *
       * @copydoc SignalHandler::SignalHandler()
       *
       * The handler to assign to is completely ignored, since there
       * can only be one instance of the SignalHandler class.
       *
       * @return A pointer to the one and only SignalHandler instance.
       */
    SignalHandler&
    operator=(const SignalHandler&)
      {
        return *ptr;
      }


      /**
       * @brief The dispatcher for interfacing signal delivery from
       * the operating system.
       *
       * This function is called back by the operating system's signal
       * dispatching mechanism whenever some registered signal is
       * raised.  Here, the signal is received and dispatched to the
       * appropriate handler's handleSignal() hook function.  Since
       * any entry point installed into the system's
       * <code>sigaction</code> table must be an external C function
       * or a static C++ member function, this function is static.
       *
       * This adaptor transforms the C-style signalling API into the
       * EventHandler, more convenient to C++ people.  Another benefit
       * is that handleSignal() should never see any signals it cannot
       * deal with.
       *
       * @param sig The raised signal.
       */
    static void
    dispatch(const int sig)
      {
        std::map<int, EventHandler*>::const_iterator i = handlers.find(sig);
        if (i != handlers.end()) {
          i -> second -> handleSignal(sig);
        }
      }


  public:
      /**
       * @brief Destructor for the SignalHandler class.
       *
       * The destructor just cleans up its internal instance
       * pointer, in case anything should go wrong.
       */
    ~SignalHandler()
      {
        ptr = 0;
      }


      /**
       * @brief Obtain a pointer to the SignalHandler singleton.
       *
       * Returns a pointer to the single SignalHandler instance.  If
       * the instance has not been created, it be will after a call to
       * this function.  Invoking instance() is perhaps equivalent to
       * <code>this</code> in a non-static context.  The name of this
       * kind of function is traditionally <code>instance</code>.
       *
       * @return A pointer to the one and only signal handler.
       */
    static SignalHandler*
    instance()
      {
        if (ptr)
          return ptr;
        return ptr = new SignalHandler();
      }


      /**
       * @brief Registers a handler for a signal.
       *
       * Installs an event handler @p eh for signal @p sig using
       * <code>sigaction()</code> from the underlying signalling API.
       * Note that the handler is <em>not</em> called directly by the
       * operating system.  It is the duty of the dispatcher,
       * implemented in dispatch(), to forward any @p sig signals to
       * @p eh.  Here, @p eh is merely entered into the dispatcher's
       * <code>map</code> of handlers, and the dispatcher is set up to
       * accept @p sig signals.  The return value from
       * <code>sigaction()</code> is ignored, since there is little to
       * do if something should go wrong.
       *
       * A pointer to any existing @p sig handler already installed in
       * the dispatcher's handler map is returned.  If there is no
       * previously registered handler, the function returns
       * <code>null</code> instead.  This will allow for <i>signal
       * chaining</i>, meaning temporary replacement of a signal
       * handler.
       *
       * @param sig The signal handled by @p eh.
       * @param eh  The handler for @p sig.
       * @return    The previously installed handler for @p sig, if
       *            any such handler exists.  <code>null</code>
       *            otherwise.
       */
    EventHandler*
    registerHandler(const int sig, EventHandler* eh)
      {
        struct sigaction act;
        std::map<int, EventHandler*>::const_iterator i = handlers.find(sig);

        EventHandler* eh_old = 0;
        if (i != handlers.end())
          eh_old = i -> second;
        handlers[sig] = eh;

        act.sa_handler = SignalHandler::dispatch;
        sigemptyset(&act.sa_mask);
        act.sa_flags = 0;
        sigaction(sig, &act, 0);
        return eh_old;
      }


      /**
       * @brief Remove the event handler for @p sig.
       *
       * Remove the EventHandler for @p sig by setting the
       * corresponding slot in the map of signal handlers to
       * <code>null</code>.  If @p eh is supplied as well, the pointer
       * in the slot will be set to @p eh.  Thus, this function undoes
       * what was previously done by registerHandler().
       *
       * If this was the last handler to be removed, the function will
       * remove the SignalHandler singleton entirely in order not to
       * leak memory.  Removal is accomplished by calling the
       * SignalHandler destructor.  This should be safe safe; if one
       * would want a new SignalHandler at a later stage, it is be
       * created on-demand by a call to instance().
       *
       * @param sig The signal for which the handler is to be removed.
       * @param eh  Pointer to the new handler for @p sig.  Defaults
       *            to <code>null</code>.
       */
    void
    removeHandler(const int sig, EventHandler* eh = 0)
      {
        std::map<int, EventHandler*>::iterator i = handlers.find(sig);
        if (i != handlers.end()) {
          if (eh)
            i -> second = eh;
          else
            handlers.erase(i);
        }
        if (handlers.size() <= 0)
          delete this;
        return;
      }
  };



    /**
     * @brief A concrete, but general EventHandler implementation.
     *
     * This is probably the most general sort of signal handler
     * imaginable.  It just sets a flag when receiving a signal.  The
     * status of the flag can be polled by signalled(), and reset by
     * reset().  As noted before, variables manipulated in signal
     * handlers should be of type <code>sig_atomic_t</code>.
     *
     * @todo This class name is probably a bit misleading.
     */
  class FlagHandler : public EventHandler
  {
  private:
    sig_atomic_t flag;

  public:

      /**
       * @brief Constructor for the FlagHandler class.
       *
       * The default constructor just ensures that the flag is not
       * set.
       */
    FlagHandler()
      : flag(0)
      {}


      /**
       * @brief Hook function for the FlagHandler class.
       *
       * The hook function is called by SignalHandler::dispatch() when
       * a signal is received.  Because SignalHandler::dispatch()
       * should ensure that this function only sees those signals it
       * is registered to handle, nothing is checked here.  In fact,
       * it just sets the flag, and it always returns zero.
       *
       * @note The received signal does not matter at all.
       *
       * @return Always returns zero.
       */
    virtual int
    handleSignal(const int)
      {
        flag = 1;
        return 0;
      }


      /**
       * @brief Reset the status flag for an instance of the
       * FlagHandler class.
       *
       * This function will just reset the flag of the FlagHandler
       * class to its initial, unset, state.
       */
    void
    reset()
      {
        flag = 0;
      }


      /**
       * @brief Accessor function for the FlagHandler class.
       *
       * Polls the status of the flag.  The flag is set as soon as a
       * signal is received.  The function returns <code>true</code>
       * if the flag is set, and <code>false</code> otherwise.
       *
       * @return <code>True</code> if the flag is set, and
       *         <code>false</code> otherwise.
       */
    bool
    signalled()
      {
        return flag != 0;
      }
  };
}
#endif /* ifndef MESORD_UI_SIGNALHANDLER_HPP */
