/* The TextGetOpt class.
 *
 * Copyright 2004-2006 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TextGetOpt.cpp 559 2008-09-10 19:46:33Z davidfange $
 */
#include "../mesord.hpp"

#ifdef WIN32
#  include "XGetOpt.hpp"
#else
#  ifdef HAVE_GETOPT_H
#    include <getopt.h>
#  else
#    include <unistd.h>
#  endif
#endif


#include <iostream>
#include <iomanip>

#include "TextGetOpt.hpp"


  /* Command line parsing using getopt(3) and getopt_long(3).  Should
   * be POSIX compliant.
   */
MesoRD::TextGetOpt::TextGetOpt(const int argc, char* const* argv)
{
    /* Only make those options available for which support is actually
     * compiled.  Relies on preprocessor string concatenation.
     */
#define OPTSTRING_BASIC ":C:EI:LPSVc:d:ehi:lo:pq:st:"

#ifdef DETERMINISTIC
#  define OPTSTRING_DETERMINISTIC "DT:m:r:x:"
#else
#  define OPTSTRING_DETERMINISTIC ""
#endif

#ifdef HAVE_VIEWER
#  define OPTSTRING_VIEWER "Gg"
#else
#  define OPTSTRING_VIEWER ""
#endif

#define OPTSTRING OPTSTRING_BASIC         \
                  OPTSTRING_DETERMINISTIC \
                  OPTSTRING_VIEWER

    /* Only use getopt(3) if getopt_long(3) is not available;
     * getopt(3) is available through POSIX.2.  <code>optstring</code>
     * begins with a colon in order to enable tracking of missing
     * option arguments.
     */
#ifdef HAVE_GETOPT_LONG
  static struct option options[] = {
    {"checkpoint-iteration",    required_argument, NULL, 'C'},
    {"output-sparse",           no_argument,       NULL, 'E'},
    {"gui",                     no_argument,       NULL, 'G'},
    {"max-iteration-scale",     required_argument, NULL, 'I'},
    {"fullstate-checkpoint",    no_argument,       NULL, 'L'},
    {"progress-bars",           no_argument,       NULL, 'P'},
    {"simulation",              no_argument,       NULL, 'S'},
    {"version",                 no_argument,       NULL, 'V'},
    {"checkpoint-time",         required_argument, NULL, 'c'},
    {"species-dimension",       required_argument, NULL, 'd'},
    {"no-output-sparse",        no_argument,       NULL, 'e'},
    {"no-gui",                  no_argument,       NULL, 'g'},
    {"help",                    no_argument,       NULL, 'h'},
    {"max-iteration",           required_argument, NULL, 'i'},
    {"no-fullstate-checkpoint", no_argument,       NULL, 'l'},
    {"old-state",               required_argument, NULL, 'o'},
    {"no-progress-bars",        no_argument,       NULL, 'p'},
    {"cubic",                   required_argument, NULL, 'q'},
    {"no-simulation",           no_argument,       NULL, 's'},
    {"max-time",                required_argument, NULL, 't'},
#ifdef DETERMINISTIC
    {"deterministic",           no_argument,       NULL, 'D'},
    {"timestep",                required_argument, NULL, 'T'},
    {"integration-method",      required_argument, NULL, 'm'},
    {"it-tolerance",            required_argument, NULL, 'r'},
    {"it-max-iterations",       required_argument, NULL, 'x'},
#endif
    {NULL,                      0,                 NULL, 0}
  };
#endif


    /* This code does its own error reporting, so <code>opterr</code>
     * is set to zero to disable getopt(3) error messages.  Loop
     * through all the arguments in <code>argv</code>.  Because
     * filenames may be interspersed with options, the getopt(3) loop,
     * which exhausts a continuous sequence of options, sits inside
     * the outer argument loop.  The index of the next argument,
     * <code>optind</code>, is initialised to <code>1</code>.
     */
  //opterr = 0;
  while (optind < argc) {
    int option;

    while ((option
#ifdef WIN32
		= getopt(argc, 
                  const_cast<TCHAR**> (static_cast< TCHAR* const*>(argv)), 
                  OPTSTRING)
#else
#  ifdef HAVE_GETOPT_LONG
             = getopt_long(argc, argv, OPTSTRING, options, NULL)
#  else
             = getopt(argc, argv, OPTSTRING)
#  endif
#endif
            ) != -1) {
      char* str = 0;

      switch (option) {
        case 'C':
          setOptionCheckpointIterations(optarg);
          break;

        case 'E':
          setOptionSparse(true);
          break;

        case 'I':
          setOptionMaxIterationsScale(optarg);
          break;

        case 'L':
          setOptionFullStateCheckPoint(true);
          break;

        case 'P':
          setOptionProgressBars(true);
          break;

        case 'S':
          setOptionSimulate(true);
          break;

        case 'V':
          setOptionVersion(true);
          break;

        case 'c':
          setOptionCheckpointTime(optarg);
          break;

        case 'd':
          setOptionSpeciesDimension(optarg);
          break;

        case 'e':
          setOptionSparse(false);
          break;

        case 'h':
          setOptionHelp(true);
          break;

        case 'i':
          setOptionMaxIterations(optarg);
          break;

        case 'l':
          setOptionFullStateCheckPoint(false);
          break;

        case 'p':
          setOptionProgressBars(false);
          break;

        case 'q':
            /* If the unit is concatenated with the size, separate the
             * two into <code>str</code> and <code>str_size</code>
             * stored in continuous memory.  If they are <em>not</em>
             * concatenated, assume the next argument in the
             * <code>argv</code> is the unit.  Therefore, increase
             * <code>optind</code> to skip over a single option
             * character.
             *
             * FIXME!! Should handle different geometries here!
             */
          if ((str = static_cast<char*>(
                 malloc(sizeof(char) * 2 * (strlen(optarg) + 1))))) {
            
            double    size     = 0;
            const int items    = sscanf(optarg, "%lf%s", &size, str);
            
              /* If there is only one item returned from sscanf, assume
               * that this is the size, and ship the entire optarg to
               * the setOptionSize function.
               */
            if (items == 1 && optind < argc && argv[optind][0] != '-') {
              setOptionSize(optarg);
              setOptionUnit(argv[optind++]);
              }
            else if (items == 2) {
              char*     str_size = &(str[strlen(str) + 1]);
              strncpy(str_size, optarg, strlen(optarg) - strlen(str));
              
              str[strlen(optarg) + 2] = '\0';
              
              setOptionSize(str_size);
              setOptionUnit(str);
            }
            free(str);
          }
          break;

        case 's':
          setOptionSimulate(false);
          break;

        case 't':
          setOptionMaxTime(optarg);
          break;

        case 'o':
          setOptionReadOldState(true);
          setOptionReadOldStateTime(optarg);
          break;

#ifdef HAVE_VIEWER
        case 'G':
          setOptionViewer(true);
          break;

        case 'g':
          setOptionViewer(false);
          break;
#endif

#ifdef DETERMINISTIC
        case 'D':
          setOptionDeterministic(true);
          break;

        case 'T':
          setOptionTimeStep(optarg);
          break;

        case 'm':
          setOptionIntegrationMethod(optarg);
          break;

        case 'r':
          setOptionItMetConvergenceTolerance(optarg);
          break;

        case 'x':
          setOptionItMetMaxIterations(optarg);
          break;
#endif

        case ':':
            //FIXME!! This does not seem to be working properly.  
            /* The required argument of an option is missing.  For
             * error reporting, one can use the last <em>known</em>
             * option character, which is stored in
             * <code>optopt</code>.
	     *
	     * Use the standard C library <code>fprintf(3)</code>
	     * function instead of stream output in order to break
	     * typesafety checks when interpreting and
	     * <code>int</code> as a <code>char</code>.
             */
	  //fprintf(stderr, "Option -%c", optopt);
#ifdef HAVE_GETOPT_LONG
          std::cerr << " (--";
          for (size_t i = 0; options[i].name != NULL; i++) {
            if (options[i].val == optopt) {
              std::cerr << options[i].name;
              break;
            }
          }
          std::cerr << ")";
#endif
          std::cerr << " requires an argument!" << std::endl;
          setOptionUsage(true);
          break;

        case '?':
            /* Encountered a non-recognised option.  Because
             * <code>optind</code> always points to the next
             * <code>argv</code> argument, <code>argv[optind -
             * 1]</code> gives the offending option.
             */
          std::cerr << "Unrecognised option " << argv[optind - 1] << std::endl;
          setOptionUsage(true);
          break;

        default:
          setOptionUsage(true);
          break;
      }
    }


      /* The filename should now be the next <code>argv</code>
       * argument.  The index of this argument is <code>optind</code>,
       * as set up by the getopt(3) call.
       */
    if (optind < argc)
      setOptionFilename(argv[optind++]);
  }

  try {
    GetOpt::fromSBML();
  } catch (MesoRD::System::ModelError& e) {
    throw MesoRD::System::ModelError(e.what());
  }

#ifdef HAVE_VIEWER
  GetOpt::setDefaultColours();
#endif
}


  /* Display the help message.  This is just a summary of all the
   * options , with a little textual help for each one.
   */
void
MesoRD::TextGetOpt::displayHelp() const
{
  typedef struct description
  {
    char* text;
    char* name;
    int   val;
    char* args;
  } description;

  const description descriptions[] = {
    {"display this help",                     "help",
     'h',                                     ""},
    {"use cubes with side size",              "cubic",
     'q',                                     "size"},
    {"simulate its iterations",               "max-iteration",
     'i',                                     "its"},
    {"scale of iteration count",              "max-iteration-scale",
     'I',                                     "scale"},
    {"simulate secs seconds",                 "max-time",
     't',                                     "secs"},
    {"checkpoint every its iterations",       "checkpoint-iteration",
     'C',                                     "its"},
    {"checkpoint every secs seconds",         "checkpoint-time",
     'c',                                     "secs"},
    {"launch simulation",                     "simulation",
     'S',                                     ""},
    {"do not start simulation",               "no-simulation",
     's',                                     ""},
    {"plot molecs molecules per dimension",   "species-dimension",
     'd',                                     "molecs"},
    {"launch progress bars",                  "progress-bars",
     'P',                                     ""},
    {"do not start progress bars",            "no-progress-bars",
     'p',                                     ""},
    {"turn on sparse output",                 "output-sparse",
     'E',                                     ""},
    {"turn off sparse output",                "no-output-sparse",
     'e',                                     ""},
    {"turn off fullstate checkpointing",      "no-fullstate-checkpoint",
     'l',                                     ""},
    {"turn on fullstate checkpoint (not with -D)",          "fullstate-checkpoint",
     'L',                                     ""},
    {"parse old system state",                "old-state",
     'o',                                     "line#"},
    {"display version information",           "version",
     'V',                                     ""},
#ifdef HAVE_VIEWER
    {"launch viewer",                         "gui",
     'G',                                     ""},
    {"do not launch viewer",                  "no-gui",
     'g',                                     ""},
#endif
#ifdef DETERMINISTIC
    {"run a deterministic simulation",        "deterministic",
     'D',                                     ""},
    {"timestep deterministic simulation",     "timestep",
     'T',                                     "secs"},
    {"use a specific integration-method",     "integration-method",
     'm',                                     "method"},
    {"tolerance in iterative method",         "it-tolerance",
     'r',                                     "tol"},
    {"max number of its in iterative method", "it-max-iterations",
     'x',                                     "its"},
#endif
    {NULL,                                    NULL,
     0,                                       NULL}
  };

  int len_args = 0;
  int len_name = 0;
  int len_text = 0;

    /* Find the maximal length of the <code>args</code>,
     * <code>name</code> and <code>text</code> members.
     */
  for (size_t i = 0; descriptions[i].text != NULL; i++) {
    len_args = std::max(len_args,
			static_cast<int>(strlen(descriptions[i].args)));
    len_name = std::max(len_name,
			static_cast<int>(strlen(descriptions[i].name)));
    len_text = std::max(len_text,
			static_cast<int>(strlen(descriptions[i].text)));
  }

    /* Print the summary.  Use the previously determined maximal
     * lengths for nicer formatting.
     *
     * Use the standard C library <code>fprintf(3)</code> function
     * instead of stream output in order to break typesafety checks
     * when interpreting and <code>int</code> as a <code>char</code>.
     */
  std::cout << std::left
            << "  Available command line options"   << std::endl
            << "----------------------------------" << std::endl;
  for (size_t i = 0; descriptions[i].text != NULL; i++) {
    fprintf(stdout, " -%c", descriptions[i].val);
    std::cout << " "   << std::setw(len_args)
              << descriptions[i].args
#ifdef HAVE_GETOPT_LONG
              << "  "  << std::setw(len_name)
              << descriptions[i].name
#endif
              << "  "  << std::setw(len_text)
              << descriptions[i].text
              << std::endl;
  }
    //FIXME!! Use address defined in configure.  
  std::cout << " "   << std::setw(len_args)
            << "Report bugs to david.fange@icm.uu.se" 
            << std::endl;
}
