/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* Copyright (c) 2004-2007, Johan Elf, David Fange, Johan Hattne
 *
 * Permission to use, copy, modify, and/or distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear
 * in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: TextProgressBar.hpp 533 2007-10-24 09:55:48Z hattne $
 */
#ifndef MESORD_UI_TEXTPROGRESSBAR_HPP
#define MESORD_UI_TEXTPROGRESSBAR_HPP 1

#include <string>

#include "ProgressBar.hpp"
#include "SignalHandler.hpp"


namespace MesoRD
{

    /**
     * @brief Concrete implementation for textual progress bars.
     *
     * This class implements the ProgressBar interface in the case
     * where display is done using the standard out file descriptor.
     * A textual progress bar has a width that presumably depends on
     * the width of the terminal where the progress bar is displayed.
     * Another parameter for textual progress bars is the update
     * frequency.  Since textual progress bars are in fact animated,
     * the configurable update frequency is the time delay between
     * each update.
     *
     * @todo This entire class hierarchy, as well as most other things
     * in this directory, should probably be moved to a new
     * <code>MesoRD::UI</code> namespace.
     */
  class TextProgressBar : public ProgressBar
  {
  private:
    /**
     * @brief The state of the progress bar spinner.
     *
     * This enum encodes the state of the textual progress bar's
     * spinner.
     *
     * <ul>
     *
     *   <li><code>horizontal</code> corresponds to
     *   <code>-</code></li>
     *
     *   <li><code>declined</code> corresponds to <code>\</code></li>
     *
     *   <li><code>vertical</code> corresponds to <code>|</code></li>
     *
     *   <li><code>inclined</code> corresponds to <code>/</code></li>
     *
     * </ul>
     *
     * To spin the spinner, the state should cycle through these
     * states in the above order.
     */
    typedef enum {
      /**
       * The horizontal direction corresponds to the <code>-</code>
       * character.
       */
      horizontal = '-',

      /**
       * The declined direction corresponds to the <code>\</code>
       * character.
       */
      declined = '\\',

      /**
       * The vertical direction corresponds to the <code>|</code>
       * character.
       */
      vertical = '|',

      /**
       * The inclined direction corresponds to the <code>/</code>
       * character.
       */
      inclined = '/'
    } SpinnerState;

    /**
     * @brief Updates the state of the "spinning wheel".
     *
     * This function takes a reference to the current state of the
     * spinner and updates it.  It returns a character corresponding
     * to the textual representation of the current state of the
     * spinner.
     *
     * @param[in]  state The current state of the wheel.
     * @param[out] state The updated state of the wheel.
     * @return           Textual representation of the updated state
     *                   of the spinner.
     */
    static inline char
    updateSpinner(SpinnerState& state)
      {
        switch (state) {
          case horizontal:
            state = declined;
            return '-';

          case declined:
            state = vertical;
            return '\\';

          case vertical:
            state = inclined;
            return '|';

          case inclined:
            state = horizontal;
            return '/';
        }

        return ' ';
      }


  private:

      /* @bug if isAlive() is renamed to alive(), then this should
       * probably become isAlive.
       */
    bool alive;

    FlagHandler* handler;
    const unsigned int period;
    SpinnerState spinner;

    /**
     * @brief The width of the progress bar.
     *
     * This is an <code>int</code>, because that is what C++'s
     * std::setw() manipulator expects.
     */
    const int width;


  public:

      /**
       * @brief Default constructor for the TextProgressBar class.
       *
       * The constructor sets up a TextProgressBar instance in the
       * <code>alive</code> state.  A few defaults for the progress
       * bar can be overridden in the constructor:
       *
       * <ul>
       *
       *   <li> The period, @p p, is the interval between updates of
       *   the progress bar.  Even though there may not be a change in
       *   progress, the little indicator animation will keep on
       *   spinning, indicating that at least the progress bar is
       *   still doing its work.</li>
       *
       *   <li>The total width of the progress bar, @p w.  The
       *   standard width of a terminal window seems to be 80
       *   characters.  Anything below that is probably acceptable,
       *   even though long titles may not display correctly with too
       *   short widths.</li>
       *
       * </ul>
       *
       * The constructor registers a handler for <code>SIGINT</code>.
       * The run() function will return upon reception of any caught
       * signals.
       *
       * @todo Perhaps it is a good idea to register the same handler
       * for <code>SIGQUIT</code> as well?
       *
       * @param p Sleep time, or <i>period</i>, for the progress bar,
       *          in seconds.
       * @param w Total width of the progress bar, in characters.
       */
    TextProgressBar(const unsigned int p = 1, const int w = 75)
      : ProgressBar(),
        alive(true), handler(0), period(p), spinner(horizontal), width(w)
      {
        handler = new FlagHandler();
        SignalHandler::instance() -> registerHandler(SIGINT, handler);
      }


      /**
       * @brief Destructor for the TextProgressBar class.
       *
       * Removes the signal handler, if present.
       */
    ~TextProgressBar()
      {
        if (handler)
          delete handler;
      }


      /**
       * @brief Draws the textual progress bar on standard out.
       *
       * This is the only non-trivial function of this class, and is
       * called repeatedly from run().  It will draw the progress bar,
       * modelled after the OpenSSH scp textual progress bar
       */
    void draw();


      /**
       * @brief Poll the status of the progress bar.
       *
       * The status may change because the process, the progress of
       * which is monitored, completes, or because a signal is
       * received.  Signals may be received either through the
       * operating system's signalling facilities or by an
       * asynchronous call to stop().  In either case, once such an
       * event has been registered, this function will return
       * <code>false</code>.  As long as the progress bar is running,
       * it instead returns <code>true</code>.
       *
       * @return <code>True</code> if the progress bar is still alive,
       *         <code>false</code> otherwise
       */
    inline bool
    isAlive()
      {
        return alive;
      }


      /**
       * @brief Animate the textual progress bar.
       *
       * This defines some sort of <i>main event loop</i> of the
       * progress bar.  The progress bar is animated in each iteration
       * step.  Animation continues until the process has gone to
       * completion, a signal from the operating system is caught or
       * the class internal <code>alive</code> flag is no longer
       * <code>true</code>.  The <code>alive</code> flag may be
       * cleared asynchronously by calling the stop() function.
       *
       * The run() function exits once it no longer animates the
       * progress bar.  The function also outputs a trailing newline,
       * so as not mess up standard out.
       *
       * @todo Maybe the run function should return once fraction
       * falls outside the set range?  Check with GTK+ code in order
       * to decide.  This will affect code in the Thread class.
       */
    inline void
    run()
      {
        while (alive
          && fraction >= 0 && fraction <= 1
          && handler       && !(handler -> signalled())) {
          draw();
          sleep(period);
        }
        alive = false;
        std::cout << std::endl;
      }


      /**
       * @brief Stop the progress bar.
       *
       * This function can be used to stop the progress bar
       * asynchronously.  Calling this function will set the class
       * internal <code>alive</code> flag to <code>false</code>,
       * meaning that progress monitoring should stop.  This will also
       * mean that any calls to alive() will subsequently return
       * <code>false</code>.  Note, however, that it will <em>not</em>
       * call the destructor of the progress bar.
       */
    inline void
    stop()
      {
        alive = false;
      }
  };
}
#endif /* ifndef MESORD_UI_TEXTPROGRESSBAR_HPP */
