/* The TestCubicSubVolume class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestCubicSubVolume.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestCubicSubVolume.hpp"
#include "../src/sparselib/comprow_double.hpp"
#include "../src/mv/mvvd.hpp"
#include "../src/random/Random.hpp"

#include "../src/CubicSubVolume.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestCubicSubVolume );


void 
TestCubicSubVolume::setUp()
{
  
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  MesoRD::Species::setUseDoubleInAmount(false);

  std::string test_filename 
    = "./sbml_source/TestSubVolume.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
  
  ut = MesoRD::UnitTable(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  pt.addParameter(p1);
  MesoRD::Parameter p2(m -> getParameter(1), &ut);
  pt.addParameter(p2);


   
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  MesoRD::Species::setUseDoubleInAmount(true);

  MesoRD::Random::Random* r = new MesoRD::Random::Random();
  
  subvolumes = ct.fromSBML(m, &pt, &ut, 1, "um", r);
  
  subvolumes_det = ct_det.fromSBML(m, &pt, &ut, 1, "um", r);

  
}


void 
TestCubicSubVolume::tearDown()
{}

void
TestCubicSubVolume::testGetSet()
{

  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::Compartment c(&ct, NULL, &st, &pt, &ut, m, 0);
  
  MesoRD::CubicSubVolume* s = 
    static_cast<MesoRD::CubicSubVolume*>(subvolumes[0]);

  s -> setNeighbour(s -> down, NULL);
  
  CPPUNIT_ASSERT(
    s -> getNeighbour(s -> down) == NULL);
  
  MesoRD::CubicSubVolume* new_neighbour = 
    static_cast<MesoRD::CubicSubVolume*>(subvolumes[2]);
  
  s -> setNeighbour(s -> down, new_neighbour);
  
  CPPUNIT_ASSERT(s -> getNeighbour(
                   s -> down) == new_neighbour);

}


void
TestCubicSubVolume::testEventDiffusion()
{
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::Compartment c(&ct, NULL, &st, &pt, &ut, m, 0);
  
  MesoRD::SubVolume* s = subvolumes[1];

  s -> setSpecies(0, (unsigned int)5);
  s -> setSpecies(1, (unsigned int)0);
  s -> setSpecies(2, (unsigned int)0);
  
  s -> updateTime(0.5);
  
  MesoRD::SubVolume* diff_target = s -> eventDiffusion(0.3);
  
  CPPUNIT_ASSERT(s -> getSpecies(0) == 4);
}

void
TestCubicSubVolume::testEventNeighbourReaction()
{
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::Compartment c(&ct, NULL, &st, &pt, &ut, m, 0);
  
  MesoRD::SubVolume* s;
  
  
  for (unsigned int i = 0; i < subvolumes.size(); i++) {
    s = subvolumes[i];
    s -> setSpecies(0, (unsigned int)5);
    s -> setSpecies(1, (unsigned int)5);
    s -> setSpecies(2, (unsigned int)0);
  }
  
  s = subvolumes[1];
  
  s -> updateTime(0.5);
  s -> eventNeighbourReaction(0.24);

  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(0) == 5);
  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(1) == 5);
  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(2) == 0);
  
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(0) == 5);
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(1) == 4);
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(2) == 1);
  
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(0) == 4);
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(1) == 5);
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(2) == 0);


    for (unsigned int i = 0; i < subvolumes.size(); i++) {
    s = subvolumes[i];
    s -> setSpecies(0, (unsigned int)5);
    s -> setSpecies(1, (unsigned int)5);
    s -> setSpecies(2, (unsigned int)0);
  }
  
  s = subvolumes[1];
  
  s -> updateTime(0.5);
  s -> eventNeighbourReaction(0.3);

  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(0) == 5);
  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(1) == 5);
  CPPUNIT_ASSERT(subvolumes[0] -> getSpecies(2) == 0);
  
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(0) == 4);
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(1) == 5);
  CPPUNIT_ASSERT(subvolumes[1] -> getSpecies(2) == 1);
  
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(0) == 5);
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(1) == 4);
  CPPUNIT_ASSERT(subvolumes[2] -> getSpecies(2) == 0);


}
