/* The TestDerivative class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestDerivative.cpp 553 2008-05-09 13:28:22Z davidfange $
 */

#include "TestDerivative.hpp"
#include "../src/sparselib/comprow_double.hpp"
#include "../src/mv/mvvd.hpp"


// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestDerivative );


void 
TestDerivative::setUp()
{
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  MesoRD::Species::setUseDoubleInAmount(true);

  std::string test_filename 
    = "./sbml_source/TestDerivative.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
  
  ut = MesoRD::UnitTable(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  pt.addParameter(p1);
  
  
  
  subvolumes = ct.fromSBML(m, &pt, &ut, 1, "um", NULL);
  
  derivative = MesoRD::Derivative(subvolumes, &ct);
}


void 
TestDerivative::tearDown()
{}

void 
TestDerivative::testGetCurrentState()
{
  
  MV_Vector_double currentState = derivative.getCurrentState();
  
  CPPUNIT_ASSERT_EQUAL((int)subvolumes.size(), 3);
  
  
  CPPUNIT_ASSERT_EQUAL(currentState.size(), 6);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0),1,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),  1,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),  1,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
  std::numeric_limits<double>::epsilon());

}


void
TestDerivative::testSetCurrentState()
{
  MV_Vector_double currentState = derivative.getCurrentState();
  
  currentState(0) = 4;
  currentState(3) = 12;
  currentState(5) = 36;
  
  derivative.setCurrentState(currentState);
  
  MV_Vector_double newState = derivative.getCurrentState();
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(0),  currentState(0),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(1),  currentState(1),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(2),  currentState(2),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(3),  currentState(3),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(4),  currentState(4),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(newState(5),  currentState(5),
    std::numeric_limits<double>::epsilon());
  
}

void
TestDerivative::testGetReactionDerivative()
{
  MV_Vector_double currentState = derivative.getCurrentState();

  MV_Vector_double rxnDerivative(6);
  for(int i=0;i<6;i++) {
    rxnDerivative(i) = 0;
  }
  
  derivative.getReactionDerivative(currentState, rxnDerivative);

  
  CPPUNIT_ASSERT_DOUBLES_EQUAL((double)rxnDerivative(0),  (double)-5.0,
    10*std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rxnDerivative(1),  0,
    10*std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rxnDerivative(2),  -5,
    10*std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rxnDerivative(3),  0,
    10*std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rxnDerivative(4),  -5,
    10*std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rxnDerivative(5),  0,
    10*std::numeric_limits<double>::epsilon());
}

void
TestDerivative::testGetDiffusionMatrix()
{
  CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  CPPUNIT_ASSERT(diffusion.size(0) == 6);
  CPPUNIT_ASSERT(diffusion.size(1) == 6);

  double a_diff = 5;
  double b_diff = 1;

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,0), (-1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,2), (1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,0), (1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,2), (-2) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,4), (1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,2), (1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,4), (-1) * a_diff,
    10 * std::numeric_limits<double>::epsilon());




  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,1), (-1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,3), (1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,1), (1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,3), (-2) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,5), (1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,3), (1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,5), (-1) * b_diff,
    10 * std::numeric_limits<double>::epsilon());


  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,1),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,3),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,4),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(0,5),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,0),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,2),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,4),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(1,5),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,1),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,3),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(2,5),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,0),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,2),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(3,4),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,0),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,1),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,3),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(4,5),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,0),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,1),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,2),  0,
    10 * std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(diffusion(5,4),  0,
    10 * std::numeric_limits<double>::epsilon());
  
}
