/* The TestIntegration class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestIntegration.cpp 553 2008-05-09 13:28:22Z davidfange $
 */

#include "TestIntegration.hpp"
#include "../src/sparselib/comprow_double.hpp"
#include "../src/mv/mvvd.hpp"
#include "../src/integration/Euler.hpp"
#include "../src/integration/EulerBackward.hpp"
#include "../src/integration/Trapez.hpp"
#include "../src/integration/RungeKutta4.hpp"
#include "../src/integration/BDF2.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestIntegration );


void 
TestIntegration::setUp()
{
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  MesoRD::Species::setUseDoubleInAmount(true);

  std::string test_filename 
    = "./sbml_source/TestDerivative.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
  
  ut = MesoRD::UnitTable(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  pt.addParameter(p1);
  
  subvolumes = ct.fromSBML(m, &pt, &ut, 1, "um", NULL);
  
  derivative = MesoRD::Derivative(subvolumes, &ct);
}


void 
TestIntegration::tearDown()
{}


void
TestIntegration::testGetSet()
{

  MesoRD::Integration::setIntegrationMethod(
    MesoRD::Integration::EulerBackward);
  CPPUNIT_ASSERT(MesoRD::Integration::getIntegrationMethod() ==
    MesoRD::Integration::EulerBackward);
  
  MesoRD::Integration::setIntegrationMethod("RungeKutta4");
  CPPUNIT_ASSERT(MesoRD::Integration::getIntegrationMethod() ==
    MesoRD::Integration::RungeKutta4);

}

void
TestIntegration::testEuler() 
{
  CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  MesoRD::Euler e(&derivative, &diffusion, 0.1, NULL);
  
  MV_Vector_double currentState = derivative.getCurrentState();

  e.updateState(currentState);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
    std::numeric_limits<double>::epsilon());

}

void
TestIntegration::testEulerBackward()
{

  CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  MV_Vector_double beginState = derivative.getCurrentState();
  
  MesoRD::EulerBackward e(&derivative, &diffusion, 0.1, NULL, 
    1e-6, 100, beginState);
  
  MV_Vector_double currentState = derivative.getCurrentState();

  e.updateState(currentState);


  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
    std::numeric_limits<double>::epsilon());  

}

void
TestIntegration::testTrapezoid()
{
  CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  MV_Vector_double beginState = derivative.getCurrentState();
  
  MesoRD::Trapez e(&derivative, &diffusion, 0.1, NULL, 
    1e-6, 100, beginState);
  
  MV_Vector_double currentState = derivative.getCurrentState();

  e.updateState(currentState);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),  0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
    std::numeric_limits<double>::epsilon());  

}

void
TestIntegration::testRungeKutta4()
{
 CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  MV_Vector_double beginState = derivative.getCurrentState();
  
  MesoRD::RungeKutta4 e(&derivative, &diffusion, 0.1, NULL);
  
  MV_Vector_double currentState = derivative.getCurrentState();

  e.updateState(currentState);
  
    //!!FIXME
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0), (double)2/3,
    //std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),(double)2/3, 
    //std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
    //CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),(double)2/3,
    //std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
    std::numeric_limits<double>::epsilon());  

} 

void
TestIntegration::testBDF2()
{
  CompRow_Mat_double diffusion = derivative.getDiffusionMatrix();
  
  MV_Vector_double beginState = derivative.getCurrentState();
  
  MesoRD::BDF2 e(&derivative, &diffusion, 0.1, NULL,
    1e-6, 100, beginState);
  
  MV_Vector_double currentState = derivative.getCurrentState();

  e.updateState(currentState);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(0),(double)2/3,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(1),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(2),(double)2/3,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(3),  2,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(4),(double)2/3,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(currentState(5),  2,
    std::numeric_limits<double>::epsilon());  

}
