/* The TestParameter class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestParameterTable.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestParameterTable.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestParameterTable );


void 
TestParameterTable::setUp()
{
  
  std::string test_filename = "./sbml_source/TestParameter.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
}


void 
TestParameterTable::tearDown()
{}


void 
TestParameterTable::testConstructor()
{
  MesoRD::ParameterTable pt;
}

void 
TestParameterTable::testAddFunctions()
{
  
  MesoRD::ParameterTable pt;
  

  pt.addParameter("test_parameter_id", "test_parameter_name",
    230, MesoRD::Unit::mole());

  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id") == MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id") == 230);
  

  MesoRD::Parameter p("test_parameter_id1", "test_parameter_name1",
    340, MesoRD::Unit::centimetreSquaredPerSecond(), false);
  
  pt.addParameter(p);
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id") == MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id") == 230);
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id1") == 
    MesoRD::Unit::centimetreSquaredPerSecond());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id1") == 340);

  
  pt.addReplaceParameter("test_parameter_id2", "test_parameter_name2", 500,
    MesoRD::Unit::perSecond());
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id") == MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id") == 230);
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id1") == 
    MesoRD::Unit::centimetreSquaredPerSecond());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id1") == 340);

  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id2") == 
    MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id2") == 500);

  pt.addReplaceParameter("test_parameter_id2", "test_parameter_name2", 400,
    MesoRD::Unit::mole());
  
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id") == MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id") == 230);
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id1") == 
    MesoRD::Unit::centimetreSquaredPerSecond());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id1") == 340);

  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id2") == 
    MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id2") == 400);


  MesoRD::Parameter p2("test_parameter_id1", "test_parameter_name1",
    350, MesoRD::Unit::perMolarPerSecond(), false);
  
  pt.addReplaceParameter(p2);

  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id") == MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id") == 230);
  
  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id1") == 
    MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id1") == 350);

  CPPUNIT_ASSERT(pt.getUnit("test_parameter_id2") == 
    MesoRD::Unit::mole());
  CPPUNIT_ASSERT(pt.getValue("test_parameter_id2") == 400);

}

void
TestParameterTable::testIsFunctions()
{

  MesoRD::ParameterTable pt;
  
  pt.addParameter("test_parameter_1", "", 0, MesoRD::Unit::dimensionless(),
    true);
  pt.addParameter("test_parameter_2", "", 0, MesoRD::Unit::mole(),
    false);
  

  CPPUNIT_ASSERT(pt.isConstant("test_parameter_1") == true);
  CPPUNIT_ASSERT(pt.isConstant("test_parameter_2") == false);
  
  CPPUNIT_ASSERT(pt.isDefined("test_parameter_1") == true);
  CPPUNIT_ASSERT(pt.isDefined("test_parameter_2") == true);
  CPPUNIT_ASSERT(pt.isDefined("test_parameter_3") == false);
  
  CPPUNIT_ASSERT(pt.isLocal("test_parameter_1") == true);
  
}

void
TestParameterTable::testParentFunction()
{
  MesoRD::ParameterTable pt1;
  MesoRD::ParameterTable pt2;
  

  pt1.addParameter("test_parameter_1", "", 23, MesoRD::Unit::mole(), false);
  pt2.addParameter("test_parameter_2", "", 34, 
    MesoRD::Unit::perSecond(), true);
  
  pt1.setParent(&pt2);

  CPPUNIT_ASSERT(pt2.isDefined("test_parameter_2"));
  CPPUNIT_ASSERT(pt1.isDefined("test_parameter_2"));
  
  CPPUNIT_ASSERT(pt2.getValue("test_parameter_2") == 34);
  CPPUNIT_ASSERT(pt2.getUnit("test_parameter_2") == MesoRD::Unit::perSecond());

  CPPUNIT_ASSERT(pt1.getValue("test_parameter_2") == 34);
  CPPUNIT_ASSERT(pt1.getUnit("test_parameter_2") == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT(pt1.isLocal("test_parameter_2") == false);
  CPPUNIT_ASSERT(pt1.isLocal("test_parameter_1") == true);
  
}

void
TestParameterTable::testGetSet()
{

  MesoRD::ParameterTable pt;
  pt.addParameter("test_parameter_1", "", 23, MesoRD::Unit::mole(), false);
  
  pt.setValue("test_parameter_1", 40);
  
  CPPUNIT_ASSERT(pt.getValue("test_parameter_1") == 40);
}
