/* The TestReactionNode class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestReactionNode.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestReactionNode.hpp"
#include "../src/SpeciesTable.hpp"
#include "../src/UnitTable.hpp"
#include "../src/Parameter.hpp"
#include "../src/ParameterTable.hpp"
#include "../src/definition.hpp"
#include <valarray>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestReactionNode );


void 
TestReactionNode::setUp()
{
  std::string test_filename = "./sbml_source/TestReactionNode.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();

  std::map<std::string, double> diffConst;
  MesoRD::Species s(diffConst, "A1", "A1","comp", ((unsigned int)14), false);
  st.addSpecies(s);

  MesoRD::UnitTable ut(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  MesoRD::Parameter p2(m -> getParameter(1), &ut);
  MesoRD::Parameter p3(m -> getParameter(2), &ut);
  MesoRD::Parameter p4(m -> getParameter(3), &ut);
  MesoRD::Parameter p5("volume_subVolume", "volume_subVolume", 
    1e-8, MesoRD::Unit::litre(), 
    true);
  MesoRD::Parameter p6("p6", "p6", 4, MesoRD::Unit::litre(), false);
  MesoRD::Parameter p7("p7", "p7", (1e-8)*MesoRD::Constant::Avogadro, 
    (MesoRD::Unit::litre()/MesoRD::Unit::mole()), true);
  
  pt.addParameter(p1);
  pt.addParameter(p2);
  pt.addParameter(p3);
  pt.addParameter(p4);
  pt.addParameter(p5);
  pt.addParameter(p6);
  pt.addParameter(p7);  
  
}


void 
TestReactionNode::tearDown()
{
}

void
TestReactionNode::testAbs()
{
  ::Reaction* r = m -> getReaction(0);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(),
      &pt, &st);
  
  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeAbs));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 5.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 5.5,
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testCeil()
{
  ::Reaction* r = m -> getReaction(1);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeCeil));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -5, 
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), -5, 
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testExp()
{
  ::Reaction* r = m -> getReaction(2);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeExp));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), exp(-5.5),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), exp(-5.5),
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testFactorial()
{
  ::Reaction* r = m -> getReaction(3);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeFactorial));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 120,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 120, 
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testFloor()
{
  ::Reaction* r = m -> getReaction(4);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeFloor));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -6,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), -6, 
    std::numeric_limits<double>::epsilon());
}


void
TestReactionNode::testLn()
{
  ::Reaction* r = m -> getReaction(5);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeLn));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), log(5), 
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), log(5),
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testLog()
{
  ::Reaction* r = m -> getReaction(6);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeLog));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), log10(5),
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), log10(5),
    std::numeric_limits<double>::epsilon());
}


void
TestReactionNode::testPower()
{
  ::Reaction* r = m -> getReaction(7);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  MesoRD::UnitFactor uf(-2, MesoRD::UnitFactor::second, +1);
  std::vector<MesoRD::UnitFactor> UnitFactors;
  UnitFactors.push_back(uf);
  MesoRD::Unit u(UnitFactors);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodePower));
  CPPUNIT_ASSERT(rn -> getUnit() == u);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 25,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 25,
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testRoot()
{
  ::Reaction* r = m -> getReaction(8);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeRoot));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 4,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 4,
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testAnd()
{
  ::Reaction* r = m -> getReaction(9);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeAnd));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testNot()
{
  ::Reaction* r = m -> getReaction(10);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeNot));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testOr()
{
  ::Reaction* r = m -> getReaction(11);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeOr));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testXor()
{
  ::Reaction* r = m -> getReaction(12);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeXor));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}


void
TestReactionNode::testAddition()
{
  ::Reaction* r = m -> getReaction(13);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeAddition));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -0.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), -0.5,
    std::numeric_limits<double>::epsilon());
  
  MesoRD::ReactionNodeAddition rn2 = MesoRD::ReactionNodeAddition(
    new MesoRD::ReactionNodeConstant(45, MesoRD::Unit::perSecond(), &pt, &st),
    new MesoRD::ReactionNodeConstant(30 ,MesoRD::Unit::perSecond(), &pt, &st),
    &pt, &st);
  
  CPPUNIT_ASSERT(rn2.getUnit() == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_int), 75,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_double), 75,
    std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT(MesoRD::ReactionNode::isAddition(rn));
}

void
TestReactionNode::testDivision()
{
  ::Reaction* r = m -> getReaction(14);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  MesoRD::Unit u(MesoRD::Unit::litre()*(MesoRD::Unit::mole()^(-1)));

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeDivision));
  CPPUNIT_ASSERT(rn -> getUnit() == u);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -1.1, 
    std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -1.1,
    std::numeric_limits<double>::epsilon());
  
  
  MesoRD::ReactionNodeDivision rn2 = MesoRD::ReactionNodeDivision(
    new MesoRD::ReactionNodeConstant(34, MesoRD::Unit::perSecond(), &pt, &st),
    new MesoRD::ReactionNodeConstant(34, MesoRD::Unit::dimensionless(), 
      &pt, &st), &pt, &st);
    
  CPPUNIT_ASSERT(rn2.getUnit() == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_int), 1,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_double), 1,
    std::numeric_limits<double>::epsilon());
  
  CPPUNIT_ASSERT(MesoRD::ReactionNode::isDivision(rn));
}

void
TestReactionNode::testMultiplication()
{
  ::Reaction* r = m -> getReaction(15);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  MesoRD::Unit u(MesoRD::Unit::perMolarPerSecond());

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeMultiplication));
  CPPUNIT_ASSERT(rn -> getUnit() == (u^2));
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -27.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), -27.5,
    std::numeric_limits<double>::epsilon());
  
  
  MesoRD::ReactionNodeMultiplication rn2 = MesoRD::ReactionNodeMultiplication(
    new MesoRD::ReactionNodeConstant(10, MesoRD::Unit::perSecond(), &pt, &st),
    new MesoRD::ReactionNodeConstant(10, MesoRD::Unit::dimensionless(), 
      &pt, &st), &pt, &st);
  

  CPPUNIT_ASSERT(rn2.getUnit() == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_int), 100,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_double), 100,
    std::numeric_limits<double>::epsilon());
  
  CPPUNIT_ASSERT(MesoRD::ReactionNode::isMultiplication(rn));
}

void
TestReactionNode::testSubtraction()
{
  ::Reaction* r = m -> getReaction(16);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  MesoRD::Unit u(MesoRD::Unit::perMolarPerSecond());

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeSubtraction));
  CPPUNIT_ASSERT(rn -> getUnit() == u);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), -10.5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), -10.5,
    std::numeric_limits<double>::epsilon());

  MesoRD::ReactionNodeSubtraction rn2 = MesoRD::ReactionNodeSubtraction(
    new MesoRD::ReactionNodeConstant(34, MesoRD::Unit::perSecond(), &pt, &st),
    new MesoRD::ReactionNodeConstant(12, MesoRD::Unit::perSecond(), &pt, &st),
    &pt, &st);
  
  CPPUNIT_ASSERT(rn2.getUnit() == MesoRD::Unit::perSecond());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_int), 22,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn2.evaluate(val_int), 22,
    std::numeric_limits<double>::epsilon());
  
  CPPUNIT_ASSERT(MesoRD::ReactionNode::isSubtraction(rn));
}

void
TestReactionNode::testEq()
{
  
  ::Reaction* r = m -> getReaction(17);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeEq));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testGeq()
{
  ::Reaction* r = m -> getReaction(18);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeGeq));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testGt()
{
  ::Reaction* r = m -> getReaction(19);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeGt));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testLeq()
{
  ::Reaction* r = m -> getReaction(20);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeLeq));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testLt()
{
  ::Reaction* r = m -> getReaction(21);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeLt));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testNeq()
{
  ::Reaction* r = m -> getReaction(22);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeNeq));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::dimensionless());
  CPPUNIT_ASSERT(rn -> evaluate(val_int));
  CPPUNIT_ASSERT(rn -> evaluate(val_double));
}

void
TestReactionNode::testConcentration()
{
  ::Reaction* r = m -> getReaction(23);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeMultiplication));
  CPPUNIT_ASSERT(rn -> getUnit() == 
    (MesoRD::Unit::mole() / MesoRD::Unit::litre()));


  std::valarray<unsigned int> val_int2(1);
  std::valarray<double> val_double2(1);
  
  val_int2 = 14;
  val_double2 = 14;
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int2), 
    14 / (MesoRD::Constant::Avogadro*1e-8),
    std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double2), 
    14/(MesoRD::Constant::Avogadro*1e-8),
    std::numeric_limits<double>::epsilon());
}

void
TestReactionNode::testParameterConstant()
{
  ::Reaction* r = m -> getReaction(24);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeConstant));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::perMolarPerSecond());
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 5,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 5,
    std::numeric_limits<double>::epsilon());

  CPPUNIT_ASSERT(MesoRD::ReactionNode::isConstant(rn));
}

void
TestReactionNode::testParameter()
{
  ::Reaction* r = m -> getReaction(25);
  MesoRD::ReactionNode* rn = 
    MesoRD::ReactionNode::fromSBML((r -> getKineticLaw()) -> getMath(), 
      &pt, &st);

  CPPUNIT_ASSERT(typeid(*rn) == typeid(MesoRD::ReactionNodeParameter));
  CPPUNIT_ASSERT(rn -> getUnit() == MesoRD::Unit::litre());
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_int), 4,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(rn -> evaluate(val_double), 4,
    std::numeric_limits<double>::epsilon());
}
