/* The TestReactionTable class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestReaction.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestReactionTable.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestReactionTable );


void 
TestReactionTable::setUp()
{
  std::string test_filename 
    = "./sbml_source/TestReaction.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
  
  std::map<std::string, double> diffConst;
  MesoRD::Species s(diffConst, "A1", "A1","comp", ((unsigned int)14), false);
  st.addSpecies(s);
  MesoRD::Species s1(diffConst, "A2", "A2","comp2", ((unsigned int)14), false);
  st.addSpecies(s1);

  MesoRD::UnitTable ut(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  MesoRD::Parameter p2(m -> getParameter(1), &ut);
  MesoRD::Parameter p3(m -> getParameter(2), &ut);
  MesoRD::Parameter p4(m -> getParameter(3), &ut);
  MesoRD::Parameter p5("volume_subVolume", "volume_subVolume", 
    1e-8, MesoRD::Unit::litre(), 
    true);
  MesoRD::Parameter p6("p6", "p6", 4, MesoRD::Unit::litre(), false);
  MesoRD::Parameter p7("p7", "p7", (1e-8)*MesoRD::Constant::Avogadro, 
    (MesoRD::Unit::litre()/MesoRD::Unit::mole()), true);
  
  pt.addParameter(p1);
  pt.addParameter(p2);
  pt.addParameter(p3);
  pt.addParameter(p4);
  pt.addParameter(p5);
  pt.addParameter(p6);
  pt.addParameter(p7);
}


void 
TestReactionTable::tearDown()
{}


void 
TestReactionTable::testGetMultiSVReactionArray()
{
  
  MesoRD::ReactionTable rt;
  
  CPPUNIT_ASSERT(rt.getNumReactions()==0);
  
  MesoRD::Reaction* r1 = new MesoRD::Reaction(m->getReaction(0), &pt, &st, NULL);
  MesoRD::Reaction* r2 = new MesoRD::Reaction(m->getReaction(1), &pt, &st, NULL);

  CPPUNIT_ASSERT(r1->getMultiSVReaction()==true);
  CPPUNIT_ASSERT(r2->getMultiSVReaction()==true);
  
  rt.addReaction(*r1);
  rt.addReaction(*r2);
 
  CPPUNIT_ASSERT(rt.getNumReactions() == 2);
  
  
  std::valarray<bool> temp = rt.getMultiSVReactionArray();
  
  CPPUNIT_ASSERT(temp.size() == 2);
  
  CPPUNIT_ASSERT(temp[0] == true);
  CPPUNIT_ASSERT(temp[1] == true);
  
}

void
TestReactionTable::testGetMultiSVReaction()
{
  
  MesoRD::ReactionTable rt;
  
  CPPUNIT_ASSERT(rt.getNumReactions()==0);
  
  MesoRD::Reaction r1(m->getReaction(0), &pt, &st, NULL);
  MesoRD::Reaction r2(m->getReaction(1), &pt, &st, NULL);

  
  CPPUNIT_ASSERT(r1.getMultiSVReaction()==true);
  CPPUNIT_ASSERT(r2.getMultiSVReaction()==true);

  rt.addReaction(*(new MesoRD::Reaction(m->getReaction(0), &pt, &st, NULL)));
  rt.addReaction(*(new MesoRD::Reaction(m->getReaction(1), &pt, &st, NULL)));

  CPPUNIT_ASSERT(rt.getMultiSVReaction(0) == true);
  CPPUNIT_ASSERT(rt.getMultiSVReaction(1) == true);
}

void 
TestReactionTable::testGetRates()
{
  MesoRD::ReactionTable rt;
  
  MesoRD::Reaction* r1 = new MesoRD::Reaction(m->getReaction(0), &pt, &st, NULL);
  MesoRD::Reaction* r2 = new MesoRD::Reaction(m->getReaction(1), &pt, &st, NULL);
  
  rt.addReaction(*r1);
  rt.addReaction(*r2);

  std::valarray<unsigned int> species((unsigned int)0, 2);
  double theRates = rt.getRates(species);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( (r1 -> getRate(species) + r2 -> getRate(species)), theRates, 10* std::numeric_limits<double>::epsilon());

  std::valarray<unsigned int> species2((unsigned int)1, 2);
  double theRates2 = rt.getRates(species2);
  CPPUNIT_ASSERT_DOUBLES_EQUAL( (r1 -> getRate(species2) + r2 -> getRate(species2)), theRates2, 10* std::numeric_limits<double>::epsilon());

}

void
TestReactionTable::testGetMultiRates()
{
  MesoRD::ReactionTable rt;
  
  MesoRD::Reaction* r1 = new MesoRD::Reaction(m->getReaction(0), &pt, &st, NULL);
  MesoRD::Reaction* r2 = new MesoRD::Reaction(m->getReaction(1), &pt, &st, NULL);
  
  rt.addReaction(*r1);
  rt.addReaction(*r2);
  
  std::valarray<unsigned int> theReactants1 = r1-> getReactants();
  CPPUNIT_ASSERT(theReactants1.size()==1);
  CPPUNIT_ASSERT(theReactants1[0] == 0);
  
  std::valarray<unsigned int> theReactants2 = r2 -> getReactants();
  CPPUNIT_ASSERT(theReactants2.size()==0);
  
  
  std::valarray<unsigned int> species((unsigned int)0, 2);
  double theRates = rt.getRates(species, species);
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL((1*r1 -> getRate(species) + 0*r2 -> getRate(species)), theRates, 10* std::numeric_limits<double>::epsilon());
}

void 
TestReactionTable::testGetMultiRatesArray()
{
  MesoRD::ReactionTable rt;
  
  MesoRD::Reaction* r1 = new MesoRD::Reaction(m->getReaction(0), &pt, &st, NULL);
  MesoRD::Reaction* r2 = new MesoRD::Reaction(m->getReaction(1), &pt, &st, NULL);
  
  rt.addReaction(*r1);
  rt.addReaction(*r2);
  
  std::valarray<unsigned int> theReactants1 = r1-> getReactants();
  CPPUNIT_ASSERT(theReactants1.size()==1);
  CPPUNIT_ASSERT(theReactants1[0] == 0);
  
  std::valarray<unsigned int> theReactants2 = r2 -> getReactants();
  CPPUNIT_ASSERT(theReactants2.size()==0);
  
  
  std::valarray<unsigned int> species((unsigned int)0, 2);
  std::valarray<double> theRates = rt.getRatesArray(species, species);

  CPPUNIT_ASSERT(theRates.size() == 1);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(1*r1 -> getRate(species), theRates[0], 10* std::numeric_limits<double>::epsilon());

}



/*
void
TestReaction::testGetCompartment()
{
  ::Reaction* r = m -> getReaction(0);
      
  CPPUNIT_ASSERT(MesoRD::Reaction::getCompartment(r, &st) == "comp");
  
  r = m -> getReaction(1);
  
  CPPUNIT_ASSERT(MesoRD::Reaction::getCompartment(r, &st) == "comp");

  r = m -> getReaction(2);
  
  CPPUNIT_ASSERT(MesoRD::Reaction::getCompartment(r, &st) == "");
}

void
TestReaction::testGetCompartmentThrow()
{
  ::Reaction* r = m -> getReaction(3);
      
  std::string test = MesoRD::Reaction::getCompartment(r, &st);
  std::cerr << test << std::endl;
  
}

void
TestReaction::testFromSBML()
{
  ::Reaction* r = m -> getReaction(4);
  
  MesoRD::Reaction reaction(r, &pt, &st, NULL);
  
  
  reaction.fromSBML(r, &st);
}

void
TestReaction::testGetRate()
{
  ::Reaction* r = m -> getReaction(4);
  
  MesoRD::Reaction reaction(r, &pt, &st, NULL);
  

  reaction.fromSBML(r, &st);
  
  std::valarray<unsigned int> val_int2(2);
  std::valarray<double> val_double2(2);
    
  val_int2[0] = 2;
  val_double2[0] = 2;
  val_int2[1] = 4;
  val_double2[1] = 4;
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(reaction.getRate(val_int2), (double)8, 
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(reaction.getRate(val_double2), (double)8, 
    std::numeric_limits<double>::epsilon());

}

void
TestReaction::testGetMultiSVReaction()
{
  ::Reaction* r = m -> getReaction(4);
  
  MesoRD::Reaction reaction(r, &pt, &st, NULL);
  

  reaction.fromSBML(r, &st);
  
  CPPUNIT_ASSERT(reaction.getMultiSVReaction());
}  
*/
