/* The TestSpeciesTable class.
 *
 * Copyright 2004-2005 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestSpeciesTable.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestSpeciesTable.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestSpeciesTable );


void 
TestSpeciesTable::setUp()
{
}


void 
TestSpeciesTable::tearDown()
{
}


void 
TestSpeciesTable::testAddSpecies()
{

  MesoRD::SpeciesTable st;

  std::map<std::string, double> diffConst;
  diffConst["comp1"] = 234;
  diffConst["comp2"] = 235;
  
  st.addSpecies(diffConst, "test_id", "test_name", "test_comp",
    (unsigned int) 45);
  
  
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 45);
  CPPUNIT_ASSERT(st.getCompartmentByIndex(0) == "test_comp");
  CPPUNIT_ASSERT(st.getId(0) == "test_id");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getDiffusion(0, "comp1"), (double)234.0,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getDiffusion(0, "comp2"), (double)235.0,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT(st.getKey(0) == "test_name");
  CPPUNIT_ASSERT(st.getSpeciesNumber() == 1);

  
  MesoRD::Species s1(diffConst, "test_species_id1", "test_species_name1", 
    "comp_3", ((unsigned int)14), false);
  
  st.addSpecies(s1);
  
  CPPUNIT_ASSERT(st.getAmountByIndex(1) == 14);
  CPPUNIT_ASSERT(st.getCompartmentByIndex(1) == "comp_3");
  CPPUNIT_ASSERT(st.getId(1) == "test_species_id1");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getDiffusion(1, "comp1"), (double)234.0,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getDiffusion(1, "comp2"), (double)235.0,
    std::numeric_limits<double>::epsilon());
  CPPUNIT_ASSERT(st.getKey(1) == "test_species_name1");
  CPPUNIT_ASSERT(st.getSpeciesNumber() == 2);

}

void
TestSpeciesTable::testIsFunctions()
{

  
MesoRD::SpeciesTable st;

  std::map<std::string, double> diffConst;
  diffConst["comp1"] = 234;
  diffConst["comp2"] = 235;
  
  MesoRD::Species s1(diffConst, "test_species_id1", "test_species_name1", 
    "comp_3", ((unsigned int)14), false);
  
  st.addSpecies(s1);
  
  CPPUNIT_ASSERT(!st.isConstantByIndex(0));
  CPPUNIT_ASSERT(st.isDefined("test_species_name1"));
  CPPUNIT_ASSERT(st.isDefinedById("test_species_id1"));
  
}

void 
TestSpeciesTable::testGetSets()
{

  MesoRD::SpeciesTable st;

  std::map<std::string, double> diffConst;
  diffConst["comp1"] = 234;
  diffConst["comp2"] = 235;
  
  MesoRD::Species s1(diffConst, "test_species_id1", "test_species_name1", 
    "comp_3", ((unsigned int)14), false);
  
  st.addSpecies(s1);
    
  st.setAmountByIndex(0, (unsigned int)456);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 456);
  
  st.setAmountById("test_species_id1", (unsigned int)4562);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 4562);
  
  st.setAmountByName("test_species_name1", (unsigned int)452);
  CPPUNIT_ASSERT(st.getAmountByName("test_species_name1") == 452);
  
  
  st.setAmountByIndex(0, (double)456);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(0),(double)456.0,
    std::numeric_limits<double>::epsilon());
  
  st.setAmountById("test_species_id1", (double)4562);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(0), (double)4562.0,
    std::numeric_limits<double>::epsilon());
  
  st.setAmountByName("test_species_name1", (double)452);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByNameDouble("test_species_name1"),
    (double)452.0, std::numeric_limits<double>::epsilon());
}


void
TestSpeciesTable::testChangeFunctions()
{

  MesoRD::SpeciesTable st;
   
  std::map<std::string, double> diffConst;
  diffConst["comp1"] = 234;
  diffConst["comp2"] = 235;
  
  MesoRD::Species s1(diffConst, "test_species_id1", "test_species_name1", 
    "comp_3", ((unsigned int)14), false);
  
  st.addSpecies(s1);
  
    
  st.changeAmountByIndex(0, (int)45);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 59);
  
  st.changeAmountById("test_species_id1", (int)3);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 62);
  
  st.changeAmountByName("test_species_name1", (int)33);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 95);
  
  st.changeAmountByIndex(0, (int)-15);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 80);
  
  st.changeAmountById("test_species_id1", (int)-10);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 70);
  
  st.changeAmountByName("test_species_name1", (int)-20);
  CPPUNIT_ASSERT(st.getAmountByIndex(0) == 50);

  
   MesoRD::Species s2(diffConst, "test_species_id2", "test_species_name2", 
    "comp_3", ((double)14), false);
  
  st.addSpecies(s2);
  
  CPPUNIT_ASSERT(st.getId(1) == "test_species_id2");
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1),14,
    std::numeric_limits<double>::epsilon());
  

  st.changeAmountByIndex(1, (double)45);  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 59,
    std::numeric_limits<double>::epsilon());

  st.changeAmountById("test_species_id2", (double)3);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 62,
    std::numeric_limits<double>::epsilon());
  
  st.changeAmountByName("test_species_name2", (double)33);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 95,
    std::numeric_limits<double>::epsilon());
  
  st.changeAmountByIndex(1, (double)-15);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 80,
    std::numeric_limits<double>::epsilon());
  
  st.changeAmountById("test_species_id2", (double)-10);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 70,
    std::numeric_limits<double>::epsilon());
  
  st.changeAmountByName("test_species_name2", (double)-20);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(st.getAmountByIndexDouble(1), 50,
    std::numeric_limits<double>::epsilon());
  
}


