/* The TestSubVolume class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestSubVolume.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestSubVolume.hpp"
#include "../src/sparselib/comprow_double.hpp"
#include "../src/mv/mvvd.hpp"

#include "../src/CubicSubVolume.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestSubVolume );


void 
TestSubVolume::setUp()
{
  
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  MesoRD::Species::setUseDoubleInAmount(false);

  std::string test_filename 
    = "./sbml_source/TestSubVolume.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
  m = d -> getModel();
  uds = m -> getListOfUnitDefinitions();
  
  ut = MesoRD::UnitTable(uds);
  MesoRD::Parameter p1(m -> getParameter(0), &ut);
  pt.addParameter(p1);
  MesoRD::Parameter p2(m -> getParameter(1), &ut);
  pt.addParameter(p2);


   
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  MesoRD::Species::setUseDoubleInAmount(true);

  subvolumes = ct.fromSBML(m, &pt, &ut, 1, "um", NULL);
  
  subvolumes_det = ct_det.fromSBML(m, &pt, &ut, 1, "um", NULL);

  
}


void 
TestSubVolume::tearDown()
{}

void
TestSubVolume::testConstructor()
{
  
  CPPUNIT_ASSERT(subvolumes.size() == 3);
  
  
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::CubicSubVolume s(NULL, 4, 12);
  
  CPPUNIT_ASSERT(MesoRD::SubVolume::getUseDoubleInSpecies() == false);
  
  CPPUNIT_ASSERT(s.getSpeciesNumber() == 12);
  CPPUNIT_ASSERT(s.getSize() == 4);
  
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  CPPUNIT_ASSERT(MesoRD::SubVolume::getUseDoubleInSpecies() == true);
  MesoRD::CubicSubVolume s2(NULL, 5,13);
  
  CPPUNIT_ASSERT(s2.getSpeciesNumber() == 13);
  CPPUNIT_ASSERT(s2.getSize() == 5);
  
}

void
TestSubVolume::testGetSet()
{
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::CubicSubVolume s(NULL, 3,12);
  
  MesoRD::Compartment c(NULL);
  
  s.setCompartment(&c);
  s.setQueueIndex(45);
  s.setSize(2);
  s.setSpecies(4, (unsigned int) 34);
    
  s.setX(12);
  s.setY(2);
  s.setZ(127);
  
  CPPUNIT_ASSERT(s.getCompartment() == &c);
  CPPUNIT_ASSERT(s.getQueueIndex() == 45);
  CPPUNIT_ASSERT(s.getSize() == 2);
  CPPUNIT_ASSERT(s.getSpecies(4) == 34);
  CPPUNIT_ASSERT(s.getX() == 12);
  CPPUNIT_ASSERT(s.getY() == 2);
  CPPUNIT_ASSERT(s.getZ() == 127);

  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  
  MesoRD::CubicSubVolume s2(NULL, 12,4);
    
  s2.setSpecies(2, (double)23);

  CPPUNIT_ASSERT(s2.getSpeciesDouble(2) == 23);
  
}

void
TestSubVolume::testChangeFunctions()
{

  MesoRD::SubVolume::setUseDoubleInSpecies(false);

  MesoRD::Compartment c(&ct, NULL, &st, &pt, &ut, m, 0);

  MesoRD::SubVolume* s = subvolumes[0];
  
  unsigned int species_number = s -> getSpecies(0);
  
  s -> changeSpecies(0,(int)2);
  
  CPPUNIT_ASSERT(s -> getSpecies(0) == species_number + 2);
  
  s -> changeSpecies(0,(int)-2);
  
  CPPUNIT_ASSERT(s -> getSpecies(0) == species_number);





  MesoRD::SubVolume::setUseDoubleInSpecies(true);

  MesoRD::Compartment c_det(&ct_det, NULL, &st, &pt, &ut, m, 0);

  MesoRD::SubVolume* s_det = subvolumes_det[0];
  
  double species_number_det = s_det -> getSpeciesDouble(0);
  
  s_det -> changeSpecies(0,(double)2);
  
  CPPUNIT_ASSERT(s_det -> getSpeciesDouble(0) == species_number_det + 2);
  
  s_det -> changeSpecies(0,(double)-2);
  
  CPPUNIT_ASSERT(s_det -> getSpeciesDouble(0) == species_number_det);

}

void
TestSubVolume::testEventReaction()
{
  
  MesoRD::SubVolume::setUseDoubleInSpecies(false);
  
  MesoRD::Compartment c(&ct, NULL, &st, &pt, &ut, m, 0);
  
  MesoRD::SubVolume* s = subvolumes[0];
  
/*
  s -> setSpecies(0, (unsigned int)4);

  s -> updateTime(0.5);

  s -> eventReaction(0.3333333);
  
  CPPUNIT_ASSERT(s -> getSpecies(0) == 3);
  
  s -> setSpecies(0, (unsigned int)4);

  s -> updateTime(0.5);

  s -> eventReaction(0.5);
  
  CPPUNIT_ASSERT(s -> getSpecies(0) == 5);
*/
}

void
TestSubVolume::testGetDerivativeVector()
{
   
  MesoRD::SubVolume::setUseDoubleInSpecies(true);
  
  MesoRD::Compartment c(&ct_det, NULL, &st, &pt, &ut, m, 0);
  
  MesoRD::SubVolume* s = subvolumes[0];
  
  s -> setSpecies(0, (double)4);
  
  MV_Vector_double currentState(2);
  currentState(0) =  s -> getSpeciesDouble(0);
  currentState(1) =  s -> getSpeciesDouble(1);


  MV_Vector_double derVector(2);
  derVector(0) = 0;
  derVector(1) = 0;
  
  std::valarray<double> temp((double)0, 2);
  
  s -> getDerivativeVector(currentState, derVector, 0, temp);
    /*
  CPPUNIT_ASSERT(derVector(0) == 20);
  CPPUNIT_ASSERT(derVector(1) == 0);
    */
}


  
