/* The TestUnit class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: TestUnit.cpp 550 2008-05-07 14:24:19Z davidfange $
 */

#include "TestUnit.hpp"
#include "../src/UnitFactor.hpp"
#include "../src/Unit.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( TestUnit );


void 
TestUnit::setUp()
{
  std::string test_filename = "./sbml_source/TestUnit.xml";
  SBMLDocument* d = readSBML(test_filename.c_str());
    //m = SBMLDocument_getModel(d);
  m = d -> getModel();
  
}


void 
TestUnit::tearDown()
{
}


void 
TestUnit::testConstructor()
{
  MesoRD::UnitFactor metre=MesoRD::UnitFactor(1, 
    MesoRD::UnitFactor::metre, (double) 1);
  
  MesoRD::UnitFactor second=MesoRD::UnitFactor(-2, 
    MesoRD::UnitFactor::second, (double) 1);

  
  std::vector<MesoRD::UnitFactor> UnitFactors;
  UnitFactors.push_back(metre);
  UnitFactors.push_back(second);
  
  MesoRD::Unit unit1(UnitFactors, "test_id", "test_name");

  MesoRD::Unit unit2( m -> getUnitDefinition(0) );

  unit1.fold();
  unit2.fold();
  

  CPPUNIT_ASSERT(unit1.getKey() == "test_id");
  CPPUNIT_ASSERT(unit2.getKey() == "metrepersecond_squared");
  CPPUNIT_ASSERT(unit1 == unit2);
  
}

void
TestUnit::testFold()
{

  MesoRD::UnitFactor metre=MesoRD::UnitFactor
    (1, MesoRD::UnitFactor::metre, (double) 10000);

  std::vector<MesoRD::UnitFactor> UnitFactors;
  UnitFactors.push_back(metre);

  MesoRD::Unit finmeter(UnitFactors, "test_id", "test_name");
  MesoRD::Unit fulmeter(m -> getUnitDefinition(1));
  
  CPPUNIT_ASSERT(fulmeter.getKey() == "fulmeter");
  CPPUNIT_ASSERT(finmeter == fulmeter);

  double multiplier = pow(MesoRD::Constant::Avogadro,-2)*1000;

  MesoRD::UnitFactor second=MesoRD::UnitFactor(2, 
    MesoRD::UnitFactor::second, (double) multiplier);

  MesoRD::UnitFactor metre2=MesoRD::UnitFactor(-2, 
    MesoRD::UnitFactor::metre, (double) 1);
  
  UnitFactors.clear();
  
  UnitFactors.push_back(second);
  UnitFactors.push_back(metre2);
  
  MesoRD::Unit test_fold1(UnitFactors, "test_fold", "test_fold");
    
  MesoRD::Unit test_fold2( m -> getUnitDefinition(2) );
  test_fold2.fold();
  
  CPPUNIT_ASSERT(test_fold2.getKey() == "test_fold");
  CPPUNIT_ASSERT(test_fold1 == test_fold2);
}


void
TestUnit::testOperators()
{

    /* Plus
     */
  MesoRD::Unit pMps1 = MesoRD::Unit::perMolarPerSecond();
  MesoRD::Unit pMps2 = MesoRD::Unit::perMolarPerSecond();
  
  CPPUNIT_ASSERT(pMps1 == (pMps1 + pMps2));
  
    /* Minus
     */
  MesoRD::Unit cm2ps1 = MesoRD::Unit::centimetreSquaredPerSecond();
  MesoRD::Unit cm2ps2 = MesoRD::Unit::centimetreSquaredPerSecond();
  
  CPPUNIT_ASSERT(cm2ps1 == (cm2ps1 - cm2ps2));
  
    /* Multiply
     */
  MesoRD::Unit mPs = MesoRD::Unit::molarPerSecond();
  MesoRD::Unit pS = MesoRD::Unit::perSecond();
  
  CPPUNIT_ASSERT( (pMps1 * mPs) == ( pS * pS));
  
    /* Divide
     */
  MesoRD::Unit mole = MesoRD::Unit::mole();
  MesoRD::Unit litre = MesoRD::Unit::litre();

  CPPUNIT_ASSERT( (pMps1 / mPs) == ( (litre/mole)*(litre/mole)));

    /* Power
     */
  CPPUNIT_ASSERT( (pMps1 / mPs) == ( (mole/litre)^(-2)));
  CPPUNIT_ASSERT( MesoRD::Unit::dimensionless() == 
    (MesoRD::Unit::dimensionless())^(-1));
  
  
  
}

void
TestUnit::testPlusThrow()
{
  
  MesoRD::Unit pMps = MesoRD::Unit::perMolarPerSecond();
  MesoRD::Unit perSecond = MesoRD::Unit::perSecond();

  MesoRD::Unit sum = pMps + perSecond;
  
}

void
TestUnit::testMinusThrow()
{
  MesoRD::Unit mPs = MesoRD::Unit::molarPerSecond();
  MesoRD::Unit mole = MesoRD::Unit::mole();
  
  MesoRD::Unit diff = mPs - mole;
}


void
TestUnit::testBuildIns()
{
  
    /* centimetreSquaredPerSecond 
     */
   MesoRD::UnitFactor metre=MesoRD::UnitFactor(2, 
     MesoRD::UnitFactor::metre, (double) 1e-4);

   MesoRD::UnitFactor second=MesoRD::UnitFactor(-1, 
     MesoRD::UnitFactor::second, (double) 1);

   std::vector<MesoRD::UnitFactor> UnitFactors;
   UnitFactors.push_back(metre);
   UnitFactors.push_back(second);
   
   MesoRD::Unit cm2ps(UnitFactors);
   cm2ps.fold();
   
   MesoRD::Unit cm2psBI = MesoRD::Unit::centimetreSquaredPerSecond();
   
   CPPUNIT_ASSERT(cm2ps == cm2psBI);
   

     /* litre
      */
   metre = MesoRD::UnitFactor(3,MesoRD::UnitFactor::metre, (double) 1e-3);

   UnitFactors.clear();
   UnitFactors.push_back(metre);
      
   MesoRD::Unit litre(UnitFactors);
   MesoRD::Unit litreBI = MesoRD::Unit::litre();
   
   CPPUNIT_ASSERT(litre == litreBI);
   

     /* mole
      */
   MesoRD::UnitFactor dimensionless =MesoRD::UnitFactor(1, 
     MesoRD::UnitFactor::dimensionless, (double) MesoRD::Constant::Avogadro);

   UnitFactors.clear();
   UnitFactors.push_back(dimensionless);
      
   MesoRD::Unit mole(UnitFactors);
   MesoRD::Unit moleBI = MesoRD::Unit::mole();
   
   CPPUNIT_ASSERT(mole == moleBI);
   
   
   
     /* molarPerSecond
      */
   dimensionless =MesoRD::UnitFactor(1, 
     MesoRD::UnitFactor::dimensionless, (double) MesoRD::Constant::Avogadro);

   metre=MesoRD::UnitFactor(-3, MesoRD::UnitFactor::metre, (double) 1e3);
   second=MesoRD::UnitFactor(-1, MesoRD::UnitFactor::second, (double) 1);

   
   UnitFactors.clear();
   UnitFactors.push_back(dimensionless);
   UnitFactors.push_back(metre);
   UnitFactors.push_back(second);
      
   MesoRD::Unit molarpersecond(UnitFactors);
   molarpersecond.fold();

   MesoRD::Unit molarpersecondBI = MesoRD::Unit::molarPerSecond();

   CPPUNIT_ASSERT(molarpersecond == molarpersecondBI);
   


     /* perMolarPerSecond
      */
   dimensionless =MesoRD::UnitFactor(-1, 
     MesoRD::UnitFactor::dimensionless, 
     (double) pow(MesoRD::Constant::Avogadro,-1));

   metre=MesoRD::UnitFactor(3, MesoRD::UnitFactor::metre, (double) 1e-3);
   second=MesoRD::UnitFactor(-1, MesoRD::UnitFactor::second, (double) 1);

   
   UnitFactors.clear();
   UnitFactors.push_back(dimensionless);
   UnitFactors.push_back(metre);
   UnitFactors.push_back(second);
      
   MesoRD::Unit permolarpersecond(UnitFactors);
   permolarpersecond.fold();

   MesoRD::Unit permolarpersecondBI = MesoRD::Unit::perMolarPerSecond();

   CPPUNIT_ASSERT(permolarpersecond == permolarpersecondBI);  
}

void
TestUnit::testConversion()
{

  MesoRD::Unit pMps = MesoRD::Unit::perMolarPerSecond();
  MesoRD::Unit Mps = MesoRD::Unit::molarPerSecond();
  MesoRD::Unit metre = MesoRD::Unit::metre();
  
  double multiplier = (pMps/Mps).toUnit(metre^(6));
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(multiplier, 
    (1e-6)*(1/(pow(MesoRD::Constant::Avogadro,2))), 
    std::numeric_limits<double>::epsilon());
}

void
TestUnit::testConversionThrow()
{
  MesoRD::Unit pMps = MesoRD::Unit::perMolarPerSecond();
  MesoRD::Unit Mps = MesoRD::Unit::molarPerSecond();
  
  double uc = pMps.toUnit(Mps);
  
}

void 
TestUnit::testConversionThrow2()
{
  MesoRD::Unit cm2ps( m -> getUnitDefinition(4) );

  double uc = cm2ps.toUnit(MesoRD::Unit::centimetreSquaredPerSecond());
}

