/* The ReactionNode class.
 *
 * Copyright 2004-2008 Johan Elf, David Fange, Johan Hattne
 *
 * This file is part of MesoRD.
 *
 * MesoRD is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MesoRD is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MesoRD; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: ReactionNode.hpp 550 2008-05-07 14:24:19Z davidfange $
 */
#ifndef MESORD_REACTIONNODE_HPP
#define MESORD_REACTIONNODE_HPP

#include <fstream>
#include <list>
#include <stdexcept>
#include <string>
#include <valarray>

#include <sbml/math/ASTNode.h>
#include <sbml/SBMLTypes.h>

#include "../Unit.hpp"

namespace MesoRD {

  class LinearElement;
  class LinearFactor;
  class LinearTerm;
  class ParameterTable;
  class SpeciesTable;
  
  class ReactionNodeAddition;
  class ReactionNodeConcentration;
  class ReactionNodeConstant;
  class ReactionNodeDivision;
  class ReactionNodeMultiplication;
  class ReactionNodeSubtraction;
  

    /********************************
     * THE REACTION NODE BASE CLASS *
     ********************************/

    /**
     * Class for evaluating mathematical expressions used in
     * description of chemical reactions.
     *
     * SBML uses MathML to represent mathematical expressions used in
     * the description of chemical reactions.  SBML does not make use
     * of all of MathML, but limits itself to a small subset.  Given
     * the length of this file, one may argue that the subset is not
     * so small after all.
     *
     * Note that none of the classes actually do deep copy in their
     * copy constructors.
     *
     * Currently unsupported:
     *
     * AST_FUNCTION_PIECEWISE
     * AST_NAME_DELAY
     * AST_NAME_TIME
     * AST_LAMBDA
     * AST_UNKNOWN
     */
  class ReactionNode {
  protected:

      /**
       * The parameter table is needed in order to look up parameters.
       * For non-constant parameter, the parameter table will be
       * consulted during evaluation as well.
       */
    const ParameterTable* parameterTable;

      /**
       * The species table is needed in order to do species-to-index
       * mapping on the participating species.  This table is not
       * consulted during evaluation, so the index of a species must
       * never change once set.
       */
    const SpeciesTable* speciesTable;

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).  The reason for its existence is purely
       * technical; it makes the implementation a little nicer.
       * Callers should not use this method, but make use of the
       * operator instead.
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    virtual std::ostream& putTo(std::ostream& os) const = 0;


  public:


      /**
       * Class for reporting invalid unit usage.
       *
       * If we attempt to violate unit rules, for example by adding a
       * metre to a second, this exception will be thrown.
       */
    class InvalidUnit : public std::runtime_error {

        /**
         * The InvalidUnit constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
    public:
      InvalidUnit(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Class for reporting errors during model building.
       *
       * If something went wrong during the model construction, this
       * exception will be thrown.
       */
    class ModelError : public std::runtime_error {
    public:

        /**
         * The ModelError constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      ModelError(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Class for reporting implementation shortcomings.
       *
       * If one tries to use a feature that is not yet implemented,
       * this exception will be thrown.
       */
    class NotYetImplemented : public std::runtime_error {
    public:

        /**
         * The NotYetImplemented constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      NotYetImplemented(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Class for reporting references to undefined identifiers.
       *
       * If the mathematics contain references to identifiers that are
       * not defined, this exception will be thrown.
       */
    class UndefinedIdentifier : public std::runtime_error {
    public:

        /**
         * The UndefinedIdentifier constructor.
         *
         * Just enables the caller to set the "what" argument.  This
         * argument can provide further information about what went
         * wrong.
         *
         * @param what More information on what went wrong.
         */
      UndefinedIdentifier(const char* what = "")
        : runtime_error(what) {}
    };


      /**
       * Constructor for the reaction node base class.
       *
       * This method will set up the parameter table and species
       * table.  These tables are needed during the construction of
       * the mathematical expression to map parameters to values and
       * species to indices.  If we have non-constant parameters, the
       * parameter table will be needed during evaluation as well.
       *
       * @param n The AST to build the mathematics from.
       * @param p The parameter table.
       * @param s The species table.
       */
    ReactionNode(
      const ASTNode*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0);

      /**
       * Destructor for the reaction node base class.
       *
       * The destructors of the derived classes should make sure that
       * the destructors of any children the node may have are called
       * as well.  This will free the memory allocated to the reaction
       * tree recursively, and will prevent leaking memory on the heap
       * on systems that do not do garbage collection.
       */
    virtual ~ReactionNode();

      /**
       * Linearises the node to an expression that is easier to
       * evaluate.
       *
       * Calling this method will attempt to bring the form of the
       * node to a form that is easier, and hopefully faster, to
       * evaluate.  If it was not possible to do this, the method
       * returns null.
       *
       * @return A pointer to the linearised node.
       */
    virtual LinearElement* linearise()
      { return 0; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    virtual double evaluate(
      const std::valarray<unsigned int>& species) const = 0;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    virtual double evaluate(
      const std::valarray<double>& species) const = 0;


      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    virtual Unit getUnit() const throw(InvalidUnit) = 0;

      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    virtual ReactionNode* fold()
      { return this; }

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.  In fact, MesoRD will rebuild the
       * AST using its own classes defined here.  It returns a pointer
       * to the parsed root node of the new AST.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    static ReactionNode* fromSBML(
      const ASTNode *n, const ParameterTable* p, const SpeciesTable* s)
      throw(ModelError, NotYetImplemented, UndefinedIdentifier);

      /**
       * Check if a node is an addition node.
       *
       * This method will check whether the argument node is an
       * addition node.  It returns a casted pointer if it is and null
       * if it is not.  This is actually what dynamic_cast<>() would
       * do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not addition.
       */
    static ReactionNodeAddition* isAddition(ReactionNode* node);

      /**
       * Check if a node is a concentration node.
       *
       * This method will check whether the argument node is a
       * concentration node.  It returns a casted pointer if it is and
       * null if it is not.  This is actually what dynamic_cast<>()
       * would do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not concentration.
       */
    static ReactionNodeConcentration* isConcentration(ReactionNode* node);

      /**
       * Check if a node is a constant node.
       *
       * This method will check whether the argument node is a
       * constant node.  It returns a casted pointer if it is and null
       * if it is not.  This is actually what dynamic_cast<>() would
       * do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not constant.
       */
    static ReactionNodeConstant* isConstant(ReactionNode* node);

      /**
       * Check if a node is a division node.
       *
       * This method will check whether the argument node is a
       * division node.  It returns a casted pointer if it is and null
       * if it is not.  This is actually what dynamic_cast<>() would
       * do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not division.
       */
    static ReactionNodeDivision* isDivision(ReactionNode* node);

      /**
       * Check if a node is a factor node.
       *
       * This method will check whether the argument node is a factor
       * element.  It returns a casted pointer if it is and null if it
       * is not.  This is actually what dynamic_cast<>() would do.
       *
       * @param element A pointer to the query element.
       * @return        A casted pointer, or null if not factor.
       */
    static LinearFactor* isFactor(LinearElement* element);

      /**
       * Check if a node is an multiplication node.
       *
       * This method will check whether the argument node is a
       * multiplication node.  It returns a casted pointer if it is
       * and null if it is not.  This is actually what
       * dynamic_cast<>() would do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not multiplication.
       */
    static ReactionNodeMultiplication* isMultiplication(ReactionNode* node);

      /**
       * Check if a node is an subtraction node.
       *
       * This method will check whether the argument node is a
       * subtraction node.  It returns a casted pointer if it is and
       * null if it is not.  This is actually what dynamic_cast<>()
       * would do.
       *
       * @param node A pointer to the query node.
       * @return     A casted pointer, or null if not subtraction.
       */
    static ReactionNodeSubtraction* isSubtraction(ReactionNode* node);

      /**
       * Check if a node is an term node.
       *
       * This method will check whether the argument node is a term
       * element.  It returns a casted pointer if it is and null if it
       * is not.  This is actually what dynamic_cast<>() would do.
       *
       * @param element A pointer to the query element.
       * @return        A casted pointer, or null if not term.
       */
    static LinearTerm* isTerm(LinearElement* element);

      /**
       * Does constant folding on a node.
       *
       * The node is folded until we are sure that no more
       * transformation affect the root node.
       *
       * @param node A pointer to the node to fold.
       * @return     A pointer to the folded node.
       */
    static ReactionNode* refold(ReactionNode* node) 
      throw( InvalidUnit );

      /**
       * Prepends a factor to the reaction tree to handle item to
       * concentration conversion.
       *
       * This method will return a pointer to the tree rooted at the
       * same place, but multiplied by Avogadro constant and the
       * volume of a sub volume.
       *
       * @return A pointer to the converted tree.
       */
    ReactionNode* toConcentration();

      /**
       * Tries to convert the reactiontree to the unit of per second 
       * which is used is the unit used for reaction rate evaluation 
       * later on in the program.  
       *
       * @ return A pointer to the converted tree.
       */
    ReactionNode* toPerSecond(std::ostream* outstream_log) 
      throw( Unit::InvalidConversion );

      /**
       * Put a reaction node to an output stream.
       *
       * This method will put a reaction node onto an output stream.
       * The result should look something like MathML.
       *
       * @param os The output stream on which to write.
       * @param n  The reaction node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNode& n)
      { return n.putTo(os); }
  };



    /***********************************
     * FUNCTION CLASSES: THE ABS CLASS *
     ***********************************/

    /**
     * The ReactionNodeAbs class.
     *
     * This class implements a unary operator which represents the
     * absolute value of its argument.
     */
  class ReactionNodeAbs : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeAbs class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeAbs(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }
    
      /**
       * ReactionNodeAbs class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeAbs(const ReactionNodeAbs& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }
    
      /**
       * ReactionNodeAbs class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeAbs()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);
    
      /**
       * Put the ReactionNodeAbs node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeAbs& n)
      { return n.putTo(os); }
  };



    /************************************
     * FUNCTION CLASSES: THE CEIL CLASS *
     ************************************/

    /**
     * The ReactionNodeCeil class.
     *
     * This class implements an operator used to denote the round-up
     * (towards positive infinity) operator.
     */
  class ReactionNodeCeil : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCeil class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeCeil(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCeil class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCeil(const ReactionNodeCeil& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeCeil class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCeil()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCeil node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCeil& n)
      { return n.putTo(os); }
  };



    /***********************************
     * FUNCTION CLASSES: THE EXP CLASS *
     ***********************************/

    /**
     * The ReactionNodeExp class.
     *
     * This class represents the exponentiation function. It takes one
     * argument.
     */
  class ReactionNodeExp : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeExp class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeExp(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeExp class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeExp(const ReactionNodeExp& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeExp class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeExp()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeExp node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeExp& n)
      { return n.putTo(os); }
  };



    /*****************************************
     * FUNCTION CLASSES: THE FACTORIAL CLASS *
     *****************************************/

    /**
     * The ReactionNodeFactorial class.
     *
     * This class implements the unary operator used to construct
     * factorials.
     */
  class ReactionNodeFactorial : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeFactorial class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeFactorial(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeFactorial class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeFactorial(const ReactionNodeFactorial& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeFactorial class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeFactorial()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeFactorial node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeFactorial& n)
      { return n.putTo(os); }
  };



    /*************************************
     * FUNCTION CLASSES: THE FLOOR CLASS *
     *************************************/

    /**
     * The ReactionNodeFloor class.
     *
     * This class implements the floor operator, which is used to
     * denote the round-down (towards negative infinity) operator.
     */
  class ReactionNodeFloor : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeFloor class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeFloor(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeFloor class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeFloor(const ReactionNodeFloor& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeFloor class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeFloor()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeFloor node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeFloor& n)
      { return n.putTo(os); }
  };



    /**********************************
     * FUNCTION CLASSES: THE LN CLASS *
     **********************************/

    /**
     * The ReactionNodeLn class.
     *
     * This class implements the ln function (natural logarithm).  It
     * takes one argument.
     */
  class ReactionNodeLn : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeLn class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeLn(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeLn class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeLn(const ReactionNodeLn& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeLn class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeLn()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeLn node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeLn& n)
      { return n.putTo(os); }
  };



    /***********************************
     * FUNCTION CLASSES: THE LOG CLASS *
     ***********************************/

    /**
     * The ReactionNodeLog class.
     *
     * This class implements the log function.  If its first argument
     * is a "logbase" element, it specifies the base and the second
     * argument is the argument to which the function is applied using
     * that base.  If no logbase element is present, the base is
     * assumed to be 10.
     */
  class ReactionNodeLog : public ReactionNode {
  private:
    ReactionNode* argument;
    ReactionNode* base;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeLog class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeLog(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; base = 0; fromSBML(n); }

      /**
       * ReactionNodeLog class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeLog(const ReactionNodeLog& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; base = n.base; }

      /**
       * ReactionNodeLog class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeLog()
      { if (argument) delete argument; if (base) delete base; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeLog node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeLog& n)
      { return n.putTo(os); }
  };



    /*************************************
     * FUNCTION CLASSES: THE POWER CLASS *
     *************************************/

    /**
     * The ReactionNodePower class.
     *
     * This class implements the binary powering operator that is used
     * to construct expressions such as a "to the power of" b.  In
     * particular, it is the operation for which a "to the power of" 2
     * is equivalent to a * a.
     */
  class ReactionNodePower : public ReactionNode {
  private:
    ReactionNode* argument;
    ReactionNode* power;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodePower class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodePower(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; power = 0; fromSBML(n); }

      /**
       * ReactionNodePower class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodePower(const ReactionNodePower& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; power = n.power; }

      /**
       * ReactionNodePower class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodePower()
      { if (argument) delete argument; if (power) delete power; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodePower node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodePower& n)
      { return n.putTo(os); }
  };



    /************************************
     * FUNCTION CLASSES: THE ROOT CLASS *
     ************************************/

    /**
     * The ReactionNodeRoot class.
     *
     * This class implements the binary operator used to construct the
     * nth root of an expression.  The first argument "a" is the
     * expression and the second object "n" denotes the root, as in (
     * a ) ^ (1/n).
     */
  class ReactionNodeRoot : public ReactionNode {
  private:
    ReactionNode* argument;
    ReactionNode* degree;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeRoot class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeRoot(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; degree = 0; fromSBML(n); }

      /**
       * ReactionNodeRoot class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeRoot(const ReactionNodeRoot& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; degree = n.degree; }

      /**
       * ReactionNodeRoot class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeRoot()
      { if (argument) delete argument; if (degree) delete degree; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeRoot node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeRoot& n)
      { return n.putTo(os); }
  };



    /*********************************************
     * BOOLEAN OPERATIONS CLASSES: THE AND CLASS *
     *********************************************/

    /**
     * The ReactionNodeAnd class.
     *
     * This class implements the n-ary logical "and" operator.  It is
     * used to construct the logical expression which were it to be
     * evaluated would have a value of "true" when all of its operands
     * have a truth value of "true", and "false" otherwise.
     */
  class ReactionNodeAnd : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeAnd class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeAnd(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeAnd class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeAnd(const ReactionNodeAnd& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeAnd class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeAnd()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeAnd node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeAnd& n)
      { return n.putTo(os); }
  };



    /*********************************************
     * BOOLEAN OPERATIONS CLASSES: THE NOT CLASS *
     *********************************************/

    /**
     * The ReactionNodeNot class.
     *
     * This class implements the unary logical "not" operator.  It
     * negates the truth value of its single argument.  For example,
     * not P is true when P is false and false when P is true.
     */
  class ReactionNodeNot : public ReactionNode {
  private:
    ReactionNode* operand;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeNot class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeNot(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operand = 0; fromSBML(n); }

      /**
       * ReactionNodeNot class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeNot(const ReactionNodeNot& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { operand = n.operand; }
    
      /**
       * ReactionNodeNot class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeNot()
      { if (operand) delete operand; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Put the ReactionNodeNot node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeNot& n)
      { return n.putTo(os); }
  };



    /********************************************
     * BOOLEAN OPERATIONS CLASSES: THE OR CLASS *
     ********************************************/

    /**
     * The ReactionNodeOr class.
     *
     * This class implements the n-ary logical "or" operator.  The
     * constructed expression has a truth value of true if at least
     * one of its arguments is true.
     */
  class ReactionNodeOr : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeOr class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    ReactionNodeOr(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeOr class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeOr(const ReactionNodeOr& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeOr class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeOr()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeOr node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeOr& n)
      { return n.putTo(os); }
 };



    /*********************************************
     * BOOLEAN OPERATIONS CLASSES: THE XOR CLASS *
     *********************************************/

    /**
     * The ReactionNodeXor class.
     *
     * This class implements the n-ary logical "xor" operator.  The
     * constructed expression has a truth value of true if an odd
     * number of its arguments are true.
     */
  class ReactionNodeXor : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;
    
  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
   std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeXor class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    ReactionNodeXor(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeXor class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    ReactionNodeXor(const ReactionNodeXor& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeXOr class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeXor()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeXor node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeXor& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * ALGEBRAIC OPERATIONS CLASSES: THE ADDITION CLASS *
     ****************************************************/

    /**
     * The ReactionNodeAddition class.
     *
     * This class implements the n-ary addition operator of an
     * algebraic structure.
     */
  class ReactionNodeAddition : public ReactionNode {
  private:
    ReactionNode* left;
    ReactionNode* right;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeAddition class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param l The left operand.
       * @param r The right operand.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeAddition(
      ReactionNode*         l,
      ReactionNode*         r,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { left = l; right = r; }

      /**
       * ReactionNodeAddition class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeAddition(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { left = 0; right = 0; fromSBML(n); }

      /**
       * ReactionNodeAddition class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeAddition(const ReactionNodeAddition& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { left = n.left; right = n.right; }

      /**
       * ReactionNodeAddition class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeAddition()
      { if (left) delete left; if (right) delete right; }

      /**
       * Linearises the node to an expression that is easier to
       * evaluate.
       *
       * Calling this method will attempt to bring the form of the
       * node to a form that is easier, and hopefully faster, to
       * evaluate.  If it was not possible to do this, the method
       * returns null.
       *
       * @return A pointer to the linearised node.
       */
    LinearElement* linearise();

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Get the left operand of the addition node.
       *
       * This method will return the left operand of the addition
       * node.
       *
       * @return The left operand.
       */
    inline ReactionNode* getLeftChild() const
      { return left; }

      /**
       * Get the right operand of the addition node.
       *
       * This method will return the right operand of the addition
       * node.
       *
       * @return The right operand.
       */
    inline ReactionNode* getRightChild() const
      { return right; }

      /**
       * Set the left operand of the addition node.
       *
       * This method will update the left operand of the addition
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setLeftChild(ReactionNode* child)
      { left = child; }

      /**
       * Set the right operand of the addition node.
       *
       * This method will update the right operand of the addition
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setRightChild(ReactionNode* child)
      { right = child; }

      /**
       * Put the ReactionNodeXor node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeAddition& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * ALGEBRAIC OPERATIONS CLASSES: THE DIVISION CLASS *
     ****************************************************/

    /**
     * The ReactionNodeDivision class.
     *
     * This class implements the binary MathML operator that is used
     * indicate the mathematical operation a "divided by" b.
     */
  class ReactionNodeDivision : public ReactionNode {
  private:
    ReactionNode* left;
    ReactionNode* right;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeDivision class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param l The left operand.
       * @param r The right operand.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeDivision(
      ReactionNode*         l,
      ReactionNode*         r,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { left = l; right = r; }

      /**
       * ReactionNodeDivision class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeDivision(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { left = 0; right = 0; fromSBML(n); }

      /**
       * ReactionNodeDivision class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeDivision(const ReactionNodeDivision& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { left = n.left; right = n.right; }

      /**
       * ReactionNodeDivision class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeDivision()
      { if (left) delete left; if (right) delete right; }

      /**
       * Linearises the node to an expression that is easier to
       * evaluate.
       *
       * Calling this method will attempt to bring the form of the
       * node to a form that is easier, and hopefully faster, to
       * evaluate.  If it was not possible to do this, the method
       * returns null.
       *
       * @return A pointer to the linearised node.
       */
    LinearElement* linearise();

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Get the left operand of the division node.
       *
       * This method will return the left operand of the division
       * node.
       *
       * @return The left operand.
       */
    inline ReactionNode* getLeftChild() const
      { return left; }

      /**
       * Get the right operand of the division node.
       *
       * This method will return the right operand of the division
       * node.
       *
       * @return The right operand.
       */
    inline ReactionNode* getRightChild() const
      { return right; }

      /**
       * Set the left operand of the division node.
       *
       * This method will update the left operand of the division
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setLeftChild(ReactionNode* child)
      { left = child; }

      /**
       * Set the right operand of the division node.
       *
       * This method will update the right operand of the division
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setRightChild(ReactionNode* child)
      { right = child; }

      /**
       * Put the ReactionNodeDivision node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeDivision& n)
      { return n.putTo(os); }
  };



    /**********************************************************
     * ALGEBRAIC OPERATIONS CLASSES: THE MULTIPLICATION CLASS *
     **********************************************************/

    /**
     * The ReactionNodeMultiplication class.
     *
     * This class implements the n-ary multiplication operator.
     */
  class ReactionNodeMultiplication : public ReactionNode {
  private:
    ReactionNode* left;
    ReactionNode* right;

  protected:
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeMultiplication class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param l The left operand.
       * @param r The right operand.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeMultiplication(
      ReactionNode*         l,
      ReactionNode*         r,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { left = l; right = r; }

      /**
       * ReactionNodeMultiplication class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    ReactionNodeMultiplication(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { left = 0; right = 0; fromSBML(n); }

      /**
       * ReactionNodeMultiplication class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    ReactionNodeMultiplication(const ReactionNodeMultiplication& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { left = n.left; right = n.right; }

      /**
       * ReactionNodeMultiplication class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    ~ReactionNodeMultiplication()
      { if (left) delete left; if (right) delete right; }

      /**
       * Linearises the node to an expression that is easier to
       * evaluate.
       *
       * Calling this method will attempt to bring the form of the
       * node to a form that is easier, and hopefully faster, to
       * evaluate.  If it was not possible to do this, the method
       * returns null.
       *
       * @return A pointer to the linearised node.
       */
    LinearElement* linearise();

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Get the left operand of the multiplication node.
       *
       * This method will return the left operand of the
       * multiplication node.
       *
       * @return The left operand.
       */
    inline ReactionNode* getLeftChild() const
      { return left; }

      /**
       * Get the right operand of the multiplication node.
       *
       * This method will return the right operand of the
       * multiplication node.
       *
       * @return The right operand.
       */
    inline ReactionNode* getRightChild() const
      { return right; }

      /**
       * Set the left operand of the multiplication node.
       *
       * This method will update the left operand of the
       * multiplication node.  The old operand should be released
       * prior to calling this method, otherwise we risk memory
       * leakage.
       *
       * @param child The new operand.
       */
    inline void setLeftChild(ReactionNode* child)
      { left = child; }

      /**
       * Set the right operand of the multiplication node.
       *
       * This method will update the right operand of the
       * multiplication node.  The old operand should be released
       * prior to calling this method, otherwise we risk memory
       * leakage.
       *
       * @param child The new operand.
       */
    inline void setRightChild(ReactionNode* child)
      { right = child; }

      /**
       * Put the ReactionNodeMultiplication node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeMultiplication& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * ALGEBRAIC OPERATIONS CLASSES: THE SUBTRACTION CLASS *
     *******************************************************/

    /**
     * The ReactionNodeSubtraction class.
     *
     * This class implements the subtraction operator for an additive
     * group.
     */
  class ReactionNodeSubtraction : public ReactionNode {
  private:
    ReactionNode* left;
    ReactionNode* right;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeSubtraction class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param l The left operand.
       * @param r The right operand.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSubtraction(
      ReactionNode*         l,
      ReactionNode*         r,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { left = l; right = r; }

      /**
       * ReactionNodeSubtraction class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSubtraction(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { left = 0; right = 0; fromSBML(n); }

      /**
       * ReactionNodeSubtraction class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeSubtraction(const ReactionNodeSubtraction& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { left  = n.left; right = n.right; }

      /**
       * ReactionNodeSubtraction class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeSubtraction()
      { if (left) delete left; if (right) delete right; }

      /**
       * Linearises the node to an expression that is easier to
       * evaluate.
       *
       * Calling this method will attempt to bring the form of the
       * node to a form that is easier, and hopefully faster, to
       * evaluate.  If it was not possible to do this, the method
       * returns null.
       *
       * @return A pointer to the linearised node.
       */
    LinearElement* linearise();

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Get the left operand of the subtraction node.
       *
       * This method will return the left operand of the subtraction
       * node.
       *
       * @return The left operand.
       */
    inline ReactionNode* getLeftChild() const
      { return left; }

      /**
       * Get the right operand of the subtraction node.
       *
       * This method will return the right operand of the subtraction
       * node.
       *
       * @return The right operand.
       */
    inline ReactionNode* getRightChild() const
      { return right; }

      /**
       * Set the left operand of the subtraction node.
       *
       * This method will update the left operand of the subtraction
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setLeftChild(ReactionNode* child)
      { left = child; }

      /**
       * Set the right operand of the subtraction node.
       *
       * This method will update the right operand of the subtraction
       * node.  The old operand should be released prior to calling
       * this method, otherwise we risk memory leakage.
       *
       * @param child The new operand.
       */
    inline void setRightChild(ReactionNode* child)
      { right = child; }

      /**
       * Put the ReactionNodeSubtraction node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
     inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeSubtraction& n)
      { return n.putTo(os); }
  };



    /*****************************************************
     * RELATIONAL OPERATIONS CLASSES: THE EQUALITY CLASS *
     *****************************************************/

    /**
     * The ReactionNodeEq class.
     *
     * This class implements the n-ary function that is used to
     * indicate that two or more quantities are equal.  There must be
     * at least two arguments.
     */
  class ReactionNodeEq : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeEq class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeEq(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeEq class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeEq(const ReactionNodeEq& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeEq class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeEq()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeEq node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeEq& n)
      { return n.putTo(os); }
  };



    /****************************************************************
     * RELATIONAL OPERATIONS CLASSES: THE GREATER-OR-EQUAL-TO CLASS *
     ****************************************************************/

    /**
     * The ReactionNodeGeq class.
     *
     * This class implements the n-ary greater than or equal to
     * function, which returns true if each argument in turn is
     * greater than or equal to the one following it.  There must be
     * at least two arguments.
     */
  class ReactionNodeGeq : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeGeq class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeGeq(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeGeq class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeGeq(const ReactionNodeGeq& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeGeq class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeGeq()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeGeq node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeGeq& n)
      { return n.putTo(os); }
  };



    /*********************************************************
     * RELATIONAL OPERATIONS CLASSES: THE GREATER-THAN CLASS *
     *********************************************************/

    /**
     * The ReactionNodeGt class.
     *
     * One may be tempted to believe that this class will serve a
     * Chappist's Gin & Tonic, with two cubes of ice and one slice of
     * lemon.  Sadly, it is not so.  This n-ary function merely
     * represents the relation "greater than" which returns true if
     * each argument in turn is greater than the one following it.
     * There must be at least two arguments.
     */
  class ReactionNodeGt : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeGt class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeGt(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeGt class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeGt(const ReactionNodeGt& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeGt class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeGt()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeGt node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeGt& n)
      { return n.putTo(os); }
  };



    /*************************************************************
     * RELATIONAL OPERATIONS CLASSES: THE LESS-OR-EQUAL-TO CLASS *
     *************************************************************/

    /**
     * The ReactionNodeLeq class.
     *
     * This class implements the n-ary function that represents the
     * relation "less than or equal to" which returns true if each
     * argument in turn is less or equal to the one following it.
     * There must be at least two arguments.
     */
  class ReactionNodeLeq : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeLeq class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeLeq(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeLeq class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeLeq(const ReactionNodeLeq& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }

      /**
       * ReactionNodeLeq class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeLeq()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeLeq node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeLeq& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * RELATIONAL OPERATIONS CLASSES: THE LESS-THAN CLASS *
     ******************************************************/

    /**
     * The ReactionNodeLt class.
     *
     * This class implements the n-ary function that represents the
     * relation "less than" which returns true if each argument in
     * turn is less than the one following it.  There must be at least
     * two arguments.
     */
  class ReactionNodeLt : public ReactionNode {
  private:
    std::list<ReactionNode*> operands;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeLt class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeLt(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { operands.clear(); fromSBML(n); }

      /**
       * ReactionNodeLt class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeLt(const ReactionNodeLt& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { for (std::list<ReactionNode*>::const_iterator i = n.operands.begin();
             i != n.operands.end(); i++) operands.push_back(*i); }
    
      /**
       * ReactionNodeLt class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeLt()
      { operands.clear(); }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeLt node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeLt& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * RELATIONAL OPERATIONS CLASSES: THE INEQUALITY CLASS *
     *******************************************************/

    /**
     * The ReactionNodeNeq class.
     *
     * This class implements the binary function that represents the
     * relation "not equal to" which returns true unless the two
     * arguments are equal.
     */
  class ReactionNodeNeq : public ReactionNode {
  private:
    ReactionNode* left;
    ReactionNode* right;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeNeq class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeNeq(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { left = 0; right = 0; fromSBML(n); }

      /**
       * ReactionNodeNeq class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeNeq(const ReactionNodeNeq& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { left = n.left; right = n.right; }

      /**
       * ReactionNodeNeq class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeNeq()
      { if (left) delete left; if (right) delete right; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeNeq node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeNeq& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCOS CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArccos class.
     *
     * This class implements the arccos function which is the inverse
     * of the cos function.  It takes one argument.
     */
  class ReactionNodeArccos : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccos class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccos(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccos class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccos(const ReactionNodeArccos& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArccos class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccos()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccos node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccos& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCOSH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArccosh class.
     *
     * This class implements the arccosh function.  It takes one
     * argument.
     */
  class ReactionNodeArccosh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccosh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccosh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccosh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccosh(const ReactionNodeArccosh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArccosh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccosh()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccosh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccosh& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCOT CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArccot class.
     *
     * This class implements the arccot function.  It takes one
     * argument.
     */
  class ReactionNodeArccot : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccot class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccot(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccot class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccot(const ReactionNodeArccot& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

       /**
       * ReactionNodeArccot class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccot()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccot node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccot& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCOTH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArccoth class.
     *
     * This class implements the arccoth function.  It takes one
     * argument.
     */
  class ReactionNodeArccoth : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccoth class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccoth(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccoth class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccoth(const ReactionNodeArccoth& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArccoth class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccoth()
      { if (argument) delete argument; }
 
      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccoth node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccoth& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCSC CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArccsc class.
     *
     * This class implements the arccsc function.  It takes one
     * argument.
     */
  class ReactionNodeArccsc : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccsc class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccsc(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccsc class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccsc(const ReactionNodeArccsc& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArccsc class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccsc()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccsc node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccsc& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCCSCH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArccsch class.
     *
     * This class implements the arccsch function.  It takes one
     * argument.
     */
  class ReactionNodeArccsch : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArccsch class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArccsch(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArccsch class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArccsch(const ReactionNodeArccsch& n)
    : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArccsch class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArccsch()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArccsch node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArccsch& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCSEC CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArcsec class.
     *
     * This class implements the arcsec function.  It takes one
     * argument.
     */
  class ReactionNodeArcsec : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArcsec class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArcsec(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArcsec class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArcsec(const ReactionNodeArcsec& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArcsec class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArcsec()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArcsec node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArcsec& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCSECH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArcsech class.
     *
     * This class implements the arcsech function.  It takes one
     * argument.
     */
  class ReactionNodeArcsech : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:
 
      /**
       * ReactionNodeArcsech class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
   inline ReactionNodeArcsech(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArcsech class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArcsech(const ReactionNodeArcsech& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArcsech class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline~ReactionNodeArcsech()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArcsech node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArcsech& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCSIN CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArcsin class.
     *
     * This class implements the arcsin function which is the inverse
     * of the sin function.  It takes one argument.
     */
  class ReactionNodeArcsin : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArcsin class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArcsin(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArcsin class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArcsin(const ReactionNodeArcsin& n) 
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArcsin class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArcsin()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArcsin node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArcsin& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCSINH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArcsinh class.
     *
     * This class implements the arcsinh function.  It takes one
     * argument.
     */
  class ReactionNodeArcsinh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArcsinh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArcsinh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArcsinh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArcsinh(const ReactionNodeArcsinh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArcsinh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArcsinh()
      { if (argument) delete argument; }
 
      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArcsinh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArcsinh& n)
      { return n.putTo(os); }
  };



    /******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCTAN CLASS *
     ******************************************************/

    /**
     * The ReactionNodeArctan class.
     *
     * This class implements the arctan function which is the inverse
     * of the tan function.  It takes one argument.
     */
  class ReactionNodeArctan : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArctan class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArctan(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArctan class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArctan(const ReactionNodeArctan& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArctan class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArctan()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArctan node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArctan& n)
      { return n.putTo(os); }
  };



    /*******************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE ARCTANH CLASS *
     *******************************************************/

    /**
     * The ReactionNodeArctanh class.
     *
     * This class implements the arctanh function.  It takes one
     * argument.
     */
  class ReactionNodeArctanh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeArctanh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeArctanh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeArctanh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeArctanh(const ReactionNodeArctanh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeArctanh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeArctanh()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeArctanh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeArctanh& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE COS CLASS *
     ***************************************************/

    /**
     * The ReactionNodeCos class.
     *
     * This class implements the cos function.  It takes one argument.
     */
  class ReactionNodeCos : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCos class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeCos(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCos class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCos(const ReactionNodeCos& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeCos class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCos()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCos node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCos& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE COSH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeCosh class.
     *
     * This class implements the cosh function.  It takes one
     * argument.
     */
  class ReactionNodeCosh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCosh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeCosh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCosh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCosh(const ReactionNodeCosh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeCosh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCosh()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCosh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCosh& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE COT CLASS *
     ***************************************************/

    /**
     * The ReactionNodeCot class.
     *
     * This class implements the cot function.  It takes one argument.
     */
  class ReactionNodeCot : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCot class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeCot(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCot class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCot(const ReactionNodeCot& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

     /**
       * ReactionNodeCot class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCot()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCot node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCot& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE COTH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeCoth class.
     *
     * This class implements the coth function.  It takes one
     * argument.
     */
  class ReactionNodeCoth : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCoth class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
     inline ReactionNodeCoth(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCoth class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCoth(const ReactionNodeCoth& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

     /**
       * ReactionNodeCoth class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCoth()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCoth node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCoth& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE CSC CLASS *
     ***************************************************/

    /**
     * The ReactionNodeCsc class.
     *
     * This class implements the csc function.  It takes one argument.
     */
  class ReactionNodeCsc : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCsc class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeCsc(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCsc class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCsc(const ReactionNodeCsc& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

     /**
       * ReactionNodeCsc class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCsc()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCsc node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCsc& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE CSCH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeCsch class.
     *
     * This class implements the csch function.  It takes one
     * argument.
     */
  class ReactionNodeCsch : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeCsch class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
     inline ReactionNodeCsch(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeCsch class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeCsch(const ReactionNodeCsch& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

     /**
       * ReactionNodeCsch class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeCsch()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeCsch node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeCsch& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE SEC CLASS *
     ***************************************************/

    /**
     * The ReactionNodeSec class.
     *
     * This class implements the sec function.  It takes one argument.
     */
  class ReactionNodeSec : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
   std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeSec class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSec(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeSec class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeSec(const ReactionNodeSec& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeSec class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
   inline ~ReactionNodeSec()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeSec node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeSec& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE SECH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeSech class.
     *
     * This class implements the sech function.  It takes one
     * argument.
     */
  class ReactionNodeSech : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeSech class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSech(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeSech class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeSech(const ReactionNodeSech& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeSech class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeSech()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeSech node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeSech& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE SIN CLASS *
     ***************************************************/

    /**
     * The ReactionNodeSin class.
     *
     * This class implements the sin function.  It takes one argument.
     */
  class ReactionNodeSin : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeSin class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSin(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeSin class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeSin(const ReactionNodeSin& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeSin class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeSin()
      { if (argument) delete argument; }
 
      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeSin node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeSin& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE SINH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeSinh class.
     *
     * This class implements the sinh function.  It takes one
     * argument.
     */
  class ReactionNodeSinh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeSinh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeSinh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeSinh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeSinh(const ReactionNodeSinh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeSinh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeSinh()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;


      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;


      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeSinh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeSinh& n)
      { return n.putTo(os); }
  };



    /***************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE TAN CLASS *
     ***************************************************/

    /**
     * The ReactionNodeTan class.
     *
     * This class implements the tan function.  It takes one argument.
     */
  class ReactionNodeTan : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeTan class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeTan(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeTan class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeTan(const ReactionNodeTan& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeTan class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeTan()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeTan node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeTan& n)
      { return n.putTo(os); }
  };



    /****************************************************
     * TRIGONOMETRIC OPERATIONS CLASSES: THE TANH CLASS *
     ****************************************************/

    /**
     * The ReactionNodeTanh class.
     *
     * This class implements the tanh function.  It takes one
     * argument.
     */
  class ReactionNodeTanh : public ReactionNode {
  private:
    ReactionNode* argument;

  protected:
 
      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeTanh class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeTanh(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { argument = 0; fromSBML(n); }

      /**
       * ReactionNodeTanh class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeTanh(const ReactionNodeTanh& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { argument = n.argument; }

      /**
       * ReactionNodeTanh class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeTanh()
      { if (argument) delete argument; }

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    ReactionNode* fold() throw( InvalidUnit );

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeTanh node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeTanh& n)
      { return n.putTo(os); }
  };



    /******************************************
     * VALUE CLASSES: THE CONCENTRATION CLASS *
     ******************************************/

    /**
     * The ReactionNodeConcentration class.
     *
     * This class implements species concentrations.  On construction,
     * the index in the species vector must be supplied.  Species
     * concentrations are looked up in the supplied species vector
     * using this index every time the node is to be evaluated.
     */
  class ReactionNodeConcentration : public ReactionNode {
  private:
    off_t index;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:
      /**
       * ReactionNodeConcentration class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeConcentration(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { index = -1; fromSBML(n); }

      /**
       * ReactionNodeConcentration class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeConcentration(const ReactionNodeConcentration& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { index = n.index; }

      /**
       * ReactionNodeConcentration class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeConcentration()
      {}
    
      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n);

      /**
       * Get the index of the species in the species vector.
       *
       * This method will return the index of the species in the
       * species vector.  Note that internally, all references to
       * species are by indices into the species vector.  This index
       * is identical to the index in the species change vector for
       * the same species.
       *
       * @return The index of the species.
       */
    inline off_t getIndex() const
      { return index; }

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeConcentration node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeConcentration& n)
      { return n.putTo(os); }
  };



    /*************************************
     * VALUE CLASSES: THE CONSTANT CLASS *
     *************************************/

    /**
     * The ReactionNodeConstant class.
     *
     * Not only does this class handle constants in the pure numerical
     * sense, such as 42 and pi.  It also takes care of parameters
     * that are defined as constant.
     */
  class ReactionNodeConstant : public ReactionNode {
  private:
    double value;
    Unit   unit;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeConstant class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param v The constant value.
       * @param u The unit of the constant.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeConstant(
      double                v,
      Unit                  u,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { value = v; unit = u; }

      /**
       * ReactionNodeConstant class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeConstant(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { value = 0.0; unit = Unit::dimensionless(); fromSBML(n); }

      /**
       * ReactionNodeConstant class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeConstant(const ReactionNodeConstant& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { value = n.value; unit = n.unit; }

      /**
       * ReactionNodeConstant class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeConstant()
      {}

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n) throw(ModelError);

      /**
       * Get the value of a constant node.
       *
       * This method will return the value of a constant node.
       *
       * @return The value of the constant node.
       */
    inline double getValue() const
      { return value; }

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Set the value of a constant node.
       *
       * This method will update the value of a constant node.
       *
       * @param v The new value for the constant node.
       */
    inline void setValue(double v)
      { value = v; }
    
      /**
       * Put the ReactionNodeConstant node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeConstant& n)
      { return n.putTo(os); }
  };



    /*************************************
     * VALUE CLASSES: THE FUNCTION CLASS *
     *************************************/

    /**
     * The ReactionNodeFunction class.
     *
     * This class is supposed to implement general functions.  Sadly,
     * a bit of work remains until it can do so.
     */
  class ReactionNodeFunction : public ReactionNode {
  private:

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    inline std::ostream& putTo(std::ostream& os) const 
      { os << "FUNCTION not yet implemented." << std::endl; return os; }

  public:

      /**
       * ReactionNodeFunction class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeFunction(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(n, p, s)
      { fromSBML(n); }

      /**
       * ReactionNodeFunction class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeFunction(const ReactionNodeFunction& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      {}

      /**
       * ReactionNodeFunction class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeFunction()
      {}

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Do constant folding on the expression beneath the node.
       *
       * Calling this method will attempt to do constant folding on
       * the expression beneath the node.  Note that this may change
       * the root node on which folding is performed.  In order to be
       * certain the node is really folded, one should call the
       * refold() method of the ReactionNode class.
       *
       * @return The root node of the constant folded node.
       */
    inline ReactionNode* fold()
      { return this; }

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    inline ReactionNode* fromSBML(const ASTNode_t*)
      { return 0; }

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    inline Unit getUnit() const throw(InvalidUnit)
      { return Unit::dimensionless(); }
        
      /**
       * Put the ReactionNodeFunction node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeFunction& n)
      { return n.putTo(os); }
  };



    /**************************************
     * VALUE CLASSES: THE PARAMETER CLASS *
     **************************************/

    /**
     * The ReactionNodeParameter class.
     *
     * Note that this class does not handle constant parameters.
     * These are handled by the constant class instead.
     */
  class ReactionNodeParameter : public ReactionNode {
  private:
    std::string id;

  protected:

      /**
       * Puts the node in XML format onto an output stream.
       *
       * This method accomplishes the exact same thing as C++'s "put
       * to" operator (<<).
       *
       * @param os The output stream to write to.
       * @return   The output stream once written to.
       */
    std::ostream& putTo(std::ostream& os) const;

  public:

      /**
       * ReactionNodeParameter class value constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param i The parameter id.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeParameter(
      std::string&          i,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0) : ReactionNode(0, p, s)
      { id = i; }

      /**
       * ReactionNodeParameter class default constructor.
       *
       * Initialises the argument and calls fromSBML() to read the data
       * for the node.
       *
       * @param n The SBML AST root node.
       * @param p The parameter table.
       * @param s The species table.
       */
    inline ReactionNodeParameter(
      const ASTNode_t*      n = 0,
      const ParameterTable* p = 0,
      const SpeciesTable*   s = 0)  : ReactionNode(n, p, s)
      { id = ""; fromSBML(n); }

      /**
       * ReactionNodeParameter class copy constructor.
       *
       * Initialises the class members from the copy.  There is no
       * need to call fromSBML() here.
       *
       * @param n The node to copy data from.
       */
    inline ReactionNodeParameter(const ReactionNodeParameter& n)
      : ReactionNode(0, n.parameterTable, n.speciesTable)
      { id = n.id; }

      /**
       * ReactionNodeParameter class destructor.
       *
       * Deletes the argument.  This way the entire tree below this
       * node is freed recursively.
       */
    inline ~ReactionNodeParameter()
      {}

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<unsigned int>& species) const;

      /**
       * Evaluate the node to given a species composition.
       *
       * Calling this method will evaluate the expression beneath the
       * node, given the supplied species configuration.
       *
       * @param species The current species vector
       * @return        The value of the node, given the configuration.
       */
    double evaluate(const std::valarray<double>& species) const;

      /**
       * Read and parse the mathematics from an SBML AST.
       *
       * This method will attempt to read and parse the information
       * contained in an SBML AST.
       *
       * @param n The SBML AST root node.
       */
    ReactionNode* fromSBML(const ASTNode_t* n) 
      throw(ModelError, UndefinedIdentifier);

      /**
       * Get the unit of the expression beneath the node.
       *
       * Calling this method will evaluate the expression beneath the
       * node in terms of units.  If the unit could not be determined,
       * an exception is thrown.
       *
       * @return The unit of the node.
       */
    Unit getUnit() const throw(InvalidUnit);

      /**
       * Put the ReactionNodeParameter node to an output stream.
       *
       * This method will put the MathML for the node onto an output
       * stream.
       *
       * @param os The output stream on which to write.
       * @param n  The node in question.
       * @return   The output stream, once written to.
       */
    inline friend std::ostream& operator<< (
      std::ostream& os, const ReactionNodeParameter& n)
      { return n.putTo(os); }
  };
}

#endif
